/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*- */
/*
 * Chroma widget set
 * (C) 2005 M. Derezynski <internalerror@gmail.com>
 *
 */

#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h>
#include <gdk/gdkx.h>
#include <X11/Xlib.h>

#include <libchroma/chroma-list.h>
#include <libchroma/chroma-list-column.h>
#include <libchroma/chroma-list-selection.h>
#include <libchroma/chroma-vscrollbar.h>
#include <libchroma/dnd.h>

#include "draw.h"
#include "libchroma-marshalers.h"

#define NONE (-1)

static void chroma_list_class_init     (ChromaListClass *klass);

static void chroma_list_init           (ChromaList      *widget);

static void chroma_list_realize        (GtkWidget       *widget);

static void chroma_list_size_allocate  (GtkWidget       *widget,
				        GtkAllocation   *allocation);

static void chroma_list_send_configure (ChromaList      *widget);

static void
chroma_list_drag_leave	    (GtkWidget		  *widget,
			     GdkDragContext	  *drag_context,
			     guint		   time_); 
 
static gboolean
chroma_list_drag_motion	    (GtkWidget		  *widget,
			     GdkDragContext	  *drag_context,
			     gint		   x,
			     gint		   y,
			     guint		   time_);

static gboolean
chroma_list_button_press    (GtkWidget		  *widget,
			     GdkEventButton	  *event);

static gboolean
chroma_list_button_release  (GtkWidget		  *widget,
			     GdkEventButton	  *event);

static gboolean
chroma_list_key_press	    (GtkWidget		  *widget,
			     GdkEventKey	  *event);


static gboolean
chroma_list_scroll	    (GtkWidget		  *widget,
			     GdkEventScroll	  *event);

static gboolean
chroma_list_motion	    (GtkWidget	          *widget,
			     GdkEventMotion       *event);

typedef enum {
    CHROMA_LIST_SIGNAL_ROW_ACTIVATED,
    CHROMA_LIST_SIGNAL_CURSOR_POSITION,
    CHROMA_LIST_SIGNAL_DND,

    CHROMA_LIST_N_SIGNALS
} ChromaListSignals;
static guint signals[CHROMA_LIST_N_SIGNALS] = {0};

typedef enum {
    CHROMA_LIST_PROP_OFFSET = 1,
    CHROMA_LIST_PROP_N_ITEMS,
    CHROMA_LIST_PROP_ROW_HEIGHT,
    CHROMA_LIST_PROP_RESERVED_SPACE,
    CHROMA_LIST_PROP_SHOW_HEADERS,

    CHROMA_LIST_N_PROPERTIES
} ChromaListProperties;

enum {
    SIGNAL_ROW_CHANGED,
    SIGNAL_ROW_DELETED,
    SIGNAL_ROW_INSERTED,
    SIGNAL_ROWS_REORDERED,

    SIGNAL_SORT_COLUMN_CHANGED, //GtkTreeSortable
    
    N_MODEL_SIGNALS
};

enum {
    SIGNAL_POSITION,
    SIGNAL_STEP_UP,
    SIGNAL_STEP_DOWN,
    
    N_ADJUSTMENT_SIGNALS
};

typedef enum {
    SCROLL_DIRECTION_UP,
    SCROLL_DIRECTION_DOWN,
} ScrollDirection;

struct _ChromaListPrivate {
    gboolean		       dispose_has_run;

    GtkTreeModel	      *model;
    gulong		       model_signals[N_MODEL_SIGNALS];
    ChromaListRowSwapFunc      model_row_swap_func;
    

    ChromaVScrollbar	      *adjustment;
    gulong		       adjustment_signals[N_ADJUSTMENT_SIGNALS];

    GList		      *selected_rows;
    GtkTreeRowReference	      *last_pressed_row;

    GtkSortType		       sort_type;
    gint		       sort_column_id;
    gint		       sort_trigger;

    ChromaListColumn	      *number_column;

    ChromaListCellDataFunc    *cell_data_funcs;
    gpointer		      *cell_data_funcs_data;

    GPtrArray		      *columns;
    gdouble		      *column_widths;
    gdouble		      *column_xalign;
    gdouble		      *column_yalign;
    gboolean		      *column_show;
    gint		       n_columns;
    gint		       n_columns_visible;
    gdouble		      *pixel_widths;

    gint		       header_size;
    gint		       row_height;
    gint		       cellspacing;
    gint		       offset;

    guint		       all_columns_mask;

    gboolean		       active_row_drag;
    gboolean		       active_column_resize;
    gboolean		       button_press;

    gint		       grabbed_position;

    gint		       adjactent_l;
    gint		       adjactent_r;
    gdouble		       resize_position;

    guint		       scroll_timeout_source_id;
    ScrollDirection	       scroll_direction; 

    /* Parameters */
    GtkSelectionMode	       selection_mode;
    ChromaListDropMode	       drop_mode;
    gboolean		       use_numbers;
    gint		       reserved_space;
    gboolean		       show_headers;

    gboolean		       drop;
    gint		       drop_row;

    /* Drawing stuff */
    GdkPixmap		      *pm_backingstore;
    cairo_surface_t	      *cs_selection;

    PangoFontMetrics	      *metrics;
};

#define COLUMN_MINIMAL_WIDTH 50.0

G_LOCK_DEFINE(PROCESS_CELLS);
G_LOCK_DEFINE(SCROLL_OFFSET);

/* Rendering Function Prototypes */
static void
cl_process_cells	      (ChromaList *self, gint *rows, guint columns); 

static void
cl_recreate_pm_backingstore   (ChromaList *self);

static void
cl_recreate_cs_selection      (ChromaList *self);

static guint
gint_count_digits (gint n)
{
    guint count = 0;

    n = ABS (n);
    do {
	count++;
	n /= 10;
    } while (n > 0);

    return count;
}

static gdouble 
pixel_width_to_percent (ChromaList	*self, 
		        gdouble		 pixel_width)
{
    gdouble  total_width;
    gdouble  width;

    total_width = GTK_WIDGET (self)->allocation.width - 
		  (self->priv->n_columns_visible*self->priv->cellspacing) -
		  self->priv->reserved_space;

    width = pixel_width / total_width; 

    return width;
}

static void 
cl_recalc_pixel_widths (ChromaList *self)
{
    gdouble   *pixel_widths;
    gdouble    total_width;
    gint       n;

    if (self->priv->pixel_widths) g_free (self->priv->pixel_widths);

    total_width  = GTK_WIDGET (self)->allocation.width - 
		   (self->priv->n_columns_visible*self->priv->cellspacing) -
		   self->priv->reserved_space;

    pixel_widths = g_new0 (gdouble, self->priv->n_columns);

    for (n = 0; n < self->priv->n_columns; n++) 
      {
	if (self->priv->column_show[n])
	    pixel_widths[n] = total_width * self->priv->column_widths[n];  
	else
	    pixel_widths[n] = 0; 
      }

    self->priv->pixel_widths = pixel_widths; 
}

void
cl_recalc_reserved_space (ChromaList *self)
{

  if (self->priv->use_numbers)
      self->priv->reserved_space = (pango_font_metrics_get_approximate_digit_width (self->priv->metrics)/PANGO_SCALE)*
				      (gint_count_digits(gtk_tree_model_iter_n_children(self->priv->model,NULL))+3);
  else
      self->priv->reserved_space = self->priv->cellspacing;
}

static void
cl_initialize_metrics (ChromaList *self)
{
    self->priv->metrics = pango_context_get_metrics (gtk_widget_get_pango_context (GTK_WIDGET (self)),
				        GTK_WIDGET (self)->style->font_desc, 
				        pango_context_get_language(gtk_widget_get_pango_context(GTK_WIDGET (self))));

    self->priv->row_height  = (pango_font_metrics_get_ascent  (self->priv->metrics)/PANGO_SCALE) + 
				(pango_font_metrics_get_descent (self->priv->metrics)/PANGO_SCALE) + 4;

    if (self->priv->show_headers)
      self->priv->header_size = self->priv->row_height + 7;
    else
      self->priv->header_size = 0; 
}

/* DnD */
static void
chroma_list_drag_end (GtkWidget	      *widget,
		      GdkDragContext  *context)

{
      CHROMA_LIST (widget)->priv->drop = FALSE;
      gtk_widget_queue_draw (widget);
}

static void
chroma_list_drag_leave (GtkWidget      *widget,
			GdkDragContext *context,
			guint	        time)
		
{
      CHROMA_LIST (widget)->priv->drop = FALSE;
      gtk_widget_queue_draw (widget);
}


static void
chroma_list_drag_data_received (GtkWidget	  *widget,
				GdkDragContext	  *context,
				gint		   x,
				gint		   y,
				GtkSelectionData   *selection_data,
				guint		   info,
				guint		   time) 
			
{
    gchar	   **uri_strv;
    gchar	    *uri_string;

    g_return_if_fail(selection_data != NULL);

    if (!selection_data->data)
      {
  	  gtk_drag_finish (context, FALSE, FALSE, time);
          return;
      }

    uri_string = g_strdup((gchar*)selection_data->data);
    g_strstrip(uri_string);
    uri_strv = g_uri_list_extract_uris (uri_string);
    gtk_drag_finish (context, TRUE, FALSE, time);

    g_signal_emit (G_OBJECT(widget), 
		   signals[CHROMA_LIST_SIGNAL_DND], 
		   0, 
		   CHROMA_LIST(widget)->priv->drop_row + CHROMA_LIST(widget)->priv->offset,
		   uri_strv);

}

static gboolean
chroma_list_drag_motion (GtkWidget	  *widget,
			 GdkDragContext	  *context,
			 gint		   x,
			 gint		   y,
			 guint		   time)


{
    GdkAtom target;
    gint    position;

    target = gtk_drag_dest_find_target (widget, context, NULL);
    if (target == GDK_NONE)
      {
	gdk_drag_status (context, 0, time);
	return FALSE;
      }

    CHROMA_LIST(widget)->priv->drop = TRUE;

    position = (y - CHROMA_LIST(widget)->priv->header_size) / CHROMA_LIST(widget)->priv->row_height;

    if (gtk_tree_model_iter_n_children (CHROMA_LIST(widget)->priv->model, NULL))
	{
	    if ((position+CHROMA_LIST(widget)->priv->offset) > chroma_list_get_length (CHROMA_LIST(widget)))
		 position = (chroma_list_get_length (CHROMA_LIST(widget)) - CHROMA_LIST(widget)->priv->offset);
	    CHROMA_LIST(widget)->priv->drop_row = position; 
	}
    else
	{
	    CHROMA_LIST(widget)->priv->drop_row = 0; 
	}

    gtk_widget_queue_draw (widget);

    return TRUE;
}

/* END DnD */

void
chroma_list_refresh (ChromaList *self)
{
    g_return_if_fail(CHROMA_IS_LIST(self));

    if (chroma_list_get_n_item_display (self) <= 0) return; 

    cl_recreate_pm_backingstore  (self);
    cl_recreate_cs_selection     (self);

    cl_recalc_reserved_space	 (self);
    cl_recalc_pixel_widths	 (self);
    cl_process_cells		 (self, NULL, self->priv->all_columns_mask);

    gtk_widget_queue_draw        (GTK_WIDGET (self));
}

void
chroma_list_set_font (ChromaList *self, PangoFontDescription *font_desc)
{
    g_return_if_fail(CHROMA_IS_LIST(self));

    gtk_widget_modify_font (GTK_WIDGET (self), font_desc); 
    cl_initialize_metrics (self);
    chroma_list_refresh (self);
}

void
chroma_list_set_selection_mode (ChromaList * self,
                                GtkSelectionMode selection_mode)
{
    g_return_if_fail(CHROMA_IS_LIST(self));

    if ((selection_mode != GTK_SELECTION_MULTIPLE) &&
        (selection_mode != GTK_SELECTION_SINGLE))
        return;

    self->priv->selection_mode = selection_mode;
}

void
chroma_list_set_cell_data_func (ChromaList		    *self,
				gint			     column, 
				ChromaListCellDataFunc	     cell_data_func, 
				gpointer		     data)
{
    g_return_if_fail (CHROMA_IS_LIST(self));
    g_return_if_fail (column < self->priv->n_columns);

    self->priv->cell_data_funcs[column] = cell_data_func;
    self->priv->cell_data_funcs_data[column] = data;
}

void
chroma_list_set_column_visible (ChromaList	  *self,
			        gint		   column,
			        gboolean	   visible) 
{
    g_return_if_fail (CHROMA_IS_LIST(self));
    g_return_if_fail (column < self->priv->n_columns);

    if (self->priv->column_show[column] != visible)
      {
	  self->priv->column_show[column] = visible; 
	
	  if (visible)
	    self->priv->n_columns_visible++;
	  else
	    self->priv->n_columns_visible--;
      }
}

void
chroma_list_set_column_header (ChromaList	  *self,
			       gint		   column,
			       const gchar	  *caption) 
{
    ChromaListColumn *list_column;

    g_return_if_fail (CHROMA_IS_LIST(self));
    g_return_if_fail (column < self->priv->n_columns);

    list_column = g_ptr_array_index (self->priv->columns, column);
    
    g_object_set (G_OBJECT(list_column), "name", caption, NULL);
}

void
chroma_list_set_column_show_header (ChromaList		*self, 
				    gint		 column, 
				    gboolean		 show) 
{
    ChromaListColumn *list_column;

    g_return_if_fail (CHROMA_IS_LIST(self));

    list_column = g_ptr_array_index (self->priv->columns, column);
    g_object_set (G_OBJECT(list_column), "show_name", show, NULL);
}

void
chroma_list_set_column_width (ChromaList	*self, 
			      gint		 column, 
			      gdouble		 width)
{
    g_return_if_fail (CHROMA_IS_LIST(self));

    self->priv->column_widths[column] = width; 
}

gdouble
chroma_list_get_column_width (ChromaList	*self, 
			      gint		 column)
			    
{
    g_return_val_if_fail (CHROMA_IS_LIST(self), 0.0);

    return self->priv->column_widths[column];
}

ChromaListColumn*
chroma_list_get_column (ChromaList *self, gint _idx)
{
    g_return_val_if_fail (CHROMA_IS_LIST(self), NULL);

    return g_ptr_array_index (self->priv->columns, _idx);
}

void
chroma_list_set_column_xalign (ChromaList *self, gint _idx, gdouble xalign)
{
    g_return_if_fail (CHROMA_IS_LIST(self));

    self->priv->column_xalign[_idx] = xalign;
    g_object_set (G_OBJECT(g_ptr_array_index (self->priv->columns, _idx)), "xalign", xalign, NULL);
}

void
chroma_list_set_column_yalign (ChromaList *self, gint _idx, gdouble yalign)
{
    g_return_if_fail (CHROMA_IS_LIST(self));

    self->priv->column_yalign[_idx] = yalign;
    g_object_set (G_OBJECT(g_ptr_array_index (self->priv->columns, _idx)), "yalign", yalign, NULL);
}

static void
cl_render_row (ChromaList	    *self,
	       GtkTreeIter	    *iter,
	       gint		     k,
	       guint		     r_columns)
{
    ChromaListColumn    *column;
    GValue	         value_o = {0,};
    GValue	         value = {0,};
    gint	         n, 
		         render_position;
    gint		 x, y, w, h;
    gboolean	         selected = FALSE, current = FALSE, clear = TRUE;
    cairo_t		*cr;
    double		 r, g, b, a;
    GtkTreeRowReference *reference;
    GtkTreePath		*path_c,
			*path_r;

    g_return_if_fail (CHROMA_IS_LIST(self));

    //We start rendering the real columns after the the reserved space for the list index numbers
    render_position = self->priv->reserved_space;

    //Check for row being selected or current
    reference = g_object_get_data (G_OBJECT(self->priv->model), "current");
    if (reference)
      {
	path_r = gtk_tree_model_get_path (GTK_TREE_MODEL(self->priv->model), iter);
	path_c = gtk_tree_row_reference_get_path (reference);

	if (path_r && path_c && (!gtk_tree_path_compare (path_r, path_c))) current = TRUE;
  
	if (path_r) gtk_tree_path_free (path_r);
	if (path_c) gtk_tree_path_free (path_c);
      }

    if (chroma_list_selection_iter_get_selected (CHROMA_LIST_SELECTION (self), iter)) selected = TRUE;

    //Create local cairo context
    cr = gdk_cairo_create (CHROMA_LIST(self)->priv->pm_backingstore);

    //Check for general row clear
    if (r_columns == self->priv->all_columns_mask)
      {
	    cairo_rectangle (cr, 0, 
			     k * self->priv->row_height + self->priv->header_size,
			     GTK_WIDGET(self)->allocation.width, 
			     self->priv->row_height);
	    chroma_gdk_color_to_rgba (&(GTK_WIDGET (self)->style->bg[GTK_STATE_NORMAL]), &r, &g, &b, &a);
	    cairo_set_source_rgba (cr, r, g, b, 1.0);
	    cairo_fill (cr);
	    clear = FALSE;
      }

    if (self->priv->use_numbers)
      {
	g_value_init (&value, G_TYPE_INT);  
	g_value_set_int (&value, k+1+self->priv->offset);

	if (selected || current)
	  {
	      g_object_set (G_OBJECT(self->priv->number_column), "foreground", &(GTK_WIDGET (self)->style->fg[GTK_STATE_PRELIGHT]), NULL);
	      g_object_set (G_OBJECT(self->priv->number_column), "background", &(GTK_WIDGET (self)->style->bg[GTK_STATE_NORMAL]), NULL);
	  }
	else
	  {
	      g_object_set (G_OBJECT(self->priv->number_column), "foreground", &(GTK_WIDGET (self)->style->fg[GTK_STATE_NORMAL]), NULL);
	      g_object_set (G_OBJECT(self->priv->number_column), "background", &(GTK_WIDGET (self)->style->bg[GTK_STATE_NORMAL]), NULL);
	  }

	g_object_set (self->priv->number_column, "font", GTK_WIDGET (self)->style->font_desc, NULL);
	g_object_set (self->priv->number_column, "row_height", self->priv->row_height, NULL);

	g_object_set (G_OBJECT(self->priv->number_column), "xalign", 1.0, NULL);
	g_object_set (G_OBJECT(self->priv->number_column), "yalign", 0.5, NULL);

	chroma_list_column_cell_render (self->priv->number_column,
					GTK_WIDGET (self),
					&value,	
					CHROMA_LIST(self)->priv->pm_backingstore,
					self->priv->reserved_space,
					0, k*self->priv->row_height + self->priv->header_size);

	if ((selected) && !(r_columns == self->priv->all_columns_mask))
	  {
	    cairo_set_source_surface (cr, self->priv->cs_selection, 0, k*self->priv->row_height + self->priv->header_size + 1);
	    cairo_rectangle (cr, 0, 
				 k*self->priv->row_height + self->priv->header_size + 1, 
				 self->priv->reserved_space, 
				 self->priv->row_height);
	    cairo_fill (cr);
	  }

	g_value_unset (&value);
      }

    /* First let the columns render the TreeModel data */
    for (n = 0; n < self->priv->n_columns; n++) 
      {
	if (!self->priv->column_show[n]) continue;
	if (!(r_columns & (1 << n))) goto RENDER_END; 

	//Clear cell background
	if (clear)
	  {
	    cairo_rectangle (cr, render_position, 
			     k * self->priv->row_height + self->priv->header_size,
			     self->priv->pixel_widths[n] + self->priv->cellspacing, 
			     self->priv->row_height);
	    chroma_gdk_color_to_rgba (&(GTK_WIDGET (self)->style->bg[GTK_STATE_NORMAL]), &r, &g, &b, &a);
	    cairo_set_source_rgba (cr, r, g, b, 1.0);
	    cairo_fill (cr);
	  }

	gtk_tree_model_get_value        (self->priv->model,
                                         iter, 
                                         n, 
                                         &value_o); 

	g_value_init (&value, G_VALUE_TYPE(&value_o));
	g_value_copy (&value_o, &value);
	g_value_unset (&value_o);

	if (self->priv->cell_data_funcs[n])
	  {
	    self->priv->cell_data_funcs[n](self,
					      self->priv->model,
					      iter,
					      &value,
					      self->priv->cell_data_funcs_data[n]);
	  }

	column = g_ptr_array_index (self->priv->columns, n);

	if (selected || current)
	  {
	      g_object_set (G_OBJECT(column), "foreground", &(GTK_WIDGET (self)->style->fg[GTK_STATE_PRELIGHT]), NULL);
	      g_object_set (G_OBJECT(column), "background", &(GTK_WIDGET (self)->style->bg[GTK_STATE_NORMAL]), NULL);
	  }
	else
	  {
	      g_object_set (G_OBJECT(column), "foreground", &(GTK_WIDGET (self)->style->fg[GTK_STATE_NORMAL]), NULL);
	      g_object_set (G_OBJECT(column), "background", &(GTK_WIDGET (self)->style->bg[GTK_STATE_NORMAL]), NULL);
	  }

	g_object_set (G_OBJECT(column), "font", GTK_WIDGET (self)->style->font_desc, NULL);
	g_object_set (G_OBJECT(column), "row_height", self->priv->row_height, NULL);

	chroma_list_column_cell_render (column,
					GTK_WIDGET (self),
					&value,
					CHROMA_LIST(self)->priv->pm_backingstore,
					self->priv->pixel_widths[n], // space allowed to render within	
					render_position, 
					k*self->priv->row_height + self->priv->header_size);

	g_value_unset (&value);

	if ((selected) && !(r_columns == self->priv->all_columns_mask))
	  {
	    cairo_set_source_surface (cr, self->priv->cs_selection, 0, k*self->priv->row_height + self->priv->header_size + 1);
	    cairo_rectangle (cr, render_position, 
				 k*self->priv->row_height + self->priv->header_size + 1, 
				 self->priv->pixel_widths[n] + self->priv->cellspacing, 
				 self->priv->row_height);
	    cairo_fill (cr);
	  }

	RENDER_END:
	render_position += (self->priv->pixel_widths[n] + self->priv->cellspacing); 
      }

    if (selected && (r_columns == self->priv->all_columns_mask))
      {
	    cairo_set_source_surface (cr, self->priv->cs_selection, 0, k*self->priv->row_height + self->priv->header_size + 1);
	    cairo_rectangle (cr, 0,
				 k*self->priv->row_height + self->priv->header_size + 1, 
				 GTK_WIDGET(self)->allocation.width,
				 self->priv->row_height);
	    cairo_fill (cr);
      }

    cairo_destroy (cr);

    x = 0; 
    w = GTK_WIDGET(self)->allocation.width; 
    y = CHROMA_LIST(self)->priv->header_size + (CHROMA_LIST(self)->priv->row_height * k);
    h = CHROMA_LIST(self)->priv->row_height; 

    gtk_widget_queue_draw_area (GTK_WIDGET(self),
				      x, y,
				      w, h);
    gdk_flush ();
}

static void
cl_process_cells (ChromaList	    *self,
		  gint		    *rows,
		  guint		     columns)
{
    GtkTreePath  *path;
    GtkTreeIter   iter;
    gint          n; 
	
    G_LOCK (PROCESS_CELLS);

    if (rows)
	  {
	    for (n = 0; rows[n] != NONE; n++)
	      {
		if ((rows[n] - self->priv->offset) >= chroma_list_get_n_item_display (self)) continue;

		path = gtk_tree_path_new_from_indices (rows[n], -1);
		if (path && gtk_tree_model_get_iter (self->priv->model, &iter, path))
		  {
		    gtk_tree_path_free (path);

		    cl_render_row (self, 
				   &iter,
				   rows[n] - self->priv->offset,
				   columns);

		  }
	      }

	      g_free (rows);
	  }
    else 
	  {
	    path = gtk_tree_path_new_from_indices (self->priv->offset, -1);
	    if (path && gtk_tree_model_get_iter (self->priv->model, &iter, path))
	      {
		gtk_tree_path_free (path);
		for (n = 0; n < (chroma_list_get_n_item_display(self)); n++)
		  {
		    cl_render_row (self, 
				&iter,
				n,
			        columns);

		    if (!gtk_tree_model_iter_next (self->priv->model, &iter)) break;
		  }
	      }
	  }

   G_UNLOCK (PROCESS_CELLS);
}

static void
cl_recreate_cs_selection (ChromaList *self)
{
    cairo_t *cr;
    cairo_pattern_t *cr_pattern;
    double r,g,b,a;

    if (self->priv->cs_selection) cairo_surface_destroy (self->priv->cs_selection);

    self->priv->cs_selection = 
	cairo_image_surface_create (CAIRO_FORMAT_ARGB32, GTK_WIDGET(self)->allocation.width - 1, self->priv->row_height);

    cr = cairo_create (self->priv->cs_selection); 

    //Clear background
    cairo_save (cr); 
    cairo_set_operator (cr, CAIRO_OPERATOR_CLEAR);
    cairo_paint (cr);
    cairo_restore (cr);

    //Create selection bar
    chroma_gdk_color_to_rgba (&(GTK_WIDGET (self)->style->fg[GTK_STATE_PRELIGHT]), &r, &g, &b, &a);
    cr_pattern = cairo_pattern_create_linear (0,
					      self->priv->row_height/2,
					      GTK_WIDGET (self)->allocation.width - 1,
					      self->priv->row_height/2);

    cairo_pattern_add_color_stop_rgba (cr_pattern, 0.0, r, g, b, 0.4);
    cairo_pattern_add_color_stop_rgba (cr_pattern, 0.3, r, g, b, 0.28);
    cairo_pattern_add_color_stop_rgba (cr_pattern, 0.6, r, g, b, 0.20);
    cairo_pattern_add_color_stop_rgba (cr_pattern, 1.0, r, g, b, 0.15);

    cairo_pattern_set_filter (cr_pattern, CAIRO_FILTER_BILINEAR);

    chroma_draw_rounded_rectangle_pattern (cr,
                        4, 0, 
                        GTK_WIDGET (self)->allocation.width-(2*self->priv->cellspacing-1) - 1, self->priv->row_height-2.5,
                        cr_pattern,
                        0.8,
                        TRUE,
                        4.5);

#if 0
    draw_rounded_rectangle_aqua (cr,
                        4, 0, 
                        GTK_WIDGET (self)->allocation.width-(2*self->priv->cellspacing-1) - 1, self->priv->row_height-2,
			&(GTK_WIDGET (self)->style->fg[GTK_STATE_PRELIGHT]),
                        0.2,
			1.0,
                        TRUE,
                        4.5);
#endif

    chroma_draw_rounded_rectangle (cr,
                        4, 0, 
                        GTK_WIDGET (self)->allocation.width-(2*self->priv->cellspacing-1) - 1, self->priv->row_height-2,
			&(GTK_WIDGET (self)->style->fg[GTK_STATE_PRELIGHT]),
                        0.5,
			1.0,
                        FALSE,
                        4.5);

    cairo_surface_reference (self->priv->cs_selection);
    cairo_pattern_destroy (cr_pattern);
    cairo_destroy (cr);
}

static void
cl_recreate_pm_backingstore (ChromaList *self)
{
    if (!GTK_WIDGET_REALIZED(GTK_WIDGET (self))) return;

    if (self->priv->pm_backingstore)
      {
        gdk_drawable_unref (self->priv->pm_backingstore);
	self->priv->pm_backingstore = NULL;
      }

    if (self->priv->model)
      {
	if ((GTK_WIDGET (self)->allocation.width > 0) && (GTK_WIDGET (self)->allocation.height > 0))
          {
	    self->priv->pm_backingstore = gdk_pixmap_new (gtk_widget_get_parent (GTK_WIDGET (self))->window,
					   GTK_WIDGET (self)->allocation.width,
					   GTK_WIDGET (self)->allocation.height,
					   gdk_drawable_get_depth (gtk_widget_get_parent(GTK_WIDGET (self))->window));

	    gdk_draw_rectangle (self->priv->pm_backingstore,
			GTK_WIDGET (self)->style->bg_gc[GTK_STATE_NORMAL],	
			TRUE,
			0, 0,
			GTK_WIDGET (self)->allocation.width,
			GTK_WIDGET (self)->allocation.height);
	  }
      }
}

static gboolean
chroma_list_configure (GtkWidget		*widget,
		       GdkEventConfigure	*event)
{
  ChromaList *self = CHROMA_LIST(widget);  
  gdouble     size;

  if (self->priv->adjustment && (chroma_list_get_n_item_display(self) != 0) && (chroma_list_get_length(self) != 0))
      {
	size = (gdouble)chroma_list_get_length (self)/(gdouble)chroma_list_get_n_item_display (self);
	chroma_vscrollbar_set_grip_size (self->priv->adjustment, size); 
      }

  if (chroma_list_get_n_item_display (self) >= gtk_tree_model_iter_n_children (self->priv->model, NULL))
      {
	self->priv->offset = 0;
      }

  chroma_list_refresh (self);

  return FALSE;
}

GtkWidget*
chroma_list_new (void)
{
    ChromaList *self;

    self = g_object_new(chroma_list_get_type(), NULL);

    return GTK_WIDGET (self);
}

static gboolean
chroma_list_button_release (GtkWidget	      *widget, 
			    GdkEventButton    *event) 

{
    ChromaList	*self = CHROMA_LIST(widget);

    self->priv->active_row_drag = FALSE;
    self->priv->active_column_resize = FALSE;
    self->priv->grabbed_position = -1;

    if (self->priv->resize_position == -1)
      {
	gdk_window_set_cursor (widget->window, NULL);
      }

    gtk_widget_queue_draw (widget);

    return FALSE;
}

static gboolean
chroma_list_key_press	    (GtkWidget		  *widget,
			     GdkEventKey	  *event)
{
    ChromaList *self = CHROMA_LIST(widget);

    g_return_val_if_fail (GTK_IS_WIDGET (widget), FALSE);
    g_return_val_if_fail (CHROMA_IS_LIST (widget), FALSE);

    self = CHROMA_LIST(widget);

    if (!self->priv->last_pressed_row)
      {
	  GtkTreePath *path;

	  path = gtk_tree_path_new_from_indices (0, -1);
	  self->priv->last_pressed_row = gtk_tree_row_reference_new (self->priv->model, path);
	  chroma_list_selection_path_set_selected (CHROMA_LIST_SELECTION(widget), path, TRUE);
	  gtk_tree_path_free (path);
	  return FALSE;
      }

    switch (event->keyval)
    {
            case GDK_Return:
            case GDK_KP_Enter:
            {
	      if (self->priv->last_pressed_row)
		{
		    GtkTreePath	     *path;

		    path = gtk_tree_row_reference_get_path (self->priv->last_pressed_row);
		    if (path)
		      {
			  gint position;

			  position = gtk_tree_path_get_indices (path)[0];
			  gtk_tree_path_free (path);
			  g_signal_emit (self, signals[CHROMA_LIST_SIGNAL_ROW_ACTIVATED], 0, position);
		      }
		}
            }
            break;

	    case GDK_Page_Up:
	    {
		  GtkTreePath *path;
		  GtkTreeIter  iter;
		  gint	       position = self->priv->offset;

		  position -= chroma_list_get_n_item_display (self);

		  position = (position < 0) ? 0 : position;

		  chroma_list_scroll_to_offset (self, position); 
		  chroma_list_selection_unselect_all (CHROMA_LIST_SELECTION(widget));	

		  path = gtk_tree_path_new_from_indices (position, -1);
		  gtk_tree_row_reference_free (self->priv->last_pressed_row);
		  self->priv->last_pressed_row = gtk_tree_row_reference_new (self->priv->model, path);
		  gtk_tree_model_get_iter (self->priv->model, &iter, path);

		  chroma_list_selection_iter_set_selected (CHROMA_LIST_SELECTION(self), &iter, TRUE);
		  gtk_tree_path_free (path);

		  g_signal_emit (self, signals[CHROMA_LIST_SIGNAL_CURSOR_POSITION], 0, position); 
	    }
	    break;

	    case GDK_Page_Down:
	    {
		  GtkTreePath *path;
		  GtkTreeIter  iter;
		  gint	       position = self->priv->offset;

		  position += chroma_list_get_n_item_display (self);

		  position = (position > (chroma_list_get_length(self)-chroma_list_get_n_item_display(self))) ? 
		       (chroma_list_get_length(self)-chroma_list_get_n_item_display(self))
		     : position;

		  chroma_list_scroll_to_offset (self, position); 
		  chroma_list_selection_unselect_all (CHROMA_LIST_SELECTION(widget));	

		  path = gtk_tree_path_new_from_indices (position, -1);
		  gtk_tree_row_reference_free (self->priv->last_pressed_row);
		  self->priv->last_pressed_row = gtk_tree_row_reference_new (self->priv->model, path);
		  gtk_tree_model_get_iter (self->priv->model, &iter, path);

		  chroma_list_selection_iter_set_selected (CHROMA_LIST_SELECTION(self), &iter, TRUE);
		  gtk_tree_path_free (path);

		  g_signal_emit (self, signals[CHROMA_LIST_SIGNAL_CURSOR_POSITION], 0, position); 
	    }
	    break;

	    case GDK_Home:
	    {
		  GtkTreePath *path;
		  GtkTreeIter  iter;
		  gint	       position = 0;

		  chroma_list_scroll_to_offset (self, position); 
		  chroma_list_selection_unselect_all (CHROMA_LIST_SELECTION(widget));	

		  path = gtk_tree_path_new_from_indices (position, -1);
		  gtk_tree_row_reference_free (self->priv->last_pressed_row);
		  self->priv->last_pressed_row = gtk_tree_row_reference_new (self->priv->model, path);
		  gtk_tree_model_get_iter (self->priv->model, &iter, path);

		  chroma_list_selection_iter_set_selected (CHROMA_LIST_SELECTION(self), &iter, TRUE);
		  gtk_tree_path_free (path);

		  g_signal_emit (self, signals[CHROMA_LIST_SIGNAL_CURSOR_POSITION], 0, position); 
	    }
	    break;

	    case GDK_End:
	    {
		  GtkTreePath *path;
		  GtkTreeIter  iter;
		  gint	       position = chroma_list_get_length (self)-1;

		  chroma_list_scroll_to_offset (self, chroma_list_get_length(self)-chroma_list_get_n_item_display(self)); 
		  chroma_list_selection_unselect_all (CHROMA_LIST_SELECTION(widget));	

		  path = gtk_tree_path_new_from_indices (position, -1);
		  gtk_tree_row_reference_free (self->priv->last_pressed_row);
		  self->priv->last_pressed_row = gtk_tree_row_reference_new (self->priv->model, path);
		  gtk_tree_model_get_iter (self->priv->model, &iter, path);

		  chroma_list_selection_iter_set_selected (CHROMA_LIST_SELECTION(self), &iter, TRUE);
		  gtk_tree_path_free (path);
	    }
	    break;

            case GDK_Up:
            {
	      gboolean multiple;
	      multiple = chroma_list_selection_multiple_selected (CHROMA_LIST_SELECTION(self)); 
   
	      if (self->priv->last_pressed_row)
		{
		    GtkTreePath	     *path;
		    GtkTreeIter	      iter;
		    gint	      position;

		    path = gtk_tree_row_reference_get_path (self->priv->last_pressed_row);
		    if (path)
		      {
			  position = gtk_tree_path_get_indices (path)[0] - 1;

			  if (position < 0)
			    {
			      gtk_tree_path_free (path);
			      return FALSE;
			    }

			  if (position < self->priv->offset)
			    {
				  chroma_list_scroll_to_offset (self, self->priv->offset-1);
			    }

			  if (!(event->state & GDK_SHIFT_MASK)) chroma_list_selection_unselect_all (CHROMA_LIST_SELECTION(widget));	
			  gtk_tree_path_free (path);
			  path = gtk_tree_path_new_from_indices (position, -1);
			  gtk_tree_row_reference_free (self->priv->last_pressed_row);
			  self->priv->last_pressed_row = gtk_tree_row_reference_new (self->priv->model, path);
			  gtk_tree_model_get_iter (self->priv->model, &iter, path);
			  chroma_list_selection_iter_set_selected (CHROMA_LIST_SELECTION(self), &iter, TRUE);
			  gtk_tree_path_free (path);
			  if (!(event->state & GDK_SHIFT_MASK))
			    {
			      g_signal_emit (self, signals[CHROMA_LIST_SIGNAL_CURSOR_POSITION], 0, position);
			    }
			  else
			    {
			      g_signal_emit (self, signals[CHROMA_LIST_SIGNAL_CURSOR_POSITION], 0, -1);
			    }
		      }
		}
            }
            break;

	    case GDK_Down:
	    {
	      gboolean multiple;
	      multiple = chroma_list_selection_multiple_selected (CHROMA_LIST_SELECTION(self)); 
 
	      if (self->priv->last_pressed_row)
		{
		    GtkTreePath	     *path;
		    GtkTreeIter	      iter;
		    gint	      position;

		    path = gtk_tree_row_reference_get_path (self->priv->last_pressed_row);
		    if (path)
		      {
			  gint max;
  
			  max = chroma_list_get_n_item_display (self)-1;
			  if ((chroma_list_get_length (self)-1) < max) max = (chroma_list_get_length (self)-1);
  
			  position = gtk_tree_path_get_indices (path)[0] + 1;

			  if (position > (chroma_list_get_length (self)-1))
			    {
				gtk_tree_path_free (path);
				return FALSE;
			    }

			  if (position > (self->priv->offset+max))
			    {
				  chroma_list_scroll_to_offset (self, self->priv->offset+1);
			    }

			  if (!(event->state & GDK_SHIFT_MASK)) chroma_list_selection_unselect_all (CHROMA_LIST_SELECTION(widget));	
			  gtk_tree_path_free (path);
			  path = gtk_tree_path_new_from_indices (position, - 1);
			  gtk_tree_row_reference_free (self->priv->last_pressed_row);
			  self->priv->last_pressed_row = gtk_tree_row_reference_new (self->priv->model, path);
			  gtk_tree_model_get_iter (self->priv->model, &iter, path);
			  chroma_list_selection_iter_set_selected (CHROMA_LIST_SELECTION(self), &iter, TRUE);
			  gtk_tree_path_free (path);
			  if (!(event->state & GDK_SHIFT_MASK))
			    {
			      g_signal_emit (self, signals[CHROMA_LIST_SIGNAL_CURSOR_POSITION], 0, position);
			    }
			  else
			    {
			      g_signal_emit (self, signals[CHROMA_LIST_SIGNAL_CURSOR_POSITION], 0, -1);
			    }
		      }
		}
	    }
	    break;
  
	  default: break;

      }

      return FALSE;
}

#define SINGLE_CLICK (event->type == GDK_BUTTON_PRESS)
#define DOUBLE_CLICK (event->type == GDK_2BUTTON_PRESS)

#define BUTTON_LEFT  (event->button == 1)
#define BUTTON_RIGHT (event->button == 3)

#define INSIDE_ROWS (event->y > self->priv->header_size)
#define INSIDE_HEADERS (event->y <= self->priv->header_size)

#define COLUMN_RESIZING (self->priv->resize_position != -1)

static gboolean
chroma_list_button_press (GtkWidget	  *widget,
			  GdkEventButton  *event)
{
    ChromaList *self = CHROMA_LIST(widget);

    g_return_val_if_fail(CHROMA_IS_LIST(self), FALSE);

    if (!self->priv->model) return FALSE;

    gtk_widget_grab_focus (widget);

    if (BUTTON_LEFT && INSIDE_ROWS && DOUBLE_CLICK && !COLUMN_RESIZING)
      {
	  gint row;

	  row = ((event->y - self->priv->header_size) / self->priv->row_height) + self->priv->offset;
	  if (row > (chroma_list_get_length (self)-1)) return FALSE;

	  if (!(event->state & GDK_SHIFT_MASK) && !(event->state & GDK_CONTROL_MASK)) 
	    {
	      g_signal_emit (self, signals[CHROMA_LIST_SIGNAL_CURSOR_POSITION], 0, row);
	      g_signal_emit (self, signals[CHROMA_LIST_SIGNAL_ROW_ACTIVATED], 0, row);
	    }

	  return FALSE;
      }

    if (BUTTON_LEFT && INSIDE_ROWS && SINGLE_CLICK && !COLUMN_RESIZING)
      {
	  GtkTreePath *path;
	  GtkTreeIter  iter;
	  gint row;

	  row = ((event->y - self->priv->header_size) / self->priv->row_height) + self->priv->offset;
	  if (row > (chroma_list_get_length (self)-1)) return FALSE;

	  path = gtk_tree_path_new_from_indices (row, -1);
	  gtk_tree_model_get_iter (self->priv->model, &iter, path);
	  gtk_tree_path_free (path);
  
	  self->priv->active_row_drag = TRUE;
	  self->priv->grabbed_position = row;

	  if (!(event->state & GDK_SHIFT_MASK) && !(event->state & GDK_CONTROL_MASK)) 
	    {
	      GtkTreePath *path;
	      gint	   n; 
	      gboolean	   multiple;
 
	      n = 0;
	      multiple = (g_list_length (self->priv->selected_rows)>1) ? TRUE : FALSE;
 
	      chroma_list_selection_unselect_all (CHROMA_LIST_SELECTION(self));	
	      chroma_list_selection_iter_set_selected (CHROMA_LIST_SELECTION(self), &iter, TRUE); 

	      gtk_tree_row_reference_free (self->priv->last_pressed_row);
	      path = gtk_tree_model_get_path (self->priv->model, &iter);
	      self->priv->last_pressed_row = gtk_tree_row_reference_new (self->priv->model, path); 
	      gtk_tree_path_free (path);
	      g_signal_emit (self, signals[CHROMA_LIST_SIGNAL_CURSOR_POSITION], 0, row);
	    }
	  else
	  if ((event->state & GDK_CONTROL_MASK) == GDK_CONTROL_MASK)
	    {
	      gboolean selected = chroma_list_selection_iter_get_selected(CHROMA_LIST_SELECTION(self), &iter);
	      chroma_list_selection_iter_set_selected (CHROMA_LIST_SELECTION(self), &iter, !selected);

	      gtk_tree_row_reference_free (self->priv->last_pressed_row);
	      path = gtk_tree_model_get_path (self->priv->model, &iter);
	      if (!selected)
		{
		  self->priv->last_pressed_row = gtk_tree_row_reference_new (self->priv->model, path); 
		}
	      else
		{
		  self->priv->last_pressed_row = NULL;
		}
	      gtk_tree_path_free (path);

	      g_signal_emit (self, signals[CHROMA_LIST_SIGNAL_CURSOR_POSITION], 0, row);
	    }
	  if ((event->state & GDK_SHIFT_MASK) == GDK_SHIFT_MASK)
	    {
	      GtkTreePath *path;
	      gint last_pressed_row;

	      if (!self->priv->last_pressed_row) return FALSE;

	      path = gtk_tree_row_reference_get_path (self->priv->last_pressed_row);
	      if (path)
		{
		  last_pressed_row = gtk_tree_path_get_indices (path)[0];
		  gtk_tree_path_free (path);
		}
	      else
		{
		  last_pressed_row = 0;
		}
	    
	      if (row < last_pressed_row) 
		{
		    GtkTreePath *path_l;
		    GtkTreeIter  iter_l;
		    gint n;
  
		    for (n = row; n <= last_pressed_row; n++)
		      {
			  path_l = gtk_tree_path_new_from_indices (n, -1);
			  gtk_tree_model_get_iter (self->priv->model, &iter_l, path_l);
		  	  gtk_tree_path_free (path_l);
  
			  chroma_list_selection_iter_set_selected (CHROMA_LIST_SELECTION(self), &iter_l, TRUE); 
		      }

		    g_signal_emit (self, signals[CHROMA_LIST_SIGNAL_CURSOR_POSITION], 0, n-1);
		}
	      else
	      if (row > last_pressed_row)
		{
		    GtkTreePath *path_l;
		    GtkTreeIter  iter_l;
		    gint n;
  
		    for (n = last_pressed_row; n <= row; n++)
		      {
			  path_l = gtk_tree_path_new_from_indices (n, -1);
			  gtk_tree_model_get_iter (self->priv->model, &iter_l, path_l);
		  	  gtk_tree_path_free (path_l);
  
			  chroma_list_selection_iter_set_selected (CHROMA_LIST_SELECTION(self), &iter_l, TRUE); 
		      }

		    g_signal_emit (self, signals[CHROMA_LIST_SIGNAL_CURSOR_POSITION], 0, n-1);
		}
	    }

	  return FALSE;
      }
    else if (BUTTON_LEFT && INSIDE_ROWS && SINGLE_CLICK && (self->priv->resize_position != -1))
      {
	  self->priv->active_column_resize = TRUE;
	  return FALSE;
      }

    if (BUTTON_LEFT && INSIDE_ROWS && SINGLE_CLICK && !COLUMN_RESIZING && (self->priv->resize_position != -1))
      {
	  self->priv->active_column_resize = TRUE;
	  return FALSE;
      }

    if (BUTTON_LEFT && INSIDE_HEADERS && GTK_IS_TREE_SORTABLE(self->priv->model))
      {
	gdouble x;
	gint    n;

	x = self->priv->reserved_space;

	for (n = 0; n < self->priv->n_columns; n ++)
	  {
	    if ((event->x >= x) &&
	        (event->x <  x+self->priv->pixel_widths[n]))
	      {
		  GtkSortType sort_type;
		  gint col;

		  if (self->priv->sort_column_id == n)
		    {
			self->priv->sort_trigger++;
			sort_type = (self->priv->sort_type == GTK_SORT_ASCENDING) ? GTK_SORT_DESCENDING : GTK_SORT_ASCENDING;
			col = n;
		    }
		  else
		    {
			self->priv->sort_trigger = 0;
			sort_type = GTK_SORT_ASCENDING;
			col = n;
		    }

		  if (self->priv->sort_trigger == 2)
		    {
			self->priv->sort_trigger = 0;
			sort_type = GTK_SORT_ASCENDING;
			col = GTK_TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID;

		    	gtk_tree_sortable_set_sort_column_id (GTK_TREE_SORTABLE(self->priv->model),
						        self->priv->sort_column_id,
							GTK_SORT_ASCENDING);
		    }

		  gtk_tree_sortable_set_sort_column_id (GTK_TREE_SORTABLE(self->priv->model),
						        col,
							sort_type);

		  return FALSE;
	      }
	    x += self->priv->pixel_widths[n] + self->priv->cellspacing;
	  }
      }

    return FALSE;
}

static gboolean
chroma_list_scroll	    (GtkWidget		  *widget,
			     GdkEventScroll	  *event)
{
    ChromaList	  *self = CHROMA_LIST (widget);

    g_return_val_if_fail (GTK_IS_WIDGET(widget), FALSE);
    g_return_val_if_fail (CHROMA_IS_LIST(self), FALSE);

    if (event->direction == GDK_SCROLL_DOWN)
      {
	chroma_list_scroll_to_offset (self, self->priv->offset+3);
      }
    else if (event->direction == GDK_SCROLL_UP)
      {
	chroma_list_scroll_to_offset (self, self->priv->offset-3);
      }

    return FALSE;
}

G_LOCK_DEFINE(MOVE_ROWS);

static gboolean
scroll_list (ChromaList *self)
{
    switch (self->priv->scroll_direction)
      {
	  case SCROLL_DIRECTION_UP:
	    {
	      gint	   position;

	      position = self->priv->offset - 1;
	      if (position < 0)
		{
		  self->priv->scroll_timeout_source_id = 0;
		  return FALSE;
		}

	      self->priv->model_row_swap_func (self->priv->model,
						  self->priv->grabbed_position,
						  position);
	      self->priv->grabbed_position = position;
	      chroma_list_scroll_to_offset (self, position);
	      break;
	    }

	  case SCROLL_DIRECTION_DOWN:
	    {
	      break;
	    }

      }

    return TRUE;
}

static void
destroy_scroll_source (ChromaList *self)
{
    //Destroy the scrolling source
    GSource *source;

    if (self->priv->scroll_timeout_source_id > 0)
      {
	  source = g_main_context_find_source_by_id (NULL, self->priv->scroll_timeout_source_id);
	  g_source_destroy (source);
	  self->priv->scroll_timeout_source_id = 0;
      }
}

static gboolean
chroma_list_motion (GtkWidget	    *widget,
		    GdkEventMotion  *event)

{
    ChromaList	     *self = CHROMA_LIST(widget);

    gint	      x, y;
    GdkModifierType   state;

    g_return_val_if_fail(CHROMA_IS_LIST(self), FALSE);

    if (!self->priv->model)
	return FALSE;

    if (event->is_hint)
      {
	gdk_window_get_pointer (event->window, &x, &y, &state);
      }
    else
      {
	x = event->x;
	y = event->y;
	state = event->state;
      }

    //Column resize
    if (self->priv->active_column_resize)
      {
	  gint    pixel_width_l,
		  pixel_width_r;

	  gdouble column_width_l,
		  column_width_r;

	  gint    delta;
	  gint    l = self->priv->adjactent_l;
	  gint    r = self->priv->adjactent_r;

	  guint   columns = 0;

	  delta = x - self->priv->resize_position;
  
	  if (delta > 0)
	    {
                  pixel_width_l  = self->priv->pixel_widths[l];
                  pixel_width_r  = self->priv->pixel_widths[r];

                  pixel_width_l += delta;
                  pixel_width_r -= delta;

                  column_width_l = pixel_width_to_percent (self, pixel_width_l);
                  column_width_r = pixel_width_to_percent (self, pixel_width_r);

		  if (pixel_width_r > COLUMN_MINIMAL_WIDTH)
		    {
			self->priv->column_widths[l] = column_width_l;
			self->priv->column_widths[r] = column_width_r;

			self->priv->pixel_widths[l] = pixel_width_l;
			self->priv->pixel_widths[r] = pixel_width_r;

			columns |= 1 << l;
			columns |= 1 << r;
			goto EXIT_LOOP;    	
		    }
	    }
	  else if (delta < 0)
	    {
                  pixel_width_l   = self->priv->pixel_widths[l];
                  pixel_width_r   = self->priv->pixel_widths[r];

                  pixel_width_l  += delta;
                  pixel_width_r  -= delta;

                  column_width_l  = pixel_width_to_percent (self, pixel_width_l);
                  column_width_r  = pixel_width_to_percent (self, pixel_width_r);

		  if (pixel_width_l > COLUMN_MINIMAL_WIDTH)
		    {
			self->priv->column_widths[l] = column_width_l;
			self->priv->column_widths[r] = column_width_r;

			self->priv->pixel_widths[l] = pixel_width_l;
			self->priv->pixel_widths[r] = pixel_width_r;

			columns |= 1 << l;
			columns |= 1 << r;
			goto EXIT_LOOP;    	
		    }
	    }
  
	  return FALSE;

	  EXIT_LOOP:	  
	  self->priv->resize_position = x; 
	  cl_process_cells (self, NULL, columns);
	  gtk_widget_queue_draw (GTK_WIDGET (self));
	  return FALSE;
      }

    //Row drag is active
    if (self->priv->active_row_drag)
      {
	gint  position,
	      position_max;

	G_LOCK(MOVE_ROWS);

	position = ((y - self->priv->header_size) / self->priv->row_height) + self->priv->offset;
	if (position < 0)
	  {
	    destroy_scroll_source (self);
	    return FALSE;
	  }

	position_max = chroma_list_get_n_item_display (self) + self->priv->offset - 1;
	if (position_max > (chroma_list_get_length (self)-1)) position_max = chroma_list_get_length (self)-1;
	if (position > position_max) 
	  {
	    G_UNLOCK(MOVE_ROWS);
	    return FALSE;
	  }

        if (!(self->priv->grabbed_position - position))
	  {
	    G_UNLOCK(MOVE_ROWS);
	    return FALSE;
	  } 
  
	if (position < self->priv->offset)
	  {
	    chroma_list_scroll_to_offset (self, self->priv->offset - 1);
	    self->priv->scroll_timeout_source_id = g_timeout_add_full (G_PRIORITY_HIGH_IDLE+20, 100, (GSourceFunc) scroll_list, self, NULL);
	    self->priv->scroll_direction = SCROLL_DIRECTION_UP;
	  }
	else
	  {
	    destroy_scroll_source (self);
	  }

	if ((position - self->priv->offset) > (chroma_list_get_n_item_display (self)-1)) 
	  {
	    G_UNLOCK(MOVE_ROWS);
	    return FALSE;
	  }

	if (GTK_IS_TREE_SORTABLE(self->priv->model))
	  {
	    gtk_tree_sortable_set_sort_column_id (GTK_TREE_SORTABLE(self->priv->model),
					      GTK_TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID, 
					      GTK_SORT_ASCENDING);
	  }

	self->priv->model_row_swap_func (self->priv->model,
					    self->priv->grabbed_position,
					    position);

        self->priv->grabbed_position = position;
	G_UNLOCK(MOVE_ROWS);
	return FALSE;
      }

    //Check if we're in between a particular header
    if ((y > self->priv->header_size) && (!self->priv->active_row_drag) && (!self->priv->active_column_resize))
      {
	gint	  l, r, n, hx; 

	hx    = self->priv->reserved_space;
	l     = 0; 
	r     = 0;

	//Find left border
	for (n = 0; n < (self->priv->n_columns-1); n++)
	  {
	      if (!self->priv->column_show[n]) continue;
	      if ((hx + self->priv->pixel_widths[n]) > x) break;

	      l = n;
	      hx += self->priv->pixel_widths[n] + self->priv->cellspacing;
	  }

	//Find right border
	for (n = l+1; n < (self->priv->n_columns-1); n++)
	  {
	      if (!self->priv->column_show[n])
		{
		    continue;
		}
	      else
		{
		    r = n;
		    break;
		}
	  }

	hx -= self->priv->cellspacing;

	if ((x >= (hx)) &&
	    (x <= (hx+self->priv->cellspacing)) &&
	    (hx > (self->priv->cellspacing+self->priv->reserved_space)) &&
	    (hx < (GTK_WIDGET(self)->allocation.width - (self->priv->cellspacing*2))))
	  {
	      GdkCursor *cursor;

	      self->priv->adjactent_l	= l;
	      self->priv->adjactent_r	= r;
	      self->priv->resize_position	= hx + self->priv->cellspacing/2; 

	      cursor = gdk_cursor_new_from_name (gdk_display_get_default (), "sb_h_double_arrow");
	      gdk_window_set_cursor (widget->window, cursor);
	      gtk_widget_queue_draw (GTK_WIDGET (self));
	  }
	else if (self->priv->resize_position != NONE)
	  {
	      self->priv->resize_position = NONE; 
	      gdk_window_set_cursor (widget->window, NULL);
	      gtk_widget_queue_draw (GTK_WIDGET (self));
	  }

	return FALSE;
      }

    return FALSE;
}

static gboolean
chroma_list_expose (GtkWidget	    *widget,
		    GdkEventExpose  *event) 
{
    ChromaList	 *self = CHROMA_LIST(widget);
    gdouble	  x;
    gint	  n;

    cairo_t	 *cr;

    double	  r,
		  g,
		  b,
		  a;

    g_return_val_if_fail(CHROMA_IS_LIST(self), FALSE);

    if (!self->priv->model)
	return FALSE;

    if (!self->priv->pm_backingstore)
        cl_recreate_pm_backingstore (self);

    gdk_draw_drawable (GTK_WIDGET (self)->window, 
		       GTK_WIDGET (self)->style->fg_gc[GTK_STATE_NORMAL],
                       self->priv->pm_backingstore, 
		       event->area.x,  event->area.y,
		       event->area.x,  event->area.y,
		       event->area.width, event->area.height); 

    //Create local cairo context
    cr = gdk_cairo_create (GTK_WIDGET (self)->window);

    //Draw column headers
    x = self->priv->reserved_space;

    if (event->area.y <= self->priv->header_size)
      for (n = 0; n < self->priv->n_columns; n ++)
	{
	  PangoLayout	  *layout;
	  PangoAttrList	  *attr_list;
	  PangoAttribute  *attr;
	  PangoRectangle   rect1, 
			   rect2;

	  gchar		  *text;
	  gboolean	   show_name;

	  g_object_get (G_OBJECT(g_ptr_array_index (self->priv->columns, n)), "show_name", &show_name, NULL);

	  if (!show_name)
	    {
	      x += self->priv->pixel_widths[n];
	      continue;
	    }

	  if (n == self->priv->sort_column_id)
	      a = 0.4;
	  else
	      a = 0.2;

#if 0
	  draw_rounded_rectangle_aqua (cr, x, (gdouble)4, 
				(gdouble)self->priv->pixel_widths[n], (gdouble)self->priv->header_size - 7, 
				 &(GTK_WIDGET (self)->style->fg[GTK_STATE_PRELIGHT]), a, 0.4, TRUE, 4.0);
#endif
	  chroma_draw_rounded_rectangle (cr, x, (gdouble)4, 
				(gdouble)self->priv->pixel_widths[n], (gdouble)self->priv->header_size - 7.5, 
				 &(GTK_WIDGET (self)->style->fg[GTK_STATE_PRELIGHT]), a, 0.4, TRUE, 4.0);

	  chroma_draw_rounded_rectangle (cr, x, (gdouble)4, 
				(gdouble)self->priv->pixel_widths[n], (gdouble)self->priv->header_size - 7, 
				 &(GTK_WIDGET (self)->style->fg[GTK_STATE_PRELIGHT]), 1.0, 0.6, FALSE, 4.0);

	  if (n == self->priv->sort_column_id)
	    {
		GdkRectangle *rect;

		rect = g_new0 (GdkRectangle, 1);
		rect->x = x + self->priv->pixel_widths[n] - self->priv->header_size + 8; 
		rect->y = 4;
		rect->width = self->priv->header_size-9; 
		rect->height = self->priv->header_size-9; 

		chroma_draw_arrow (cr, 
			    rect, 
			    &(GTK_WIDGET (self)->style->fg[GTK_STATE_PRELIGHT]),
			    (self->priv->sort_type == GTK_SORT_ASCENDING) ? GTK_ARROW_DOWN : GTK_ARROW_UP);

		g_free (rect);

	    }

	  g_object_get (G_OBJECT(g_ptr_array_index (self->priv->columns, n)), "name", &text, NULL); 

	  if (text)
	    {
		chroma_gdk_color_to_rgba (&(GTK_WIDGET (self)->style->fg[GTK_STATE_PRELIGHT]), &r, &g, &b, &a);

		attr_list = pango_attr_list_new();
		attr = pango_attr_foreground_new (r * 65535, g * 65535, b * 65535);
		pango_attr_list_insert (attr_list, attr);

		layout = pango_cairo_create_layout (cr);
		pango_layout_set_attributes (layout, attr_list);

		pango_layout_set_markup (layout, text, strlen(text));

		pango_layout_set_font_description (layout, GTK_WIDGET (self)->style->font_desc); 
		pango_layout_set_ellipsize (layout, PANGO_ELLIPSIZE_END);
		pango_layout_set_alignment (layout, PANGO_ALIGN_LEFT);
		pango_layout_set_width (layout, (self->priv->pixel_widths[n]-8) * PANGO_SCALE);
		pango_layout_get_pixel_extents (layout, &rect1, &rect2);

		cairo_move_to (cr, x+6., ((gdouble)((self->priv->header_size - rect2.height)/2)-1.)+2);

		cairo_set_source_rgb (cr, r,g,b);
		pango_cairo_show_layout (cr, layout);

		g_object_unref (layout);
		g_free (text);
	    }

	  x += self->priv->pixel_widths[n] + self->priv->cellspacing;
	}

    if (self->priv->resize_position != -1)
      {
	  gdouble x, y, w, h;

	  x = self->priv->resize_position;
	  y = self->priv->header_size;
	  w = 1;
	  h = GTK_WIDGET (self)->allocation.height - self->priv->header_size;

	  chroma_gdk_color_to_rgba (&(GTK_WIDGET (self)->style->fg[GTK_STATE_PRELIGHT]), &r, &g, &b, &a);

	  cairo_save (cr);
	  cairo_set_line_width (cr, 0.5);
	  cairo_move_to (cr, x, y);
	  cairo_line_to (cr, x, y+h);
	  cairo_set_source_rgba (cr, r, g, b, 0.75);
	  cairo_set_line_cap (cr, CAIRO_LINE_CAP_ROUND);
	  cairo_set_line_join (cr, CAIRO_LINE_JOIN_ROUND);
	  cairo_stroke(cr);
	  cairo_restore (cr);
      }

    if (self->priv->drop)
      {
        gint x, y, w;

        x = self->priv->cellspacing; 
        y = (self->priv->row_height * self->priv->drop_row) + self->priv->header_size + 2;
        w = GTK_WIDGET (self)->allocation.width-(2*self->priv->cellspacing);

	chroma_gdk_color_to_rgba (&(GTK_WIDGET (self)->style->fg[GTK_STATE_PRELIGHT]), &r, &g, &b, &a);

        cairo_set_line_width(cr, 0.7);
        cairo_move_to (cr, x, y-0.5);
        cairo_line_to (cr, x + w, y-0.5);
        cairo_set_source_rgba (cr, r, g, b, 1.0);
        cairo_set_line_cap (cr, CAIRO_LINE_CAP_ROUND);
        cairo_set_line_join (cr, CAIRO_LINE_JOIN_ROUND);
        cairo_stroke(cr);
      }

    cairo_destroy (cr);

    if (GTK_WIDGET_HAS_FOCUS(widget))
      {
	gtk_paint_focus  (widget->style,
                          widget->window, 
                          GTK_STATE_PRELIGHT, 
                          NULL, 
                          widget, 
                          "", 
                          1, 
                          0, 
                          widget->allocation.width-2, 
                          widget->allocation.height); 
      }

    return FALSE;
}

static gboolean
chroma_list_leave_notify (GtkWidget *widget,
                          GdkEventCrossing *event)
{
    g_return_val_if_fail (GTK_IS_WIDGET (widget), FALSE);
    g_return_val_if_fail (CHROMA_IS_LIST (widget), FALSE);

    CHROMA_LIST(widget)->priv->resize_position = -1;
    gdk_window_set_cursor (widget->window, NULL);

    gtk_widget_queue_draw (widget);

    return FALSE;
}

/* Convenience stuff for setting properties */
void
chroma_list_set_use_numbers (ChromaList	      *self,
			     gboolean	       use_numbers)
{
    g_return_if_fail(CHROMA_IS_LIST(self));

    self->priv->use_numbers = use_numbers;
    
    chroma_list_refresh (self);
}

gint
chroma_list_get_n_item_display (ChromaList *self)
{
    gint n_items;

    g_return_val_if_fail(CHROMA_IS_LIST(self), -1);

    if (!self->priv->row_height) return 0;

    n_items = (GTK_WIDGET (self)->allocation.height - self->priv->header_size) / self->priv->row_height;

    return n_items;
}

gint
chroma_list_get_length (ChromaList *self)
{
    g_return_val_if_fail(CHROMA_IS_LIST(self), -1);

    if (!self->priv->model)
        return 0;

    return gtk_tree_model_iter_n_children (self->priv->model, NULL); 
}

GtkTreeModel*
chroma_list_get_model (ChromaList *self)
{
    g_return_val_if_fail(CHROMA_IS_LIST(self), NULL);

    return self->priv->model;
}

void
on_row_changed      (GtkTreeModel  *treemodel,
                      GtkTreePath   *arg1,
                      GtkTreeIter   *arg2,
                      ChromaList    *self)
{
  gint *rows;
  gint  index;

  index = gtk_tree_path_get_indices(arg1)[0];

  if ((index >= self->priv->offset) &&
      (index < (self->priv->offset+chroma_list_get_n_item_display (self))))
    {
      rows = realloc (NULL, 2*sizeof(gint));
      rows[0] = index;
      rows[1] = -1;
      cl_process_cells (self, rows, self->priv->all_columns_mask);
    }
}

void
on_row_deleted      (GtkTreeModel  *treemodel,
                      GtkTreePath   *path,
                      ChromaList    *self)
{
  cl_recalc_reserved_space (self);
  cl_recalc_pixel_widths (self);

  chroma_vscrollbar_set_grip_size (self->priv->adjustment, 
				   (gdouble)chroma_list_get_length(self)/(gdouble)chroma_list_get_n_item_display(self));

  if (chroma_list_get_length (self) < chroma_list_get_n_item_display (self))
      g_object_set (G_OBJECT(self), "offset", (gint)0, NULL);
}

void
on_row_inserted      (GtkTreeModel   *treemodel,
                       GtkTreePath    *arg1,
                       GtkTreeIter    *arg2,
                       ChromaList     *self)
{
  if (GTK_WIDGET_VISIBLE(self->priv->adjustment))
    chroma_vscrollbar_set_grip_size (self->priv->adjustment, 
				     (gdouble)chroma_list_get_length(self)/(gdouble)chroma_list_get_n_item_display(self));
}

void
on_rows_reordered      (GtkTreeModel	*treemodel,
		         GtkTreePath	*arg1,
			 GtkTreeIter	*arg2,
                         gint	  	*arg3,
                         ChromaList	*self)
{
  gint *rows;
  gint  n,
	start,
	end,
	ctr;

  ctr   = 0;
  start = self->priv->offset; 
  end   = self->priv->offset + chroma_list_get_n_item_display (self); 
  rows = NULL;

  for (n = 0; n < chroma_list_get_length (self); n++)
    {
      if ((arg3[n] != n) && ((arg3[n] >= start) && (arg3[n] <end)))
	{
	  rows = realloc (rows, (ctr*sizeof(gint)) + sizeof(gint));
	  rows[ctr] = arg3[n];
	  ctr++;
	}
    }

  rows = realloc (rows, (ctr*sizeof(gint)) + sizeof(gint));
  rows[ctr] = -1;

  cl_process_cells (self, rows, self->priv->all_columns_mask);
  // gtk_widget_queue_draw (GTK_WIDGET (self));
}

void
on_adjustment_position (ChromaVScrollbar *adjustment,
			gdouble		  position,
			gpointer	  data)
{
    gint offset = 0, n_items = 0, n_item_display = 0;

    ChromaList *self = CHROMA_LIST (data);

    G_LOCK(SCROLL_OFFSET);

    n_items =
      chroma_list_get_length (self);
    n_item_display =
      chroma_list_get_n_item_display (self);

    if (n_items < n_item_display) return;

    offset = (n_items - n_item_display) * position;

    if (offset >= 0)
      {
	if (self->priv->offset != offset)
	  g_object_set (G_OBJECT(self), "offset", offset, NULL);
      }

    G_UNLOCK(SCROLL_OFFSET);
}

static void
on_adjustment_step_up	(ChromaVScrollbar   *scrollbar,
			 ChromaList	    *self)
{
    gint offset;

    offset = self->priv->offset - 3;
    offset = (offset < 0) ? 0 : offset;

    chroma_list_scroll_to_offset (self, offset); 
}

static void
on_adjustment_step_down (ChromaVScrollbar   *scrollbar,
			 ChromaList	    *self)
{
    gint offset;

    offset = self->priv->offset + 3;
    offset = (offset >= chroma_list_get_length (self)) ? (chroma_list_get_length (self)-1) : offset;

    chroma_list_scroll_to_offset (self, offset); 
}


static void
on_sort_column_changed	(GtkTreeSortable *treesortable,
			 ChromaList	 *self)
{
    gtk_tree_sortable_get_sort_column_id (treesortable,
					  &(self->priv->sort_column_id),
					  &(self->priv->sort_type));

    cl_process_cells (self, NULL, self->priv->all_columns_mask);
    gtk_widget_queue_draw (GTK_WIDGET (self));
}

void
chroma_list_set_model (ChromaList	      *self,
		       GtkTreeModel	      *model,
		       ChromaListRowSwapFunc   model_row_swap_func)
{
    ChromaListColumn *column;
    gint n;

    g_return_if_fail(CHROMA_IS_LIST(self));

    if (self->priv->model)
      {
	for (n = 0; n < N_MODEL_SIGNALS; n++)
	  {
	    if (self->priv->model_signals[n])
		g_signal_handler_disconnect (self->priv->model, self->priv->model_signals[n]); 
	  }
	g_object_unref (self->priv->model);

	if (self->priv->columns && (self->priv->n_columns > 0))
	  {
	    for (n = 0; n < self->priv->n_columns; n++) 
	      {
		g_object_unref (g_ptr_array_index (self->priv->columns, n));
		g_ptr_array_remove_fast (self->priv->columns, 0);
	      }
	    g_ptr_array_free (self->priv->columns, FALSE);

	    g_free (self->priv->column_widths);
	    g_free (self->priv->column_xalign);
	    g_free (self->priv->column_yalign);
	 }
      }

    self->priv->model = model;
    if (!model) return;

    g_object_ref (self->priv->model);

    self->priv->model_row_swap_func = model_row_swap_func;

    self->priv->model_signals[SIGNAL_ROW_CHANGED] =
	    g_signal_connect_data           ((gpointer)self->priv->model,
                                             "row-changed", 
                                             G_CALLBACK(on_row_changed), 
                                             self, 
                                             NULL, 
                                             0); 

    self->priv->model_signals[SIGNAL_ROW_DELETED] =
	    g_signal_connect_data           ((gpointer)self->priv->model,
                                             "row-deleted", 
                                             G_CALLBACK(on_row_deleted), 
                                             self, 
                                             NULL, 
                                             0); 

    self->priv->model_signals[SIGNAL_ROW_INSERTED] =
	    g_signal_connect_data           ((gpointer)self->priv->model,
                                             "row-inserted", 
                                             G_CALLBACK(on_row_inserted), 
                                             self, 
                                             NULL, 
                                             0); 

    self->priv->model_signals[SIGNAL_ROWS_REORDERED] =
	    g_signal_connect_data           ((gpointer)self->priv->model,
                                             "rows-reordered", 
                                             G_CALLBACK(on_rows_reordered), 
                                             self, 
                                             NULL, 
                                             0); 

    if (GTK_IS_TREE_SORTABLE(self->priv->model))
      {
	  self->priv->model_signals[SIGNAL_SORT_COLUMN_CHANGED] =
	    g_signal_connect_data           ((gpointer)self->priv->model,
                                             "sort-column-changed", 
                                             G_CALLBACK(on_sort_column_changed), 
                                             self, 
                                             NULL, 
                                             0); 
	  gtk_tree_sortable_get_sort_column_id (GTK_TREE_SORTABLE(self->priv->model),
					  &(self->priv->sort_column_id),
					  &(self->priv->sort_type));
      }

    self->priv->columns = g_ptr_array_new();
    self->priv->n_columns = gtk_tree_model_get_n_columns (self->priv->model); 
    self->priv->all_columns_mask = 0; 
    for (n = 0; n < self->priv->n_columns; n++) self->priv->all_columns_mask |= 1 << n; 

    self->priv->n_columns_visible = gtk_tree_model_get_n_columns (self->priv->model); 

    self->priv->column_show = g_new0(gboolean, self->priv->n_columns); 
    for (n = 0; n < self->priv->n_columns; self->priv->column_show[n++] = TRUE); 

    self->priv->column_widths = g_new0(gdouble, self->priv->n_columns); 
    self->priv->column_xalign = g_new0(gdouble, self->priv->n_columns); 
    self->priv->column_yalign = g_new0(gdouble, self->priv->n_columns); 

    self->priv->cell_data_funcs = g_new0 (ChromaListCellDataFunc, self->priv->n_columns);
    self->priv->cell_data_funcs_data = g_new0 (gpointer, self->priv->n_columns);

    for (n = 0; n < self->priv->n_columns; n++) 
      {
	column = chroma_list_column_new();		

	g_object_set (column, "font", GTK_WIDGET (self)->style->font_desc, NULL);
	g_object_set (column, "foreground", &(GTK_WIDGET (self)->style->fg[GTK_STATE_NORMAL]), NULL);
	g_object_set (column, "background", &(GTK_WIDGET (self)->style->bg[GTK_STATE_NORMAL]), NULL);

	g_ptr_array_add (self->priv->columns, column);

	self->priv->column_yalign[n] = 0.5;
	self->priv->column_xalign[n] = 0.0;
	self->priv->column_widths[n] = 1.0;
      }

    g_object_set (self, "offset", (gint)0, NULL);

    cl_recalc_reserved_space (self);
    cl_recalc_pixel_widths (self);
    cl_recreate_pm_backingstore (self);

    gtk_widget_queue_draw (GTK_WIDGET (self));
}

void
chroma_list_set_adjustment (ChromaList	      *self,
			    ChromaVScrollbar  *adjustment)
{
    gint n;

    g_return_if_fail(CHROMA_IS_LIST(self));

    if (self->priv->adjustment)
      {
	for (n = 0; n < N_MODEL_SIGNALS; n++)
	    g_signal_handler_disconnect (self->priv->adjustment, self->priv->adjustment_signals[n]); 

	g_object_unref (self->priv->adjustment);
      }

    self->priv->adjustment = adjustment;
    g_object_ref (adjustment);

    self->priv->adjustment_signals[SIGNAL_POSITION] =
	    g_signal_connect_data           ((gpointer)self->priv->adjustment,
                                             "position", 
                                             G_CALLBACK(on_adjustment_position), 
                                             self, 
                                             NULL, 
                                             0); 

    self->priv->adjustment_signals[SIGNAL_STEP_UP] =
	    g_signal_connect_data           ((gpointer)self->priv->adjustment,
                                             "step_up", 
                                             G_CALLBACK(on_adjustment_step_up), 
                                             self, 
                                             NULL, 
                                             0); 

    self->priv->adjustment_signals[SIGNAL_STEP_DOWN] =
	    g_signal_connect_data           ((gpointer)self->priv->adjustment,
                                             "step_down", 
                                             G_CALLBACK(on_adjustment_step_down), 
                                             self, 
                                             NULL, 
                                             0); 

}

void
chroma_list_set_drop_mode   (ChromaList		*self,
			     ChromaListDropMode  mode)
{
    g_return_if_fail(CHROMA_IS_LIST(self));

    self->priv->drop_mode = mode;
}

gint
chroma_list_get_row_height    (ChromaList     *self)
{
    g_return_val_if_fail(CHROMA_IS_LIST(self), -1);

    return self->priv->row_height;
}

void
chroma_list_set_cursor_position	    (ChromaList		*self,
				     gint		 offset_)
{
    GtkTreePath		  *path;

    path = gtk_tree_path_new_from_indices (offset_, -1); 
    chroma_list_selection_unselect_all (CHROMA_LIST_SELECTION(self));
  
    if (self->priv->last_pressed_row) gtk_tree_row_reference_free (self->priv->last_pressed_row);
    self->priv->last_pressed_row = gtk_tree_row_reference_new (self->priv->model, path);
    chroma_list_selection_path_set_selected (CHROMA_LIST_SELECTION(self), path, TRUE);
    gtk_tree_path_free (path);

    g_signal_emit (self, signals[CHROMA_LIST_SIGNAL_CURSOR_POSITION], 0, offset_); 
}

void
chroma_list_scroll_to_offset  (ChromaList     *self,
			       gint	       offset)
{
      gdouble position, total_area, scroll_area;

      g_return_if_fail (CHROMA_IS_LIST (self));

      total_area  = ((chroma_list_get_length (self) - chroma_list_get_n_item_display (self)) * self->priv->row_height);
      scroll_area = offset * self->priv->row_height; 

      position = scroll_area / total_area; 

      if (position < 0) position = 0;
      if (position > 1) position = 1;

      if (offset < 0) offset = 0;
      if (offset > (chroma_list_get_length (self) - chroma_list_get_n_item_display (self)))
          offset = (chroma_list_get_length (self) - chroma_list_get_n_item_display (self));

      g_object_set (G_OBJECT(self->priv->adjustment), "position_noemit", position, NULL);

      if (offset > 0)
	{
	    g_object_set (G_OBJECT(self), "offset", offset, NULL);
	}
}

/* GObject stuff */
static void
chroma_list_set_property (GObject	*object,
                          guint		 property_id,
                          const GValue	*value,
                          GParamSpec    *pspec)
{
    ChromaList *self = (ChromaList *) object;

    switch (property_id) {

    case CHROMA_LIST_PROP_OFFSET:
      {
	    gint *rows = NULL;
	    gint  delta;

	    if (!self->priv->model) return;

	    delta = g_value_get_int(value) - self->priv->offset; 
	    if (!delta) return;

	    if (abs(delta) >= chroma_list_get_n_item_display (self))
              {
                self->priv->offset = g_value_get_int(value);

		cl_process_cells (self, rows, self->priv->all_columns_mask);
		// gtk_widget_queue_draw (GTK_WIDGET (self));
		break;
	      }

	    /* If the delta is positive, we need to move a lower part upwards */
	    if (delta > 0)
              {
		    gint  start,
			  height,
			  n;
		  
		    start  = abs(delta);
		    height = chroma_list_get_n_item_display(self)-start;

		    gdk_draw_drawable (self->priv->pm_backingstore,

				       GTK_WIDGET (self)->style->fg_gc[GTK_STATE_NORMAL],
				       self->priv->pm_backingstore,

				       0, (start  * self->priv->row_height) + self->priv->header_size,
				       0, self->priv->header_size,
				      -1, (height * self->priv->row_height));

		    self->priv->offset = g_value_get_int(value);

		    rows = g_new0 (gint,start+1);
	
		    for (n = 0; n < start; n++)
		      {
			rows[n] = (chroma_list_get_n_item_display(self)-1) + self->priv->offset - n; 
	              }
		    rows[start] = -1; 
	      }
	    else if (delta < 0)
              {
		    gint  start,
			  height,
			  n;

		    start  = abs(delta);
		    height = chroma_list_get_n_item_display(self)-start;

		    gdk_draw_drawable (self->priv->pm_backingstore,

				       GTK_WIDGET (self)->style->fg_gc[GTK_STATE_NORMAL],
				       self->priv->pm_backingstore,

				       0, self->priv->header_size,
				       0, start  * self->priv->row_height + self->priv->header_size,
				      -1, height * self->priv->row_height);

		    self->priv->offset = g_value_get_int(value);

		    rows = g_new0 (gint,start+1);

		    for (n = 0; n < start; n++)
		      {
			rows[n] = n + self->priv->offset;
	              }
		    rows[start] = -1; 
	      }

	    cl_process_cells (self, rows, self->priv->all_columns_mask);
	    gtk_widget_queue_draw (GTK_WIDGET (self));
	    gdk_flush ();
        }
        break;

      case CHROMA_LIST_PROP_SHOW_HEADERS:
	{
          self->priv->show_headers = g_value_get_boolean (value);

	  if (self->priv->show_headers)
	    self->priv->header_size = self->priv->row_height + 7;
          else
	    self->priv->header_size = 0; 

	  gtk_widget_queue_draw (GTK_WIDGET (self));
	}
	break;

    default:
        /* We don't have any other property... */
        g_assert(FALSE);
        break;
    }
}

static void
chroma_list_get_property(GObject    *object,
                         guint	     property_id,
                         GValue	    *value,
			 GParamSpec *pspec)
{
    switch (property_id)
    {

    case CHROMA_LIST_PROP_OFFSET:
        {
	    g_value_set_int (value, CHROMA_LIST(object)->priv->offset);
	}
    break;
 

    case CHROMA_LIST_PROP_N_ITEMS:
        {
            if (CHROMA_LIST(object)->priv->model == NULL)
                g_value_set_int(value, (gint) 0);
            else
                g_value_set_int(value, gtk_tree_model_iter_n_children (CHROMA_LIST(object)->priv->model, NULL));
        }
        break;

    case CHROMA_LIST_PROP_RESERVED_SPACE:
        {
            g_value_set_int(value, CHROMA_LIST(object)->priv->reserved_space);
        }
        break;

    case CHROMA_LIST_PROP_ROW_HEIGHT:
        {
            g_value_set_int(value, CHROMA_LIST(object)->priv->row_height);
        }
        break;

    case CHROMA_LIST_PROP_SHOW_HEADERS:
	{
          g_value_set_boolean (value, CHROMA_LIST(object)->priv->show_headers);
	}
	break;

    default:
        /* We don't have any other property... */
        g_assert(FALSE);
        break;
    }
}

static void
chroma_list_dispose(GObject * obj)
{
    ChromaList *self = (ChromaList *) obj;

    if (self->priv->dispose_has_run) return; 

    self->priv->dispose_has_run = TRUE;
    if (self->priv->cs_selection) cairo_surface_destroy (self->priv->cs_selection);
}

static void
chroma_list_finalize(GObject * obj)
{
    ChromaList *self = (ChromaList *) obj;

    /*
     * Here, complete object destruction.
     * You might not need to do much...
     */
    g_free(self->priv);
}

static GObject *
chroma_list_constructor(GType type,
                        guint n_construct_properties,
                        GObjectConstructParam * construct_properties)
{
    GObject *obj;

    {
        /* Invoke parent constructor. */
        ChromaListClass *klass;
        GObjectClass *parent_class;
        klass = CHROMA_LIST_CLASS(g_type_class_peek(CHROMA_TYPE_LIST));
        parent_class = G_OBJECT_CLASS(g_type_class_peek_parent(klass));
        obj = parent_class->constructor(type,
                                        n_construct_properties,
                                        construct_properties);
    }

    /* do stuff. */

    return obj;
}

static void
chroma_list_class_init (ChromaListClass *class)
{
    GObjectClass    *gobject_class = G_OBJECT_CLASS (class);
    GtkWidgetClass  *widget_class  = GTK_WIDGET_CLASS (class);

    widget_class->realize		= chroma_list_realize;
    widget_class->size_allocate		= chroma_list_size_allocate;

    widget_class->drag_leave		= chroma_list_drag_leave;
    widget_class->drag_motion		= chroma_list_drag_motion;
    widget_class->drag_end		= chroma_list_drag_end;
    widget_class->drag_data_received	= chroma_list_drag_data_received;

    widget_class->button_press_event	= chroma_list_button_press; 
    widget_class->button_release_event	= chroma_list_button_release;
    widget_class->key_press_event	= chroma_list_key_press;

    widget_class->scroll_event		= chroma_list_scroll;

    widget_class->motion_notify_event	= chroma_list_motion;
    widget_class->expose_event		= chroma_list_expose;
    widget_class->leave_notify_event	= chroma_list_leave_notify;
    widget_class->configure_event	= chroma_list_configure;

    gobject_class->set_property		= chroma_list_set_property;
    gobject_class->get_property		= chroma_list_get_property;
    gobject_class->dispose		= chroma_list_dispose;
    gobject_class->finalize		= chroma_list_finalize;
    gobject_class->constructor		= chroma_list_constructor;

    g_object_class_install_property (gobject_class,
                                     CHROMA_LIST_PROP_SHOW_HEADERS, g_param_spec_boolean ("show_headers",
							      				  "show_headers",
											  "Show Headers?", TRUE, G_PARAM_READABLE));

    g_object_class_install_property (gobject_class,
                                     CHROMA_LIST_PROP_RESERVED_SPACE, g_param_spec_int("reserved_space",
										       "reserved_space",
							                               "reserved_space", 0, G_MAXINT, 0, G_PARAM_READABLE));

    g_object_class_install_property (gobject_class,
                                     CHROMA_LIST_PROP_OFFSET, g_param_spec_int("offset",
									       "offset",
									       "offset", 0, G_MAXINT, 0, G_PARAM_READWRITE));

    g_object_class_install_property (gobject_class,
                                     CHROMA_LIST_PROP_N_ITEMS, g_param_spec_int("n_items",
									        "n_items",
									        "n_items", 0, G_MAXINT, 0, G_PARAM_READABLE));


    g_object_class_install_property (gobject_class,
                                    CHROMA_LIST_PROP_ROW_HEIGHT, g_param_spec_int("row_height",
									          "row_height",
										  "row_height", 0, G_MAXINT, 0, G_PARAM_READABLE));

    signals[CHROMA_LIST_SIGNAL_CURSOR_POSITION] =
        g_signal_new("cursor-position",
                     G_OBJECT_CLASS_TYPE(gobject_class),
                     G_SIGNAL_RUN_FIRST,
                     G_STRUCT_OFFSET(ChromaListClass, cursor_position),
                     NULL, NULL,
                     g_cclosure_marshal_VOID__INT, G_TYPE_NONE, 1, G_TYPE_INT);

    signals[CHROMA_LIST_SIGNAL_ROW_ACTIVATED] =
        g_signal_new("row_activated",
                     G_OBJECT_CLASS_TYPE(gobject_class),
                     G_SIGNAL_RUN_FIRST,
                     G_STRUCT_OFFSET(ChromaListClass, row_activated),
                     NULL, NULL,
                     g_cclosure_marshal_VOID__INT,
                     G_TYPE_NONE, 1, G_TYPE_INT);

    signals[CHROMA_LIST_SIGNAL_DND] =
        g_signal_new("dnd",
                     G_OBJECT_CLASS_TYPE(gobject_class),
                     G_SIGNAL_RUN_FIRST,
                     G_STRUCT_OFFSET(ChromaListClass, dnd),
                     NULL, NULL,
                     libchroma_VOID__INT_BOXED,
                     G_TYPE_NONE, 2, G_TYPE_INT, G_TYPE_STRV);

}

static void
on_list_selection_path_selected	    (ChromaListSelectionInterface   *selection,
				     GtkTreePath		    *path,
				     gpointer			     data)
{
      gint *rows;
      gint  position;

      position = gtk_tree_path_get_indices (path)[0];

      if ((position < CHROMA_LIST(data)->priv->offset) ||
	  (position > (CHROMA_LIST(data)->priv->offset+(chroma_list_get_n_item_display (CHROMA_LIST(data)))))) return;

      rows = g_new0 (gint,2);
      rows[0] = position;
      rows[1] = NONE; 

      cl_process_cells (CHROMA_LIST (data), rows, CHROMA_LIST(data)->priv->all_columns_mask);
}

static void
on_list_selection_path_unselected   (ChromaListSelectionInterface   *selection,
				     GtkTreePath		    *path,
				     gpointer			     data)
{
      gint *rows;
      gint  position;

      position = gtk_tree_path_get_indices (path)[0];
  
      if ((position < CHROMA_LIST(data)->priv->offset) ||
	  (position > CHROMA_LIST(data)->priv->offset+(chroma_list_get_n_item_display (CHROMA_LIST(data))))) return;

      rows = g_new0 (gint,2);
      rows[0] = position;
      rows[1] = NONE; 

      cl_process_cells (CHROMA_LIST (data), rows, CHROMA_LIST(data)->priv->all_columns_mask);
}

static void
chroma_list_init (ChromaList *self)
{
    self->priv = g_new(ChromaListPrivate, 1);
    self->priv->dispose_has_run = FALSE;

    self->priv->model	  = NULL;

    self->priv->adjustment	  = NULL;

    self->priv->use_numbers	  = FALSE;

    self->priv->offset	  = 0;
    self->priv->n_columns	  = 0;

    self->priv->row_height	  = 0; 
    self->priv->reserved_space = 0;

    self->priv->cellspacing = 4;

    self->priv->selection_mode = GTK_SELECTION_MULTIPLE;
    self->priv->drop_mode	  = CHROMA_LIST_DROP_ROWS;

    self->priv->active_row_drag	= FALSE;
    self->priv->active_column_resize = FALSE;
    self->priv->grabbed_position	= 0;
    self->priv->button_press		= FALSE;

    self->priv->drop	    = FALSE;
    self->priv->drop_row = NONE;

    self->priv->scroll_timeout_source_id = 0;

    self->priv->pm_backingstore  = NULL;
    self->priv->cs_selection	    = NULL;

    self->priv->columns	 = NULL;
    self->priv->column_widths = NULL;

    self->priv->number_column = chroma_list_column_new();		

    self->priv->show_headers = TRUE;

    self->priv->sort_type = GTK_SORT_ASCENDING;
    self->priv->sort_column_id = GTK_TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID; 
    self->priv->sort_trigger = 0;

    self->priv->selected_rows    = NULL;
    self->priv->last_pressed_row = NULL; 
    self->priv->pixel_widths	    = NULL; 

    self->priv->adjactent_l      = NONE;
    self->priv->adjactent_r      = NONE;
    self->priv->resize_position  = NONE;

    cl_initialize_metrics (self);
    cl_drag_dest_set (GTK_WIDGET (self));

    g_signal_connect (CHROMA_LIST_SELECTION (self),
		      "path-selected",
		      G_CALLBACK (on_list_selection_path_selected),
		      self);

    g_signal_connect (CHROMA_LIST_SELECTION (self),
		      "path-unselected",
		      G_CALLBACK (on_list_selection_path_unselected),
		      self);
}

void
chroma_list_size (ChromaList *widget_list,
		       gint            width,
		       gint            height)
{
  g_return_if_fail (CHROMA_IS_LIST (widget_list));

  GTK_WIDGET (widget_list)->requisition.width = width;
  GTK_WIDGET (widget_list)->requisition.height = height;

  gtk_widget_queue_resize (GTK_WIDGET (widget_list));
}

static void
chroma_list_realize (GtkWidget *widget)
{
  ChromaList *widget_list;
  GdkWindowAttr attributes;
  gint attributes_mask;

  g_return_if_fail (CHROMA_IS_LIST (widget));

  widget_list = CHROMA_LIST (widget);
  GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED | GTK_CAN_FOCUS);

  attributes.window_type = GDK_WINDOW_CHILD;
  attributes.x = widget->allocation.x;
  attributes.y = widget->allocation.y;
  attributes.width = widget->allocation.width;
  attributes.height = widget->allocation.height;
  attributes.wclass = GDK_INPUT_OUTPUT;
  attributes.visual = gtk_widget_get_visual (widget);
  attributes.colormap = gtk_widget_get_colormap (widget);
  attributes.event_mask = gtk_widget_get_events (widget) | GDK_EXPOSURE_MASK | GDK_ALL_EVENTS_MASK;

  attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;

  widget->window = gdk_window_new (gtk_widget_get_parent_window (widget), &attributes, attributes_mask);
  gdk_window_set_user_data (widget->window, widget_list);

  widget->style = gtk_style_attach (widget->style, widget->window);
  gtk_style_set_background (widget->style, widget->window, GTK_STATE_NORMAL);

  chroma_list_send_configure (CHROMA_LIST (widget));
}

static void
chroma_list_size_allocate (GtkWidget     *widget,
				GtkAllocation *allocation)
{
  g_return_if_fail (CHROMA_IS_LIST (widget));
  g_return_if_fail (allocation != NULL);

  widget->allocation = *allocation;

  if (GTK_WIDGET_REALIZED (widget))
    {
      gdk_window_move_resize (widget->window,
			      allocation->x, allocation->y,
			      allocation->width, allocation->height);

      chroma_list_send_configure (CHROMA_LIST (widget));
    }
}

static void
chroma_list_send_configure (ChromaList *self)
{
  GtkWidget *widget;
  GdkEvent *event = gdk_event_new (GDK_CONFIGURE);

  widget = GTK_WIDGET (self);

  event->configure.window = g_object_ref (widget->window);
  event->configure.send_event = TRUE;
  event->configure.x = widget->allocation.x;
  event->configure.y = widget->allocation.y;
  event->configure.width = widget->allocation.width;
  event->configure.height = widget->allocation.height;

  gtk_widget_event (widget, event);
  gdk_event_free (event);
}

static gboolean
list_selection_path_get_selected (ChromaListSelectionInterface *self, 
				  GtkTreePath		       *path)
{
    g_return_val_if_fail (CHROMA_IS_LIST(self), FALSE);

    GtkTreeModel	  *model;
    GtkTreeRowReference	  *rref;
    GList		  *iter_list;
    gboolean		   found = FALSE;

    g_return_val_if_fail (CHROMA_IS_LIST_SELECTION(self), FALSE);

    model = CHROMA_LIST(self)->priv->model;
    rref = gtk_tree_row_reference_new (model, path);

    iter_list = CHROMA_LIST(self)->priv->selected_rows;

    for ( ; iter_list; iter_list = iter_list->next)
      {
	  GtkTreePath		*path_cmp;
	  GtkTreeRowReference   *rref_cmp;

	  rref_cmp = (GtkTreeRowReference *)iter_list->data; 
	  if (!rref_cmp) continue;
	  path_cmp = gtk_tree_row_reference_get_path (rref_cmp);
	  if (!path_cmp) continue;

	  if (!gtk_tree_path_compare (path, path_cmp))
	    {
		gtk_tree_path_free (path_cmp);
		found = TRUE;
		break;
	    }
      }

    gtk_tree_row_reference_free (rref);

    return found;
}

static gboolean
list_selection_iter_get_selected (ChromaListSelectionInterface *self, 
				  GtkTreeIter		       *iter)
{
    g_return_val_if_fail (CHROMA_IS_LIST(self), FALSE);

    GtkTreePath		  *path;
    GtkTreeModel	  *model;
    GtkTreeRowReference	  *rref;
    GList		  *iter_list;
    gboolean		   found = FALSE;

    g_return_val_if_fail (CHROMA_IS_LIST_SELECTION(self), FALSE);

    model = CHROMA_LIST(self)->priv->model;
    path = gtk_tree_model_get_path (model, iter);
    rref = gtk_tree_row_reference_new (model, path);

    iter_list = CHROMA_LIST(self)->priv->selected_rows;

    for ( ; iter_list; iter_list = iter_list->next)
      {
	  GtkTreePath		*path_cmp;
	  GtkTreeRowReference   *rref_cmp;

	  rref_cmp = (GtkTreeRowReference *)iter_list->data; 
	  if (!rref_cmp) continue;
	  path_cmp = gtk_tree_row_reference_get_path (rref_cmp);
	  if (!path_cmp) continue;

	  if (!gtk_tree_path_compare (path, path_cmp))
	    {
		gtk_tree_path_free (path_cmp);
		found = TRUE;
		break;
	    }
      }

    gtk_tree_path_free (path);
    gtk_tree_row_reference_free (rref);

    return found;
}

static void
list_selection_iter_set_selected (ChromaListSelectionInterface *self, 
				  GtkTreeIter		       *iter, 
				  gboolean		        selected)
{
    GtkTreePath		  *path;
    GtkTreeModel	  *model;
    GtkTreeRowReference	  *rref;

    g_return_if_fail (CHROMA_IS_LIST_SELECTION(self));

    model = CHROMA_LIST(self)->priv->model;
    path = gtk_tree_model_get_path (model, iter);
    rref = gtk_tree_row_reference_new (model, path);

    if (selected)
      {
	  if (!chroma_list_selection_iter_get_selected (self, iter))
	    {
		CHROMA_LIST(self)->priv->selected_rows = g_list_append (CHROMA_LIST(self)->priv->selected_rows, rref);
		chroma_list_selection_path_selected (self, path);
	    }
      }
    else
      {
	  GList *iter_list;
	  iter_list = CHROMA_LIST(self)->priv->selected_rows;

	  for ( ; iter_list; iter_list = iter_list->next)
	    {
		GtkTreePath	      *path_cmp;
		GtkTreeRowReference   *rref_cmp;

		rref_cmp = (GtkTreeRowReference *)iter_list->data; 
		if (!rref_cmp) continue;
		path_cmp = gtk_tree_row_reference_get_path (rref_cmp);
		if (!path_cmp) continue;

		if (!gtk_tree_path_compare (path, path_cmp))
		  {
		      CHROMA_LIST(self)->priv->selected_rows = g_list_remove_link (CHROMA_LIST(self)->priv->selected_rows, iter_list);
		      gtk_tree_path_free (path_cmp);
		      gtk_tree_row_reference_free (rref_cmp);
		      chroma_list_selection_path_unselected (self, path);
		      break;
		  }
	    }

	gtk_tree_row_reference_free (rref);
      }

    gtk_tree_path_free (path);
}

static void
list_selection_path_set_selected (ChromaListSelectionInterface *self, 
				  GtkTreePath		       *path, 
				  gboolean		        selected)
{
    GtkTreeModel	  *model;
    GtkTreeRowReference	  *rref;

    g_return_if_fail (CHROMA_IS_LIST_SELECTION(self));

    model = CHROMA_LIST(self)->priv->model;
    rref = gtk_tree_row_reference_new (model, path);

    if (selected)
      {
	  if (!chroma_list_selection_path_get_selected (self, path))
	    {
	      CHROMA_LIST(self)->priv->selected_rows = g_list_append (CHROMA_LIST(self)->priv->selected_rows, rref);
	      chroma_list_selection_path_selected (self, path);
	    }
      }
    else
      {
	  GList *iter_list;
	  iter_list = CHROMA_LIST(self)->priv->selected_rows;

	  for ( ; iter_list; iter_list = iter_list->next)
	    {
		GtkTreePath	      *path_cmp;
		GtkTreeRowReference   *rref_cmp;

		rref_cmp = (GtkTreeRowReference *)iter_list->data; 
		if (!rref_cmp) continue;
		path_cmp = gtk_tree_row_reference_get_path (rref_cmp);
		if (!path_cmp) continue;

		if (!gtk_tree_path_compare (path, path_cmp))
		  {
		      CHROMA_LIST(self)->priv->selected_rows = g_list_remove_link (CHROMA_LIST(self)->priv->selected_rows, iter_list);
		      gtk_tree_path_free (path_cmp);
		      gtk_tree_row_reference_free (rref_cmp);
		      chroma_list_selection_path_unselected (self, path);
		      break;
		  }
	    }

	gtk_tree_row_reference_free (rref);
      }
}

static void
list_selection_unselect_all (ChromaListSelectionInterface *self) 
{
    GList *iter_list;
    GList *path_list;

    iter_list = CHROMA_LIST(self)->priv->selected_rows;
    path_list = NULL;

    for ( ; iter_list; iter_list = iter_list->next)
      {
	  GtkTreeRowReference   *rref;
	  GtkTreePath		*path;

	  rref = (GtkTreeRowReference *)iter_list->data; 
	  path = gtk_tree_row_reference_get_path (rref);
	  if (path) path_list = g_list_append (path_list, path);
      }

    for ( ; path_list; path_list = path_list->next)
      {
	  chroma_list_selection_path_set_selected (self, (GtkTreePath*) path_list->data, FALSE);
      }
    g_list_foreach (g_list_first(path_list), (GFunc)gtk_tree_path_free, NULL);

    g_list_foreach (CHROMA_LIST(self)->priv->selected_rows, (GFunc)gtk_tree_row_reference_free, NULL);
    CHROMA_LIST(self)->priv->selected_rows = NULL;
}

static void
list_selection_select_all   (ChromaListSelectionInterface *self) 
{
    GtkTreeIter	iter;
    gboolean	proceed;    

    proceed = gtk_tree_model_get_iter_first (CHROMA_LIST(self)->priv->model, &iter);

    while (proceed)
      {
	  chroma_list_selection_iter_set_selected (self, &iter, TRUE);
	  proceed = gtk_tree_model_iter_next (CHROMA_LIST(self)->priv->model, &iter);
      }
}

static GList*
list_selection_get_selected_paths (ChromaListSelectionInterface *self) 
{
    g_return_val_if_fail (CHROMA_IS_LIST(self), FALSE);

    GList		  *iter_list;
    GList		  *paths = NULL;

    g_return_val_if_fail (CHROMA_IS_LIST_SELECTION(self), FALSE);

    iter_list = CHROMA_LIST(self)->priv->selected_rows;

    for ( ; iter_list; iter_list = iter_list->next)
      {
	  GtkTreePath		*path;
	  GtkTreeRowReference   *rref;

	  rref = (GtkTreeRowReference *)iter_list->data; 
	  if (!rref) continue;
	  path = gtk_tree_row_reference_get_path (rref);
	  if (!path) continue;

	  paths = g_list_append (paths, path);
      }

    return paths;
}

static void
list_selection_remove_selected (ChromaListSelectionInterface *self) 
{
    GtkTreeModel	  *model;
    GList		  *iter_list;

    g_return_if_fail (CHROMA_IS_LIST_SELECTION(self));

    iter_list = CHROMA_LIST(self)->priv->selected_rows;
    model = CHROMA_LIST(self)->priv->model;

    for ( ; iter_list; iter_list = iter_list->next)
      {
	  GtkTreePath		*path;
	  GtkTreeRowReference   *rref;
	  GtkTreeIter		 iter;

	  rref = (GtkTreeRowReference *)iter_list->data; 
	  if (!rref) continue;
	  path = gtk_tree_row_reference_get_path (rref);
	  if (!path) continue;

	  gtk_tree_model_get_iter (model, &iter, path); 
	  gtk_list_store_remove (GTK_LIST_STORE(model), &iter);
      }

    g_list_foreach (CHROMA_LIST(self)->priv->selected_rows, (GFunc)gtk_tree_row_reference_free, NULL);
    CHROMA_LIST(self)->priv->selected_rows = NULL;
}

static gboolean
list_selection_multiple_selected (ChromaListSelectionInterface *self) 
{
    g_return_val_if_fail (CHROMA_IS_LIST_SELECTION(self), FALSE);

    if (CHROMA_LIST(self)->priv->selected_rows == NULL) return FALSE;

    if (CHROMA_LIST(self)->priv->selected_rows->next != NULL) return TRUE;

    return FALSE;
}

static void
chroma_list_selection_init (ChromaListSelectionInterface *iface)
{
    iface->get_selected_paths = list_selection_get_selected_paths; 

    iface->iter_get_selected  = list_selection_iter_get_selected; 
    iface->iter_set_selected  = list_selection_iter_set_selected; 

    iface->path_get_selected  = list_selection_path_get_selected; 
    iface->path_set_selected  = list_selection_path_set_selected; 

    iface->select_all	      = list_selection_select_all; 
    iface->unselect_all	      = list_selection_unselect_all; 

    iface->remove_selected    = list_selection_remove_selected; 

    iface->multiple_selected  = list_selection_multiple_selected; 
}

GType
chroma_list_get_type (void)
{
    static GType list_type = 0;

    if (!list_type)
     {
         static const GTypeInfo list_info = {
             sizeof (ChromaListClass),
             NULL,
             NULL,
             (GClassInitFunc) chroma_list_class_init,
             NULL,
             NULL,
             sizeof (ChromaList),
             0,
             (GInstanceInitFunc) chroma_list_init,
         };

         static const GInterfaceInfo list_selection_info = {
             (GInterfaceInitFunc) chroma_list_selection_init,
             NULL,
             NULL
         };

         list_type =
             g_type_register_static (GTK_TYPE_WIDGET, "ChromaList",
                                     &list_info, 0);

         g_type_add_interface_static (list_type,
                                      CHROMA_TYPE_LIST_SELECTION, 
				      &list_selection_info);

     }

    return list_type;
}
