/*  libhrel - Holyrel library

    Copyright (C) 2005 Chong Kai Xiong <descender@phreaker.net>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the
    Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA  02111-1307  USA.
*/

#include <glib.h>
#include <glib-object.h>
#include <libhrel/types.h>
#include <libhrel/relation.h>
#include <stdio.h>

typedef struct _Track Track;

struct _Track
{
  gchar *name;
  gchar *uri;
  gchar *artist;
  guint  count;
};

static HRelation *tracks;
static HRelation *sorted_tracks;
static HRelation *filtered_tracks;
static HRelation *artists;

static void
track_print (Track *track)
{
  printf ("Name: %s\n", track->name);
  printf ("Artist: %s\n", track->artist);
  printf ("Play count: %d\n", track->count);
  printf ("URI: %s\n", track->uri);
}

static void
track_free (Track *track)
{
  if (!track)
    return;

  g_free (track->name);
  g_free (track->uri);
  g_free (track->artist);
  g_free (track);
}


static void
tracks_on_insert_tuple (HRelation *relation,
                        HTuple    *tuple,
                        gpointer   data)
{
  printf ("Inserting tuple: ");
  h_tuple_print (tuple);
  printf ("\n");
}

static void
tracks_on_delete_tuple (HRelation *relation,
                        HTuple    *tuple,
                        gpointer   data)
{
  printf ("Deleting tuple: ");
  h_tuple_print (tuple);
  printf ("\n");
}

static void
tracks_init (void)
{
  tracks = h_relation_base_new ("track",
                                "track has name, uri, artist, play count",
                                "name"  , G_TYPE_STRING,
                                "uri"   , G_TYPE_STRING,
                                "artist", G_TYPE_STRING,
                                "count" , G_TYPE_INT,
                                NULL,
                                "uri",
                                NULL);

  sorted_tracks = h_relation_sort_by_attrs (tracks, "artist", "uri", NULL);
  h_relation_set_name (sorted_tracks, "sorted tracks");
  h_relation_set_description (sorted_tracks, "tracks sorted by artist");

  filtered_tracks = h_relation_restrict_attrs (tracks, "artist", "Ehren Starks", NULL);
  h_relation_set_name (filtered_tracks, "filtered tracks");
  h_relation_set_description (filtered_tracks, "tracks by Ehren Starks");

  artists = h_relation_project (tracks, "artist", NULL);
  h_relation_set_name (artists, "artist");
  h_relation_set_description (artists, "artists");

  g_signal_connect (tracks, "insert-tuple",
                    G_CALLBACK (tracks_on_insert_tuple),
                    NULL);

  g_signal_connect (tracks, "delete-tuple",
                    G_CALLBACK (tracks_on_delete_tuple),
                    NULL);
}

static void
tracks_destroy (void)
{
  h_relation_unref (tracks);
  h_relation_unref (sorted_tracks);
  h_relation_unref (filtered_tracks);
  h_relation_unref (artists);
}

guint
tracks_get_count (void)
{
  return h_relation_get_n_tuples (tracks);
}

static void
tracks_add_track (Track *track)
{
  h_relation_base_insert_tuplev ((HRelationBase *) tracks,
                                 "name"  , track->name,
                                 "uri"   , track->uri,
                                 "artist", track->artist,
                                 "count" , track->count);
}

static void
tracks_delete_track_with_uri (const gchar *uri)
{
  h_relation_base_delete_tuplev ((HRelationBase *) tracks,
                                 "uri", uri);
}

static Track *
tracks_get_track_from_uri (const gchar *uri)
{
  HTuple *tuple;

  tuple = h_relation_base_get_tuplev ((HRelationBase *) tracks,
                                      "uri", uri);

  if (tuple)
    {
      /* marshal tuple into a Track */

      Track *track = g_new (Track, 1);

      track->name   = (gchar *) h_tuple_get_value (tuple, "name");
      track->uri    = (gchar *) h_tuple_get_value (tuple, "uri");
      track->artist = (gchar *) h_tuple_get_value (tuple, "artist");
      track->count  = GPOINTER_TO_UINT (h_tuple_get_value (tuple, "count"));

      h_tuple_free (tuple);
      return track;
    }

  return NULL;
}

static void
tracks_print (void)
{
  h_relation_print (tracks);
}

static void
tracks_print_sorted (void)
{
  h_relation_print (sorted_tracks);
}

static void
tracks_print_filtered (void)
{
  h_relation_print (filtered_tracks);
}

static void
artists_print (void)
{
  h_relation_print (artists);
}

static void
track_demo (void)
{
  static Track track_data[] =
  {
    {
      "Vincent",
      "file:///home/descender/Documents/music/Don McLean/Vincent.ogg",
      "Don McLean",
      1
    },

    {
      "Before I was gone",
      "file:///home/descender/Documents/music/Ehren Starks/Before I was gone.ogg",
      "Ehren Starks",
      3
    },

    {
      "New Year's Love",
      "file:///home/descender/Documents/music/Ehren Starks/New Year's Love.ogg",
      "Ehren Starks",
      5
    },

    {
      "Love, Me",
      "file:///home/descender/Documents/music/Collin Raye - Love Me.ogg",
      "Collin Raye",
      4
    },

    {
      "Eternal Snow",
      "file:///home/descender/Documents/music/soundtracks/Full Moon wo Sagashite/Eternal Snow.mp3",
      "Myco",
      6
    }
  };

  Track *track;
  gint   i;

  tracks_init ();

  for (i = 0; i < G_N_ELEMENTS (track_data); i++)
    tracks_add_track (&track_data[i]);

  printf ("\n"
          "Track list\n"
          "----------\n");
  tracks_print ();

  track = tracks_get_track_from_uri ("file:///home/descender/Documents/music/Don McLean/Vincent.ogg");
  track_print (track);
  track_free (track);

  track = tracks_get_track_from_uri ("file:///home/descender/Documents/music/Ehren Starks/Before I was gone.ogg");
  track_print (track);
  track_free (track);

  printf ("\nTrack count: %d\n", tracks_get_count ());
  tracks_delete_track_with_uri ("file:///home/descender/Documents/music/Don McLean/Vincent.ogg");
  printf ("\nNew track count: %d\n", tracks_get_count ());

  printf ("\n"
          "Artists\n"
          "-------\n");
  artists_print ();

  printf ("\n"
          "Sorted tracks relation\n"
          "----------------------\n");
  tracks_print_sorted ();

  printf ("\n"
          "Filtered tracks relation\n"
          "------------------------\n");
  tracks_print_filtered ();

  tracks_destroy ();
}

gint
main (void)
{
  h_type_init ();

  track_demo ();

  return 0;
}
