/*  libhrel - Holyrel library

    Copyright (C) 2005 Chong Kai Xiong <descender@phreaker.net>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the
    Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA  02111-1307  USA.
*/

#include <config.h>

#include "hash_index.h"
#include "hash_table.h"
#include "misc.h"

/**
 * HIndexFunc:
 * @tuple: a #HTuple key
 * @value: value
 * @user_data: user data
 **/

/**
 * HValueFunc:
 * @value: value
 * @user_data: user data
 **/

/**
 * HHashIndex:
 *
 * A hash table index.
 **/

struct _HHashIndex
{
  HHashTable *hash_table;
  GSList     *list;
};

/**
 * h_hash_index_new:
 * @key_destroy_func: key destructor, or %NULL if none
 * @value_destroy_func: value destructor, or %NULL if none
 * @...: name of an attribute, followed by another, and so on, then %NULL; if no names are listed, all attributes are used for hashing
 *
 * Creates a new hash index.
 *
 * Return value: a newly created #HHashIndex
 **/
HHashIndex *
h_hash_index_new (GDestroyNotify  key_destroy_func,
                  GDestroyNotify  value_destroy_func,
                  ...)
{
  va_list     args;
  HHashIndex *index;
  GSList     *list;

  va_start (args, value_destroy_func);
  list = h_va_list_to_slist (args);
  va_end (args);

  index = h_hash_index_new_l (key_destroy_func, value_destroy_func, list);
  g_slist_free (list);

  return index;
}

/**
 * h_hash_index_new_l:
 * @key_destroy_func: key destructor, or %NULL if none
 * @value_destroy_func: value destructor, or %NULL if none
 * @list: list of attribute names
 *
 * A variation of h_hash_index_new(), where the attribute name list is
 * stored in a #GSList.
 *
 * Return value: a newly created #HHashIndex
 **/
HHashIndex *
h_hash_index_new_l (GDestroyNotify  key_destroy_func,
                    GDestroyNotify  value_destroy_func,
                    GSList         *list)
{
  HHashIndex *index;

  index = g_new (HHashIndex, 1);

  index->list = h_string_list_copy (list);

  index->hash_table = h_hash_table_new_full ((HHashFunc) h_tuple_hash_elems_l,
                                             index->list,
                                             (HEqualFunc) h_tuple_equal_elems_l,
                                             index->list,
                                             key_destroy_func,
                                             value_destroy_func);

  return index;
}

/**
 * h_hash_index_free:
 * @index: #HHashIndex to free
 *
 * Frees @index.
 **/
void
h_hash_index_free (HHashIndex *index)
{
  g_return_if_fail (index != NULL);

  g_slist_foreach (index->list, (GFunc) g_free, NULL);
  g_slist_free (index->list);

  h_hash_table_destroy (index->hash_table);

  g_free (index);
}

/**
 * h_hash_index_get_attrs:
 * @index: a #HHashIndex
 *
 * Returns the list of attributes used for hashing.
 *
 * Return value: list of attribute names. This must not be freed.
 **/
GSList *
h_hash_index_get_attrs (HHashIndex *index)
{
  g_return_val_if_fail (index != NULL, NULL);

  return index->list;
}

/**
 * h_hash_index_lookup:
 * @index: a #HHashIndex
 * @tuple: a #HTuple key to lookup
 *
 * Looks up a key in @index. Note that this function cannot
 * distinguish between a key is not present and one which is present
 * and has the value %NULL.
 *
 * Return value: associated value, or %NULL if the key is not found
 **/
gpointer
h_hash_index_lookup (HHashIndex *index,
                     HTuple     *tuple)
{
  g_return_val_if_fail (index != NULL, NULL);

  return h_hash_table_lookup (index->hash_table, tuple);
}

/**
 * h_hash_index_insert:
 * @index: a #HHashIndex
 * @tuple: a #HTuple key to insert
 * @value: value associated with key
 *
 * Inserts a new key and value into @index. If the key already exists,
 * its current value is replaced with the new value. If
 * @value_destroy_func was supplied when creating @index, the old
 * value is freed using it. If @key_destroy_func was supplied when
 * creating @index, the passed key is freed using it.
 **/
void
h_hash_index_insert (HHashIndex *index,
                     HTuple     *tuple,
                     gpointer    value)
{
  g_return_if_fail (index != NULL);

  h_hash_table_insert (index->hash_table, tuple, value);
}

/**
 * h_hash_index_replace:
 * @index: a #HHashIndex
 * @tuple: a #HTuple key to insert
 * @value: value associated with key
 *
 * Inserts a new key and value into @index. This is a version of
 * h_hash_index_insert() which replaces the old key with the
 * passed key and calls @key_destroy_func on the former.
 **/
void
h_hash_index_replace (HHashIndex *index,
                      HTuple     *tuple,
                      gpointer    value)
{
  g_return_if_fail (index != NULL);

  h_hash_table_replace (index->hash_table, tuple, value);
}

/**
 * h_hash_index_remove:
 * @index: a #HHashIndex
 * @tuple: a #HTuple key to remove
 *
 * Removes a new key and its associated value from @index.
 *
 * Return value: TRUE if key was found and removed
 **/
gboolean
h_hash_index_remove (HHashIndex *index,
                     HTuple     *tuple)
{
  g_return_val_if_fail (index != NULL, FALSE);

  return h_hash_table_remove (index->hash_table, tuple);
}

/**
 * h_hash_index_get_size:
 * @index: a #HHashIndex
 *
 * Returns the number of elements in @index.
 *
 * Return value: number of elements
 **/
guint
h_hash_index_get_size (HHashIndex *index)
{
  g_return_val_if_fail (index != NULL, 0);

  return h_hash_table_size (index->hash_table);
}

/**
 * h_hash_index_foreach:
 * @index: a #HHashIndex
 * @func: function called for each key/value pair in @index
 * @user_data: custom data to pass to @func
 *
 * Performs a function over each key/value pair in @index.
 **/
void
h_hash_index_foreach (HHashIndex *index,
                      HIndexFunc  func,
                      gpointer    user_data)
{
  g_return_if_fail (index != NULL);

  h_hash_table_foreach (index->hash_table, (HHFunc) func, user_data);
}

typedef struct
{
  HValueFunc func;
  gpointer   data;
}
IterValueArgs;

static void
iter_value (HTuple        *key,
            gpointer       value,
            IterValueArgs *args)
{
  (*args->func) (value, args->data);
}

/**
 * h_hash_index_for_each_value:
 * @index: a #HHashIndex
 * @func: function called for each value in @index
 * @user_data: custom data to pass to @func
 *
 * Performs a function over each key/value pair in @index.
 **/
void
h_hash_index_for_each_value (HHashIndex *index,
                             HValueFunc  func,
                             gpointer    user_data)
{
  IterValueArgs args;

  g_return_if_fail (index != NULL);
  g_return_if_fail (func != NULL);

  args.func = func;
  args.data = user_data;

  h_hash_index_foreach (index, (HIndexFunc) iter_value, &args);
}
