/*  libhrel - Holyrel library
 *
 *  Copyright (C) 2005 Chong Kai Xiong <descender@phreaker.net>
 *
 *  HHashTable is an adaptation of GHashTable to allow users to pass
 *  custom data to the hash function and key comparator.
 */

/* GLIB - Library of useful routines for C programming
 * Copyright (C) 1995-1997  Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * Modified by the GLib Team and others 1997-2000.  See the AUTHORS
 * file for a list of people on the GLib Team.  See the ChangeLog
 * files for a list of changes.  These files are distributed with
 * GLib at ftp://ftp.gtk.org/pub/gtk/.
 */

#include <config.h>
#include "hash_table.h"

/* Note: these numbers should be prime to reduce collisions */
#define HASH_TABLE_MIN_SIZE 11
#define HASH_TABLE_MAX_SIZE 13845163


typedef struct _HHashNode HHashNode;

struct _HHashNode
{
  gpointer   key;
  gpointer   value;
  HHashNode *next;
};

struct _HHashTable
{
  gint             size;
  gint             nnodes;
  HHashNode      **nodes;
  HHashFunc        hash_func;
  gpointer         hash_func_data;
  HEqualFunc       key_equal_func;
  gpointer         key_equal_func_data;
  GDestroyNotify   key_destroy_func;
  GDestroyNotify   value_destroy_func;
};

#define H_HASH_TABLE_RESIZE(hash_table)                         \
   G_STMT_START {                                               \
     if ((hash_table->size >= 3 * hash_table->nnodes &&         \
          hash_table->size > HASH_TABLE_MIN_SIZE) ||            \
         (3 * hash_table->size <= hash_table->nnodes &&         \
          hash_table->size < HASH_TABLE_MAX_SIZE))              \
           h_hash_table_resize (hash_table);                    \
   } G_STMT_END

static void        h_hash_table_resize      (HHashTable     *hash_table);
static HHashNode **h_hash_table_lookup_node (HHashTable     *hash_table,
                                             gconstpointer   key);
static HHashNode  *h_hash_node_new          (gpointer        key,
                                             gpointer        value);
static void        h_hash_node_destroy      (HHashNode      *hash_node,
                                             GDestroyNotify  key_destroy_func,
                                             GDestroyNotify  value_destroy_func);
static void        h_hash_nodes_destroy     (HHashNode      *hash_node,
                                             GDestroyNotify  key_destroy_func,
                                             GDestroyNotify  value_destroy_func);

static guint h_hash_table_foreach_remove_or_steal (HHashTable     *hash_table,
                                                   GHRFunc         func,
                                                   gpointer        user_data,
                                                   gboolean        notify);

#ifndef DISABLE_MEM_POOLS
G_LOCK_DEFINE_STATIC (h_hash_global);

static GMemChunk *node_mem_chunk = NULL;
static HHashNode *node_free_list = NULL;
#endif

/**
 * h_hash_table_new:
 * @hash_func: a function to create a hash value from a key.
 *   Hash values are used to determine where keys are stored within the
 *   #HHashTable data structure.
 * @hash_func_data: user data passed to @hash_func.
 * @key_equal_func: a function to check two keys for equality.  This is
 *   used when looking up keys in the #HHashTable.
 * @key_equal_func_data: user data passed to @key_equal_func.
 *
 * Creates a new #HHashTable.
 *
 * Return value: a new #HHashTable.
 **/
HHashTable*
h_hash_table_new (HHashFunc    hash_func,
                  gpointer     hash_func_data,
                  HEqualFunc   key_equal_func,
                  gpointer     key_equal_func_data)
{
  return h_hash_table_new_full (hash_func, hash_func_data,
                                key_equal_func, key_equal_func_data,
                                NULL, NULL);
}

/**
 * h_hash_table_new_full:
 * @hash_func: a function to create a hash value from a key.
 * @hash_func_data: user data passed to @hash_func.
 * @key_equal_func: a function to check two keys for equality.
 * @key_equal_func_data: user data passed to @key_equal_func.
 * @key_destroy_func: a function to free the memory allocated for the key
 *   used when removing the entry from the #HHashTable or %NULL if you
 *   don't want to supply such a function.
 * @value_destroy_func: a function to free the memory allocated for the
 *   value used when removing the entry from the #HHashTable or %NULL if
 *   you don't want to supply such a function.
 *
 * Creates a new #HHashTable like h_hash_table_new() and allows to specify
 * functions to free the memory allocated for the key and value that get
 * called when removing the entry from the #HHashTable.
 *
 * Return value: a new #HHashTable.
 **/
HHashTable *
h_hash_table_new_full (HHashFunc       hash_func,
                       gpointer        hash_func_data,
                       HEqualFunc      key_equal_func,
                       gpointer        key_equal_func_data,
                       GDestroyNotify  key_destroy_func,
                       GDestroyNotify  value_destroy_func)
{
  HHashTable *hash_table;

  hash_table = g_new (HHashTable, 1);

  hash_table->size                = HASH_TABLE_MIN_SIZE;
  hash_table->nnodes              = 0;
  hash_table->hash_func           = hash_func;
  hash_table->hash_func_data      = hash_func_data;
  hash_table->key_equal_func      = key_equal_func;
  hash_table->key_equal_func_data = key_equal_func_data;
  hash_table->key_destroy_func    = key_destroy_func;
  hash_table->value_destroy_func  = value_destroy_func;

  hash_table->nodes = g_new0 (HHashNode *, hash_table->size);

  return hash_table;
}

/**
 * h_hash_table_destroy:
 * @hash_table: a #HHashTable.
 *
 * Destroys the #HHashTable. If keys and/or values are dynamically
 * allocated, you should either free them first or create the
 * #HHashTable using h_hash_table_new_full(). In the latter case the
 * destroy functions you supplied will be called on all keys and
 * values before destroying the #HHashTable.
 **/
void
h_hash_table_destroy (HHashTable *hash_table)
{
  guint i;

  g_return_if_fail (hash_table != NULL);

  for (i = 0; i < hash_table->size; i++)
    h_hash_nodes_destroy (hash_table->nodes[i],
                          hash_table->key_destroy_func,
                          hash_table->value_destroy_func);

  g_free (hash_table->nodes);
  g_free (hash_table);
}

static inline HHashNode**
h_hash_table_lookup_node (HHashTable    *hash_table,
                          gconstpointer  key)
{
  HHashNode **node;

  node = &hash_table->nodes
    [(*hash_table->hash_func) (key, hash_table->hash_func_data)
     % hash_table->size];

  while (*node && !(*hash_table->key_equal_func) ((*node)->key, key, hash_table->key_equal_func_data))
      node = &(*node)->next;

  return node;
}

/**
 * h_hash_table_lookup:
 * @hash_table: a #HHashTable.
 * @key: the key to look up.
 *
 * Looks up a key in a #HHashTable. Note that this function cannot
 * distinguish between a key that is not present and one which is present
 * and has the value %NULL. If you need this distinction, use
 * h_hash_table_lookup_extended().
 *
 * Return value: the associated value, or %NULL if the key is not found.
 **/
gpointer
h_hash_table_lookup (HHashTable   *hash_table,
                     gconstpointer key)
{
  HHashNode *node;

  g_return_val_if_fail (hash_table != NULL, NULL);

  node = *h_hash_table_lookup_node (hash_table, key);

  return node ? node->value : NULL;
}

/**
 * h_hash_table_lookup_extended:
 * @hash_table: a #HHashTable.
 * @lookup_key: the key to look up.
 * @orig_key: returns the original key.
 * @value: returns the value associated with the key.
 *
 * Looks up a key in the #HHashTable, returning the original key and the
 * associated value and a #gboolean which is %TRUE if the key was found. This
 * is useful if you need to free the memory allocated for the original key,
 * for example before calling h_hash_table_remove().
 *
 * Return value: %TRUE if the key was found in the #HHashTable.
 **/
gboolean
h_hash_table_lookup_extended (HHashTable    *hash_table,
                              gconstpointer  lookup_key,
                              gpointer      *orig_key,
                              gpointer      *value)
{
  HHashNode *node;

  g_return_val_if_fail (hash_table != NULL, FALSE);

  node = *h_hash_table_lookup_node (hash_table, lookup_key);

  if (node)
    {
      if (orig_key)
        *orig_key = node->key;
      if (value)
        *value = node->value;
      return TRUE;
    }
  else
    return FALSE;
}

/**
 * h_hash_table_insert:
 * @hash_table: a #HHashTable.
 * @key: a key to insert.
 * @value: the value to associate with the key.
 *
 * Inserts a new key and value into a #HHashTable.
 *
 * If the key already exists in the #HHashTable its current value is replaced
 * with the new value. If you supplied a @value_destroy_func when creating the
 * #HHashTable, the old value is freed using that function. If you supplied
 * a @key_destroy_func when creating the #HHashTable, the passed key is freed
 * using that function.
 **/
void
h_hash_table_insert (HHashTable *hash_table,
                     gpointer    key,
                     gpointer    value)
{
  HHashNode **node;

  g_return_if_fail (hash_table != NULL);

  node = h_hash_table_lookup_node (hash_table, key);

  if (*node)
    {
      /* do not reset node->key in this place, keeping
       * the old key is the intended behaviour.
       * h_hash_table_replace() can be used instead.
       */

      /* free the passed key */
      if (hash_table->key_destroy_func)
        hash_table->key_destroy_func (key);

      if (hash_table->value_destroy_func)
        hash_table->value_destroy_func ((*node)->value);

      (*node)->value = value;
    }
  else
    {
      *node = h_hash_node_new (key, value);
      hash_table->nnodes++;
      H_HASH_TABLE_RESIZE (hash_table);
    }
}

/**
 * h_hash_table_replace:
 * @hash_table: a #HHashTable.
 * @key: a key to insert.
 * @value: the value to associate with the key.
 *
 * Inserts a new key and value into a #HHashTable similar to
 * h_hash_table_insert(). The difference is that if the key already exists
 * in the #HHashTable, it gets replaced by the new key. If you supplied a
 * @value_destroy_func when creating the #HHashTable, the old value is freed
 * using that function. If you supplied a @key_destroy_func when creating the
 * #HHashTable, the old key is freed using that function.
 **/
void
h_hash_table_replace (HHashTable *hash_table,
                      gpointer    key,
                      gpointer    value)
{
  HHashNode **node;

  g_return_if_fail (hash_table != NULL);

  node = h_hash_table_lookup_node (hash_table, key);

  if (*node)
    {
      if (hash_table->key_destroy_func)
        hash_table->key_destroy_func ((*node)->key);

      if (hash_table->value_destroy_func)
        hash_table->value_destroy_func ((*node)->value);

      (*node)->key   = key;
      (*node)->value = value;
    }
  else
    {
      *node = h_hash_node_new (key, value);
      hash_table->nnodes++;
      H_HASH_TABLE_RESIZE (hash_table);
    }
}

/**
 * h_hash_table_remove:
 * @hash_table: a #HHashTable.
 * @key: the key to remove.
 *
 * Removes a key and its associated value from a #HHashTable.
 *
 * If the #HHashTable was created using h_hash_table_new_full(), the
 * key and value are freed using the supplied destroy functions, otherwise
 * you have to make sure that any dynamically allocated values are freed
 * yourself.
 *
 * Return value: %TRUE if the key was found and removed from the #HHashTable.
 **/
gboolean
h_hash_table_remove (HHashTable    *hash_table,
                     gconstpointer  key)
{
  HHashNode **node, *dest;

  g_return_val_if_fail (hash_table != NULL, FALSE);

  node = h_hash_table_lookup_node (hash_table, key);
  if (*node)
    {
      dest = *node;
      (*node) = dest->next;
      h_hash_node_destroy (dest,
                           hash_table->key_destroy_func,
                           hash_table->value_destroy_func);
      hash_table->nnodes--;

      H_HASH_TABLE_RESIZE (hash_table);

      return TRUE;
    }

  return FALSE;
}

/**
 * h_hash_table_steal:
 * @hash_table: a #HHashTable.
 * @key: the key to remove.
 *
 * Removes a key and its associated value from a #HHashTable without
 * calling the key and value destroy functions.
 *
 * Return value: %TRUE if the key was found and removed from the #HHashTable.
 **/
gboolean
h_hash_table_steal (HHashTable    *hash_table,
                    gconstpointer  key)
{
  HHashNode **node, *dest;

  g_return_val_if_fail (hash_table != NULL, FALSE);

  node = h_hash_table_lookup_node (hash_table, key);
  if (*node)
    {
      dest = *node;
      (*node) = dest->next;
      h_hash_node_destroy (dest, NULL, NULL);
      hash_table->nnodes--;

      H_HASH_TABLE_RESIZE (hash_table);

      return TRUE;
    }

  return FALSE;
}

/**
 * h_hash_table_foreach_remove:
 * @hash_table: a #HHashTable.
 * @func: the function to call for each key/value pair.
 * @user_data: user data to pass to the function.
 *
 * Calls the given function for each key/value pair in the #HHashTable.
 * If the function returns %TRUE, then the key/value pair is removed from the
 * #HHashTable. If you supplied key or value destroy functions when creating
 * the #HHashTable, they are used to free the memory allocated for the removed
 * keys and values.
 *
 * Return value: the number of key/value pairs removed.
 **/
guint
h_hash_table_foreach_remove (HHashTable *hash_table,
                             HHFunc      func,
                             gpointer    user_data)
{
  g_return_val_if_fail (hash_table != NULL, 0);
  g_return_val_if_fail (func != NULL, 0);

  return h_hash_table_foreach_remove_or_steal (hash_table, func, user_data, TRUE);
}

/**
 * h_hash_table_foreach_steal:
 * @hash_table: a #HHashTable.
 * @func: the function to call for each key/value pair.
 * @user_data: user data to pass to the function.
 *
 * Calls the given function for each key/value pair in the #HHashTable.
 * If the function returns %TRUE, then the key/value pair is removed from the
 * #HHashTable, but no key or value destroy functions are called.
 *
 * Return value: the number of key/value pairs removed.
 **/
guint
h_hash_table_foreach_steal (HHashTable *hash_table,
                            GHRFunc     func,
                            gpointer    user_data)
{
  g_return_val_if_fail (hash_table != NULL, 0);
  g_return_val_if_fail (func != NULL, 0);

  return h_hash_table_foreach_remove_or_steal (hash_table, func, user_data, FALSE);
}

static guint
h_hash_table_foreach_remove_or_steal (HHashTable *hash_table,
                                      GHRFunc     func,
                                      gpointer    user_data,
                                      gboolean    notify)
{
  HHashNode *node, *prev;
  guint i;
  guint deleted = 0;

  for (i = 0; i < hash_table->size; i++)
    {
    restart:

      prev = NULL;

      for (node = hash_table->nodes[i]; node; prev = node, node = node->next)
        {
          if ((* func) (node->key, node->value, user_data))
            {
              deleted += 1;

              hash_table->nnodes -= 1;

              if (prev)
                {
                  prev->next = node->next;
                  h_hash_node_destroy (node,
                                       notify ? hash_table->key_destroy_func : NULL,
                                       notify ? hash_table->value_destroy_func : NULL);
                  node = prev;
                }
              else
                {
                  hash_table->nodes[i] = node->next;
                  h_hash_node_destroy (node,
                                       notify ? hash_table->key_destroy_func : NULL,
                                       notify ? hash_table->value_destroy_func : NULL);
                  goto restart;
                }
            }
        }
    }

  H_HASH_TABLE_RESIZE (hash_table);

  return deleted;
}

/**
 * h_hash_table_foreach:
 * @hash_table: a #HHashTable.
 * @func: the function to call for each key/value pair.
 * @user_data: user data to pass to the function.
 *
 * Calls the given function for each of the key/value pairs in the
 * #HHashTable.  The function is passed the key and value of each
 * pair, and the given @user_data parameter.  The hash table may not
 * be modified while iterating over it (you can't add/remove
 * items). To remove all items matching a predicate, use
 * h_hash_table_foreach_remove().
 **/
void
h_hash_table_foreach (HHashTable *hash_table,
                      HHFunc      func,
                      gpointer    user_data)
{
  HHashNode *node;
  gint i;

  g_return_if_fail (hash_table != NULL);
  g_return_if_fail (func != NULL);

  for (i = 0; i < hash_table->size; i++)
    for (node = hash_table->nodes[i]; node; node = node->next)
      (* func) (node->key, node->value, user_data);
}

/**
 * h_hash_table_find:
 * @hash_table: a #HHashTable.
 * @predicate:  function to test the key/value pairs for a certain property.
 * @user_data:  user data to pass to the function.
 *
 * Calls the given function for key/value pairs in the #HHashTable until
 * @predicate returns %TRUE.  The function is passed the key and value of
 * each pair, and the given @user_data parameter. The hash table may not
 * be modified while iterating over it (you can't add/remove items).
 *
 * Return value: The value of the first key/value pair is returned, for which
 * func evaluates to %TRUE. If no pair with the requested property is found,
 * %NULL is returned.
 *
 * Since: 2.4
 **/
gpointer
h_hash_table_find (HHashTable      *hash_table,
                   GHRFunc          predicate,
                   gpointer         user_data)
{
  HHashNode *node;
  gint i;

  g_return_val_if_fail (hash_table != NULL, NULL);
  g_return_val_if_fail (predicate != NULL, NULL);

  for (i = 0; i < hash_table->size; i++)
    for (node = hash_table->nodes[i]; node; node = node->next)
      if (predicate (node->key, node->value, user_data))
        return node->value;
  return NULL;
}

/**
 * h_hash_table_size:
 * @hash_table: a #HHashTable.
 *
 * Returns the number of elements contained in the #HHashTable.
 *
 * Return value: the number of key/value pairs in the #HHashTable.
 **/
guint
h_hash_table_size (HHashTable *hash_table)
{
  g_return_val_if_fail (hash_table != NULL, 0);

  return hash_table->nnodes;
}

static void
h_hash_table_resize (HHashTable *hash_table)
{
  HHashNode **new_nodes;
  HHashNode *node;
  HHashNode *next;
  guint hash_val;
  gint new_size;
  gint i;

  new_size = g_spaced_primes_closest (hash_table->nnodes);
  new_size = CLAMP (new_size, HASH_TABLE_MIN_SIZE, HASH_TABLE_MAX_SIZE);

  new_nodes = g_new0 (HHashNode*, new_size);

  for (i = 0; i < hash_table->size; i++)
    for (node = hash_table->nodes[i]; node; node = next)
      {
        next = node->next;

        hash_val = (*hash_table->hash_func) (node->key, hash_table->hash_func_data) % new_size;

        node->next = new_nodes[hash_val];
        new_nodes[hash_val] = node;
      }

  g_free (hash_table->nodes);
  hash_table->nodes = new_nodes;
  hash_table->size = new_size;
}

static HHashNode*
h_hash_node_new (gpointer key,
                 gpointer value)
{
  HHashNode *hash_node;

#ifdef DISABLE_MEM_POOLS
  hash_node = g_new (HHashNode, 1);
#else
  G_LOCK (h_hash_global);
  if (node_free_list)
    {
      hash_node = node_free_list;
      node_free_list = node_free_list->next;
    }
  else
    {
      if (!node_mem_chunk)
        node_mem_chunk = g_mem_chunk_new ("hash node mem chunk",
                                          sizeof (HHashNode),
                                          1024, G_ALLOC_ONLY);

      hash_node = g_chunk_new (HHashNode, node_mem_chunk);
    }
  G_UNLOCK (h_hash_global);
#endif

  hash_node->key = key;
  hash_node->value = value;
  hash_node->next = NULL;

  return hash_node;
}

static void
h_hash_node_destroy (HHashNode      *hash_node,
                     GDestroyNotify  key_destroy_func,
                     GDestroyNotify  value_destroy_func)
{
  if (key_destroy_func)
    key_destroy_func (hash_node->key);
  if (value_destroy_func)
    value_destroy_func (hash_node->value);

#ifdef ENABLE_GC_FRIENDLY
  hash_node->key = NULL;
  hash_node->value = NULL;
#endif /* ENABLE_GC_FRIENDLY */

#ifdef DISABLE_MEM_POOLS
  g_free (hash_node);
#else
  G_LOCK (h_hash_global);
  hash_node->next = node_free_list;
  node_free_list = hash_node;
  G_UNLOCK (h_hash_global);
#endif
}

static void
h_hash_nodes_destroy (HHashNode *hash_node,
                      GFreeFunc  key_destroy_func,
                      GFreeFunc  value_destroy_func)
{
#ifdef DISABLE_MEM_POOLS
  while (hash_node)
    {
      HHashNode *next = hash_node->next;

      if (key_destroy_func)
        key_destroy_func (hash_node->key);
      if (value_destroy_func)
        value_destroy_func (hash_node->value);

      g_free (hash_node);
      hash_node = next;
    }
#else
  if (hash_node)
    {
      HHashNode *node = hash_node;

      while (node->next)
        {
          if (key_destroy_func)
            key_destroy_func (node->key);
          if (value_destroy_func)
            value_destroy_func (node->value);

#ifdef ENABLE_GC_FRIENDLY
          node->key = NULL;
          node->value = NULL;
#endif /* ENABLE_GC_FRIENDLY */

          node = node->next;
        }

      if (key_destroy_func)
        key_destroy_func (node->key);
      if (value_destroy_func)
        value_destroy_func (node->value);

#ifdef ENABLE_GC_FRIENDLY
      node->key = NULL;
      node->value = NULL;
#endif /* ENABLE_GC_FRIENDLY */

      G_LOCK (h_hash_global);
      node->next = node_free_list;
      node_free_list = hash_node;
      G_UNLOCK (h_hash_global);
    }
#endif
}
