/*  libhrel - Holyrel library

    Copyright (C) 2005 Chong Kai Xiong <descender@phreaker.net>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the
    Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA  02111-1307  USA.
*/

#include <config.h>

#include <glib.h>
#include <glib-object.h>
#include <glib/gi18n.h>
#include <stdarg.h>
#include <stdio.h>

#include "types.h"
#include "relation.h"

#define H_RELATION_GET_PRIVATE(obj) \
  G_TYPE_INSTANCE_GET_PRIVATE ((obj), H_TYPE_RELATION, HRelationPrivate)

typedef struct _HRelationPrivate HRelationPrivate;

struct _HRelationPrivate
{
  gchar *name;
  gchar *description;
};

/**
 * HAttribute:
 * @name: name
 * @type: type ID
 *
 * Holds the complete description of an attribute.
 **/

/**
 * HAttrFunc:
 * @attr: an #HAttribute
 * @user_data: user data
 *
 * #HAttribute function.
 **/

/**
 * HTupleFunc:
 * @tuple: a #HTuple
 * @user_data: user data
 *
 * #HTuple function.
 **/

/**
 * HRelation:
 *
 * Type representing a relation.
 **/
G_DEFINE_TYPE (HRelation, h_relation, G_TYPE_OBJECT)

static void h_relation_set_property (GObject      *object,
                                     guint         property,
                                     const GValue *value,
                                     GParamSpec   *spec);
static void h_relation_get_property (GObject      *object,
                                     guint         property,
                                     GValue       *value,
                                     GParamSpec   *spec);
static void h_relation_dispose      (GObject      *object);
static void h_relation_finalize     (GObject      *object);
static void h_relation_insert_tuple (HRelation    *self,
                                     HTuple       *tuple);
static void h_relation_delete_tuple (HRelation    *self,
                                     HTuple       *tuple);

enum
{
  PROP_NONE,
  PROP_NAME,
  PROP_DESCRIPTION,
  PROP_LAST
};

enum
{
  SIGNAL_INSERT_TUPLE,
  SIGNAL_DELETE_TUPLE,
  SIGNAL_LAST
};

static guint signals[SIGNAL_LAST];

static void
h_relation_class_init (HRelationClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (klass, sizeof (HRelationPrivate));

  object_class->set_property = h_relation_set_property;
  object_class->get_property = h_relation_get_property;
  object_class->dispose = h_relation_dispose;
  object_class->finalize = h_relation_finalize;

  klass->insert_tuple = h_relation_insert_tuple;
  klass->delete_tuple = h_relation_delete_tuple;

  g_object_class_install_property (object_class, PROP_NAME,
                                   g_param_spec_string ("name",
                                                        Q_("name"),
                                                        Q_("Name of relation"),
                                                        NULL,
                                                        G_PARAM_READABLE | G_PARAM_WRITABLE));

  g_object_class_install_property (object_class, PROP_DESCRIPTION,
                                   g_param_spec_string ("description",
                                                        Q_("description"),
                                                        Q_("Description of relation"),
                                                        NULL,
                                                        G_PARAM_READABLE | G_PARAM_WRITABLE));

  signals[SIGNAL_INSERT_TUPLE] = g_signal_new ("insert-tuple",
                                               H_TYPE_RELATION,
                                               G_SIGNAL_RUN_FIRST,
                                               G_STRUCT_OFFSET (HRelationClass, insert_tuple),
                                               NULL,
                                               NULL,
                                               g_cclosure_marshal_VOID__BOXED,
                                               G_TYPE_NONE,
                                               1,
                                               H_TYPE_TUPLE | G_SIGNAL_TYPE_STATIC_SCOPE);

  signals[SIGNAL_DELETE_TUPLE] = g_signal_new ("delete-tuple",
                                               H_TYPE_RELATION,
                                               G_SIGNAL_RUN_FIRST,
                                               G_STRUCT_OFFSET (HRelationClass, delete_tuple),
                                               NULL,
                                               NULL,
                                               g_cclosure_marshal_VOID__BOXED,
                                               G_TYPE_NONE,
                                               1,
                                               H_TYPE_TUPLE | G_SIGNAL_TYPE_STATIC_SCOPE);
}

static void
h_relation_init (HRelation *self)
{
  self->floating = TRUE;
}

static void
h_relation_dispose (GObject *object)
{
  G_OBJECT_CLASS (h_relation_parent_class)->dispose (object);
}

static void
h_relation_finalize (GObject *object)
{
  HRelationPrivate *private = H_RELATION_GET_PRIVATE (object);

  g_free (private->description);
  g_free (private->name);

  G_OBJECT_CLASS (h_relation_parent_class)->finalize (object);
}

static void
h_relation_set_property (GObject      *object,
                         guint         property,
                         const GValue *value,
                         GParamSpec   *spec)
{
  HRelationPrivate *private = H_RELATION_GET_PRIVATE (object);

  switch (property)
    {
    case PROP_NAME:
      private->name = g_strdup (g_value_get_string (value));
      break;
    case PROP_DESCRIPTION:
      private->description = g_strdup (g_value_get_string (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property, spec);
    }
}

static void
h_relation_get_property (GObject    *object,
                         guint       property,
                         GValue     *value,
                         GParamSpec *spec)
{
  HRelationPrivate *private = H_RELATION_GET_PRIVATE (object);

  switch (property)
    {
    case PROP_NAME:
      g_value_set_string (value, private->name);
      break;
    case PROP_DESCRIPTION:
      g_value_set_string (value, private->description);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property, spec);
    }
}

static void
append_attr_str (HAttribute *attr,
                 GString    *str)
{
  g_string_append_printf (str, "%s (%s)\n",
                          attr->name,
                          g_type_name (attr->type));
}

static void
append_tuple_str (HTuple  *tuple,
                  GString *str)
{
  gchar *tuple_str;

  tuple_str = h_tuple_printable_str (tuple);
  g_string_append_printf (str, "%s\n", tuple_str);
  g_free (tuple_str);
}

/**
 * h_relation_set_name
 * @relation: a #HRelation
 * @name: name
 *
 * Set the name of @relation.
 **/
void
h_relation_set_name (HRelation   *self,
                     const gchar *name)
{
  g_return_if_fail (H_IS_RELATION (self));

  g_object_set (G_OBJECT (self), "name", name, NULL);
}

/**
 * h_relation_set_description:
 * @relation: a #HRelation
 * @desc: description
 *
 * Set the description of @relation.
 **/
void
h_relation_set_description (HRelation   *self,
                            const gchar *desc)
{
  g_return_if_fail (H_IS_RELATION (self));

  g_object_set (G_OBJECT (self), "description", desc, NULL);
}

/**
 * h_relation_get_name:
 * @relation: a #HRelation
 *
 * Returns the name of @relation.
 *
 * Return value: a copy of the name string, or %NULL if no name is
 * set.
 **/
gchar *
h_relation_get_name (HRelation *self)
{
  gchar *name;

  g_return_val_if_fail (H_IS_RELATION (self), NULL);

  g_object_get (G_OBJECT (self), "name", &name, NULL);
  return name;
}

/**
 * h_relation_get_description:
 * @relation: a #HRelation
 *
 * Returns the description of @relation.
 *
 * Return value: a copy of the description string, or %NULL if no
 * description exists
 **/
gchar *
h_relation_get_description (HRelation *self)
{
  gchar *desc;

  g_return_val_if_fail (H_IS_RELATION (self), NULL);

  g_object_get (G_OBJECT (self), "description", &desc, NULL);
  return desc;
}

/**
 * h_relation_get_attr:
 * @relation: a #HRelation
 * @name: name of attribute
 *
 * Returns an attribute of @relation.
 *
 * Return value: attribute
 **/
HAttribute *
h_relation_get_attr (HRelation   *self,
                     const gchar *name)
{
  g_return_val_if_fail (H_IS_RELATION (self), NULL);
  g_return_val_if_fail (name != NULL, NULL);

  return H_RELATION_GET_CLASS (self)->get_attr (self, name);
}

/**
 * h_relation_print:
 * @relation: a #HRelation
 *
 * Prints @relation to stdout.
 **/
void
h_relation_print (HRelation *self)
{
  gchar *str;

  g_return_if_fail (H_IS_RELATION (self));

  str = h_relation_printable_str (self);
  printf ("%s", str);
  g_free (str);
}

/**
 * h_relation_printable_str:
 * @relation: a #HRelation
 *
 * Returns a printable string of @relation.
 *
 * Return value: printable string
 **/
gchar *
h_relation_printable_str (HRelation *self)
{
  HRelationPrivate *private;
  GString *str;
  gchar   *self_str;

  g_return_val_if_fail (H_IS_RELATION (self), NULL);

  private = H_RELATION_GET_PRIVATE (self);

  str = g_string_new ("");

  g_string_append_printf (str, "Name: %s\n", H_STR_NULL (private->name));
  g_string_append_printf (str, "Description: %s\n", H_STR_NULL (private->description));

  self_str = H_RELATION_GET_CLASS (self)->printable_str (self);
  g_string_append (str, self_str);
  g_free (self_str);

  g_string_append (str, "Attributes: \n");
  h_relation_for_each_attr (self, (HAttrFunc) append_attr_str, str);

  g_string_append (str, "Value: \n");
  h_relation_for_each_tuple (self, (HTupleFunc) append_tuple_str, str);

  return g_string_free (str, FALSE);
}

/**
 * h_relation_for_each_attr:
 * @relation: a #HRelation
 * @func: function called for each attribute in @relation
 * @user_data: custom data to pass to @func
 *
 * Performs a function over each attribute in @relation.
 **/
void
h_relation_for_each_attr (HRelation *self,
                          HAttrFunc  func,
                          gpointer   user_data)
{
  g_return_if_fail (H_IS_RELATION (self));

  H_RELATION_GET_CLASS (self)->for_each_attr (self, func, user_data);
}

/**
 * h_relation_for_each_tuple:
 * @relation: a #HRelation
 * @func: function called for each tuple in @relation
 * @user_data: custom data to pass to @func
 *
 * Performs a function over each tuple in @relation.
 **/
void
h_relation_for_each_tuple (HRelation  *self,
                           HTupleFunc  func,
                           gpointer    user_data)
{
  g_return_if_fail (H_IS_RELATION (self));

  H_RELATION_GET_CLASS (self)->for_each_tuple (self, func, user_data);
}

/**
 * h_relation_get_n_tuples:
 * @relation: a #HRelation
 *
 * Returns the number of tuples in @relation. This is an alternative
 * name for h_relation_get_cardinality().
 *
 * Return value: number of tuples
 *
 **/

/**
 * h_relation_get_cardinality:
 * @relation: a #HRelation
 *
 * Returns the cardinality of @relation.
 *
 * Return value: cardinality
 **/
guint
h_relation_get_cardinality (HRelation *self)
{
  g_return_val_if_fail (H_IS_RELATION (self), 0);

  return H_RELATION_GET_CLASS (self)->get_cardinality (self);
}

/**
 * h_relation_get_n_attrs:
 * @relation: a #HRelation
 *
 * Returns the number of attributes in @relation. This is an
 * alternative name for h_relation_get_degree().
 *
 * Return value: number of attributes
 **/

/**
 * h_relation_get_degree:
 * @relation: a #HRelation
 *
 * Returns the degree of @relation.
 *
 * Return value: degree
 **/
guint
h_relation_get_degree (HRelation *self)
{
  g_return_val_if_fail (H_IS_RELATION (self), 0);

  return H_RELATION_GET_CLASS (self)->get_degree (self);
}

/**
 * h_relation_has_same_attrs:
 * @relation1: a #HRelation
 * @relation2: a #HRelation
 *
 * Checks whether two relations have the same set of attributes.
 *
 * Return value: TRUE if yes, FALSE if no
 **/

typedef struct
{
  HRelation *other;
  gboolean   result;
}
CompareAttrArgs;

static void
compare_attr (HAttribute      *attr,
              CompareAttrArgs *args)
{
  HAttribute *other_attr;

  if (!args->result)
    return;

  other_attr = h_relation_get_attr (args->other, attr->name);
  if (!other_attr)
    {
      args->result = FALSE;
      return;
    }

  args->result = (attr->type == other_attr->type);
}

gboolean
h_relation_has_same_attrs (HRelation *relation1,
                           HRelation *relation2)
{
  CompareAttrArgs args;

  g_return_val_if_fail (H_IS_RELATION (relation1), FALSE);
  g_return_val_if_fail (H_IS_RELATION (relation2), FALSE);

  args.other  = relation2;
  args.result = TRUE;

  h_relation_for_each_attr (relation1, (HAttrFunc) compare_attr, &args);

  return args.result;
}

/**
 * h_relation_has_tuple:
 * @relation: a #HRelation
 * @tuple: a #HTuple
 *
 * Checks whether a relation contains a tuple.
 *
 * Return value: TRUE if yes, FALSE if no
 **/
gboolean
h_relation_has_tuple (HRelation *self,
                      HTuple    *tuple)
{
  g_return_val_if_fail (H_IS_RELATION (self), FALSE);
  g_return_val_if_fail (tuple != NULL, FALSE);

  return H_RELATION_GET_CLASS (self)->has_tuple (self, tuple);
}

/**
 * H_RELATION_FLOATING:
 * @relation: a #HRelation
 *
 * Checks whether @relation is unowned i.e. floating.
 *
 * Return value: TRUE if @relation is floating, FALSE otherwise
 **/

/**
 * h_relation_sink:
 * @relation: a #HRelation
 *
 * Removes the floating reference from @relation if it exists,
 * otherwise does nothing.
 **/
void
h_relation_sink (HRelation *self)
{
  g_return_if_fail (H_IS_RELATION (self));

  self->floating = FALSE;
}

/**
 * h_relation_sink_or_ref:
 * @relation: a #HRelation
 *
 * Removes the floating reference from @relation if it exists,
 * otherwise increase its reference count.
 *
 * Return value: @relation
 **/
HRelation *
h_relation_sink_or_ref (HRelation *relation)
{
  if (H_RELATION_FLOATING (relation))
    h_relation_sink (relation);
  else
    g_object_ref (relation);

  return relation;
}

/**
 * h_relation_ref:
 * @relation: a #HRelation
 *
 * Increases the reference count of @relation.
 *
 * Returns: #HRelation
 **/

/**
 * h_relation_unref:
 * @relation: a #HRelation
 *
 * Decreases the reference count of @relation. @relation will be
 * destroyed when its reference count goes to 0.
 *
 * Returns: #HRelation
 **/

static void
h_relation_insert_tuple (HRelation *self,
                         HTuple    *tuple)
{
  /* do nothing */
}

static void
h_relation_delete_tuple (HRelation *self,
                         HTuple    *tuple)
{
  /* do nothing */
}
