/*  libhrel - Holyrel library

    Copyright (C) 2005 Chong Kai Xiong <descender@phreaker.net>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the
    Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA  02111-1307  USA.
*/

#include <config.h>

#include "relation_base.h"
#include "types.h"
#include "misc.h"
#include <stdio.h>
#include <string.h>

#define H_RELATION_BASE_GET_PRIVATE(obj) \
  (G_TYPE_INSTANCE_GET_PRIVATE ((obj), H_TYPE_RELATION_BASE, HRelationBasePrivate))

typedef struct _HRelationBasePrivate HRelationBasePrivate;

struct _HRelationBasePrivate
{
  HAttrMap   *attrs;
  HHashIndex *tuples;
  GSList     *primary_key;
};

/**
 * HRelationBase:
 *
 * A #HRelation representing a base relation.
 **/
G_DEFINE_TYPE (HRelationBase, h_relation_base, H_TYPE_RELATION)

static void   h_relation_base_dispose         (GObject      *object);
static void   h_relation_base_finalize        (GObject      *object);
static gchar *h_relation_base_printable_str   (HRelation    *relation);
static HAttribute *h_relation_base_get_attr   (HRelation    *relation,
                                               const gchar  *name);
static guint  h_relation_base_get_degree      (HRelation    *relation);
static guint  h_relation_base_get_cardinality (HRelation    *relation);
static void   h_relation_base_for_each_attr   (HRelation    *relation,
                                               HAttrFunc     func,
                                               gpointer      user_data);
static void   h_relation_base_for_each_tuple  (HRelation    *relation,
                                               HTupleFunc    func,
                                               gpointer      user_data);
static gboolean h_relation_base_has_tuple     (HRelation    *relation,
                                               HTuple       *tuple);

static void
h_relation_base_class_init (HRelationBaseClass *klass)
{
  GObjectClass   *object_class = G_OBJECT_CLASS (klass);
  HRelationClass *relation_class = H_RELATION_CLASS (klass);

  g_type_class_add_private (klass, sizeof (HRelationBasePrivate));

  object_class->dispose = h_relation_base_dispose;
  object_class->finalize = h_relation_base_finalize;

  relation_class->printable_str = h_relation_base_printable_str;
  relation_class->get_attr = h_relation_base_get_attr;
  relation_class->get_degree = h_relation_base_get_degree;
  relation_class->get_cardinality = h_relation_base_get_cardinality;
  relation_class->has_tuple = h_relation_base_has_tuple;
  relation_class->for_each_attr = h_relation_base_for_each_attr;
  relation_class->for_each_tuple = h_relation_base_for_each_tuple;
}

static void
h_relation_base_init (HRelationBase *self)
{
  /* nothing to do */
}

static void
h_relation_base_dispose (GObject *object)
{
  G_OBJECT_CLASS (h_relation_base_parent_class)->dispose (object);
}

static void
h_relation_base_finalize (GObject *object)
{
  HRelationBasePrivate *private = H_RELATION_BASE_GET_PRIVATE (object);

  g_slist_foreach (private->primary_key, (GFunc) g_free, NULL);
  g_slist_free (private->primary_key);

  h_attr_map_free (private->attrs);
  h_hash_index_free (private->tuples);

  G_OBJECT_CLASS (h_relation_base_parent_class)->finalize (object);
}

static HAttribute *
h_relation_base_get_attr (HRelation   *relation,
                          const gchar *name)
{
  HRelationBasePrivate *private = H_RELATION_BASE_GET_PRIVATE (relation);

  return h_attr_map_get (private->attrs, name);
}

static HAttrMap *
build_attr_map (va_list *args)
{
  HAttrMap    *map;
  const gchar *name;
  GType        type;

  /* Collect attribute list into a name-attribute associative map */

  map = h_attr_map_new ();

  while (1)
    {
      name = va_arg (*args, const gchar *);
      if (!name)
        break;

      type = va_arg (*args, GType);

#ifdef H_ENABLE_TYPE_CHECK
      if (h_attr_map_get (map, name) || !h_type_known (type))
        {
          h_attr_map_free (map);
          return NULL;
        }
#endif

      h_attr_map_insert (map, h_attr_new (name, type));
    }

  return map;
}

#ifdef H_ENABLE_TYPE_CHECK
static gboolean
check_key (GHashTable *map,
           GSList     *key)
{
  if (!key)
    return FALSE;

  while (key)
    {
      if (!h_attr_map_get (map, (const gchar *) key->data))
        return FALSE;

      key = key->next;
    }

  return TRUE;
}
#endif

static GSList *
duplicate_key_in_place (GSList *key)
{
  GSList *iter;

  for (iter = key; iter; iter = iter->next)
    iter->data = g_strdup (iter->data);

  return key;
}

/**
 * h_relation_base_new:
 * @name: name
 * @description: description
 * @...: name and type of an attribute, followed by another and so on
 *   then %NULL, followed by the name of a primary key attribute,
 *   followed by another and so on, then %NULL.
 *
 * Returns a new base relation with name @name, description
 * @description, specified attributes and primary key attributes.
 *
 * Return value: a base relation or %NULL
 **/
HRelation *
h_relation_base_new (const gchar *name,
                     const gchar *description,
                     ...)
{
  HRelationBase        *relation;
  HRelationBasePrivate *private;
  GHashTable           *attrs;
  GSList               *primary_key;
  va_list               args;

  g_return_val_if_fail (name != NULL, NULL);
  g_return_val_if_fail (description != NULL, NULL);

  va_start (args, description);

  attrs = build_attr_map (&args);
  primary_key = h_va_list_to_slist (args);

  va_end (args);

#ifdef H_ENABLE_TYPE_CHECK
  if (h_attr_map_get_size (attrs) == 0 || !check_key (attrs, primary_key))
    {
      h_attr_map_free (attrs);
      g_slist_free (primary_key);
      return NULL;
    }
#endif

  relation = g_object_new (H_TYPE_RELATION_BASE,
                           "name", name,
                           "description", description,
                           NULL);

  private = H_RELATION_BASE_GET_PRIVATE (relation);

  private->primary_key = duplicate_key_in_place (primary_key);
  private->attrs = attrs;

  private->tuples = h_hash_index_new_l (NULL,
                                        (GDestroyNotify) h_tuple_free,
                                        primary_key);

  return H_RELATION (relation);
}

#ifdef H_ENABLE_TYPE_CHECK

typedef struct
{
  HTuple   *tuple;
  gboolean  match;
}
CheckAttrArgs;

static void
check_attr (const HAttribute *attr,
            CheckAttrArgs    *args)
{
  HTupleElem *elem;

  if (!args->match)
    return;

  elem = h_tuple_get (args->tuple, attr->name);

  if (!elem)
    {
      args->match = FALSE;
      return;
    }

  if (elem->type != G_TYPE_NONE &&
      elem->type != attr->type)
    args->match = FALSE;
}

static gboolean
h_relation_base_check_tuple (HRelationBase *self,
                             HTuple        *tuple)
{
  CheckAttrArgs args;

  if (h_tuple_get_size (tuple) !=
      h_relation_base_get_degree ((HRelation *) self))
    return FALSE;

  args.tuple = tuple;
  args.match = TRUE;

  h_relation_base_for_each_attr ((HRelation *) self,
                                 (HAttrFunc) check_attr,
                                 &args);

  return args.match;
}

#endif /* #ifdef H_ENABLE_TYPE_CHECK */

typedef struct
{
  HTuple *tuple;
  HTuple *copy;
}
CopyAttrArgs;

static void
copy_attr (const HAttribute *attr,
           CopyAttrArgs     *args)
{
  h_tuple_insert (args->copy,
                  attr->name,
                  attr->type,
                  h_tuple_get (args->tuple, attr->name)->value);
}

static void
copy_insert_tuple (HRelationBase *self,
                   HTuple        *tuple)
{
  HRelationBasePrivate *private = H_RELATION_BASE_GET_PRIVATE (self);
  CopyAttrArgs args;

  args.tuple = tuple;
  args.copy = h_tuple_new ();

  h_relation_base_for_each_attr ((HRelation *) self,
                                 (HAttrFunc) copy_attr,
                                 &args);

  h_hash_index_replace (private->tuples, args.copy, args.copy);

  g_signal_emit_by_name (self, "insert-tuple", args.copy);
}

static HTuple *
key_new_v (HRelationBase *relation,
           va_list        args)
{
  HTuple      *key;
  gint         i, key_size;
  const gchar *name;
  gpointer     value;

#ifdef H_ENABLE_TYPE_CHECK
  GSList      *node;
#endif

  key_size = g_slist_length (h_relation_base_get_primary_key (relation));

  key = h_tuple_new ();

  for (i = 0; i < key_size; i++)
    {
      name  = va_arg (args, const gchar *);
      value = va_arg (args, gpointer);

#ifdef H_ENABLE_TYPE_CHECK
      node = g_slist_find_custom (private->primary_key, name, (GCompareFunc) strcmp);
      if (!node)
        {
          g_message ("Key does not have attribute: %s", name);
          h_tuple_free (key);
          return NULL;
        }
#endif

      h_tuple_insert (key, name,
                      h_relation_base_get_attr ((HRelation *) relation, name)->type,
                      value);
    }

  return key;
}

/**
 * h_relation_base_insert_tuple:
 * @relation: a #HRelationBase
 * @tuple: a #HTuple to insert
 *
 * Inserts a tuple into @relation.
 *
 * Return value: TRUE if @tuple was successfully inserted, FALSE
 * otherwise
 **/
gboolean
h_relation_base_insert_tuple (HRelationBase *self,
                              HTuple        *tuple)
{
  g_return_val_if_fail (self != NULL, FALSE);

#ifdef H_ENABLE_TYPE_CHECK
  if (!h_relation_base_check_tuple (self, tuple))
    {
      g_message ("Invalid tuple type");
      return FALSE;
    }
#endif

  copy_insert_tuple (self, tuple);

  return TRUE;
}

/**
 * h_relation_base_insert_tuplev:
 * @relation: a #HRelationBase
 * @...: An attribute name, followed by its corresponding value,
 *   followed by another attribute name, and so on.
 *
 * Inserts a tuple into @relation.
 *
 * Return value: TRUE if tuple was successfully inserted, FALSE
 * otherwise
 **/
gboolean
h_relation_base_insert_tuplev (HRelationBase *self,
                               ...)
{
  va_list  args;
  HTuple  *tuple;

  g_return_val_if_fail (self != NULL, FALSE);

  va_start (args, self);
  tuple = h_tuple_n_new_v (FALSE,
                           h_relation_base_get_degree ((HRelation *) self),
                           args);
  va_end (args);

  copy_insert_tuple (self, tuple);
  h_tuple_free (tuple);

  return TRUE;
}

/**
 * h_relation_base_get_tuple:
 * @relation: a #HRelationBase
 * @key: primary key value
 *
 * Returns a tuple in @relation that has the primary key @key.
 *
 * Return value: a #HTuple storing the tuple or %NULL if no tuples have
 * the specified key.
 **/
HTuple *
h_relation_base_get_tuple (HRelationBase *self,
                           HTuple        *key)
{
  HRelationBasePrivate *private;
  HTuple *tuple;

  g_return_val_if_fail (H_IS_RELATION_BASE (self), FALSE);

  private = H_RELATION_BASE_GET_PRIVATE (self);

  tuple = (HTuple *) h_hash_index_lookup (private->tuples, key);

  return tuple ? h_tuple_clone (tuple) : NULL;
}

/**
 * h_relation_base_get_tuplev:
 * @relation: a #HRelationBase
 * @...: name and value of first key attribute, followed by the second
 *   and so on, then %NULL
 *
 * Returns a tuple in @relation that has the primary key @key.
 *
 * Return value: a #HTuple storing the tuple or %NULL if no tuples have
 * the specified key.
 **/
HTuple *
h_relation_base_get_tuplev (HRelationBase *self,
                            ...)
{
  va_list  args;
  HTuple  *key;
  HTuple  *tuple;

  g_return_val_if_fail (H_IS_RELATION_BASE (self), NULL);

  va_start (args, self);
  key = key_new_v (self, args);
  va_end (args);

  g_return_val_if_fail (key != NULL, NULL);

  tuple = h_relation_base_get_tuple (self, key);
  h_tuple_free (key);

  return tuple;
}

/**
 * h_relation_base_delete_tuple:
 * @relation: a #HRelationBase
 * @key: primary key value
 *
 * Deletes a tuple in @relation that has the primary key @key.
 *
 * Return value: TRUE if tuple was deleted, FALSE if no such tuple
 * exists.
 **/
gboolean
h_relation_base_delete_tuple (HRelationBase *self,
                              HTuple        *key)
{
  HRelationBasePrivate *private;
  HTuple *tuple;

  g_return_val_if_fail (self != NULL, FALSE);

  private = H_RELATION_BASE_GET_PRIVATE (self);

  tuple = (HTuple *) h_hash_index_lookup (private->tuples, key);
  g_return_val_if_fail (tuple != NULL, FALSE);

  g_signal_emit_by_name (self, "delete-tuple", tuple);

  return h_hash_index_remove (private->tuples, key);
}

/**
 * h_relation_base_delete_tuplev:
 * @relation: a #HRelationBase
 * @...: name and value of first key attribute, followed by the second
 *   and so on, then %NULL.
 *
 * Deletes the tuple in @relation that has the specified primary key.
 *
 * Return value: TRUE if tuple was deleted, FALSE if no such tuple
 * exists.
 **/
gboolean
h_relation_base_delete_tuplev (HRelationBase *self,
                               ...)
{
  va_list   args;
  HTuple   *key;
  gboolean  result;

  g_return_val_if_fail (H_IS_RELATION_BASE (self), FALSE);

  va_start (args, self);
  key = key_new_v (self, args);
  va_end (args);

  g_return_val_if_fail (key != NULL, FALSE);

  result = h_relation_base_delete_tuple (self, key);
  h_tuple_free (key);

  return result;
}

/**
 * h_relation_base_get_primary_key:
 * @relation: a #HRelationBase
 *
 * Returns the primary key of @relation.
 *
 * Return value: primary key (must not be freed)
 **/
GSList *
h_relation_base_get_primary_key (HRelationBase *self)
{
  HRelationBasePrivate *private = H_RELATION_BASE_GET_PRIVATE (self);

  /* Should we clone the key? */
  return private->primary_key;
}

static gchar *
h_relation_base_printable_str (HRelation *relation)
{
  HRelationBasePrivate *private = H_RELATION_BASE_GET_PRIVATE (relation);
  GString *str;
  GSList  *iter;

  str = g_string_new ("Primary key: ");

  for (iter = private->primary_key; iter; iter = iter->next)
    g_string_append_printf (str, "%s ", (const gchar *) iter->data);

  g_string_append_c (str, '\n');

  return g_string_free (str, FALSE);
}

static guint
h_relation_base_get_cardinality (HRelation *relation)
{
  HRelationBasePrivate *private = H_RELATION_BASE_GET_PRIVATE (relation);

  return h_hash_index_get_size (private->tuples);
}

static guint
h_relation_base_get_degree (HRelation *relation)
{
  HRelationBasePrivate *private = H_RELATION_BASE_GET_PRIVATE (relation);

  return h_attr_map_get_size (private->attrs);
}

static void
h_relation_base_for_each_attr (HRelation *relation,
                               HAttrFunc  func,
                               gpointer   user_data)
{
  HRelationBasePrivate *private = H_RELATION_BASE_GET_PRIVATE (relation);

  h_attr_map_foreach (private->attrs, func, user_data);
}

static void
h_relation_base_for_each_tuple (HRelation  *relation,
                                HTupleFunc  func,
                                gpointer    user_data)
{
  HRelationBasePrivate *private = H_RELATION_BASE_GET_PRIVATE (relation);

  h_hash_index_for_each_value (private->tuples, (HValueFunc) func, user_data);
}

static gboolean
h_relation_base_has_tuple (HRelation *relation,
                           HTuple    *tuple)
{
  HRelationBasePrivate *private = H_RELATION_BASE_GET_PRIVATE (relation);
  HTuple *stored_tuple;

  stored_tuple = h_hash_index_lookup (private->tuples, tuple);

  return stored_tuple && h_tuple_equal (stored_tuple, tuple);
}
