/*  libhrel - Holyrel library

    Copyright (C) 2005 Chong Kai Xiong <descender@phreaker.net>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the
    Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA  02111-1307  USA.
*/

#include <config.h>

#include <glib-object.h>
#include "relation_project.h"
#include "misc.h"
#include <stdio.h>

#define H_RELATION_PROJECT_GET_PRIVATE(obj) \
  (G_TYPE_INSTANCE_GET_PRIVATE ((obj), H_TYPE_RELATION_PROJECT, HRelationProjectPrivate))

typedef struct _HRelationProjectPrivate HRelationProjectPrivate;

struct _HRelationProjectPrivate
{
  HRelation  *to_project;
  HAttrMap   *attrs;
  HHashIndex *index;
};

/**
 * HRelationProject:
 *
 * A #HRelation representing a projected relation.
 **/
G_DEFINE_TYPE (HRelationProject, h_relation_project, H_TYPE_RELATION)

static void   h_relation_project_dispose        (GObject          *object);
static void   h_relation_project_finalize       (GObject          *object);
static gchar *h_relation_project_printable_str  (HRelation        *relation);
static HAttribute *h_relation_project_get_attr  (HRelation        *relation,
                                                 const gchar      *name);
static guint h_relation_project_get_degree      (HRelation        *relation);
static guint h_relation_project_get_cardinality (HRelation        *relation);
static void  h_relation_project_for_each_attr   (HRelation        *relation,
                                                 HAttrFunc         func,
                                                 gpointer          user_data);
static void  h_relation_project_for_each_tuple  (HRelation        *relation,
                                                 HTupleFunc        func,
                                                 gpointer          user_data);
static gboolean h_relation_project_has_tuple    (HRelation        *relation,
                                                 HTuple           *tuple);
static void  h_relation_project_insert_tuple    (HRelationProject *self,
                                                 HTuple           *tuple);
static void  h_relation_project_delete_tuple    (HRelationProject *self,
                                                 HTuple           *tuple);

static void
h_relation_project_class_init (HRelationProjectClass *klass)
{
  GObjectClass   *object_class = G_OBJECT_CLASS (klass);
  HRelationClass *relation_class = H_RELATION_CLASS (klass);

  g_type_class_add_private (klass, sizeof (HRelationProjectPrivate));

  object_class->dispose = h_relation_project_dispose;
  object_class->finalize = h_relation_project_finalize;

  relation_class->printable_str = h_relation_project_printable_str;
  relation_class->get_attr = h_relation_project_get_attr;
  relation_class->get_degree = h_relation_project_get_degree;
  relation_class->get_cardinality = h_relation_project_get_cardinality;
  relation_class->for_each_attr = h_relation_project_for_each_attr;
  relation_class->for_each_tuple = h_relation_project_for_each_tuple;
  relation_class->has_tuple = h_relation_project_has_tuple;
}

static void
h_relation_project_init (HRelationProject *self)
{
  HRelationProjectPrivate *private = H_RELATION_PROJECT_GET_PRIVATE (self);

  private->to_project = NULL;
  private->attrs = NULL;
  private->index = NULL;
}

static void
h_relation_project_dispose (GObject *object)
{
  G_OBJECT_CLASS (h_relation_project_parent_class)->dispose (object);
}

static void
h_relation_project_finalize (GObject *object)
{
  HRelationProject        *self    = H_RELATION_PROJECT (object);
  HRelationProjectPrivate *private = H_RELATION_PROJECT_GET_PRIVATE (self);

  g_signal_handlers_disconnect_by_func (private->to_project,
                                        (gpointer) h_relation_project_insert_tuple,
                                        self);

  g_signal_handlers_disconnect_by_func (private->to_project,
                                        (gpointer) h_relation_project_delete_tuple,
                                        self);

  h_attr_map_free (private->attrs);
  h_hash_index_free (private->index);

  g_object_unref (private->to_project);

  G_OBJECT_CLASS (h_relation_project_parent_class)->finalize (object);
}

static gchar *
h_relation_project_printable_str (HRelation *relation)
{
  HRelationProjectPrivate *private = H_RELATION_PROJECT_GET_PRIVATE (relation);
  GString *str;
  gchar   *temp_str;

  str = g_string_new ("Relation to project:\n");

  temp_str = h_relation_get_name (private->to_project);
  g_string_append_printf (str, "  Name: %s\n", H_STR_NULL (temp_str));
  g_free (temp_str);

  temp_str = h_relation_get_description (private->to_project);
  g_string_append_printf (str, "  Description: %s\n", H_STR_NULL (temp_str));
  g_free (temp_str);

  return g_string_free (str, FALSE);
}

static HAttribute *
h_relation_project_get_attr (HRelation   *relation,
                             const gchar *name)
{
  HRelationProjectPrivate *private = H_RELATION_PROJECT_GET_PRIVATE (relation);

  return h_attr_map_get (private->attrs, name);
}

static guint
h_relation_project_get_degree (HRelation *relation)
{
  HRelationProjectPrivate *private = H_RELATION_PROJECT_GET_PRIVATE (relation);

  return h_attr_map_get_size (private->attrs);
}

static guint
h_relation_project_get_cardinality (HRelation *relation)
{
  HRelationProjectPrivate *private = H_RELATION_PROJECT_GET_PRIVATE (relation);

  return h_hash_index_get_size (private->index);
}

static void
h_relation_project_for_each_attr (HRelation *relation,
                                  HAttrFunc  func,
                                  gpointer   user_data)
{
  HRelationProjectPrivate *private = H_RELATION_PROJECT_GET_PRIVATE (relation);

  h_attr_map_foreach (private->attrs, func, user_data);
}

static void
h_relation_project_for_each_tuple (HRelation  *relation,
                                   HTupleFunc  func,
                                   gpointer    user_data)
{
  HRelationProjectPrivate *private = H_RELATION_PROJECT_GET_PRIVATE (relation);

  h_hash_index_for_each_value (private->index, (HValueFunc) func, user_data);
}

static gboolean
h_relation_project_has_tuple (HRelation *relation,
                              HTuple    *tuple)
{
  HRelationProjectPrivate *private = H_RELATION_PROJECT_GET_PRIVATE (relation);

  return (h_hash_index_lookup (private->index, tuple) != NULL);
}

static HTuple *
insert_index_entry (HTuple           *tuple,
                    HRelationProject *self)
{
  HRelationProjectPrivate *private = H_RELATION_PROJECT_GET_PRIVATE (self);
  HTuple *copy;

  copy = h_tuple_clone_elems_l (tuple, h_hash_index_get_attrs (private->index));
  h_hash_index_replace (private->index, copy, copy);

  return copy;
}

static void
h_relation_project_insert_tuple (HRelationProject *self,
                                 HTuple           *tuple)
{
  HRelationProjectPrivate *private = H_RELATION_PROJECT_GET_PRIVATE (self);

  if (!h_hash_index_lookup (private->index, tuple))
    {
      HTuple *new_tuple;

      new_tuple = insert_index_entry (tuple, self);

      g_signal_emit_by_name (G_OBJECT (self), "insert-tuple", new_tuple);
    }
}

static void
h_relation_project_delete_tuple (HRelationProject *self,
                                 HTuple           *tuple)
{
  HRelationProjectPrivate *private = H_RELATION_PROJECT_GET_PRIVATE (self);

  if (h_hash_index_remove (private->index, tuple))
    {
      HTuple *new_tuple;

      new_tuple = h_tuple_clone_elems_l (tuple, h_hash_index_get_attrs (private->index));

      g_signal_emit_by_name (G_OBJECT (self), "delete-tuple", new_tuple);

      h_tuple_free (new_tuple);
    }
}

static void
build_index (HRelationProject *self,
             GSList           *list)
{
  HRelationProjectPrivate *private = H_RELATION_PROJECT_GET_PRIVATE (self);

  private->index = h_hash_index_new_l ((GDestroyNotify) h_tuple_free, NULL, list);

  h_relation_for_each_tuple (private->to_project,
                             (HTupleFunc) insert_index_entry,
                             self);
}

static void
attr_map_copy_insert_attr (HAttribute *attr,
                           HAttrMap   *map)
{
  h_attr_map_insert (map, h_attr_copy (attr));
}

static HAttrMap *
attr_map_from_name_list (HRelation *relation,
                         GSList    *list)
{
  HAttrMap    *map;
  const gchar *name;

#ifdef H_ENABLE_TYPE_CHECK
  GSList      *iter;

  for (iter = list; iter; iter = iter->next)
    {
      name = (const gchar *) list->data;

      if (!relation_get_attr (relation, name))
        {
          g_message ("Underlying relation has no attribute '%s'", name);
          return NULL;
        }
    }
#endif /* H_ENABLE_TYPE_CHECK */

  map = h_attr_map_new ();

  if (list)
    {
      while (list)
        {
          name = (const gchar *) list->data;

          h_attr_map_insert (map,
                             h_attr_copy (h_relation_get_attr (relation, name)));

          list = list->next;
        }
    }
  else
    {
      h_relation_for_each_attr (relation,
                                (HAttrFunc) attr_map_copy_insert_attr,
                                map);
    }

  return map;
}

/**
 * h_relation_project:
 * @relation: relation to project
 * @...: name of an attribute, followed by another and so on, then a
 *   %NULL
 *
 * Creates a relation on @relation projected over a set of
 * attributes.
 *
 * Return value: projected relation
 **/
HRelation *
h_relation_project (HRelation *to_project,
                    ...)
{
  va_list    args;
  GSList    *list;
  HRelation *relation;

  g_return_val_if_fail (H_IS_RELATION (to_project), NULL);

  va_start (args, to_project);
  list = h_va_list_to_slist (args);
  va_end (args);

  relation = h_relation_project_l (to_project, list);
  g_slist_free (list);

  return relation;
}

/**
 * h_relation_project_l:
 * @relation: relation to project
 * @list: list of attribute names
 *
 * A variation of h_relation_project() taking a #GSList of attribute
 * names.
 *
 * Return value: projected relation
 **/
HRelation *
h_relation_project_l (HRelation *to_project,
                      GSList    *list)
{
  HRelationProject        *relation;
  HRelationProjectPrivate *private;

  g_return_val_if_fail (H_IS_RELATION (to_project), NULL);

  relation = g_object_new (H_TYPE_RELATION_PROJECT, NULL);
  private = H_RELATION_PROJECT_GET_PRIVATE (relation);

  private->to_project =  h_relation_sink_or_ref (to_project);

  private->attrs = attr_map_from_name_list (to_project, list);

  build_index (relation, list);

  g_signal_connect_swapped (G_OBJECT (to_project), "insert-tuple",
                            G_CALLBACK (h_relation_project_insert_tuple),
                            relation);

  g_signal_connect_swapped (G_OBJECT (to_project), "delete-tuple",
                            G_CALLBACK (h_relation_project_delete_tuple),
                            relation);

  return (HRelation *) relation;
}
