/*  libhrel - Holyrel library

    Copyright (C) 2005 Chong Kai Xiong <descender@phreaker.net>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the
    Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA  02111-1307  USA.
*/

#include <config.h>

#include "relation_restrict.h"
#include "misc.h"
#include <stdio.h>

#define H_RELATION_RESTRICT_GET_PRIVATE(obj) \
  (G_TYPE_INSTANCE_GET_PRIVATE ((obj), H_TYPE_RELATION_RESTRICT, HRelationRestrictPrivate))

typedef struct _HRelationRestrictPrivate HRelationRestrictPrivate;

struct _HRelationRestrictPrivate
{
  HRelation      *to_restrict;
  HTuplePredFunc  func;
  gpointer        func_data;
  HHashIndex     *index;
};

/**
 * HTuplePredFunc:
 * @tuple: a #HTuple
 * @user_data: user data
 *
 * A #HTuple predicate function.
 *
 * Return value: either TRUE or FALSE
 **/

/**
 * HRelationRestrict:
 *
 * A #HRelation representing a restricted relation.
 **/
G_DEFINE_TYPE (HRelationRestrict, h_relation_restrict, H_TYPE_RELATION)

static void   h_relation_restrict_dispose         (GObject           *object);
static void   h_relation_restrict_finalize        (GObject           *object);
static gchar *h_relation_restrict_printable_str   (HRelation         *relation);
static HAttribute *h_relation_restrict_get_attr   (HRelation         *relation,
                                                   const gchar       *name);
static guint  h_relation_restrict_get_degree      (HRelation         *relation);
static guint  h_relation_restrict_get_cardinality (HRelation         *relation);
static void   h_relation_restrict_for_each_attr   (HRelation         *relation,
                                                   HAttrFunc          func,
                                                   gpointer           user_data);
static void   h_relation_restrict_for_each_tuple  (HRelation         *relation,
                                                   HTupleFunc         func,
                                                   gpointer           user_data);
static gboolean h_relation_restrict_has_tuple     (HRelation         *relation,
                                                   HTuple            *tuple);
static void  h_relation_restrict_insert_tuple     (HRelationRestrict *self,
                                                   HTuple            *tuple);
static void  h_relation_restrict_delete_tuple     (HRelationRestrict *self,
                                                   HTuple            *tuple);

static void
h_relation_restrict_class_init (HRelationRestrictClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  HRelationClass *h_relation_class = H_RELATION_CLASS (klass);

  g_type_class_add_private (klass, sizeof (HRelationRestrictPrivate));

  object_class->dispose  = h_relation_restrict_dispose;
  object_class->finalize = h_relation_restrict_finalize;

  h_relation_class->printable_str = h_relation_restrict_printable_str;
  h_relation_class->get_attr = h_relation_restrict_get_attr;
  h_relation_class->get_degree = h_relation_restrict_get_degree;
  h_relation_class->get_cardinality = h_relation_restrict_get_cardinality;
  h_relation_class->for_each_attr = h_relation_restrict_for_each_attr;
  h_relation_class->for_each_tuple = h_relation_restrict_for_each_tuple;
  h_relation_class->has_tuple = h_relation_restrict_has_tuple;
}

static void
h_relation_restrict_init (HRelationRestrict *self)
{
  HRelationRestrictPrivate *private = H_RELATION_RESTRICT_GET_PRIVATE (self);

  private->to_restrict = NULL;
  private->func = NULL;
  private->func_data = NULL;
}

static void
h_relation_restrict_dispose (GObject *object)
{
  G_OBJECT_CLASS (h_relation_restrict_parent_class)->dispose (object);
}

static void
h_relation_restrict_finalize (GObject *object)
{
  HRelationRestrict        *self    = H_RELATION_RESTRICT (object);
  HRelationRestrictPrivate *private = H_RELATION_RESTRICT_GET_PRIVATE (self);

  g_signal_handlers_disconnect_by_func (private->to_restrict,
                                        (gpointer) h_relation_restrict_insert_tuple,
                                        self);

  g_signal_handlers_disconnect_by_func (private->to_restrict,
                                        (gpointer) h_relation_restrict_delete_tuple,
                                        self);

  h_hash_index_free (private->index);

  g_object_unref (private->to_restrict);

  G_OBJECT_CLASS (h_relation_restrict_parent_class)->finalize (object);
}

static gchar *
h_relation_restrict_printable_str (HRelation *relation)
{
  HRelationRestrictPrivate *private = H_RELATION_RESTRICT_GET_PRIVATE (relation);
  GString *str;
  gchar   *temp_str;

  str = g_string_new ("Relation to restrict:\n");

  temp_str = h_relation_get_name (private->to_restrict);
  g_string_append_printf (str, "  Name: %s\n", H_STR_NULL (temp_str));
  g_free (temp_str);

  temp_str = h_relation_get_description (private->to_restrict);
  g_string_append_printf (str, "  Description: %s\n", H_STR_NULL (temp_str));
  g_free (temp_str);

  return g_string_free (str, FALSE);
}

static HAttribute *
h_relation_restrict_get_attr (HRelation   *relation,
                              const gchar *name)
{
  HRelationRestrictPrivate *private = H_RELATION_RESTRICT_GET_PRIVATE (relation);

  return h_relation_get_attr (private->to_restrict, name);
}

static guint
h_relation_restrict_get_degree (HRelation *relation)
{
  HRelationRestrictPrivate *private = H_RELATION_RESTRICT_GET_PRIVATE (relation);

  return h_relation_get_degree (private->to_restrict);
}

static guint
h_relation_restrict_get_cardinality (HRelation *relation)
{
  HRelationRestrictPrivate *private = H_RELATION_RESTRICT_GET_PRIVATE (relation);

  return h_hash_index_get_size (private->index);
}

static void
h_relation_restrict_for_each_attr (HRelation *relation,
                                   HAttrFunc  func,
                                   gpointer   user_data)
{
  HRelationRestrictPrivate *private = H_RELATION_RESTRICT_GET_PRIVATE (relation);

  h_relation_for_each_attr (private->to_restrict, func, user_data);
}

static void
h_relation_restrict_for_each_tuple (HRelation  *relation,
                                    HTupleFunc  func,
                                    gpointer    user_data)
{
  HRelationRestrictPrivate *private = H_RELATION_RESTRICT_GET_PRIVATE (relation);

  h_hash_index_for_each_value (private->index, (HValueFunc) func, user_data);
}

static gboolean
h_relation_restrict_has_tuple (HRelation *relation,
                               HTuple    *tuple)
{
  HRelationRestrictPrivate *private = H_RELATION_RESTRICT_GET_PRIVATE (relation);

  return (h_hash_index_lookup (private->index, tuple) != NULL);
}

static void
h_relation_restrict_insert_tuple (HRelationRestrict *self,
                                  HTuple            *tuple)
{
  HRelationRestrictPrivate *private = H_RELATION_RESTRICT_GET_PRIVATE (self);

  if ((*private->func) (tuple, private->func_data))
    {
      h_hash_index_insert (private->index, tuple, tuple);
      g_signal_emit_by_name (self, "insert-tuple", tuple);
    }
}

static void
h_relation_restrict_delete_tuple (HRelationRestrict *self,
                                  HTuple            *tuple)
{
  HRelationRestrictPrivate *private = H_RELATION_RESTRICT_GET_PRIVATE (self);

  if (h_hash_index_remove (private->index, tuple))
    g_signal_emit_by_name (self, "delete-tuple", tuple);
}

static void
insert_tuple (HTuple            *tuple,
              HRelationRestrict *self)
{
  HRelationRestrictPrivate *private = H_RELATION_RESTRICT_GET_PRIVATE (self);

  if ((*private->func) (tuple, private->func_data))
    h_hash_index_insert (private->index, tuple, tuple);
}

static void
build_index (HRelationRestrict *self)
{
  HRelationRestrictPrivate *private = H_RELATION_RESTRICT_GET_PRIVATE (self);

  private->index = h_hash_index_new (NULL, NULL, NULL);

  h_relation_for_each_tuple (private->to_restrict,
                             (HTupleFunc) insert_tuple,
                             self);
}

/**
 * h_relation_restrict:
 * @relation: a #HRelation to perform restrict on
 * @predicate_func: tuple predicate function
 * @user_data: Custom data to pass to @predicate_func
 *
 * Restricts @relation to the set of tuples evaluating to TRUE with
 * @predicate_func.
 *
 * Return value: restricted #HRelation
 **/
HRelation *
h_relation_restrict (HRelation     *to_restrict,
                     HTuplePredFunc predicate_func,
                     gpointer       user_data)
{
  HRelationRestrict        *relation;
  HRelationRestrictPrivate *private;

  g_return_val_if_fail (to_restrict != NULL, NULL);
  g_return_val_if_fail (predicate_func != NULL, NULL);

  relation = g_object_new (H_TYPE_RELATION_RESTRICT, NULL);

  private = H_RELATION_RESTRICT_GET_PRIVATE (relation);

  private->to_restrict = h_relation_sink_or_ref (to_restrict);

  private->func = predicate_func;
  private->func_data = user_data;

  build_index (relation);

  g_signal_connect_swapped (to_restrict, "insert-tuple",
                            G_CALLBACK (h_relation_restrict_insert_tuple),
                            relation);

  g_signal_connect_swapped (to_restrict, "delete-tuple",
                            G_CALLBACK (h_relation_restrict_delete_tuple),
                            relation);

  return H_RELATION (relation);
}

typedef struct
{
  HTuple *tuple;
  GSList *list;
}
TupleHasArgs;

static gboolean
tuple_has (HTuple       *tuple,
           TupleHasArgs *args)
{
  return h_tuple_equal_elems_l (tuple, args->tuple, args->list);
}

static TupleHasArgs *
tuple_has_args_new (HRelation *relation,
                    GSList    *list)
{
  TupleHasArgs *args;
  GSList         *iter;
  const gchar    *name;

  args = g_new (TupleHasArgs, 1);

  args->tuple = h_tuple_new ();
  args->list  = NULL;

  for (iter = list; iter; iter = iter->next->next)
    {
      name = (const gchar *) list->data;

      h_tuple_insert (args->tuple, name,
                      h_relation_get_attr (relation, name)->type,
                      list->next->data);

      args->list = g_slist_append (args->list, g_strdup (name));
    }

  return args;
}

static void
tuple_has_args_free (TupleHasArgs *args)
{
  h_string_list_free (args->list);
  h_tuple_free (args->tuple);
  g_free (args);
}

/**
 * h_relation_restrict_attrs:
 * @relation: relation to restrict
 * @...: name and value of an attribute, followed by another and so
 *   on, then a %NULL.
 *
 * Restricts @relation to the set of tuples which contain the
 * attribute values listed.
 *
 * Return value: restricted relation
 **/
HRelation *
h_relation_restrict_attrs (HRelation *to_restrict,
                           ...)
{
  va_list    args;
  GSList    *list;
  HRelation *relation;

  g_return_val_if_fail (H_IS_RELATION (to_restrict), NULL);

  va_start (args, to_restrict);
  list = h_va_list_to_slist (args);
  va_end (args);

  relation = h_relation_restrict_attrs_l (to_restrict, list);
  g_slist_free (list);

  return relation;
}

/**
 * h_relation_restrict_attrs_l:
 * @relation: relation to restrict
 * @list: list of attribute names and their values
 *
 * A variation of h_relation_restrict_attrs(), taking a #GSList for
 * the attribute list.
 *
 * Return value: restricted relation
 **/
HRelation *
h_relation_restrict_attrs_l (HRelation *to_restrict,
                             GSList    *list)
{
  HRelation    *relation;
  TupleHasArgs *args;

  g_return_val_if_fail (H_IS_RELATION (to_restrict), NULL);

  args = tuple_has_args_new (to_restrict, list);

  relation = h_relation_restrict (to_restrict, (HTuplePredFunc) tuple_has,
                                  args);

  g_object_set_data_full (G_OBJECT (relation), "tuple-has-args", args,
                          (GDestroyNotify) tuple_has_args_free);

  return relation;
}
