/*  libhrel - Holyrel library

    Copyright (C) 2005 Chong Kai Xiong <descender@phreaker.net>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the
    Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA  02111-1307  USA.
*/

#include <config.h>

#include "relation_sort.h"
#include "misc.h"
#include <stdio.h>

#define H_RELATION_SORT_GET_PRIVATE(obj) \
  (G_TYPE_INSTANCE_GET_PRIVATE ((obj), H_TYPE_RELATION_SORT, HRelationSortPrivate))

typedef struct _HRelationSortPrivate HRelationSortPrivate;

struct _HRelationSortPrivate
{
  HRelation        *to_sort;
  GCompareDataFunc  compare_func;
  gpointer          compare_func_data;
  GTree            *index;
};

typedef struct
{
  HTupleFunc func;
  gpointer   data;
}
IndexIterArgs;

/**
 * HRelationSort:
 *
 * A #HRelation representing a "sorted relation".
 **/
G_DEFINE_TYPE (HRelationSort, h_relation_sort, H_TYPE_RELATION)

static void   h_relation_sort_dispose         (GObject       *object);
static void   h_relation_sort_finalize        (GObject       *object);
static gchar *h_relation_sort_printable_str   (HRelation     *relation);
static HAttribute *h_relation_sort_get_attr   (HRelation     *relation,
                                               const gchar   *name);
static guint  h_relation_sort_get_degree      (HRelation     *relation);
static guint  h_relation_sort_get_cardinality (HRelation     *relation);
static void   h_relation_sort_for_each_attr   (HRelation     *relation,
                                               HAttrFunc      func,
                                               gpointer       user_data);
static void   h_relation_sort_for_each_tuple  (HRelation     *relation,
                                               HTupleFunc     func,
                                               gpointer       user_data);
static gboolean h_relation_sort_has_tuple     (HRelation     *relation,
                                               HTuple        *tuple);
static void   h_relation_sort_insert_tuple    (HRelationSort *self,
                                               HTuple        *tuple);
static void   h_relation_sort_delete_tuple    (HRelationSort *self,
                                               HTuple        *tuple);

static void index_build   (HRelationSort   *self);
static void index_insert  (HTuple          *tuple,
                           HRelationSort   *relation);
static void index_delete  (HTuple          *tuple,
                           HRelationSort   *relation);
static void index_iterate (HTuple          *key,
                           GList           *list,
                           IndexIterArgs   *args);

static void
h_relation_sort_class_init (HRelationSortClass *klass)
{
  GObjectClass   *object_class = G_OBJECT_CLASS (klass);
  HRelationClass *relation_class = H_RELATION_CLASS (klass);

  g_type_class_add_private (klass, sizeof (HRelationSortPrivate));

  object_class->dispose  = h_relation_sort_dispose;
  object_class->finalize = h_relation_sort_finalize;

  relation_class->printable_str = h_relation_sort_printable_str;
  relation_class->get_attr = h_relation_sort_get_attr;
  relation_class->get_degree = h_relation_sort_get_degree;
  relation_class->get_cardinality = h_relation_sort_get_cardinality;
  relation_class->for_each_attr = h_relation_sort_for_each_attr;
  relation_class->for_each_tuple = h_relation_sort_for_each_tuple;
  relation_class->has_tuple = h_relation_sort_has_tuple;
}

static void
h_relation_sort_init (HRelationSort *self)
{
  HRelationSortPrivate *private = H_RELATION_SORT_GET_PRIVATE (self);

  private->to_sort = NULL;
  private->compare_func = NULL;
  private->index = NULL;
}

static void
h_relation_sort_dispose (GObject *object)
{
  G_OBJECT_CLASS (h_relation_sort_parent_class)->dispose (object);
}

static void
h_relation_sort_finalize (GObject *object)
{
  HRelationSort        *self    = H_RELATION_SORT (object);
  HRelationSortPrivate *private = H_RELATION_SORT_GET_PRIVATE (self);

  g_signal_handlers_disconnect_by_func (private->to_sort,
                                        (gpointer) h_relation_sort_insert_tuple,
                                        self);

  g_signal_handlers_disconnect_by_func (private->to_sort,
                                        (gpointer) h_relation_sort_delete_tuple,
                                        self);

  g_object_unref (private->to_sort);

  g_tree_destroy (private->index);

  G_OBJECT_CLASS (h_relation_sort_parent_class)->finalize (object);
}

static gchar *
h_relation_sort_printable_str (HRelation *relation)
{
  HRelationSortPrivate *private = H_RELATION_SORT_GET_PRIVATE (relation);
  GString *str;
  gchar   *temp_str;

  str = g_string_new ("Relation to sort:\n");

  temp_str = h_relation_get_name (private->to_sort);
  g_string_append_printf (str, "  Name: %s\n", H_STR_NULL (temp_str));
  g_free (temp_str);

  temp_str = h_relation_get_description (private->to_sort);
  g_string_append_printf (str, "  Description: %s\n", H_STR_NULL (temp_str));
  g_free (temp_str);

  return g_string_free (str, FALSE);
}

static HAttribute *
h_relation_sort_get_attr (HRelation   *relation,
                          const gchar *name)
{
  HRelationSortPrivate *private = H_RELATION_SORT_GET_PRIVATE (relation);

  return h_relation_get_attr (private->to_sort, name);
}

static guint
h_relation_sort_get_degree (HRelation *relation)
{
  HRelationSortPrivate *private = H_RELATION_SORT_GET_PRIVATE (relation);

  return h_relation_get_degree (private->to_sort);
}

static guint
h_relation_sort_get_cardinality (HRelation *relation)
{
  HRelationSortPrivate *private = H_RELATION_SORT_GET_PRIVATE (relation);

  return h_relation_get_cardinality (private->to_sort);
}

static void
h_relation_sort_for_each_attr (HRelation *relation,
                               HAttrFunc  func,
                               gpointer   user_data)
{
  HRelationSortPrivate *private = H_RELATION_SORT_GET_PRIVATE (relation);

  h_relation_for_each_attr (private->to_sort, func, user_data);
}

static void
h_relation_sort_for_each_tuple (HRelation  *relation,
                                HTupleFunc  func,
                                gpointer    user_data)
{
  HRelationSortPrivate *private = H_RELATION_SORT_GET_PRIVATE (relation);
  IndexIterArgs         args;

  args.func = func;
  args.data = user_data;

  g_tree_foreach (private->index, (GTraverseFunc) index_iterate, &args);
}

static gboolean
h_relation_sort_has_tuple (HRelation *relation,
                           HTuple    *tuple)
{
  HRelationSortPrivate *private = H_RELATION_SORT_GET_PRIVATE (relation);
  GList *list;

  list = g_tree_lookup (private->index, tuple);

  return (h_list_find_custom_with_data (list, tuple,
                                        private->compare_func,
                                        private->compare_func_data) != NULL);
}

static void
h_relation_sort_insert_tuple (HRelationSort *self,
                              HTuple        *tuple)
{
  index_insert (tuple, self);

  g_signal_emit_by_name (self, "insert-tuple", tuple);
}

static void
h_relation_sort_delete_tuple (HRelationSort *self,
                              HTuple        *tuple)
{
  index_delete (tuple, self);

  g_signal_emit_by_name (self, "delete-tuple", tuple);
}

/**
 * h_relation_sort:
 * @relation: a relation to sort
 * @compare_func: function to compare a pair of tuples in @relation
 * @user_data: Custom data passed to @compare_func
 *
 * Creates a sorted relation.
 *
 * Return value: sorted relation
 **/
HRelation *
h_relation_sort (HRelation        *to_sort,
                 GCompareDataFunc  compare_func,
                 gpointer          user_data)
{
  HRelationSort        *relation;
  HRelationSortPrivate *private;

  g_return_val_if_fail (H_IS_RELATION (to_sort), NULL);
  g_return_val_if_fail (compare_func != NULL, NULL);

  relation = g_object_new (H_TYPE_RELATION_SORT, NULL);

  private = H_RELATION_SORT_GET_PRIVATE (relation);

  private->to_sort = h_relation_sink_or_ref (to_sort);

  private->compare_func = compare_func;
  private->compare_func_data = user_data;

  g_signal_connect_swapped (to_sort, "insert-tuple",
                            G_CALLBACK (h_relation_sort_insert_tuple),
                            relation);

  g_signal_connect_swapped (to_sort, "delete-tuple",
                            G_CALLBACK (h_relation_sort_delete_tuple),
                            relation);

  index_build (relation);

  return H_RELATION (relation);
}

/**
 * h_relation_sort_by_attrs:
 * @relation: relation to sort
 * @...: name of first attribute, followed by the second and so on,
 *   then a %NULL
 *
 * Creates a relation sorted by attributes. Each pair of tuples are
 * compared attribute by attribute in the order listed.
 *
 * Return value: sorted relation
 **/
HRelation *
h_relation_sort_by_attrs (HRelation *to_sort,
                          ...)
{
  va_list    args;
  GSList    *list;
  HRelation *relation;

  g_return_val_if_fail (H_IS_RELATION (to_sort), NULL);

  va_start (args, to_sort);
  list = h_va_list_to_slist (args);
  va_end (args);

  relation = h_relation_sort_by_attrs_l (to_sort, list);
  g_slist_free (list);

  return relation;
}

/**
 * h_relation_sort_by_attrs_l:
 * @relation: relation to sort
 * @list: list of attribute names
 *
 * A variation of h_relation_sort_by_attrs(), taking a #GSList for the
 * attribute name list.
 *
 * Return value: sorted relation
 **/
HRelation *
h_relation_sort_by_attrs_l (HRelation *to_sort,
                            GSList    *list)
{
  HRelation *relation;
  GSList    *list_copy;

  g_return_val_if_fail (H_IS_RELATION (to_sort), NULL);

  list_copy = h_string_list_copy (list);

  relation = h_relation_sort (to_sort,
                              (GCompareDataFunc) h_tuple_compare_elems_l,
                              list_copy);

  g_object_set_data_full (G_OBJECT (relation), "sort-attrs", list_copy,
                          (GDestroyNotify) h_string_list_free);

  return relation;
}


static void
index_iterate (HTuple        *key,
               GList         *list,
               IndexIterArgs *args)
{
  GList *iter;

  for (iter = list; iter != NULL; iter = iter->next)
    {
      args->func ((HTuple *) iter->data, args->data);
    }
}

static void
index_insert (HTuple        *tuple,
              HRelationSort *relation)
{
  HRelationSortPrivate *private = H_RELATION_SORT_GET_PRIVATE (relation);
  GList *list;

  list = g_tree_lookup (private->index, tuple);

  if (list)
    g_tree_steal (private->index, tuple);

  list = g_list_append (list, tuple);
  g_tree_insert (private->index, tuple, list);
}

static void
index_delete (HTuple        *tuple,
              HRelationSort *relation)
{
  HRelationSortPrivate *private = H_RELATION_SORT_GET_PRIVATE (relation);
  GList *list, *node;

  list = g_tree_lookup (private->index, tuple);
  if (!list)
    return;

  node = h_list_find_custom_with_data (list, tuple,
                                       private->compare_func,
                                       private->compare_func_data);
  g_return_if_fail (node != NULL);

  list = g_list_delete_link (list, node);

  g_tree_steal (private->index, tuple);

  if (list)
    g_tree_insert (private->index, tuple, list);
}

static void
index_build (HRelationSort *self)
{
  HRelationSortPrivate *private = H_RELATION_SORT_GET_PRIVATE (self);

  private->index = g_tree_new_full (private->compare_func,
                                    private->compare_func_data,
                                    NULL,
                                    (GDestroyNotify) g_list_free);

  h_relation_for_each_tuple (private->to_sort,
                             (HTupleFunc) index_insert,
                             self);
}
