/*  libhrel - Holyrel library

    Copyright (C) 2005 Chong Kai Xiong <descender@phreaker.net>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the
    Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA  02111-1307  USA.
*/

#include <config.h>
#include "relation_union.h"
#include "hash_index.h"
#include "misc.h"
#include <stdio.h>

#define H_RELATION_UNION_GET_PRIVATE(obj) \
  (G_TYPE_INSTANCE_GET_PRIVATE ((obj), H_TYPE_RELATION_UNION, HRelationUnionPrivate))

typedef struct _HRelationUnionPrivate HRelationUnionPrivate;

struct _HRelationUnionPrivate
{
  HRelation  *relation1;
  HRelation  *relation2;
  HHashIndex *index;
};

/**
 * HRelationUnion:
 *
 * A #HRelation representing the union of two relations.
 **/
G_DEFINE_TYPE (HRelationUnion, h_relation_union, H_TYPE_RELATION)

static void   h_relation_union_dispose         (GObject        *object);
static void   h_relation_union_finalize        (GObject        *object);
static gchar *h_relation_union_printable_str   (HRelation      *relation);
static HAttribute *h_relation_union_get_attr   (HRelation      *relation,
                                                const gchar    *name);
static guint  h_relation_union_get_degree      (HRelation      *relation);
static guint  h_relation_union_get_cardinality (HRelation      *relation);
static void   h_relation_union_for_each_attr   (HRelation      *relation,
                                                HAttrFunc       func,
                                                gpointer        user_data);
static void   h_relation_union_for_each_tuple  (HRelation      *relation,
                                                HTupleFunc      func,
                                                gpointer        user_data);
static gboolean h_relation_union_has_tuple     (HRelation      *relation,
                                                HTuple         *tuple);
static void   h_relation_union_insert_tuple    (HRelationUnion *self,
                                                HTuple         *tuple,
                                                HRelation      *relation);
static void   h_relation_union_delete_tuple    (HRelationUnion *self,
                                                HTuple         *tuple,
                                                HRelation      *relation);

static void
h_relation_union_class_init (HRelationUnionClass *klass)
{
  GObjectClass   *object_class = G_OBJECT_CLASS (klass);
  HRelationClass *relation_class = H_RELATION_CLASS (klass);

  g_type_class_add_private (klass, sizeof (HRelationUnionPrivate));

  object_class->dispose = h_relation_union_dispose;
  object_class->finalize = h_relation_union_finalize;

  relation_class->printable_str = h_relation_union_printable_str;
  relation_class->get_attr = h_relation_union_get_attr;
  relation_class->get_degree = h_relation_union_get_degree;
  relation_class->get_cardinality = h_relation_union_get_cardinality;
  relation_class->for_each_attr = h_relation_union_for_each_attr;
  relation_class->for_each_tuple = h_relation_union_for_each_tuple;
  relation_class->has_tuple = h_relation_union_has_tuple;
}

static void
h_relation_union_init (HRelationUnion *self)
{
  HRelationUnionPrivate *private = H_RELATION_UNION_GET_PRIVATE (self);

  private->relation1 = NULL;
  private->relation2 = NULL;
  private->index     = NULL;
}

static void
h_relation_union_dispose (GObject *object)
{
  G_OBJECT_CLASS (h_relation_union_parent_class)->dispose (object);
}

static void
h_relation_union_finalize (GObject *object)
{
  HRelationUnion        *self = H_RELATION_UNION (object);
  HRelationUnionPrivate *private = H_RELATION_UNION_GET_PRIVATE (self);

  g_signal_handlers_disconnect_by_func (private->relation1,
                                        (gpointer) h_relation_union_insert_tuple,
                                        self);

  g_signal_handlers_disconnect_by_func (private->relation1,
                                        (gpointer) h_relation_union_delete_tuple,
                                        self);

  g_signal_handlers_disconnect_by_func (private->relation2,
                                        (gpointer) h_relation_union_insert_tuple,
                                        self);

  g_signal_handlers_disconnect_by_func (private->relation2,
                                        (gpointer) h_relation_union_delete_tuple,
                                        self);

  h_hash_index_free (private->index);

  g_object_unref (private->relation1);
  g_object_unref (private->relation2);

  G_OBJECT_CLASS (h_relation_union_parent_class)->finalize (object);
}

static gchar *
h_relation_union_printable_str (HRelation *relation)
{
  HRelationUnionPrivate *private = H_RELATION_UNION_GET_PRIVATE (relation);
  GString *str;
  gchar   *temp_str;

  str = g_string_new ("Relational union of:\n");

  temp_str = h_relation_get_name (private->relation1);
  g_string_append_printf (str, "  1. Name: %s\n", H_STR_NULL (temp_str));
  g_free (temp_str);

  temp_str = h_relation_get_description (private->relation1);
  g_string_append_printf (str, "     Description: %s\n", H_STR_NULL (temp_str));
  g_free (temp_str);

  temp_str = h_relation_get_name (private->relation2);
  g_string_append_printf (str, "  2. Name: %s\n", H_STR_NULL (temp_str));
  g_free (temp_str);

  temp_str = h_relation_get_description (private->relation2);
  g_string_append_printf (str, "     Description: %s\n", H_STR_NULL (temp_str));
  g_free (temp_str);

  return g_string_free (str, FALSE);
}

static HAttribute *
h_relation_union_get_attr (HRelation   *relation,
                           const gchar *name)
{
  HRelationUnionPrivate *private = H_RELATION_UNION_GET_PRIVATE (relation);

  return h_relation_get_attr (private->relation1, name);
}

static guint
h_relation_union_get_degree (HRelation *relation)
{
  HRelationUnionPrivate *private = H_RELATION_UNION_GET_PRIVATE (relation);

  return h_relation_get_degree (private->relation1);
}

static guint
h_relation_union_get_cardinality (HRelation *relation)
{
  HRelationUnionPrivate *private = H_RELATION_UNION_GET_PRIVATE (relation);

  return h_hash_index_get_size (private->index);
}

static void
h_relation_union_for_each_attr  (HRelation *relation,
                                 HAttrFunc  func,
                                 gpointer   user_data)
{
  HRelationUnionPrivate *private = H_RELATION_UNION_GET_PRIVATE (relation);

  h_relation_for_each_attr (private->relation1, func, user_data);
}

static void
h_relation_union_for_each_tuple (HRelation  *relation,
                                 HTupleFunc  func,
                                 gpointer    user_data)
{
  HRelationUnionPrivate *private = H_RELATION_UNION_GET_PRIVATE (relation);

  h_hash_index_for_each_value (private->index, (HValueFunc) func, user_data);
}

static gboolean
h_relation_union_has_tuple (HRelation *relation,
                            HTuple    *tuple)
{
  HRelationUnionPrivate *private = H_RELATION_UNION_GET_PRIVATE (relation);

  return (h_hash_index_lookup (private->index, tuple) != NULL);
}

static void
insert_index_entry (HTuple         *tuple,
                    HRelationUnion *self)
{
  HRelationUnionPrivate *private = H_RELATION_UNION_GET_PRIVATE (self);
  HTuple *copy = h_tuple_clone (tuple);

  h_hash_index_replace (private->index, copy, copy);
}

static void
h_relation_union_insert_tuple (HRelationUnion *self,
                               HTuple         *tuple,
                               HRelation      *relation)
{
  HRelationUnionPrivate *private = H_RELATION_UNION_GET_PRIVATE (self);

  if (!h_hash_index_lookup (private->index, tuple))
    {
      insert_index_entry (tuple, self);

      g_signal_emit_by_name (G_OBJECT (self), "insert-tuple", tuple);
    }
}

static void
h_relation_union_delete_tuple (HRelationUnion *self,
                               HTuple         *tuple,
                               HRelation      *relation)
{
  HRelationUnionPrivate *private = H_RELATION_UNION_GET_PRIVATE (self);

  if (private->relation1 != private->relation2)
    {
      if (relation == private->relation1 &&
          h_relation_has_tuple (private->relation2, tuple))
        return;

      if (relation == private->relation2 &&
          h_relation_has_tuple (private->relation1, tuple))
        return;
    }

  h_hash_index_remove (private->index, tuple);

  g_signal_emit_by_name (G_OBJECT (self), "delete-tuple", tuple);
}

static void
build_index (HRelationUnion *self)
{
  HRelationUnionPrivate *private = H_RELATION_UNION_GET_PRIVATE (self);

  private->index = h_hash_index_new ((GDestroyNotify) h_tuple_free, NULL, NULL);

  h_relation_for_each_tuple (private->relation1,
                             (HTupleFunc) insert_index_entry,
                             self);

  h_relation_for_each_tuple (private->relation2,
                             (HTupleFunc) insert_index_entry,
                             self);
}

/**
 * h_relation_union:
 * @relation1: a #HRelation
 * @relation2: a #HRelation
 *
 * Creates the union of two relations.
 *
 * Return value: union of @relation1 and @relation2
 **/
HRelation *
h_relation_union (HRelation *relation1,
                  HRelation *relation2)
{
  HRelationUnion        *relation;
  HRelationUnionPrivate *private;

  g_return_val_if_fail (H_IS_RELATION (relation1), NULL);
  g_return_val_if_fail (H_IS_RELATION (relation2), NULL);

#ifdef H_ENABLE_TYPE_CHECK
  if (!h_relation_has_same_attrs (relation1, relation2))
    {
      g_message ("Relations cannot be union'ed, they have different sets of attributes");
      return NULL;
    }
#endif /* H_ENABLE_TYPE_CHECK */

  relation = g_object_new (H_TYPE_RELATION_UNION, NULL);
  private = H_RELATION_UNION_GET_PRIVATE (relation);

  private->relation1 = h_relation_sink_or_ref (relation1);
  private->relation2 = h_relation_sink_or_ref (relation2);

  build_index (relation);

  g_signal_connect_swapped (G_OBJECT (relation1), "insert-tuple",
                            G_CALLBACK (h_relation_union_insert_tuple),
                            relation);

  g_signal_connect_swapped (G_OBJECT (relation1), "delete-tuple",
                            G_CALLBACK (h_relation_union_delete_tuple),
                            relation);

  g_signal_connect_swapped (G_OBJECT (relation2), "insert-tuple",
                            G_CALLBACK (h_relation_union_insert_tuple),
                            relation);

  g_signal_connect_swapped (G_OBJECT (relation2), "delete-tuple",
                            G_CALLBACK (h_relation_union_delete_tuple),
                            relation);

  return H_RELATION (relation);
}
