/*  libhrel - Holyrel library

    Copyright (C) 2005 Chong Kai Xiong <descender@phreaker.net>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the
    Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA  02111-1307  USA.
*/

#include <config.h>

#include "tuple.h"
#include "types.h"
#include "misc.h"
#include <stdio.h>

#define DIRECT_COMPARE(a, b) \
    (((const gchar *) a) - ((const gchar *) b))

/**
 * HTuple:
 *
 * Tuple type.
 **/

/**
 * HTupleElem:
 * @type: type
 * @value: value of @type
 *
 * Stores type and value of a tuple (#HTuple) component.
 **/

typedef struct
{
  HTuple   *other;
  gboolean  result;
}
CheckCommonElemArgs;

static void     append_elem_str   (const gchar         *key,
                                   const HTupleElem    *elem,
                                   GString             *str);
static void     check_common_elem (const gchar         *name,
                                   const HTupleElem    *elem,
                                   CheckCommonElemArgs *args);
static void     combine_hash      (const gchar         *name,
                                   const HTupleElem    *elem,
                                   guint               *hash);
static void     copy_elem         (const gchar         *name,
                                   const HTupleElem    *elem,
                                   HTuple              *tuple);
static gpointer copy_value        (GType                type,
                                   gconstpointer        value);
static void     free_elem         (HTupleElem          *elem);

GType
h_tuple_get_type (void)
{
  static GType type = G_TYPE_INVALID;

  if (!type)
    {
      type = g_boxed_type_register_static ("HTuple",
                                           (GBoxedCopyFunc) h_tuple_clone,
                                           (GBoxedFreeFunc) h_tuple_free);
    }

  return type;
}

/**
 * h_tuple_new:
 *
 * Creates a new empty tuple.
 *
 * Return value: a newly created empty #HTuple, or %NULL.
 **/
HTuple *
h_tuple_new (void)
{
  return g_hash_table_new_full (g_str_hash, g_str_equal,
                                (GDestroyNotify) g_free,
                                (GDestroyNotify) free_elem);
}

/**
 * h_tuple_newv:
 * @typed: set to FALSE to create a typeless tuple
 * @...: name, type (omitted if typeless), value of first component,
 *   followed by the second, and so on, then %NULL
 *
 * Constructs a tuple from a list of components.
 *
 * Return value: newly constructed #HTuple, or %NULL
 **/
HTuple *
h_tuple_newv (gboolean typed,
              ...)
{
  va_list  args;
  HTuple  *tuple;

  va_start (args, typed);
  tuple = h_tuple_new_v (typed, args);
  va_end (args);

  return tuple;
}

/**
 * h_tuple_new_v:
 * @typed: set to FALSE to create a typeless tuple
 * @args: argument list
 *
 * Constructs a tuple from an argument list.
 *
 * Return value: a newly constructed #HTuple, or %NULL
 **/
HTuple *
h_tuple_new_v (gboolean typed,
               va_list  args)
{
  HTuple   *tuple;
  gchar    *name;
  GType     type;
  gpointer  value;

  tuple = h_tuple_new ();

  while (1)
    {
      name = va_arg (args, gchar *);
      if (!name)
        break;

      type  = typed ? va_arg (args, GType) : G_TYPE_NONE;
      value = va_arg (args, gpointer);

      h_tuple_insert (tuple, name, type, value);
    }

  return tuple;
}

/**
 * h_tuple_n_new_v:
 * @typed: set to FALSE to create a typeless tuple
 * @size: size
 * @args: argument list
 *
 * Constructs an n-tuple from an argument list.
 *
 * Return value: a newly constructed #HTuple, or %NULL
 **/
HTuple *
h_tuple_n_new_v (gboolean typed,
                 guint    size,
                 va_list  args)
{
  HTuple   *tuple;
  gchar    *name;
  GType     type;
  gpointer  value;

  tuple = h_tuple_new ();

  while (size--)
    {
      name  = va_arg (args, gchar *);
      type  = typed ? va_arg (args, GType) : G_TYPE_NONE;
      value = va_arg (args, gpointer);

      h_tuple_insert (tuple, name, type, value);
    }

  return tuple;
}

static HTupleElem *
h_tuple_elem_new (GType         type,
                  gconstpointer value)
{
  HTupleElem *elem;

  elem = g_new (HTupleElem, 1);
  elem->type = type;
  elem->value = copy_value (type, value);

  return elem;
}

/**
 * h_tuple_insert:
 * @tuple: a #HTuple
 * @name: name of component
 * @type: type of component
 * @value: value of @type
 *
 * Adds a new atribute named @name, of type @type and value @value.
 **/
void
h_tuple_insert (HTuple      *tuple,
                const gchar *name,
                GType        type,
                gpointer     value)
{
  g_return_if_fail (tuple != NULL);
  g_return_if_fail (name != NULL);

  if (type != G_TYPE_NONE)
    g_return_if_fail (h_type_known (type));

  g_hash_table_insert (tuple, g_strdup (name), h_tuple_elem_new (type, value));
}

/**
 * h_tuple_get:
 * @tuple: a #HTuple
 * @name: name of component
 *
 * Returns the type and value of an component in @tuple, or %NULL if
 * component doesn't exist.
 *
 * Return value: a #HTupleElem containing the type and value.
 **/
HTupleElem *
h_tuple_get (HTuple      *tuple,
             const gchar *name)
{
  g_return_val_if_fail (tuple != NULL, NULL);
  g_return_val_if_fail (name != NULL, NULL);

  return (HTupleElem *) g_hash_table_lookup (tuple, name);
}

/**
 * h_tuple_get_value:
 * @tuple: a #HTuple
 * @name: name of component
 *
 * Returns the value of an component in @tuple, or %NULL if component
 * doesn't exist.
 *
 * Return value: a copy of the component value
 **/
gpointer
h_tuple_get_value (HTuple      *tuple,
                   const gchar *name)
{
  HTupleElem *elem;

  g_return_val_if_fail (tuple != NULL, NULL);
  g_return_val_if_fail (name != NULL, NULL);

  elem = h_tuple_get (tuple, name);
  return elem ? h_value_clone (elem->type, elem->value) : NULL;
}

/**
 * h_tuple_get_size:
 * @tuple: a #HTuple
 *
 * Returns the degree ('size') of @tuple.
 *
 * Return value: degree ('size') of @tuple
 **/
guint
h_tuple_get_size (HTuple *tuple)
{
  g_return_val_if_fail (tuple != NULL, 0);

  return g_hash_table_size (tuple);
}

/**
 * h_tuple_clone:
 * @tuple: a #HTuple
 *
 * Returns a copy of @tuple.
 *
 * Return value: copy of @tuple
 **/
HTuple *
h_tuple_clone (HTuple *tuple)
{
  HTuple *copy;

  g_return_val_if_fail (tuple != NULL, NULL);

  copy = h_tuple_new ();
  g_hash_table_foreach (tuple, (GHFunc) copy_elem, copy);
  return copy;
}

/**
 * h_tuple_clone_elems_l:
 * @tuple: a #HTuple
 * @list: list of attribute names
 *
 * Constructs a subset of @tuple, with attributes listed in @list.
 *
 * Return value: subset of @tuple.
 **/

HTuple *
h_tuple_clone_elems_l (HTuple *tuple,
                       GSList *list)
{
  HTuple      *copy;
  const gchar *name;
  HTupleElem  *elem;

#ifdef H_ENABLE_TYPE_CHECK
  GSList      *iter;
#endif /* H_ENABLE_TYPE_CHECK */

  if (!list)
    return h_tuple_clone (tuple);

#ifdef H_ENABLE_TYPE_CHECK
  for (iter = list; iter; iter = iter->next)
    {
      name = (const gchar *) iter->data;

      if (!h_tuple_get (tuple, name))
        {
          g_message ("Tuple has no component named '%s'", name);
          return NULL;
        }
    }
#endif /* H_ENABLE_TYPE_CHECK */

  copy = h_tuple_new ();

  while (list)
    {
      name = (const gchar *) list->data;
      elem = h_tuple_get (tuple, name);

      h_tuple_insert (copy, name, elem->type, elem->value);
      list = list->next;
    }

  return copy;
}

/**
 * h_tuple_free:
 * @tuple: a #HTuple
 *
 * Free memory used by @tuple.
 **/
void
h_tuple_free (HTuple *tuple)
{
  if (tuple)
    g_hash_table_destroy (tuple);
}

/**
 * h_tuple_print:
 * @tuple: a #HTuple
 *
 * Prints the content of @tuple to stdout.
 **/
void
h_tuple_print (HTuple *tuple)
{
  gchar *str;

  g_return_if_fail (tuple != NULL);

  str = h_tuple_printable_str (tuple);
  printf ("%s", str);
  g_free (str);
}

/**
 * h_tuple_printable_str:
 * @tuple: a #HTuple
 *
 * Gets a printable string of @tuple.
 *
 * Return value: Printable string
 **/
gchar *
h_tuple_printable_str (HTuple *tuple)
{
  GString *str;

  g_return_val_if_fail (tuple != NULL, NULL);

  str = g_string_new ("(");
  g_hash_table_foreach (tuple, (GHFunc) append_elem_str, str);
  g_string_append_c (str, ')');

  return g_string_free (str, FALSE);
}

/**
 * h_tuple_compare_elem:
 * @tuple1: a #HTuple
 * @tuple2: a #HTuple
 * @name: name of component to compare
 *
 * Compares the values of a common component in @tuple1 and @tuple2.
 *
 * Return value: negative if @tuple1 is smaller, 0 if equal, positive
 * if greater
 **/
gint
h_tuple_compare_elem (HTuple      *tuple1,
                      HTuple      *tuple2,
                      const gchar *name)
{
  GType type;
  gpointer value1, value2;

  g_return_val_if_fail (tuple1 != NULL, 0);
  g_return_val_if_fail (tuple2 != NULL, 0);

#ifdef H_ENABLE_TYPE_CHECK
  g_return_val_if_fail (h_tuple_get (tuple1, name) != NULL &&
                        h_tuple_get (tuple2, name) != NULL,
                        0);
  g_return_val_if_fail (h_tuple_get (tuple1, name)->type !=
                        h_tuple_get (tuple2, name)->type,
                        0);
#endif /* H_ENABLE_TYPE_CHECK */

  type = h_tuple_get (tuple1, name)->type;

  value1 = h_tuple_get (tuple1, name)->value;
  value2 = h_tuple_get (tuple2, name)->value;

  return (type != G_TYPE_NONE) ?
    h_value_compare (type, value1, value2) :
    DIRECT_COMPARE (value1, value2);
}

/**
 * h_tuple_compare_elems:
 * @tuple1: a #HTuple
 * @tuple2: a #HTuple
 * @...: name of 1st component, name of 2nd component, and so on,
 *   ending with a %NULL.
 *
 * Compares the values of each common component in @tuple1 and @tuple2
 * in the order listed.
 *
 * Return value: negative if @tuple1 is smaller, 0 if equal, positive
 * if greater
 **/
gint
h_tuple_compare_elems (HTuple *tuple1,
                       HTuple *tuple2,
                       ...)
{
  va_list  args;
  GSList  *list;
  gint     result;

  va_start (args, tuple2);
  list = h_va_list_to_slist (args);
  va_end (args);

  result = h_tuple_compare_elems_l (tuple1, tuple2, list);
  g_slist_free (list);

  return result;
}

/**
 * h_tuple_compare_elems_l:
 * @tuple1: a #HTuple
 * @tuple2: a #HTuple
 * @list: list of component names
 *
 * A variation of h_tuple_compare_elems(), where the component name
 * list is stored as a #GSList.
 *
 * Return value: negative if @tuple1 is smaller, 0 if equal, positive
 * if greater
 **/
gint
h_tuple_compare_elems_l (HTuple *tuple1,
                         HTuple *tuple2,
                         GSList *list)
{
  gint result = 0;

  g_return_val_if_fail (tuple1 != NULL, 0);
  g_return_val_if_fail (tuple2 != NULL, 0);
  g_return_val_if_fail (list   != NULL, 0);

  while (list && !result)
    {
      result = h_tuple_compare_elem (tuple1, tuple2, (const gchar *) list->data);
      list = list->next;
    }

  return result;
}

typedef struct
{
  HTuple   *tuple;
  gboolean  result;
}
CompareElemParams;

static void
compare_elem (const gchar        *name,
              const HTupleElem   *elem,
              CompareElemParams  *params)
{
  HTupleElem *other;

  if (!params->result)
    return;

  other = h_tuple_get (params->tuple, name);

  if (!other || elem->type != other->type)
    {
      params->result = FALSE;
      return;
    }

  params->result = (h_value_compare (elem->type, elem->value, other->value) == 0);
}

/**
 * h_tuple_equal:
 * @tuple1: a #HTuple
 * @tuple2: a #HTuple
 *
 * Compares @tuple1 and @tuple2 for equality.
 *
 * Return value: TRUE if equal, FALSE if not
 **/
gboolean
h_tuple_equal (HTuple *tuple1,
               HTuple *tuple2)
{
  CompareElemParams params;

  g_return_val_if_fail (tuple1 != NULL && tuple2 != NULL, FALSE);

  params.tuple = tuple2;
  params.result = TRUE;
  g_hash_table_foreach (tuple1, (GHFunc) compare_elem, &params);

  return params.result;
}

/**
 * h_tuple_equal_elem:
 * @tuple1: a #HTuple
 * @tuple2: a #HTuple
 * @name: name of component to compare
 *
 * Compares a common component in @tuple1 and @tuple2 for equality.
 *
 * Return value: TRUE if equal, FALSE if not
 **/
gboolean
h_tuple_equal_elem (HTuple      *tuple1,
                    HTuple      *tuple2,
                    const gchar *name)
{
  g_return_val_if_fail (tuple1 != NULL, FALSE);
  g_return_val_if_fail (tuple2 != NULL, FALSE);

  if (!h_tuple_get (tuple1, name) || !h_tuple_get (tuple2, name))
      return FALSE;

  if (h_tuple_get (tuple1, name)->type != h_tuple_get (tuple2, name)->type)
      return FALSE;

  return (h_value_compare (h_tuple_get (tuple1, name)->type,
                           h_tuple_get (tuple1, name)->value,
                           h_tuple_get (tuple2, name)->value) == 0);
}

/**
 * h_tuple_equal_elems:
 * @tuple1: a #HTuple
 * @tuple2: a #HTuple
 * @...: first component name, followed by the second and so on, then %NULL
 *
 * Compares common components in @tuple1 and @tuple2 for equality. If
 * no component names are provided, all components are compared
 * (equivalent to h_tuple_equal()).
 *
 * Return value: TRUE if equal, FALSE if not
 **/
gboolean
h_tuple_equal_elems (HTuple *tuple1,
                     HTuple *tuple2,
                     ...)
{
  va_list  args;
  GSList  *list;
  gboolean result;

  va_start (args, tuple2);
  list = h_va_list_to_slist (args);
  va_end (args);

  result = h_tuple_equal_elems_l (tuple1, tuple2, list);
  g_slist_free (list);

  return result;
}

/**
 * h_tuple_equal_elems_l:
 * @tuple1: a #HTuple
 * @tuple2: a #HTuple
 * @list: name list of components to compare
 *
 * A variation of h_tuple_equal_elems() with the component name list is
 * stored as a #GSList.
 *
 * Return value: TRUE if equal, FALSE if not
 **/
gboolean
h_tuple_equal_elems_l (HTuple *tuple1,
                       HTuple *tuple2,
                       GSList *list)
{
  g_return_val_if_fail (tuple1 != NULL, FALSE);
  g_return_val_if_fail (tuple2 != NULL, FALSE);

  if (!list)
    return h_tuple_equal (tuple1, tuple2);

  while (list)
    {
      if (h_tuple_compare_elem (tuple1, tuple2, (const gchar *) list->data))
        return FALSE;

      list = list->next;
    }

  return TRUE;
}

/**
 * h_tuple_hash:
 * @tuple: a #HTuple
 *
 * Returns the hash of @tuple.
 *
 * Return value: hash number
 **/
guint
h_tuple_hash (HTuple *tuple)
{
  guint hash = 0;

  g_hash_table_foreach (tuple, (GHFunc) combine_hash, &hash);

  return hash;
}

/**
 * h_tuple_hash_elems:
 * @tuple: a #HTuple
 * @...: a component name, followed by another, and so on, ending with
 *   a %NULL
 *
 * Returns the hash of a subset of @tuple. If no components were
 * specified, all components are used to generate the hash. The hash
 * is independent of the order in which component names are listed.
 *
 * Return value: hash number
 **/
guint
h_tuple_hash_elems (HTuple *tuple,
                    ...)
{
  va_list   args;
  GSList   *list;
  guint     hash;

  va_start (args, tuple);
  list = h_va_list_to_slist (args);
  va_end (args);

  hash = h_tuple_hash_elems_l (tuple, list);
  g_slist_free (list);

  return hash;
}

/**
 * h_tuple_hash_elems_l
 * @tuple: a #HTuple
 * @list: list of component names
 *
 * A variation of h_tuple_hash_elems(), where the component name list
 * is store as a #GSList.
 *
 * Return value: hash number
 **/
guint
h_tuple_hash_elems_l (HTuple *tuple,
                      GSList *list)
{
  /* Jey (jeyk) suggested using XOR to combine the hashes. Assuming
   * the hash of each component is well distributed (i.e. for every
   * bit, the chance of being 0 or 1 is 0.5), the result should be
   * just as well distributed. */

  const HTupleElem *elem;
  guint             hash = 0;

  g_return_val_if_fail (tuple != NULL, 0);

  if (!list)
    return h_tuple_hash (tuple);

  while (list)
    {
      elem = h_tuple_get (tuple, (const gchar *) list->data);
      hash ^= (*h_type_hash_func (elem->type)) (elem->value);

      list = list->next;
    }

  return hash;
}

/**
 * h_tuple_have_common_attrs:
 * @tuple1: a #HTuple
 * @tuple2: a #HTuple
 *
 * Checks whether @tuple1 and @tuple2 have common attributes.
 *
 * Return value: TRUE if common attributes exist, FALSE if not
 **/
gboolean
h_tuple_have_common_attrs (HTuple *tuple1,
                           HTuple *tuple2)
{
  CheckCommonElemArgs args;

  g_return_val_if_fail (tuple1 != NULL, FALSE);
  g_return_val_if_fail (tuple2 != NULL, FALSE);

  args.other = tuple2;
  args.result = FALSE;

  g_hash_table_foreach (tuple1, (GHFunc) check_common_elem, &args);

  return args.result;
}

/**
 * h_tuple_product:
 * @tuple1: a #HTuple
 * @tuple2: a #HTuple
 *
 * Computes the Cartesian product of 2 tuples. @tuple1 and @tuple2
 * must not have common attributes.
 *
 * Return value: product of @tuple1 and @tuple2, or %NULL if they
 * have common attributes
 **/
HTuple *
h_tuple_product (HTuple *tuple1,
                 HTuple *tuple2)
{
  HTuple *tuple;

  g_return_val_if_fail (tuple1 != NULL, NULL);
  g_return_val_if_fail (tuple2 != NULL, NULL);

  if (h_tuple_have_common_attrs (tuple1, tuple2))
    return NULL;

  tuple = h_tuple_clone (tuple1);

  g_hash_table_foreach (tuple2, (GHFunc) copy_elem, tuple);

  return tuple;
}

static void
append_elem_str (const gchar      *key,
                 const HTupleElem *elem,
                 GString          *str)
{
  if (elem->type != G_TYPE_NONE)
    {
      gchar *value_str = h_value_str (elem->type, elem->value);
      g_string_append_printf (str, " %s:%s:%s ", key, g_type_name (elem->type), value_str);
      g_free (value_str);
    }
  else
    g_string_append_printf (str, " %s::0x%d ", key, GPOINTER_TO_INT (elem->value));
}

static void
check_common_elem (const gchar         *name,
                   const HTupleElem    *elem,
                   CheckCommonElemArgs *args)
{
  if (args->result)
    return;

  args->result = (h_tuple_get (args->other, name) != NULL);
}

static void
combine_hash (const gchar      *name,
              const HTupleElem *elem,
              guint            *hash)
{
  *hash ^= (*h_type_hash_func (elem->type)) (elem->value);
}

static void
copy_elem (const gchar      *name,
           const HTupleElem *elem,
           HTuple           *tuple)
{
  HTupleElem *copy = g_new (HTupleElem, 1);

  copy->type = elem->type;
  copy->value = copy_value (elem->type, elem->value);

  g_hash_table_insert (tuple, g_strdup (name), copy);
}

static gpointer
copy_value (GType         type,
            gconstpointer value)
{
  return (type != G_TYPE_NONE) ?
    h_value_clone (type, value) :
    (gpointer) value;
}

static void
free_elem (HTupleElem *elem)
{
  g_return_if_fail (elem != NULL);

  if (elem->type != G_TYPE_NONE)
    h_value_free (elem->type, elem->value);

  g_free (elem);
}

