/*  libhrel - Holyrel library

    Copyright (C) 2005 Chong Kai Xiong <descender@phreaker.net>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the
    Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA  02111-1307  USA.
*/

#include <config.h>

#include "types.h"
#include <string.h>

/**
 * HTypeInfo:
 * @size: size in bytes
 * @clone: clone function
 * @compare: comparator
 * @equal: equality function
 * @hash: hash function
 * @free: destructor
 * @str: string function
 **/

/**
 * HCloneFunc:
 * @value: value
 *
 * Returns a copy of @value
 *
 * Return value: copy
 **/

/**
 * HStrFunc:
 * @value: value
 *
 * Returns the pretty string representation of @value.
 *
 * Return value: pretty string
 **/

static gint     int_compare (gint a, gint b);
static gchar   *int_to_str  (gint a);
static gchar   *str_to_str  (const gchar *str);

typedef struct _HTypeInfoPair HTypeInfoPair;

struct _HTypeInfoPair
{
  GType     type;
  HTypeInfo info;
};

static HTypeInfoPair base_types[] =
{
  {
    G_TYPE_INT,
    {
      sizeof (gint),
      (HCloneFunc)     h_identity,
      (GCompareFunc)   int_compare,
      (GEqualFunc)     g_direct_equal,
      (GHashFunc)      g_direct_hash,
      (GDestroyNotify) h_dummy,
      (HStrFunc)       int_to_str
    }
  },

  {
    G_TYPE_STRING,
    {
      sizeof (gchar *),
      (HCloneFunc)     g_strdup,
      (GCompareFunc)   g_strcasecmp,
      (GEqualFunc)     g_str_equal,
      (GHashFunc)      g_str_hash,
      (GDestroyNotify) g_free,
      (HStrFunc)       str_to_str
    }
  }
};

static GHashTable *domain_types = NULL;

/**
 * h_type_get_info:
 * @type: type ID
 *
 * Returns information on @type.
 *
 * Return value: type information
 **/
HTypeInfo *
h_type_get_info (GType type)
{
    return (HTypeInfo *) g_hash_table_lookup (domain_types, GINT_TO_POINTER (type));
}

/**
 * h_type_known:
 * @type: type ID
 *
 * Checks whether @type is known to the type system.
 *
 * Return value: TRUE if yes, FALSE if no
 **/
gboolean
h_type_known (GType type)
{
  return (h_type_get_info (type) != NULL);
}

/**
 * h_type_compare_func:
 * @type: type ID
 *
 * Returns the comparator for @type.
 *
 * Return value: comparator, or %NULL if @type is unknown
 **/
GCompareFunc
h_type_compare_func (GType type)
{
  HTypeInfo *info = h_type_get_info (type);
  return info ? info->compare : NULL;
}

/**
 * h_type_hash_func:
 * @type: type ID
 *
 * Returns the hash function for @type.
 *
 * Return value: hash function, or %NULL if @type is unknown
 **/
GHashFunc
h_type_hash_func (GType type)
{
  HTypeInfo *info = h_type_get_info (type);
  return info ? info->hash : NULL;
}


/**
 * h_type_equal_func:
 * @type: type ID
 *
 * Returns the equality function for @type.
 *
 * Return value: equality function, or %NULL if @type is unknown
 **/
GEqualFunc
h_type_equal_func (GType type)
{
  HTypeInfo *info = h_type_get_info (type);
  return info ? info->equal : NULL;
}

/**
 * h_type_free_func:
 * @type: type ID
 *
 * Returns the destructor for @type.
 *
 * Return value: destructor, or %NULL if @type is unknown
 **/
GDestroyNotify
h_type_free_func (GType type)
{
  HTypeInfo *info = h_type_get_info (type);
  return info ? info->free : NULL;
}

/**
 * h_value_clone:
 * @type: type of @value
 * @value: a value
 *
 * Returns a copy of @value.
 *
 * Return value: copy
 **/
gpointer
h_value_clone (GType         type,
               gconstpointer value)
{
  HTypeInfo *info = h_type_get_info (type);
  return info ? info->clone (value) : NULL;
}

/**
 * h_value_str:
 * @type: type of @value
 * @value: a value
 *
 * Returns the pretty string representation of @value.
 *
 * Return value: pretty string representation
 **/
gchar *
h_value_str (GType         type,
             gconstpointer value)
{
  HTypeInfo *info = h_type_get_info (type);
  return info ? info->str (value) : NULL;
}

/**
 * h_value_free:
 * @type: type of @value
 * @value: a value
 *
 * Frees the memory allocated to @value.
 **/
void
h_value_free (GType    type,
              gpointer value)
{
  GDestroyNotify free_func = h_type_free_func (type);

  g_return_if_fail (free_func != NULL);
  free_func (value);
}

/**
 * h_value_compare:
 * @type: type of @value1 and @value2
 * @value1: left operand
 * @value2: right operand
 *
 * Compares @value1 and @value2.
 *
 * Return value: -1 if @value1 is greater, 1 if @value2 is greater, 0 if equal
 **/
gint
h_value_compare (GType         type,
                 gconstpointer value1,
                 gconstpointer value2)
{
  GCompareFunc compare_func = h_type_compare_func (type);

  g_return_val_if_fail (compare_func != NULL, 0);
  return compare_func (value1, value2);
}

/**
 * h_value_set_gvalue:
 * @type: type of @value
 * @value: a value
 * @gvalue: a #GValue
 *
 * Stores a value of type @type into a GValue.
 *
 * Return value: the #GValue passed in, or %NULL if @type is unknown
 **/
GValue *
h_value_set_gvalue (GType          type,
                    gconstpointer  value,
                    GValue        *gvalue)
{
  /* FIXME: generalize this */

  g_value_init (gvalue, type);

  switch (type)
    {
    case G_TYPE_INT:
      g_value_set_int (gvalue, GPOINTER_TO_INT (value));
      break;

    case G_TYPE_STRING:
      g_value_set_string (gvalue, (const gchar*) value);
      break;

    default:
      return NULL;
    }

  return gvalue;
}

/**
 * h_type_init:
 *
 * Initializes the type system. It is safe to call this multiple times.
 **/
void
h_type_init (void)
{
  static gboolean has_run = FALSE;

  gint i;

  if (has_run)
    return;

  has_run = TRUE;

  g_type_init ();

  domain_types = g_hash_table_new_full (g_direct_hash, g_direct_equal,
                                        NULL, g_free);

  for (i = 0; i < G_N_ELEMENTS (base_types); i++)
    h_type_add (base_types[i].type, &base_types[i].info);
}

/**
 * h_type_add:
 * @type: type ID
 * @info: information for @type
 *
 * Adds a new type to the system.
 *
 * Returns: TRUE if successful, FALSE if @type is already known
 **/
gboolean
h_type_add (GType            type,
            const HTypeInfo *info)
{
  HTypeInfo *info_copy;

  if (h_type_known (type))
    return FALSE;

  info_copy = g_new (HTypeInfo, 1);
  memcpy (info_copy, info, sizeof (HTypeInfo));

  g_hash_table_insert (domain_types, GINT_TO_POINTER (type), info_copy);
  return TRUE;
}

/**
 * h_type_remove:
 * @type: type ID
 *
 * Removes @type from the system.
 *
 * Return value: TRUE if type is known and removed, otherwise FALSE
 **/
gboolean
h_type_remove (GType type)
{
  return g_hash_table_remove (domain_types, GINT_TO_POINTER (type));
}


/* int comparator function */
static gint
int_compare (gint a,
             gint b)
{
  return (a - b);
}

/* int to string representation */
static gchar *
int_to_str (gint value)
{
  return g_strdup_printf ("%d", value);
}

/* string to string representation */
static gchar *
str_to_str (const gchar *value)
{
  return g_strconcat ("\"", value, "\"", NULL);
}

/**
 * h_dummy:
 * @value: value
 *
 * A dummy function that does nothing.
 **/
void
h_dummy (gconstpointer value)
{}

/**
 * h_identity:
 * @value: value
 *
 * Returns @value.
 *
 * Return value: @value
 **/
gpointer
h_identity (gconstpointer value)
{
  return (gpointer) value;
}

/**
 * h_direct_compare:
 * @value1: left operand
 * @value2: right operand
 *
 * Compares @value1 and @value2 using the builtin C operators <, and >.
 *
 * Return value: negative if @value1 is smaller, 0 if equal, positive
 * if greater
 **/
gint
h_direct_compare (gconstpointer value1,
                  gconstpointer value2)
{
  /* NOTE: this fails if system is LLP64 and values to compare are
   * pointers or long longs */

  /* NOTE 2: need to test and see how negative values are truncated */

  return ((gchar *) value1 - (gchar *) value2);
}
