/* GTK - The GIMP Toolkit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * Modified by the GTK+ Team and others 1997-2000.  See the AUTHORS
 * file for a list of people on the GTK+ Team.  See the ChangeLog
 * files for a list of changes.  These files are distributed with
 * GTK+ at ftp://ftp.gtk.org/pub/gtk/.
 */

#include <config.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include <gtk/gtk.h>
#include <gtk/gtkprivate.h>

#include <libskinned/bmp_tooltips.h>

#if 0
#include "gtkmain.h"
#include "gtkmenuitem.h"
#include "gtkprivate.h"
#include "gtkwidget.h"
#include "gtkwindow.h"
#include "gtkstyle.h"
#include "gtktooltips.h"
#include "gtkalias.h"
#endif

#define PIl 3.1415926535897932384626433832795029L

#define DEFAULT_SHOW_DELAY 200      /* Default show delay in ms */
#define DEFAULT_HIDE_DELAY 15000     /* Default hide delay in ms */
#define STICKY_DELAY 0              /* Delay before popping up next tip
                                     * if we're sticky
                                     */
#define STICKY_REVERT_DELAY 1000    /* Delay before sticky tooltips revert
				     * to normal
                                     */

static void bmp_tooltips_class_init        (BmpTooltipsClass *klass);
static void bmp_tooltips_init              (BmpTooltips      *tooltips);
static void bmp_tooltips_destroy           (GtkObject        *object);

static void bmp_tooltips_event_handler     (GtkWidget   *widget,
                                            GdkEvent    *event);
static void bmp_tooltips_widget_unmap      (GtkWidget   *widget,
                                            gpointer     data);
static void bmp_tooltips_widget_remove     (GtkWidget   *widget,
                                            gpointer     data);
static void bmp_tooltips_set_active_widget (BmpTooltips *tooltips,
                                            GtkWidget   *widget);
static gint bmp_tooltips_timeout           (gpointer     data);

static gint bmp_tooltips_paint_window      (BmpTooltips *tooltips);
static void bmp_tooltips_draw_tips         (BmpTooltips *tooltips);
static void bmp_tooltips_unset_tip_window  (BmpTooltips *tooltips);
static void
bmp_tooltips_hide_tip (GtkWidget *widget);
 
static gboolean get_keyboard_mode          (GtkWidget   *widget);

static GtkObjectClass *parent_class;
static const gchar  tooltips_data_key[] = "_BmpTooltipsData";
static const gchar  tooltips_info_key[] = "_BmpTooltipsInfo";

static GtkWidget *previous_widget;
static gulong handler_id = 0;

GType
bmp_tooltips_get_type (void)
{
  static GType tooltips_type = 0;

  if (!tooltips_type)
    {
      static const GTypeInfo tooltips_info =
      {
	sizeof (BmpTooltipsClass),
	NULL,		/* base_init */
	NULL,		/* base_finalize */
	(GClassInitFunc) bmp_tooltips_class_init,
	NULL,		/* class_finalize */
	NULL,		/* class_data */
	sizeof (BmpTooltips),
	0,		/* n_preallocs */
	(GInstanceInitFunc) bmp_tooltips_init,
      };

      tooltips_type = g_type_register_static (GTK_TYPE_OBJECT, "BmpTooltips",
					      &tooltips_info, 0);
    }

  return tooltips_type;
}

static void
bmp_tooltips_class_init (BmpTooltipsClass *class)
{
  GtkObjectClass *object_class;

  object_class = (GtkObjectClass*) class;

  parent_class = g_type_class_peek_parent (class);

  object_class->destroy = bmp_tooltips_destroy;
}

static void
bmp_tooltips_init (BmpTooltips *tooltips)
{
  tooltips->tip_window = NULL;
  tooltips->active_tips_data = NULL;
  tooltips->tips_data_list = NULL;

  tooltips->delay = DEFAULT_SHOW_DELAY;
  tooltips->enabled = TRUE;
  tooltips->timer_tag = 0;
  tooltips->use_sticky_delay = FALSE;
  tooltips->last_popdown.tv_sec = -1;
  tooltips->last_popdown.tv_usec = -1;
  tooltips->draw_hotspot = FALSE;
}

BmpTooltips *
bmp_tooltips_new (void)
{
  return g_object_new (BMP_TYPE_TOOLTIPS, NULL);
}

static void
bmp_tooltips_destroy_data (BmpTooltipsData *tooltipsdata)
{
  g_free (tooltipsdata->tip_text);
  g_free (tooltipsdata->tip_private);
  g_free (tooltipsdata->tip_icon_stock_id);

  g_signal_handlers_disconnect_by_func ((gpointer)tooltipsdata->widget,
					(GClosure*)bmp_tooltips_event_handler,
					(gpointer)tooltipsdata);
  g_signal_handlers_disconnect_by_func ((gpointer)tooltipsdata->widget,
					(GClosure*)bmp_tooltips_widget_unmap,
					(gpointer)tooltipsdata);
  g_signal_handlers_disconnect_by_func ((gpointer)tooltipsdata->widget,
					(GClosure*)bmp_tooltips_widget_remove,
					(gpointer)tooltipsdata);

  g_object_set_data (G_OBJECT (tooltipsdata->widget), tooltips_data_key, NULL);
  g_object_unref (tooltipsdata->widget);
  g_free (tooltipsdata);
}

static void
tip_window_display_closed (GdkDisplay  *display,
			   gboolean     was_error,
			   BmpTooltips *tooltips)
{
  bmp_tooltips_unset_tip_window (tooltips);
}

static void
disconnect_tip_window_display_closed (BmpTooltips *tooltips)
{
  g_signal_handlers_disconnect_by_func ((gpointer)gtk_widget_get_display (tooltips->tip_window),
					(GClosure*)tip_window_display_closed,
					(gpointer)tooltips);
}

static void
bmp_tooltips_unset_tip_window (BmpTooltips *tooltips)
{
  if (tooltips->tip_window)
    {
      disconnect_tip_window_display_closed (tooltips);

      gtk_widget_destroy (tooltips->tip_window);
      tooltips->tip_window = NULL;
    }
}

static void
bmp_tooltips_destroy (GtkObject *object)
{
  BmpTooltips *tooltips = BMP_TOOLTIPS (object);
  GList *current;
  BmpTooltipsData *tooltipsdata;

  g_return_if_fail (tooltips != NULL);

  if (tooltips->timer_tag)
    {
      g_source_remove (tooltips->timer_tag);
      tooltips->timer_tag = 0;
    }

  if (tooltips->tips_data_list != NULL)
    {
      current = g_list_first (tooltips->tips_data_list);
      while (current != NULL)
	{
	  tooltipsdata = (BmpTooltipsData*) current->data;
	  current = current->next;
	  bmp_tooltips_widget_remove (tooltipsdata->widget, tooltipsdata);
	}
    }

  bmp_tooltips_unset_tip_window (tooltips);

  GTK_OBJECT_CLASS (parent_class)->destroy (object);
}

static void
bmp_tooltips_update_screen (BmpTooltips *tooltips,
			    gboolean     new_window)
{
  gboolean screen_changed = FALSE;

  if (tooltips->active_tips_data &&
      tooltips->active_tips_data->widget)
    {
      GdkScreen *screen = gtk_widget_get_screen (tooltips->active_tips_data->widget);

      screen_changed = (screen != gtk_widget_get_screen (tooltips->tip_window));

      if (screen_changed)
	{
	  if (!new_window)
	    disconnect_tip_window_display_closed (tooltips);

	  gtk_window_set_screen (GTK_WINDOW (tooltips->tip_window), screen);
	}
    }

  if (screen_changed || new_window)
    g_signal_connect (gtk_widget_get_display (tooltips->tip_window), "closed",
		      G_CALLBACK (tip_window_display_closed), tooltips);

}

void
bmp_tooltips_force_window (BmpTooltips *tooltips)
{
  GtkWidget *alignment;

  g_return_if_fail (BMP_IS_TOOLTIPS (tooltips));

  if (!tooltips->tip_window)
    {
      tooltips->tip_window = gtk_window_new (GTK_WINDOW_POPUP);
      bmp_tooltips_update_screen (tooltips, TRUE);
      gtk_widget_set_app_paintable (tooltips->tip_window, TRUE);
      gtk_window_set_resizable (GTK_WINDOW (tooltips->tip_window), FALSE);
      gtk_widget_set_name (tooltips->tip_window, "gtk-tooltips");
      gtk_container_set_border_width (GTK_CONTAINER (tooltips->tip_window), 4);

      g_signal_connect_swapped (tooltips->tip_window,
				"expose_event",
				G_CALLBACK (bmp_tooltips_paint_window),
				tooltips);

      tooltips->tip_hbox = gtk_hbox_new (FALSE, 0);

      tooltips->tip_alignment = gtk_alignment_new (0.5, 0.5, 1.0, 1.0);
      gtk_alignment_set_padding (GTK_ALIGNMENT(tooltips->tip_alignment), 2, 2, 2, 12);
      gtk_container_add (GTK_CONTAINER (tooltips->tip_alignment), tooltips->tip_hbox);

      alignment = gtk_alignment_new (0.5,0.0,0.5,0.5);
      gtk_alignment_set_padding (GTK_ALIGNMENT (alignment), 1, 0, 0, 4);

      tooltips->tip_image = gtk_image_new ();
      gtk_misc_set_alignment (GTK_MISC (tooltips->tip_image), 0.5, 0.0);
      gtk_widget_show (tooltips->tip_image);

      gtk_container_add (GTK_CONTAINER (alignment), tooltips->tip_image);

      tooltips->tip_label = gtk_label_new (NULL);
      gtk_label_set_line_wrap (GTK_LABEL (tooltips->tip_label), TRUE);
      gtk_label_set_justify (GTK_LABEL (tooltips->tip_label), GTK_JUSTIFY_FILL);
      gtk_misc_set_alignment (GTK_MISC (tooltips->tip_label), 0.5, 0.5);
      gtk_widget_show (tooltips->tip_label);

      gtk_box_pack_start (GTK_BOX(tooltips->tip_hbox), GTK_WIDGET (alignment), FALSE, FALSE, 4);
      gtk_box_pack_start (GTK_BOX(tooltips->tip_hbox), tooltips->tip_label, FALSE, FALSE, 0);

      gtk_widget_show_all (tooltips->tip_alignment);
      gtk_container_add (GTK_CONTAINER (tooltips->tip_window), tooltips->tip_alignment);

      g_signal_connect (tooltips->tip_window,
			"destroy",
			G_CALLBACK (gtk_widget_destroyed),
			&tooltips->tip_window);
    }
}

void
bmp_tooltips_enable (BmpTooltips *tooltips)
{
  g_return_if_fail (tooltips != NULL);

  tooltips->enabled = TRUE;
}

void
bmp_tooltips_disable (BmpTooltips *tooltips)
{
  g_return_if_fail (tooltips != NULL);

  bmp_tooltips_set_active_widget (tooltips, NULL);

  tooltips->enabled = FALSE;
}

void
bmp_tooltips_set_delay (BmpTooltips *tooltips,
                        guint         delay)
{
  g_return_if_fail (tooltips != NULL);

  tooltips->delay = delay;
}

void
bmp_tooltips_set_draw_hotspot (BmpTooltips *tooltips,
			       gboolean     draw_hotspot)
{
  g_return_if_fail (tooltips != NULL);

  tooltips->draw_hotspot = draw_hotspot;
}

BmpTooltipsData*
bmp_tooltips_data_get (GtkWidget       *widget)
{
  g_return_val_if_fail (widget != NULL, NULL);

  return g_object_get_data (G_OBJECT (widget), tooltips_data_key);
}

void
bmp_tooltips_set_tip (BmpTooltips *tooltips,
		      GtkWidget   *widget,
		      const gchar *icon_stock_id,
		      const gchar *tip_text,
		      const gchar *tip_private,
		      gboolean rectangular)
{
  BmpTooltipsData *tooltipsdata;

  g_return_if_fail (BMP_IS_TOOLTIPS (tooltips));
  g_return_if_fail (widget != NULL);

  tooltipsdata = bmp_tooltips_data_get (widget);

  if (!tip_text)
    {
      if (tooltipsdata)
	bmp_tooltips_widget_remove (tooltipsdata->widget, tooltipsdata);
      return;
    }

  if (tooltips->active_tips_data
      && tooltips->active_tips_data->widget == widget
      && GTK_WIDGET_DRAWABLE (tooltips->active_tips_data->widget))
    {
      g_free (tooltipsdata->tip_text);
      g_free (tooltipsdata->tip_private);
      g_free (tooltipsdata->tip_icon_stock_id);

      tooltipsdata->tip_text = g_strdup (tip_text);
      tooltipsdata->tip_private = g_strdup (tip_private);
      tooltipsdata->tip_icon_stock_id = g_strdup (icon_stock_id);
      tooltipsdata->rectangular = rectangular;

      bmp_tooltips_draw_tips (tooltips);
    }
  else
    {
      g_object_ref (widget);

      if (tooltipsdata)
        bmp_tooltips_widget_remove (tooltipsdata->widget, tooltipsdata);

      tooltipsdata = g_new0 (BmpTooltipsData, 1);

      tooltipsdata->tooltips = tooltips;
      tooltipsdata->widget = widget;

      tooltipsdata->tip_text = g_strdup (tip_text);
      tooltipsdata->tip_private = g_strdup (tip_private);
      tooltipsdata->tip_icon_stock_id = g_strdup (icon_stock_id);
      tooltipsdata->rectangular = rectangular;

      tooltips->tips_data_list = g_list_append (tooltips->tips_data_list,
                                                tooltipsdata);
      g_signal_connect_after (widget, "event-after",
                              G_CALLBACK (bmp_tooltips_event_handler),
			      tooltipsdata);

      g_object_set_data (G_OBJECT (widget), tooltips_data_key,
                         tooltipsdata);

      g_signal_connect (widget, "unmap",
			G_CALLBACK (bmp_tooltips_widget_unmap),
			tooltipsdata);

      g_signal_connect (widget, "unrealize",
			G_CALLBACK (bmp_tooltips_widget_unmap),
			tooltipsdata);

      g_signal_connect (widget, "destroy",
                        G_CALLBACK (bmp_tooltips_widget_remove),
			tooltipsdata);
    }
}

void
bmp_tooltips_set_text (BmpTooltips *tooltips,
		       GtkWidget   *widget,
		       const gchar *tip_text)
{
  BmpTooltipsData *tooltipsdata;

  g_return_if_fail (BMP_IS_TOOLTIPS (tooltips));
  g_return_if_fail (widget != NULL);

  tooltipsdata = bmp_tooltips_data_get (widget);

  g_free (tooltipsdata->tip_text);
  tooltipsdata->tip_text = g_strdup (tip_text);
  bmp_tooltips_draw_tips (tooltips);
}


void
bmp_tooltips_set_icon (BmpTooltips *tooltips,
		       GtkWidget   *widget,
		       const gchar *icon_stock_id)
{
  BmpTooltipsData *tooltipsdata;

  g_return_if_fail (BMP_IS_TOOLTIPS (tooltips));
  g_return_if_fail (widget != NULL);

  tooltipsdata = bmp_tooltips_data_get (widget);

  g_free (tooltipsdata->tip_icon_stock_id);
  tooltipsdata->tip_icon_stock_id = g_strdup (icon_stock_id);
  bmp_tooltips_draw_tips (tooltips);
}

static gint
bmp_tooltips_paint_window (BmpTooltips *tooltips)
{
  GtkRequisition req;

  gtk_widget_size_request (tooltips->tip_window, &req);
  gtk_paint_flat_box (tooltips->tip_window->style, tooltips->tip_window->window,
		      GTK_STATE_NORMAL, GTK_SHADOW_OUT,
		      NULL, GTK_WIDGET(tooltips->tip_window), "tooltip",
		      0, 0, req.width, req.height);

  return FALSE;
}

static gboolean
on_tooltip_widget_expose (GtkWidget      *widget,
			  GdkEventExpose *event,
			  gpointer user_data)
{
    cairo_t *cr;
    gboolean rectangular = GPOINTER_TO_INT (user_data);

    g_return_val_if_fail (GTK_IS_WIDGET (widget), FALSE);

    cr = gdk_cairo_create (widget->window);

    if (rectangular)
    {
	cairo_rectangle (cr, 1, 1, widget->allocation.width-1, widget->allocation.height-1);
	cairo_set_source_rgba (cr, 1., 0., 0., 0.35);
	cairo_fill (cr);

	cairo_set_line_width (cr, 1.0);
	cairo_rectangle (cr, 0, 0, widget->allocation.width, widget->allocation.height);
	cairo_set_source_rgba (cr, 1., 0., 0., 0.3);
	cairo_stroke (cr);
    }
    else
    {
	gdouble x, y, w, h, r = 0.0;

	w = widget->allocation.width;
	h = widget->allocation.height;

	x = w/2;
	y = w/2;

	if (w>h)
	    r = w/3;
	else if (w<h)
	    r = h/3;
	else if (w==h)
	    r = w/3;

	cairo_arc (cr, x, y - 3., r, 0., 2. * PIl);
	cairo_set_source_rgba (cr, 1., 0., 0., 0.35);
	cairo_fill (cr);
    }

    cairo_destroy (cr);

    return FALSE;
}

static void
bmp_tooltips_draw_tips (BmpTooltips *tooltips)
{
  GtkRequisition requisition;
  GtkWidget *widget;
  GtkStyle *style;
  gint x, y, w, h;
  BmpTooltipsData *data;
  gboolean keyboard_mode;
  GdkScreen *screen;
  GdkScreen *pointer_screen;
  gint monitor_num, px, py;
  GdkRectangle monitor;
#if 0
  gchar *markup;
#endif

  if (!tooltips->active_tips_data) return;
  if (!tooltips->active_tips_data->widget) return;

  if (!tooltips->tip_window)
    bmp_tooltips_force_window (tooltips);
  else if (GTK_WIDGET_VISIBLE (tooltips->tip_window))
    g_get_current_time (&tooltips->last_popdown);

  gtk_widget_ensure_style (tooltips->tip_window);
  style = tooltips->tip_window->style;

  widget = tooltips->active_tips_data->widget;
  g_object_set_data (G_OBJECT (tooltips->tip_window), tooltips_info_key,
                     tooltips);

  keyboard_mode = get_keyboard_mode (widget);

  bmp_tooltips_update_screen (tooltips, FALSE);

  screen = gtk_widget_get_screen (widget);

  data = tooltips->active_tips_data;

#if 0
  markup = g_strconcat ("<span size=\"small\">", data->tip_text, "</span>", NULL);
  gtk_label_set_markup (GTK_LABEL (tooltips->tip_label), markup);
#endif

  gtk_label_set_markup (GTK_LABEL (tooltips->tip_label), data->tip_text);
#if 0
  g_free (markup);
#endif
  gtk_image_set_from_stock (GTK_IMAGE (tooltips->tip_image), data->tip_icon_stock_id, GTK_ICON_SIZE_MENU);

  gtk_widget_size_request (tooltips->tip_window, &requisition);
  w = requisition.width;
  h = requisition.height;

  gdk_window_get_origin (widget->window, &x, &y);
  if (GTK_WIDGET_NO_WINDOW (widget))
    {
      x += widget->allocation.x;
      y += widget->allocation.y;
    }

  x += widget->allocation.width / 2;

  if (!keyboard_mode)
    gdk_window_get_pointer (gdk_screen_get_root_window (screen),
			    &x, NULL, NULL);

  x -= (w / 2 + 4);

  gdk_display_get_pointer (gdk_screen_get_display (screen),
			   &pointer_screen, &px, &py, NULL);
  if (pointer_screen != screen)
    {
      px = x;
      py = y;
    }
  monitor_num = gdk_screen_get_monitor_at_point (screen, px, py);
  gdk_screen_get_monitor_geometry (screen, monitor_num, &monitor);

  if ((x + w) > monitor.x + monitor.width)
    x -= (x + w) - (monitor.x + monitor.width);
  else if (x < monitor.x)
    x = monitor.x;

  if ((y + h + widget->allocation.height + 4) > monitor.y + monitor.height)
    y = y - h - 4;
  else
    y = y + widget->allocation.height + 4;

  gtk_window_move (GTK_WINDOW (tooltips->tip_window), x, y);
  gtk_widget_show (tooltips->tip_window);
  g_timeout_add (DEFAULT_HIDE_DELAY, (GSourceFunc)bmp_tooltips_hide_tip, widget);
}

static gint
bmp_tooltips_timeout (gpointer data)
{
  BmpTooltips *tooltips = (BmpTooltips *) data;

  GDK_THREADS_ENTER ();

  if (tooltips->active_tips_data != NULL &&
      GTK_WIDGET_DRAWABLE (tooltips->active_tips_data->widget))
    bmp_tooltips_draw_tips (tooltips);

  GDK_THREADS_LEAVE ();

  return FALSE;
}

static void
bmp_tooltips_set_active_widget (BmpTooltips *tooltips,
                                GtkWidget   *widget)
{
  if (tooltips->tip_window)
    {
      if (GTK_WIDGET_VISIBLE (tooltips->tip_window))
	g_get_current_time (&tooltips->last_popdown);
      gtk_widget_hide (tooltips->tip_window);
      if (handler_id != 0)
    {
	g_signal_handler_disconnect (previous_widget, handler_id);
	gtk_widget_queue_draw (previous_widget);
	handler_id = 0;
    }
      }
  if (tooltips->timer_tag)
    {
      g_source_remove (tooltips->timer_tag);
      tooltips->timer_tag = 0;
    }

  tooltips->active_tips_data = NULL;

  if (widget)
    {
      GList *list;

      for (list = tooltips->tips_data_list; list; list = list->next)
	{
	  BmpTooltipsData *tooltipsdata;

	  tooltipsdata = list->data;

	  if (tooltipsdata->widget == widget &&
	      GTK_WIDGET_DRAWABLE (widget))
	    {
	      tooltips->active_tips_data = tooltipsdata;
	      break;
	    }
	}

    if (tooltips->draw_hotspot)
    {
	handler_id =
	    g_signal_connect_data (widget,
                                   "expose-event",
                                   G_CALLBACK (on_tooltip_widget_expose),
                                   GINT_TO_POINTER(tooltips->active_tips_data->rectangular),
                                   NULL,
                                   G_CONNECT_AFTER);
	previous_widget = widget;

	gtk_widget_queue_draw (widget);
    }

    }
  else
    {
      tooltips->use_sticky_delay = FALSE;

      if (handler_id != 0)
    {
	g_signal_handler_disconnect (previous_widget, handler_id);
	gtk_widget_queue_draw (previous_widget);
	handler_id = 0;
    }

    }
}

static void
bmp_tooltips_show_tip (GtkWidget *widget)
{
  BmpTooltipsData *tooltipsdata;

  tooltipsdata = bmp_tooltips_data_get (widget);

  if (tooltipsdata &&
      (!tooltipsdata->tooltips->active_tips_data ||
       tooltipsdata->tooltips->active_tips_data->widget != widget))
    {
      bmp_tooltips_set_active_widget (tooltipsdata->tooltips, widget);
      bmp_tooltips_draw_tips (tooltipsdata->tooltips);
    }
}

static void
bmp_tooltips_hide_tip (GtkWidget *widget)
{
  BmpTooltipsData *tooltipsdata;

  tooltipsdata = bmp_tooltips_data_get (widget);

  if (tooltipsdata &&
      (tooltipsdata->tooltips->active_tips_data &&
       tooltipsdata->tooltips->active_tips_data->widget == widget))
    bmp_tooltips_set_active_widget (tooltipsdata->tooltips, NULL);
}

static gboolean
bmp_tooltips_recently_shown (BmpTooltips *tooltips)
{
  GTimeVal now;
  glong msec;

  g_get_current_time (&now);
  msec = (now.tv_sec  - tooltips->last_popdown.tv_sec) * 1000 +
	  (now.tv_usec - tooltips->last_popdown.tv_usec) / 1000;
  return (msec < STICKY_REVERT_DELAY);
}

static gboolean
get_keyboard_mode (GtkWidget *widget)
{
  GtkWidget *toplevel = gtk_widget_get_toplevel (widget);
  if (GTK_IS_WINDOW (toplevel))
    return GPOINTER_TO_UINT (g_object_get_data (G_OBJECT (toplevel), "gtk-tooltips-keyboard-mode"));
  else
    return FALSE;
}

static void
start_keyboard_mode (GtkWidget *widget)
{
  GtkWidget *toplevel = gtk_widget_get_toplevel (widget);
  if (GTK_IS_WINDOW (toplevel))
    {
      GtkWidget *focus = GTK_WINDOW (toplevel)->focus_widget;

      g_object_set_data (G_OBJECT (toplevel), "gtk-tooltips-keyboard-mode", GUINT_TO_POINTER (TRUE));

      if (focus)
	bmp_tooltips_show_tip (focus);
    }
}

static void
stop_keyboard_mode (GtkWidget *widget)
{
  GtkWidget *toplevel = gtk_widget_get_toplevel (widget);
  if (GTK_IS_WINDOW (toplevel))
    {
      GtkWidget *focus = GTK_WINDOW (toplevel)->focus_widget;
      if (focus)
	bmp_tooltips_hide_tip (focus);

      g_object_set_data (G_OBJECT (toplevel), "gtk-tooltips-keyboard-mode", GUINT_TO_POINTER (FALSE));
    }
}

static void
bmp_tooltips_start_delay (BmpTooltips *tooltips,
			  GtkWidget   *widget)
{
  BmpTooltipsData *old_tips_data;

  old_tips_data = tooltips->active_tips_data;
  if (tooltips->enabled &&
      (!old_tips_data || old_tips_data->widget != widget))
    {
      guint delay;

      bmp_tooltips_set_active_widget (tooltips, widget);

      if (tooltips->use_sticky_delay &&
	  bmp_tooltips_recently_shown (tooltips))
	delay = STICKY_DELAY;
      else
	delay = tooltips->delay;
      tooltips->timer_tag = g_timeout_add (delay,
					   bmp_tooltips_timeout,
					   (gpointer) tooltips);
    }
}

static void
bmp_tooltips_event_handler (GtkWidget *widget,
                            GdkEvent  *event)
{
  BmpTooltips *tooltips;
  BmpTooltipsData *old_tips_data;
  GtkWidget *event_widget;
  gboolean keyboard_mode = get_keyboard_mode (widget);

  if ((event->type == GDK_LEAVE_NOTIFY || event->type == GDK_ENTER_NOTIFY) &&
      event->crossing.detail == GDK_NOTIFY_INFERIOR)
    return;

  old_tips_data = bmp_tooltips_data_get (widget);
  tooltips = old_tips_data->tooltips;

  if (keyboard_mode)
    {
      switch (event->type)
	{
	case GDK_FOCUS_CHANGE:
	  if (event->focus_change.in)
	    bmp_tooltips_show_tip (widget);
	  else
	    bmp_tooltips_hide_tip (widget);
	  break;
	default:
	  break;
	}
    }
  else
    {
      if (event->type != GDK_KEY_PRESS && event->type != GDK_KEY_RELEASE)
	{
	  event_widget = gtk_get_event_widget (event);
	  if (event_widget != widget)
	    return;
	}

      switch (event->type)
	{
	case GDK_EXPOSE:
	  /* do nothing */
	  break;
	case GDK_ENTER_NOTIFY:
	  if (!(GTK_IS_MENU_ITEM (widget) && GTK_MENU_ITEM (widget)->submenu))
	    bmp_tooltips_start_delay (tooltips, widget);
	  break;

	case GDK_LEAVE_NOTIFY:
	  {
	    gboolean use_sticky_delay;

	    use_sticky_delay = tooltips->tip_window &&
	      GTK_WIDGET_VISIBLE (tooltips->tip_window);
	    bmp_tooltips_set_active_widget (tooltips, NULL);
	    tooltips->use_sticky_delay = use_sticky_delay;
	  }
	  break;

	case GDK_MOTION_NOTIFY:
	  /* Handle menu items specially ... pend popup for each motion
	   * on other widgets, we ignore motion.
	   */
	  if (GTK_IS_MENU_ITEM (widget) && !GTK_MENU_ITEM (widget)->submenu)
	    {
	      /* Completely evil hack to make sure we get the LEAVE_NOTIFY
	       */
	      GTK_PRIVATE_SET_FLAG (widget, GTK_LEAVE_PENDING);
	      bmp_tooltips_set_active_widget (tooltips, NULL);
	      bmp_tooltips_start_delay (tooltips, widget);
	      break;
	    }
	  break;		/* ignore */
	case GDK_BUTTON_PRESS:
	case GDK_BUTTON_RELEASE:
	case GDK_KEY_PRESS:
	case GDK_KEY_RELEASE:
	case GDK_PROXIMITY_IN:
	case GDK_SCROLL:
	  bmp_tooltips_set_active_widget (tooltips, NULL);
	  break;
	default:
	  break;
	}
    }
}

static void
bmp_tooltips_widget_unmap (GtkWidget *widget,
			   gpointer   data)
{
  BmpTooltipsData *tooltipsdata = (BmpTooltipsData *)data;
  BmpTooltips *tooltips = tooltipsdata->tooltips;

  if (tooltips->active_tips_data &&
      (tooltips->active_tips_data->widget == widget))
    bmp_tooltips_set_active_widget (tooltips, NULL);
}

static void
bmp_tooltips_widget_remove (GtkWidget *widget,
			    gpointer   data)
{
  BmpTooltipsData *tooltipsdata = (BmpTooltipsData*) data;
  BmpTooltips *tooltips = tooltipsdata->tooltips;

  bmp_tooltips_widget_unmap (widget, data);
  tooltips->tips_data_list = g_list_remove (tooltips->tips_data_list,
					    tooltipsdata);
  bmp_tooltips_destroy_data (tooltipsdata);
}

void
_bmp_tooltips_toggle_keyboard_mode (GtkWidget *widget)
{
  if (get_keyboard_mode (widget))
    stop_keyboard_mode (widget);
  else
    start_keyboard_mode (widget);
}

/**
 * bmp_tooltips_get_info_from_tip_window:
 * @tip_window: a #GtkWindow
 * @tooltips: the return location for the tooltips which are displayed
 *    in @tip_window, or %NULL
 * @current_widget: the return location for the widget whose tooltips
 *    are displayed, or %NULL
 *
 * Determines the tooltips and the widget they belong to from the window in
 * which they are displayed.
 *
 * This function is mostly intended for use by accessibility technologies;
 * applications should have little use for it.
 *
 * Return value: %TRUE if @tip_window is displaying tooltips, otherwise %FALSE.
 *
 * Since: 2.4
 **/
gboolean
bmp_tooltips_get_info_from_tip_window (GtkWindow    *tip_window,
                                       BmpTooltips **tooltips,
                                       GtkWidget   **current_widget)
{
  BmpTooltips  *current_tooltips;
  gboolean has_tips;

  g_return_val_if_fail (GTK_IS_WINDOW (tip_window), FALSE);

  current_tooltips = g_object_get_data (G_OBJECT (tip_window), tooltips_info_key);

  has_tips = current_tooltips != NULL;

  if (tooltips)
    *tooltips = current_tooltips;
  if (current_widget)
    *current_widget = has_tips ? current_tooltips->active_tips_data->widget : NULL;

  return has_tips;
}

#define __BMP_TOOLTIPS_C__
