/*  BMPx - The Dumb Music Player
 *  Copyright (C) 2005 BMPx development team.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * The BMPx project hereby grant permission for non-gpl compatible GStreamer
 * plugins to be used and distributed together with GStreamer and BMPx. This
 * permission are above and beyond the permissions granted by the GPL license
 * BMPx is covered by.
 */


#include <config.h>

#include <stdio.h>
#ifndef __USE_BSD
# define __USE_BSD	/* Needed for mkdtemp () */
#endif

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <strings.h>

#ifdef ENABLE_NLS
# include <libintl.h>
#endif

#include <bmp/objects.h>
#include <bmp/util.h>
#include <bmp/uri.h>
#include <bmp/vfs.h>
#include <bmp/plugin.h>
#include <bmp/plugin_interfaces.h>

#include "libcontainer_pls.h"

#ifdef HAVE_FTS_H
#  include <fts.h>
#endif

static VFSHandle *handle = NULL;

gboolean
container_pls_can_process (BmpPluginContainerInterface *self, const gchar *uri)
{
    BMPURI *bmp_uri;
    gboolean rv;

    bmp_uri = bmp_uri_new (uri);
    rv = FALSE;
    
    if (str_has_suffix_nocase (bmp_uri->path, "pls")) rv = TRUE;
    bmp_uri_delete (bmp_uri);

    return rv;
}

gboolean
container_pls_handle_set (BmpPluginContainerInterface *self, VFSHandle *_handle)
{
    if (_handle == NULL) 
	return FALSE;
    
    if (!_handle->valid)
	return FALSE;

    handle = _handle;

    return TRUE;
}

VFSHandle*
container_pls_handle_get (BmpPluginContainerInterface *self)
{
    return handle;
}

gboolean
container_pls_handle_remove (BmpPluginContainerInterface *self)
{
    handle = NULL;

    return TRUE;
}

gboolean 
container_pls_handle_write (BmpPluginContainerInterface *self)
{
    GString *out;
    GError *error = NULL;
    gint n = 1;

    g_return_val_if_fail (handle != NULL, FALSE);
    g_return_val_if_fail (VFS_HANDLE_VALID(handle), FALSE);

    if (handle->buffer) g_free (handle->buffer);
    handle->buffer = NULL;

    out = g_string_new ("");

    g_string_append (out, "[playlist]\n");
    g_string_append_printf (out, "numberofentries=%d\n", g_list_length (handle->uri_list));

    for ( ; handle->uri_list ; handle->uri_list = handle->uri_list->next)
	{
	  BMPURI *uri;	
	  gchar *uri_fq;

	  uri = bmp_uri_new ((gchar*)handle->uri_list->data);
	  bmp_uri_unescape (uri);
	  uri_fq = bmp_uri_get_string (uri);

	  g_string_append_printf (out, "File%d=%s\n", n, uri_fq);

	  g_free (uri_fq);
	  bmp_uri_delete (uri);

	  g_string_append_printf (out, "Title%d=\n", n);
	  g_string_append_printf (out, "Length%d=-1\n", n++);
	}

    g_string_append (out, "Version=2\n");
    
    handle->buffer = g_strdup(out->str);
    g_string_free (out, TRUE);

    if (!g_file_set_contents (handle->filename, handle->buffer, strlen(handle->buffer), &error)) 
      { 
	g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "%s: Unable to save PLS '%s': %s", G_STRLOC, handle->filename, error->message);
	g_error_free (error);
	error = NULL;
	return FALSE;
      }

    return TRUE;
}

gboolean 
container_pls_handle_read (BmpPluginContainerInterface *self)
{
    GList *uri_list = NULL;
    gchar *num_entries = NULL;
    gint num_entries_i, n;

    g_return_val_if_fail (handle != NULL, FALSE);
    g_return_val_if_fail (VFS_HANDLE_VALID(handle), FALSE);

    if ((!handle->buffer) || (!strlen(handle->buffer)))
      {
	handle->buffer = NULL;
	return TRUE;
      }

    num_entries =
	      read_ini_string_buffer (handle->buffer, "playlist", "numberofentries");
    if (!num_entries)
      {
        handle->uri_list = NULL;
	return FALSE;
      }

    num_entries_i =
	      strtol (num_entries, NULL, 10);
    g_free (num_entries);

    for (n = 0; n < num_entries_i; n++)
      {
	gchar *value, *key;

	key =
	    g_strdup_printf ("File%d", n+1);
	value =
	    read_ini_string_buffer (handle->buffer, "playlist", key);

	uri_list =
	    g_list_append (uri_list, g_strdup (value));
      }

    handle->uri_list = uri_list;
    return TRUE;
}

const gchar**
container_pls_get_export_data (BmpPluginContainerInterface *self)
{
	static gchar* _export_data[] = {
		  "PLS Playlist", // Description
		  "pls"		  // Default suffix
	};

	return (const gchar**)_export_data;
}


static void
container_pls_interface_init (gpointer         g_iface,
                              gpointer         iface_data)
{
        BmpPluginContainerInterface *klass = (BmpPluginContainerInterface*)g_iface;

	klass->can_process     = container_pls_can_process;
	klass->handle_set      = container_pls_handle_set;
	klass->handle_get      = container_pls_handle_get;
	klass->handle_remove   = container_pls_handle_remove;
	klass->handle_read     = container_pls_handle_read;
	klass->handle_write    = container_pls_handle_write;
	klass->get_export_data = container_pls_get_export_data;
}

/* Metadata IFace */
static const gchar**
get_metadata (BmpPlugMetadataInterface *self)
{
  static gchar* _metadata[] = {
    "PLS Playlist Plugin",
    "Milosz Derezynski",
    "(C) GPL 2005",
    "http://bmpx.berlios.de",
    "internalerror@gmail.com",
  };

  return (const gchar**)_metadata;
}

static void
metadata_init (gpointer         g_iface,
               gpointer         iface_data)
{
        BmpPlugMetadataInterface *klass = (BmpPlugMetadataInterface *)g_iface;

        klass->get_metadata = get_metadata;
}


GType plugin_container_pls_type = 0;

void 
plugin_container_pls_register_type (GTypeModule *module)
{
	    static const GTypeInfo object_info = {
                        sizeof (BmpPluginContainerPLSClass),
                        NULL,   /* base_init */
                        NULL,   /* base_finalize */
                        NULL,   /* class_init */
                        NULL,   /* class_finalize */
                        NULL,   /* class_data */
                        sizeof (BmpPluginContainerPLS),
                        0,      /* n_preallocs */
                        NULL    /* instance_init */
                };

                plugin_container_pls_type = g_type_module_register_type (module,
					     G_TYPE_OBJECT,
					     "BmpPluginContainerPLSType",
					     &object_info, 0);

                static const GInterfaceInfo interface_container_pls_info = {
                        (GInterfaceInitFunc) container_pls_interface_init,    /* interface_init */
                        NULL,                                       /* interface_finalize */
                        NULL                                        /* interface_data */
                };

                static const GInterfaceInfo interface_metadata_info = {
                        (GInterfaceInitFunc) metadata_init,    /* interface_init */
                        NULL,                                       /* interface_finalize */
                        NULL                                        /* interface_data */
                };

                g_type_module_add_interface (module,
					     plugin_container_pls_type,
                                             BMP_TYPE_PLUG_METADATA,
                                             &interface_metadata_info);


                g_type_module_add_interface (module,
					     plugin_container_pls_type,
                                             BMP_TYPE_PLUGIN_CONTAINER,
                                             &interface_container_pls_info);
}
