/*  BMPx - The Dumb Music Player
 *  Copyright (C) 2005 BMPx development team.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * The BMPx project hereby grant permission for non-gpl compatible GStreamer
 * plugins to be used and distributed together with GStreamer and BMPx. This
 * permission are above and beyond the permissions granted by the GPL license
 * BMPx is covered by.
 */

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <strings.h>

#ifdef ENABLE_NLS
# include <libintl.h>
#endif

#include <ne_basic.h>
#include <ne_compress.h>
#include <ne_socket.h>
#include <ne_utils.h>
#include <ne_props.h>
#include <ne_session.h>
#include <ne_request.h>

#include <bmp/util.h>
#include <bmp/uri.h>
#include <bmp/vfs.h>
#include <bmp/plugin.h>
#include <bmp/plugin_interfaces.h>

#include "libtransport_http.h"

typedef struct _Options Options;
struct _Options {
    GList   *specs;
  
    GQuark   timeout_q;
    gint     timeout;
  
    GQuark   use_proxy_q;
    gboolean use_proxy;

    GQuark   proxy_host_q;
    gchar   *proxy_host;

    GQuark   proxy_port_q;
    gint     proxy_port;
};

#define _OPTION_TIMEOUT	      "timeout"
#define _OPTION_USE_PROXY     "useproxy"
#define _OPTION_PROXY_HOST    "proxyhost"
#define _OPTION_PROXY_PORT    "proxyport"

static VFSHandle *handle  = NULL;
static Options   *options = NULL;

gboolean
transport_http_can_process (BmpPluginTransportInterface *self, const gchar *uri)
{
    if (!g_ascii_strncasecmp("http",uri,4))
	return TRUE;

    return FALSE;
}

gboolean
transport_http_handle_set (BmpPluginTransportInterface *self, VFSHandle *_handle)
{
    if (_handle == NULL) 
	return FALSE;
    
    if (!_handle->valid)
	return FALSE;

    handle = _handle;

    return TRUE;
}

VFSHandle*
transport_http_handle_get (BmpPluginTransportInterface *self)
{
    return handle;
}

gboolean
transport_http_handle_remove (BmpPluginTransportInterface *self)
{
    handle = NULL;

    return TRUE;
}

typedef struct _ReadHandle ReadHandle;
struct _ReadHandle {
    BMPURI    *uri;
    GMutex  *handle_lock;
    GMutex  *thread_lock;
    GCond   *cond;

    char    *memory;
    size_t   size;
};
#define READ_HANDLE(_data) ((ReadHandle*)(_data))

static int
read_block (void *data, const char *buf, size_t len)
{
  ReadHandle *rhandle = READ_HANDLE(data); 
 
  if (!len)
    {
	g_mutex_unlock (rhandle->handle_lock);
	return 0;
    }

  rhandle->memory = realloc (rhandle->memory, rhandle->size+len);	
  memcpy  (rhandle->memory+rhandle->size, buf, len); 
  rhandle->size += len;

  return 0;
}

gpointer
neon_read_thread (gpointer data)
{
    ne_session			     *sess;
    ne_request			     *req;

    gchar			     *path;

    ReadHandle			     *rhandle = READ_HANDLE(data);

    BMPURI			     *uri = rhandle->uri;

    if (!uri->port) uri->port = 80;

    path = g_strconcat (uri->path, "?", uri->query, NULL);
    sess = ne_session_create (uri->scheme, uri->hostname, uri->port);

    ne_set_useragent (sess, "BMPx/0.13");
    ne_set_read_timeout (sess, options->timeout);

    if (options->use_proxy)
      {
        g_message ("%s: setting proxy %s port %d", G_STRFUNC, options->proxy_host, options->proxy_port);
	ne_session_proxy (sess, options->proxy_host, options->proxy_port);
      }

    req = ne_request_create (sess, "GET", path); 

    ne_add_response_body_reader (req, ne_accept_always, read_block, rhandle); 

    g_mutex_lock (rhandle->handle_lock);
    ne_request_dispatch (req);

    /* wait while we busy fetching the data and make sure signals are passed */
    while (!g_mutex_trylock(rhandle->handle_lock))
      {
	while (gtk_events_pending ())
	  gtk_main_iteration ();
      }

    g_mutex_unlock (rhandle->handle_lock);
    g_mutex_free (rhandle->handle_lock);

    g_free (path);
    ne_request_destroy (req);
    ne_session_destroy (sess);

    g_mutex_lock (rhandle->thread_lock);
    g_cond_signal (rhandle->cond);
    g_mutex_unlock (rhandle->thread_lock);

    g_thread_exit (0);

    return NULL;
}

gboolean
transport_http_handle_read (BmpPluginTransportInterface *self)
{
    GThread			     *thread;
    BMPURI			     *uri;
    ReadHandle			     *rhandle;
    GTimeVal			      _tv;

    uri = bmp_uri_new (handle->uri);

    if ((!uri->path) || (strlen(uri->path)<=1))
      {
	bmp_uri_delete (uri);
	handle->buffer = NULL;
	handle->uri_list = g_list_append (handle->uri_list, g_strdup(handle->uri));
	handle->need_container = FALSE;
	return TRUE;
      }

    rhandle = g_new0 (ReadHandle,1);
    rhandle->memory = NULL;  
    rhandle->size = 0;
    rhandle->handle_lock = g_mutex_new ();
    rhandle->thread_lock = g_mutex_new ();
    rhandle->uri = uri;
    rhandle->cond = g_cond_new ();

    g_mutex_lock (rhandle->thread_lock);
    thread = g_thread_create (neon_read_thread,
			      rhandle,
			      FALSE,
			      NULL);

    while (42)
      {
	while (gtk_events_pending ())
	  gtk_main_iteration ();

	g_get_current_time (&_tv);  
	g_time_val_add (&_tv, 10000);

        if (g_cond_timed_wait(rhandle->cond, rhandle->thread_lock, &_tv))
          break;
      }

    g_mutex_unlock (rhandle->thread_lock);
    g_mutex_free (rhandle->thread_lock);
    g_cond_free (rhandle->cond);
    bmp_uri_delete (rhandle->uri);

    handle->buffer = rhandle->memory;
    handle->buffer_size = rhandle->size;

    g_free (rhandle);

    return TRUE;
}

static void
transport_http_interface_init (gpointer         g_iface,
                               gpointer         iface_data)
{
        BmpPluginTransportInterface *klass = (BmpPluginTransportInterface *)g_iface;

	klass->can_process    = transport_http_can_process;
	klass->handle_set     = transport_http_handle_set;
	klass->handle_get     = transport_http_handle_get;
	klass->handle_remove  = transport_http_handle_remove;
	klass->handle_read    = transport_http_handle_read;

	ne_sock_init ();
}

static GList*
prop_get_props (BmpPlugPropertyInterface *self)
{
	return options->specs;
}

static gboolean
prop_set_prop  (BmpPlugPropertyInterface *self,
		GQuark		          prop,
	        GValue			 *value)
{
    if (prop == options->timeout_q)
      {
	  options->timeout = g_value_get_int (value);
	  return TRUE;
      }

    if (prop == options->use_proxy_q)
      {
	  options->use_proxy = g_value_get_boolean (value);
	  return TRUE;
      }

    if (prop == options->proxy_host_q)
      {
	  if (options->proxy_host) g_free (options->proxy_host);
	  options->proxy_host = g_value_dup_string (value);
	  return TRUE;
      }

    if (prop == options->proxy_port_q)
      {
	  options->proxy_port_q = g_value_get_int (value);
	  return TRUE;
      }

    return FALSE;
}

static void
prop_init (gpointer         g_iface,
           gpointer         iface_data)
{
        BmpPlugPropertyInterface *klass = (BmpPlugPropertyInterface *)g_iface;

        klass->get_props = prop_get_props;
	klass->set_prop  = prop_set_prop;

	options = g_new0 (Options,1);
	options->specs = NULL;

	/* Network timeout */ 
	options->specs = g_list_append (options->specs,
				        g_param_spec_int (_OPTION_TIMEOUT,
							  _OPTION_TIMEOUT,
							  "Network timeout value in seconds",
							  30, 600, 30, BMP_PLUG_PROPERTY_PARAM_FLAGS)); 
	options->timeout = 30;
	options->timeout_q = g_quark_from_static_string (_OPTION_TIMEOUT);

	/* Use HTTP Proxy */ 
	options->specs = g_list_append (options->specs,
				        g_param_spec_boolean (_OPTION_USE_PROXY,
							      _OPTION_USE_PROXY,
							      "Use HTTP Proxy",
							       FALSE, BMP_PLUG_PROPERTY_PARAM_FLAGS)); 
	options->use_proxy   = FALSE;
	options->use_proxy_q = g_quark_from_static_string (_OPTION_USE_PROXY);

	/* HTTP Proxy Hostname */ 
	options->specs = g_list_append (options->specs,
				        g_param_spec_string  (_OPTION_PROXY_HOST,
							      _OPTION_PROXY_HOST,
							      "Proxy Hostname",
							      "proxy.example.com", BMP_PLUG_PROPERTY_PARAM_FLAGS)); 
	options->proxy_host   = g_strdup ("proxy.example.com"); 
	options->proxy_host_q = g_quark_from_static_string (_OPTION_PROXY_HOST);

	/* HTTP Proxy Port */ 
	options->specs = g_list_append (options->specs,
				        g_param_spec_int     (_OPTION_PROXY_PORT,
							      _OPTION_PROXY_PORT,
							      "Proxy Port",
							      0, 65535, 3128, BMP_PLUG_PROPERTY_PARAM_FLAGS)); 

	options->proxy_port   = 3128;
	options->proxy_port_q = g_quark_from_static_string (_OPTION_PROXY_PORT);

}

static gchar* _metadata[] = {
    "HTTP Transport",
    "Milosz Derezynski",
    "(C) GPL 2005",
    "http://bmpx.beep-media-player.org",
    "internalerror@gmail.com",
};

const gchar**
get_metadata (BmpPlugMetadataInterface *self)
{
    return (const gchar**)_metadata;
}

static void
metadata_init (gpointer         g_iface,
               gpointer         iface_data)
{
        BmpPlugMetadataInterface *klass = (BmpPlugMetadataInterface *)g_iface;

        klass->get_metadata = get_metadata;
}



GType plugin_transport_http_type = 0;

void 
plugin_transport_http_register_type (GTypeModule *module)
{
	    static const GTypeInfo object_info = {
                        sizeof (BmpPluginTransportHttpClass),
                        NULL,   /* base_init */
                        NULL,   /* base_finalize */
                        NULL,   /* class_init */
                        NULL,   /* class_finalize */
                        NULL,   /* class_data */
                        sizeof (BmpPluginTransportHttp),
                        0,      /* n_preallocs */
                        NULL    /* instance_init */
                };

                static const GInterfaceInfo interface_transport_http_info = {
                        (GInterfaceInitFunc) transport_http_interface_init,
                        NULL,						
                        NULL					
                };

                static const GInterfaceInfo interface_property_info = {
                        (GInterfaceInitFunc) prop_init,		   
                        NULL,                                     
                        NULL                                     
                };

                static const GInterfaceInfo interface_metadata_info = {
                        (GInterfaceInitFunc) metadata_init,    
                        NULL,                                 
                        NULL                                 
                };

                plugin_transport_http_type = g_type_module_register_type (module,
					     G_TYPE_OBJECT,
					     "BmpPluginTransportHttpType",
					     &object_info, 0);

                g_type_module_add_interface (module,
					     plugin_transport_http_type,
                                             BMP_TYPE_PLUG_METADATA,
                                             &interface_metadata_info);

                g_type_module_add_interface (module,
					     plugin_transport_http_type,
                                             BMP_TYPE_PLUGIN_TRANSPORT,
                                             &interface_transport_http_info);

                g_type_module_add_interface (module,
					     plugin_transport_http_type,
                                             BMP_TYPE_PLUG_PROPERTY,
                                             &interface_property_info);

}
