/*  BMPx - The Dumb Music Player
 *  Copyright (C) 2005 BMPx development team.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  This file also contains some code originating in the XMMS project and is
 *  copyright the XMMS project authors.
 *
 *  $Id$
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

#include <bmp/util.h>
#include <bmp/uri.h>
#include <bmp/file_utils.h>

#ifndef __USE_BSD
# define __USE_BSD	/* Needed for mkdtemp () */
#endif

#include "main.h"

#ifdef HAVE_FTS_H
#  include <fts.h>
#endif

#include <glib/gstdio.h>

typedef gchar * (*ArchiveExtractFunc) (const gchar *, const gchar *);

typedef enum
{
  ARCHIVE_UNKNOWN = 0,
  ARCHIVE_DIR,
  ARCHIVE_TAR,
  ARCHIVE_TGZ,
  ARCHIVE_ZIP,
  ARCHIVE_TBZ2
}
ArchiveType;

typedef struct
{
  ArchiveType type;
  const gchar *ext;
}
ArchiveExtensionType;

static ArchiveExtensionType archive_extensions[] =
{
  {ARCHIVE_TAR, "tar"},
  {ARCHIVE_ZIP, "wsz"},
  {ARCHIVE_ZIP, "zip"},
  {ARCHIVE_TGZ, "tar.gz"},
  {ARCHIVE_TGZ, "tgz"},
  {ARCHIVE_TBZ2, "tar.bz2"},
  {ARCHIVE_TBZ2, "bz2"},
  {ARCHIVE_UNKNOWN, NULL}
};

static gchar *archive_extract_tar (const gchar * archive,
				   const gchar * dest);
static gchar *archive_extract_zip (const gchar * archive,
				   const gchar * dest);
static gchar *archive_extract_tgz (const gchar * archive,
				   const gchar * dest);
static gchar *archive_extract_tbz2 (const gchar * archive,
				    const gchar * dest);

static ArchiveExtractFunc archive_extract_funcs[] =
{
  NULL,
  NULL,
  archive_extract_tar,
  archive_extract_tgz,
  archive_extract_zip,
  archive_extract_tbz2
};

/*
 * escape_shell_chars ()
 *
 * Escapes characters that are special to the shell inside double quotes.
 */
static gchar*
escape_shell_chars (const gchar * string)
{
  const gchar *special = "$`\"\\";    /* Characters to escape */
  const gchar *in = string;
  gchar *out, *escaped;
  gint num = 0;

  while (*in != '\0')
    if (strchr (special, *in++))
      num++;

  escaped = g_malloc (strlen (string) + num + 1);

  in = string;
  out = escaped;

  while (*in != '\0')
    {
      if (strchr (special, *in))
        *out++ = '\\';
      *out++ = *in++;
    }

  *out = '\0';

  return escaped;
}


/*
 * find <file> in directory <dirname> or subdirectories.  return
 * pointer to complete filename which has to be freed by calling
 * "g_free ()" after use. Returns NULL if file could not be found.
 */

typedef struct
{
  const gchar *to_match;
  gchar       *match;
  gboolean     found;
  gint	       depth_max;
  gint	       depth;
}
FindFileContext;

static gboolean
find_file_func (const gchar *path,
                const gchar *basename,
                gpointer     data)
{
  FindFileContext *context = data;

  context->depth++;

  if (strlen (path) > FILENAME_MAX)
    {
      g_warning (G_STRLOC ": ignoring path: name too long (%s)", path);
      return TRUE;
    }

  if (g_file_test (path, G_FILE_TEST_IS_REGULAR) && (path[0] != '.'))
    {
      if (! g_ascii_strcasecmp (basename, context->to_match))
        {
          context->match = g_strdup (path);
          context->found = TRUE;
          return TRUE;
        }
    }
  else if ( g_file_test (path, G_FILE_TEST_IS_DIR) && ((context->depth_max != -1) && (context->depth < context->depth_max)) && (path[0] != '.'))
    {
      dir_foreach (path, find_file_func, context, NULL);
      if (context->found)
        return TRUE;
    }

  return FALSE;
}

gchar *
find_file_recursively (const gchar *path,
                       const gchar *filename,
		       gint depth_max)
{
  FindFileContext context;

  context.to_match = filename;
  context.match = NULL;
  context.found = FALSE;
  context.depth = -1;
  context.depth_max = depth_max;

  dir_foreach (path, find_file_func, &context, NULL);
  return context.match;
}

/* FIXME: these functions can be generalised into a function using a
 * command lookup table */

static gchar *archive_extract_tar (const gchar * archive,
				   const gchar * dest)
{
  return g_strdup_printf ("%s >/dev/null xf \"%s\" -C %s", "tar",
                          archive, dest);
}

static gchar *archive_extract_zip (const gchar * archive,
				   const gchar * dest)
{
  return g_strdup_printf ("%s >/dev/null -o -j \"%s\" -d %s", "unzip",
                          archive, dest);
}

static gchar *archive_extract_tgz (const gchar * archive,
				   const gchar * dest)
{
  return g_strdup_printf ("%s >/dev/null xzf \"%s\" -C %s", "tar",
                          archive, dest);
}

static gchar *archive_extract_tbz2 (const gchar * archive,
				    const gchar * dest)
{
  return g_strdup_printf ("bzip2 -dc \"%s\" | %s >/dev/null xf - -C %s",
                          archive, "tar", dest);
}


ArchiveType archive_get_type (const gchar * filename)
{
    gint i = 0;

    if (g_file_test (filename, G_FILE_TEST_IS_DIR))
	return ARCHIVE_DIR;

    while (archive_extensions[i].ext)
	{
	    if (g_str_has_suffix (filename, archive_extensions[i].ext))
		{
		    return archive_extensions[i].type;
		}
	    i++;
	}

    return ARCHIVE_UNKNOWN;
}

gboolean
file_is_archive (const gchar *filename)
{
    return (archive_get_type (filename) > ARCHIVE_DIR);
}

gchar*
archive_basename (const gchar *str)
{
  gint i = 0;

  while (archive_extensions[i].ext)
    {
      if (str_has_suffix_nocase (str, archive_extensions[i].ext))
      {
        const gchar *end = g_strrstr (str, archive_extensions[i].ext);
        if (end)
	  /* The -1 is for the '.' in the extension */
          return g_strndup (str, end - str - 1);
      }
      i++;
    }

  return NULL;
}

gchar*
create_glob (const gchar *suffix)
{
  GString *out;
  gint n = 0;

  out = g_string_new ("");

  while (suffix[n])
    {
      g_string_append_printf (out, "[%c%c]", g_ascii_tolower (suffix[n]), g_ascii_toupper (suffix[n]));
      n++;
    }

  return g_string_free (out, FALSE); 
}

/*
   decompress_archive

   Decompresses the archive "filename" to a temporary directory,
   returns the path to the temp dir, or NULL if failed, watch out tho,
   doesn't actually check if the system command succeeds :-|
   */
gchar *
archive_decompress (const gchar *filename)
{
  gchar *tmpdir, *cmd, *escaped_filename;
  ArchiveType type;

  if ((type = archive_get_type (filename)) <= ARCHIVE_DIR)
    return NULL;

  tmpdir = g_build_filename (g_get_tmp_dir (), "bmp.XXXXXXXX", NULL);
  if (!mkdtemp (tmpdir))
    {
      g_message ("Unable to decompress archive '%s': could not create "
                 "temporary directory: %s", filename, g_strerror (errno));
      g_free (tmpdir);
      return NULL;
    }

  escaped_filename = escape_shell_chars (filename);
  cmd = (*archive_extract_funcs[type]) (escaped_filename, tmpdir);
  g_free (escaped_filename);

  if (!cmd)
  {
    g_message (G_STRLOC ": extraction function is NULL!");
    g_free (tmpdir);
    return NULL;
  }

  system (cmd);
  g_free (cmd);

  return tmpdir;
}


#ifdef HAVE_FTS_H

void
del_directory (const gchar *dirname)
{
  gchar *const argv[2] = { (gchar *) dirname, NULL };
  FTS *fts;
  FTSENT *p;

  fts = fts_open (argv, FTS_PHYSICAL, (gint (*) ()) NULL);
  while ((p = fts_read (fts)))
    {
      switch (p->fts_info)
        {
          case FTS_D:
            break;
          case FTS_DNR:
	  case FTS_ERR:
            break;
	  case FTS_DP:
            rmdir (p->fts_accpath);
            break;
          default:
            unlink (p->fts_accpath);
	    break;
        }
    }

  fts_close (fts);
}

#else                           /* !HAVE_FTS */

static gboolean
del_directory_func (const gchar *path,
                    const gchar *basename,
                    gpointer     params)
{
  if (!strcmp (basename, ".") || !strcmp (path, ".."))
    return FALSE;

  if (g_file_test (path, G_FILE_TEST_IS_DIR))
    {
      dir_foreach (path, del_directory_func, NULL, NULL);
      rmdir (path);
      return FALSE;
    }

  unlink (path);

  return FALSE;
}

void
del_directory (const gchar *path)
{
  dir_foreach (path, del_directory_func, NULL, NULL);
  rmdir (path);
}

#endif                          /* ifdef HAVE_FTS */

gboolean
collect_path (const gchar  *path,
	      const gchar  *basename,
              GList       **items)
{
  if (g_file_test (path, G_FILE_TEST_IS_REGULAR))
    {
      if (!bmp_play_is_audio_file (bmp_play_engine, basename))
        return FALSE;

      *items = g_list_append (*items, g_strdup (path));
    }
  else if (g_file_test (path, G_FILE_TEST_IS_DIR))
    {
      dir_foreach (path, (DirForeachFunc) collect_path, items, NULL);
    }

  return FALSE;
}


gboolean
dir_foreach (const gchar    *path,
             DirForeachFunc  function,
	     gpointer        user_data,
             GError        **error)
{
  GDir *dir;
  const gchar *entry;
  gchar *entry_fullpath;

  if (!(dir = g_dir_open (path, 0, error)))
    return FALSE;

  while ((entry = g_dir_read_name (dir)))
    {
      entry_fullpath = g_build_filename (path, entry, NULL);

      if ((*function) (entry_fullpath, entry, user_data))
        {
          g_free (entry_fullpath);
          break;
        }

      g_free (entry_fullpath);
    }

  g_dir_close (dir);

  return TRUE;
}

void
remove_path (const gchar *path)
{
  GError    *error = NULL;
  GDir	    *dir;
  gboolean   proceed = TRUE;

  dir = g_dir_open (path, 0, &error);

  while (proceed)  
    {
	const gchar *entry;
	gchar       *path_fq;

	entry = g_dir_read_name (dir);
	if (!entry)
	  {
	    proceed = FALSE;
	    break;
	  }
    
	path_fq = g_build_filename (path, entry, NULL);

	if (g_file_test (path_fq, G_FILE_TEST_IS_DIR))
	    remove_path (path_fq);

	g_remove (path_fq);
	g_free (path_fq);
    }

  g_dir_close (dir);
}

void
copy_file (const gchar  *source,
	   const gchar  *dest)
{

#define CHUNK_SIZE 524288 /* 512k */

    FILE	*FH_SRC, 
		*FH_DEST;
    size_t	 size_r,
		 size_w;
    gboolean     proceed = TRUE;

    FH_SRC  = fopen (source, "r"); 
    FH_DEST = fopen (dest,   "w");

    while (proceed)
      {
	 gchar	*mem;

	 mem = g_malloc0 (CHUNK_SIZE);
	 size_r = fread (mem, 1, CHUNK_SIZE, FH_SRC); 
	 while (gtk_events_pending ()) gtk_main_iteration();
  
	 if ((size_r < CHUNK_SIZE) && (feof (FH_SRC) || ferror (FH_SRC)))
	    {
		proceed = FALSE;
	    }

	size_w = fwrite (mem, 1, size_r, FH_DEST);
	while (gtk_events_pending ()) gtk_main_iteration();

	g_free (mem);
      }

    fclose (FH_SRC);
    fclose (FH_DEST);
}

