/*  BMPx - The Dumb Music Player
 *  Copyright (C) 2005 BMPx development team.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  $Id$
 *
 * The BMPx project hereby grant permission for non-gpl compatible GStreamer
 * plugins to be used and distributed together with GStreamer and BMPx. This
 * permission are above and beyond the permissions granted by the GPL license
 * BMPx is covered by.
 */


#include <config.h>

#include <glib.h>
#include <gmodule.h>
#include <stdlib.h>
#include <glib/gi18n.h>

#include <bmp/util.h>
#include <bmp/plugin.h>
#include <bmp/plugin_interfaces.h>

#include "gc.h"
#include "main.h"

#include "loader.h"

GHashTable   *plugins_table[BMP_N_PLUGIN_TYPES];

enum {
    LIB_BASENAME,
    LIB_PLUGNAME,
    LIB_SUFFIX
};

#define IS_MODULE(basename) (str_has_suffix_nocase(basename, G_MODULE_SUFFIX))

typedef struct {
    gchar *name;
    gint   table_type;
} ForeachTuple;

gboolean load_plugins (const gchar *path,
		       const gchar *basename,
		       gpointer     user_data)
{
    GObject   *plugin_instance;
    BmpPlugin *plugin;

    ForeachTuple *tuple = (ForeachTuple*)user_data;

    gchar **subs, **_save;
    gchar  *mpath, *dir, *name;

    NGC *gc;

    if (!IS_MODULE(basename))
	return FALSE;

    gc = gc_new();
    gc_mark (gc);

    g_message ("%s: %s plugin: %s", G_STRLOC, tuple->name, basename);
    _save = subs = gc_collect_d (gc, g_strsplit_set (basename, "_.", -1), (GDestroyNotify)g_strfreev);

    dir =
	gc_collect (gc, g_build_filename (PLUGIN_DIR,tuple->name,NULL));
    name =
	gc_collect (gc, g_strconcat (tuple->name,"_",subs[LIB_PLUGNAME],NULL));
    mpath = 
	gc_collect (gc, g_module_build_path (dir, name));

    plugin =
        bmp_plugin_new (mpath);	

    if (!plugin)
      return FALSE;

    g_type_module_use (G_TYPE_MODULE(plugin));

    if (!bmp_plugin_valid (plugin))
      {
	gc_free (gc);
	gc_destroy (gc);
	return FALSE;
      }

    plugin_instance =
	bmp_plugin_create_instance (plugin);

#if 0
    if (BMP_IS_PLUG_PROPERTY(plugin_instance))
      {
	GList *props = bmp_plug_property_get_props (BMP_PLUG_PROPERTY(plugin_instance));

	for ( ; props ; props = props->next)
	  {
	      GParamSpec *spec = G_PARAM_SPEC(props->data);

	      if (G_PARAM_SPEC_VALUE_TYPE(spec) == G_TYPE_INT)
		{
		    g_message ("INT Property, name '%s', description '%s'", g_param_spec_get_name (spec),
									     g_param_spec_get_blurb (spec));

		    g_message ("\tDefault value:\t\t%d", G_PARAM_SPEC_INT(spec)->default_value);
		    g_message ("\tMinimum:\t\t%d", G_PARAM_SPEC_INT(spec)->minimum);
		    g_message ("\tMaximum:\t\t%d", G_PARAM_SPEC_INT(spec)->maximum);
		}

	      if (G_PARAM_SPEC_VALUE_TYPE(spec) == G_TYPE_STRING)
		{
		    g_message ("STRING Property, name '%s', description '%s'", g_param_spec_get_name (spec),
									     g_param_spec_get_blurb (spec));

		    g_message ("\tDefault value:\t\t%d", G_PARAM_SPEC_STRING(spec)->default_value);
		}

	  }
      }
#endif

    g_hash_table_insert (plugins_table[tuple->table_type], g_strdup(subs[LIB_PLUGNAME]), plugin_instance);

    gc_free (gc); 
    gc_destroy (gc);
    return FALSE;
}


static gchar* plugin_type_names[] = {
    "flow",
    "container",
    "transport",
    NULL,
};

static gchar* plugin_type_descs[] = {
    N_("Flow"),
    N_("Container"),
    N_("Transport"),
    NULL,
};

const gchar*
plugin_type_get_name (PluginType type)
{
    return plugin_type_names[type];
}

const gchar*
plugin_type_get_desc (PluginType type)
{
    return _(plugin_type_descs[type]);
}

void
plugins_initialize (void)
{
    GError *error = NULL;
    gint n;

    
    for (n = 0; n < BMP_N_PLUGIN_TYPES; n++)
      {
	 plugins_table[n] = 
	    g_hash_table_new (g_str_hash, g_str_equal);
      }
   
    for (n = 0; plugin_type_names[n]; n++)
     {
	ForeachTuple *tuple;
	gchar *path;

	path = 
	    g_build_filename (PLUGIN_DIR, plugin_type_names[n], NULL);
	tuple =
	    g_new0 (ForeachTuple,1);
	
	tuple->name = plugin_type_names[n];
	tuple->table_type = n;

	dir_foreach (path, load_plugins, tuple, &error);

	if (error)
	  {
	    g_message ("%s: %s", G_STRLOC, error->message);
	    g_error_free (error);
	    error = NULL;
	  }

	g_free (tuple);
	g_free (path);
    }
}

void
plugins_shutdown (void)
{
#if 0
    GError *error = NULL;
    gint n;

    n = 0;
    while (plugin_type_names[n])
     {
	n++;
     }
#endif
}

/* Tracklist exporting */
void
container_h_tuple_free (gpointer data)
{
    ContainerHTuple *tuple = (ContainerHTuple*)data;

    g_free (tuple->suffix);
    g_free (tuple->name);
    g_free (tuple->desc);
    g_free (tuple);
}

static void
for_each_container (gpointer key, gpointer value, GList **containers)
{
    BmpPluginContainerInterface *container = (BmpPluginContainerInterface*)value;

    if (BMP_PLUGIN_CONTAINER_GET_IFACE (container)->get_export_data != NULL) 
      {
	ContainerHTuple *tuple = g_new0 (ContainerHTuple,1);	
	const gchar** info;

	info = bmp_plugin_container_get_export_data (container);

	tuple->name = g_strdup (key);
	tuple->desc = g_strdup (info[0]);
	tuple->suffix = g_strdup (info[1]);
    
	*containers = g_list_append (*containers, tuple); 
      }
}

GList*
plugins_find_containers (void)
{
  GList *containers = NULL;

  g_hash_table_foreach (plugins_table[BMP_PLUGIN_TYPE_CONTAINER], (GHFunc)for_each_container, &containers);

  return containers;
}
