/*  BMPx - The Dumb Music Player
 *  Copyright (C) 2005 BMPx development team.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  $Id$
 *
 * The BMPx project hereby grant permission for non-gpl compatible GStreamer
 * plugins to be used and distributed together with GStreamer and BMPx. This
 * permission are above and beyond the permissions granted by the GPL license
 * BMPx is covered by.
 */

/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

#include <config.h>

#include <glib.h>
#include <glib/gi18n.h>
#include <glib/gprintf.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <time.h>
#include <unistd.h>
#include <string.h>

#include "main.h"
#include "logger.h"

G_DEFINE_TYPE (BmpLogger, bmp_logger, G_TYPE_OBJECT)

struct _BmpLoggerPrivate
{
        gboolean     dispose_has_run;
	GHashTable  *domains;
	gchar	    *timestr;
};

typedef struct _BmpLogInfo BmpLogInfo;
struct _BmpLogInfo {
    gchar	    *domain;
    gchar	    *filename;
    FILE	    *file;
    GMutex	    *lock;
    GLogLevelFlags   flags;
    guint	     id;
};

/* GObject */
static void
bmp_logger_init (BmpLogger *self)
{
    self->private = g_new (BmpLoggerPrivate, 1);
    self->private->dispose_has_run = FALSE;
    self->private->domains = g_hash_table_new (g_str_hash, g_str_equal);
    self->private->timestr = get_timestamp_str ();
}

static GObject *
bmp_logger_constructor (GType                  type,
                        guint                  n_construct_properties,
                        GObjectConstructParam *construct_properties)
{
        GObject *obj;

        {
                BmpLoggerClass *klass;
                GObjectClass *parent_class;
                klass = BMP_LOGGER_CLASS (g_type_class_peek (BMP_TYPE_LOGGER));
                parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
                obj = parent_class->constructor (type,
                                                 n_construct_properties,
                                                 construct_properties);
        }

        return obj;
}

static void
for_each_logfile (gpointer _k, gpointer _v, gpointer _d)
{
  BmpLogInfo  *info = (BmpLogInfo *)_v;

  g_log (info->domain, G_LOG_LEVEL_INFO, "Logging stopped for [%s] at [%s]", info->domain, get_timestamp_str());
  g_log_remove_handler (info->domain, info->id);

  if (info->file)
    {
      const gchar *end = "\n";
      fwrite (end, sizeof (gchar), strlen (end), info->file);
    }

  fclose (info->file);

  g_free (info->filename);
  g_free (info->domain);
  g_free (info);
}

static void
bmp_logger_dispose (GObject *obj)
{
        BmpLogger *self = (BmpLogger *)obj;

        if (self->private->dispose_has_run) return;
        self->private->dispose_has_run = TRUE;

	g_hash_table_foreach (self->private->domains, (GHFunc)for_each_logfile, self);
}

static void
bmp_logger_finalize (GObject *obj)
{
        BmpLogger *self = (BmpLogger *)obj;
        g_free (self->private);
}

typedef struct _LogHandler LogHandler;
struct _LogHandler
{
  gchar		  *domain;
  gchar		  *file;
  GLogLevelFlags   level;
  guint		   id;
};
static LogHandler log_handlers[] = {
  {BMPX_SYSTEM_LOG_DOMAIN,	  "system.log",   LOG_ALL_LEVELS},
  {"Glib",			  "glib.log",     LOG_ALL_LEVELS},
  {"Gtk",			  "gtk.log",      LOG_ALL_LEVELS},
  {BMPX_SCROBBLER_LOG_DOMAIN,	  "lastfm.log",   LOG_ALL_LEVELS}
};
static guint log_handler_count = G_N_ELEMENTS (log_handlers);


BmpLogger*
bmp_logger_new (void)
{
  BmpLogger *self;
  gint i;

  self = g_object_new (bmp_logger_get_type (), NULL);

  for (i = 0; i < log_handler_count; i++)
    {
      bmp_logger_add_log (self,
			  log_handlers[i].domain,
			  log_handlers[i].file,
			  log_handlers[i].level);
    }

  return self;
}

static void
bmp_logger_class_init (BmpLoggerClass *g_class)
{
        GObjectClass *gobject_class = G_OBJECT_CLASS (g_class);
        gobject_class->dispose = bmp_logger_dispose;
        gobject_class->finalize = bmp_logger_finalize;
        gobject_class->constructor = bmp_logger_constructor;
}

/* Logger */
gchar*
get_timestamp_str (void)
{
  time_t     current_time;
  struct tm  current_tm;
  gchar     *stamp;
  gchar      tmp[256];
  size_t     len;

  current_time = time (NULL);
  localtime_r (&current_time, &current_tm);
  len = strftime (tmp, sizeof (tmp), "%F %T", &current_tm);

  stamp = g_malloc0 (len + 1);

  strncpy (stamp, tmp, len);
  g_strstrip (stamp);

  return stamp;
}

static size_t
get_filesize (const gchar *filename)
{
  struct stat info;

  if (stat (filename, &info))
    return 0;

  return info.st_size;
}

static void
log_to_file (const gchar    *domain,
             GLogLevelFlags  level,
             const gchar    *message,
             gpointer        data)
{
  BmpLogInfo *info = (BmpLogInfo *)data;
  FILE  *file = info->file;
  gchar *timestamp;

  if (!file)
    {
      g_printerr (G_STRLOC ": file is NULL!\n");
      return;
    }

  g_mutex_lock (info->lock);

  timestamp = get_timestamp_str ();
  g_fprintf (file, "%s", timestamp);
  g_free (timestamp);

  if (domain)
    g_fprintf (file, " %s: ", domain);

  if (message)
    g_fprintf (file, "%s\n", message);
  else
    g_fprintf (file, "(null)\n");

  fflush (file);

  g_mutex_unlock (info->lock);
}

gboolean
bmp_logger_add_log (BmpLogger		*self,
		    const gchar		*domain,
		    const gchar		*filename,
		    GLogLevelFlags	 flags)
{
  BmpLogInfo *info;

  g_return_val_if_fail (filename != NULL, FALSE);

  info = g_new0 (BmpLogInfo, 1);
  info->lock = g_mutex_new ();
  info->filename = g_build_filename (BMP_PATH_LOGS_DIR, filename, NULL);
  info->domain = g_strdup (domain);
  info->flags = flags;

  /* truncate file when size limit is reached */
  if (get_filesize (info->filename) < BMP_LOGGER_FILE_MAX_SIZE)
    info->file = fopen (info->filename, "at");
  else
    info->file = fopen (info->filename, "w+t");

  if (!info->file)
    {
      g_printerr (_("Unable to create log file '%s' for domain '%s'!\n"), info->filename, domain);
      return FALSE;
    }

  info->id = g_log_set_handler (domain,
				flags,
				log_to_file,
				info);
  g_hash_table_insert (self->private->domains, g_strdup (domain), info);

  g_log (domain, G_LOG_LEVEL_INFO, "Logging started for [%s] at [%s], PID[%d]", domain, self->private->timestr, getpid());
  return TRUE;
}


