/*  BMPx - The Dumb Music Player
 *  Copyright (C) 2003-2006 BMPx development team.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  $Id$
 *
 * The BMPx project hereby grant permission for non-gpl compatible GStreamer
 * plugins to be used and distributed together with GStreamer and BMPx. This
 * permission are above and beyond the permissions granted by the GPL license
 * BMPx is covered by.
 */

#include <config.h>

#ifdef HAVE_GUI
#include <gtk/gtk.h>
#else
#include <glib.h>
#endif

#include <glib/gi18n.h>
#include <glib/gstdio.h>

#include <unistd.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <sys/stat.h>

#include <tag_c.h>
#include <musicbrainz/mb_c.h>

#include <libhrel/types.h>
#include <libhrel/relation.h>

#include <bmp/metadata.h>

G_DEFINE_TYPE (BmpMetadataCache, bmp_metadata_cache, G_TYPE_OBJECT)

#include <bmp/util.h>
#include <bmp/file_utils.h>
#include <bmp/metadata_common.h>
#include <bmp/rtfs.h>
#include <bmp/uri.h>

#ifdef HAVE_HAL
#  include <bmp/hal.h>
#endif

#include <libxml/tree.h>
#include <libxml/parser.h>
#include <libxml/xmlreader.h>
#include <libxml/xpath.h>
#include <libxml/xpathInternals.h>

#include <xcs/xml.h>

#include "bookmarks.h"
#include "bmp-marshalers.h"
#include "main.h"

enum {
    BMP_METADATA_CACHE_SIGNAL_INSERT_TUPLE,
    BMP_METADATA_CACHE_SIGNAL_DELETE_TUPLE,

    BMP_METADATA_CACHE_SIGNAL_TAG_ADDED,
    BMP_METADATA_CACHE_SIGNAL_TAG_DELETED,

    BMP_METADATA_CACHE_SIGNAL_TAG_VALUE_INSERT,
    BMP_METADATA_CACHE_SIGNAL_TAG_VALUE_REMOVE,

    BMP_METADATA_N_SIGNALS
};
static guint signals[BMP_METADATA_N_SIGNALS] = {0};

enum {
    LIST_ARTIST,
    LIST_ALBUM
};

struct _BmpMetadataCachePrivate {
        gboolean dispose_has_run;

	RTFS	    *rtfs;
	HRelation   *relation[N_REALMS];
	GHashTable  *tags;
};

struct _BmpMetadatumDefine {

    gchar *title;
    gchar *title_plural;
    gchar *description;
    gchar *id;
    GType  type;

    gboolean filter;

    gint   default_int;
    gchar* default_char;
};
typedef struct _BmpMetadatumDefine BmpMetadatumDefine;

static GHashTable  *bmp_metadatum_define_name_to_id = NULL;

/* We're using GST-style tag identifiers here */
static BmpMetadatumDefine bmp_metadatum_defines[] = {
    { N_("Location"),          N_("Locations"),    N_("The location of the track (file, internet stream, etc.)"),   "location",       G_TYPE_STRING, TRUE , 0, N_("(Unknown Location)")},
    { N_("Artist"),            N_("Artists"),	   N_("The artist/peformer of the track"),			    "artist",         G_TYPE_STRING, TRUE,  0, N_("(Unknown Artist)")},
    { N_("Album"),             N_("Albums"),	   N_("The album of the track"),				    "album",          G_TYPE_STRING, TRUE,  0, N_("(Unknown Album)")},
    { N_("Title"),             N_("Titles"),	   N_("The title of the current track"),			    "title",          G_TYPE_STRING, TRUE,  0, N_("(Unknown Title)")},
    { N_("Track"),             N_("Tracks"),	   N_("The track number of the track"),				    "track-number",   G_TYPE_INT,    FALSE, 0, ""},
    { N_("Time"),              N_("Times"),	   N_("The length of the track"),				    "time",           G_TYPE_INT,    FALSE, 0, ""}, 
    { N_("Genre"),             N_("Genres"),	   N_("The genre (kind of music) of the track"),		    "genre",          G_TYPE_STRING, FALSE, 0, N_("(Unknown Genre)")},
    { N_("Comment"),           N_("Comments"),	   N_("Comments added to the track"),				    "comment",        G_TYPE_STRING, FALSE, 0, N_("(Unknown Comment)")},
    { N_("Rating"),            N_("Ratings"),	   N_("The rating of the track"),				    "rating",         G_TYPE_INT,    FALSE, 0, ""},
    { N_("Date"),              N_("Dates"),	   N_("The date (year) of the track"),				    "date",           G_TYPE_INT,    FALSE, 0, ""},
    { N_("MTIME"),	       N_("MTIMEs"),	   N_("The modification time of the track"),			    "mtime",	      G_TYPE_LONG  , FALSE, 0, ""},
    { N_("Bitrate"),           N_("Bitrates"),	   N_("The bitrate of the track"),				    "bitrate",        G_TYPE_INT,    FALSE, 0, ""},
    { N_("Samplerate"),        N_("Samplerates"),  N_("The samplerate of the track"),				    "samplerate",     G_TYPE_INT,    FALSE, 0, ""},
    { N_("Play count"),        N_("Play counts"),  N_("The number of times this track was played"),		    "count",	      G_TYPE_INT,    FALSE, 0, ""},
    { N_("Volume UDI"),        N_("Volume UDIs"),  N_("HAL Volume UDI this track resides on"),			    "volume-udi",     G_TYPE_STRING, FALSE, 0, ""}, 
    { N_("Device UDI"),        N_("Device UDIs"),  N_("HAL Device UDI this track resides on"),			    "device-udi",     G_TYPE_STRING, FALSE, 0, ""}, 
};

static gchar* bmp_tag_defines[] = {
    "artist",
    "album",
    "title",
};

static gint bmp_tag_defines_to_datum_map[] = {
    -1, 
    BMP_TAG_ARTIST, 
    BMP_TAG_ALBUM,
    BMP_TAG_TITLE,
    -1, 
    -1, 
    -1, 
    -1, 
    -1, 
    -1, 
    -1, 
    -1, 
    -1, 
    -1, 
};

static gint bmp_datum_to_tag_defines_map[] = {
    BMP_DATUM_ARTIST,
    BMP_DATUM_ALBUM,
    BMP_DATUM_TITLE,
};

static void
tags_save (BmpMetadataCache *self);

static void 
tags_load (BmpMetadataCache *self);
 
gint
bmp_metadata_datum_to_tag_id (gint n)
{
    return bmp_tag_defines_to_datum_map[n];
}

gint
bmp_metadata_tag_id_to_datum (gint n)
{
    return bmp_datum_to_tag_defines_map[n];
}

const gchar*
bmp_metadata_get_tag_define (gint n)
{
  return (const gchar*)bmp_tag_defines[n];
}

/* RowGUID */
gboolean
row_guid_equal_func (RowGUID *guid_a, RowGUID *guid_b)
{
    g_return_val_if_fail (guid_a != NULL, FALSE);
    g_return_val_if_fail (guid_b != NULL, FALSE);
  
    if ((guid_a->counter == guid_b->counter) &&
        (guid_a->hash    == guid_b->hash))
    return TRUE;

    return FALSE;
}

guint
row_guid_hash_func (RowGUID *guid)
{
    guint  hash;
    gchar *hash_str;

    g_return_val_if_fail (guid != NULL, FALSE);

    hash_str = g_strdup_printf ("%u %u", guid->counter, guid->hash); 
    hash = g_str_hash (hash_str);
    g_free (hash_str);

    return hash;
}

static guint guid_counter = 0;

/* FileItem */
BmpFileItem*
bmp_file_item_new (const gchar	    *uri,
		   MetadataRealm     realm)
{
    BmpFileItem *item;
    RowGUID	*row_guid;

    g_return_val_if_fail (uri != NULL, NULL);

    item = g_new0(BmpFileItem,1);
    row_guid = g_new0(RowGUID,1);

    row_guid->counter =	guid_counter++;
    row_guid->hash = g_str_hash(uri);

    item->uri    = g_strdup (uri);
    item->guid	 = row_guid;
    item->realm  = realm;

    return item;
}

BmpFileItem*
bmp_file_item_copy (BmpFileItem *item)
{
    BmpFileItem *_copy;

    _copy = g_new0 (BmpFileItem, 1);

    _copy->uri = g_strdup (item->uri);
    _copy->guid = g_new0 (RowGUID, 1);
    _copy->guid->counter = item->guid->counter; 
    _copy->guid->hash = item->guid->hash;
    _copy->realm = item->realm;

    return _copy;
}

void
bmp_file_item_free (BmpFileItem *item)
{
    if (!item) return;

    g_free (item->uri);
    g_free (item->guid);
    g_free (item);
}

GType
bmp_file_item_get_type (void)
{
    static GType our_type = 0;

    if (our_type == 0)
        our_type = g_boxed_type_register_static ("BmpFileItem",
                                                 (GBoxedCopyFunc)
                                                 bmp_file_item_copy,
                                                 (GBoxedFreeFunc)
                                                 bmp_file_item_free);
    return our_type;
}
/* END FileItem */



/* Tags stuff */
static void
collect_tagnames (gpointer   k,
		  gpointer   v,
		  GList	   **list)
{
  *list = g_list_append (*list, g_strdup(k));
}

GList*
bmp_metadata_cache_tags_list (BmpMetadataCache *self)
{
  GList *tagnames = NULL;

  g_hash_table_foreach (self->private->tags, (GHFunc)collect_tagnames, &tagnames); 

  return tagnames;
}

gboolean
bmp_metadata_cache_tag_delete (BmpMetadataCache *self, 
			       const gchar      *tagname, 
			       GError	       **error)
{
  GHashTable *subs;

  g_return_val_if_fail (BMP_IS_METADATA_CACHE(self), FALSE);

  if (!g_hash_table_lookup (self->private->tags, tagname))  
    {
      *error = g_error_new (g_quark_from_static_string("BmpMetadataTags"),
			    BMP_TAGS_ENOEXIST,
			    "Tag doesn't exist");
      return FALSE;
    }

  subs = g_hash_table_lookup (self->private->tags, tagname);
  g_hash_table_remove (self->private->tags, tagname);
  g_hash_table_destroy (subs);

  g_signal_emit (G_OBJECT(self), signals[BMP_METADATA_CACHE_SIGNAL_TAG_DELETED], 0, tagname);
  
  return TRUE;
}

gboolean
bmp_metadata_cache_tag_new (BmpMetadataCache *self, 
			    const gchar	     *tagname, 
			    GError	    **error)
{
  GHashTable *subs;
  GHashTable *n_tags[BMP_N_TAGS];
  gint	      n;

  g_return_val_if_fail (BMP_IS_METADATA_CACHE(self), FALSE);

  if (g_hash_table_lookup (self->private->tags, tagname))  
    {
      *error = g_error_new (g_quark_from_static_string("BmpMetadataTags"),
			    BMP_TAGS_EEXISTS,
			    "Tag exists");
      return FALSE;
    }

  subs = g_hash_table_new (g_str_hash, g_str_equal);
  g_hash_table_insert (self->private->tags, g_strdup (tagname), subs);
 
  for (n = 0; n < BMP_N_TAGS; n++) 
    {
      n_tags[n] = g_hash_table_new (g_str_hash, g_str_equal);
      g_hash_table_insert (subs, g_strdup (bmp_tag_defines[n]), n_tags[n]);    
    }

  // created
  g_signal_emit (G_OBJECT(self), signals[BMP_METADATA_CACHE_SIGNAL_TAG_ADDED], 0, tagname);

  return TRUE;
}

gboolean
bmp_metadata_cache_tag_insert (BmpMetadataCache	  *self,
			       const gchar	  *tagname,
			       BmpMetadatumTagId   metadatum_tag_id,
			       TagDataTuple	  *data_tuple, 
			       GError	         **error)
{
  GHashTable *tag,
	     *sub;

  g_return_val_if_fail (BMP_IS_METADATA_CACHE(self), FALSE);

  if (*error != NULL)
    {
      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "GError is not NULL");
      return FALSE;
    }

  tag = g_hash_table_lookup (self->private->tags, tagname);  

  if (!tag)
    {
      *error = g_error_new (g_quark_from_static_string("BmpMetadataTags"),
			    BMP_TAGS_ENOEXIST,
			    "Tag does not exist");
      return FALSE;
    }

  sub = g_hash_table_lookup (tag, bmp_tag_defines[metadatum_tag_id]);

  if (!g_hash_table_lookup (sub, data_tuple->tag[metadatum_tag_id]))
    {
      g_hash_table_replace (sub, g_strdup (data_tuple->tag[metadatum_tag_id]), data_tuple); 
      g_signal_emit (G_OBJECT(self), signals[BMP_METADATA_CACHE_SIGNAL_TAG_VALUE_INSERT], 0, tagname, metadatum_tag_id, data_tuple); 
    }

  return TRUE;
}

gboolean
bmp_metadata_cache_tag_remove (BmpMetadataCache	  *self,
			       const gchar	  *tagname,
			       BmpMetadatumTagId   metadatum_tag_id,
			       TagDataTuple	  *data_tuple, 
			       GError	         **error)
{
  GHashTable *tag,
	     *sub;

  g_return_val_if_fail (BMP_IS_METADATA_CACHE(self), FALSE);

  if (*error != NULL)
    {
      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "GError is not NULL");
      return FALSE;
    }

  tag = g_hash_table_lookup (self->private->tags, tagname);  

  if (!tag)
    {
      *error = g_error_new (g_quark_from_static_string("BmpMetadataTags"),
			    BMP_TAGS_ENOEXIST,
			    "Tag does not exist");
      return FALSE;
    }

  sub = g_hash_table_lookup (tag, bmp_tag_defines[metadatum_tag_id]);

  g_hash_table_remove (sub, data_tuple->tag[metadatum_tag_id]);

  g_signal_emit (G_OBJECT(self), signals[BMP_METADATA_CACHE_SIGNAL_TAG_VALUE_REMOVE], 0, tagname, metadatum_tag_id, data_tuple); 

  return TRUE;
}

gboolean
bmp_metadata_cache_tag_check (BmpMetadataCache   *self,
			      const gchar       *tagname,
			      BmpMetadatumTagId	 metadatum_tag_id,
			      const gchar	*value,
			      gboolean		*exists,
			      GError	       **error)
{
  GHashTable *tag,
	     *subs;

  g_return_val_if_fail (BMP_IS_METADATA_CACHE(self), FALSE);

  if (*error != NULL)
    {
      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "GError is not NULL");
      return FALSE;
    }

  tag = g_hash_table_lookup (self->private->tags, tagname);  

  if (!tag)
    {
      *error = g_error_new (g_quark_from_static_string("BmpMetadataTags"),
			    BMP_TAGS_ENOEXIST,
			    "Tag does not exist");
      return FALSE;
    }

  subs = g_hash_table_lookup (tag, bmp_tag_defines[metadatum_tag_id]); 

  if (g_hash_table_lookup (subs, value))
    *exists = TRUE;
  else
    *exists = FALSE;

  return TRUE;
}

/* High-level Tags API */
typedef struct _FindTagsTuple FindTagsTuple;

struct _FindTagsTuple {
    HTuple		*tuple;
    BmpMetadatumTagId	 metadatum_tag_id;
    GList		*tags;
};

static void
find_tags_for_tuple (gpointer _k,
		     gpointer _v,
		     gpointer data)
{
    FindTagsTuple *ttuple  = (FindTagsTuple*)data;
    const gchar	  *tagname = (const gchar*)_k;
    GHashTable	  *sub     = (GHashTable*)_v;
    HTupleElem	  *elem;

    sub = g_hash_table_lookup (sub, bmp_tag_defines[ttuple->metadatum_tag_id]); 
    elem = h_tuple_get (ttuple->tuple, bmp_tag_defines[ttuple->metadatum_tag_id]); 

    if (g_hash_table_lookup (sub, elem->value))
      {
	ttuple->tags = g_list_append (ttuple->tags, g_strdup(tagname));
      }
}

GList*
bmp_metadata_cache_tuple_get_tags (BmpMetadataCache	    *self,
				   HTuple		    *tuple,
				   BmpMetadatumTagId	     metadatum_tag_id)
{
  g_return_val_if_fail (BMP_IS_METADATA_CACHE(self), FALSE);

  FindTagsTuple	  *ttuple   = g_new0 (FindTagsTuple,1);

  ttuple->tuple = tuple;
  ttuple->tags = NULL;

  g_hash_table_foreach (self->private->tags, (GHFunc) find_tags_for_tuple, ttuple);

  return ttuple->tags;
}


gboolean
bmp_metadata_cache_tuple_matches_tag (BmpMetadataCache	    *self,
				      HTuple		    *tuple,
				      BmpMetadatumTagId	     metadatum_tag_id,
				      const gchar	    *tagname) 
{
  gboolean tuple_does_match = TRUE;

  g_return_val_if_fail (BMP_IS_METADATA_CACHE(self), FALSE);

  HTupleElem  *elem;
  const gchar *id;
  const gchar *value;
  GError      *error = NULL;

  id = bmp_tag_defines[metadatum_tag_id];
  elem = h_tuple_get (tuple, id);

  value = elem->value;

  if (!bmp_metadata_cache_tag_check (self,
				     tagname, 
				     metadatum_tag_id, 
				     value, 
				     &tuple_does_match, 
				     &error)) 
    {	      
	  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "%s(%d: %s) Error during tuple/tags matching: %s", __FILE__, __LINE__, __FUNCTION__, error->message);
	  g_error_free (error);
	  error = NULL;
    }

   return tuple_does_match;
}

gboolean
bmp_metadata_cache_tag_tuple (BmpMetadataCache	    *self,
			      HTuple		    *tuple,
			      BmpMetadatumTagId	     metadatum_tag_id,
			      const gchar	    *tagname,
			      GError		   **error)
{
  GError	*error_local = NULL;
  GHashTable	*tag;
  HTupleElem	*elem;
  TagDataTuple	*data_tuple;
  gint		 n;

  g_return_val_if_fail (BMP_IS_METADATA_CACHE(self), FALSE);

  if (*error != NULL)
    {
      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "GError is not NULL");
      return FALSE;
    }

  tag = g_hash_table_lookup (self->private->tags, tagname);  

  if (!tag)
    {
      *error = g_error_new (g_quark_from_static_string("BmpMetadataTags"),
			    BMP_TAGS_ENOEXIST,
			    "Tag does not exist");
      return FALSE;
    }

  data_tuple = g_new0 (TagDataTuple,1);
  for (n = 0; n < BMP_N_TAGS; n++)
    {
	elem = h_tuple_get (tuple, bmp_tag_defines[n]);
	data_tuple->tag[n] = g_strdup (elem->value);
    }

  bmp_metadata_cache_tag_insert  (self,
			          tagname,
				  metadatum_tag_id,
				  data_tuple,			
				  &error_local);
				  

  return TRUE;
}

gboolean
bmp_metadata_cache_untag_tuple (BmpMetadataCache    *self,
			        HTuple		    *tuple,
			        BmpMetadatumTagId    metadatum_tag_id,
			        const gchar	    *tagname,
			        GError		   **error)
{
  GError	  *error_local = NULL;
  GHashTable	  *tag;
  HTupleElem	  *elem;
  TagDataTuple	  *data_tuple;
  gint		   n;

  g_return_val_if_fail (BMP_IS_METADATA_CACHE(self), FALSE);

  if (*error != NULL)
    {
      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "GError is not NULL");
      return FALSE;
    }

  tag = g_hash_table_lookup (self->private->tags, tagname);  

  if (!tag)
    {
      *error = g_error_new (g_quark_from_static_string("BmpMetadataTags"),
			    BMP_TAGS_ENOEXIST,
			    "Tag does not exist");
      return FALSE;
    }

  data_tuple = g_new0 (TagDataTuple,1);
  for (n = 0; n < BMP_N_TAGS; n++)
    {
	elem = h_tuple_get (tuple, bmp_tag_defines[n]);
	data_tuple->tag[n] = g_strdup (elem->value);
    }


  bmp_metadata_cache_tag_remove  (self,
			          tagname,
				  metadatum_tag_id,
				  data_tuple,		
  				  &error_local);
				  
  return TRUE;
}
/* Tags END! */

static GHashTable*
tuple_to_hash_table (HTuple *tuple)
{
  GHashTable *array;
  GValue *value;
  HTupleElem *elmt;
  const gchar *key_name;
  gint n;

  array = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free); 

  for (n = 0; n < G_N_ELEMENTS(bmp_metadatum_defines); n++)
    {
      key_name = 
	  bmp_metadata_get_id_static (n); 
      elmt = 
	  h_tuple_get (tuple, key_name); 

      value = g_new0 (GValue,1);
      g_value_init (value, elmt->type);

      if (elmt->type == G_TYPE_INT)
  	g_value_set_int (value, GPOINTER_TO_INT(elmt->value));
      else if (elmt->type == G_TYPE_STRING)
	g_value_set_string (value, (gchar*)(elmt->value));

      g_hash_table_insert (array, g_strdup(key_name), value); 
    }

  return array;
}

GHashTable*
bmp_metadata_cache_get_hash_value_table (BmpMetadataCache	  *self,
					 BmpFileItem		  *item) 
{
  HTuple      *tuple;
  GHashTable  *metadata_ht;

  tuple = bmp_metadata_cache_metadata_get (self, item);
  if (!tuple) return NULL;

  metadata_ht = tuple_to_hash_table (tuple);
  h_tuple_free (tuple);
  return metadata_ht;
}


gint
bmp_metadata_get_id_by_name (const gchar *id_string)
{
    return GPOINTER_TO_INT (g_hash_table_lookup (bmp_metadatum_define_name_to_id, id_string)); 
}

gchar*
bmp_metadata_get_id (BmpMetadatumId metadatum_id)
{
    return g_strdup(bmp_metadatum_defines[metadatum_id].id);
}

GType
bmp_metadata_get_type (BmpMetadatumId metadatum_id)
{
    return bmp_metadatum_defines[metadatum_id].type;
}


const gchar*
bmp_metadata_get_id_static (BmpMetadatumId metadatum_id)
{
    return bmp_metadatum_defines[metadatum_id].id;
}

gchar*
bmp_metadata_get_title (BmpMetadatumId metadatum_id)
{
    return g_strdup(_(bmp_metadatum_defines[metadatum_id].title));
}

const gchar*
bmp_metadata_get_title_static (BmpMetadatumId metadatum_id)
{
    return _(bmp_metadatum_defines[metadatum_id].title);
}

gchar*
bmp_metadata_get_title_plural (BmpMetadatumId metadatum_id)
{
    return g_strdup(_(bmp_metadatum_defines[metadatum_id].title_plural));
}

const gchar*
bmp_metadata_get_default_char (BmpMetadatumId metadatum_id)
{
    return g_strdup(_(bmp_metadatum_defines[metadatum_id].default_char));
}

gint
bmp_metadata_get_default_int (BmpMetadatumId metadatum_id)
{
    return bmp_metadatum_defines[metadatum_id].default_int;
}

gboolean
bmp_metadata_get_filter (BmpMetadatumId metadatum_id)
{
    return bmp_metadatum_defines[metadatum_id].filter;
}


HTuple*
bmp_metadata_tuple_create_default (const gchar *uri)
{
    HTuple *tuple;
    gint    n;

    tuple = h_tuple_new (); 
    for (n = 0; n < G_N_ELEMENTS(bmp_metadatum_defines); n++)
      {
	GType type = bmp_metadata_get_type (n);
	h_tuple_insert (tuple,
			bmp_metadata_get_id_static (n), 
			type, 
			(type == G_TYPE_INT) ? GINT_TO_POINTER (bmp_metadata_get_default_int (n)) 
			: ""); 

      }

    return tuple;
}

typedef enum {
    MRR_OK,
    MRR_FILE_NOT_FOUND,
    MRR_UNABLE_TO_READ,
} MetadataReadResult;

static MetadataReadResult 
taglib_metadata_get	(BmpMetadataCache	*self,
		         const gchar		*uri,
			 HTuple		       **tuple) 
{
    TagLib_File			  *file = NULL;
    TagLib_Tag			  *tag  = NULL;
    TagLib_AudioProperties const  *audio = NULL;
    BMPURI			  *bmp_uri;
    struct stat			   file_stat;
    const gchar			  *aux;
    gchar			  *filename, 
				  *aux_check,
				  *uri_str;

    bmp_uri = bmp_uri_new (uri);
    bmp_uri_unescape (bmp_uri);
    filename = bmp_uri->path;

    if (!filename)
      {
	bmp_uri_delete (bmp_uri);
	return MRR_FILE_NOT_FOUND; 
      }

    uri_str = bmp_uri_get_string (bmp_uri);
    *tuple = bmp_metadata_tuple_create_default (uri_str);
    g_free (uri_str);

    stat (filename, &file_stat);

    h_tuple_insert (*tuple, 
		    bmp_metadata_get_id_static (BMP_DATUM_MTIME),    
		    G_TYPE_LONG,    
		    (gpointer)((glong)file_stat.st_mtime));

    h_tuple_insert (*tuple, 
		    bmp_metadata_get_id_static (BMP_DATUM_LOCATION), 
		    G_TYPE_STRING, 
		    (gpointer)uri); 

    file = taglib_file_new (filename);
    if (file)
      {
         tag = taglib_file_tag (file);
         if (!tag)
           {
	      taglib_file_free (file);
	      return MRR_UNABLE_TO_READ; 
           }

         audio = taglib_file_audioproperties (file);
	 h_tuple_insert (*tuple, 
			 bmp_metadata_get_id_static (BMP_DATUM_TRACK),    
			 G_TYPE_INT,    
			 GINT_TO_POINTER(taglib_tag_track (tag))); 

	 h_tuple_insert (*tuple, 
			 bmp_metadata_get_id_static (BMP_DATUM_DATE),     
			 G_TYPE_INT,    
			 GINT_TO_POINTER(taglib_tag_year  (tag))); 

	 /* G_TYPE_STRING metadata */
	 aux = taglib_tag_artist (tag);
	 aux_check = aux ? g_strstrip(g_strdup(aux)) : NULL; 
	 if (aux && strlen(aux_check))
	   {
	      h_tuple_insert (*tuple, 
			      bmp_metadata_get_id_static (BMP_DATUM_ARTIST), 
			      G_TYPE_STRING, 
			      (gpointer)aux); 
	   }
	 g_free (aux_check);

	 aux = taglib_tag_album (tag);
	 aux_check = aux ? g_strstrip(g_strdup(aux)) : NULL; 
	 if (aux && strlen(aux_check))
	   {
	      h_tuple_insert (*tuple, 
			      bmp_metadata_get_id_static (BMP_DATUM_ALBUM), 
			      G_TYPE_STRING, 
			      (gpointer)aux); 
	   }
	 g_free (aux_check);

	 aux = taglib_tag_title (tag);
	 aux_check = aux ? g_strstrip(g_strdup(aux)) : NULL; 
	 if (aux && strlen(aux_check))
	   {
	      h_tuple_insert (*tuple, 
			      bmp_metadata_get_id_static (BMP_DATUM_TITLE), 
			      G_TYPE_STRING, 
			      (gpointer)aux); 
	   }
	  else
	   {
	      gchar  *basename;
	      gchar **fragments;

	      basename = g_path_get_basename (bmp_uri->path); 
	      fragments = g_strsplit (basename, ".", -1);

	      h_tuple_insert (*tuple,
			      bmp_metadata_get_id_static (BMP_DATUM_TITLE),
			      G_TYPE_STRING,
			      (gpointer)fragments[0]);
#if 0
	      g_strfreev (fragments);
	      g_free (basename);
#endif
	    }
	 g_free (aux_check);

	 aux = taglib_tag_genre (tag);
	 aux_check = aux ? g_strstrip(g_strdup(aux)) : NULL; 
	 if (aux && strlen(aux_check))
	   {
	      h_tuple_insert (*tuple, 
			      bmp_metadata_get_id_static (BMP_DATUM_GENRE), 
			      G_TYPE_STRING, 
			      (gpointer)aux); 
	   }
	 g_free (aux_check);

	 aux = taglib_tag_comment (tag);
	 aux_check = aux ? g_strstrip(g_strdup(aux)) : NULL; 
	 if (aux && strlen(aux_check))
	   {
	      h_tuple_insert (*tuple, 
			      bmp_metadata_get_id_static (BMP_DATUM_COMMENT), 
			      G_TYPE_STRING, 
			      (gpointer)aux); 
	   }
	 g_free (aux_check);

	 h_tuple_insert (*tuple, 
			 bmp_metadata_get_id_static (BMP_DATUM_RATING),    
			 G_TYPE_INT,    
			 GINT_TO_POINTER(0)); 

	 h_tuple_insert (*tuple, 
			 bmp_metadata_get_id_static (BMP_DATUM_COUNT),     
			 G_TYPE_INT,    
			 GINT_TO_POINTER(0));

	 h_tuple_insert (*tuple, 
			 bmp_metadata_get_id_static (BMP_DATUM_BITRATE),     
			 G_TYPE_INT,    
			 GINT_TO_POINTER(taglib_audioproperties_bitrate(audio)));

	 h_tuple_insert (*tuple, 
			 bmp_metadata_get_id_static (BMP_DATUM_SAMPLERATE),     
			 G_TYPE_INT,    
			 GINT_TO_POINTER(taglib_audioproperties_samplerate(audio)));

	 h_tuple_insert (*tuple, 
		         bmp_metadata_get_id_static (BMP_DATUM_TIME),     
			 G_TYPE_INT,    
			 GINT_TO_POINTER(taglib_audioproperties_length (audio))); 


         taglib_tag_free_strings ();
         taglib_file_free (file);
      }
    else
      {
	h_tuple_free (*tuple);
	bmp_uri_delete (bmp_uri);
	return MRR_UNABLE_TO_READ; 
      }

    bmp_uri_delete (bmp_uri);
    return MRR_OK; 
}

/* Private API */
static void
tracks_on_insert_tuple (HRelation     *relation,
			HTuple	      *tuple,
			gpointer       data)
{
    g_signal_emit (G_OBJECT(data), signals[BMP_METADATA_CACHE_SIGNAL_INSERT_TUPLE], 0, tuple); 
}

static void
tracks_on_delete_tuple (HRelation     *relation,
			HTuple	      *tuple,
			gpointer       data)
{
    g_signal_emit (G_OBJECT(data), signals[BMP_METADATA_CACHE_SIGNAL_DELETE_TUPLE], 0, tuple); 
}


/* Media Cache functions */
static gint
long_compare (gconstpointer a,
	      gconstpointer b)
{
  glong va, vb;

  va = (glong)(a);
  vb = (glong)(a);

  if (va == vb) return  0;
  if (va <  vb) return -1;
  
  return 1;
}

static gchar*
long_to_str (gconstpointer v)
	      
{
  return g_strdup_printf ("%ld", (glong)v);
}

static void
bmp_metadata_cache_init (BmpMetadataCache *self)
{
    HTypeInfo *info;
    gint n;

    self->private = g_new (BmpMetadataCachePrivate, 1);
    self->private->dispose_has_run = FALSE; 

    h_type_init ();

    /* We do this dynamically to accomodate for different size of glong on 32bit and 64bit platforms */
    info = g_new0 (HTypeInfo, 1);
    info->size	  = sizeof (G_TYPE_LONG);
    info->clone	  = (HCloneFunc) h_identity; 
    info->compare = (GCompareFunc) long_compare; 
    info->equal	  = g_direct_equal; 
    info->hash	  = g_direct_hash; 
    info->free	  = (GDestroyNotify) h_dummy; 
    info->str	  = long_to_str; 
    if (!h_type_add (G_TYPE_LONG, info)) g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "Failed to add G_TYPE_LONG to libhrel type system.");

    g_free (info);

    self->private->relation[REALM_PERMANENT] = create_relation_base_track (); 
    self->private->relation[REALM_TRANSIENT] = create_relation_base_track (); 

    g_signal_connect (self->private->relation[REALM_PERMANENT], "insert-tuple",
			  G_CALLBACK (tracks_on_insert_tuple),
			  self);

    g_signal_connect (self->private->relation[REALM_PERMANENT], "delete-tuple",
			  G_CALLBACK (tracks_on_delete_tuple),
			  self);


    bmp_metadatum_define_name_to_id = g_hash_table_new (g_str_hash, g_str_equal); 
    for (n = 0; n < G_N_ELEMENTS(bmp_metadatum_defines); n++)
      {
	      g_hash_table_insert (bmp_metadatum_define_name_to_id, bmp_metadata_get_id (n), GINT_TO_POINTER(n));
      }

    self->private->tags = NULL; 
    self->private->rtfs = rtfs_open (self->private->relation[REALM_PERMANENT], BMP_PATH_USER_DIR, "library", FALSE); 

    bookmarks_load ();

    g_message("running: %s", G_OBJECT_TYPE_NAME(self));
}

static void
bmp_metadata_cache_set_property (GObject      *object,
                        guint         property_id,
                        const GValue *value,
                        GParamSpec   *pspec)
{
        switch (property_id) {
        default:
                g_assert (FALSE);
                break;
        }
}

static void
bmp_metadata_cache_get_property (GObject      *object,
                        guint         property_id,
                        GValue       *value,
                        GParamSpec   *pspec)
{
        switch (property_id) {
        default:
                g_assert (FALSE);
                break;
        }
}

static GObject *
bmp_metadata_cache_constructor (GType                  type,
                       guint                  n_construct_properties,
                       GObjectConstructParam *construct_properties)
{
        GObject *obj;

        {
                /* Invoke parent constructor. */
                BmpMetadataCacheClass *klass;
                GObjectClass *parent_class;
                klass = BMP_METADATA_CACHE_CLASS (g_type_class_peek (BMP_TYPE_METADATA_CACHE));
                parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
                obj = parent_class->constructor (type,
                                                 n_construct_properties,
                                                 construct_properties);
        }

        /* do stuff. */

        return obj;
}

static void
bmp_metadata_cache_dispose (GObject *obj)
{
        BmpMetadataCache *self = (BmpMetadataCache *)obj;

        if (self->private->dispose_has_run) return;
        self->private->dispose_has_run = TRUE;

	bookmarks_save ();
	tags_save (self);

	rtfs_close (self->private->rtfs);
	h_relation_unref (self->private->relation[REALM_TRANSIENT]);

	g_message("stopped: %s", G_OBJECT_TYPE_NAME(self));
}

static void
bmp_metadata_cache_finalize (GObject *obj)
{
        BmpMetadataCache *self = (BmpMetadataCache *)obj;

        g_free (self->private);
}

BmpMetadataCache*
bmp_metadata_cache_new (void)
{
    BmpMetadataCache *self;

    self = g_object_new (bmp_metadata_cache_get_type (), NULL);
    return self;
}

static void
bmp_metadata_cache_class_init (BmpMetadataCacheClass *g_class)
{
        GObjectClass *gobject_class = G_OBJECT_CLASS (g_class);

        gobject_class->set_property = bmp_metadata_cache_set_property;
        gobject_class->get_property = bmp_metadata_cache_get_property;
        gobject_class->dispose = bmp_metadata_cache_dispose;
        gobject_class->finalize = bmp_metadata_cache_finalize;
        gobject_class->constructor = bmp_metadata_cache_constructor;

        signals[BMP_METADATA_CACHE_SIGNAL_INSERT_TUPLE] =
	    g_signal_new ("insert-tuple",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      0, 
                      G_STRUCT_OFFSET (BmpMetadataCacheClass, insert_tuple),
                      NULL, NULL, g_cclosure_marshal_VOID__POINTER,
                      G_TYPE_NONE, 1, G_TYPE_POINTER);

        signals[BMP_METADATA_CACHE_SIGNAL_DELETE_TUPLE] =
	    g_signal_new ("delete-tuple",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      0, 
                      G_STRUCT_OFFSET (BmpMetadataCacheClass, delete_tuple),
                      NULL, NULL, g_cclosure_marshal_VOID__POINTER,
                      G_TYPE_NONE, 1, G_TYPE_POINTER);

        signals[BMP_METADATA_CACHE_SIGNAL_TAG_ADDED] =
	    g_signal_new ("tag-added",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
                      G_STRUCT_OFFSET (BmpMetadataCacheClass, tag_added),
                      NULL, NULL, g_cclosure_marshal_VOID__STRING,
                      G_TYPE_NONE, 1, G_TYPE_STRING);

        signals[BMP_METADATA_CACHE_SIGNAL_TAG_DELETED] =
	    g_signal_new ("tag-deleted",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
                      G_STRUCT_OFFSET (BmpMetadataCacheClass, tag_deleted),
                      NULL, NULL, g_cclosure_marshal_VOID__STRING,
                      G_TYPE_NONE, 1, G_TYPE_STRING);

        signals[BMP_METADATA_CACHE_SIGNAL_TAG_VALUE_INSERT] =
	    g_signal_new ("tag-value-insert",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
                      G_STRUCT_OFFSET (BmpMetadataCacheClass, tag_value_insert),
                      NULL, NULL, bmpx_VOID__STRING_INT_POINTER,
                      G_TYPE_NONE, 3, G_TYPE_STRING, G_TYPE_INT, G_TYPE_POINTER);

        signals[BMP_METADATA_CACHE_SIGNAL_TAG_VALUE_REMOVE] =
	    g_signal_new ("tag-value-remove",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
                      G_STRUCT_OFFSET (BmpMetadataCacheClass, tag_value_remove),
                      NULL, NULL, bmpx_VOID__STRING_INT_POINTER,
                      G_TYPE_NONE, 3, G_TYPE_STRING, G_TYPE_INT, G_TYPE_POINTER);

}

static void 
tags_load (BmpMetadataCache *self)
{
  GError	*error = NULL;
  GDir		*master_dir;
  gchar		*path; 
  gboolean	 proceed = TRUE;

  self->private->tags = g_hash_table_new (g_str_hash, g_str_equal);

  path = g_build_filename (BMP_PATH_USER_DIR, BMP_TAGS_DIR_BASENAME, NULL); 

  if (!g_file_test (path, G_FILE_TEST_EXISTS))
    {
      g_mkdir (path, 0700);
      return;
    }
  else if (!g_file_test (path, G_FILE_TEST_IS_DIR))
    {
      g_unlink (path); // No mercy 
      g_mkdir (path, 0700);
      return;
    }

  master_dir = g_dir_open (path, 0, &error);

  if (error)
    {
      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "%s: Couldn't open tags directory: %s", G_STRLOC, error->message);
      g_error_free (error);
      error = NULL;
      return; 
    }

  //We currently assume every directory is a tag. We need some further checking for that like a stamp or something
  while (proceed)
    {
      GHashTable  *tag;
      const gchar *tagname;
      gint	   n;

      tagname = g_dir_read_name (master_dir);
      if (!tagname)
	{
	    proceed = FALSE;
	    break;
	}
      
      tag = g_hash_table_new (g_str_hash, g_str_equal); 

      //Insert into master table
      g_hash_table_insert (self->private->tags, g_strdup(tagname), tag);

      g_signal_emit (G_OBJECT(self), signals[BMP_METADATA_CACHE_SIGNAL_TAG_ADDED], 0, tagname); 

      for (n = 0; n < BMP_N_TAGS; n++) 
	{
	    GError	  *error = NULL;
	    GMappedFile	  *mappedfile;
	    GHashTable	  *sub;
	    GString	  *tagfile;
	    const gchar	  *id;
	    gsize	   length,
			   counter = 0;
	    gchar	  *buffer;

	    id = bmp_metadata_get_tag_define (n);

	    sub = g_hash_table_new (g_str_hash, g_str_equal);
	    g_hash_table_insert (tag, g_strdup(id), sub);

	    tagfile = g_string_new (path);
	    g_string_append (tagfile, G_DIR_SEPARATOR_S);
	    g_string_append (tagfile, tagname);
	    g_string_append (tagfile, G_DIR_SEPARATOR_S);
	    g_string_append (tagfile, id);
	    g_string_append (tagfile, ".mti");

	    mappedfile = g_mapped_file_new (tagfile->str,
					    FALSE,
					    &error);
	   
	    if (error) 
	      {
		// g_log (G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "%s: Unable to mmap file '%s': %s", G_STRLOC, tagfile->str, error->message);
		g_error_free (error);
		error = NULL;
		continue;
	      }

	    length = g_mapped_file_get_length (mappedfile);
	    buffer = g_mapped_file_get_contents (mappedfile);

	    while (counter < length) 
	      {
		  TagDataTuple	  *data_tuple = g_new0 (TagDataTuple,1);
		  gint		   m;

		  for (m = 0; m < BMP_N_TAGS; m++)
		    {
		      gchar *value;
		      value = (gchar*)buffer;   

		      data_tuple->tag[m] = g_strdup(value);

		      buffer += strlen(value)+1;
		      counter += strlen(value)+1;
		    }

		    g_signal_emit (G_OBJECT(self), 
				   signals[BMP_METADATA_CACHE_SIGNAL_TAG_VALUE_INSERT],
				   0,
				   tagname,
				   n,
				   data_tuple); 

		    g_hash_table_insert (sub, g_strdup(data_tuple->tag[n]), data_tuple); 

	      }

	    g_mapped_file_free (mappedfile);
	}

    }
}

static void
process_sub (gpointer _k,
	     gpointer _v,
	     gpointer data)
{
  TagDataTuple	  *data_tuple = (TagDataTuple*)_v;
  gint n;

  for (n = 0; n < BMP_N_TAGS; n++)
    {
      fwrite (data_tuple->tag[n], sizeof(char), strlen((gchar*)data_tuple->tag[n])+1, (FILE*)data); 
    }
}

static void
process_master (gpointer _k,
	        gpointer _v,
	        gpointer data)
{
  GHashTable  *tag = (GHashTable*)_v;
  const gchar *key = (const gchar*)_k;
  const gchar *tags_path = (const gchar*)data;
  gchar       *path;
  gint	       n;
  FILE	      *fh;

  path = g_build_filename (tags_path, key, NULL);

   if (!g_file_test (path, G_FILE_TEST_EXISTS))
    {
      g_mkdir (path, 0700);
    }
  else if (!g_file_test (path, G_FILE_TEST_IS_DIR))
    {
      g_unlink (path); // No mercy 
      g_mkdir (path, 0700);
    }

   for (n = 0; n < BMP_N_TAGS; n++) 
    {
	GString	    *tagfile;
	GHashTable  *sub;
	const gchar *id;

	id = bmp_metadata_get_tag_define (n);
	sub = g_hash_table_lookup (tag, id);

	tagfile = g_string_new (path);
	g_string_append (tagfile, G_DIR_SEPARATOR_S);
	g_string_append (tagfile, id);
	g_string_append (tagfile, ".mti");

	fh = fopen (tagfile->str, "w");
	g_hash_table_foreach (sub, (GHFunc) process_sub, fh);
	fclose (fh);
  
	g_string_free (tagfile, TRUE);
    }
}

static void
tags_save (BmpMetadataCache *self)
{
  gchar *path; 

  if (!self->private->tags) return;

  path = g_build_filename (BMP_PATH_USER_DIR, BMP_TAGS_DIR_BASENAME, NULL); 

  if (!g_file_test (path, G_FILE_TEST_EXISTS))
    {
      g_mkdir (path, 0700);
    }
  else 
    {
      remove_path (path);
      g_mkdir (path, 0700);
    }

  g_hash_table_foreach (self->private->tags, (GHFunc) process_master, path);

  g_free (path);
}

/** MusicBrainz AudioCD metadata stuff **/
#define BYTES_PER_SECTOR 2352
#define BYTES_PER_SECOND (44100 / 8) / 16 / 2

static int
get_duration_from_sectors (int sectors)
{
  return (sectors * BYTES_PER_SECTOR / BYTES_PER_SECOND);
}

enum {
  MB_FAILURE,
  MB_OK,
  MB_OFFLINE
};

static GList*
get_album_data_offline (musicbrainz_t o)
{
  GList *tuples = NULL;
  int	 num_tracks,
	 i;

  if (!mb_Query (o, MBQ_GetCDTOC))
    {
      g_log (BMPX_SYSTEM_LOG_DOMAIN, G_LOG_LEVEL_INFO, "%s: Unable to get read CD TOC", G_STRLOC);
      return NULL;
    }

  num_tracks = mb_GetResultInt (o, MBE_TOCGetLastTrack);
  for (i = 1; i <= num_tracks; i++)
    {
      HTuple *tuple;
      gchar *uri, *title;

      uri = g_strdup_printf ("cdda:///%d", i);

      tuple = bmp_metadata_tuple_create_default (uri);

      h_tuple_insert (tuple,
		      bmp_metadata_get_id_static (BMP_DATUM_LOCATION),
		      G_TYPE_STRING,
		      (gpointer)uri);
      g_free (uri);

      title = g_strdup_printf ("Track %d", i);
      h_tuple_insert (tuple,
		      bmp_metadata_get_id_static (BMP_DATUM_TITLE),
		      G_TYPE_STRING,
		      (gpointer)title);
      g_free (title);

      h_tuple_insert (tuple,
		      bmp_metadata_get_id_static (BMP_DATUM_ARTIST),
		      G_TYPE_STRING,
		      (gpointer)bmp_metadata_get_default_char (BMP_DATUM_ARTIST));

      h_tuple_insert (tuple,
		      bmp_metadata_get_id_static (BMP_DATUM_ALBUM),
		      G_TYPE_STRING,
		      (gpointer)bmp_metadata_get_default_char (BMP_DATUM_ALBUM));

      h_tuple_insert (tuple,
		      bmp_metadata_get_id_static (BMP_DATUM_TRACK),
		      G_TYPE_INT,
		      (gpointer)GINT_TO_POINTER(i));

      h_tuple_insert (tuple,
		      bmp_metadata_get_id_static (BMP_DATUM_TIME),
		      G_TYPE_INT,
		      (gpointer)
		      GINT_TO_POINTER(get_duration_from_sectors(mb_GetResultInt1 (o, MBE_TOCGetTrackNumSectors, i+1))));

      tuples = g_list_append (tuples, tuple);      
  }

  return tuples;
}

static gint 
get_album_data (GList **tuples)
{
  musicbrainz_t o;

  char error[256], data[256], temp[256], *cdindex = NULL;
  int ret, numTracks, trackNum, i, isMultipleArtist = 0;

  gchar *album_name  = NULL,
	*artist_name = NULL;

  // Create the musicbrainz object, which will be needed for subsequent calls
  o = mb_New();

#ifdef WIN32
    mb_WSAInit(o);
#endif

  mb_UseUTF8(o, 1);

  // Set the proper server to use. Defaults to mm.musicbrainz.org:80
  if (getenv("MB_SERVER"))
        mb_SetServer(o, getenv("MB_SERVER"), 80);

  // Check to see if the debug env var has been set 
  if (getenv("MB_DEBUG"))
        mb_SetDebug(o, atoi(getenv("MB_DEBUG")));

  // Tell the server to only return 2 levels of data, unless the MB_DEPTH env var is set
  if (getenv("MB_DEPTH"))
        mb_SetDepth(o, atoi(getenv("MB_DEPTH")));
  else
        mb_SetDepth(o, 4);

  ret = mb_Query (o, MBQ_GetCDTOC);
  if (!ret)
    {
        mb_GetQueryError(o, error, 256);
        g_message("%s: Query failed: %s", G_STRLOC, error);
        mb_Delete(o);
        return MB_FAILURE;
    }

  ret = mb_GetResultData(o, MBE_TOCGetCDIndexId, data, 256);
  if (!ret)
      {
        mb_GetQueryError(o, error, 256);
        g_message ("%s: Query failed: %s", G_STRLOC, error);
        mb_Delete(o);
        return MB_FAILURE;
      }

  cdindex = g_strdup(data);

  ret = mb_Query (o, MBQ_GetCDInfo);
  if (!ret)
    {  
        mb_GetQueryError(o, error, 256);
        g_message ("%s: Query failed: %s", G_STRLOC, error);
        g_message ("%s: Getting offline data", G_STRLOC);
	*tuples = get_album_data_offline (o);
        mb_Delete(o);
        return MB_OFFLINE;
    }

  // Select the first album
  if (!mb_Select1(o, MBS_SelectAlbum, 1))
    {
        g_message ("%s: Getting offline data", G_STRLOC);
	*tuples = get_album_data_offline (o);
	mb_Delete(o);
	return MB_OFFLINE;
    }

  // Pull back the album id to see if we got the album
  if (!mb_GetResultData(o, MBE_AlbumGetAlbumId, data, 256))
    {
      g_message ("%s: exiting", G_STRLOC);
      goto get_album_data_early_exit;
    }

  mb_GetIDFromURL(o, data, temp, 256);

  // Extract the album name
  if (mb_GetResultData(o, MBE_AlbumGetAlbumName, data, 256))
    {
      album_name = g_strdup (data);
    }
  else
    {
      album_name = NULL;
    }

   // Extract the number of tracks
   numTracks = mb_GetResultInt(o, MBE_AlbumGetNumTracks);

   if (numTracks < 1)
    {
        g_message ("%s: Getting offline data", G_STRLOC);
	*tuples = get_album_data_offline (o);
	mb_Delete(o);
	return MB_OFFLINE;
    }

  // Check to see if there is more than one artist for this album
  for(i = 1; i <= numTracks; i++)
    {
            if (!mb_GetResultData1(o, MBE_AlbumGetArtistId, data, 256, i))
		break; 

            if (i == 1)
                strcpy(temp, data);

            if (strcmp(temp, data))
            {
                isMultipleArtist = 1;
		break; 
            }
    }

  if (!isMultipleArtist)
    {
            // Extract the artist name from the album
            if (mb_GetResultData1(o, MBE_AlbumGetArtistName, data, 256, 1))
	      {
		  artist_name = g_strdup (data);
	      }
	    else
	      {
		  artist_name = NULL;
	      } 
      
            // Extract the artist id from the album
            if (mb_GetResultData1(o, MBE_AlbumGetArtistId, data, 256, 1))
            {
		  mb_GetIDFromURL(o, data, temp, 256);
            }
    }

  for(i = 1; i <= numTracks; i++)
    {
	    HTuple *tuple;
	    gchar *uri;

	    uri = g_strdup_printf ("cdda://%s/%d", cdindex, i);
	    tuple = bmp_metadata_tuple_create_default (uri);

		  h_tuple_insert (tuple,
			    bmp_metadata_get_id_static (BMP_DATUM_LOCATION),
			    G_TYPE_STRING,
			    (gpointer)uri);

		  h_tuple_insert (tuple,
			    bmp_metadata_get_id_static (BMP_DATUM_ALBUM),
			    G_TYPE_STRING,
			    (gpointer)album_name);

		  // Extract the track name from the album.
		  if (mb_GetResultData1(o, MBE_AlbumGetTrackName, data, 256, i))
		  {
	  	    h_tuple_insert (tuple,
				    bmp_metadata_get_id_static (BMP_DATUM_TITLE),
				    G_TYPE_STRING,
				    (gpointer)data);
		  }  


		  // Extract the album id from the track. Just use the
		  // first album that this track appears on
		  if (mb_GetResultData1(o, MBE_AlbumGetTrackId, data, 256, i))
		  {

		    // Extract the track number
		    trackNum = mb_GetOrdinalFromList(o, MBE_AlbumGetTrackList, data);
		    if (trackNum > 0 && trackNum < 100)
		      {
		        h_tuple_insert (tuple,
				    bmp_metadata_get_id_static (BMP_DATUM_TRACK),
				    G_TYPE_INT,
				    (gpointer)trackNum);

		      }
		}

		// If its a multple artist album, print out the artist for each track
		if (isMultipleArtist)
		{
		  // Extract the artist name from this track
		  if (mb_GetResultData1(o, MBE_AlbumGetArtistName, data, 256, i))
		    {
		      h_tuple_insert (tuple,
				    bmp_metadata_get_id_static (BMP_DATUM_ARTIST),
				    G_TYPE_STRING,
				    (gpointer)data);
		    }
		}
	      else
		{
		    h_tuple_insert (tuple,
				    bmp_metadata_get_id_static (BMP_DATUM_ARTIST),
				    G_TYPE_STRING,
				    (gpointer)artist_name);
		}

	g_free (uri);
	*tuples = g_list_append (*tuples, tuple);
    }

  ret = mb_Query (o, MBQ_GetCDTOC);
  if (!ret)
    {
        mb_GetQueryError(o, error, 256);
        g_message("%s: Query failed: %s", G_STRLOC, error);
        mb_Delete(o);
        return MB_FAILURE;
    }

  for(i = 1; i <= numTracks; i++)
    {
      HTuple *tuple;

      tuple = g_list_nth_data (*tuples, i-1);

      h_tuple_insert (tuple,
		      bmp_metadata_get_id_static (BMP_DATUM_TIME),
		      G_TYPE_INT,
		      (gpointer)
		      GINT_TO_POINTER(get_duration_from_sectors(mb_GetResultInt1 (o, MBE_TOCGetTrackNumSectors, i+1))));

      h_tuple_insert (tuple,
		      bmp_metadata_get_id_static (BMP_DATUM_BITRATE),
		      G_TYPE_INT,
		      (gpointer)
		      GINT_TO_POINTER(1411));

      h_tuple_insert (tuple,
		      bmp_metadata_get_id_static (BMP_DATUM_SAMPLERATE),
		      G_TYPE_INT,
		      (gpointer)
		      GINT_TO_POINTER(44100));
    }

    get_album_data_early_exit:

    g_free (cdindex);
    mb_Delete(o);
    return MB_OK;
}

static HTuple*
fetch_metadata (BmpMetadataCache    *self,
		const gchar	    *uri)
{
    HTuple *tuple;
    BMPURIProtocol protocol;

    tuple = NULL;
    protocol = g_uri_get_protocol (uri);

    switch (protocol)
      {
	  case BMP_URI_PROTOCOL_FILE:
	    {
	      MetadataReadResult mrr;
	      mrr = taglib_metadata_get (self, uri, &tuple);
	      if ((mrr != MRR_OK) && (mrr != MRR_FILE_NOT_FOUND))
		{
		    tuple = bmp_play_metadata_get (bmp_play_engine, uri); 
		}
	    }
	    break;

	  case BMP_URI_PROTOCOL_HTTP:
	    {
	      tuple = bmp_metadata_tuple_create_default (uri);
	      h_tuple_insert (tuple,
			  bmp_metadata_get_id_static (BMP_DATUM_LOCATION),
			  G_TYPE_STRING,
			  (gpointer)uri);

	      h_tuple_insert (tuple,
			  bmp_metadata_get_id_static (BMP_DATUM_TITLE),
			  G_TYPE_STRING,
			  (gpointer)uri);
	    }
	    break;

	  default: g_message ("%s: Can't fetch metadata for URI '%s'", G_STRLOC, uri); break; 
      }

    return tuple;
}

gboolean
tuple_check_sanity (HTuple *tuple)
{
    HTupleElem *elem;
    gboolean	sane;
    gint	n;

    static gint check_data[]={
	BMP_DATUM_ARTIST,
	BMP_DATUM_ALBUM,
	BMP_DATUM_TITLE
    };

    sane = TRUE;

    for (n = 0; n < G_N_ELEMENTS(check_data); n++)
      {
	  elem = h_tuple_get (tuple, bmp_metadata_get_id_static (check_data[n]));
          if (!elem->value || !strlen(elem->value))
	    {
	      sane = FALSE;
	      break;
	    }
      }

    return sane;
}

gboolean
bmp_metadata_cache_cache_audiocd  (BmpMetadataCache	  *self)
{
    GList   *tuples = NULL,
	    *iter   = NULL;
    HTuple  *tuple  = NULL;
    gint     rval;

    rval = get_album_data (&tuples);
    if (rval == MB_FAILURE) return FALSE;
 
    iter = tuples;
    for ( ; iter ; iter = iter->next)
      {
	if (rval == MB_OFFLINE)
	  {
	      h_relation_base_insert_tuple (H_RELATION_BASE(self->private->relation[REALM_TRANSIENT]), (HTuple*)(iter->data));
	  }
	else
	  {
	      tuple = h_relation_base_get_tuple (H_RELATION_BASE(self->private->relation[REALM_PERMANENT]), (HTuple*)(iter->data)); 
	      if (tuple)
		{
		  h_tuple_free (tuple);
		  break;
		}

	      h_relation_base_insert_tuple (H_RELATION_BASE(self->private->relation[REALM_PERMANENT]), (HTuple*)(iter->data));
	  }
      }

    g_list_foreach (tuples, (GFunc)h_tuple_free, NULL);

    return TRUE;
}

void
bmp_metadata_cache_metadata_cache (BmpMetadataCache	  *self,
				   const gchar		  *uri)
{
    HTuple	*tuple = NULL;
    HTupleElem	*elem;
    BMPURI	*bmp_uri;
    struct stat  tstat;
    
    tuple = rtfs_read_entry (self->private->rtfs, (gpointer)uri);
    bmp_uri = bmp_uri_new (uri);
    bmp_uri_unescape (bmp_uri);

#if 0
    if (tuple && g_uri_test (uri, BMP_URI_TEST_PROTOCOL, PROTOCOL_FILE))
      {
	glong mtime;

	stat (bmp_uri->path, &tstat);

        elem = h_tuple_get (tuple, bmp_metadata_get_id_static (BMP_DATUM_MTIME));
	mtime = (glong)(elem->value);

	if (tstat.st_mtime != mtime)
	  { 
	      h_relation_base_delete_tuple (H_RELATION_BASE(self->private->relation[REALM_PERMANENT]), tuple);
	      bmp_metadata_cache_metadata_get (self, uri);
	  }
      }
#endif
 
    if (!tuple)
    {
	tuple = fetch_metadata (self, uri);

	if ((tuple && !xcs_key_fetch_boolean (xcs, "bmp", "library-no-incomplete-add")) ||
            (tuple &&  xcs_key_fetch_boolean (xcs, "bmp", "library-no-incomplete-add") && tuple_check_sanity(tuple)))
	  {
#ifdef HAVE_HAL
	   if (BMP_URI_IS_LOCAL(uri)) 
	      {
		  BmpHalVolumeData  *v_data = NULL;

		  if (bmp_hal_find_volume_data_for_path (bmp_hal, bmp_uri->path, &v_data))
		    {
			
		      h_tuple_insert (tuple, 
				      bmp_metadata_get_id_static (BMP_DATUM_HAL_VOLUME_UDI), 
				      G_TYPE_STRING, 
				      (gpointer)v_data->volume_udi); 

		      h_tuple_insert (tuple, 
				      bmp_metadata_get_id_static (BMP_DATUM_HAL_DEVICE_UDI), 
				      G_TYPE_STRING, 
				      (gpointer)v_data->device_udi); 

		      h_relation_base_insert_tuple (H_RELATION_BASE(self->private->relation[REALM_PERMANENT]), tuple);
		    }
	      }
#else
	    h_relation_base_insert_tuple (H_RELATION_BASE(self->private->relation[REALM_PERMANENT]), tuple);
#endif
	  }
     }

    bmp_uri_delete (bmp_uri);
}

HTuple*
bmp_metadata_cache_get_cached   (BmpMetadataCache	*self,
				 const gchar		*uri) //This function returns cached items ONLY and ONLY from the PERMANENT realm
{
    HTuple *tuple;

    tuple = rtfs_read_entry (self->private->rtfs, (gpointer)uri);
    return tuple;
}

void
bmp_metadata_cache_metadata_remove	    (BmpMetadataCache	*self,
					     const gchar	*uri)
{
    if (!h_relation_base_delete_tuplev (H_RELATION_BASE(self->private->relation[REALM_PERMANENT]),
				   bmp_metadata_get_id_static (BMP_METADATA_PKEY),
				  (gpointer)uri)) 
	{
	    g_log (G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "%s: Unable to delete tuple with pkey '%s'", G_STRLOC, uri);
	}
}

HTuple*
bmp_metadata_cache_metadata_get (BmpMetadataCache	*self,
				 BmpFileItem		*item) 
{
    HTuple	      *tuple = NULL;
    HTupleElem	      *elem;
    BMPURI	      *bmp_uri;
    struct stat	       tstat;

    if (item->realm == REALM_PERMANENT)
      {
	tuple = rtfs_read_entry (self->private->rtfs, (gpointer)item->uri);
      }
    else
    if (item->realm == REALM_TRANSIENT)
      {
	tuple = rtfs_read_entry (self->private->rtfs, (gpointer)item->uri);
	if (!tuple) //If it's a transient item it might have been added from the permanent realm so we need to check there first
	  {
	      tuple = h_relation_base_get_tuplev (H_RELATION_BASE(self->private->relation[REALM_TRANSIENT]), 
						  bmp_metadata_get_id_static (BMP_METADATA_PKEY), 
						  (gpointer)item->uri); 
	  }
      }

    bmp_uri = bmp_uri_new (item->uri);
    bmp_uri_unescape (bmp_uri);

#if 0
    if (tuple && g_uri_test (uri, BMP_URI_TEST_PROTOCOL, PROTOCOL_FILE))
      {
	glong mtime;

	stat (bmp_uri->path, &tstat);

        elem = h_tuple_get (tuple, bmp_metadata_get_id_static (BMP_DATUM_MTIME));
	mtime = (glong)(elem->value);

	if (tstat.st_mtime != mtime)
	  { 
	      h_relation_base_delete_tuple (H_RELATION_BASE(self->private->relation[REALM_PERMANENT]), tuple);
	      bmp_metadata_cache_metadata_get (self, uri);
	  }
      }
#endif
 
    if (!tuple)
    {
	tuple = fetch_metadata (self, item->uri);
	if (tuple)
	  {
#ifdef HAVE_HAL
	   BmpHalVolumeData  *v_data = NULL;

	   if (BMP_URI_IS_LOCAL(item->uri)) 
	      {
		  if (bmp_hal_find_volume_data_for_path (bmp_hal, bmp_uri->path, &v_data))
		    {
			
		      h_tuple_insert (tuple, 
				      bmp_metadata_get_id_static (BMP_DATUM_HAL_VOLUME_UDI), 
				      G_TYPE_STRING, 
				      (gpointer)v_data->volume_udi); 

		      h_tuple_insert (tuple, 
				      bmp_metadata_get_id_static (BMP_DATUM_HAL_DEVICE_UDI), 
				      G_TYPE_STRING, 
				      (gpointer)v_data->device_udi); 
		    }
		  else
		    {
		      g_log (G_LOG_DOMAIN,
			     G_LOG_LEVEL_CRITICAL,
			     "%s: Unable to acquire v_data for path '%s'!",
			     G_STRLOC,
			     bmp_uri->path);
		    }
	      }

	    if (xcs_key_fetch_boolean (xcs, "bmp", "library-no-incomplete-add"))
	      {
		if (!tuple_check_sanity(tuple)) item->realm = REALM_TRANSIENT;
	      }

	    if (v_data && (!v_data->is_disc))
	      {
		h_relation_base_insert_tuple (H_RELATION_BASE(self->private->relation[item->realm]), tuple);
	      }
#else
	    h_relation_base_insert_tuple (H_RELATION_BASE(self->private->relation[item->realm]), tuple);
#endif
	  }
	else
	  {
		item->realm = REALM_TRANSIENT;
		tuple = bmp_metadata_tuple_create_default (item->uri);
	        h_tuple_insert (tuple, 
		    bmp_metadata_get_id_static (BMP_DATUM_LOCATION), 
		    G_TYPE_STRING, 
		    (gpointer)item->uri); 
	        h_tuple_insert (tuple, 
		    bmp_metadata_get_id_static (BMP_DATUM_TITLE), 
		    G_TYPE_STRING, 
		    (gpointer)item->uri); 
	  }
     }

    bmp_uri_delete (bmp_uri);
    return tuple; 
}

#ifdef HAVE_HAL
gboolean
bmp_metadata_cache_check_tuple_volume	(BmpMetadataCache     *self,
					 HTuple		      *tuple,
					 gboolean	      *is_present)
{
    HTupleElem *elem; 

    elem = h_tuple_get (tuple, bmp_metadata_get_id_static (BMP_DATUM_HAL_VOLUME_UDI));
    
    if (elem->value)
      {
	*is_present = bmp_hal_volume_is_present (bmp_hal, (const gchar*)elem->value);
	return TRUE;
      }
    else
      {
	return FALSE;
      }
    
}
#endif

void
bmp_metadata_cache_load_rtfs (BmpMetadataCache *self)
{
    static gboolean loaded = FALSE;

    if (loaded) return;
    rtfs_read_filesystem (self->private->rtfs);
    loaded = TRUE;
}

void
bmp_metadata_cache_load_tags (BmpMetadataCache *self)
{
    tags_load (self);
}

void
bmp_metadata_cache_for_each_tuple (BmpMetadataCache *self, HTupleFunc func, gpointer data)
{
    h_relation_for_each_tuple (self->private->relation[REALM_PERMANENT], func, data);
}

gchar*
bmp_metadata_cache_get_titlestring (BmpMetadataCache	  *self,
				    BmpFileItem		  *item) 
{
    GString	  *string_out   = NULL;
    GString	  *string_field = NULL;
    HTuple	  *tuple;
    gchar	  *field;
    const gchar	  *format;
    gint	   n;
    gboolean	   hit = FALSE;

    tuple = bmp_metadata_cache_metadata_get (bmp_metadata_cache, item);
   
    if (!tuple)
      {
	return g_strdup("");
      }

    format = xcs_key_fetch (xcs, "bmp", "titlestring-format");
    string_out = g_string_new ("");

    if (!format)
      {
	return g_strdup("");
      }

    while (*format) {

	/* Begin sequence */
	if ((*format == '%') && (*(format+1) != 0) && (*(format+1) == '{'))
	  {
	    string_field =
		g_string_new ("");
	    format += 2;

	    while (*format && (*format != '}') && (*format != ' '))
	      {
		g_string_append_c (string_field, *format);
		format++;
	      }

	    field =
		g_string_free (string_field, FALSE);

	    if (!*format)
	      {
	        g_free (field);
		return g_string_free (string_out, FALSE);
	      }

	    /* Append metadata field */
	    for (n = 0; n < BMP_DATUM_NONE; n++)
	      {
		if (!g_ascii_strcasecmp (field, bmp_metadata_get_id_static(n)))
                  {
		    HTupleElem *elem = NULL;

		    elem = h_tuple_get (tuple, field);

		    if (elem)
		      {
			gchar *value = NULL; 

			if (elem->type == G_TYPE_INT)
			  {
			    value = g_strdup_printf ("%d", GPOINTER_TO_INT(elem->value));
			  }
			else if (elem->type == G_TYPE_STRING)
			  {
			    if (elem->value)
			      value = g_strdup (elem->value);
			    else
			      value = g_strdup ("");
			  }
			
			if (value)
			  {
			    g_strstrip (value);
			    if (strlen(value)>0 && strcmp(value, "0") != 0)
			      {
				hit = TRUE;
				g_string_append (string_out, value);
			      }
			    g_free (value);
			  }
		      }
		    break;
		  }
	      }

	    g_free (field);
	  }
	else
	  {
	    g_string_append_c (string_out, *format);
	  }
	format++;
    }

    if (!hit)
      {
	HTupleElem  *elem;
	gchar	    *filename, *basename, *ret;
	g_string_free (string_out, TRUE);
  
	elem = h_tuple_get (tuple, bmp_metadata_get_id_static (BMP_METADATA_PKEY));

	if (!g_ascii_strncasecmp("file:",elem->value,5))
	  {
	    filename = g_filename_from_uri (elem->value, NULL, NULL);

	    if (filename)
	      {
		basename = g_path_get_basename (filename);
		g_free (filename);
		if (basename)
		  {
		    h_tuple_free (tuple);
		    return basename;
		  }
	      }
	  }

	ret = g_strdup (elem->value);
	h_tuple_free (tuple);
	return ret; 
      }
    else
      {
	return g_string_free (string_out, FALSE);
      }
}

Query*
bmp_metadata_cache_query_initialize (BmpMetadataCache *self)
{
    Query *query;

    g_return_val_if_fail (BMP_IS_METADATA_CACHE(self), NULL);

    /* Initialize query */
    query = g_new0 (Query,1);
    query->query_atoms = NULL;

    return query;
}

gboolean
bmp_metadata_cache_query_append_int (BmpMetadataCache *self, Query* query, BmpMetadatumId metadatum_id, gboolean exact_match, gint datum) 
{
    GType      type;
    QueryAtom *query_atom;

    g_return_val_if_fail (BMP_IS_METADATA_CACHE(self), FALSE);

    type = bmp_metadata_get_type (metadatum_id);

    if (type != G_TYPE_INT)
      {
		g_log (G_LOG_DOMAIN, 
		       G_LOG_LEVEL_CRITICAL, 
		       "%s(%d): System inconsistency: Metadatum %d is not of type %ld", 
		       __FILE__, 
		       __LINE__, 
		       metadatum_id,
		       (gulong)type);

		return FALSE;
      }

    query_atom = g_new0 (QueryAtom,1);

    query_atom->metadatum_id =
	    metadatum_id; 
    query_atom->exact_match =
	    exact_match; 
    query_atom->match_value =
	    g_new0 (GValue,1); 

    g_value_init (query_atom->match_value, type); 
    g_value_set_int (query_atom->match_value, datum); 

    query->query_atoms = g_list_append (query->query_atoms, query_atom);

    return TRUE;
}

gboolean
bmp_metadata_cache_query_append_string (BmpMetadataCache *self, Query* query, BmpMetadatumId metadatum_id, gboolean exact_match, gchar *datum) 
{
    GType      type;
    QueryAtom *query_atom;

    g_return_val_if_fail (BMP_IS_METADATA_CACHE(self), FALSE);

    type = bmp_metadata_get_type (metadatum_id);

    if (type != G_TYPE_STRING)
      {
		g_log (G_LOG_DOMAIN, 
		       G_LOG_LEVEL_CRITICAL, 
		       "%s(%d): System inconsistency: Metadatum %d is not of type %ld", 
		       __FILE__, 
		       __LINE__, 
		       metadatum_id,
		       (gulong)type);

		return FALSE;
      }

    query_atom = g_new0 (QueryAtom,1);

    query_atom->metadatum_id =
	    metadatum_id; 
    query_atom->exact_match =
	    exact_match; 
    query_atom->match_value =
	    g_new0 (GValue,1); 

    g_value_init (query_atom->match_value, type); 
    g_value_set_string (query_atom->match_value, datum); 

    query->query_atoms = g_list_append (query->query_atoms, query_atom);

    return TRUE;
}

Query*
bmp_metadata_cache_query_construct (BmpMetadataCache *self, gint n_atoms, ...)
{
    Query      *query;
    GType	type;
    va_list	args;
    gint	n;
    gboolean	fatal_error = FALSE;

    g_return_val_if_fail (BMP_IS_METADATA_CACHE(self), NULL);

    /* Initialize query */
    query = bmp_metadata_cache_query_initialize (self); 

    va_start (args, n_atoms);

    for (n = 0; n < n_atoms; n++)
      {
	BmpMetadatumId metadatum_id;
	gboolean       exact_match;

	metadatum_id = va_arg (args, gint);
	exact_match = va_arg (args, gboolean);

	type = bmp_metadata_get_type (metadatum_id);	

	switch (type)
	  {
	    case G_TYPE_INT:
	      {
		  fatal_error = !bmp_metadata_cache_query_append_int (self, query, metadatum_id, exact_match, va_arg (args, gint));
	      }
	      break;

	    case G_TYPE_STRING:
	      {
		  fatal_error = !bmp_metadata_cache_query_append_string (self, query, metadatum_id, exact_match, va_arg (args, gchar*));
	      }
	      break;

	    default:
	      {
		g_log (G_LOG_DOMAIN, 
		       G_LOG_LEVEL_CRITICAL, 
		       "%s(%d): System inconsistency: Metadatum %d is not of type %ld", 
		       __FILE__, 
		       __LINE__, 
		       metadatum_id,
		       (gulong)type);

		fatal_error = TRUE;

		break;
	      }
	  }

	if (fatal_error) break;
    }

    va_end (args);

    // FIXME: Free query on error
    return fatal_error ? NULL : query;
}

gboolean
query_predicate_func (HTuple	 *tuple,
		      Query	 *query) 
{
  GList *_iter_list = query->query_atoms;
  gboolean hit = TRUE;

  for ( ; _iter_list; _iter_list = _iter_list->next)
    {
      QueryAtom  *atom = (QueryAtom*)_iter_list->data;
      HTupleElem *elem = h_tuple_get (tuple, bmp_metadata_get_id_static (atom->metadatum_id)); 

      switch (elem->type)
	{
	    case G_TYPE_INT:
	      {
		  if (GPOINTER_TO_INT(elem->value) != g_value_get_int (atom->match_value))
		      hit = FALSE;
	      }
	      break;

	    case G_TYPE_STRING:
	      {
		  if (!elem->value)
		    {
		      hit = TRUE;
		      break;
		    }
		
		  if (!atom->exact_match)
		    {
		      if (!match_keys((gchar*)elem->value, 
				g_value_get_string (atom->match_value))) 
		      hit = FALSE;
		    }
		  else
		    {
		      if (g_utf8_collate((gchar*)elem->value, 
					 g_value_get_string (atom->match_value))) 
		      hit = FALSE;
		    }
	      }
	      break;

	    default:
	      {
		g_log (G_LOG_DOMAIN, 
		       G_LOG_LEVEL_CRITICAL, 
		       "%s(%d): System inconsistency: Metadatum %d is not of type %ld", 
		       __FILE__, 
		       __LINE__, 
		       atom->metadatum_id,
		       (gulong)elem->type);
		break;
	      }
      }

      if (!hit)
	break;
    }

    return hit;
}

static void
query_atom_free (gpointer data)
{
  QueryAtom   *atom = (QueryAtom*)data;

  g_value_unset (atom->match_value); 
  g_free (atom->match_value);
}

void
bmp_metadata_cache_query_free (BmpMetadataCache *self, Query *query)
{
    g_return_if_fail (BMP_IS_METADATA_CACHE (self));

    g_list_foreach (query->query_atoms, (GFunc)query_atom_free, NULL);
}

HRelation*
bmp_metadata_cache_query_run (BmpMetadataCache *self, Query *query) 
{
    HRelation *relation;

    g_return_val_if_fail (BMP_IS_METADATA_CACHE (self), NULL);

    relation = h_relation_restrict (self->private->relation[REALM_PERMANENT],
				    (HTuplePredFunc)query_predicate_func,
				    query);	

    return relation;
}

HRelation*
bmp_metadata_cache_query_list_datum   (BmpMetadataCache	  *self,
				       gint		   metadatum_id,
				       gint		   metadatum_id_predicate,
				       gchar		  *predicate_value)
{
  HRelation *projected;

  g_return_val_if_fail (BMP_IS_METADATA_CACHE(self), NULL); 

  bmp_metadata_cache_load_rtfs (self);

  if (metadatum_id_predicate != -1)
    {
	projected = h_relation_project (self->private->relation[REALM_PERMANENT],
				  bmp_metadata_get_id_static (metadatum_id),
				  bmp_metadata_get_id_static (metadatum_id_predicate),
				  NULL);
    }
  else
    {
	projected = h_relation_project (self->private->relation[REALM_PERMANENT],
				  bmp_metadata_get_id_static (metadatum_id),
				  NULL);
    }


  return projected;
}



#ifdef HAVE_GUI
#ifdef USE_AMAZON
static void
area_prepared      (GdkPixbufLoader *loader,
                    gint	    *done)
{
  g_atomic_int_inc (done);
}

#define AMAZON_FORMAT_STRINBMP_URI "http://webservices.amazon.com/onca/xml?Service=AWSECommerceService&SubscriptionId=1E90RVC80K4MVNTCHG02&Operation=ItemSearch&Artist=%s&Title=%s&SearchIndex=Music&ResponseGroup=Images&Version=2004-11-10"

GdkPixbuf*
bmp_metadata_cache_get_cover_from_tuple (BmpMetadataCache      *self,
				         HTuple		       *tuple)
{
    GError		  *error = NULL;
    GdkPixbufLoader	  *loader;
    GdkPixbuf		  *image = NULL;
    VFSHandle		  *handle;
    HTupleElem		  *elem_artist;
    HTupleElem		  *elem_album;

    BMPURI	          *uri;
    gchar		  *uri_string;
    gchar		  *uri_img;
    gint		   done;
    xmlDocPtr		   doc;
    xmlXPathObjectPtr	   xpathObj;
    xmlNodeSetPtr	   nv;
    xmlNodePtr		   node;

    g_return_val_if_fail (BMP_IS_METADATA_CACHE (self), NULL);
    g_return_val_if_fail (tuple != NULL, NULL);

    elem_artist = h_tuple_get (tuple, bmp_metadata_get_id_static (BMP_DATUM_ARTIST));
    elem_album = h_tuple_get (tuple, bmp_metadata_get_id_static (BMP_DATUM_ALBUM));

    uri_string = g_strdup_printf (AMAZON_FORMAT_STRINBMP_URI, (gchar*)elem_artist->value, (gchar *)elem_album->value);
    uri = bmp_uri_new (uri_string);
    bmp_uri_escape (uri);

    handle = vfs_handle_new ();
    handle->uri = bmp_uri_get_string (uri); 
    bmp_uri_delete (uri);

    vfs_handle_read (handle, VFS_NEED_CONTAINER_FORCE_NONE);
    if ((!handle->buffer) || (handle->buffer_size == 0))
      {
	vfs_handle_free (handle);
	return NULL;
      }

    doc = xmlParseMemory (handle->buffer, handle->buffer_size);
    if (!doc)
      {
	vfs_handle_free (handle);
	return NULL;
      }
      
    xpathObj = xml_execute_xpath_expression (doc, BAD_CAST "//amazon:MediumImage/amazon:URL", BAD_CAST "amazon=http://webservices.amazon.com/AWSECommerceService/2004-11-10");
    nv = xpathObj->nodesetval;

    if (!nv)
      {
	  xmlXPathFreeObject (xpathObj);
	  return NULL;
      }

    if (!nv->nodeNr)
      {
	  xmlXPathFreeObject (xpathObj);
	  return NULL;
      }

    node = nv->nodeTab[0];
    if (!node->children)
      {
	  xmlXPathFreeObject (xpathObj);
	  return NULL;
      }

    uri_img = (gchar*)XML_GET_CONTENT (node->children);
    xmlXPathFreeObject (xpathObj);

    //Load the image
    handle = vfs_handle_new ();
    handle->uri = uri_img; 
    vfs_handle_read (handle, VFS_NEED_CONTAINER_FORCE_NONE);

    if ((!handle->buffer) || (handle->buffer_size == 0))
      {
	vfs_handle_free (handle);
	return NULL;
      }

    loader = gdk_pixbuf_loader_new_with_type ("jpeg", &error);
    if (error)
      {
	  g_message (error->message);
	  g_error_free (error);
	  error = NULL;
      }

    done = 0;
    g_object_connect (G_OBJECT (loader),
		      "signal::area-prepared",
		      G_CALLBACK (area_prepared),
		      &done,
		      NULL);

    gdk_pixbuf_loader_write (loader,
			     (const guchar *)handle->buffer,
			     handle->buffer_size, 
			     &error);
    if (error)
      {
	  g_message (error->message);
	  g_error_free (error);
	  error = NULL;
      }

    gdk_pixbuf_loader_close (loader, &error);
    if (error)
      {
	  g_message (error->message);
	  g_error_free (error);
	  error = NULL;
      }

    image = gdk_pixbuf_loader_get_pixbuf (loader);
    g_object_ref (image);
    g_object_unref (loader);
    vfs_handle_free (handle);

    //FIXME: Use Mutex and GCond
    while (!g_atomic_int_get(&done))
      {
	  while (gtk_events_pending()) gtk_main_iteration ();
      }

    return image;
}
#endif
#endif
