/*  BMPx - The Dumb Music Player
 *  Copyright (C) 2005 BMPx development team.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  $Id$
 *
 * The BMPx project hereby grant permission for non-gpl compatible GStreamer
 * plugins to be used and distributed together with GStreamer and BMPx. This
 * permission are above and beyond the permissions granted by the GPL license
 * BMPx is covered by.
 */

#include <config.h>
#include <string.h>

#include <glib.h>
#include <glib-object.h>
#include <glib/gi18n.h>

#include <gst/gst.h>
#include <gst/gstelement.h>
#include <gst/interfaces/mixer.h>
#include <gst/interfaces/mixertrack.h>
#include <gst/interfaces/mixeroptions.h>

#include <bmp/uri.h>
#include <bmp/util.h>

#include <goa/libgoa.h>

#include <libhrel/types.h>
#include <libhrel/tuple.h>
#include <libhrel/relation.h>
#include <libhrel/relation_base.h>

#include "play.h"

#define BMP_PLAY_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), BMP_TYPE_PLAY, BmpPlayPrivate))

#define BMP_GST_BUFFER_TIME   ((gint64) 50000)
#define BMP_GST_PLAY_TIMEOUT  ((gint64) 3000000000)

G_DEFINE_TYPE (BmpPlay, bmp_play, G_TYPE_OBJECT)

#include "main.h"

enum
{
  BMP_PLAY_SIGNAL_EOS,
  BMP_PLAY_SIGNAL_ERROR,
  BMP_PLAY_SIGNAL_BUFFER,
  BMP_PLAY_SIGNAL_TITLE,
  BMP_PLAY_SIGNAL_SEEK,

  BMP_PLAY_N_SIGNALS
};

enum
{
  BMP_PLAY_PROP_STREAM	    = 1,
  BMP_PLAY_PROP_PLAYSTATUS, 
  BMP_PLAY_PROP_HAS_CDDA, 
  BMP_PLAY_PROP_HAS_HTTP,

  BMP_PLAY_PROP_MUTE,
  BMP_PLAY_PROP_VOLUME,
  BMP_PLAY_PROP_POSITION,

  BMP_PLAY_PROP_SANE,	

  BMP_PLAY_N_PROPS
};

enum 
{
  HAVE_GNOMEVFS,
  HAVE_NEON,
};

enum
{
  P_FILE,
  P_HTTP,
  P_CDDA,

  N_PIPES
};

enum
{
  PIPELINE,

  SRC,
  DECODER,
  CONVERT,
  SINK,
  VOLUME,

  N_ELEMENTS
};

typedef struct _BmpPlayPrivate BmpPlayPrivate;
struct _BmpPlayPrivate
{
  gboolean dispose_has_run;

  GstElement *elements[N_PIPES][N_ELEMENTS];
  gint	      current_pipe;

  gchar	     *mrl;
  gint	      playstatus;
  gdouble     volume;
  gboolean    mute;

  guint	      source;
  gboolean    sending;

  gboolean    has_cdda;
  gboolean    has_http;
  gint	      http_sink;

  gint	      message_domain_id;

  gboolean    sane;

  gchar	    **extensions;

  /* data upload */

  GPtrArray *fifo_buffer;
  GPtrArray *fifo_buffer_timestamps;
  gint fifo_buffer_length;
  gint fifo_buffer_new_items;

  gint timeout_src_id;
};

G_LOCK_DEFINE (fifo_buffer);
G_LOCK_DEFINE (fifo_buffer_length);


static guint signals[BMP_PLAY_N_SIGNALS] = { 0, 0, 0, 0 };

static GObject *bmp_play_constructor (GType type,
                                      guint n_construct_properties,
                                      GObjectConstructParam *construct_properties);

static gboolean bmp_play_bus_watch (GstBus	*bus,
                                    GstMessage	*message,
                                    BmpPlay	*self);

static void
pipeline_create		   (BmpPlay	*self,
			    gint	 pipe);

static void
send_stream_pos_stop (BmpPlay *self)
{
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);
  GSource *source;

  if (private->source)
    {
	  source = g_main_context_find_source_by_id (NULL, private->source);
	  if (source)
	    {
	      g_source_destroy (source);
	      private->sending = FALSE;
	    }
    }

  g_object_set (G_OBJECT (bmp_system_control), "stream_pos", (gint) 0, NULL);
}

static gboolean
send_stream_pos (BmpPlay *self)
{
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);
  gint            stream_pos = g_object_get_int (G_OBJECT(self), "position"); 

  //FIXME: Not sure this should be 'uploaded', i think SystemControl should rather 'download' this from BmpPlay (?)
  if (!private->elements[private->current_pipe][PIPELINE])
    {
      g_object_set (G_OBJECT (bmp_system_control), "stream_pos", (gint) 0, NULL);
      private->sending = FALSE;
      return FALSE;
    }

  if ((stream_pos >= 0) && (GST_STATE(private->elements[private->current_pipe][PIPELINE]) >= GST_STATE_PLAYING))
    {
      g_object_set (G_OBJECT (bmp_system_control), "stream_pos", stream_pos, NULL);
    }
  else
    {
      g_object_set (G_OBJECT (bmp_system_control), "stream_pos", (gint) 0, NULL);
      private->sending = FALSE;
      return FALSE;
    }

  return TRUE;
}

#if 0
#define LV_BUFFER_SIZE 1024

static gboolean
handoff_cb (GstPad	      *pad,
	    GstMiniObject     *obj,
	    BmpPlay	      *self)
{
    BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);
    gint n_buffers;
    gint v;

    short *data;
    GstBuffer *arg0;

    if (!GST_IS_BUFFER(obj))
        return TRUE;

    arg0 = GST_BUFFER(obj);

    n_buffers = arg0->size / (LV_BUFFER_SIZE * sizeof (short));

    for (v = 0; v < n_buffers; ++v)
     {
         data =
             g_memdup (arg0->data +
                       ((LV_BUFFER_SIZE * sizeof (short)) * v),
                       LV_BUFFER_SIZE * sizeof (short));

         G_LOCK (fifo_buffer);
         g_ptr_array_add (private->fifo_buffer, data);
         G_UNLOCK (fifo_buffer);
     }

    G_LOCK (fifo_buffer_length);
    private->fifo_buffer_length += n_buffers;
    G_UNLOCK (fifo_buffer_length);

    private->fifo_buffer_new_items += n_buffers;

    return TRUE;
}

static gboolean
handoff_send_buffer (BmpPlay *self)
{
    BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);
    short *data;
    gint n;

    g_return_val_if_fail (BMP_IS_PLAY (self), FALSE);

    if (!private->fifo_buffer_length)
        return TRUE;

#if 0
    r_count = (play->private->fifo_buffer_new_items / 2) - 1;
#endif

    /* We take data from in between the last shot */
    data = g_ptr_array_index (private->fifo_buffer, 0);

    g_signal_emit (self,
                   signals[BMP_PLAY_SIGNAL_BUFFER], 0, data, NULL);

    G_LOCK (fifo_buffer);
    for (n = 0; n < private->fifo_buffer_new_items; ++n)
     {
         g_free (g_ptr_array_index (private->fifo_buffer, n));
     }
    g_ptr_array_remove_range (private->fifo_buffer, 0, private->fifo_buffer_new_items);
    G_UNLOCK (fifo_buffer);

    G_LOCK (fifo_buffer_length);
    private->fifo_buffer_length -= private->fifo_buffer_new_items;
    G_UNLOCK (fifo_buffer_length);

    private->fifo_buffer_new_items = 0;

    return (!(private->playstatus & PLAYSTATUS_STOPPED));
}
#endif

void
bmp_play_set_volume (BmpPlay *self,
                     gint     volume)
{
  g_return_if_fail (BMP_IS_PLAY (self));

  g_object_set (G_OBJECT(self), "volume", volume, NULL);
}

gint
bmp_play_get_volume (BmpPlay *self)
{
  g_return_val_if_fail (BMP_IS_PLAY (self), 0);

  return g_object_get_int (self, "volume");
}

static void
nullify_elements    (BmpPlay *self)
{
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);
  gint n;

  for (n = 0; n < N_ELEMENTS; n++) 
    {
	private->elements[P_FILE][n] = NULL;
	private->elements[P_HTTP][n] = NULL;
	private->elements[P_CDDA][n] = NULL;
    }
}

static void
pipelines_destroy (BmpPlay *self)
{
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);
  g_return_if_fail (BMP_IS_PLAY (self));

  if (private->elements[P_FILE][PIPELINE])
    {
      gst_element_set_state (private->elements[P_FILE][PIPELINE], GST_STATE_NULL);
      gst_object_unref (private->elements[P_FILE][PIPELINE]);
    }

  if (private->elements[P_HTTP][PIPELINE])
    {
      gst_element_set_state (private->elements[P_HTTP][PIPELINE], GST_STATE_NULL);
      gst_object_unref (private->elements[P_HTTP][PIPELINE]);
    }

  if (private->elements[P_CDDA][PIPELINE])
    {
      gst_element_set_state (private->elements[P_CDDA][PIPELINE], GST_STATE_NULL);
      gst_object_unref (private->elements[P_CDDA][PIPELINE]);
    }

  nullify_elements (self);
}

#if 0
static void
get_stream_title (GObject *gobject,
                  GParamSpec *arg1,
                  gpointer user_data)
{
}
#endif

static gulong handler_id = 0;
static void
link_pad (GstElement *element,
	  GstPad     *pad,
	  gboolean    last,
	  gpointer    data)
{
  BmpPlay	  *self = PLAY (data);
  BmpPlayPrivate  *private = BMP_PLAY_GET_PRIVATE (self);
  GstPad	  *sinkpad;

  sinkpad = gst_element_get_pad (private->elements[private->current_pipe][CONVERT], "sink");
  gst_pad_link (pad, sinkpad);
}

static void
pipeline_create (BmpPlay	*self,
		 gint		 pipe)
{
  BmpPlayPrivate	    *private = BMP_PLAY_GET_PRIVATE (self);
  gchar			    *device, *sink;
  const gchar		    *device_propname;
  guint64		     
			     buffer_time;
  gboolean		     set_buffer_time;

  sink = xcs_key_fetch_string (xcs, "audio", "sink");
  device = xcs_key_fetch_string (xcs, "audio", "device");

  set_buffer_time = FALSE;
  buffer_time = 200000;

#ifdef HAVE_ALSA
  if (!strcmp(sink, "alsasink"))
    {
      set_buffer_time = TRUE;
      buffer_time = xcs_key_fetch_int (xcs, "audio", "alsa-buffer-time");
      device_propname = "device";
    }
  else
#endif
  if (!strcmp(sink, "osssink"))
    {
      set_buffer_time = TRUE;
      buffer_time = xcs_key_fetch_int (xcs, "audio", "oss-buffer-time");
      device_propname = "device";
    }
  else
  if (!strcmp(sink, "esdsink"))
    {
      set_buffer_time = TRUE;
      buffer_time = xcs_key_fetch_int (xcs, "audio", "esd-buffer-time");
      device_propname = "host";

      if (!strlen(device))
	{
	  g_free (device);
	  device = NULL;
	}
    }
#ifdef HAVE_SUNAUDIO
  else
  if (!strcmp(sink, "sunaudiosink"))
    {
      set_buffer_time = TRUE;
      buffer_time = xcs_key_fetch_int (xcs, "audio", "sun-buffer-time");
      device_propname = "device";
    }
#endif

  if (pipe == P_FILE)
    {
      //Create FILE pipeline
      private->elements[P_FILE][SRC] =
	    gst_element_factory_make ("filesrc", "src");
      private->elements[P_FILE][DECODER] =
	    gst_element_factory_make ("decodebin", "decoder");
      private->elements[P_FILE][CONVERT] =
	    gst_element_factory_make ("audioconvert", "audioconvert");
      private->elements[P_FILE][VOLUME] =
	    gst_element_factory_make ("volume", "volume");
      private->elements[P_FILE][SINK] =
	    gst_element_factory_make (sink, "sink");

#if 0
      gst_pad_add_buffer_probe (gst_element_get_pad (private->elements[P_FILE][SINK], "sink"), handoff_cb, self);
#endif


      g_object_set (G_OBJECT(private->elements[P_FILE][SINK]), 
		     device_propname, 
		     device, 
		     NULL);

      if (set_buffer_time)
	{
	  g_object_set (G_OBJECT(private->elements[P_FILE][SINK]),
		    "buffer-time", 
		     buffer_time,
		     NULL); 
	}
  
      private->elements[P_FILE][PIPELINE] = gst_pipeline_new ("pipeline_file"); 

      gst_bin_add_many (GST_BIN (private->elements[P_FILE][PIPELINE]),
		    private->elements[P_FILE][SRC],	
		    private->elements[P_FILE][DECODER],	
		    private->elements[P_FILE][CONVERT],
		    private->elements[P_FILE][VOLUME],
		    private->elements[P_FILE][SINK],
		    NULL);

      gst_element_link_many (private->elements[P_FILE][SRC], 
			 private->elements[P_FILE][DECODER], NULL);

      gst_element_link_many (private->elements[P_FILE][CONVERT], 
			 private->elements[P_FILE][VOLUME], 
			 private->elements[P_FILE][SINK], NULL);

      handler_id = g_signal_connect (G_OBJECT(private->elements[P_FILE][DECODER]),
		    "new-decoded-pad", 
		    G_CALLBACK (link_pad),
		    self);

      gst_bus_add_watch (gst_pipeline_get_bus (GST_PIPELINE(private->elements[P_FILE][PIPELINE])),
			(GstBusFunc) bmp_play_bus_watch, self);

      goto pipeline_create_early_exit;
    }
  else if (pipe == P_HTTP)
    {
      //Create HTTP pipeline
      private->elements[P_HTTP][SRC] = gst_element_factory_make ("neonhttpsrc",  "src");

      if (!private->elements[P_HTTP][SRC])
	{
	  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "%s: neonhttpsrc not available, trying gnomevfssrc", G_STRLOC);
	  private->elements[P_HTTP][PIPELINE] = NULL;
	}
      else
	{
	  private->http_sink = HAVE_NEON;
	  goto http_setup_pipeline;
	}

      private->elements[P_HTTP][SRC] = gst_element_factory_make ("gnomevfssrc",  "src");
      if (!private->elements[P_HTTP][SRC])
	{
	  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "%s: No HTTP support available!", G_STRLOC);
	  private->elements[P_HTTP][PIPELINE] = NULL;
	  goto pipeline_create_early_exit;
	}
      else
	{
	  private->http_sink = HAVE_GNOMEVFS;
	}

      http_setup_pipeline:
      private->has_http = TRUE;
      private->elements[P_HTTP][DECODER] = gst_element_factory_make ("decodebin", "decoder");
      private->elements[P_HTTP][CONVERT] = gst_element_factory_make ("audioconvert", "audioconvert");
      private->elements[P_HTTP][VOLUME] = gst_element_factory_make ("volume", "volume");
      private->elements[P_HTTP][SINK] = gst_element_factory_make (sink, "sink");

#if 0
      gst_pad_add_buffer_probe (gst_element_get_pad (private->elements[P_HTTP][SINK], "sink"), handoff_cb, self);
#endif

      g_object_set (G_OBJECT(private->elements[P_HTTP][SINK]), 
		     device_propname, 
		     device, 
		     NULL);

      if (set_buffer_time)
	    {
	      g_object_set (G_OBJECT(private->elements[P_HTTP][SINK]),
		    "buffer-time", 
		     buffer_time,
		     NULL); 
	    }

      private->elements[P_HTTP][PIPELINE] = gst_pipeline_new ("pipeline_http"); 

      gst_bin_add_many (GST_BIN (private->elements[P_HTTP][PIPELINE]),
		    private->elements[P_HTTP][SRC],	
		    private->elements[P_HTTP][DECODER],	
		    private->elements[P_HTTP][CONVERT],
		    private->elements[P_HTTP][VOLUME],
		    private->elements[P_HTTP][SINK],
		    NULL);

      gst_element_link_many (private->elements[P_HTTP][SRC], 
			 private->elements[P_HTTP][DECODER], NULL);

      gst_element_link_many (private->elements[P_HTTP][CONVERT], 
			 private->elements[P_HTTP][VOLUME], 
			 private->elements[P_HTTP][SINK], NULL);

      handler_id = g_signal_connect (G_OBJECT(private->elements[P_HTTP][DECODER]),
		    "new-decoded-pad", 
		    G_CALLBACK (link_pad),
		    self);

      gst_bus_add_watch (gst_pipeline_get_bus (GST_PIPELINE(private->elements[P_HTTP][PIPELINE])), 
			    (GstBusFunc) bmp_play_bus_watch, self);
      goto pipeline_create_early_exit;
    }
  else if (pipe == P_CDDA)
    {
      //Create CDDA pipeline
      private->elements[P_CDDA][SRC] = gst_element_factory_make ("cdparanoiasrc",  "src");

      if (!private->elements[P_CDDA][SRC])
      {
	g_log (G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "%s: No CDDA support available!", G_STRLOC);
	private->elements[P_CDDA][PIPELINE] = NULL;
      }
    else
      {
	private->has_cdda = TRUE;

	private->elements[P_CDDA][CONVERT] = gst_element_factory_make ("audioconvert", "audioconvert");
	private->elements[P_CDDA][VOLUME] = gst_element_factory_make ("volume",	    "volume");
	private->elements[P_CDDA][SINK] = gst_element_factory_make (sink, "sink");

#if 0
      gst_pad_add_buffer_probe (gst_element_get_pad (private->elements[P_HTTP][SINK], "sink"), handoff_cb, self);
#endif

	g_object_set (G_OBJECT(private->elements[P_CDDA][SINK]), 
		     device_propname, 
		     device, 
		     NULL);

	if (set_buffer_time)
	    {
	      g_object_set (G_OBJECT(private->elements[P_CDDA][SINK]),
		    "buffer-time", 
		     buffer_time,
		     NULL); 
	    }

	private->elements[P_CDDA][PIPELINE] = gst_pipeline_new ("pipeline_cdda"); 

	gst_bin_add_many (GST_BIN (private->elements[P_CDDA][PIPELINE]),
		    private->elements[P_CDDA][SRC],	
		    private->elements[P_CDDA][CONVERT],
		    private->elements[P_CDDA][VOLUME],
		    private->elements[P_CDDA][SINK],
		    NULL);

	gst_element_link_many (private->elements[P_CDDA][SRC], 
			     private->elements[P_CDDA][CONVERT], 
			     private->elements[P_CDDA][VOLUME], 
			     private->elements[P_CDDA][SINK], NULL);

	gst_bus_add_watch (gst_pipeline_get_bus (GST_PIPELINE(private->elements[P_CDDA][PIPELINE])),
			  (GstBusFunc) bmp_play_bus_watch, self);
	goto pipeline_create_early_exit;
      }
    }

  pipeline_create_early_exit:

  g_free (sink);
  g_free (device);
}

void
tag_foreach (const GstTagList *list,
             const gchar      *tag,
             HTuple           *tuple)
{
  const GValue *value;
  gint i, count;

  count = gst_tag_list_get_tag_size (list, tag);

  for (i = 0; i < count; i++)
    {
      if (gst_tag_get_type (tag) == G_TYPE_STRING)
	{
	    gchar *str;
	    gst_tag_list_get_string_index (list, tag, i, &str);      

	    if (g_utf8_validate (str, strlen (str), NULL))
	      {
		h_tuple_insert (tuple,
			    tag,	
			    G_TYPE_STRING,
			    (gpointer)str);
	      }
	}
      else
	{
	    value = gst_tag_list_get_value_index (list, tag, i); 
	    if (G_VALUE_HOLDS (value, G_TYPE_INT))
	      { 
		  h_tuple_insert (tuple,
				  tag,	
				  G_TYPE_INT,
				  GINT_TO_POINTER(g_value_get_int(value)));	
	      }
	    else
	    if (G_VALUE_HOLDS (value, G_TYPE_UINT))
	      {
		gint _v_int;
		_v_int = g_value_get_uint (value);
		h_tuple_insert (tuple,
				tag,	
				G_TYPE_INT,
				GINT_TO_POINTER(_v_int));
	      }
	}
    }
}

typedef struct _TagReadTuple TagReadTuple;
struct _TagReadTuple {
    BmpPlay	  *self;
    HTuple	  *tuple;
    gboolean	   eos;
};
#define TAG_READ_TUPLE(_data) ((TagReadTuple*)(_data))

static gboolean
bmp_metadata_bus_handler (GstBus		  *bus,
			  GstMessage		  *message,
			  TagReadTuple		  *tuple)
{
  GstTagList *tag_list;
  
  switch (GST_MESSAGE_TYPE (message))
    {
      case GST_MESSAGE_EOS:
	{
	  tuple->eos = TRUE;
	}
	break;

      case GST_MESSAGE_TAG:
	{
	  gst_message_parse_tag (message, &tag_list);
	  if (tag_list)
	    {
	      gst_tag_list_foreach (tag_list, (GstTagForeachFunc) tag_foreach, tuple->tuple);
	      gst_tag_list_free (tag_list);
	    }
	}
	break;

      case GST_MESSAGE_ERROR:
	{
	  GError *error = NULL;
	  gst_message_parse_error (message, &error, NULL);
	  g_print ("%s ERROR: %s\n", G_STRLOC, error->message);
	  g_error_free (error);
	}
	break;

      default:
	break;
    }

  return FALSE;
}

static void 
bmp_metadata_event_loop (TagReadTuple	*tuple,
			 GstElement	*element,
			 gboolean	 block)
{
    GstBus *bus;
    gboolean done = FALSE;

    bus = gst_element_get_bus (element);
    g_return_if_fail (bus != NULL);

    while (!done && !tuple->eos)
      {
	  GstMessage *message;

	  if (block)
	      message = gst_bus_poll (bus, GST_MESSAGE_ANY, -1);
	  else
	      message = gst_bus_pop  (bus);

	  if (message == NULL)
	    {
	      gst_object_unref (bus);
	      return;
	    }

	  done = bmp_metadata_bus_handler (bus, message, tuple);
	  gst_message_unref (message);
      }

    gst_object_unref (bus);
}

HTuple*
bmp_play_metadata_get  (BmpPlay     *self,
                        const gchar *uri)
{
  GstElement	  *element, *fakesink, *videosink;
  HTuple	  *tuple;
  TagReadTuple	  *ttuple;
  GstState	   state_ret;
  GstQuery        *query;
  GstFormat        format = GST_FORMAT_TIME;
  gint		   change_timeout;
  gint64	   length_in_nanoseconds;
  gint		   length;
  gchar		  *uri_str;

  g_return_val_if_fail (BMP_IS_PLAY (self), NULL);
  
  if (str_has_suffix_nocase (uri, "jpg")) return NULL;

  ttuple = g_new0 (TagReadTuple, 1);
  ttuple->tuple = bmp_metadata_tuple_create_default (uri);
  ttuple->self = self;
  ttuple->eos = FALSE;

  h_tuple_insert (ttuple->tuple,
		  bmp_metadata_get_id_static (BMP_DATUM_TITLE),	
		  G_TYPE_STRING,
		  (gpointer)uri);	

  h_tuple_insert (ttuple->tuple,
		  bmp_metadata_get_id_static (BMP_DATUM_LOCATION),	
		  G_TYPE_STRING,
		  (gpointer)uri);	

  element =
      gst_element_factory_make ("playbin",  "playbin0");
  fakesink =
      gst_element_factory_make ("fakesink", "fakesink0");
  videosink =
      gst_element_factory_make ("fakesink", "fakesink1");

  uri_str = g_uri_get_unescaped (uri);
  g_object_set (G_OBJECT (element), "uri", uri_str, "audio-sink", fakesink, "video-sink", videosink, NULL);
  g_free (uri_str);

  state_ret = gst_element_set_state (element, GST_STATE_PAUSED);
  change_timeout = 5;
  while (state_ret == GST_STATE_CHANGE_ASYNC && change_timeout > 0)
    {
	GstState state;
	state_ret = gst_element_get_state (GST_ELEMENT (element), &state, NULL, 1 * GST_SECOND);
	change_timeout--;
    }

  if (state_ret != GST_STATE_CHANGE_FAILURE && state_ret != GST_STATE_CHANGE_ASYNC)
    {
	  /* Post application specific message so we'll know when to stop
	  * the message loop */
	  GstBus *bus;
	  bus = gst_element_get_bus (GST_ELEMENT (element));
	  if (bus)
	    {
		gst_bus_post (bus, 
		gst_message_new_application (GST_OBJECT (element), NULL));
		gst_object_unref (bus);
	    }

	  /* Poll the bus for messages */
	  bmp_metadata_event_loop (ttuple, GST_ELEMENT (element), FALSE);
    }

  query = gst_query_new_duration (format);
  if (gst_element_query (GST_ELEMENT(element), query))
    {
      gst_element_get_state (GST_ELEMENT(element), NULL, NULL, 100 * GST_MSECOND); 
      gst_query_parse_duration (query, &format, &length_in_nanoseconds);
      length = length_in_nanoseconds / GST_SECOND; 
      h_tuple_insert (ttuple->tuple,
		  bmp_metadata_get_id_static (BMP_DATUM_TIME),	
		  G_TYPE_INT,
		  (gpointer)GINT_TO_POINTER(length));	
    }
  gst_query_unref (query);

  gst_element_set_state (GST_ELEMENT (element), GST_STATE_NULL);
  gst_object_unref (GST_ELEMENT (element));

  tuple = ttuple->tuple;
  g_free (ttuple);

  return tuple;
}

static gboolean
stop_stream (BmpPlay *self)
{
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);

  g_return_val_if_fail (BMP_IS_PLAY (self), FALSE);
  g_return_val_if_fail ((private->elements[private->current_pipe][PIPELINE] != NULL), FALSE);

  gst_element_set_state (private->elements[private->current_pipe][PIPELINE], GST_STATE_READY);
  send_stream_pos_stop (self);
  return TRUE;
}

static gboolean
play_stream (BmpPlay *self)
{
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);

  g_return_val_if_fail (BMP_IS_PLAY (self), FALSE);
  g_return_val_if_fail ((private->elements[private->current_pipe][PIPELINE] != NULL), FALSE);
 
  if (g_uri_test (private->mrl, BMP_URI_TEST_PROTOCOL, PROTOCOL_HTTP) && (!private->elements[P_HTTP][PIPELINE]))
    {
	g_object_set (G_OBJECT(self), "playstatus", PLAYSTATUS_STOPPED, NULL);
	return FALSE;
    }

  if (g_uri_test (private->mrl, BMP_URI_TEST_PROTOCOL, PROTOCOL_CDDA) && (!private->elements[P_CDDA][PIPELINE]))
    {
	g_object_set (G_OBJECT(self), "playstatus", PLAYSTATUS_STOPPED, NULL);
	return FALSE;
    }

  if (gst_element_set_state (private->elements[private->current_pipe][PIPELINE], GST_STATE_PLAYING) ==
      GST_STATE_CHANGE_FAILURE)
    {
      g_message ("%s: Error in changing state to PLAYING", G_STRFUNC);
      gst_element_set_state (private->elements[private->current_pipe][PIPELINE], GST_STATE_READY);
      return FALSE;
    }

  g_object_set (private->elements[private->current_pipe][VOLUME], "mute", private->mute, NULL);
  g_object_set (private->elements[private->current_pipe][VOLUME], "volume", private->volume, NULL);

#if 0
  if (self->private->timeout_src_id)
    {
	source = g_main_context_find_source_by_id (NULL, self->private->timeout_src_id);
	if (source)
	  {
	    g_source_destroy (source);
	  }
    }
  private->timeout_src_id = g_timeout_add (60, (GSourceFunc) handoff_send_buffer, self);
#endif

  return TRUE;
}

static gboolean
pause_stream (BmpPlay *self)
{
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);

  g_return_val_if_fail (BMP_IS_PLAY (self), FALSE);
  g_return_val_if_fail ((private->elements[private->current_pipe][PIPELINE] != NULL), FALSE);

  if  (GST_STATE (private->elements[private->current_pipe][PIPELINE]) == GST_STATE_PAUSED)
    {
	g_object_set (G_OBJECT(self), "playstatus", PLAYSTATUS_PLAYING, NULL);
    }

  return gst_element_set_state (private->elements[private->current_pipe][PIPELINE], GST_STATE_PAUSED);
}

static gboolean
seek_stream (BmpPlay *self,
             gint     offset)
{
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);
  GstEvent	 *seek_event;
  guint64	  seek_offset;

  g_return_val_if_fail (BMP_IS_PLAY (self), FALSE);
  g_return_val_if_fail ((private->elements[private->current_pipe][PIPELINE] != NULL), FALSE);

  seek_offset = (offset * GST_SECOND);
  seek_event = gst_event_new_seek (1.0, GST_FORMAT_TIME, GST_SEEK_FLAG_FLUSH, 
			GST_SEEK_TYPE_SET, seek_offset, GST_SEEK_TYPE_NONE, 0);

  if (gst_element_send_event (private->elements[private->current_pipe][PIPELINE], seek_event))
    {
	gst_pipeline_set_new_stream_time (GST_PIPELINE(private->elements[private->current_pipe][PIPELINE]), 0);	
	gst_element_get_state (GST_ELEMENT(private->elements[private->current_pipe][PIPELINE]), NULL, NULL, 50 * GST_MSECOND); 
	g_signal_emit (G_OBJECT(self), signals[BMP_PLAY_SIGNAL_SEEK], 0, offset);
	return TRUE;
    }
  else
    {
	return FALSE;
    }

#if 0
  if (gst_element_seek (private->elements[private->current_pipe][PIPELINE], 1.0, GST_FORMAT_TIME, GST_SEEK_FLAG_FLUSH, 
			GST_SEEK_TYPE_SET, seek_offset, GST_SEEK_TYPE_NONE, -1))
    {
	gst_pipeline_set_new_stream_time (GST_PIPELINE(private->elements[private->current_pipe][PIPELINE]), 0);	
	gst_element_get_state (GST_PIPELINE(private->elements[private->current_pipe][PIPELINE]), NULL, NULL, 50 * GST_MSECOND); 
	g_signal_emit (G_OBJECT(self), signals[BMP_PLAY_SIGNAL_SEEK], 0, offset);
	return TRUE;
    }
  else
    {
	return FALSE;
    }
#endif

  g_assert_not_reached ();
}

GstState
bmp_play_get_element_state (BmpPlay *self)
{
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);

  g_return_val_if_fail (BMP_IS_PLAY (self), GST_STATE_NULL);

  return GST_STATE (private->elements[private->current_pipe][PIPELINE]);
}

static gint
get_time (BmpPlay *self)
{
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);
  GstQuery       *query;
  GstFormat       format = GST_FORMAT_TIME;
  gint64          time_in_nanoseconds;
  gint            time_in_seconds;

  g_return_val_if_fail (BMP_IS_PLAY (self), 0);
  g_return_val_if_fail ((private->elements[private->current_pipe][PIPELINE] != NULL), 0);

  if (!(GST_STATE (private->elements[private->current_pipe][PIPELINE]) == GST_STATE_PLAYING)) return 0;

  query = gst_query_new_position (format);
  gst_element_query (private->elements[private->current_pipe][SINK], query);
  gst_query_parse_position (query, &format, &time_in_nanoseconds);
  time_in_seconds = time_in_nanoseconds / GST_SECOND; 

  gst_query_unref (query);
  return time_in_seconds;
}

gint
bmp_play_get_length (BmpPlay *self)
{
  BmpPlayPrivate *private;
  GstQuery       *query;
  GstFormat       format = GST_FORMAT_TIME;
  gint64          length_in_nanoseconds;

  g_return_val_if_fail (BMP_IS_PLAY (self), 0);

  private = BMP_PLAY_GET_PRIVATE (self);

  if (!private->elements[private->current_pipe][PIPELINE]) return 0;
  if (!(GST_STATE (private->elements[private->current_pipe][PIPELINE]) == GST_STATE_PLAYING)) return 0;

  query = gst_query_new_duration (format);
  gst_element_query (private->elements[private->current_pipe][PIPELINE], query);
  gst_query_parse_duration (query, &format, &length_in_nanoseconds);
  length_in_nanoseconds = length_in_nanoseconds / GST_SECOND; 

  gst_query_unref (query);
  return length_in_nanoseconds;
}

gint
bmp_play_get_status (BmpPlay *self)
{
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);
  gint status;

  g_return_val_if_fail (BMP_IS_PLAY (self), PLAYSTATUS_STOPPED);

  if (!private->elements[private->current_pipe][PIPELINE]) return 0;

  status = GST_STATE (private->elements[private->current_pipe][PIPELINE]);

  if (status == GST_STATE_READY)
    return PLAYSTATUS_STOPPED;

  if (status == GST_STATE_PLAYING)
    return PLAYSTATUS_PLAYING;

  return status;
}

static gboolean
playstatus_notify_cb (BmpPlay *self)
{
  gint status;

  g_return_val_if_fail (BMP_IS_PLAY (self), FALSE);
  g_object_get (self, "playstatus", &status, NULL);

  if (status & PLAYSTATUS_SEEKING)
    {
      if (!pause_stream (self)); //ERROR
      return TRUE;
    }

  if (status & PLAYSTATUS_STOPPED)
    {
      if (!stop_stream (self)); //ERROR 
      return TRUE;
    }

  if (status & PLAYSTATUS_WAITING)
    {
      if (!stop_stream (self)); //ERROR
      return TRUE;
    }

  if (status & PLAYSTATUS_PAUSED)
    {
      if (!pause_stream (self)); //ERROR
      return TRUE;
    }

  if (status & PLAYSTATUS_PLAYING)
    {
     if (!play_stream (self)); //ERROR
      return TRUE;
    }

  return FALSE;
}

static void
bmp_play_set_property (GObject      *object,
                       guint         property_id,
                       const GValue *value,
                       GParamSpec   *pspec)
{
  BmpPlay        *self    = (BmpPlay *) object;
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);

  switch (property_id)
  {

  case BMP_PLAY_PROP_POSITION:
    {
	seek_stream (self, g_value_get_int (value));
    }
    break;

  case BMP_PLAY_PROP_VOLUME:
    {
      gdouble _volume;

      _volume = g_value_get_int (value) / 100.;
      private->volume = _volume;
      g_object_set (BMP_PLAY_GET_PRIVATE (self)->elements[private->current_pipe][VOLUME], "volume", _volume, NULL);
    }
    break;

  case BMP_PLAY_PROP_MUTE:
    {
      private->mute = g_value_get_boolean (value); 

      if (private->elements[P_FILE])
	{
	    g_object_set (G_OBJECT(private->elements[P_FILE][VOLUME]),
		    "mute",
		    private->mute,
		    NULL);
	}

      if (private->elements[P_CDDA])
	{
	    g_object_set (G_OBJECT(private->elements[P_HTTP][VOLUME]),
		    "mute",
		    private->mute,
		    NULL);
	}

      if (private->elements[P_HTTP])
	{
	    g_object_set (G_OBJECT(private->elements[P_HTTP][VOLUME]),
		    "mute",
		    private->mute,
		    NULL);
	}
    }
    break;

  case BMP_PLAY_PROP_PLAYSTATUS:
    {
      private->playstatus = g_value_get_int (value);
    }
    break;

  case BMP_PLAY_PROP_STREAM:
    {
      gchar   *filename, 
	      *host;

      if (!private->sane) return;

      g_free (private->mrl);
      private->mrl = g_value_dup_string (value);
      if (g_uri_test (private->mrl, BMP_URI_TEST_PROTOCOL, PROTOCOL_FILE))
	{
	  private->current_pipe = P_FILE;
	  filename = g_filename_from_uri (private->mrl, &host, NULL);      
	  g_object_set (private->elements[private->current_pipe][SRC], "location", filename, NULL);
	  g_free (filename);
	}
      else if (g_uri_test (private->mrl, BMP_URI_TEST_PROTOCOL, PROTOCOL_HTTP))
	{
	  private->current_pipe = P_HTTP;
    
	  if (private->http_sink == HAVE_NEON)
	    {
		g_object_set (private->elements[private->current_pipe][SRC], "location", private->mrl, NULL);
	    }
	  if (private->http_sink == HAVE_GNOMEVFS)
	    {
		g_object_set (private->elements[private->current_pipe][SRC], "location", private->mrl, NULL);
	    }
	}
      else if (g_uri_test (private->mrl, BMP_URI_TEST_PROTOCOL, PROTOCOL_CDDA))
	{
	  BMPURI *bmp_uri;
	  guint track;

	  bmp_uri = bmp_uri_new (private->mrl);
	  track = (guint)atoi(bmp_uri->path+1);

	  private->current_pipe = P_CDDA;
	  g_object_set (private->elements[private->current_pipe][SRC], "track", track, NULL);
	  bmp_uri_delete (bmp_uri);
	}

    }
    break;

  default:
    g_assert (FALSE);
  }
}

static void
bmp_play_get_property (GObject    *object,
                       guint       property_id,
                       GValue     *value,
                       GParamSpec *pspec)
{
  BmpPlay        *self = (BmpPlay *) object;
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);

  switch (property_id)
    {
    case BMP_PLAY_PROP_POSITION:
      {
	g_value_set_int (value, get_time(self));
      }
      break;

    case BMP_PLAY_PROP_SANE:
      {
	g_value_set_boolean (value, private->sane); 
      }
      break;

    case BMP_PLAY_PROP_VOLUME:
      {
	gdouble volume;
	g_object_get (private->elements[private->current_pipe][VOLUME], "volume", &volume, NULL);
	g_value_set_int (value, (gint)(volume * 100));
      }
      break;

    case BMP_PLAY_PROP_MUTE:
      g_value_set_boolean (value, private->mute);
      break;

    case BMP_PLAY_PROP_HAS_CDDA:
      g_value_set_boolean (value, private->has_cdda);
      break;

    case BMP_PLAY_PROP_HAS_HTTP:
      g_value_set_boolean (value, private->has_http);
      break;

    case BMP_PLAY_PROP_STREAM:
      g_value_set_string (value, private->mrl);
      break;

    case BMP_PLAY_PROP_PLAYSTATUS:
      g_value_set_int (value, private->playstatus);
      break;

    default:
      g_assert (FALSE);
    }
}

static void
bmp_play_dispose (GObject *obj)
{
  BmpPlay        *self    = (BmpPlay *) obj;
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);

  if (private->dispose_has_run) return;

  private->dispose_has_run = TRUE;
  pipelines_destroy (self);

  g_message ("stopped: %s", G_OBJECT_TYPE_NAME(obj));
}

static void
bmp_play_finalize (GObject *obj)
{
  /* nothing to do */
}

static void
bmp_play_class_init (BmpPlayClass * klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (klass, sizeof (BmpPlayPrivate));

  gobject_class->set_property = bmp_play_set_property;
  gobject_class->get_property = bmp_play_get_property;
  gobject_class->dispose = bmp_play_dispose;
  gobject_class->finalize = bmp_play_finalize;
  gobject_class->constructor = bmp_play_constructor;

  g_object_class_install_property (gobject_class, BMP_PLAY_PROP_SANE,
                                   g_param_spec_boolean ("sane",
                                                         "sane",
                                                         "sane",
                                                          FALSE,
                                                          G_PARAM_READABLE));

  g_object_class_install_property (gobject_class, BMP_PLAY_PROP_MUTE,
                                   g_param_spec_boolean ("mute",
                                                         "mute",
                                                         "mute",
                                                          FALSE,
                                                          G_PARAM_READWRITE));

  g_object_class_install_property (gobject_class, BMP_PLAY_PROP_HAS_CDDA,
                                   g_param_spec_boolean ("has_cdda",
                                                         "has_cdda",
                                                         "has_cdda",
                                                          FALSE,
                                                          G_PARAM_READABLE));

  g_object_class_install_property (gobject_class, BMP_PLAY_PROP_HAS_HTTP,
                                   g_param_spec_boolean ("has_http",
                                                         "has_http",
                                                         "has_http",
                                                          FALSE,
                                                          G_PARAM_READABLE));

  g_object_class_install_property (gobject_class, BMP_PLAY_PROP_STREAM,
                                   g_param_spec_string ("mrl",
                                                        "mrl",
                                                        "mrl",
                                                        "",
                                                        G_PARAM_READWRITE));

  g_object_class_install_property (gobject_class, BMP_PLAY_PROP_PLAYSTATUS,
                                   g_param_spec_int ("playstatus",
                                                     "playstatus",
                                                     "player playing status",
                                                     0, 128, 0,
                                                     G_PARAM_READWRITE));

  g_object_class_install_property (gobject_class, BMP_PLAY_PROP_VOLUME,
                                   g_param_spec_int ("volume",
                                                     "volume",
                                                     "volume",
                                                     0, 100, 0,
                                                     G_PARAM_READWRITE));

  g_object_class_install_property (gobject_class, BMP_PLAY_PROP_POSITION,
                                   g_param_spec_int ("position",
                                                     "position",
                                                     "position",
                                                     0, G_MAXINT, 0,
                                                     G_PARAM_READWRITE));


  signals[BMP_PLAY_SIGNAL_TITLE] =
   g_signal_new ("title",
                  G_OBJECT_CLASS_TYPE (gobject_class),
                  G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
                  G_STRUCT_OFFSET (BmpPlayClass, title),
                  NULL, NULL, g_cclosure_marshal_VOID__STRING,
                  G_TYPE_NONE, 1, G_TYPE_STRING);

  signals[BMP_PLAY_SIGNAL_BUFFER] =
    g_signal_new ("buffer",
                  G_OBJECT_CLASS_TYPE (gobject_class),
                  G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
                  G_STRUCT_OFFSET (BmpPlayClass, buffer),
                  NULL, NULL, g_cclosure_marshal_VOID__POINTER,
                  G_TYPE_NONE, 1, G_TYPE_POINTER);

  signals[BMP_PLAY_SIGNAL_EOS] =
    g_signal_new ("eos",
                  G_OBJECT_CLASS_TYPE (gobject_class),
                  G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
                  G_STRUCT_OFFSET (BmpPlayClass, eos),
                  NULL, NULL, g_cclosure_marshal_VOID__VOID, G_TYPE_NONE,
                  0);

  signals[BMP_PLAY_SIGNAL_ERROR] =
    g_signal_new ("error",
                  G_OBJECT_CLASS_TYPE (gobject_class),
                  G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
                  G_STRUCT_OFFSET (BmpPlayClass, error),
                  NULL, NULL, g_cclosure_marshal_VOID__POINTER,
                  G_TYPE_NONE, 1, G_TYPE_POINTER);

  signals[BMP_PLAY_SIGNAL_SEEK] =
    g_signal_new ("seek",
                  G_OBJECT_CLASS_TYPE (gobject_class),
                  G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
                  G_STRUCT_OFFSET (BmpPlayClass, seek),
                  NULL, NULL, g_cclosure_marshal_VOID__INT,
                  G_TYPE_NONE, 1, G_TYPE_INT);
}

BmpPlay *
bmp_play_new (void)
{
    BmpPlay *play;

    play = g_object_new (BMP_TYPE_PLAY, NULL);
    return play;
}

static GObject *
bmp_play_constructor (GType type,
                      guint n_construct_properties,
                      GObjectConstructParam * construct_properties)
{
  GObject      *obj;
  BmpPlayClass *klass;
  GObjectClass *parent_class;

  klass = PLAY_CLASS (g_type_class_peek (BMP_TYPE_PLAY));

  parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));

  obj = parent_class->constructor (type,
                                   n_construct_properties,
                                   construct_properties);
  return obj;
}

static gboolean
_idle_emit_eos (BmpPlay * play)
{
  g_return_val_if_fail (BMP_IS_PLAY (play), FALSE);

  g_signal_emit (G_OBJECT (play), signals[BMP_PLAY_SIGNAL_EOS], 0, NULL);

  return FALSE;
}

static gboolean
bmp_play_bus_watch (GstBus     *bus,
                    GstMessage *message,
                    BmpPlay    *self)
{
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);
  GstState old_state, new_state, pending_state;

  switch (GST_MESSAGE_TYPE (message))
    {
    case GST_MESSAGE_STATE_CHANGED:
      {
        gst_message_parse_state_changed (message, &old_state,
                                         &new_state, &pending_state);

	if ((old_state < GST_STATE_PLAYING) && (new_state == GST_STATE_PLAYING) && (!private->sending))
	  {
	      private->source = g_timeout_add_full (G_PRIORITY_DEFAULT, 250, (GSourceFunc) send_stream_pos, self, NULL);
	      private->sending = TRUE;
	  }
      }
      break;

    case GST_MESSAGE_ERROR:
      {
	GstElement  *element;
        GError	    *error = NULL;
	gchar	    *debug_string,
		    *name,
		    *location;
	GString	    *message_str;

	element = GST_ELEMENT(GST_MESSAGE_SRC (message));
	name = gst_object_get_name (GST_OBJECT(element));
	location = g_object_get_string (private->elements[private->current_pipe][SRC], "location"); 

	message_str = g_string_new ("<b>ERROR ");
	g_string_append_printf (message_str, "[Element: <i>%s</i>]</b>\n", name);

        gst_message_parse_error	(message, 
				 &error,
				 &debug_string);

	if (error->domain == GST_CORE_ERROR)
	  {
		      switch (error->code)
			{
			    case GST_CORE_ERROR_MISSING_PLUGIN:
			      {
				  g_string_append (message_str, "(E101) There is no plugin available to play this track\n");
				  break;
			      }

			    case GST_CORE_ERROR_SEEK:
			      {
				  g_string_append (message_str, "(E102) An error occured during seeking\n");
				  break;
			      }

			    case GST_CORE_ERROR_STATE_CHANGE:
			      {
				  typedef struct _StringPairs StringPairs; 
				  struct _StringPairs {
				    gchar   *state_pre;
				    gchar   *state_post;
				  };

				  static StringPairs string_pairs[] = {
				      {"NULL",	  "READY"},
				      {"READY",	  "PAUSED"},
				      {"PAUSED",  "PLAYING"},
				      {"PLAYING", "PAUSED"},
				      {"PAUSED",  "READY"},
				      {"READY",	  "NULL"},
				  };
			      
				  g_string_append_printf (message_str, "(E103) An error occured during changing the state from %s to %s (note that these are internal states and don't neccesarily reflect what you might understand by e.g. 'play' or 'pause'\n",
							   string_pairs[GST_STATE_PENDING(message->src)].state_pre,
							   string_pairs[GST_STATE_PENDING(message->src)].state_post);

				  break;
			      }

			    case GST_CORE_ERROR_PAD:
			      {
				  g_string_append (message_str, "(E104) An error occured during pad linking/unlinking\n");
				  break;
			      }

			    case GST_CORE_ERROR_NEGOTIATION:
			      {
				  g_string_append (message_str, "(E105) An error occured during element(s) negotiation\n");
				  break;
			      }

			    default: break;
			}
	  }
	else if (error->domain == GST_RESOURCE_ERROR)
	  {
		      switch (error->code)
			{

			    case GST_RESOURCE_ERROR_SEEK:
			      {
				  g_string_append (message_str, "(E201) An error occured during seeking\n");
				  break;
			      }

			    case GST_RESOURCE_ERROR_NOT_FOUND:
			    case GST_RESOURCE_ERROR_OPEN_READ:
			      {
				  g_string_append (message_str, "(E202) Couldn't open the stream/track\n");
				  break;
			      }

			    default: break;
			}
	    }
	else if (error->domain == GST_STREAM_ERROR)
	    {
		      switch (error->code)
			{
			    case GST_STREAM_ERROR_CODEC_NOT_FOUND:
			      {
				  g_string_append (message_str, "(E301) There is no codec installed to handle this stream\n");
				  break;
			      }

			    case GST_STREAM_ERROR_DECODE:
			      {
				  g_string_append (message_str, "(E302) There was an error decoding the stream\n");
				  break;
			      }

			    case GST_STREAM_ERROR_DEMUX:
			      {
				  g_string_append (message_str, "(E303) There was an error demultiplexing the stream\n");
				  break;
			      }

			    case GST_STREAM_ERROR_FORMAT:
			      {
				  g_string_append (message_str, "(E304) There was an error parsing the format of the stream\n");
				  break;
			      }

			    default: break;
			}
	    }
	  

	g_string_append_printf (message_str, "\n<b>Current URI:</b>\n%s\n\n<b>Detailed debugging information:</b>\n%s", location, debug_string);
	g_free (debug_string);
        g_error_free (error);

	bmp_system_control_message_dispatch (bmp_system_control,
					     private->message_domain_id, 
					     GTK_MESSAGE_ERROR,
					     message_str->str);

	g_string_free (message_str, TRUE); 

	g_object_set (G_OBJECT(self), "playstatus", PLAYSTATUS_STOPPED, NULL);
	pipelines_destroy (self);

	if (strcmp(name, "sink"))
	  {
	    pipeline_create (self, P_FILE);
	    pipeline_create (self, P_HTTP);
	    pipeline_create (self, P_CDDA);
	  }
	else
	  {
	    private->sane = FALSE;
	  }

	g_free (name);

	break;
      }

    case GST_MESSAGE_WARNING:
      {
        GError  *error = NULL;
	gchar   *debug;

        gst_message_parse_warning (message, &error, &debug);
        g_print ("%s WARNING: %s (%s)\n", G_STRLOC, error->message, debug);
        g_error_free (error);
      }
    break;

    case GST_MESSAGE_EOS:
        {
            /* end-of-stream */
            g_idle_add ((GSourceFunc) _idle_emit_eos, self);
        }
        break;

    default:
        /* unhandled message */
        break;
    }

    return TRUE; 
}

void
bmp_play_reset_pipes  (BmpPlay *self)
{
  BmpPlayPrivate  *private = BMP_PLAY_GET_PRIVATE (self);
  GstElement	  *element;
  gchar		  *sink;

  pipelines_destroy (self);

  //We check if the sink can be created
  sink = xcs_key_fetch_string (xcs, "audio", "sink");
  element = gst_element_factory_make (sink, "sinktest0"); 

  if (!element)
    {
	bmp_system_control_message_dispatch (bmp_system_control,
				       private->message_domain_id, 
				       GTK_MESSAGE_ERROR,
				       _("BMP was not able to initialize the playback system. Please check the audio settings in the Preferences panel."));
	g_free (sink);
	private->sane = FALSE;  
	return;
    }

  //FIXME: Possibly even more sanity checks
  g_free (sink);
  private->sane = TRUE;  

  pipeline_create (self, P_FILE);
  pipeline_create (self, P_HTTP);
  pipeline_create (self, P_CDDA);
}

static void
bmp_play_init (BmpPlay *self)
{
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);
  gint		  volume;

  private->dispose_has_run = FALSE;
  private->playstatus = PLAYSTATUS_STOPPED;
  private->mrl = NULL;
  private->sending = FALSE;
  private->source = 0;
  private->current_pipe = P_FILE;
  private->has_cdda = FALSE;
  private->has_http = FALSE;

  private->sane = FALSE;

#if 0
  private->fifo_buffer = g_ptr_array_new ();
  private->fifo_buffer_timestamps = g_ptr_array_new ();
  private->fifo_buffer_length = 0;
  private->fifo_buffer_new_items = 0;
#endif

  private->mute = FALSE;
  private->timeout_src_id = 0;

  bmp_system_control_volume_get (bmp_system_control, &volume, NULL); 
  private->volume = volume/100.;
 
  private->message_domain_id =
	bmp_system_control_message_domain_register (bmp_system_control,
						    "play", 
						    _("BMP: Audio"));

  nullify_elements (self);
  bmp_play_reset_pipes (self);

  g_signal_connect (G_OBJECT (self), "notify::playstatus",
                    G_CALLBACK (playstatus_notify_cb), NULL);

  private->extensions = NULL;
  //Check supported extensions
  {
    GList  *factories;
    gchar **list;
    gint    n, k;

    k = 0;
    factories = gst_type_find_factory_get_list ();
    for (;factories;factories=factories->next)
      {
	list = gst_type_find_factory_get_extensions (factories->data);
	if (!list) continue;
	for (n = 0; list[n]; n++) 
	  {
	    private->extensions = realloc (private->extensions, (sizeof(char*))*(k+1));
	    private->extensions[k] = g_strdup(list[n]);
	    k++;
	  }
      }
    private->extensions = realloc (private->extensions, (sizeof(char*))*(k+1));
    private->extensions[k] = NULL; 
  }
}

gboolean
bmp_play_is_audio_file (BmpPlay	      *self,
			const gchar   *uri)
{
  BmpPlayPrivate *private = BMP_PLAY_GET_PRIVATE (self);

  if (!uri) return FALSE;

  if (g_uri_test (uri, BMP_URI_TEST_PROTOCOL, PROTOCOL_FILE))
    {
      gchar *filename;
      filename = g_filename_from_uri (uri, NULL, NULL);

      if (g_file_test (filename, G_FILE_TEST_IS_DIR))
	{
	  g_free (filename);
	  return FALSE;
	}
      g_free (filename);
    } 

  return str_has_suffixes_nocase (uri, private->extensions);
}
