/*  BMPx - The Dumb Music Player

 *  Copyright (C) 2003-2006 BMP developers
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  $Id$
 *
 * The BMPx project hereby grant permission for non-gpl compatible GStreamer
 * plugins to be used and distributed together with GStreamer and BMPx. This
 * permission are above and beyond the permissions granted by the GPL license
 * BMPx is covered by.
 */


#include <config.h>

#ifdef HAVE_GUI
#include <gtk/gtk.h>
#endif

#include <glib/gstdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>

#include <bmp/util.h>
#include <bmp/vfs.h>
#include <bmp/uri.h>
#include <bmp/plugin.h>
#include <bmp/plugin_interfaces.h>
#include <bmp/metadata.h>
#include <bmp/playlist.h>

#include <goa/libgoa.h>

#include <libhrel/types.h>
#include <libhrel/tuple.h>

#include "bmp-marshalers.h"

G_DEFINE_TYPE (BmpPlaylist, bmp_playlist, G_TYPE_OBJECT)

#include "loader.h"
#include "main.h"

typedef enum {
        BMP_PLAYLIST_SIGNAL_TRACKLIST_LIST_SORTED,
	BMP_PLAYLIST_SIGNAL_TRACKLIST_ITEMS_REMOVED,
	BMP_PLAYLIST_SIGNAL_TRACKLIST_ITEMS_ADDED,

	BMP_PLAYLIST_N_SIGNALS
} BmpPlaylistSignals;

static guint signals[BMP_PLAYLIST_N_SIGNALS] = { 0, 0, 0 };

typedef enum {
	BMP_PLAYLIST_PROP_0,

	BMP_PLAYLIST_N_PROPERTIES
} BmpPlaylistProperties;

struct _BmpPlaylistPrivate
{
        gboolean      dispose_has_run;

	GtkListStore *playlist;
};

static GHashTable *row_guid_history_map = NULL;
static GHashTable *row_guid_index_map   = NULL;

static gint vcolumns[] = {
    BMP_DATUM_ARTIST,
    BMP_DATUM_ALBUM,
    BMP_DATUM_TRACK,
    BMP_DATUM_TITLE,
    BMP_DATUM_GENRE,
    BMP_DATUM_DATE,
    BMP_DATUM_TIME
};

static void
save_playlist (BmpPlaylist *self);

static void
load_playlist (BmpPlaylist *self);

gint
vcolumns_get_n_items (void)
{
  return G_N_ELEMENTS (vcolumns);
}

gint
vcolumns_get_item_nth (gint n)
{
  return vcolumns[n];
}

/* History system */

#define PLAYBACK_HISTORY_LIMIT 1000

static void
history_free (BmpPlaybackHistory *history)
{
}

BmpPlaybackHistory*
bmp_playback_history_new (void)
{
    BmpPlaybackHistory *history;

    history = g_new0 (BmpPlaybackHistory,1);
    history->history = g_queue_new (); 
    history->mark = 0; 
    return history;
}

void
bmp_playback_history_free (BmpPlaybackHistory *history)
{
    if (!history) return;
    history_free (history);
}

static gboolean
history_item_prepend (BmpPlaybackHistory     *history,
		      BmpPluginFlowInterface *flow_plugin)
{
    BmpFileItem		  *item;
    GtkTreeModel	  *model;
    GtkTreeRowReference	  *reference;
    GtkTreePath		  *path;
    GtkTreeIter		   iter;

    g_return_val_if_fail (history != NULL, FALSE);

    model = GTK_TREE_MODEL(bmp_playlist_get_playlist (bmp_playlist));

    if (!bmp_plugin_flow_item_prev ((BmpPluginFlowInterface*)flow_plugin,
				    GTK_LIST_STORE(model),
				    &iter))
	return FALSE;

    if (history->history->length+1 > PLAYBACK_HISTORY_LIMIT)
	g_queue_pop_tail (history->history);

    path = gtk_tree_model_get_path (model, &iter);

    reference = gtk_tree_row_reference_new (model, path);
    g_assert (reference);
    gtk_tree_path_free (path);

    g_queue_push_head (history->history, reference);

    gtk_tree_model_get (model, &iter, TUPLE_COLUMN, &item, -1);
    g_hash_table_insert (row_guid_history_map, item->guid, g_queue_peek_nth_link (history->history, 0)); 

    history->mark = 1;
    return TRUE;
}

static gboolean
history_item_append (BmpPlaybackHistory	    *history,
		     BmpPluginFlowInterface *flow_plugin)
{
    BmpFileItem		  *item;
    GtkTreeModel	  *model;
    GtkTreeRowReference	  *reference;
    GtkTreePath		  *path;
    GtkTreeIter		   iter;

    g_return_val_if_fail (history != NULL, FALSE);

    model = GTK_TREE_MODEL(bmp_playlist_get_playlist (bmp_playlist));

    if (!bmp_plugin_flow_item_next ((BmpPluginFlowInterface*)flow_plugin,
				    GTK_LIST_STORE(model),
				    &iter))
	{
	    return FALSE;
	}


    if (history->history->length+1 > PLAYBACK_HISTORY_LIMIT)
	g_queue_pop_head (history->history);

    path = gtk_tree_model_get_path (model, &iter);
    reference = gtk_tree_row_reference_new (model, path);
    g_assert (reference);
    gtk_tree_path_free (path);

    g_queue_push_tail (history->history, reference);

    gtk_tree_model_get (model, &iter, TUPLE_COLUMN, &item, -1);
    g_hash_table_insert (row_guid_history_map, item->guid, g_queue_peek_nth_link (history->history, history->history->length-1)); 

    history->mark++;
    return TRUE;
}

void
bmp_playback_history_rewind (BmpPlaybackHistory *history) 
{
    g_return_val_if_fail (history != NULL, -1);

    history->mark = 1; 
}

void
bmp_playback_history_set (BmpPlaybackHistory *history,
			  gint		      position)

{
    GtkTreeModel	  *model;
    GtkTreeRowReference	  *reference;
    GtkTreePath		  *path;

    g_return_if_fail (history != NULL);

    model = GTK_TREE_MODEL(bmp_playlist_get_playlist (bmp_playlist));
    path = gtk_tree_path_new_from_indices (position, -1); 
    reference = gtk_tree_row_reference_new (model, path);
    g_assert (reference);
    gtk_tree_path_free (path);

    if (history->mark < history->history->length)
      {
	  gint n;

	  for (n = 0; n < (history->history->length - history->mark); n++)
	    {
		g_queue_pop_tail (history->history);
	    }
#if 0
	  GList *tail_prev_link,
		*tail_link;	

	  tail_prev_link = g_list_nth ((GList*)history->history, history->mark-1);
	  tail_link = g_list_nth ((GList*)history->history, history->mark);

	  tail_prev_link->next = NULL;    
	  tail_link->prev = NULL;
	  g_list_foreach (tail_link, (GFunc)gtk_tree_row_reference_free, NULL);

	  history->history->head = g_list_first (tail_link);
	  history->history->length = g_list_length (history->history->head);
#endif
      }

    g_queue_push_tail (history->history, reference);
    history->mark++;
}

gint
bmp_playback_history_prev (BmpPlaybackHistory     *history, 
			   BmpPluginFlowInterface *flow_plugin)
{
    GtkTreePath		  *path;
    GtkTreeRowReference	  *reference;
    gint row;

    g_return_val_if_fail (history != NULL, -1);

    row = -1;
    item_back_iteration:

    if (history->mark == 0)
      {
	if (history_item_prepend (history, flow_plugin))
	  {
	    reference = g_queue_peek_nth (history->history, history->mark-1);
	    g_assert (reference);
	    path = gtk_tree_row_reference_get_path (reference);
	    g_assert (path);

	    row = gtk_tree_path_get_indices (path)[0]; 
	    gtk_tree_path_free (path);
	  }
      }
    else
      {
	history->mark--;

	if (history->mark == 0)
	  {
	    goto item_back_iteration; 
	  }
	else
	  {
	    reference = g_queue_peek_nth (history->history, history->mark-1);
	    path = gtk_tree_row_reference_get_path (reference);
	    if (!path)
	      {
		g_queue_pop_nth (history->history, history->mark);

		if (gtk_tree_row_reference_valid (reference)) gtk_tree_row_reference_free (reference);

		if (history->mark > history->history->length)
		    history->mark = history->history->length;

		goto item_back_iteration;
	      }

	    row = gtk_tree_path_get_indices (path)[0]; 
	    gtk_tree_path_free (path);
	  }
      }

    return row;
}

gint
bmp_playback_history_next (BmpPlaybackHistory	    *history, 
			   BmpPluginFlowInterface   *flow_plugin) 
{
    GtkTreePath		  *path;
    GtkTreeRowReference	  *reference;
    gint row;

    g_return_val_if_fail (history != NULL, -1);

    row = -1;
    item_forward_iteration:

    if (history->mark == history->history->length)
      {
	if (history_item_append (history, flow_plugin))
	  {
	    reference = g_queue_peek_nth (history->history, history->mark-1);
	    if (!reference)
	      {
		abort ();
	      }

	    path = gtk_tree_row_reference_get_path (reference);
	    if (!path)
	      {
		abort ();
	      }

	    row = gtk_tree_path_get_indices (path)[0]; 
	    gtk_tree_path_free (path);
	  }
      }
    else
      {
	history->mark++;

	reference = g_queue_peek_nth (history->history, history->mark-1);
	g_assert (reference);

	path = gtk_tree_row_reference_get_path (reference);
	if (!path)
	  {
	    g_queue_pop_nth (history->history, history->mark-1);

	    if (gtk_tree_row_reference_valid (reference)) gtk_tree_row_reference_free (reference);

	    if (history->mark > history->history->length)
		history->mark = history->history->length;

	    goto item_forward_iteration;
	  }

	row = gtk_tree_path_get_indices (path)[0]; 
	gtk_tree_path_free (path);
      }

    return row;
}

void
bmp_playlist_tracklist_sort (BmpPlaylist      *self,
			     gint	       metadatum_id)
{
#if 0
        g_signal_emit (G_OBJECT (self),
                   signals[BMP_PLAYLIST_SIGNAL_TRACKLIST_LIST_SORTED], 0);
#endif
}


void
bmp_playlist_sort_current_list (BmpPlaylist *self)
{
//	bmp_playlist_tracklist_sort (self, -1);
}

void
bmp_playlist_sort_current_list_by_tag (BmpPlaylist *self, BmpMetadatumId metadatum_id)
{
//	bmp_playlist_tracklist_sort (self, metadatum_id);
}

static void
update_row_guid_mappings (GtkTreeModel  *model,
                          GtkTreePath	*path,
                          gpointer	 data)
{
    BmpFileItem		  *item;
    GtkTreeIter		   iter;
    BmpPlaybackHistory	  *history;

    gtk_tree_model_get_iter (model, &iter, path);
    gtk_tree_model_get (model, &iter, TUPLE_COLUMN, &item, -1);

    history = bmp_system_control_get_history (bmp_system_control);

#if 0
    link = g_hash_table_lookup (row_guid_history_map, item->guid);
    link_p = g_queue_pop_nth_link (history->history, g_queue_link_index (history->history, link));

    gtk_tree_row_reference_free (link_p->data);
#endif

    g_hash_table_remove (row_guid_history_map, item->guid);
    g_hash_table_remove (row_guid_index_map, item->guid);

#if 0
    if (history->mark > history->history->length)
      {
	history->mark = history->history->length;
      }
#endif
}

static void
on_playlist_sort_column_changed   (GtkTreeSortable *treesortable,
                                   gpointer         data)      
{
    GtkSortType  order;
    gint	 column;

    gtk_tree_sortable_get_sort_column_id (treesortable, &column, &order);

#if 0
        g_signal_emit (G_OBJECT (data),
                   signals[BMP_PLAYLIST_SIGNAL_TRACKLIST_LIST_SORTED], 0, column);
#endif
}

static void
bmp_playlist_init (BmpPlaylist  *self)
{
        self->private = g_new0(BmpPlaylistPrivate, 1);

        self->private->dispose_has_run = FALSE;
	self->private->playlist = gtk_list_store_new (8, 
						      G_TYPE_STRING, 
						      G_TYPE_STRING,
						      G_TYPE_INT,
						      G_TYPE_STRING,
						      G_TYPE_STRING,
						      G_TYPE_INT,
						      G_TYPE_INT,
						      BMP_TYPE_FILE_ITEM);

	g_object_set_data (G_OBJECT(self->private->playlist), "current", NULL);

	row_guid_history_map = g_hash_table_new ((GHashFunc)row_guid_hash_func, (GEqualFunc)row_guid_equal_func);
	row_guid_index_map = g_hash_table_new ((GHashFunc)row_guid_hash_func, (GEqualFunc)row_guid_equal_func);

#if 0
        g_object_connect (G_OBJECT(self->private->playlist),
			  "signal::row-deleted",
			  G_CALLBACK(update_row_guid_mappings),
			  self,
			  NULL);
#endif

        g_object_connect (G_OBJECT(self->private->playlist),
			  "signal::sort-column-changed",
			  G_CALLBACK(on_playlist_sort_column_changed),
			  self,
			  NULL);


        g_message("running: %s", G_OBJECT_TYPE_NAME(self));
}

static void
bmp_playlist_set_property (GObject      *object,
                        guint         property_id,
                        const GValue *value,
                        GParamSpec   *pspec)
{
        switch (property_id) {

        default:
                /* We don't have any other property... */
                g_assert (FALSE);
                break;
        }
}

static void
bmp_playlist_get_property (GObject      *object,
                        guint         property_id,
                        GValue       *value,
                        GParamSpec   *pspec)
{
        switch (property_id) {

        default:
                /* We don't have any other property... */
                g_assert (FALSE);
                break;
        }
}

static GObject *
bmp_playlist_constructor (GType                  type,
                       guint                  n_construct_properties,
                       GObjectConstructParam *construct_properties)
{
        GObject *obj;

        {
                /* Invoke parent constructor. */
                BmpPlaylistClass *klass;
                GObjectClass *parent_class;
                klass = BMP_PLAYLIST_CLASS (g_type_class_peek (BMP_TYPE_PLAYLIST));
                parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
                obj = parent_class->constructor (type,
                                                 n_construct_properties,
                                                 construct_properties);
        }

        /* do stuff. */

        return obj;
}

static void
bmp_playlist_dispose (GObject *obj)
{
    BmpPlaylist *self = (BmpPlaylist*) obj;

    if (self->private->dispose_has_run) return;

    self->private->dispose_has_run = TRUE;
    save_playlist (self);

    g_message("stopped: %s", G_OBJECT_TYPE_NAME(obj));
}

static void
bmp_playlist_finalize (GObject *object)
{
    BmpPlaylist *self = (BmpPlaylist*)object;
    g_free (self->private);
}

BmpPlaylist*
bmp_playlist_new (void)
{
  BmpPlaylist *self;

  self = g_object_new (bmp_playlist_get_type (), NULL);

  return self;
}

static void
bmp_playlist_class_init (BmpPlaylistClass *g_class)

{
    GObjectClass *gobject_class = G_OBJECT_CLASS (g_class);

    gobject_class->set_property = bmp_playlist_set_property;
    gobject_class->get_property = bmp_playlist_get_property;
    gobject_class->dispose	= bmp_playlist_dispose;
    gobject_class->finalize	= bmp_playlist_finalize;
    gobject_class->constructor	= bmp_playlist_constructor;

    signals[BMP_PLAYLIST_SIGNAL_TRACKLIST_LIST_SORTED] =
        g_signal_new ("tracklist-list-sorted",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      0,
                      G_STRUCT_OFFSET (BmpPlaylistClass, tracklist_list_sorted),
                      NULL, NULL, g_cclosure_marshal_VOID__INT, G_TYPE_NONE,
                      1, G_TYPE_INT); 

    signals[BMP_PLAYLIST_SIGNAL_TRACKLIST_ITEMS_REMOVED] =
        g_signal_new ("tracklist-items-removed",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      0,
                      G_STRUCT_OFFSET (BmpPlaylistClass, tracklist_items_removed),
                      NULL, NULL, bmpx_VOID__POINTER_INT, G_TYPE_NONE,
                      2, G_TYPE_POINTER, G_TYPE_INT);

    signals[BMP_PLAYLIST_SIGNAL_TRACKLIST_ITEMS_ADDED] =
        g_signal_new ("tracklist-items-added",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      0,
                      G_STRUCT_OFFSET (BmpPlaylistClass, tracklist_items_added),
                      NULL, NULL, bmpx_VOID__POINTER_INT, G_TYPE_NONE,
                      2, G_TYPE_POINTER, G_TYPE_INT);
}

GtkListStore*
bmp_playlist_get_playlist (BmpPlaylist *self)
{
	return self->private->playlist;
}

gint
bmp_playlist_get_row_by_guid (BmpPlaylist *self, RowGUID *guid)
{
    BmpFileItem *item;
    GtkTreeIter  iter;
    gboolean	 found = FALSE;
    gboolean	 run = TRUE;
    gint	 n = 0;

    g_return_val_if_fail (BMP_IS_PLAYLIST (self), -1);

#if 0
    if (g_hash_table_lookup_extended (row_guid_index_map,
				      guid,
				      &original_key,
				      &original_value)) 	      
      {
	  return GPOINTER_TO_INT(original_value);
      }
    else
      {
	  return -1;
      }
#endif

    run = gtk_tree_model_get_iter_first (GTK_TREE_MODEL(self->private->playlist), &iter);
    
    while (run) 
      {
	RowGUID *guid_cmp;

	gtk_tree_model_get (GTK_TREE_MODEL(self->private->playlist), &iter, TUPLE_COLUMN, &item, -1);
	guid_cmp = item->guid;

	if (row_guid_equal_func (guid, guid_cmp))
	 {
	    found = TRUE;
	    break;
	 }
	n++;
	run = gtk_tree_model_iter_next (GTK_TREE_MODEL(self->private->playlist), &iter);
      }

    return found?n:-1;
}

BmpFileItem*
bmp_playlist_get_row_by_index (BmpPlaylist	  *self,
			       gint		   index)
{
	BmpFileItem	*item;
	GtkTreeIter	 iter;

	if (!gtk_tree_model_iter_nth_child (GTK_TREE_MODEL(self->private->playlist), &iter, NULL, index))
	    return NULL;

	gtk_tree_model_get (GTK_TREE_MODEL(self->private->playlist), &iter, TUPLE_COLUMN, &item, -1); 

	return item;
}


static gint
bmp_playlist_insert_tracklist_item_position_R (BmpPlaylist	  *self,
					       const gchar	  *uri,
					       gint		   index)
{
    BmpFileItem	*item;
    VFSHandle	*handle;

    if (bmp_play_is_audio_file(bmp_play_engine, uri))
      {
	MetadataRealm	  realm;	
	GtkTreeIter	  iter;

	if (g_uri_test (uri, BMP_URI_TEST_PROTOCOL, PROTOCOL_FILE))
	  {
	      gchar *filename;

	      filename = g_filename_from_uri (uri, NULL, NULL);
	      if (!filename || !g_file_test (filename, G_FILE_TEST_EXISTS)) return index;
	      g_free (filename);
	  }

	if (xcs_key_fetch_boolean (xcs, "bmp", "library-add-automatically"))  
	  {
	    realm  = g_uri_test (uri, BMP_URI_TEST_PROTOCOL, PROTOCOL_HTTP) ? REALM_TRANSIENT : REALM_PERMANENT;
	  }
	else
	  {
	    realm = REALM_TRANSIENT;
	  }

	item = bmp_file_item_new (uri, realm);

	gtk_list_store_insert (GTK_LIST_STORE(self->private->playlist), &iter, index); 
	gtk_list_store_set (GTK_LIST_STORE(self->private->playlist), &iter, TUPLE_COLUMN, item, -1);
	g_hash_table_replace (row_guid_index_map, item->guid, GINT_TO_POINTER(index++));
	g_boxed_free (BMP_TYPE_FILE_ITEM, item);
      }
    else
      {
	GList *items = NULL;
	gboolean has_http = FALSE;

        handle = vfs_handle_new ();
	handle->uri = g_strdup (uri); 

	if (!vfs_handle_read (handle, VFS_NEED_CONTAINER_AUTO))
	    {
		vfs_handle_free (handle);
		return index;
	    }
	
	has_http = g_object_get_boolean (bmp_play_engine, "has_http");
	for ( ; handle->uri_list; handle->uri_list = handle->uri_list->next)
	    {
		MetadataRealm	  realm;	
		gchar		  *uri;

		uri = (gchar *)handle->uri_list->data;
		if (g_uri_test (uri, BMP_URI_TEST_PROTOCOL, PROTOCOL_HTTP) && !has_http) continue;

		if (g_uri_test (uri, BMP_URI_TEST_PROTOCOL, PROTOCOL_FILE))
		  {
		    gchar *filename;

		    filename = g_filename_from_uri (uri, NULL, NULL);
		    if (!filename || !g_file_test (filename, G_FILE_TEST_EXISTS)) continue; 
		    g_free (filename);
		  }

		if (xcs_key_fetch_boolean (xcs, "bmp", "library-add-automatically"))  
		  {
		    realm = REALM_PERMANENT;
		    if (g_uri_test (uri, BMP_URI_TEST_PROTOCOL, PROTOCOL_HTTP))
		      {
			realm = REALM_TRANSIENT;
		      } 
		    else if (g_uri_test (uri, BMP_URI_TEST_PROTOCOL, PROTOCOL_CDDA))
		      { 
			BMPURI *bmp_uri;

			bmp_uri = bmp_uri_new (uri);

			if (!bmp_uri->hostname || (bmp_uri->hostname && !strlen(bmp_uri->hostname)))
			  {
			    realm = REALM_TRANSIENT;
			  }

			bmp_uri_delete (bmp_uri);
		      }
		  }
		else
		  {
		    realm = REALM_TRANSIENT;
		  }

		items = g_list_append (items, bmp_file_item_new (uri, realm));
	    }

	 for ( ; items; items = items->next)
	    {
		GtkTreeIter iter;

		item = (BmpFileItem*)items->data;

		gtk_list_store_insert (GTK_LIST_STORE(self->private->playlist), &iter, index); 
		gtk_list_store_set (GTK_LIST_STORE(self->private->playlist), &iter, TUPLE_COLUMN, item, -1);
		g_hash_table_replace (row_guid_index_map, item->guid, GINT_TO_POINTER(index++));
		g_boxed_free (BMP_TYPE_FILE_ITEM, item);
	    }

	vfs_handle_free (handle);
      }

    return index;
}

gint
bmp_playlist_insert_tracklist_items_position (BmpPlaylist *self, const gchar **uri_list, gint _idx)
{
    gint n = 0, n_added_rows = 0, _idx_original = _idx;
    GArray *added_rows;

    while (uri_list[n])
      {
        _idx = bmp_playlist_insert_tracklist_item_position_R (self, uri_list[n], _idx);
	n++;
      }

    added_rows = g_array_new (FALSE, FALSE, sizeof(gint));

    for (n = 0; n < (_idx - _idx_original); n++)
      {
	gint val = n + _idx_original;
	g_array_append_val (added_rows, val);
      }

    n_added_rows = (_idx - _idx_original);

    g_signal_emit (G_OBJECT (self),
                   signals[BMP_PLAYLIST_SIGNAL_TRACKLIST_ITEMS_ADDED],
                   0, added_rows, n_added_rows);

    return _idx;
}

gint
bmp_playlist_insert_tracklist_items (BmpPlaylist *self, const gchar **uri_list)
{
   return bmp_playlist_insert_tracklist_items_position (self, uri_list, gtk_tree_model_iter_n_children (GTK_TREE_MODEL(self->private->playlist), NULL)); 
}

void
bmp_playlist_tracklist_remove_rows_all (BmpPlaylist *self)
{
    GArray		    *removed_rows;
    gint		     n_removed_rows = -1;

    gtk_tree_row_reference_free (g_object_get_data (G_OBJECT(self->private->playlist), "current")); 
    g_object_set_data (G_OBJECT(self->private->playlist), "current", NULL);
    gtk_list_store_clear (self->private->playlist);

    removed_rows = g_array_new (FALSE,FALSE,sizeof(gint));

    g_signal_emit (G_OBJECT (self),
                   signals[BMP_PLAYLIST_SIGNAL_TRACKLIST_ITEMS_REMOVED],
                   0, removed_rows, n_removed_rows);
}

void
bmp_playlist_tracklist_keep_predicated (BmpPlaylist *self, 
					HTuple *tuple_cmp,
					BmpMetadatumId metadatum_id)
{
#if 0
	GList	    *new_items = NULL, *iter_list = NULL;
	BmpListType *list;
	BmpFileItem *item;
	gint	     n = 0, 
		     new_current = 0,
		     new_count = 0, 
		     n_removed_rows = 0;
	GArray	    *removed_rows;

	list = self->private->playlist; 

	removed_rows = g_array_new (FALSE, FALSE, sizeof(gint));

	for (n = 0; n < g_list_length(list->items); n++)
	  {
		HTuple *tuple;
		item = LIST_GET_NTH_ITEM (list, n);

		tuple = item->tuple; 

		if (h_tuple_equal_elem (tuple, tuple_cmp, bmp_metadata_get_id_static (metadatum_id)))
		  {
			new_items = g_list_append (new_items, item);
			if (n == list->current)	new_current = new_count;
			new_count++;
		  }
		else
                  {
		    g_array_append_val (removed_rows, n);
		    n_removed_rows++;
		  }
	   }

	list->items = new_items;
	list->current = new_current; 

	iter_list = list->items;
	for ( ; iter_list; iter_list = iter_list->next)
	  {
	    list->row_set_selected_func ((gpointer)iter_list->data, FALSE);
	  }
	list->multiple_selected = FALSE;
	list->last_selected = -1;

        g_signal_emit (G_OBJECT (self),
                   signals[BMP_PLAYLIST_SIGNAL_TRACKLIST_ITEMS_REMOVED],
                   0, removed_rows, n_removed_rows);

	/* FIXME: Free old items */
#endif
}


void
bmp_playlist_tracklist_remove_predicated (BmpPlaylist *self, 
					  HTuple *tuple_cmp,
					  BmpMetadatumId metadatum_id)
{
#if 0
	GList	    *new_items = NULL, *iter_list = NULL;
	BmpListType *list;
	BmpFileItem *item;
	gint	     n = 0, 
		     new_current = 0,
		     new_count = 0, 
		     n_removed_rows = 0;
	GArray	    *removed_rows;

	list = self->private->playlist; 

	removed_rows = g_array_new (FALSE, FALSE, sizeof(gint));

	for (n = 0; n < g_list_length(list->items); n++)
	  {
		HTuple *tuple;
		item = LIST_GET_NTH_ITEM (list, n);

		tuple = item->tuple; 

		if (!h_tuple_equal_elem (tuple, tuple_cmp, bmp_metadata_get_id_static (metadatum_id)))
		  {
			new_items = g_list_append (new_items, item);
			if (n == list->current)	new_current = new_count;
			new_count++;
		  }
		else
                  {
		    g_array_append_val (removed_rows, n);
		    n_removed_rows++;
		  }
	   }

	list->items = new_items;
	list->current = new_current; 

	iter_list = list->items;
	for ( ; iter_list; iter_list = iter_list->next)
	  {
	    list->row_set_selected_func ((gpointer)iter_list->data, FALSE);
	  }
	list->multiple_selected = FALSE;
	list->last_selected = -1;

        g_signal_emit (G_OBJECT (self),
                   signals[BMP_PLAYLIST_SIGNAL_TRACKLIST_ITEMS_REMOVED],
                   0, removed_rows, n_removed_rows);

	/* FIXME: Free old items */
#endif
}


void
bmp_playlist_tracklist_remove_selected (BmpPlaylist *self, GList *paths)
{
	GArray *removed_rows;
	gint    n_removed_rows;

	removed_rows = g_array_new (FALSE, FALSE, sizeof(gint));
	n_removed_rows = g_list_length (paths);
	  
	for ( ; paths ; paths = paths->next) 
	  {
	      GtkTreeIter iter;
	      gint	  row = gtk_tree_path_get_indices (paths->data)[0];

	      if (gtk_tree_model_get_iter (GTK_TREE_MODEL(self->private->playlist), &iter, paths->data))
		{
		  gtk_list_store_remove (GTK_LIST_STORE(self->private->playlist), &iter);
		  g_array_append_val (removed_rows, row); 
		}	
	  }

        g_signal_emit (G_OBJECT (self),
                   signals[BMP_PLAYLIST_SIGNAL_TRACKLIST_ITEMS_REMOVED],
                   0, removed_rows, n_removed_rows);
}

void
bmp_playlist_tracklist_remove_rows (BmpPlaylist *self, GArray *rows, gint n_rows)
{
#if 0
	GList *new_items;
	BmpListType *list;
	BmpFileItem *item;
	GHashTable *mapping;
	gint n, new_count, new_current;

	list = self->private->playlist;

	new_items =
	    NULL;
	new_count =
	    0;
	new_current =
	    -1;
	mapping = g_hash_table_new (g_direct_hash, g_direct_equal);

	for (n = 0; n < n_rows; n++)
	  {
	    gint val;
	    val =
		&g_array_index (rows, gint, n);
	    g_hash_table_insert (mapping, GINT_TO_POINTER(val),
				 GINT_TO_POINTER(0x01));
	  }

	for (n = 0; n < g_list_length(list->items); n++)
	  {
		item = LIST_GET_NTH_ITEM (list, n);

		if (! g_hash_table_lookup (mapping, GINT_TO_POINTER(n)))
		  {
			new_items = g_list_append (new_items, item);
			if (n == list->current)
				new_current = new_count;
			new_count++;
		  }
	  }

	list->items = new_items;
	list->current = new_current;
	list->last_selected = 0;

        g_signal_emit (G_OBJECT (self),
                   signals[BMP_PLAYLIST_SIGNAL_TRACKLIST_ITEMS_REMOVED],
                   0, rows, n_rows);

	/* FIXME: Free old items */
#endif
}

void
bmp_playlist_tracklist_remove_unselected (BmpPlaylist *self)
{
#if 0
	GList *new_items;
	BmpListType *list;
	BmpFileItem *item;
	gint n, new_count, new_current, n_removed_rows = 0;
	GArray *removed_rows;

	list = self->private->playlist;

	new_items =
	    NULL;
	new_count =
	    0;
	new_current =
	    -1;
	removed_rows =
	    g_array_new (FALSE, FALSE, sizeof(gint));

	for (n = 0; n < g_list_length(list->items); n++)
	  {
		item = LIST_GET_NTH_ITEM (list, n);

		if (item->selected)
		  {
			new_items = g_list_append (new_items, item);
			if (n == list->current)
				new_current = new_count;
			new_count++;
		  }
		else
	          {
			g_array_append_val (removed_rows, n);
			n_removed_rows++;
		  }
	  }

	list->items = new_items;
	list->current = new_current;
	list->last_selected = 0;

        g_signal_emit (G_OBJECT (self),
                   signals[BMP_PLAYLIST_SIGNAL_TRACKLIST_ITEMS_REMOVED],
                   0, removed_rows, n_removed_rows);

	/* FIXME: Free old items */
#endif
}

static void
load_playlist (BmpPlaylist *self)
{
	gchar	 **uri_strv;
	gint	   new_count = 0;

	uri_strv = g_new0 (gchar*,2);
	uri_strv[0] = g_strconcat ("file://", BMP_PATH_USER_DIR, "/tracklist.xspf", NULL);
	uri_strv[1] = NULL;

	bmp_system_control_add_uri_list (bmp_system_control, 
					(const gchar**)uri_strv, 
					-1, 
					FALSE,
					FALSE,
					0,
					&new_count, 
					NULL);
	g_strfreev (uri_strv);
}

static void
save_playlist (BmpPlaylist *self)
{
	VFSHandle   *handle;
	GList	    *uri_list = NULL;

	GtkTreeIter  iter;
	gboolean     proceed;

	proceed = gtk_tree_model_get_iter_first (GTK_TREE_MODEL(self->private->playlist), &iter);

	while (proceed)
	  {
	    BmpFileItem *item;

	    BMPURI	*uri;
	    gchar	*filename,
			*uri_esc;

	    gtk_tree_model_get (GTK_TREE_MODEL(self->private->playlist), &iter, TUPLE_COLUMN, &item, -1);	 

	    uri = bmp_uri_new (item->uri); 
	    bmp_uri_escape (uri);
	    uri_esc = bmp_uri_get_string (uri);
  
	    filename = g_filename_from_uri (item->uri, NULL, NULL);
	    if (!BMP_URI_IS_LOCAL(item->uri) || (filename && g_file_test (filename, G_FILE_TEST_EXISTS))) 
	      {
		  uri_list = g_list_append (uri_list, g_strdup(item->uri)); 
	      }

	    g_free (uri_esc); 
	    g_free (filename);

	    g_boxed_free (BMP_TYPE_FILE_ITEM, item);

	    proceed = gtk_tree_model_iter_next (GTK_TREE_MODEL(self->private->playlist), &iter);
	}

	handle = vfs_handle_new ();
	handle->filename = g_build_filename (BMP_PATH_USER_DIR, "tracklist.xspf", NULL);
	handle->uri_list = uri_list;
  
	if (!vfs_handle_write (handle, "xspf"))
	  {
		g_message("%s: Error saving playlist %s", G_STRLOC, handle->filename); 
	  }

	vfs_handle_free (handle);
}

gint
bmp_playlist_get_length (BmpPlaylist *self)
{
	return gtk_tree_model_iter_n_children (GTK_TREE_MODEL(self->private->playlist), NULL);
}

gchar*
bmp_playlist_get_uri_nth (BmpPlaylist	    *self,
		          gint		     index)
{
    BmpFileItem		    *item;
    GtkTreeIter		     iter;

    gtk_tree_model_iter_nth_child (GTK_TREE_MODEL(self->private->playlist), &iter, NULL, index);

    gtk_tree_model_get (GTK_TREE_MODEL(self->private->playlist), &iter, TUPLE_COLUMN, &item, -1); 
    g_return_val_if_fail (item != NULL, NULL);

    return g_strdup (item->uri);
}


gchar*
bmp_playlist_get_current_uri (BmpPlaylist *self)
{
    BmpFileItem		    *item;
    GtkTreeRowReference	    *reference;
    GtkTreePath		    *path;
    GtkTreeIter		     iter;

    reference = g_object_get_data (G_OBJECT(self->private->playlist), "current");
    if (!reference) return NULL;

    path = gtk_tree_row_reference_get_path (reference);
    gtk_tree_model_get_iter (GTK_TREE_MODEL(self->private->playlist), &iter, path);
    gtk_tree_path_free (path);

    gtk_tree_model_get (GTK_TREE_MODEL(self->private->playlist), &iter, TUPLE_COLUMN, &item, -1); 
    g_return_val_if_fail (item != NULL, NULL);

    return g_strdup (item->uri);
}

gchar*
bmp_playlist_get_title_nth (BmpPlaylist		*self,
			    gint		 index)
{
    BmpFileItem		    *item;
    GtkTreeIter		     iter;
    gchar		    *tstring;

    gtk_tree_model_iter_nth_child (GTK_TREE_MODEL(self->private->playlist), &iter, NULL, index);
    gtk_tree_model_get (GTK_TREE_MODEL(self->private->playlist), &iter, TUPLE_COLUMN, &item, -1); 

    g_return_val_if_fail (item != NULL, NULL);

    tstring = bmp_metadata_cache_get_titlestring (bmp_metadata_cache, item); 
    g_boxed_free (BMP_TYPE_FILE_ITEM, item);

    return tstring;
}

gchar*
bmp_playlist_get_current_title (BmpPlaylist *self)
{
    BmpFileItem		    *item;
    GtkTreeRowReference	    *reference;
    GtkTreePath		    *path;
    GtkTreeIter		     iter;
    gchar		    *tstring;

    reference = g_object_get_data (G_OBJECT(self->private->playlist), "current");
    if (!reference) return NULL; 

    path = gtk_tree_row_reference_get_path (reference);
    if (!path) return NULL; 

    gtk_tree_model_get_iter (GTK_TREE_MODEL(self->private->playlist), &iter, path);
    gtk_tree_path_free (path);

    gtk_tree_model_get (GTK_TREE_MODEL(self->private->playlist), &iter, TUPLE_COLUMN, &item, -1); 
    g_return_val_if_fail (item != NULL, NULL);

    tstring = bmp_metadata_cache_get_titlestring (bmp_metadata_cache, item); 
    g_boxed_free (BMP_TYPE_FILE_ITEM, item);

    return tstring;

}

void
bmp_playlist_load_playlist (BmpPlaylist *self)
{
    g_return_if_fail (self != NULL);
    g_return_if_fail (BMP_IS_PLAYLIST(self));

    load_playlist (self);
}
