/*  BMPx - The Dumb Music Player
 *  Copyright (C) 2005 BMPx development team.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * The BMPx project hereby grant permission for non-gpl compatible GStreamer
 * plugins to be used and distributed together with GStreamer and BMPx. This
 * permission are above and beyond the permissions granted by the GPL license
 * BMPx is covered by.
 */


#include <config.h>

#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/file.h>

#include <glib.h>

#include "main.h"
#include <bmp/rtfs.h>
#include <bmp/metadata.h>

#include <libhrel/types.h>
#include <libhrel/tuple.h>
#include <libhrel/relation.h>
#include <libhrel/relation_base.h>

#define GUINTPTR_TO_POINTER(d) ((gpointer)(d))
#define GPOINTER_TO_UINTPTR(d) ((uintptr_t)(d))

typedef enum {

    _RTFS_IE_RANGE

} RTFSInternalError;

#ifdef ARCH_IA32
static inline gint
ffz (unsigned long word)
{
      gint _rv;

      __asm__("bsfl %1,%0"
                 :"=r" (word)
                 :"r" (~word));

      _rv = word;
      return _rv; 
}
#else
static gint
ffz (guint8 value)
{
    gint n = 0;
    while (n < 8) 
      {
	  guint8 _match = 0;   
	  
	  _match = value & (1 << n);
     
	  if (_match == 0) 
	    break;

	  n++;
      }
    return n;
}
#endif

static gint 
_map_block_alloc (RTFS      *rtfs,
		  uintptr_t   b_idx)
{
    guint8   byte;
    guint8   bit;

    if (b_idx > rtfs->size)
      return _RTFS_IE_RANGE;

    byte = b_idx >> 3;
    bit  = b_idx  % 8; 

#ifdef RTFS_DEBUG
    value = rtfs->bitmap->data[byte];
    g_message ("%s: b_idx: %u, byte: %hd, bit: %hd, value: %u (PRE)", G_STRFUNC, b_idx, byte, bit, value);  

    value |= (1 << bit);
    g_message ("%s: b_idx: %u, byte: %hd, bit: %hd, value: %u (POST)", G_STRFUNC, b_idx, byte, bit, value);  

    rtfs->bitmap->data[byte] = value;
#else
    rtfs->bitmap->data[byte] |= (1 << bit); 
#endif

    return 0;
}

static gint 
_map_block_free (RTFS	   *rtfs,
		 uintptr_t   b_idx)
{
    uintptr_t  byte;
    uintptr_t  bit;

    if (b_idx > rtfs->size)
      return _RTFS_IE_RANGE;

    byte = b_idx >> 3;
    bit  = b_idx  % 8; 

    rtfs->bitmap->data[byte] &= ~ (1 << bit); 

    return 0;
}

static uintptr_t
_get_uintptr_t (unsigned char **data)
{
    uintptr_t	*v_ptr;
    uintptr_t	 v;

    v_ptr = (uintptr_t*)(*data);
    v = *v_ptr;

    *data += sizeof(uintptr_t);

    return v;
}

static char* 
_get_string (unsigned char **data)
{
    char *string;

    string = (char *)(*data);

    *data += strlen(string)+1; 

    return string;
}

static gint
rtfs_grow (RTFS *rtfs)

{
    guint8   *bitfield = g_new0 (guint8,1); 
    gpointer  zero;

    g_mutex_lock(rtfs->lock_disk);

    //Grow bitmap
    g_byte_array_append (rtfs->bitmap, bitfield, 1);

    //Grow disk
    fseek (rtfs->disk, 0L, SEEK_END);
    zero = g_malloc0 (RTFS_BLOCKSIZE * 8); 
    fwrite (zero, RTFS_BLOCKSIZE * 8, 1, rtfs->disk);

    rtfs->size += 8; 

    g_mutex_unlock(rtfs->lock_disk);

    return 0;
}

/* This function _guarantees_ that it can allocate n_blocks as it will grow the filesystem if there is no space left */
/* What's important to note though is that the caller _has_ to actually use the blocks or they will remain allocated, but
   unused. */
static uintptr_t*
allocate_blocks (RTFS	  *rtfs,
		 gint	   n_blocks)
{
   uintptr_t *blocks = g_new0 (uintptr_t, n_blocks); 
   gint	     idx_bitmap = 0;
   gint	     idx_blocks = 0;
   gint	     n = 0;

   g_mutex_lock(rtfs->lock_bitmap);

   find_blocks:
   while ((n != n_blocks) && (idx_bitmap < rtfs->bitmap->len))
    {
#ifdef RTFS_DEBUG
	g_message ("%s: bitmap chunk %d is %d", G_STRFUNC, idx_bitmap, rtfs->bitmap->data[idx_bitmap]);
#endif
	if (rtfs->bitmap->data[idx_bitmap] != 255)
	  {
	      gint bit = ffz (rtfs->bitmap->data[idx_bitmap]);
	      blocks[idx_blocks] = (idx_bitmap * 8) + bit; 
#ifdef RTFS_DEBUG
	      g_message ("%s: allocating block %u", G_STRFUNC, blocks[idx_blocks]); 
#endif
	      _map_block_alloc (rtfs, blocks[idx_blocks]);
	      idx_blocks++;
	      n++;
	      continue;
	  }
	idx_bitmap++;
    } 

    //Do we have enough blocks or did the FS run out of space?
    if ((n != n_blocks) || (n == 0))
      {
	  rtfs_grow (rtfs);
	  goto find_blocks;
      }

    g_mutex_unlock(rtfs->lock_bitmap);

    return blocks;
}

static void
rtfs_read_block (RTFS	    *rtfs, 
		 gpointer    memory, 
		 uintptr_t    index)
{
    gpointer buffer;

    buffer = g_hash_table_lookup (rtfs->buffer, GUINTPTR_TO_POINTER(index)); 

    if (buffer)
      {
	  memcpy (memory, buffer, RTFS_BLOCKSIZE);
	  return;
      }

    g_mutex_lock(rtfs->lock_disk);

    if (fseek (rtfs->disk, (long)(index * RTFS_BLOCKSIZE), SEEK_SET)) 
      {
	  int errsv = errno;
	  char *errstr = strerror (errsv);
	  g_message ("Error: %s", errstr);
      }

    fread (memory, RTFS_BLOCKSIZE, 1, rtfs->disk);

    buffer = g_malloc0 (RTFS_BLOCKSIZE);
    memcpy (buffer, memory, RTFS_BLOCKSIZE);
    g_hash_table_insert (rtfs->buffer, GUINTPTR_TO_POINTER(index), buffer);

    g_mutex_unlock(rtfs->lock_disk);
}

static void
rtfs_buffer_block (RTFS	       *rtfs, 
		   uintptr_t     index)
{
    gpointer memory;

    g_mutex_lock(rtfs->lock_disk);

    if (fseek (rtfs->disk, (long)(index * RTFS_BLOCKSIZE), SEEK_SET)) 
      {
	  int errsv = errno;
	  char *errstr = strerror (errsv);
	  g_message ("Error: %s", errstr);
      }

    memory = g_malloc0 (RTFS_BLOCKSIZE);
    fread (memory, RTFS_BLOCKSIZE, 1, rtfs->disk);
    g_hash_table_replace (rtfs->buffer, GUINTPTR_TO_POINTER(index), memory);

    g_mutex_unlock(rtfs->lock_disk);
}


static void
rtfs_write_block (RTFS	    *rtfs, 
		  gpointer   memory, 
		  uintptr_t   index)
{
    gpointer buffer;

    buffer = g_malloc0 (RTFS_BLOCKSIZE);
    memcpy (buffer, memory, RTFS_BLOCKSIZE);

    g_hash_table_replace (rtfs->buffer, GUINTPTR_TO_POINTER(index), buffer);
    g_hash_table_replace (rtfs->dirty_blocks, GUINTPTR_TO_POINTER(index), GINT_TO_POINTER(1));
}

static void
relation_insert_dirent (RTFS *rtfs, RTFS_DirEnt *dirent)
{
  HTuple  *record;
  char	  *memory_linear;
  gint	   n;

  memory_linear = g_malloc0 (RTFS_BLOCKSIZE * dirent->n_blocks);
  for (n = 0; n < dirent->n_blocks; n++)  
    {
	gpointer mem = memory_linear+(n*RTFS_BLOCKSIZE);
	rtfs_read_block (rtfs, mem, dirent->chain[n]);
    }

  record = h_tuple_new (); 

  for (n = 0; n < BMP_DATUM_NONE; n++)
    {
      gint *idptr;
      gint  id;
      
      idptr = (gint*)(memory_linear);
      id = *idptr;
      memory_linear += sizeof(gint);

      if (bmp_metadata_get_type (id) == G_TYPE_INT)
	{
	  gint *vptr = (gint*)(memory_linear);
	  gint  v = *vptr;
	 
	  h_tuple_insert (record,
			  bmp_metadata_get_id_static (id),
			  G_TYPE_INT,
			  GINT_TO_POINTER(v)); 

	  memory_linear += sizeof(gint); 
	}
      else
      if (bmp_metadata_get_type (id) == G_TYPE_LONG)
	{
	  glong *vptr = (glong*)(memory_linear);
	  glong  v = *vptr;

	  h_tuple_insert (record,
			  bmp_metadata_get_id_static (id),
			  G_TYPE_LONG,
			  (gpointer)(v));

	  memory_linear += sizeof(glong); 
	}
      else
      if (bmp_metadata_get_type (id) == G_TYPE_STRING)
	{
          char *value;
 
          value = (char*)(memory_linear);

	  if (!strlen(value))
	    h_tuple_insert (record,
			  bmp_metadata_get_id_static (id),
			  G_TYPE_STRING,
			  ""); 
	  else
	    h_tuple_insert (record,
			  bmp_metadata_get_id_static (id),
			  G_TYPE_STRING,
			  value); 

           /* string length + terminating 0 */
	  memory_linear += strlen(value) + 1;
	}
    }
      
   if (!h_relation_base_get_tuple (H_RELATION_BASE(rtfs->relation), record))
      {
        h_relation_base_insert_tuple (H_RELATION_BASE(rtfs->relation), record);
      }
}

static void
read_directory_record (RTFS *rtfs, unsigned char **data, gboolean preload)
{
    RTFS_DirEnt *dirent;
    int		 n;

    dirent = g_new0 (RTFS_DirEnt,1);
    dirent->n_blocks = _get_uintptr_t (data);
    dirent->chain = g_new0 (uintptr_t, dirent->n_blocks);

    for (n = 0; n < dirent->n_blocks; n++)
      {
	 dirent->chain[n] = _get_uintptr_t (data); 
#ifdef RTFS_DEBUG
	 g_message ("RTFS(%s):   record at %p: block %d at disk-block %u", rtfs->name, (gpointer)dirent, n, dirent->chain[n]);
#endif
      }

    dirent->name = g_strdup (_get_string (data));
#ifdef RTFS_DEBUG
    g_message ("RTFS(%s):   record at %p: name: %s", rtfs->name, (gpointer)dirent, dirent->name);
#endif

    g_mutex_lock(rtfs->lock_directory);
    g_hash_table_insert (rtfs->directory, g_strdup (dirent->name), dirent);
    g_mutex_unlock(rtfs->lock_directory);

#ifdef HAVE_GUI
    while (gtk_events_pending()) gtk_main_iteration ();
#endif

    if (preload) relation_insert_dirent (rtfs, dirent);
}

typedef struct _RTFS_BUFFER RTFS_BUFFER;
struct RTFS_BUFFER {
    size_t  len;
    char   *data;
};

static void 
rtfs_insert_tuple (HRelation   *relation,
		   HTuple      *tuple,
		   RTFS	       *rtfs) 
{
  HTupleElem		*elem;
  RTFS_DirEnt		*dirent;	
  uintptr_t	        *blocks;
  char			*memory_linear;
  struct RTFS_BUFFER	 b;
  gint			 n;
  gint			 n_blocks;

  b.len = 0;
  b.data = NULL;

  elem = h_tuple_get (tuple, bmp_metadata_get_id_static (BMP_METADATA_PKEY)); 
  if (g_hash_table_lookup (rtfs->directory, elem->value))
    {
	return;
    }

  for (n = 0; n < BMP_DATUM_NONE; n++)
    {
      elem = h_tuple_get (tuple, bmp_metadata_get_id_static (n)); 

      b.data = realloc (b.data, sizeof(gint)+b.len); 
      memcpy (b.data+b.len, &n, sizeof(gint)); 
      b.len += sizeof(gint); 

      if (elem->type == G_TYPE_INT)
	{
	  gint _v;

	  _v = GPOINTER_TO_INT(elem->value);

	  b.data = realloc (b.data, sizeof(gint)+b.len); 
	  memcpy (b.data+b.len, &_v, sizeof(gint)); 
	  b.len += sizeof(gint); 
	}
      else
      if (elem->type == G_TYPE_LONG)
	{
	  glong _v;

	  _v = (glong)(elem->value);

	  b.data = realloc (b.data, sizeof(glong)+b.len); 
	  memcpy (b.data+b.len, &_v, sizeof(glong)); 
	  b.len += sizeof(glong); 
	}
      else
      if (elem->type == G_TYPE_STRING)
	{
	  if (elem->value)
	    {
	      b.data = realloc (b.data, strlen(elem->value)+1+b.len); 
	      memcpy (b.data+b.len, elem->value, strlen(elem->value)+1);
	      b.len += (strlen(elem->value)+1);
	    }
	  else
	    {
	      gchar *zero = g_strdup (""); 

	      b.data = realloc (b.data, 2+b.len); 
	      memcpy (b.data+b.len, zero, 2);
	      b.len += 2; 
	      g_free (zero);
	    }
	}
    }

    n_blocks = b.len / RTFS_BLOCKSIZE; 
    if ((n_blocks * RTFS_BLOCKSIZE) < b.len) n_blocks++;

    blocks = allocate_blocks (rtfs, n_blocks);

    //Create a linear region and then write the blocks into the free locations
    memory_linear = g_malloc0 (n_blocks * RTFS_BLOCKSIZE); 
    memcpy (memory_linear, b.data, b.len); 
  
    dirent = g_new0 (RTFS_DirEnt,1);
    dirent->n_blocks = n_blocks;
    dirent->chain = blocks;

    elem = h_tuple_get (tuple, bmp_metadata_get_id_static (BMP_METADATA_PKEY)); 
    dirent->name = g_strdup (elem->value); 

    for (n = 0; n < n_blocks; n++)    
      {
	  gpointer memory = memory_linear+(n*RTFS_BLOCKSIZE);
#ifdef RTFS_DEBUG
	  g_message ("RTFS(%s):   allocated block #%u for record '%s'", rtfs->name, blocks[n], dirent->name);
#endif
	  rtfs_write_block (rtfs, memory, blocks[n]); 
      }

    g_free (memory_linear);
    g_free (b.data);

    g_mutex_lock(rtfs->lock_directory);
    g_hash_table_insert (rtfs->directory, g_strdup(dirent->name), dirent);
    g_mutex_unlock(rtfs->lock_directory);
}

static void 
rtfs_delete_tuple (HRelation  *relation,
		   HTuple     *tuple,
		   RTFS	      *rtfs)
{
  HTupleElem		*elem = NULL;
  RTFS_DirEnt		*dirent = NULL;	
  gint			 n;

  elem = h_tuple_get (tuple, bmp_metadata_get_id_static (BMP_METADATA_PKEY)); 
  dirent = g_hash_table_lookup (rtfs->directory, elem->value);

  if (!dirent) 
    {
      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "%s: Unable to locate directory entry for pkey '%s'", G_STRLOC, elem->value);
      return;
    }

  g_mutex_lock(rtfs->lock_bitmap);
  g_mutex_lock(rtfs->lock_directory);

  for (n = 0; n < dirent->n_blocks; n++)
    {
	_map_block_free (rtfs, dirent->chain[n]);
    }

  g_free (dirent->chain); 
  g_free (dirent->name);
  g_free (dirent);

  if (!g_hash_table_remove (rtfs->directory, elem->value))
    {
      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "%s: Unable to remove directory entry for pkey '%s'", G_STRLOC, elem->value);
    }

  g_mutex_unlock(rtfs->lock_directory);
  g_mutex_unlock(rtfs->lock_bitmap);
}

RTFS*
rtfs_open (HRelation	    *relation,
	   const gchar	    *basepath,
	   const gchar	    *name,
	   gboolean	     preload)
{
    GError	  *error = NULL;
    GMappedFile	  *mfile;
    RTFS	  *rtfs;
    gchar	  *path_aux;

    rtfs = g_new0 (RTFS, 1); 
  
    path_aux = g_strconcat (name, RTFS_BITMAP, NULL);
    rtfs->path[P_BITMAP] = g_build_filename (basepath, path_aux, NULL); 
    g_free (path_aux);

    path_aux = g_strconcat (name, RTFS_DISK, NULL);
    rtfs->path[P_DISK] = g_build_filename (basepath, path_aux, NULL); 
    g_free (path_aux);

    path_aux = g_strconcat (name, RTFS_DIRECTORY, NULL);
    rtfs->path[P_DIRECTORY] = g_build_filename (basepath, path_aux, NULL); 
    g_free (path_aux);

    //FIXME: _new_full()
    rtfs->directory = g_hash_table_new (g_str_hash, g_str_equal);

    //FIXME: _new_full()
    rtfs->buffer = g_hash_table_new_full (g_direct_hash, g_direct_equal, NULL, (GDestroyNotify)g_free);
    rtfs->dirty_blocks = g_hash_table_new (g_direct_hash, g_direct_equal);

    rtfs->relation = relation;

    rtfs->lock_disk = g_mutex_new ();
    rtfs->lock_bitmap = g_mutex_new ();
    rtfs->lock_directory = g_mutex_new ();

    rtfs->name = g_strdup (name);

    if ((stat (rtfs->path[P_BITMAP], &rtfs->stats[P_BITMAP]) == -1) && (errno == ENOENT))
      {
	gpointer* zero;
	gulong    isize = RTFS_INITIAL_SIZE / 8; 

	//Bitmap doesn't exist. We override everything and restart the RTFS from zero
	g_mutex_lock(rtfs->lock_bitmap);
	rtfs->bitmap = g_byte_array_sized_new (isize);
	g_byte_array_set_size (rtfs->bitmap, isize); 
	memset (rtfs->bitmap->data, 0x00, isize);
	rtfs->size = RTFS_INITIAL_SIZE;
	g_mutex_unlock(rtfs->lock_bitmap);

	//Initialize disk to init/default size
	g_mutex_lock(rtfs->lock_disk);
	rtfs->disk = fopen (rtfs->path[P_DISK], "w+");
	zero = g_malloc0 (RTFS_BLOCKSIZE * RTFS_INITIAL_SIZE);
	fwrite (zero, RTFS_BLOCKSIZE * RTFS_INITIAL_SIZE, 1, rtfs->disk);

	g_mutex_unlock(rtfs->lock_disk);

	goto rtfs_return;
      }
    else
      {
 	uintptr_t        i_rec_size;
	uintptr_t       *i_rec_size_ptr;

	unsigned char  *current_ptr;
	uintptr_t	n;

	//Read bitmap
#ifdef RTFS_DEBUG
	g_message ("RTFS(%s): reading bitmap...", name);
#endif

        mfile = g_mapped_file_new (rtfs->path[P_BITMAP], FALSE, &error); 
	if (error)    
	  { 
	    g_error_free (error);
	    g_free (rtfs->path[P_BITMAP]);
	    return NULL;
	  }
	g_mutex_lock(rtfs->lock_bitmap);
	rtfs->bitmap = g_byte_array_sized_new (rtfs->stats[P_BITMAP].st_size); 
	current_ptr = (unsigned char*)g_mapped_file_get_contents (mfile);
	memcpy (rtfs->bitmap->data, current_ptr, rtfs->stats[P_BITMAP].st_size); 
	g_byte_array_set_size (rtfs->bitmap, rtfs->stats[P_BITMAP].st_size);
	g_mapped_file_free (mfile);
	g_mutex_unlock(rtfs->lock_bitmap);

#ifdef RTFS_DEBUG
	g_message ("RTFS(%s): ...done", name);
#endif


	//Open disk
#ifdef RTFS_DEBUG
	g_message ("RTFS(%s): Open disk...", name);
#endif
	stat (rtfs->path[P_DISK], &rtfs->stats[P_DISK]);
	rtfs->size = rtfs->stats[P_DISK].st_size / RTFS_BLOCKSIZE;
	rtfs->disk = fopen (rtfs->path[P_DISK], "r+");

#ifdef RTFS_DEBUG
	g_message ("RTFS(%s): ...done", name);
#endif

	for (n = 0; n < rtfs->bitmap->len * 8; n++)
	    rtfs_buffer_block (rtfs, n);

	//Read directory 
#ifdef RTFS_DEBUG
	g_message ("RTFS(%s): Read directory...", name);
	g_message ("RTFS(%s):   mapping file '%s'...", name, rtfs->path[P_DIRECTORY]);
#endif

        mfile = g_mapped_file_new (rtfs->path[P_DIRECTORY], FALSE, &error); 
	if (error)    
	  { 
	    g_error_free (error);
	    g_free (rtfs->path[P_DIRECTORY]);
	    return NULL;
	  }

#ifdef RTFS_DEBUG
	g_message ("RTFS(%s):   mapped", name);
#endif

	current_ptr = (unsigned char*)g_mapped_file_get_contents (mfile);
	i_rec_size_ptr = (uintptr_t*)current_ptr;	
	i_rec_size = *i_rec_size_ptr;
	current_ptr += sizeof(uintptr_t);

        for (n = 0; n < i_rec_size; n++)
	  {
#ifdef RTFS_DEBUG
	    g_message ("RTFS(%s):   reading record #%d", name, n);
#endif
	    read_directory_record (rtfs, &current_ptr, preload);
	  }

	g_mapped_file_free (mfile);
      }
#ifdef RTFS_DEBUG
    g_message ("RTFS(%s): ...done", name);
#endif

    rtfs_return:

    rtfs->handler_id_insert = g_signal_connect (G_OBJECT(relation),
						"insert-tuple",
						G_CALLBACK(rtfs_insert_tuple),
						rtfs);

    rtfs->handler_id_delete = g_signal_connect (G_OBJECT(relation),
						"delete-tuple",
						G_CALLBACK(rtfs_delete_tuple),
						rtfs);
  
    return rtfs;
}

static void
write_directory_records (gpointer _k, gpointer _v, FILE *handle)
{
    RTFS_DirEnt *dirent = (RTFS_DirEnt*)_v;
    gint         n;

    fwrite (&(dirent->n_blocks), sizeof(uintptr_t), 1, handle); 
    for (n = 0; n < dirent->n_blocks; n++)
      {
	  fwrite (&(dirent->chain[n]), sizeof(uintptr_t), 1, handle); 
      }

    fwrite (dirent->name, sizeof(char), strlen(dirent->name)+1, handle);
}

static void
write_dirty_blocks (gpointer _k, gpointer _v, gpointer _data)
{
  RTFS	    *rtfs  = (RTFS *)_data;
  uintptr_t  index = GPOINTER_TO_UINTPTR(_k);
  gpointer   mem   = g_hash_table_lookup (rtfs->buffer, _k);

  if (fseek (rtfs->disk, (long)(index * RTFS_BLOCKSIZE), SEEK_SET)) 
    {
	  int errsv = errno;
	  char *errstr = strerror (errsv);
	  g_message ("Error: %s", errstr);
    }

  fwrite (mem, RTFS_BLOCKSIZE, 1, rtfs->disk);
  g_mutex_unlock(rtfs->lock_disk);
}

gint
rtfs_sync (RTFS *rtfs)
{
	GError	  *error = NULL;
	FILE	  *handle;
	uintptr_t   i_rec_size;

	//Write bitmap
	g_file_set_contents (rtfs->path[P_BITMAP],
			     (const gchar*)rtfs->bitmap->data,
			     rtfs->bitmap->len,
			     &error);

	//Write directory
	handle = fopen (rtfs->path[P_DIRECTORY], "w");

	i_rec_size = g_hash_table_size (rtfs->directory);
	fwrite (&i_rec_size, sizeof(uintptr_t), 1, handle); 

	g_mutex_lock(rtfs->lock_directory);
	g_hash_table_foreach (rtfs->directory, 
			      (GHFunc)write_directory_records, 
			      handle);
	g_mutex_unlock(rtfs->lock_directory);
	fclose (handle);

	g_mutex_lock(rtfs->lock_disk);
	g_hash_table_foreach (rtfs->dirty_blocks, 
			      (GHFunc)write_dirty_blocks, 
			      rtfs);

	g_hash_table_destroy (rtfs->buffer);
	g_hash_table_destroy (rtfs->dirty_blocks);

	rtfs->buffer = g_hash_table_new_full (g_direct_hash, g_direct_equal, NULL, (GDestroyNotify)g_free);
	rtfs->dirty_blocks = g_hash_table_new (g_direct_hash, g_direct_equal);

	g_mutex_unlock(rtfs->lock_disk);

	return 0;
}

gint
rtfs_close (RTFS *rtfs)
{
  rtfs_sync (rtfs);
  fclose (rtfs->disk);

  return 0;
  //Free resources
}

HTuple*
rtfs_read_entry (RTFS		*rtfs,
		 gpointer	 pkey)
{
    HTuple *tuple = NULL;

    tuple = h_relation_base_get_tuplev (H_RELATION_BASE(rtfs->relation), 
					bmp_metadata_get_id_static (BMP_METADATA_PKEY), 
				        pkey); 
    if (tuple)
      {
	return tuple;
      }
    else
      {
	RTFS_DirEnt *dirent;
	//Try to look it up in the directory
	
	dirent = g_hash_table_lookup (rtfs->directory, (gpointer)pkey);
	if (!dirent) return NULL;

	g_signal_handler_block (rtfs->relation, rtfs->handler_id_insert);
	relation_insert_dirent (rtfs, dirent);		
	g_signal_handler_unblock (rtfs->relation, rtfs->handler_id_insert);

	tuple = h_relation_base_get_tuplev (H_RELATION_BASE(rtfs->relation), 
					bmp_metadata_get_id_static (BMP_METADATA_PKEY), 
				        pkey);

	return tuple;
      }

  return NULL;
}

static void
insert_entry (gpointer _k, gpointer _v, RTFS *rtfs)
{
    relation_insert_dirent (rtfs, (RTFS_DirEnt*)_v);		
#ifdef HAVE_GUI
    while (gtk_events_pending()) gtk_main_iteration ();
#endif
}

void
rtfs_read_filesystem (RTFS *rtfs)
{
    g_hash_table_foreach (rtfs->directory, (GHFunc) insert_entry, rtfs);
}
