/*  BMPx - The Dumb Music Player
 *  Copyright (C) 2005 BMPx development team.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  $Id$
 *
 * The BMPx project hereby grant permission for non-gpl compatible GStreamer
 * plugins to be used and distributed together with GStreamer and BMPx. This
 * permission are above and beyond the permissions granted by the GPL license
 * BMPx is covered by.
 */
/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

#include <config.h>

#ifdef HAVE_DBUS
# include <dbus/dbus-glib.h>
#endif

#ifdef HAVE_GUI
#include <gtk/gtk.h>
#include <libchroma/chroma-list.h>
#include <libchroma/chroma-list-selection.h>
#include "ui.h"
#include "ui_util.h"
#endif

#include <glib/gi18n.h>
#include <string.h>

#include <bmp/dbus.h>
#include <bmp/playlist.h>
#include <bmp/plugin.h>
#include <bmp/plugin_interfaces.h>
#include <bmp/util.h>
#include <bmp/vfs.h>

#include <goa/libgoa.h>
#include <xcs/xcs.h>

#include "error.h"
#include "loader.h"
#include "logger.h"
#include "play.h"
#include "resource_manager.h"
#include "system_control.h"

#include "bmp-marshalers.h"

G_DEFINE_TYPE (BmpSystemControl, bmp_system_control, G_TYPE_OBJECT)

#include "main.h"

#define MPRIS_IDENTITY "BMPx"

static gboolean proceed = TRUE;

typedef enum {
    BMP_SYSTEM_CONTROL_SIGNAL_TRACK_CHANGE,
    BMP_SYSTEM_CONTROL_SIGNAL_SEEK_EVENT,
    BMP_SYSTEM_CONTROL_SIGNAL_SET_STREAM_POS,
    BMP_SYSTEM_CONTROL_SIGNAL_SET_TITLE,
    BMP_SYSTEM_CONTROL_SIGNAL_SET_PLAYSTATUS,
    BMP_SYSTEM_CONTROL_SIGNAL_SET_VOLUME,
    BMP_SYSTEM_CONTROL_SIGNAL_SET_SHUFFLE,
    BMP_SYSTEM_CONTROL_SIGNAL_SET_REPEAT,
    BMP_SYSTEM_CONTROL_SIGNAL_SET_EQ,
    BMP_SYSTEM_CONTROL_SIGNAL_LIST_DELETED,
    BMP_SYSTEM_CONTROL_SIGNAL_LIST_ADDED,
    BMP_SYSTEM_CONTROL_SIGNAL_LIST_RENAMED,
    BMP_SYSTEM_CONTROL_SIGNAL_TRACKLIST_LIST_SORTED,
    BMP_SYSTEM_CONTROL_SIGNAL_TRACKLIST_ITEMS_REMOVED,
    BMP_SYSTEM_CONTROL_SIGNAL_TRACKLIST_ITEMS_ADDED,
    BMP_SYSTEM_CONTROL_SIGNAL_TRACKLIST_ROWS_SWAPPED,
    BMP_SYSTEM_CONTROL_SIGNAL_SET_STOP_AFTER_CURRENT,
    BMP_SYSTEM_CONTROL_SIGNAL_STARTUP_COMPLETE,
    BMP_SYSTEM_CONTROL_SIGNAL_SHUTDOWN_COMPLETE,
    BMP_SYSTEM_CONTROL_SIGNAL_SHUTDOWN_REQUEST,
    BMP_SYSTEM_CONTROL_N_SIGNALS
} BmpSystemControlSignals;

static guint signals[BMP_SYSTEM_CONTROL_N_SIGNALS] = { 0 };

typedef enum {
    BMP_SYSTEM_CONTROL_PROPERTY_VOLUME = 1,
    BMP_SYSTEM_CONTROL_PROPERTY_BITRATE,
    BMP_SYSTEM_CONTROL_PROPERTY_SAMPLERATE,
    BMP_SYSTEM_CONTROL_PROPERTY_STREAM_POS,
    BMP_SYSTEM_CONTROL_PROPERTY_MOVING,
    BMP_SYSTEM_CONTROL_PROPERTY_REPEAT,
    BMP_SYSTEM_CONTROL_PROPERTY_SHUFFLE,
    BMP_SYSTEM_CONTROL_PROPERTY_CURRENT_TITLE,
    BMP_SYSTEM_CONTROL_PROPERTY_CURRENT_URI,
    BMP_SYSTEM_CONTROL_PROPERTY_CURRENT_INDEX,
    BMP_SYSTEM_CONTROL_PROPERTY_STOP_AFTER_CURRENT,
    BMP_SYSTEM_CONTROL_N_PROPERTIES
} BmpSystemControlProperties;

struct _BmpSystemControlPrivate {
    gboolean dispose_has_run;

    GList		*message_domains;
    gint		 msg_domain_id;

    GAsyncQueue		*message_queue;
    guint		 message_queue_source_id;

    BmpPlaybackHistory  *playback_history;
    GObject		*flow_plugin;

    GHashTable		*remote_queries;
    guint		 ticket_id;

    gint		 sc_msg_domain_id;

    /* Properties */
    BmpFileItem	*current_item;
    gchar	*current_title;
    gint	 current_index;

    gint	 bitrate;
    gint	 playlist;
    gint	 samplerate;
    gint	 stream_pos;
    gint	 volume;

    gboolean	 moving;
    gboolean	 repeat;
    gboolean	 shuffle;
    gboolean	 stop_after_current;

    /*< private >*/
    gboolean	 anonymous;
};

typedef struct {
    gint list;
    gint track;
} bmp_message_list_track_t;

typedef struct {
    BmpEqEnum eq;
    gint value;
} bmp_message_eq_t;

static gint eq_values[BMP_N_EQ_PARAMS];

static gchar *eq_configuration_keynames[] = {
    "eq-30",
    "eq-60",
    "eq-125",
    "eq-250",
    "eq-500",
    "eq-1000",
    "eq-2000",
    "eq-4000",
    "eq-8000",
    "eq-16000"
};

static gint n_eq_configuration_keynames =
G_N_ELEMENTS (eq_configuration_keynames);

#ifdef HAVE_DBUS
static gboolean _no_remote = FALSE;
#include "system_control_glue.h"
#endif

static gboolean
play_track	  (BmpSystemControl	    *self,
		   gint			     position);
 
static void
play_seek	  (BmpSystemControl	    *self,
		   gint			     position);

gboolean
bmp_system_control_identity (BmpSystemControl  *self,
			     gchar	      **name,
			     GError	   **error)
{
  *name = g_strdup(MPRIS_IDENTITY);

   return TRUE;
}

/* DBus/Remote query api */
gboolean
bmp_system_control_query_request_ticket (BmpSystemControl  *self,
					 guint		   *ticket,
					 GError		  **error)
{
    Query *query;

    self->private->ticket_id++;
    *ticket = self->private->ticket_id;

    query = bmp_metadata_cache_query_initialize (bmp_metadata_cache);
    g_hash_table_insert (self->private->remote_queries, GUINT_TO_POINTER (*ticket), query);
    return TRUE;
}

gboolean
bmp_system_control_query_append_atom    (BmpSystemControl  *self,
					 guint		    ticket,
					 gint		    metadatum_id,
					 GValue		   *value,
					 gboolean	    exact_match,
					 GError		  **error)
{
    Query *query;

    query = g_hash_table_lookup (self->private->remote_queries, GUINT_TO_POINTER(ticket));

    if (!query)
      {
	    *error = g_error_new (g_quark_from_static_string (BMP_DBUS_INTERFACE),
				  BMP_DBUS_ERROR_QUERY_INVALID_TICKET_ID, 
				  "Invalid Ticket ID");
	    return FALSE;
      }

    if (G_VALUE_TYPE(value) == G_TYPE_STRING)
      {
	  gboolean rv;

	  rv = bmp_metadata_cache_query_append_string  (bmp_metadata_cache,
							query,
							metadatum_id,
							exact_match,
							(gchar *)g_value_get_string (value));
      } 
    else
    if (G_VALUE_TYPE(value) == G_TYPE_INT)
      {
	  gboolean rv;

	  rv = bmp_metadata_cache_query_append_int     (bmp_metadata_cache,
							query,
							metadatum_id,
							exact_match,
							g_value_get_int (value));
      }
    else
      {
	    *error = g_error_new (g_quark_from_static_string (BMP_DBUS_INTERFACE),
				  BMP_DBUS_ERROR_QUERY_INVALID_ATOM_TYPE, 
				  "Invalid Atom Type");
	    return FALSE;
      }

    return TRUE;
}

static void
append_uri_list_glist (HTuple	  *tuple,
		       GList	 **list)
{
    HTupleElem	  *elem;

    elem = h_tuple_get (tuple, bmp_metadata_get_id_static (BMP_METADATA_PKEY));

    *list = g_list_append (*list, g_strdup(elem->value));
}

gboolean
bmp_system_control_query_finish_and_run (BmpSystemControl  *self,
					 guint		    ticket,
					 gchar		 ***uri_list,
					 GError		  **error)
{
    HRelation	  *result;
    Query	  *query;
    GList	  *uri_list_glist = NULL; 

    query = g_hash_table_lookup (self->private->remote_queries, GUINT_TO_POINTER(ticket));

    if (!query)
      {
	    *error = g_error_new (g_quark_from_static_string (BMP_DBUS_INTERFACE),
				  BMP_DBUS_ERROR_QUERY_INVALID_TICKET_ID, 
				  "Invalid Ticket ID");
	    return FALSE;
      }

    result = bmp_metadata_cache_query_run (bmp_metadata_cache, query); 
    h_relation_for_each_tuple (result, (HTupleFunc)append_uri_list_glist, &uri_list_glist); 

    *uri_list = glist_to_strv (uri_list_glist); 
   
    h_relation_unref (result);
    bmp_metadata_cache_query_free (bmp_metadata_cache, query);    
    g_hash_table_remove (self->private->remote_queries, GUINT_TO_POINTER(ticket));
 
    return TRUE;
}

gboolean
bmp_system_control_query_cancel         (BmpSystemControl  *self,
					 guint		    ticket,
					 GError		  **error)
{
    Query	  *query;

    query = g_hash_table_lookup (self->private->remote_queries, GUINT_TO_POINTER(ticket));

    if (!query)
      {
	    *error = g_error_new (g_quark_from_static_string (BMP_DBUS_INTERFACE),
				  BMP_DBUS_ERROR_QUERY_INVALID_TICKET_ID, 
				  "Invalid Ticket ID");
	    return FALSE;
      }

    bmp_metadata_cache_query_free (bmp_metadata_cache, query);    
    g_hash_table_remove (self->private->remote_queries, GUINT_TO_POINTER(ticket));
 
    return TRUE;
}

typedef struct {
    GList *data;
    gint   metadatum_id;
} ProjectResult;

static void
collect_projected_values (HTuple	  *tuple,
			  ProjectResult	  *project_result)
{
    HTupleElem	*elem;

    elem = h_tuple_get (tuple, bmp_metadata_get_id_static (project_result->metadatum_id));
    project_result->data = g_list_append (project_result->data, g_strdup(elem->value));
}

gboolean
bmp_system_control_query_list_datum     (BmpSystemControl  *self,
					 gint		    metadatum_id,
					 gint		    metadatum_id_predicate,
					 gchar		   *predicate_value,
					 gchar		 ***datum_list,
					 GError		  **error)
{
    HRelation	    *projected;
    ProjectResult   *project_result;

    g_return_val_if_fail (BMP_IS_SYSTEM_CONTROL(self), FALSE);

    projected = bmp_metadata_cache_query_list_datum (bmp_metadata_cache, 
						     metadatum_id,
						     metadatum_id_predicate,
						     predicate_value);    

    project_result = g_new0 (ProjectResult, 1);

    project_result->data = NULL;
    project_result->metadatum_id = metadatum_id;

    h_relation_for_each_tuple (projected, (HTupleFunc)collect_projected_values, project_result);

    *datum_list = glist_to_strv (project_result->data);

    return TRUE;
}

/* END Query API */

gboolean
bmp_system_control_send_startup_complete (BmpSystemControl *self)
{
    g_return_val_if_fail (BMP_IS_SYSTEM_CONTROL(self), FALSE);

    g_signal_emit (G_OBJECT (self),
		   signals[BMP_SYSTEM_CONTROL_SIGNAL_STARTUP_COMPLETE], 0);

    if (xcs_key_fetch_boolean (xcs, "bmp", "resume-on-startup"))
      {
	  gint	  track,
		  time;

	  time  = xcs_key_fetch_int (xcs, "bmp", "resume-on-startup-time");
	  track = xcs_key_fetch_int (xcs, "bmp", "resume-on-startup-track");

	  if ((time != -1) && (track != -1))
	    {
	      gint n;
	      play_track (self, track);

	      for (n = 0; n < 50; n++)
		{
		  while (g_main_context_pending(NULL)) g_main_context_iteration (NULL, TRUE);
		  g_usleep (250);
		}

	      play_seek  (self, time);
	    }
      }

    return FALSE;
}

gboolean
bmp_system_control_send_shutdown_complete (BmpSystemControl *self)
{
    g_return_val_if_fail (BMP_IS_SYSTEM_CONTROL(self), FALSE);

    g_signal_emit (G_OBJECT (self),
		   signals[BMP_SYSTEM_CONTROL_SIGNAL_SHUTDOWN_COMPLETE], 0);

    return TRUE;
}

gboolean
bmp_system_control_add_uri_list (BmpSystemControl   *self,
				 const gchar	   **uri_list,
				 gint		     position,
				 gboolean	     clear,
				 gboolean	     start_playback,
				 gint		     index,
				 gint		    *position_out,
				 GError		   **error)
{
    g_return_val_if_fail (BMP_IS_SYSTEM_CONTROL (self), TRUE);

    gtk_tree_sortable_set_sort_column_id (GTK_TREE_SORTABLE(bmp_playlist_get_playlist(bmp_playlist)),
					  GTK_TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID,
					  GTK_SORT_ASCENDING);
    bmp_system_control_app_busy (self);

    if (clear) bmp_playlist_tracklist_remove_rows_all (bmp_playlist);

    index = (index == -1) ? bmp_playlist_get_length (bmp_playlist) : index;

    *position_out = (position == -1) ? bmp_playlist_insert_tracklist_items (bmp_playlist, uri_list) : 
				       bmp_playlist_insert_tracklist_items_position (bmp_playlist, uri_list, position);

    if (start_playback) bmp_system_control_play_track (self, index, NULL);

    bmp_system_control_app_idle (self);

    return TRUE;
}


gboolean
bmp_system_control_get_metadata_for_list_item (BmpSystemControl	  *self,
					       gint		   item_idx, 
					       GHashTable	 **metadata,
					       GError		 **error)
{
    BmpFileItem *item;

    item = bmp_playlist_get_row_by_index (bmp_playlist, item_idx); 
    
    if (!item)
	{
#ifdef HAVE_DBUS
	    *error = g_error_new (g_quark_from_static_string (BMP_DBUS_INTERFACE),
				  0, 
				  "Invalid item index specified.");
#else
	    *error = g_error_new (g_quark_from_static_string ("BMPx"),
				  0, 
				  "Invalid item index specified.");
#endif
	    return FALSE;
	}

    *metadata = bmp_metadata_cache_get_hash_value_table (bmp_metadata_cache, item);

    return TRUE;
}

gboolean
bmp_system_control_tracklist_remove_items (BmpSystemControl  *self,
					   GArray	     *rows,	
					   gint		      n_rows,
					   GError	    **error)
{
    bmp_playlist_tracklist_remove_rows (bmp_playlist, rows, n_rows);

    return TRUE;
}

gboolean
bmp_system_control_send_status (BmpSystemControl *self, GError ** error)
{
    g_signal_emit (G_OBJECT (self),
		   signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_PLAYSTATUS],
		   0, g_object_get_int (bmp_play_engine, "playstatus"));

    return TRUE;
}


gboolean
bmp_system_control_tracklist_sort (BmpSystemControl  *self,
				   gint		      metadatum_id,
				   GError	    **error)
{
    g_return_val_if_fail (BMP_IS_SYSTEM_CONTROL (self), FALSE);

    bmp_playlist_tracklist_sort (bmp_playlist, metadatum_id);

    return TRUE;
}

gboolean
bmp_system_control_eq_get (BmpSystemControl *self,
			   gint eq, gint * value, GError ** error)
{
    *value = eq_values[eq];

    return TRUE;
}

gboolean
bmp_system_control_volume_get (BmpSystemControl  *self,
			       gint		 *volume,
			       GError		**error)
{
    *volume = self->private->volume;

    return TRUE;
}

gboolean
bmp_system_control_get_current_title (BmpSystemControl *self,
				      gchar ** title, GError ** error)
{
    *title = g_strdup (self->private->current_title); 

    return TRUE;
}

gboolean
bmp_system_control_get_current_uri (BmpSystemControl	*self,
				    gchar	       **uri,
				    GError	       **error)
{
    if (!self->private->current_item)
      {
#ifdef HAVE_DBUS
	    *error = g_error_new (g_quark_from_static_string (BMP_DBUS_INTERFACE),
				  0, 
				 "No current URI"); 
#else
	    *error = g_error_new (g_quark_from_static_string ("BMPx"),
				  0, 
				 "No current URI");
#endif
	*uri = NULL;
	return FALSE;
      }

    *uri = g_strdup (self->private->current_item->uri);

    return TRUE;
}

gboolean
bmp_system_control_get_current_track (BmpSystemControl  *self,
				      gint	        *idx,
				      GError	       **error)
{
    *idx = g_object_get_int (self, "current_index");
  
    return TRUE;
}

gboolean
bmp_system_control_clear_tracklist		(BmpSystemControl      *self,
						 GError		      **error)
{
    bmp_playlist_tracklist_remove_rows_all (bmp_playlist);
    bmp_system_control_clear_playback_history (self, NULL);
    bmp_system_control_update_title (bmp_system_control);

    g_object_set (G_OBJECT(self), "current_index", (gint)-1, NULL);    

    return TRUE;
}
      

gboolean
bmp_system_control_get_tracklist (BmpSystemControl *self,
				  char ***array,
				  GError ** error)
{
    GtkListStore      *list; 
    GtkTreeIter	       iter;
    gint	       n, l;
    gboolean	       loop;

    list = bmp_playlist_get_playlist (bmp_playlist); 

    l = gtk_tree_model_iter_n_children (GTK_TREE_MODEL(list), NULL);
    *array = g_new0 (gchar*, l+1); 

    loop = gtk_tree_model_get_iter_first (GTK_TREE_MODEL(list), &iter);
    n = 0;

    //FIXME: This is not completely safe because rows might get removed while we're doing this.
    while (loop)
      {
	    BmpFileItem *item;

	    gtk_tree_model_get (GTK_TREE_MODEL(list), &iter, TUPLE_COLUMN, &item, -1);
	    (*array)[n] = g_strdup (item->uri); 
	    loop = gtk_tree_model_iter_next (GTK_TREE_MODEL(list), &iter);
	    n++;
      }

    (*array)[l] = NULL;

    return TRUE;
}

gboolean
bmp_system_control_export_tracklist (BmpSystemControl *self,
				     const gchar      *container,
				     gboolean	       all,
				     const gchar      *filename,
				     GError	     **error)
{
    GtkListStore      *list;
    GtkTreeIter	       iter;
    VFSHandle	      *handle;
    gboolean	       loop;

    handle = vfs_handle_new ();
    handle->filename = g_strdup (filename);
    list = bmp_playlist_get_playlist (bmp_playlist); 

#ifdef HAVE_GUI
    if ((all) || (!bmp_ui))
#else
    if (1)
#endif
      {
	loop = gtk_tree_model_get_iter_first (GTK_TREE_MODEL(list), &iter);

	while (loop)
	  {
	    BmpFileItem	*item;

	    gtk_tree_model_get (GTK_TREE_MODEL(list), &iter, TUPLE_COLUMN, &item, -1);
	    handle->uri_list = g_list_append (handle->uri_list, g_strdup (item->uri));
	    loop = gtk_tree_model_iter_next (GTK_TREE_MODEL(list), &iter);
	  }
      }
#ifdef HAVE_GUI
    else
      {
	  GList	*paths;

	  paths = chroma_list_selection_get_selected_paths (CHROMA_LIST_SELECTION (c_tracklist));

	  for ( ; paths ; paths = paths->next )
	    {
		BmpFileItem	*item;
		GtkTreeIter	 iter;

		gtk_tree_model_get_iter (GTK_TREE_MODEL(list), &iter, (GtkTreePath*)paths->data);
		gtk_tree_model_get (GTK_TREE_MODEL(list), &iter, TUPLE_COLUMN, &item, -1);
		handle->uri_list = g_list_append (handle->uri_list, g_strdup (item->uri));
	    }

	  g_list_foreach (g_list_first (paths), (GFunc)gtk_tree_path_free, NULL);
      }
#endif

    vfs_handle_write (handle, container); 
    vfs_handle_free (handle);

    return TRUE;
}

/* Our stuff */
static void
emit_row_changed_on_index	(BmpSystemControl   *self,
				 gint		     index)  
{
    GtkListStore *model = bmp_playlist_get_playlist (bmp_playlist);
    GtkTreePath	 *path;
    GtkTreeIter   iter;

    if (index < 0) return;

    path = gtk_tree_path_new_from_indices (index, -1);
    gtk_tree_model_get_iter (GTK_TREE_MODEL(model), &iter, path);
    gtk_tree_model_row_changed (GTK_TREE_MODEL(model), path, &iter);
    gtk_tree_path_free (path);
}

static gboolean
play_track (BmpSystemControl	    *self,
	    gint		     position)
{
    GtkListStore      *list;
    GtkTreePath	      *path;
    BmpFileItem	      *item;
    gpointer	       data;

    item = bmp_playlist_get_row_by_index (bmp_playlist, position);
    if (!item) return FALSE;

    list = bmp_playlist_get_playlist (bmp_playlist);
    data = g_object_get_data (list, "current");
    g_object_set_data (list, "current", NULL);
    if (data)
      {
	path = gtk_tree_row_reference_get_path ((GtkTreeRowReference*)(data));
	if (path)
	  {
	    emit_row_changed_on_index (self, gtk_tree_path_get_indices (path)[0]); 
	    g_object_set_data (list, "current", data);
	    gtk_tree_path_free (path);
	  }
	else
	    gtk_tree_row_reference_free ((GtkTreeRowReference*)data);
      }
    
    g_object_set (G_OBJECT(self), "current_index", position, NULL);
    bmp_playback_history_set (self->private->playback_history, position); 

    g_object_set (G_OBJECT (bmp_play_engine), "playstatus", PLAYSTATUS_WAITING, NULL);
    g_object_set (G_OBJECT (bmp_play_engine), "mrl", item->uri, NULL);
    g_object_set (G_OBJECT (bmp_play_engine), "playstatus", PLAYSTATUS_PLAYING, NULL);

    self->private->anonymous = FALSE;
    emit_row_changed_on_index (self, position);
    bmp_system_message_send (self, BMP_SYSTEM_MESSAGE_TRACK_CHANGE, NULL, NULL);
    bmp_system_message_send (self, BMP_SYSTEM_MESSAGE_UPDATE_TITLE, NULL, NULL);

    return TRUE;
}

static void
play_current (BmpSystemControl *self)
{
#ifdef HAVE_GUI
    if (c_tracklist)
      {
	GList	 *list;
	gint	  current = 0;

	list = chroma_list_selection_get_selected_paths (CHROMA_LIST_SELECTION(c_tracklist));

	if ((!list) || (g_list_length(list)>1))
	  {
	    GtkTreeRowReference *reference;

	    reference = g_object_get_data (G_OBJECT(bmp_playlist_get_playlist(bmp_playlist)), "current");
	    if (reference)
	      {
		GtkTreePath *path;
		 
		path = gtk_tree_row_reference_get_path (reference); 
		if (path)
		  {
		    current = gtk_tree_path_get_indices (path)[0];
		    gtk_tree_path_free (path);
		  }
		else
		  {
		    gtk_tree_row_reference_free (reference);
		    g_object_set_data (G_OBJECT(bmp_playlist_get_playlist(bmp_playlist)), "current", NULL);
		  }
	      }
	  }
	else
	  {
	    current = gtk_tree_path_get_indices (list->data)[0];
	    g_list_foreach (list, (GFunc)gtk_tree_path_free, NULL);
	  }

	play_track (self, current); 
      }
#else
    GList		*list;
    GtkTreeRowReference *reference;
    gint		 current = 0;

    if (g_object_get_int (G_OBJECT(bmp_play_engine), "playstatus") == PLAYSTATUS_PLAYING)
      {
	play_track (self, self->private->current_index); 
      }

    reference = g_object_get_data (G_OBJECT(bmp_playlist_get_playlist(bmp_playlist)), "current");
    if (reference)
      {
	  GtkTreePath *path;
		 
	  path = gtk_tree_row_reference_get_path (reference); 
	  if (path)
	    {
		current = gtk_tree_path_get_indices (path)[0];
		gtk_tree_path_free (path);
	    }
	  else
	    {
		gtk_tree_row_reference_free (reference);
		g_object_set_data (G_OBJECT(bmp_playlist_get_playlist(bmp_playlist)), "current", NULL);
	    }
      }

    play_track (self, current); 
#endif
}

static void
play_next (BmpSystemControl *self)
{
    GtkListStore  *list;
    GtkTreePath	  *path;
    BmpFileItem	  *item;
    gboolean	   repeat;
    gint	   position;
    gpointer	   data;

    if (self->private->stop_after_current)
      {
	g_object_set (G_OBJECT (bmp_play_engine), "playstatus", PLAYSTATUS_STOPPED, NULL);
	g_object_set (G_OBJECT (self), "stop_after_current", FALSE, NULL);
	return;
      }

    list = bmp_playlist_get_playlist (bmp_playlist);
    data = g_object_get_data (list, "current");
    g_object_set_data (list, "current", NULL);
    if (data)
      {
	path = gtk_tree_row_reference_get_path ((GtkTreeRowReference*)(data));
	if (path)
	  {
	    emit_row_changed_on_index (self, gtk_tree_path_get_indices (path)[0]); 
	    gtk_tree_path_free (path);
	    g_object_set_data (list, "current", data);
	  }
	else
	  gtk_tree_row_reference_free ((GtkTreeRowReference*)data);
      }

    repeat = g_object_get_boolean (self, "repeat");
    position = bmp_playback_history_next (self->private->playback_history,
					  (BmpPluginFlowInterface*)self->private->flow_plugin); 
    if (position < 0)
      {
	if (!repeat) 
	  {
	    g_object_set (G_OBJECT (bmp_play_engine), "playstatus", PLAYSTATUS_STOPPED, NULL);
	    return;
	  }

	bmp_playback_history_rewind (self->private->playback_history); 
	position = 0;
      }

    item = bmp_playlist_get_row_by_index (bmp_playlist, position);
    if (!item) return;

    g_object_set (G_OBJECT (bmp_play_engine), "playstatus", PLAYSTATUS_WAITING, NULL);
    g_object_set (G_OBJECT (bmp_play_engine), "mrl", item->uri, NULL);
    g_object_set (G_OBJECT (bmp_play_engine), "playstatus", PLAYSTATUS_PLAYING, NULL);

    self->private->anonymous = FALSE;
    g_object_set (G_OBJECT(self), "current_index", position, NULL);
    emit_row_changed_on_index (self, position);

    bmp_system_message_send (self, BMP_SYSTEM_MESSAGE_TRACK_CHANGE, NULL, NULL);
    bmp_system_message_send (self, BMP_SYSTEM_MESSAGE_UPDATE_TITLE, NULL, NULL);
}

static void
play_prev (BmpSystemControl *self)
{
    GtkListStore  *list;
    GtkTreePath	  *path;
    BmpFileItem	  *item;
    gboolean	   repeat;
    gint	   position;
    gpointer	   data;

    if (self->private->stop_after_current)
      {
	g_object_set (G_OBJECT (self), "stop_after_current", FALSE, NULL);
	return;
      }

    list = bmp_playlist_get_playlist (bmp_playlist);
    data = g_object_get_data (list, "current");
    g_object_set_data (list, "current", NULL);
    if (data)
      {
	path = gtk_tree_row_reference_get_path ((GtkTreeRowReference*)(data));
	if (path)
	  {
	    emit_row_changed_on_index (self, gtk_tree_path_get_indices (path)[0]); 
	    gtk_tree_path_free (path);
	    g_object_set_data (list, "current", data);
	  }
	else
	  gtk_tree_row_reference_free ((GtkTreeRowReference*)data);
      }

    repeat = g_object_get_boolean (self, "repeat");
    position = bmp_playback_history_prev (self->private->playback_history,
					  (BmpPluginFlowInterface*)self->private->flow_plugin); 
    if (position < 0) 
      {
	if (!repeat) return;
	bmp_playback_history_rewind (self->private->playback_history); 
	position = 0;
      }

    item = bmp_playlist_get_row_by_index (bmp_playlist, position);
    if (!item)
      {
	g_object_set (G_OBJECT (bmp_play_engine), "playstatus", PLAYSTATUS_STOPPED, NULL);
	return;
      }

    g_object_set (G_OBJECT (bmp_play_engine), "playstatus", PLAYSTATUS_WAITING, NULL);
    g_object_set (G_OBJECT (bmp_play_engine), "mrl", item->uri, NULL);
    g_object_set (G_OBJECT (bmp_play_engine), "playstatus", PLAYSTATUS_PLAYING, NULL);

    self->private->anonymous = FALSE;
    g_object_set (G_OBJECT(self), "current_index", position, NULL);
    emit_row_changed_on_index (self, position);

    bmp_system_message_send (self, BMP_SYSTEM_MESSAGE_TRACK_CHANGE, NULL, NULL);
    bmp_system_message_send (self, BMP_SYSTEM_MESSAGE_UPDATE_TITLE, NULL, NULL);
}

static void
play_pause (BmpSystemControl *self)
{
    g_return_if_fail (BMP_IS_SYSTEM_CONTROL (self));

    g_object_set (G_OBJECT (bmp_play_engine), "playstatus", PLAYSTATUS_PAUSED, NULL);
}

static void
play_stop (BmpSystemControl *self)
{
    g_return_if_fail (BMP_IS_SYSTEM_CONTROL (self));

    g_object_set (G_OBJECT (bmp_play_engine), "playstatus", PLAYSTATUS_STOPPED, NULL);
}

static void
play_seek	  (BmpSystemControl	*self,
		   gint		         position)
{
    g_return_if_fail (BMP_IS_SYSTEM_CONTROL (self));

    g_object_set (G_OBJECT(bmp_play_engine), "position", position, NULL);
}

static void
play_seek_percent (BmpSystemControl	*self,
		   gdouble		 position)
{
    gdouble percent,
	    length,
	    position_new;

    g_return_if_fail (BMP_IS_SYSTEM_CONTROL (self));

    length = bmp_play_get_length (bmp_play_engine);
    percent = length / 100.;

    position_new = position * percent;

    g_object_set (G_OBJECT(bmp_play_engine), "position", (gint)position_new, NULL);
}

static void
play_volume (BmpSystemControl *self, gint volume)
{
    g_return_if_fail (BMP_IS_SYSTEM_CONTROL (self));

    g_object_set (self, "volume", volume, NULL);
}

static void
play_eq (BmpSystemControl *self, BmpEqEnum eq, gint value)
{
    g_return_if_fail (BMP_IS_SYSTEM_CONTROL (self));

#if 0
    bmp_play_set_eq (bmp_play_engine, eq, value);
    eq_values[eq] = value;
    g_signal_emit (G_OBJECT (self),
		   signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_EQ],
		   0, eq, value); 
#endif
}

static void
configuration_save (BmpSystemControl *self)
{
    /* SAVE CONFIGURATION */
    xcs_key_set_int (xcs, "bmp", "volume", self->private->volume);
}

static gboolean
process_messages (BmpSystemControl *self)
{
    BmpSystemMessage *message;
    gint queue_length;

    g_return_val_if_fail (BMP_IS_SYSTEM_CONTROL (self), FALSE);

    message = g_async_queue_try_pop (self->private->message_queue);

    if (!message)
      {
	self->private->message_queue_source_id = 0;
	return FALSE;
      }

    switch (message->message_type)
	{
	case BMP_SYSTEM_MESSAGE_PAUSE:
		{
		    play_pause (self);
		}
	    break;

	case BMP_SYSTEM_MESSAGE_STOP:
		{
		    play_stop (self);
		}
	    break;

	case BMP_SYSTEM_MESSAGE_NEXT:
		{
		    play_next (self);
		}
	    break;

	case BMP_SYSTEM_MESSAGE_CURRENT:
		{
		    play_current (self);
		}
	    break;

	case BMP_SYSTEM_MESSAGE_PREV:
		{
		    play_prev (self);
		}
	    break;

	case BMP_SYSTEM_MESSAGE_PLAY_TRACK:
		{
		    play_track (self, GPOINTER_TO_INT (message->data)); 
		}
	    break;

	case BMP_SYSTEM_MESSAGE_TRACK_CHANGE:
		{
		    g_signal_emit (self, signals[BMP_SYSTEM_CONTROL_SIGNAL_TRACK_CHANGE], 0); 
		}
	    break;

	case BMP_SYSTEM_MESSAGE_SEEK:
		{
		    play_seek (self, GPOINTER_TO_INT (message->data));
		}
	    break;

	case BMP_SYSTEM_MESSAGE_SEEK_PERCENT:
		{
		    play_seek_percent (self, GPOINTER_TO_INT (message->data));
		}
	    break;

	case BMP_SYSTEM_MESSAGE_VOLUME:
		{
		    play_volume (self, GPOINTER_TO_INT (message->data));
		}
	    break;

	case BMP_SYSTEM_MESSAGE_EQ:
		{
		    bmp_message_eq_t *message_eq;

		    message_eq = message->data;

		    play_eq (self, message_eq->eq, message_eq->value);
		}
	    break;

	case BMP_SYSTEM_MESSAGE_BITRATE:
		{
		    g_object_set (self, "bitrate",
				  GPOINTER_TO_INT (message->data), NULL);
		}
	    break;

	case BMP_SYSTEM_MESSAGE_SAMPLERATE:
		{
		    g_object_set (self, "samplerate",
				  GPOINTER_TO_INT (message->data), NULL);
		}
	    break;

	case BMP_SYSTEM_MESSAGE_SAVE_STATE:
		{
		    configuration_save (self);
		}
	    break;

	case BMP_SYSTEM_MESSAGE_SHOW_PL_WINDOW:
		{
/* FIXME: Should we ignore these properties when HAVE_GUI is not defined, or actually not implement them? */
#ifdef HAVE_GUI
		    gtk_toggle_action_set_active (GTK_TOGGLE_ACTION
						  (gtk_action_group_get_action
						   (bmp_toggle_actions,
						    BMP_TOGGLE_ACTION_PL)),
						  GPOINTER_TO_INT (message->
								   data));
#endif
		}
	    break;
	
	case BMP_SYSTEM_MESSAGE_UPDATE_TITLE:
	    {
	        GtkListStore	    *tracklist;
		GtkTreeRowReference *reference;
		GtkTreePath	    *path;
		GtkTreeIter	     iter;
		BmpFileItem	    *item;
		GList		    *list;
		HTuple		    *tuple;
		gchar		    *title;
	  	gint		     current;
		gboolean	     multiple;

		if (g_object_get_int (G_OBJECT(bmp_play_engine), "playstatus") == PLAYSTATUS_WAITING) break;
		if (g_object_get_int (G_OBJECT(bmp_play_engine), "playstatus") == PLAYSTATUS_PAUSED) break;

		tracklist = bmp_playlist_get_playlist (bmp_playlist);
		list = NULL; /*Shadup GCC*/

		if (bmp_ui)
		  {
		    multiple = chroma_list_selection_multiple_selected (CHROMA_LIST_SELECTION(c_tracklist));
		    list = chroma_list_selection_get_selected_paths (CHROMA_LIST_SELECTION(c_tracklist));
		  }

		reference = g_object_get_data (G_OBJECT(tracklist), "current");
		if ((g_object_get_int (G_OBJECT(bmp_play_engine), "playstatus") == PLAYSTATUS_PLAYING) && reference)
		  {
			path = gtk_tree_row_reference_get_path (reference);
			if (!path)
			  {
			    gtk_tree_row_reference_free (reference);
			    g_object_set_data (G_OBJECT(tracklist), "current", NULL);
			    self->private->anonymous = TRUE;
			    break;
			  }
		  }
		else
		  {
			if (bmp_ui && list)
			  {
			    path = gtk_tree_path_copy (list->data); 
			  }
			else
			  {
			    if (g_object_get_int (G_OBJECT(bmp_play_engine), "playstatus") == PLAYSTATUS_PLAYING)
			      {
				  self->private->anonymous = TRUE;
				  g_signal_emit (self, signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_TITLE], 0);
				  break;
			      }
			    
			      if (self->private->current_title)
				  g_free (self->private->current_title);

			      if (self->private->current_item)
				  g_boxed_free (BMP_TYPE_FILE_ITEM, self->private->current_item); 

			      self->private->current_title = NULL; 
			      self->private->current_item  = NULL; 

			      g_signal_emit (self, signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_TITLE], 0);
			      break;
			  }
		    }

		current = gtk_tree_path_get_indices (path)[0];
		title = bmp_playlist_get_title_nth (bmp_playlist, current);
		gtk_tree_model_get_iter (GTK_TREE_MODEL(tracklist), &iter, path);
		gtk_tree_path_free (path);

		if (title)
		  {
			if (self->private->current_title && (!strcmp(title, self->private->current_title)))
			  {
			      g_free (title);
			      break;
			  }

  
			if ((g_object_get_int (G_OBJECT(bmp_play_engine), "playstatus") == PLAYSTATUS_PLAYING) && self->private->anonymous)
			  {
			      g_free (title);
			      break;
			  }
		  }

		if (self->private->current_title) g_free (self->private->current_title);
		if (self->private->current_item)  g_boxed_free (BMP_TYPE_FILE_ITEM, self->private->current_item);

		gtk_tree_model_get (GTK_TREE_MODEL(tracklist), &iter, TUPLE_COLUMN, &item, -1);
		self->private->current_title = title; 
		self->private->current_item  = item; 

		g_signal_emit (self, signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_TITLE], 0);

		tuple = bmp_metadata_cache_metadata_get (bmp_metadata_cache, item); 
		if (tuple)
		  {
		      HTupleElem  *elem;
		      gint	   bitrate,
				   samplerate;

		      elem = h_tuple_get (tuple, bmp_metadata_get_id_static (BMP_DATUM_BITRATE));	
		      bitrate = GPOINTER_TO_INT(elem->value);

		      elem = h_tuple_get (tuple, bmp_metadata_get_id_static (BMP_DATUM_SAMPLERATE));	
		      samplerate = GPOINTER_TO_INT(elem->value);

		      h_tuple_free (tuple);

		      self->private->bitrate = bitrate;
		      self->private->samplerate = samplerate;
		  }
	      }
	    break;

	case BMP_SYSTEM_MESSAGE_STOP_AFTER_CURRENT:
		{
		    gboolean stop;

		    stop = GPOINTER_TO_INT (message->data);

		    g_object_set (G_OBJECT (self), "stop_after_current", stop, NULL);
		}
	    break;

	case BMP_SYSTEM_MESSAGE_REPEAT:
		{
		    gboolean repeat;

		    repeat = GPOINTER_TO_INT (message->data);

		    g_object_set (G_OBJECT (self), "repeat", repeat, NULL);
		}
	    break;

	case BMP_SYSTEM_MESSAGE_SHUFFLE:
		{
		    gboolean shuffle;

		    shuffle = GPOINTER_TO_INT (message->data);
		    g_object_set (G_OBJECT (self), "shuffle", shuffle, NULL);
		}
	    break;

	case BMP_SYSTEM_MESSAGE_QUIT:
		{
		    gboolean rval;
 
		    rval = FALSE; 
		    g_signal_emit (G_OBJECT(self), signals[BMP_SYSTEM_CONTROL_SIGNAL_SHUTDOWN_REQUEST], 0, &rval);

		    if (!rval)
		      {
			  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "SHUTTING DOWN"); 

			  if ((xcs_key_fetch_boolean (xcs, "bmp", "resume-on-startup") && (!self->private->anonymous)) && 
			    (g_object_get_int (bmp_play_engine, "playstatus") == PLAYSTATUS_PLAYING))
			    {
			      xcs_key_set_int (xcs, "bmp", "resume-on-startup-time", self->private->stream_pos);
			      xcs_key_set_int (xcs, "bmp", "resume-on-startup-track", self->private->current_index);
			    }
			  else 
			    {
			      xcs_key_set_int (xcs, "bmp", "resume-on-startup-time", -1); 
			      xcs_key_set_int (xcs, "bmp", "resume-on-startup-track", -1); 
			    }

			  configuration_save (self);

#ifdef HAVE_GUI
			  if (bmp_ui)
			    {
			      g_object_unref (bmp_ui);
			      _bmp_ui = NULL;
			    }

			  gtk_main_quit ();
#else
			  g_main_loop_quit (mainloop);
#endif
		      }
		}
	    break;

	case BMP_SYSTEM_MESSAGE_UI_START:
		{
#ifdef HAVE_GUI
		    GdkDisplay *display;

		    if (message->data)
		      {
			display = gdk_display_open ((gchar*)message->data);
			if (display)
			  {
			    gdk_display_manager_set_default_display (gdk_display_manager_get (), display);
			  }
			else 
			  gdk_display_open_default_libgtk_only();
		      }
		    else
			gdk_display_open_default_libgtk_only();
 
		    _bmp_ui = bmp_ui_new ();
		    if (bmp_ui_finalize_init (bmp_ui) != BMP_ERROR_SUCCESS)
			{
			    g_object_unref (bmp_ui);
			    _bmp_ui = NULL;
			    gtk_main_quit ();
			}

		    bmp_system_control_update_title (self);
#endif
		}
	    break;

	case BMP_SYSTEM_MESSAGE_UI_STOP:
		{
#ifdef HAVE_GUI
		    if (bmp_ui)
		      {
			g_object_unref (bmp_ui);
			_bmp_ui = NULL;
		      }
#endif
		}
	    break;
	}

    g_free (message);

    queue_length = g_async_queue_length (self->private->message_queue);

    if (queue_length == 0)
      {
	self->private->message_queue_source_id = 0;
	return FALSE;
      }

    return TRUE;
}

typedef gpointer SystemMessageData;

/* Public API */
void
bmp_system_message_send (BmpSystemControl	*self,
			 BmpSystemMessageType    message_type,
			 SystemMessageData	 data, 
		         GDestroyNotify          data_destroy)
{
    BmpSystemMessage *message;

    g_return_if_fail (BMP_IS_SYSTEM_CONTROL (self));

    message = g_new0 (BmpSystemMessage, 1);
    message->message_type = message_type;
    message->data = data;
    message->data_destroy = data_destroy;

    g_async_queue_push (self->private->message_queue, message);

    if (self->private->message_queue_source_id != 0) return;
    self->private->message_queue_source_id = g_idle_add ((GSourceFunc) process_messages, self);
}

void
bmp_system_control_update_title (BmpSystemControl *self)
				    
{
    bmp_system_message_send (self, BMP_SYSTEM_MESSAGE_UPDATE_TITLE, NULL, NULL);
}

void
bmp_system_control_show_pl_window (BmpSystemControl *self, gboolean show)
{
    bmp_system_message_send (self, 
			     BMP_SYSTEM_MESSAGE_SHOW_PL_WINDOW,
			     (SystemMessageData)GINT_TO_POINTER(show),
			     NULL);
}

gboolean
bmp_system_control_raise_ui	(BmpSystemControl *self, GError **error)
{
#ifdef HAVE_GUI
    bmp_ui_raise_windows (bmp_ui);
#endif
    return TRUE;
}

gboolean
bmp_system_control_startup	(BmpSystemControl *self, GError **error)
{

    return TRUE;
}

gboolean
bmp_system_control_stop_after_current_set (BmpSystemControl *self, gboolean stop, GError **error)
{
    bmp_system_message_send (self,
			     BMP_SYSTEM_MESSAGE_STOP_AFTER_CURRENT,
			     (SystemMessageData)GINT_TO_POINTER(stop),
			     NULL);

    return TRUE;
}

gboolean
bmp_system_control_repeat_set (BmpSystemControl *self, gboolean repeat, GError **error)
{
    bmp_system_message_send (self,
			     BMP_SYSTEM_MESSAGE_REPEAT,
			     (SystemMessageData)GINT_TO_POINTER(repeat),
			     NULL);

    return TRUE;
}

gboolean
bmp_system_control_repeat_get (BmpSystemControl *self, gboolean *repeat, GError **error)
{
    *repeat = self->private->repeat;

    return TRUE;
}


gboolean
bmp_system_control_shuffle_set (BmpSystemControl *self, gboolean shuffle, GError **error)
{
    bmp_system_message_send (self,
			     BMP_SYSTEM_MESSAGE_SHUFFLE,
			     (SystemMessageData)GINT_TO_POINTER(shuffle),
			     NULL);
    return TRUE;
}

gboolean
bmp_system_control_shuffle_get (BmpSystemControl *self, gboolean *shuffle, GError **error)
{
    *shuffle = self->private->shuffle;
    return TRUE;
}

gboolean
bmp_system_control_ui_start (BmpSystemControl *self, gchar *display, GError **error)
{
    bmp_system_message_send (self, BMP_SYSTEM_MESSAGE_UI_START, display, NULL);
    return TRUE;
}

gboolean
bmp_system_control_ui_stop (BmpSystemControl *self, GError **error)
{
    bmp_system_message_send (self, BMP_SYSTEM_MESSAGE_UI_STOP, NULL, NULL);

    return TRUE;
}

void
bmp_system_control_quit (BmpSystemControl *self)
{
    bmp_system_message_send (self, BMP_SYSTEM_MESSAGE_QUIT, NULL, NULL);
}

gboolean
bmp_system_control_volume_set (BmpSystemControl *self, gint volume, GError **error)
{
    g_object_set (self, "volume", volume, NULL);
    return TRUE;
}

gboolean
bmp_system_control_eq_set (BmpSystemControl *self, BmpEqEnum eq, gint value, GError **error)
{
    bmp_message_eq_t *message;

#define MAX_EQ_BANDS 10

    if (eq > MAX_EQ_BANDS)
      {
      }

    message = g_new0 (bmp_message_eq_t, 1);

    message->eq = eq;
    message->value = value;

    bmp_system_message_send (self,    
			     BMP_SYSTEM_MESSAGE_EQ,  
			     message, 
			     (GDestroyNotify)g_free);

    return TRUE;
}

void
bmp_system_control_save_state (BmpSystemControl *self)
{
    bmp_system_message_send (self,
			     BMP_SYSTEM_MESSAGE_SAVE_STATE,
			     NULL,
			     NULL);
}

static gboolean
check_playback_sanity	     (BmpSystemControl	      *self,
			      GError		     **error)
{
  if (!g_object_get_boolean (bmp_play_engine, "sane"))
	{
	  bmp_system_control_message_dispatch (bmp_system_control,
				       self->private->sc_msg_domain_id, 
				       GTK_MESSAGE_ERROR,
				       ("The audio system is not properly initialized. Please check the audio settings in the Preferences panel."));
	  if (error != NULL) 
	    {
	      *error = g_error_new (g_quark_from_static_string (BMP_DBUS_INTERFACE),
				    0, 
				    "Unable to initiate playback: check audio settings");
	    }
	  return FALSE;
	}

    return TRUE;
}


gboolean
bmp_system_control_play_track (BmpSystemControl	      *self,
			       gint		       index,
			       GError		     **error)
{
    if (!check_playback_sanity (self, error)) return FALSE;

    bmp_system_message_send (self,
			     BMP_SYSTEM_MESSAGE_PLAY_TRACK,
			     (SystemMessageData)GINT_TO_POINTER(index),
			     NULL);
    return TRUE;
}

gboolean
bmp_system_control_seek (BmpSystemControl	      *self,
			 gint			       position,
			 GError			     **error)
{
#if 0
    bmp_system_message_send (self,
			     BMP_SYSTEM_MESSAGE_SEEK,
			     (SystemMessageData)GINT_TO_POINTER(position),
			     NULL);
#endif

    if (!check_playback_sanity (self, error)) return FALSE;

    play_seek (self, position);
    return TRUE;
}

gboolean
bmp_system_control_seek_percent (BmpSystemControl     *self,
			         gdouble	       position,
				 GError		     **error)
{
#if 0
    bmp_system_message_send (self,
			     BMP_SYSTEM_MESSAGE_SEEK_PERCENT,
			     (SystemMessageData)GINT_TO_POINTER(position),
			     NULL);
#endif

    if (!check_playback_sanity (self, error)) return FALSE;

    play_seek_percent (self, position);
    return TRUE;
}

gboolean
bmp_system_control_pause      (BmpSystemControl	      *self,
			       GError		     **error)
{
    if (!check_playback_sanity (self, error)) return FALSE;

    bmp_system_message_send (self, BMP_SYSTEM_MESSAGE_PAUSE, NULL, NULL);
    return TRUE;
}

gboolean
bmp_system_control_stop      (BmpSystemControl	      *self,
			      GError		     **error)
{
    if (!check_playback_sanity (self, error)) return FALSE;

    bmp_system_message_send (self, BMP_SYSTEM_MESSAGE_STOP, NULL, NULL);
    return TRUE;
}

gboolean
bmp_system_control_go_prev   (BmpSystemControl	      *self,
			      GError		     **error)
{
    if (!check_playback_sanity (self, error)) return FALSE;

    bmp_system_message_send (self, BMP_SYSTEM_MESSAGE_PREV, NULL, NULL);
    return TRUE;
}

gboolean
bmp_system_control_go_next   (BmpSystemControl	      *self,
			      GError		     **error)
{
    if (!check_playback_sanity (self, error)) return FALSE;

    bmp_system_message_send (self, BMP_SYSTEM_MESSAGE_NEXT, NULL, NULL);
    return TRUE;
}

gboolean
bmp_system_control_play	     (BmpSystemControl     *self,
			      GError		  **error)
{
    if (!check_playback_sanity (self, error)) return FALSE;

    bmp_system_message_send (self, BMP_SYSTEM_MESSAGE_CURRENT, NULL, NULL);
    return TRUE;
}

/* BmpPlay callbacks/handlers */

static void
on_play_eos (BmpPlay	      *play,
	     BmpSystemControl *control)
{
    bmp_system_message_send (control, BMP_SYSTEM_MESSAGE_NEXT, NULL, NULL);
}

static void
on_play_playstatus (BmpPlay	      *play,
		    GParamSpec	      *spec,
		    BmpSystemControl  *self)
{
    bmp_system_control_update_title (self);

    g_signal_emit (G_OBJECT (self),
		   signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_PLAYSTATUS],
		   0, g_object_get_int (bmp_play_engine, "playstatus"));
}

static void	  
on_play_seek	  (BmpPlay	      *play,
		   gint		       offset,
		   BmpSystemControl   *self)
{
    g_signal_emit (G_OBJECT (self),
		   signals[BMP_SYSTEM_CONTROL_SIGNAL_SEEK_EVENT],
		   0, offset); 
}

void
bmp_system_control_app_busy (BmpSystemControl *self)
{
#ifdef HAVE_GUI
    BmpWindowPlaylist *w_playlist;

    if (!bmp_ui)
      return;

    if (!bmp_rm)
      return;

    w_playlist = RM_GET_ITEM("windows","window-playlist",NOLOCK);

    if (w_playlist)
      bmp_window_set_busy (w_playlist->window);	
#endif
}

void bmp_system_control_app_idle (BmpSystemControl *self)
{
#ifdef HAVE_GUI
    BmpWindowPlaylist *w_playlist;

    if (!bmp_ui)
      return;

    if (!bmp_rm)
      return;

    w_playlist = RM_GET_ITEM("windows","window-playlist",NOLOCK);

    if (w_playlist)
      bmp_window_playlist_set_cursors (w_playlist);	
#endif
}

/* GObject stuff */

/* Proxy BmpPlaylist signals */
static void
playlist_callback_tracklist_list_sorted (BmpPlaylist	    *playlist, 
					 gint		     column,
					 BmpSystemControl   *self) 
{
    /* FIXME: Why do we need this? */
    if (signals[BMP_SYSTEM_CONTROL_SIGNAL_TRACKLIST_LIST_SORTED] > 0)
        g_signal_emit (G_OBJECT (self),
                   signals[BMP_SYSTEM_CONTROL_SIGNAL_TRACKLIST_LIST_SORTED], 0, column);
}

static void
playlist_callback_tracklist_items_removed (BmpPlaylist	    *playlist, 
					   GArray	    *removed_rows,  
					   gint		     n_removed_rows,
					   BmpSystemControl *self)
{
    /* FIXME: Why do we need this? */
    if (signals[BMP_SYSTEM_CONTROL_SIGNAL_TRACKLIST_ITEMS_REMOVED] > 0)
      {
        g_signal_emit (G_OBJECT (self),
                   signals[BMP_SYSTEM_CONTROL_SIGNAL_TRACKLIST_ITEMS_REMOVED],
                   0, removed_rows, n_removed_rows);

	bmp_system_control_update_title (self);
      }
}

static void
playlist_callback_tracklist_items_added (BmpPlaylist	  *playlist, 
					 GArray		  *added_rows,  
				         gint		   n_added_rows,
					 BmpSystemControl *self) 
{

    /* FIXME: Why do we need this? */
    if (signals[BMP_SYSTEM_CONTROL_SIGNAL_TRACKLIST_ITEMS_ADDED] > 0)
      {
	gint offset;

        g_signal_emit (G_OBJECT (self),
                   signals[BMP_SYSTEM_CONTROL_SIGNAL_TRACKLIST_ITEMS_ADDED],
                   0, added_rows, n_added_rows);
#ifdef HAVE_GUI
	if (c_tracklist)	
	  {
	    g_object_get (G_OBJECT(c_tracklist), "offset", &offset, NULL);
	    if (offset > 0)
	      chroma_list_scroll_to_offset (CHROMA_LIST(c_tracklist), offset);
	  }
#endif
      }
}

gboolean
bmp_system_control_clear_playback_history (BmpSystemControl *self, GError **error)
{
    g_return_val_if_fail (BMP_IS_SYSTEM_CONTROL (self), FALSE);

    bmp_playback_history_free (self->private->playback_history);
    self->private->playback_history = bmp_playback_history_new();

    bmp_plugin_flow_reset_state (BMP_PLUGIN_FLOW(self->private->flow_plugin));
    return TRUE;
}

/* User Messaging system */
typedef struct _BmpUserMessageDomain BmpUserMessageDomain;

struct _BmpUserMessageDomain {
	gint	 id;
	gchar	*name_id;
	gchar   *description;
};

gint
bmp_system_control_message_domain_register (BmpSystemControl	    *self,
					    const gchar		    *id,
					    const gchar		    *description)
{
    BmpUserMessageDomain *domain;
    gint new_id;

    new_id = self->private->msg_domain_id;
    self->private->msg_domain_id++;

    domain = g_new0 (BmpUserMessageDomain, 1);

    domain->id = new_id;
    domain->name_id = g_strdup (id);
    domain->description = g_strdup (description);
   
    self->private->message_domains = g_list_append (self->private->message_domains, domain); 

    return new_id;
}

gint
bmp_system_control_message_dispatch	    (BmpSystemControl	    *self,
					     gint		     domain_id,
					     GtkMessageType	     msgtype,
					     const gchar	    *text)
{
    BmpUserMessageDomain *domain;

#ifdef HAVE_GUI
    GtkWidget		 *dialog;
    gint		  response;
#endif

    domain = g_list_nth_data (self->private->message_domains, domain_id);

#ifdef HAVE_GUI
    if (msgtype == GTK_MESSAGE_QUESTION)
	{
	    dialog = gtk_message_dialog_new (NULL,
					     GTK_DIALOG_MODAL,
					     msgtype,
					     GTK_BUTTONS_YES_NO,
					     "");
					     
	}
    else
	{
	    dialog = gtk_message_dialog_new (NULL,
					     GTK_DIALOG_MODAL,
					     msgtype,
					     GTK_BUTTONS_OK,
					     "");
	}

    gtk_message_dialog_set_markup (GTK_MESSAGE_DIALOG(dialog), text);
    gtk_widget_realize (dialog);
    gtk_window_set_title (GTK_WINDOW(dialog), domain->description);
    gtk_window_set_position (GTK_WINDOW(dialog), GTK_WIN_POS_CENTER);

    response = gtk_dialog_run (GTK_DIALOG(dialog));
    gtk_widget_destroy (dialog);

    if (msgtype != GTK_MESSAGE_QUESTION) // Pointless if the GUI isn't running
	{
#endif

	    static gchar *m_strings[] = {
		"INFO",
		"WARNING",
		"QUESTION",
		"ERROR"
	    };

	    g_log (G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "%s: [%s] from '%s': %s", 
						    G_STRLOC,
						    m_strings[msgtype],
						    domain->description,
						    text);
#ifdef HAVE_GUI
	}

    return response;
#else
    return 0;
#endif
}

void
bmp_system_control_rows_swapped (BmpSystemControl *self,
				 gint row_a, 
				 gint row_b) 
{
    g_return_if_fail (BMP_IS_SYSTEM_CONTROL (self));

#ifdef HAVE_DBUS
    if (_no_remote) return;

    g_signal_emit (G_OBJECT (self),
		   signals[BMP_SYSTEM_CONTROL_SIGNAL_TRACKLIST_ROWS_SWAPPED],
		   0, row_a, row_b); 
#endif
}

static void
bmp_system_control_init (BmpSystemControl *self)
{
    self->private = g_new (BmpSystemControlPrivate, 1);
    self->private->dispose_has_run = FALSE;

    self->private->message_queue = g_async_queue_new ();
    self->private->message_queue_source_id = 0; 

    self->private->volume = xcs_key_fetch_int (xcs, "bmp", "volume");

    self->private->bitrate = 0;
    self->private->samplerate = 0;

    self->private->moving = FALSE; //FIXME

    self->private->repeat = FALSE;
    self->private->shuffle = FALSE;

    self->private->current_index = -1;
    self->private->current_title = NULL;
    self->private->current_item = NULL;

    self->private->stop_after_current = FALSE;
    self->private->anonymous = FALSE;
    self->private->playback_history = bmp_playback_history_new();
    self->private->flow_plugin = g_hash_table_lookup (plugins_table[BMP_PLUGIN_TYPE_FLOW], "linear");

    self->private->remote_queries = g_hash_table_new (g_direct_hash, g_direct_equal);
    self->private->ticket_id = 0;

    self->private->message_domains = NULL;
    self->private->msg_domain_id = 0;

    g_assert (self->private->flow_plugin != NULL);

    g_object_connect (G_OBJECT(xcs),
		      "swapped-signal::changed::bmp-titlestring-format",
		      G_CALLBACK(bmp_system_control_update_title),
		      self,
		      NULL);

    self->private->sc_msg_domain_id =
	bmp_system_control_message_domain_register (self,
						    "bmp", 
						    "BMP");

    g_message ("running: %s", G_OBJECT_TYPE_NAME(self));
}

#ifdef HAVE_DBUS
void
bmp_system_control_initialize_dbus_connection	(BmpSystemControl *self)
{
    DBusGConnection *bus;
    DBusGProxy *bus_proxy;
    gboolean connected = FALSE;
    guint request_name_result;
    GError *error = NULL;

    bus = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
    if (!bus)
	{
	    g_message ("running: %s (DBUS Error: '%s')", G_OBJECT_TYPE_NAME(self), error->message);
	    g_error_free (error);
	    return;
	}

    bus_proxy =
      dbus_g_proxy_new_for_name (bus,
				 "org.freedesktop.DBus",
				 "/org/freedesktop/DBus",
				 "org.freedesktop.DBus");

    if (!dbus_g_proxy_call (bus_proxy, "RequestName", &error,
			    G_TYPE_STRING, BMP_DBUS_INTERFACE,
			    G_TYPE_UINT, 0,
			    G_TYPE_INVALID,
			    G_TYPE_UINT, &request_name_result,
			    G_TYPE_INVALID))
	{
	    g_error ("%s: Failed RequestName request: %s", G_STRFUNC, error->message);
	    g_error_free (error);
	    error = NULL;
	}

    switch (request_name_result)
	{
	case DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER:
		{
		    dbus_g_connection_register_g_object (bus, BMP_DBUS_PATH,
							 G_OBJECT (self));
		    connected = TRUE;
		}
	    break;

	case DBUS_REQUEST_NAME_REPLY_EXISTS:
		{
		    g_message
		      ("%s: BMPx session already running, not registering SystemControl object to DBus.",
		       G_STRFUNC);
		    connected = FALSE;
		    proceed = FALSE;
		}
	    break;
	}

    if (connected)
	{
	    g_message ("%s (DBus:%s)", G_OBJECT_TYPE_NAME(self), BMP_DBUS_INTERFACE);
	}
}
#endif

static void
bmp_system_control_set_property (GObject * object,
				 guint property_id,
				 const GValue * value, GParamSpec * pspec)
{
    BmpSystemControl *self = BMP_SYSTEM_CONTROL (object);

    switch (property_id)
	{

	case BMP_SYSTEM_CONTROL_PROPERTY_CURRENT_INDEX:
		{
		    GtkListStore	  *list;
		    GtkTreePath		  *path;
		    GtkTreeRowReference	  *reference;
		    gint		   index = g_value_get_int (value);

		    list = bmp_playlist_get_playlist (bmp_playlist);
		    self->private->current_index = index;

		    reference = g_object_get_data (G_OBJECT(list), "current");  
		    gtk_tree_row_reference_free (reference);
		    g_object_set_data (G_OBJECT(list), "current", NULL); 

		    if (index != -1)
		      {
			path = gtk_tree_path_new_from_indices (index, -1);
			reference = gtk_tree_row_reference_new (GTK_TREE_MODEL(list), path);
		  	g_object_set_data (G_OBJECT(list), "current", reference); 
		      }
		}
	    break;

	case BMP_SYSTEM_CONTROL_PROPERTY_STREAM_POS:
		{
		    self->private->stream_pos = g_value_get_int (value);
		    g_signal_emit (self,
				   signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_STREAM_POS],
				   0, self->private->stream_pos);
		}
	    break;

	case BMP_SYSTEM_CONTROL_PROPERTY_BITRATE:
		{
		    self->private->bitrate = g_value_get_int (value);
		}
	    break;

	case BMP_SYSTEM_CONTROL_PROPERTY_SAMPLERATE:
		{
		    self->private->samplerate = g_value_get_int (value);
		}
	    break;

	case BMP_SYSTEM_CONTROL_PROPERTY_VOLUME:
		{
		    self->private->volume = g_value_get_int (value);
		    bmp_play_set_volume (bmp_play_engine, self->private->volume);
		    g_signal_emit (self,
				   signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_VOLUME],
				   0, self->private->volume);
		}
	    break;

	case BMP_SYSTEM_CONTROL_PROPERTY_MOVING:
		{
		    self->private->moving = g_value_get_boolean (value);
		}
	    break;

	case BMP_SYSTEM_CONTROL_PROPERTY_STOP_AFTER_CURRENT:
		{
		    self->private->stop_after_current = g_value_get_boolean (value);
#ifdef HAVE_GUI
		    GtkToggleAction *action;

		    action =
		      GTK_TOGGLE_ACTION (gtk_action_group_get_action
					 (bmp_toggle_actions,
					  BMP_TOGGLE_ACTION_STOP_AFTER_CURRENT));

		    if (gtk_toggle_action_get_active (action) !=
			self->private->stop_after_current)
			gtk_toggle_action_set_active (action,
						      self->private->stop_after_current);
#endif

		    g_signal_emit (self,
				   signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_STOP_AFTER_CURRENT],
				   0, self->private->stop_after_current); 
		}
	    break;


	case BMP_SYSTEM_CONTROL_PROPERTY_REPEAT:
		{
		    self->private->repeat = g_value_get_boolean (value);
#ifdef HAVE_GUI
		    GtkToggleAction *action;

		    action =
		      GTK_TOGGLE_ACTION (gtk_action_group_get_action
					 (bmp_toggle_actions,
					  BMP_TOGGLE_ACTION_REPEAT));

		    gtk_toggle_action_set_active (action, self->private->repeat); 
#endif
		    xcs_key_set_boolean (xcs, "bmp", "repeat", self->private->repeat);
		    g_signal_emit (self,
				   signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_REPEAT],
				   0, self->private->repeat); 
		}
	    break;

	case BMP_SYSTEM_CONTROL_PROPERTY_SHUFFLE:
		{
		    self->private->shuffle = g_value_get_boolean (value);
#ifdef HAVE_GUI
		    GtkToggleAction *action;

		    action =
		      GTK_TOGGLE_ACTION (gtk_action_group_get_action
					 (bmp_toggle_actions,
					  BMP_TOGGLE_ACTION_SHUFFLE));

		    gtk_toggle_action_set_active (action, self->private->shuffle);
						     
#endif
		    self->private->flow_plugin = g_hash_table_lookup (plugins_table[BMP_PLUGIN_TYPE_FLOW], 
						 self->private->shuffle ? "shuffle" : "linear");

		    xcs_key_set_boolean (xcs, "bmp", "shuffle", self->private->shuffle);

		    g_signal_emit (self,
				   signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_SHUFFLE],
				   0, self->private->shuffle); 
		}
	    break;

	default:
	    /* We don't have any other property... */
	    g_assert (FALSE);
	    break;
	}
}

static void
bmp_system_control_get_property (GObject * object,
				 guint property_id,
				 GValue * value, GParamSpec * pspec)
{
    BmpSystemControl *self = BMP_SYSTEM_CONTROL (object);

    switch (property_id)
	{

	case BMP_SYSTEM_CONTROL_PROPERTY_CURRENT_INDEX:
		{
		    g_value_set_int (value, self->private->current_index);
		}
	    break;

	case BMP_SYSTEM_CONTROL_PROPERTY_STREAM_POS:
		{
		    g_value_set_int (value, self->private->stream_pos);
		}
	    break;

	case BMP_SYSTEM_CONTROL_PROPERTY_BITRATE:
		{
		    g_value_set_int (value, self->private->bitrate);
		}
	    break;

	case BMP_SYSTEM_CONTROL_PROPERTY_SAMPLERATE:
		{
		    g_value_set_int (value, self->private->samplerate);
		}
	    break;


	case BMP_SYSTEM_CONTROL_PROPERTY_VOLUME:
		{
		    g_value_set_int (value, self->private->volume);
		}
	    break;

	case BMP_SYSTEM_CONTROL_PROPERTY_MOVING:
		{
		    g_value_set_boolean (value, self->private->moving);
		}
	    break;

	case BMP_SYSTEM_CONTROL_PROPERTY_REPEAT:
		{
		    g_value_set_boolean (value, self->private->repeat);
		}
	    break;

	case BMP_SYSTEM_CONTROL_PROPERTY_SHUFFLE:
		{
		    g_value_set_boolean (value, self->private->shuffle);
		}
	    break;

	case BMP_SYSTEM_CONTROL_PROPERTY_CURRENT_TITLE:
		{
		    g_value_set_string (value, self->private->current_title);
		}
	    break;

	default:
	    /* We don't have any other property... */
	    g_assert (FALSE);
	    break;
	}
}

static GObject *
bmp_system_control_constructor (GType type,
				guint n_construct_properties,
				GObjectConstructParam * construct_properties)
{
    GObject *obj;

	{
	    /* Invoke parent constructor. */
	    BmpSystemControlClass *klass;
	    GObjectClass *parent_class;
	    klass =
	      BMP_SYSTEM_CONTROL_CLASS (g_type_class_peek
					(BMP_TYPE_SYSTEM_CONTROL));
	    parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
	    obj = parent_class->constructor (type,
					     n_construct_properties,
					     construct_properties);
	}

    /* do stuff. */

    return obj;
}

static void
bmp_system_control_dispose (GObject * obj)
{
    BmpSystemControl *self = (BmpSystemControl *) obj;

    if (self->private->dispose_has_run) return;
    self->private->dispose_has_run = TRUE;

    g_message ("stopped: %s", G_OBJECT_TYPE_NAME(self));
}

static void
bmp_system_control_finalize (GObject * obj)
{
    BmpSystemControl *self = (BmpSystemControl *) obj;

    /*
     * Here, complete object destruction.
     * You might not need to do much...
     */
    g_free (self->private);
}

//FIXME: This doesn't really work out
static void
on_play_set_title (BmpPlay *play, gchar *text, BmpSystemControl *self)
{
    if (self->private->current_title) g_free (self->private->current_title);

    self->private->current_title = g_strdup (text);
    g_signal_emit (self, signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_TITLE], 0);
}

BmpSystemControl *
bmp_system_control_new (gboolean no_remote)
{
    BmpSystemControl *self;

#ifdef HAVE_DBUS
    if (!no_remote)
	dbus_g_object_type_install_info (BMP_TYPE_SYSTEM_CONTROL,
					 &dbus_glib_bmp_system_control_object_info);
    _no_remote = no_remote;
#endif

    self = g_object_new (bmp_system_control_get_type (), NULL);

    /* Proxy BmpPlaylist signals */
    g_object_connect(G_OBJECT(bmp_playlist),
		"signal::tracklist-list-sorted",
		G_CALLBACK(playlist_callback_tracklist_list_sorted),
	 	self,
		"signal::tracklist-items-removed",
		G_CALLBACK(playlist_callback_tracklist_items_removed),
		self,
		"signal::tracklist-items-added",
		G_CALLBACK(playlist_callback_tracklist_items_added),
		self,
		NULL);

    if (!proceed)
	return NULL;
    else
	return self;
}

void
bmp_system_control_init_finalize (BmpSystemControl *self)
{
    /* BmpPlay stuff */
    g_object_connect(G_OBJECT(bmp_play_engine),
		"signal::eos",
		G_CALLBACK(on_play_eos),
		self,
		"signal::notify::playstatus",
		G_CALLBACK(on_play_playstatus),
		self,
		"signal::title",
		G_CALLBACK(on_play_set_title),
	 	self,
		"signal::seek",
		G_CALLBACK(on_play_seek),
	 	self,
		NULL);
}

static void
bmp_system_control_class_init (BmpSystemControlClass * g_class)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (g_class);

    gobject_class->set_property = bmp_system_control_set_property;
    gobject_class->get_property = bmp_system_control_get_property;
    gobject_class->dispose = bmp_system_control_dispose;
    gobject_class->finalize = bmp_system_control_finalize;
    gobject_class->constructor = bmp_system_control_constructor;

    g_object_class_install_property (gobject_class,
				     BMP_SYSTEM_CONTROL_PROPERTY_BITRATE,
			             g_param_spec_int ("bitrate",
						       "bitrate",
						       "Current stream's bitrate.",
							0, G_MAXINT, 0, G_PARAM_READWRITE));

    g_object_class_install_property (gobject_class,
				     BMP_SYSTEM_CONTROL_PROPERTY_SAMPLERATE,
			             g_param_spec_int ("samplerate",
						       "samplerate",
						       "Current stream's samplerate.",
							0, G_MAXINT, 0, G_PARAM_READWRITE));


    g_object_class_install_property (gobject_class,
				     BMP_SYSTEM_CONTROL_PROPERTY_VOLUME,
				     g_param_spec_int ("volume",
						       "volume",
						       "Current volume.", 
							0, 100, 0, G_PARAM_READWRITE));

    g_object_class_install_property (gobject_class,
				     BMP_SYSTEM_CONTROL_PROPERTY_STREAM_POS,
			             g_param_spec_int ("stream_pos",
						       "stream_pos",
						       "Current stream position.",
							0, G_MAXINT, 0, G_PARAM_READWRITE));

    g_object_class_install_property (gobject_class,
				     BMP_SYSTEM_CONTROL_PROPERTY_CURRENT_TITLE,
				     g_param_spec_string ("current_title",
							  "current_title",
							  "Current title",
							   NULL, G_PARAM_READABLE));

    g_object_class_install_property (gobject_class,
				     BMP_SYSTEM_CONTROL_PROPERTY_CURRENT_URI,
			             g_param_spec_string ("current_uri",
							  "current_uri",
							  "Current URI",
							   NULL, G_PARAM_READABLE));

    g_object_class_install_property (gobject_class,
				     BMP_SYSTEM_CONTROL_PROPERTY_CURRENT_INDEX,
			             g_param_spec_int    ("current_index",
							  "current_index",
							  "Current track index",
							   -1, G_MAXINT, 0, G_PARAM_READWRITE));

    g_object_class_install_property (gobject_class,
				     BMP_SYSTEM_CONTROL_PROPERTY_MOVING,
			             g_param_spec_boolean ("moving",
							   "moving",
							   "Windows moving.", 
							   FALSE, G_PARAM_READWRITE));

    g_object_class_install_property (gobject_class,
				     BMP_SYSTEM_CONTROL_PROPERTY_REPEAT,
			             g_param_spec_boolean ("repeat",
							   "repeat",
							   "Player repeat state",
							   FALSE, G_PARAM_READWRITE));

    g_object_class_install_property (gobject_class,
				     BMP_SYSTEM_CONTROL_PROPERTY_STOP_AFTER_CURRENT,
				     g_param_spec_boolean ("stop_after_current",
							   "stop_after_current",
							   "Stop after current track",
						           FALSE, G_PARAM_READWRITE));

    g_object_class_install_property (gobject_class,
				     BMP_SYSTEM_CONTROL_PROPERTY_SHUFFLE,
				     g_param_spec_boolean ("shuffle",
							   "shuffle",
							   "Player shuffle state",
							   FALSE, G_PARAM_READWRITE));

    signals[BMP_SYSTEM_CONTROL_SIGNAL_SEEK_EVENT] =
      g_signal_new ("seek-event",
		    G_OBJECT_CLASS_TYPE (gobject_class),
		    G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
		    G_STRUCT_OFFSET (BmpSystemControlClass, seek_event),
		    NULL, NULL, g_cclosure_marshal_VOID__INT,
		    G_TYPE_NONE, 1, G_TYPE_INT);

    signals[BMP_SYSTEM_CONTROL_SIGNAL_STARTUP_COMPLETE] =
      g_signal_new ("startup-complete",
		    G_OBJECT_CLASS_TYPE (gobject_class),
		    G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
		    G_STRUCT_OFFSET (BmpSystemControlClass, startup_complete),
		    NULL, NULL, g_cclosure_marshal_VOID__VOID,
		    G_TYPE_NONE, 0);

    signals[BMP_SYSTEM_CONTROL_SIGNAL_SHUTDOWN_COMPLETE] =
      g_signal_new ("shutdown-complete",
		    G_OBJECT_CLASS_TYPE (gobject_class),
		    G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
		    G_STRUCT_OFFSET (BmpSystemControlClass, shutdown_complete),
		    NULL, NULL, g_cclosure_marshal_VOID__VOID,
		    G_TYPE_NONE, 0);

    signals[BMP_SYSTEM_CONTROL_SIGNAL_SHUTDOWN_REQUEST] =
      g_signal_new ("shutdown-request",
		    G_OBJECT_CLASS_TYPE (gobject_class),
		    0,
		    G_STRUCT_OFFSET (BmpSystemControlClass, shutdown_request),
		    NULL, NULL, bmpx_BOOLEAN__VOID,
		    G_TYPE_BOOLEAN, 0);

    signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_VOLUME] =
      g_signal_new ("set-volume",
		    G_OBJECT_CLASS_TYPE (gobject_class),
		    G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
		    G_STRUCT_OFFSET (BmpSystemControlClass, set_volume),
		    NULL, NULL, g_cclosure_marshal_VOID__INT,
		    G_TYPE_NONE, 1, G_TYPE_INT);

    signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_SHUFFLE] =
      g_signal_new ("set-shuffle",
		    G_OBJECT_CLASS_TYPE (gobject_class),
		    G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
		    G_STRUCT_OFFSET (BmpSystemControlClass, set_shuffle),
		    NULL, NULL, g_cclosure_marshal_VOID__BOOLEAN,
		    G_TYPE_NONE, 1, G_TYPE_BOOLEAN);

    signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_EQ] =
      g_signal_new ("set-eq",
		    G_OBJECT_CLASS_TYPE (gobject_class),
		    G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
		    G_STRUCT_OFFSET (BmpSystemControlClass, set_eq),
		    NULL, NULL, bmpx_VOID__INT_INT,
		    G_TYPE_NONE, 2, G_TYPE_INT, G_TYPE_INT);

    signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_REPEAT] =
      g_signal_new ("set-repeat",
		    G_OBJECT_CLASS_TYPE (gobject_class),
		    G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
		    G_STRUCT_OFFSET (BmpSystemControlClass, set_repeat),
		    NULL, NULL, g_cclosure_marshal_VOID__BOOLEAN,
		    G_TYPE_NONE, 1, G_TYPE_BOOLEAN);

    signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_PLAYSTATUS] =
      g_signal_new ("set-playstatus",
		    G_OBJECT_CLASS_TYPE (gobject_class),
		    G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
		    G_STRUCT_OFFSET (BmpSystemControlClass, set_playstatus),
		    NULL, NULL, g_cclosure_marshal_VOID__INT,
		    G_TYPE_NONE, 1, G_TYPE_INT);

    signals[BMP_SYSTEM_CONTROL_SIGNAL_TRACK_CHANGE] =
      g_signal_new ("track-change",
		    G_OBJECT_CLASS_TYPE (gobject_class),
		    G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
		    G_STRUCT_OFFSET (BmpSystemControlClass, track_change),
		    NULL, NULL, g_cclosure_marshal_VOID__VOID,
		    G_TYPE_NONE, 0);


    signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_STREAM_POS] =
      g_signal_new ("set-stream-pos",
		    G_OBJECT_CLASS_TYPE (gobject_class),
		    G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
		    G_STRUCT_OFFSET (BmpSystemControlClass, set_stream_pos),
		    NULL, NULL, g_cclosure_marshal_VOID__INT,
		    G_TYPE_NONE, 1, G_TYPE_INT);

    signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_TITLE] =
      g_signal_new ("set-title",
		    G_OBJECT_CLASS_TYPE (gobject_class),
		    G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
		    G_STRUCT_OFFSET (BmpSystemControlClass, set_title),
		    NULL, NULL, g_cclosure_marshal_VOID__VOID,
		    G_TYPE_NONE, 0);


    signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_STOP_AFTER_CURRENT] =
      g_signal_new ("set-stop-after-current",
		    G_OBJECT_CLASS_TYPE (gobject_class),
		    G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
		    G_STRUCT_OFFSET (BmpSystemControlClass,
				     set_stop_after_current), NULL, NULL,
		    g_cclosure_marshal_VOID__BOOLEAN, G_TYPE_NONE, 1,
		    G_TYPE_BOOLEAN);

    signals[BMP_SYSTEM_CONTROL_SIGNAL_TRACKLIST_LIST_SORTED] =
        g_signal_new ("tracklist-list-sorted",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      G_SIGNAL_RUN_FIRST,
                      G_STRUCT_OFFSET (BmpSystemControlClass, tracklist_list_sorted),
                      NULL, NULL, g_cclosure_marshal_VOID__INT, G_TYPE_NONE,
                      1, G_TYPE_INT);

    signals[BMP_SYSTEM_CONTROL_SIGNAL_TRACKLIST_ROWS_SWAPPED] =
        g_signal_new ("tracklist-rows-swapped",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      G_SIGNAL_RUN_FIRST,
                      G_STRUCT_OFFSET (BmpSystemControlClass, tracklist_rows_swapped),
                      NULL, NULL, bmpx_VOID__INT_INT, G_TYPE_NONE,
                      2, G_TYPE_INT, G_TYPE_INT);

#ifdef HAVE_DBUS
    if (_no_remote)
#endif
     {
	signals[BMP_SYSTEM_CONTROL_SIGNAL_TRACKLIST_ITEMS_REMOVED] =
	    g_signal_new ("tracklist-items-removed",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      G_SIGNAL_RUN_FIRST,
                      G_STRUCT_OFFSET (BmpSystemControlClass, tracklist_items_removed),
                      NULL, NULL, bmpx_VOID__POINTER_INT, G_TYPE_NONE,
                      2, G_TYPE_POINTER, G_TYPE_INT);


	signals[BMP_SYSTEM_CONTROL_SIGNAL_TRACKLIST_ITEMS_ADDED] =
	    g_signal_new ("tracklist-items-added",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      G_SIGNAL_RUN_FIRST,
                      G_STRUCT_OFFSET (BmpSystemControlClass, tracklist_items_added),
                      NULL, NULL, bmpx_VOID__POINTER_INT, G_TYPE_NONE,
                      2, G_TYPE_POINTER, G_TYPE_INT);
    }
#ifdef HAVE_DBUS
  else
    {
	signals[BMP_SYSTEM_CONTROL_SIGNAL_TRACKLIST_ITEMS_REMOVED] =
	    g_signal_new ("tracklist-items-removed",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      G_SIGNAL_RUN_FIRST,
                      G_STRUCT_OFFSET (BmpSystemControlClass, tracklist_items_removed),
                      NULL, NULL, bmpx_VOID__POINTER_INT, G_TYPE_NONE,
                      2, DBUS_TYPE_G_INT_ARRAY, G_TYPE_INT);


	signals[BMP_SYSTEM_CONTROL_SIGNAL_TRACKLIST_ITEMS_ADDED] =
	    g_signal_new ("tracklist-items-added",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      G_SIGNAL_RUN_FIRST,
                      G_STRUCT_OFFSET (BmpSystemControlClass, tracklist_items_added),
                      NULL, NULL, bmpx_VOID__POINTER_INT, G_TYPE_NONE,
                      2, DBUS_TYPE_G_INT_ARRAY, G_TYPE_INT);
    }
#endif
}

BmpPlaybackHistory*
bmp_system_control_get_history (BmpSystemControl *self)
{
    return self->private->playback_history;
}
