/*  BMPx - The Dumb Music Player
 *  Copyright (C) 2005 BMPx development team.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * The BMPx project hereby grant permission for non-gpl compatible GStreamer
 * plugins to be used and distributed together with GStreamer and BMPx. This
 * permission are above and beyond the permissions granted by the GPL license
 * BMPx is covered by.
 */


#include <stdlib.h>
#include <glib/gi18n.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <glade/glade.h>
#include <string.h>

#include "system_control.h"
#include "resource_manager.h"

#include "ui.h"
#include "ui_util.h"

#include "main.h"

#include <bmp/dbus.h>
#include <bmp/util.h>
#include <bmp/uri.h>
#include <bmp/playlist.h>
#include <bmp/metadata.h>

#include <xcs/xml.h>
#include <xcs/xcs.h>
#include <goa/libgoa.h>

#include "glade.h"
#include "loader.h"
#include "dnd.h"
#include "wm.h"

#include "embedded-images/media-pixbufs.h"

#include "ui_dialog_library.h"
#include "signal_handler_pool.h"

/* GObject stuff */
G_DEFINE_TYPE (BmpLibrary, bmp_library, G_TYPE_OBJECT)

struct _BmpLibraryPrivate {

    gboolean	       dispose_has_run;

    gboolean	       adding;

    GladeXML	      *xml;
    SignalHandlerPool *pool;

    GtkTreeModel      *tree_store_browse_filter; 
    GtkTreeModel      *tree_store_browse; 
    GtkTreeModel      *tree_store_tags;

    GHashTable	      *map_root_browse;
    GHashTable	      *map_root_tags;
    GHashTable	      *map_root_playlists; 

    GdkPixbuf	      *pb[BMP_N_TAGS];
    GdkPixbuf	      *pb_starred[BMP_N_TAGS];

    gboolean	       shown; 
    gboolean	       running;

    gboolean	       clear;
    gboolean	       start_playback;
    gint	       start_playback_track;

    gboolean	       drag_inside;
};

static gchar* page_names[] = {
	"tree_view_browse",
	"tree_view_tags"
};
	
static void
window_library_create (BmpLibrary *this);

static gboolean
on_system_control_shutdown_request    (BmpSystemControl	  *control,
				       BmpLibrary	  *this)
{
    GtkWidget *dialog;

    if (!this->private->adding) return FALSE;

    dialog = gtk_message_dialog_new (GTK_WINDOW(this->window),
				     GTK_DIALOG_MODAL,
				     GTK_MESSAGE_INFO,
				     GTK_BUTTONS_OK,
				     _("BMP is currently adding files to the library and can't shut down right now.\n\nPlease either forcibly abort the adding process or wait until it's finished."));
				     
    gtk_window_set_title (GTK_WINDOW(dialog), _("Music Library: Adding Files - BMP"));

    gtk_dialog_run (GTK_DIALOG(dialog));
    gtk_widget_destroy (dialog);

    return TRUE;
}

BmpLibrary*
bmp_library_new (void)
{
    BmpLibrary *this;

    this = g_object_new (BMP_TYPE_LIBRARY, NULL);
    return this;
}

static GObject*
bmp_library_constructor (GType type,
	                   guint n_construct_properties,
         	           GObjectConstructParam * construct_properties)
{
    GObject *obj;

    {
        BmpLibraryClass *klass;
        GObjectClass *parent_class;
        klass = BMP_LIBRARY_CLASS (g_type_class_peek (BMP_TYPE_LIBRARY));
        parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
        obj =
            parent_class->constructor (type,
                                       n_construct_properties,
                                       construct_properties);
    }

    return obj;
}

static void
bmp_library_dispose (GObject *object)
{
    BmpLibrary *this = (BmpLibrary *)object;

    if (this->private->tree_store_browse)
      g_object_unref (this->private->tree_store_browse);

    if (this->private->tree_store_browse_filter)
      g_object_unref (this->private->tree_store_browse_filter);

    if (this->private->tree_store_tags)
      g_object_unref (this->private->tree_store_tags);

    if (this->private->map_root_browse)
      g_hash_table_destroy (this->private->map_root_browse);

    if (this->private->map_root_tags)
      g_hash_table_destroy (this->private->map_root_tags);

    if (this->private->map_root_playlists)
      g_hash_table_destroy (this->private->map_root_playlists); 

    signal_handler_pool_free (this->private->pool);
    g_object_unref (this->private->xml);
    gtk_widget_destroy (this->window);
}

static void
bmp_library_class_init (BmpLibraryClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

#if 0
    gobject_class->set_property = bmp_play_set_property;
    gobject_class->get_property = bmp_play_get_property;
    gobject_class->finalize = bmp_play_finalize;
#endif
    gobject_class->dispose	= bmp_library_dispose;
    gobject_class->constructor	= bmp_library_constructor;

}

static void
bmp_library_init (BmpLibrary *this)
{
    this->private = g_new (BmpLibraryPrivate, 1);

    this->private->dispose_has_run = FALSE;

    this->private->adding = FALSE;

    this->private->pool = signal_handler_pool_new ("UiLibrary");

    this->private->tree_store_browse	    = NULL; 
    this->private->tree_store_browse_filter = NULL; 
    this->private->map_root_browse	    = NULL;

    this->private->tree_store_tags	    = NULL;
    this->private->map_root_tags	    = NULL;

    this->private->map_root_playlists	    = NULL; 

    this->private->shown		= FALSE; 
    this->private->running		= FALSE; 

    this->private->clear		= FALSE;
    this->private->start_playback	= FALSE;
    this->private->start_playback_track = 0; 

    this->private->pb[0] = gdk_pixbuf_new_from_inline (-1, icon_artist, FALSE, NULL);
    this->private->pb[1] = gdk_pixbuf_new_from_inline (-1, icon_album,  FALSE, NULL);
    this->private->pb[2] = gdk_pixbuf_new_from_inline (-1, icon_title,  FALSE, NULL);
    this->private->pb_starred[0] = gdk_pixbuf_new_from_inline (-1, icon_artist_starred, FALSE, NULL);
    this->private->pb_starred[1] = gdk_pixbuf_new_from_inline (-1, icon_album_starred,  FALSE, NULL);
    this->private->pb_starred[2] = gdk_pixbuf_new_from_inline (-1, icon_title_starred,  FALSE, NULL);

    window_library_create (this);
}

#if 0
static GtkTargetEntry dnd_entries[] = {
    {"application/x-bmpx-db-query", 0, BMP_DROP_QUERY},
};
#endif

enum {
      BROWSE_COL_IMG,
      BROWSE_COL_DISP,
      BROWSE_COL_KEY,
      BROWSE_COL_URI,
      BROWSE_COL_MAP,
      BROWSE_COL_KEYS,
      BROWSE_COL_TRACK,
      BROWSE_COL_LENGTH,
      BROWSE_COL_GENRE,
      BROWSE_COL_BITRATE,
};

static gint browse_levels[] = {
    BMP_DATUM_ARTIST,
    BMP_DATUM_ALBUM,
    BMP_DATUM_TITLE
};

enum {
    TAG_COL_PIXBUF, 
    TAG_COL_NAME,
    TAG_COL_MAP,
    TAG_COL_TUPLE,
    TAG_COL_DATUM
};

enum {
    TAB_BROWSE,
    TAB_TAGS
};

#if 0
static void
ui_dialog_library_playlist_new_from_file (GtkAction	  *action,
					  BmpLibrary	  *this);
#endif

#if 0
static GtkActionEntry bmp_actions_LIBRARY[] = {
  { "dummy-plmanager", NULL, "dummy" },

#if 0
  { BMP_ACTION_LIBRARY_PLAYLIST_NEW_FROM_FILE,
    GTK_STOCK_OPEN,
    N_("New Playlist from File"), "",
    N_("New Playlist from File"),
    G_CALLBACK(ui_dialog_library_playlist_new_from_file)},

  { BMP_ACTION_LIBRARY_PLAYLIST_NEW_FROM_LIBRARY,
    GTK_STOCK_FIND,
    N_("New Playlist from Library"), "",
    N_("New Playlist from Library"),
    G_CALLBACK(NULL)}
#endif

};
#endif

static gboolean
tree_store_browse_visible_func (GtkTreeModel	*model,
                                GtkTreeIter	*iter,
                                BmpLibrary	*this);


static gchar*
tuple_get_elem_value (HTuple	      *tuple,
		      BmpMetadatumId   metadatum_id) 
{
    HTupleElem	  *elem;
    gchar	  *key;

    elem = h_tuple_get (tuple, bmp_metadata_get_id_static (metadatum_id));

    if ((elem->value) && (strlen(elem->value)>0))
      {
	    key = g_strdup (elem->value);
	    g_strstrip (key);
      }
    else
      {
	    key = g_strdup (bmp_metadata_get_default_char(metadatum_id));
      }

    return key;
}
	
static void
browse_view_insert_tuple (BmpMetadataCache	  *cache,
			  HTuple		  *tuple,
			  BmpLibrary		  *this)
{
    HTupleElem	  *elem;
    GtkTreeIter	   iter[BMP_N_TAGS];
    GHashTable*	   map[BMP_N_TAGS];
    GHashTable*	   map_c;
    gint	   n;
    gchar	  *key;
    gchar	  *filename;

    elem = h_tuple_get (tuple, bmp_metadata_get_id_static (BMP_METADATA_PKEY));

    if (!g_uri_test (elem->value, BMP_URI_TEST_PROTOCOL, PROTOCOL_FILE)) return;

    filename = g_filename_from_uri (elem->value, NULL, NULL);
    if (!g_file_test (filename, G_FILE_TEST_EXISTS))
	{
	    g_free (filename);
	    return;
	}
    g_free (filename);

    map[0] = this->private->map_root_browse;
    map_c = map[0];

    for (n = 0; n < BMP_N_TAGS; n++)
      {
	key = tuple_get_elem_value (tuple, browse_levels[n]); 
	elem = h_tuple_get (tuple, bmp_metadata_get_id_static (BMP_METADATA_PKEY));
	if (!attr_row_map_mapping_exists (map_c, key))
          {	
	      GHashTable *keys;

	      if (n == 0)
		  gtk_tree_store_append (GTK_TREE_STORE(this->private->tree_store_browse), &iter[n], NULL); 
	      else
		  gtk_tree_store_append (GTK_TREE_STORE(this->private->tree_store_browse), &iter[n], &iter[n-1]); 

	      keys = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);

	      gtk_tree_store_set (GTK_TREE_STORE(this->private->tree_store_browse), 
			     &iter[n], 
			     BROWSE_COL_KEY, 
			     (n == 2) ? elem->value : key, 
			     BROWSE_COL_DISP,
			     key,
			     BROWSE_COL_URI, 
			     elem->value, 
			     BROWSE_COL_MAP, 
			     attr_row_map_new (),
			     BROWSE_COL_KEYS,
			     keys, 
			     -1); 

	      attr_row_map_mapping_insert (map_c, (n == 2) ? elem->value : key, GTK_TREE_MODEL(this->private->tree_store_browse), &iter[n]);

	      if (n < 2) 
		{
		    HTupleElem	*elem_local;
		    gchar	*value;

		    elem_local = h_tuple_get (tuple, bmp_metadata_get_id_static (browse_levels[n+1])); 
		    value = g_strdup (elem_local->value);
		    g_strstrip (value);
		    g_hash_table_replace (keys, value, GINT_TO_POINTER(1)); 
		}
	  }
	else
	  {
	    if (!attr_row_map_mapping_get (map_c, key, GTK_TREE_MODEL(this->private->tree_store_browse), &iter[n])) 
	      {
		  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "%s: Mapping for key '%p' disappeared", G_STRLOC, key); 
		  g_free (key);
		  break;
	      }

	    if (n == 2)
	      {
		    //We check if the URI is identical too
		    gchar  *uri;
		    gtk_tree_model_get (GTK_TREE_MODEL(this->private->tree_store_browse),
				  &iter[n],
				  BROWSE_COL_URI,
				  &uri,
				  -1);

		    if (strcmp((char *)elem->value, uri))
		      {
			  gtk_tree_store_append (GTK_TREE_STORE(this->private->tree_store_browse), &iter[n], &iter[n-1]); 
			  gtk_tree_store_set (GTK_TREE_STORE(this->private->tree_store_browse), 
			     &iter[n], 
			     BROWSE_COL_KEY, 
			     elem->value, 
			     BROWSE_COL_DISP,
			     key,
			     BROWSE_COL_URI, 
			     elem->value, 
			     BROWSE_COL_MAP, 
			     attr_row_map_new (),
			     -1); 

			  attr_row_map_mapping_insert (map_c, elem->value, GTK_TREE_MODEL(this->private->tree_store_browse), &iter[n]);
		      }
	      }
	    else if (n < 2) 
	      {
		    HTupleElem	*elem_local;
		    GHashTable  *keys;
		    gchar	*value;

		    gtk_tree_model_get (GTK_TREE_MODEL(this->private->tree_store_browse),
				  &iter[n],
				  BROWSE_COL_KEYS,
				  &keys,
				  -1);

		    elem_local = h_tuple_get (tuple, bmp_metadata_get_id_static (browse_levels[n+1])); 
		    value = g_strdup (elem_local->value);
		    g_strstrip (value);
		    g_hash_table_replace (keys, value, GINT_TO_POINTER(1)); 
	      }
	  }

	g_free (key);
	gtk_tree_model_get (GTK_TREE_MODEL(this->private->tree_store_browse), &iter[n], BROWSE_COL_MAP, &map_c, -1); 
    }
}

G_LOCK_DEFINE(DELETE_TUPLE);

static void
browse_view_delete_tuple (BmpMetadataCache	  *cache,
			  HTuple		  *tuple,
			  BmpLibrary		  *this)
{
    HTupleElem	  *elem;
    GtkTreeIter	   iter[BMP_N_TAGS];
    GHashTable*	   map[BMP_N_TAGS];
    gchar	  *key[BMP_N_TAGS];

    G_LOCK (DELETE_TUPLE);

    elem = h_tuple_get (tuple, bmp_metadata_get_id_static (BMP_METADATA_PKEY));

    if (g_ascii_strncasecmp (elem->value, "file:", 5))
	{
	    G_UNLOCK (DELETE_TUPLE);
	    return;
	}

    map[0] = this->private->map_root_browse;

    key[0] = tuple_get_elem_value (tuple, browse_levels[0]); 
#if 0
    g_message ("%s: key[0]: %s", G_STRLOC, key[0]);
#endif
    if (!attr_row_map_mapping_get (map[0], key[0], this->private->tree_store_browse, &iter[0]))
	{
	    g_free (key[0]);
	    G_UNLOCK (DELETE_TUPLE);
	    return;
	}

    gtk_tree_model_get (GTK_TREE_MODEL(this->private->tree_store_browse), &iter[0], BROWSE_COL_MAP, &map[1], -1); 
    key[1] = tuple_get_elem_value (tuple, browse_levels[1]); 
#if 0
    g_message ("%s: key[1]: %s", G_STRLOC, key[1]);
#endif
    if (!attr_row_map_mapping_get (map[1], key[1], this->private->tree_store_browse, &iter[1]))
	{
	    g_free (key[0]);
	    g_free (key[1]);
	    G_UNLOCK (DELETE_TUPLE);
	    return;
	}

    gtk_tree_model_get (GTK_TREE_MODEL(this->private->tree_store_browse), &iter[1], BROWSE_COL_MAP, &map[2], -1); 
    elem = h_tuple_get (tuple, bmp_metadata_get_id_static (BMP_METADATA_PKEY));
    key[2] = g_strdup (elem->value);
#if 0
    g_message ("%s: key[2]: %s", G_STRLOC, key[2]);
#endif

    if (!attr_row_map_mapping_get (map[2], key[2], this->private->tree_store_browse, &iter[2]))
	{
	    g_free (key[0]);
	    g_free (key[1]);
	    g_free (key[2]);
	    G_UNLOCK (DELETE_TUPLE);
	    return;
	}

    gtk_tree_store_remove (GTK_TREE_STORE(this->private->tree_store_browse), &iter[2]);
    attr_row_map_mapping_remove_nodestroy (map[2], key[2]);

    if (gtk_tree_model_iter_n_children (this->private->tree_store_browse, &iter[1]) == 0)
      {
	attr_row_map_mapping_remove_nodestroy (map[1], key[1]);
	gtk_tree_store_remove (GTK_TREE_STORE(this->private->tree_store_browse), &iter[1]);
      }

    if (gtk_tree_model_iter_n_children (GTK_TREE_MODEL(this->private->tree_store_browse), &iter[0]) == 0)
      {
	attr_row_map_mapping_remove_nodestroy (map[0], key[0]);
	gtk_tree_store_remove (GTK_TREE_STORE(this->private->tree_store_browse), &iter[0]);
      }

    g_free (key[0]);
    g_free (key[1]);
    g_free (key[2]);

    G_UNLOCK (DELETE_TUPLE);
}

static void
browse_view_insert_tuple_proxy_func (HTuple		  *tuple,
				     BmpLibrary		  *this)	      
{
    browse_view_insert_tuple (bmp_metadata_cache, tuple, this);
    while (gtk_events_pending()) gtk_main_iteration ();
}

static gint
tree_store_browse_sort_func (GtkTreeModel *model,
			     GtkTreeIter  *a,
			     GtkTreeIter  *b,
			     gpointer      user_data)
{
    HTuple *tuple_a, *tuple_b;
    GtkTreePath *path_a, *path_b;
    gchar *uri_a, *uri_b;
    gchar *key_a, *_key_a;
    gchar *key_b, *_key_b;
    gint cmp_result = 0;

    gtk_tree_model_get (model, a, BROWSE_COL_DISP, &_key_a, BROWSE_COL_URI, &uri_a, -1);
    gtk_tree_model_get (model, b, BROWSE_COL_DISP, &_key_b, BROWSE_COL_URI, &uri_b, -1);

    tuple_a = bmp_metadata_cache_get_cached (bmp_metadata_cache, uri_a);
    tuple_b = bmp_metadata_cache_get_cached (bmp_metadata_cache, uri_b);
    g_free (uri_a);
    g_free (uri_b);

    key_a = g_utf8_casefold (_key_a, strlen (_key_a));
    key_b = g_utf8_casefold (_key_b, strlen (_key_b));

    path_a =
      gtk_tree_model_get_path (model, a);

    path_b =
      gtk_tree_model_get_path (model, b);

    if (path_b && (gtk_tree_path_get_depth (path_b) == 3))
      {
	HTupleElem *elem_a, *elem_b;
	gint track_a, track_b;

	if ((!tuple_a) || (!tuple_b))
	  return 0;

	elem_a = h_tuple_get (tuple_a, bmp_metadata_get_id_static (BMP_DATUM_TRACK));
	elem_b = h_tuple_get (tuple_b, bmp_metadata_get_id_static (BMP_DATUM_TRACK));
	track_a = GPOINTER_TO_INT (elem_a->value);
	track_b = GPOINTER_TO_INT (elem_b->value);

	gtk_tree_path_free (path_a);
	gtk_tree_path_free (path_b);

	if (track_a  < track_b)
	  cmp_result = -1;
  
	if (track_a == track_b)
	  cmp_result =  0;

	if (track_a >  track_b)
	  cmp_result =  1;

	  goto done;
      }
    else 
      {
	if (path_a)
	  gtk_tree_path_free (path_a);

	if (path_b)
	  gtk_tree_path_free (path_b);
      }

    if ((key_a != NULL) && (key_b == NULL))
      {
	cmp_result = 1;
	goto done;
      }

    if ((key_a == NULL) && (key_b == NULL))
      {
	cmp_result = 0;
	goto done;
      }

    if ((key_a == NULL) && (key_b != NULL))
      {
	cmp_result = -1;
	goto done;
      }
    
    cmp_result = g_utf8_collate (key_a, key_b);

    done:

    h_tuple_free (tuple_a);
    h_tuple_free (tuple_b);

    return cmp_result;
}

static void
on_tag_added (BmpMetadataCache *cache,
	      gchar	       *tagname,
	      BmpLibrary       *this)
{
    GtkTreeIter	   iter, iter_child;
    GdkPixbuf	  *pb_bookmark;
    GHashTable	  *mapping_sub;
    gint	   n;

    pb_bookmark = gdk_pixbuf_new_from_file (DATA_DIR "/images/icon-bookmark.png", NULL);
    gtk_tree_store_append (GTK_TREE_STORE(this->private->tree_store_tags), &iter, NULL);
    attr_row_map_mapping_insert (this->private->map_root_tags, tagname, GTK_TREE_MODEL(this->private->tree_store_tags), &iter);

    mapping_sub = attr_row_map_new (); 
    gtk_tree_store_set (GTK_TREE_STORE(this->private->tree_store_tags), &iter, 0, pb_bookmark, 1, tagname, 2, mapping_sub, -1);

    for (n = 0; n < BMP_N_TAGS; n++) 
      {
	GHashTable  *mapping_values;
	gchar	    *name;

	gtk_tree_store_append (GTK_TREE_STORE(this->private->tree_store_tags), &iter_child, &iter);
	name = bmp_metadata_get_title_plural (bmp_metadata_tag_id_to_datum(n));
	attr_row_map_mapping_insert (mapping_sub, name, GTK_TREE_MODEL(this->private->tree_store_tags), &iter_child);

	mapping_values = attr_row_map_new ();
	gtk_tree_store_set (GTK_TREE_STORE(this->private->tree_store_tags), &iter_child, 
			      TAG_COL_PIXBUF, 
			      this->private->pb_starred[n], 
			      TAG_COL_NAME,
			      name, 
			      TAG_COL_MAP,
			      mapping_values, 
			      -1); 
      }
}

static void
on_tag_deleted (BmpMetadataCache *cache,
	        gchar	         *tagname,
	        BmpLibrary	 *this)
{
    GtkTreeIter	   iter;
    attr_row_map_mapping_get (this->private->map_root_tags, tagname, GTK_TREE_MODEL(this->private->tree_store_tags), &iter);
    if (!gtk_tree_store_remove (GTK_TREE_STORE(this->private->tree_store_tags), &iter))
	{
	    //FIXME: Add some error handling here.. shouldn't probably happen ever anyway
	}

    gtk_tree_model_filter_refilter (GTK_TREE_MODEL_FILTER(this->private->tree_store_browse_filter));
}

static void
on_tag_value_insert (BmpMetadataCache	*cache,
		     const gchar	*tagname,
		     BmpMetadatumTagId   metadatum_tag_id,
		     TagDataTuple	*data_tuple, 
		     BmpLibrary		*this)
{
    GtkTreeIter	   iter, iter_child;
    GHashTable	  *mapping_sub, *mapping_tags;
    gchar	  *name;

    if (!attr_row_map_mapping_get (this->private->map_root_tags, tagname, GTK_TREE_MODEL(this->private->tree_store_tags), &iter))
      {
	g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "%s: Tag with name '%s' does not exist", G_STRLOC, tagname);
	return;
      }

    gtk_tree_model_get (GTK_TREE_MODEL(this->private->tree_store_tags), &iter, 2, &mapping_sub, -1);

    name = bmp_metadata_get_title_plural (bmp_metadata_tag_id_to_datum(metadatum_tag_id));

    if (!attr_row_map_mapping_get (mapping_sub, name, GTK_TREE_MODEL(this->private->tree_store_tags), &iter))
      {
	g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "%s: Tag Id %d does not exist", G_STRLOC, metadatum_tag_id);
	g_free (name);
	return;
      }
    g_free (name);

    gtk_tree_model_get (GTK_TREE_MODEL(this->private->tree_store_tags), &iter, TAG_COL_MAP, &mapping_tags, -1); 

    gtk_tree_store_append (GTK_TREE_STORE(this->private->tree_store_tags), &iter_child, &iter);
    attr_row_map_mapping_insert (mapping_tags, data_tuple->tag[metadatum_tag_id], GTK_TREE_MODEL(this->private->tree_store_tags), &iter_child);

    gtk_tree_store_set (GTK_TREE_STORE(this->private->tree_store_tags), &iter_child, 
			TAG_COL_PIXBUF, 
			this->private->pb[metadatum_tag_id], 
			TAG_COL_NAME, 
			data_tuple->tag[metadatum_tag_id], 
			TAG_COL_MAP, 
			NULL, 
			TAG_COL_TUPLE, 
			data_tuple,
			TAG_COL_DATUM,
			metadatum_tag_id,
			-1); 
}

static void
on_tag_value_remove (BmpMetadataCache	*cache,
		     const gchar	*tagname,
		     BmpMetadatumTagId   metadatum_tag_id,
		     TagDataTuple	*data_tuple, 
		     BmpLibrary		*this)
{
    GtkTreeIter	   iter, iter_child;
    GHashTable	  *mapping_sub, *mapping_tags;
    gchar	  *name;

    if (!attr_row_map_mapping_get (this->private->map_root_tags, tagname, GTK_TREE_MODEL(this->private->tree_store_tags), &iter))
      {
	g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "%s: Tag with name '%s' does not exist", G_STRLOC, tagname);
	return;
      }

    gtk_tree_model_get (GTK_TREE_MODEL(this->private->tree_store_tags), &iter, 2, &mapping_sub, -1);

    name = bmp_metadata_get_title_plural (bmp_metadata_tag_id_to_datum(metadatum_tag_id));

    if (!attr_row_map_mapping_get (mapping_sub, name, GTK_TREE_MODEL(this->private->tree_store_tags), &iter))
      {
	g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "%s: Tag Id %d does not exist", G_STRLOC, metadatum_tag_id);
	g_free (name);
	return;
      }
    g_free (name);

    gtk_tree_model_get (GTK_TREE_MODEL(this->private->tree_store_tags), &iter, TAG_COL_MAP, &mapping_tags, -1);

    if (attr_row_map_mapping_get (mapping_tags, data_tuple->tag[metadatum_tag_id],
        GTK_TREE_MODEL(this->private->tree_store_tags), &iter_child))

    gtk_tree_store_remove (GTK_TREE_STORE(this->private->tree_store_tags), &iter_child);
}

static void
on_music_remove (GtkWidget	  *widget, 
		 BmpLibrary	  *this)
{
    GtkWidget 	      *window;
    GtkWidget         *notebook;
    GtkWidget	      *tree_view;
    GtkTreeModel      *store;
    GtkTreeSelection  *selection;
    GList	      *paths, *uris;
    GtkTreeIter	       iter,
		       iter_children,
		       iter_children_children;
    gint	       current_page;

    window = glade_xml_get_widget_warn (this->private->xml, "dialog_library");
    notebook = glade_xml_get_widget_warn (this->private->xml, "notebook_library");
    current_page = gtk_notebook_get_current_page (GTK_NOTEBOOK(notebook));

    bmp_window_set_busy (GTK_WIDGET(window));
    switch (current_page)
      {
        case TAB_BROWSE: 
	  {
	    tree_view =
		glade_xml_get_widget_warn (this->private->xml, "tree_view_browse");
	    store =
		gtk_tree_view_get_model (GTK_TREE_VIEW(tree_view));
	    selection =
		gtk_tree_view_get_selection (GTK_TREE_VIEW(tree_view));
	    paths =
		gtk_tree_selection_get_selected_rows (selection, NULL);

	    paths = g_list_sort (paths, (GCompareFunc) gtk_tree_path_compare);

	    uris = NULL;

	    while (paths)
	      {
		gtk_tree_model_get_iter (store, &iter, paths->data);

	      switch (gtk_tree_path_get_depth (paths->data))
		  {
			case 1:
			  {
			    gboolean run_1 = TRUE, run_2 = TRUE;
			    run_1 = gtk_tree_model_iter_children (store, &iter_children, &iter);
			    while (run_1)
			      {
				run_2 = gtk_tree_model_iter_children (store, &iter_children_children, &iter_children);
				while (run_2)
				  {
				    gchar *uri;
				    gtk_tree_model_get (store, &iter_children_children, BROWSE_COL_URI, &uri, -1); 
				    uris = g_list_append (uris, uri);
				    run_2 = gtk_tree_model_iter_next (store, &iter_children_children);
				  }
				  run_1 = gtk_tree_model_iter_next (store, &iter_children);
			      }
			    break;
			  }

			case 2:
			  {
			    gboolean run = TRUE;
			    run = gtk_tree_model_iter_children (store, &iter_children, &iter);
			    while (run)
			      {
				gchar *uri;
				gtk_tree_model_get (store, &iter_children, BROWSE_COL_URI, &uri, -1); 
				uris = g_list_append (uris, uri);
				run = gtk_tree_model_iter_next (store, &iter_children);
			      }
			    break;
			  }
	
			case 3:
			  {
			    gchar *uri;
			    gtk_tree_model_get (store, &iter, BROWSE_COL_URI, &uri, -1); 
			    uris = g_list_append (uris, uri);
			    break;
			  }
			
			default: break; 
		  }
	      paths = paths->next;
	  }

	for ( ; uris ; uris = uris->next )
	  {
	      bmp_metadata_cache_metadata_remove (bmp_metadata_cache, uris->data); 
	  }

	g_list_foreach (paths, (GFunc)gtk_tree_path_free, NULL);
	break;
      }

    default: break;
  }

  bmp_window_set_idle (GTK_WIDGET(window));
}


static void
on_music_add (GtkWidget	    *widget,
	      BmpLibrary    *this) 
{
    GtkWidget	  *dialog, 
		  *notebook_library_add,
		  *window;

    GList	  *items = NULL;
    GError	  *error = NULL;
    gchar	  *folder;
    gint	   rv;

    window = glade_xml_get_widget_warn (this->private->xml, "dialog_library");
    dialog = gtk_file_chooser_dialog_new (_("Select Path To Add"),
                                          GTK_WINDOW (glade_xml_get_widget_warn (this->private->xml, "dialog_library")),
                                          GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER,
                                          GTK_STOCK_CANCEL,
                                          GTK_RESPONSE_CANCEL,
                                          GTK_STOCK_ADD,
                                          GTK_RESPONSE_OK,
                                          NULL);

    gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER(dialog),  
					 xcs_key_fetch (xcs, 
					 "bmp", 
					 "file-chooser-path"));

    rv = gtk_dialog_run (GTK_DIALOG (dialog));

    if (rv == GTK_RESPONSE_CANCEL)
      {
	gtk_widget_destroy (dialog);
	return;
      }

    folder = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER(dialog));
    gtk_widget_destroy (dialog);

    this->private->adding = TRUE;

    notebook_library_add = glade_xml_get_widget_warn (this->private->xml, "notebook_library_add");
    gtk_notebook_set_current_page (GTK_NOTEBOOK(notebook_library_add), 1);
    bmp_window_set_busy (GTK_WIDGET(window));

    dir_foreach (folder, (DirForeachFunc) collect_path, &items, &error);

    for ( ; items ; items = items->next)
      {
	gchar *uri;
	 
	uri = g_filename_to_uri (items->data, NULL, NULL);
	bmp_metadata_cache_metadata_cache (bmp_metadata_cache, uri);
	g_free (uri);

	while (gtk_events_pending()) gtk_main_iteration();
      }

    g_list_foreach (g_list_first(items), (GFunc)g_free, NULL);
    g_free (folder);

    gtk_notebook_set_current_page (GTK_NOTEBOOK(notebook_library_add), 0);
    bmp_window_set_idle (GTK_WIDGET(window));

    this->private->adding = FALSE;
}

static GtkTreeViewColumn* column_browse[4]; /* 0 for the main stuff, 1: BROWSE_COL_GENRE, 2: BROWSE_COL_BITRATE */
static GtkCellRenderer*   renderer_browse[6]; /* 0-2 for the main stuff, 3: BROWSE_COL_GENRE, 4: BROWSE_COL */

static GtkTreeViewColumn* column_tags[2]; 
static GtkCellRenderer*   renderer_tags[2];

static GtkTreeViewColumn* column_modify[1]; 
static GtkCellRenderer*   renderer_modify[2];

#if 0
static void 
query_create (GtkTreeModel *store, GtkTreePath *path, gchar **_name, gchar **_uri)
{
    GtkTreeIter  iter,
		 iter_parent;

    *_name = NULL;
    *_uri  = NULL;

    gint depth = gtk_tree_path_get_depth (path);
    gtk_tree_model_get_iter (store, &iter, path); 

    switch (depth)
	{
	    case 1:
	      {
			    HTuple	  *tuple;
			    HTupleElem	  *elem;
			    BMPURI	  *bmp_uri;
			    GString	  *query_uri;
			    gchar	  *uri;

			    query_uri = g_string_new ("query://");
			    g_string_append (query_uri, BMP_DBUS_INTERFACE); 
			    g_string_append (query_uri, "/"); 

			    gtk_tree_model_get (store, &iter, BROWSE_COL_URI, &uri, -1); 
			    tuple = bmp_metadata_cache_get_cached (bmp_metadata_cache, uri);
			    elem = h_tuple_get (tuple, bmp_metadata_get_id_static (browse_levels[depth-1]));
			    h_tuple_free (tuple);

			    g_string_append_printf (query_uri, "?%s=", bmp_metadata_get_id_static (browse_levels[depth-1])); 
			    g_string_append (query_uri, elem->value);

			    bmp_uri = bmp_uri_new (query_uri->str);
			    g_string_free (query_uri, TRUE);
			    bmp_uri_escape (bmp_uri);

			    *_name = g_strdup (elem->value);
			    *_uri  = bmp_uri_get_string (bmp_uri);
      
			    bmp_uri_delete (bmp_uri);
			    return;
	    } break;
 
	  case 2:
	    {
			    HTuple	  *tuple;
			    HTupleElem	  *elem;
			    GString	  *query_uri;
			    BMPURI	  *bmp_uri;
			    gchar	  *uri;
			    gchar	  *l1,
					  *l2;

			    gtk_tree_model_get (store, &iter, BROWSE_COL_URI, &uri, -1); 
			    tuple = bmp_metadata_cache_get_cached (bmp_metadata_cache, uri);

			    query_uri = g_string_new ("query://");
			    g_string_append (query_uri, BMP_DBUS_INTERFACE); 
			    g_string_append (query_uri, "/"); 

			    elem = h_tuple_get (tuple, bmp_metadata_get_id_static (browse_levels[depth-1]));
			    l2 = g_strdup (elem->value);

			    h_tuple_free (tuple);

			    gtk_tree_model_iter_parent (store, &iter_parent, &iter);
			    gtk_tree_model_get (store, &iter_parent, BROWSE_COL_URI, &uri, -1); 
			    tuple = bmp_metadata_cache_get_cached (bmp_metadata_cache, uri);

			    elem = h_tuple_get (tuple, bmp_metadata_get_id_static (browse_levels[depth-2]));
			    l1 = g_strdup (elem->value);

			    g_string_append_printf (query_uri, "?%s=", bmp_metadata_get_id_static (browse_levels[depth-2])); 
			    g_string_append (query_uri, l1);
			    g_string_append_printf (query_uri, "&%s=", bmp_metadata_get_id_static (browse_levels[depth-1])); 
			    g_string_append (query_uri, l2); 

			    h_tuple_free (tuple);

			    bmp_uri = bmp_uri_new (query_uri->str);
			    g_string_free (query_uri, TRUE);

			    bmp_uri_escape (bmp_uri);

			    *_name = g_strconcat (l1, " / ", l2, NULL);
			    *_uri  = bmp_uri_get_string (bmp_uri); 

			    bmp_uri_delete (bmp_uri);
			    return;
	    }
	 break;

	case 3:
	    {
			    HTuple	*tuple;
			    HTupleElem	*elem;
			    GString	*query_uri;
			    BMPURI      *bmp_uri;
			    gchar       *uri;
			    gchar	*l1,
					*l2,
					*l3;

			    query_uri = g_string_new ("query://");
			    g_string_append (query_uri, BMP_DBUS_INTERFACE); 
			    g_string_append (query_uri, "/"); 

			    gtk_tree_model_get (store, &iter, BROWSE_COL_URI, &uri, -1); 
			    tuple = bmp_metadata_cache_get_cached (bmp_metadata_cache, uri);
			    elem = h_tuple_get (tuple, bmp_metadata_get_id_static (browse_levels[depth-1]));
			    l3 = g_strdup (elem->value);

			    h_tuple_free (tuple);

			    gtk_tree_model_iter_parent (store, &iter_parent, &iter);
			    gtk_tree_model_get (store, &iter_parent, BROWSE_COL_URI, &uri, -1); 
			    tuple = bmp_metadata_cache_get_cached (bmp_metadata_cache, uri);
			    elem = h_tuple_get (tuple, bmp_metadata_get_id_static (browse_levels[depth-2]));
			    l2 = g_strdup (elem->value);

			    h_tuple_free (tuple);

			    gtk_tree_model_iter_parent (store, &iter, &iter_parent);
			    gtk_tree_model_get (store, &iter, BROWSE_COL_URI, &uri, -1); 
			    tuple = bmp_metadata_cache_get_cached (bmp_metadata_cache, uri);
			    elem = h_tuple_get (tuple, bmp_metadata_get_id_static (browse_levels[depth-3]));
			    l1 = g_strdup (elem->value);

			    h_tuple_free (tuple);

			    g_string_append_printf (query_uri, "?%s=", bmp_metadata_get_id_static (browse_levels[depth-3])); 
			    g_string_append (query_uri, l1);
			    g_string_append_printf (query_uri, "&%s=", bmp_metadata_get_id_static (browse_levels[depth-2])); 
			    g_string_append (query_uri, l2); 
			    g_string_append_printf (query_uri, "&%s=", bmp_metadata_get_id_static (browse_levels[depth-1])); 
			    g_string_append (query_uri, l3); 

			    bmp_uri = bmp_uri_new (query_uri->str);
			    g_string_free (query_uri, TRUE);

			    bmp_uri_escape (bmp_uri);

			    *_name = g_strconcat (l1, " / ", l2, " / ", l3, NULL);
			    *_uri  = bmp_uri_get_string (bmp_uri);
    
			    bmp_uri_delete (bmp_uri);
			    return;
	      } break;
			
	      default: break; 
	}
}
#endif

static void
on_button_tag_clicked (GtkButton	*button, 
		       BmpLibrary	*this)
{
    GladeXML	      *xml;
    GtkWidget	      *image_type,
		      *label_name,
		      *dialog_set_tags,
		      *library_notebook;

    GtkWidget*	       tree_view;
    GtkTreeSelection  *selection;
  
    GtkTreeModel      *store;
    GtkTreeIter	       iter;

    HTuple	      *tuple;
    HTupleElem	      *elem;

    GList	      *tags = NULL,
		      *paths = NULL;

    BmpMetadatumId     metadatum_id;
    BmpMetadatumTagId  metadatum_tag_id;

    gint	       current_page = -1;
    gint	       depth;

    gchar	      *uri;

    gboolean	       proceed = TRUE;

    xml = this->private->xml;
    library_notebook = glade_xml_get_widget_warn (xml, "notebook_library");
    dialog_set_tags = glade_xml_get_widget_warn (xml, "dialog_set_tags");

    current_page = gtk_notebook_get_current_page (GTK_NOTEBOOK(library_notebook));

    if (current_page != TAB_BROWSE)
	return; //Shouldn't happen

    tree_view =	glade_xml_get_widget_warn (xml, "tree_view_browse");
    store = gtk_tree_view_get_model (GTK_TREE_VIEW(tree_view));
    selection = gtk_tree_view_get_selection (GTK_TREE_VIEW(tree_view));
    paths = gtk_tree_selection_get_selected_rows (selection, NULL);

    if (!paths)
	return; //Shouldn't happen.

    depth = gtk_tree_path_get_depth (paths->data);

    gtk_tree_model_get_iter (store, &iter, paths->data);
    gtk_tree_model_get (store, &iter, BROWSE_COL_URI, &uri, -1);
    tuple = bmp_metadata_cache_get_cached (bmp_metadata_cache, uri);

    image_type = glade_xml_get_widget (xml, "image_type");
    gtk_image_set_from_pixbuf (GTK_IMAGE(image_type), this->private->pb[depth-1]);

    tree_view =	glade_xml_get_widget_warn (xml, "tree_view_modify_tags");
    store = GTK_TREE_MODEL(gtk_list_store_new (2, G_TYPE_BOOLEAN, G_TYPE_STRING));
    tags = bmp_metadata_cache_tags_list (bmp_metadata_cache);

    tags = g_list_sort (tags, (GCompareFunc)g_utf8_collate);

    metadatum_id = browse_levels[depth-1];
    metadatum_tag_id = bmp_metadata_datum_to_tag_id (metadatum_id);

    elem = h_tuple_get (tuple, bmp_metadata_get_id_static (metadatum_id));
    label_name = glade_xml_get_widget (xml, "label_name");
    gtk_label_set_text (GTK_LABEL(label_name), elem->value);

    for ( ; tags ; tags = tags->next)
      {
	  gboolean tagged;

	  gtk_list_store_append (GTK_LIST_STORE(store), &iter);

	  tagged = bmp_metadata_cache_tuple_matches_tag (bmp_metadata_cache,
							 tuple,
							 metadatum_tag_id,
							 (const gchar*)tags->data);

	  gtk_list_store_set (GTK_LIST_STORE(store), &iter, 0, tagged, 1, (gchar*)tags->data, -1);
      }

    gtk_tree_view_set_model (GTK_TREE_VIEW(tree_view), store);
    gtk_dialog_run (GTK_DIALOG(dialog_set_tags));

    proceed = gtk_tree_model_get_iter_first (store, &iter); 

    while (proceed)
      { 
	  GError   *error = NULL;
	  gchar	   *id;
	  gboolean  checked;

	  gtk_tree_model_get (store, &iter, 0, &checked, 1, &id, -1); 

	  if (checked)
	    {
		if (!bmp_metadata_cache_tag_tuple (bmp_metadata_cache,
						   tuple,
						   metadatum_tag_id,
						   id,
						   &error))
		  {
		    g_log (G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "%s: %s", G_STRLOC, error->message);
		    g_error_free (error);
		    error = NULL;
		  }
	    }
	  else
	    {
		if (!bmp_metadata_cache_untag_tuple (bmp_metadata_cache,
						   tuple,
						   metadatum_tag_id,
						   id,
						   &error))
		  {
		    g_log (G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "%s: %s", G_STRLOC, error->message);
		    g_error_free (error);
		    error = NULL;
		  }
	    }

	  proceed = gtk_tree_model_iter_next (store, &iter);
      }

    h_tuple_free (tuple);
    gtk_widget_hide (dialog_set_tags);
    gtk_tree_model_filter_refilter (GTK_TREE_MODEL_FILTER(this->private->tree_store_browse_filter));
}

static gchar*
get_query_sublevel (GtkTreeModel *store, GtkTreeIter *iter)
{
  GString	      *query_uri;
  BMPURI		      *uri;
  gchar		      *uri_str;
  TagDataTuple	      *data_tuple;
  BmpMetadatumTagId    metadatum_tag_id;

  gtk_tree_model_get (store, iter, TAG_COL_TUPLE, &data_tuple, TAG_COL_DATUM, &metadatum_tag_id, -1); 

  query_uri = g_string_new ("query://");

  g_string_append (query_uri, BMP_DBUS_INTERFACE);
  g_string_append (query_uri, "/"); 

  g_string_append_printf (query_uri, "?%s=%s", bmp_metadata_get_id_static (BMP_DATUM_ARTIST), data_tuple->tag[BMP_TAG_ARTIST]); 

  if (metadatum_tag_id > BMP_TAG_ARTIST)
      g_string_append_printf (query_uri, "&%s=%s", bmp_metadata_get_id_static (BMP_DATUM_ALBUM), data_tuple->tag[BMP_TAG_ALBUM]); 
  if (metadatum_tag_id > BMP_TAG_ALBUM)
      g_string_append_printf (query_uri, "&%s=%s", bmp_metadata_get_id_static (BMP_DATUM_TITLE), data_tuple->tag[BMP_TAG_TITLE]); 

  uri = bmp_uri_new (query_uri->str);
  bmp_uri_escape (uri);
  uri_str = bmp_uri_get_string (uri); 

  bmp_uri_delete (uri); 
  g_string_free (query_uri, TRUE);

  return uri_str; 
}

static void
append_to_uniquelist (HTuple	       *tuple,
		      BmpUniqueList   **ulist)
{
  HTupleElem *elem;

  elem = h_tuple_get (tuple, bmp_metadata_get_id_static (BMP_DATUM_LOCATION));
  bmp_unique_list_insert (*ulist, g_strdup(elem->value));
}

static void
on_button_enqueue_clicked (GtkButton	  *button, 
			   BmpLibrary	  *this)
{
    GtkWidget 	      *window;
    GtkWidget         *notebook;
    GtkWidget	      *tree_view;
    GtkTreeModel      *store;
    GtkTreeSelection  *selection;
    GList	      *paths = NULL;
    GList	      *uris = NULL;
    gchar	     **uri_list;
    GtkTreeIter	       iter, iter_children, iter_children_children;
    gint	       new_ctr = 0, current_page = -1;

    window = glade_xml_get_widget_warn (this->private->xml, "dialog_library");
    notebook = glade_xml_get_widget_warn (this->private->xml, "notebook_library");
    current_page = gtk_notebook_get_current_page (GTK_NOTEBOOK(notebook));

    bmp_window_set_busy (GTK_WIDGET(window));

    switch (current_page)
      {
	case TAB_TAGS:
	  {
	    GtkTreePath	*tree_path;

	    tree_view =
		glade_xml_get_widget_warn (this->private->xml, "tree_view_tags");
	    selection =
		gtk_tree_view_get_selection (GTK_TREE_VIEW(tree_view));

	    if (!gtk_tree_selection_get_selected (selection, &store, &iter))
	      goto enqueue_early_exit;

	    tree_path = gtk_tree_model_get_path (store, &iter);
	    switch (gtk_tree_path_get_depth(tree_path))
		{
		    case 1:
			{
			    // Here we cant use get_query_sublevel() because we must avoid dupes:
			    // An album might be basically tagged twice, once as the album explicitely,
			    // and secondly implicitly inside the artist of this album.
			    BmpUniqueList   *ulist;
			    gboolean	     proceed;
  
			    ulist = bmp_unique_list_new (G_TYPE_STRING);

			    //First descend into all subs
			    proceed = gtk_tree_model_iter_children (store, &iter_children, &iter);
			    while (proceed)
			      {
				    gboolean proceed_s = gtk_tree_model_iter_children (store, &iter_children_children, &iter_children);
				    while (proceed_s)
				      {
					  TagDataTuple	      *data_tuple;
					  Query		      *query;
					  HRelation	      *result;
					  BmpMetadatumTagId    metadatum_tag_id;

					  gtk_tree_model_get (store, &iter_children_children, TAG_COL_TUPLE, &data_tuple, TAG_COL_DATUM, &metadatum_tag_id, -1); 

					  query = bmp_metadata_cache_query_initialize (bmp_metadata_cache);
					  bmp_metadata_cache_query_append_string (bmp_metadata_cache,
										  query,
										  bmp_metadata_tag_id_to_datum (metadatum_tag_id),
										  TRUE,
										  data_tuple->tag[metadatum_tag_id]);

					  result = bmp_metadata_cache_query_run (bmp_metadata_cache,
										 query);   


				result = h_relation_sort (
					 h_relation_sort (
                                         h_relation_sort (result,
                                                          (GCompareDataFunc) h_tuple_compare_elem,
                                                          (gpointer)bmp_metadata_get_id_static (BMP_DATUM_TRACK)),
                                                 (GCompareDataFunc) h_tuple_compare_elem,
                                                 (gpointer)bmp_metadata_get_id_static (BMP_DATUM_ARTIST)),

                              (GCompareDataFunc) h_tuple_compare_elem,
                              (gpointer)bmp_metadata_get_id_static (BMP_DATUM_ALBUM));



					  h_relation_for_each_tuple (result, (HTupleFunc)append_to_uniquelist, &ulist);

					  h_relation_unref (result);
					  bmp_metadata_cache_query_free (bmp_metadata_cache, query);
					  
					  proceed_s = gtk_tree_model_iter_next (store, &iter_children_children);
				      }

				    proceed = gtk_tree_model_iter_next (store, &iter_children);
			      }
			    uri_list = glist_to_strv (bmp_unique_list_to_glist(ulist));
			    bmp_system_control_add_uri_list (bmp_system_control,
					    (const gchar**)uri_list,
					    -1,
					    this->private->clear,
					    this->private->start_playback,
					    this->private->start_playback_track,
					    &new_ctr,
					    NULL);
			    bmp_unique_list_free (ulist);
			}
			break;

		    case 2:
			{
			    gboolean proceed = TRUE;
  
			    proceed = gtk_tree_model_iter_children (store, &iter_children, &iter);

			    while (proceed)
			      {
				  uris = g_list_append (uris, get_query_sublevel (store, &iter_children)); 
				  proceed = gtk_tree_model_iter_next (store, &iter_children);
			      }

			    uri_list = glist_to_strv (uris);
			    bmp_system_control_add_uri_list (bmp_system_control,
					    (const gchar**)uri_list,
					    -1,
					    this->private->clear,
					    this->private->start_playback,
					    this->private->start_playback_track,
					    &new_ctr,
					    NULL);
			}
			break;

		    case 3:
			{
			    uris = g_list_append (uris, get_query_sublevel (store, &iter)); 
			    uri_list = glist_to_strv (uris);
			    bmp_system_control_add_uri_list (bmp_system_control,
					    (const gchar**)uri_list,
					    -1,
					    this->private->clear,
					    this->private->start_playback,
					    this->private->start_playback_track,
					    &new_ctr,
					    NULL);
			}
			break;
  
		    default: break;
		}

		g_list_foreach (uris, (GFunc)g_free, NULL);
		g_list_foreach (paths, (GFunc)gtk_tree_path_free, NULL);
		g_list_free (paths);
		g_list_free (uris);

	  }
	break;

        case TAB_BROWSE: 
	  {
	    tree_view =
		glade_xml_get_widget_warn (this->private->xml, "tree_view_browse");
	    store =
		gtk_tree_view_get_model (GTK_TREE_VIEW(tree_view));
	    selection =
		gtk_tree_view_get_selection (GTK_TREE_VIEW(tree_view));
	    paths =
		gtk_tree_selection_get_selected_rows (selection, NULL);

	    if (!paths)
	      {
		gtk_tree_selection_select_all (selection);
		paths =
		    gtk_tree_selection_get_selected_rows (selection, NULL);
	      }

	    while (paths)
	      {
		gtk_tree_model_get_iter (store, &iter, paths->data);

		switch (gtk_tree_path_get_depth (paths->data))
		  {
			case 1:
			  {
			    gboolean run_1 = TRUE, run_2 = TRUE;
			    run_1 = gtk_tree_model_iter_children (store, &iter_children, &iter);
			    while (run_1)
			      {
				run_2 = gtk_tree_model_iter_children (store, &iter_children_children, &iter_children);
				while (run_2)
				  {
				    gchar *uri;
				    gtk_tree_model_get (store, &iter_children_children, BROWSE_COL_URI, &uri, -1); 
				    uris = g_list_append (uris, uri);
				    run_2 = gtk_tree_model_iter_next (store, &iter_children_children);
				  }
				  run_1 = gtk_tree_model_iter_next (store, &iter_children);
			      }
			    break;
			  } 
  
			case 2:
			  {
			    gboolean run = TRUE;
			    run = gtk_tree_model_iter_children (store, &iter_children, &iter);
			    while (run)
			      {
				gchar *uri;
				gtk_tree_model_get (store, &iter_children, BROWSE_COL_URI, &uri, -1); 
				uris = g_list_append (uris, uri); 
				run = gtk_tree_model_iter_next (store, &iter_children);
			      }
			    break;
			  }
	
			case 3:
			  {
			    gchar *uri;
			    gtk_tree_model_get (store, &iter, BROWSE_COL_URI, &uri, -1); 
			    uris = g_list_append (uris, uri); 
			    break;
			  }

			default: break; 
		  }

	      paths = paths->next;
	  }

	uri_list = glist_to_strv (uris);
	bmp_system_control_add_uri_list (bmp_system_control,
					    (const gchar**)uri_list,
					    -1,
					    this->private->clear,
					    this->private->start_playback,
					    this->private->start_playback_track,
					    &new_ctr,
					    NULL);

	g_list_foreach (uris, (GFunc)g_free, NULL);
	g_list_foreach (paths, (GFunc)gtk_tree_path_free, NULL);
	g_list_free (paths);
	g_list_free (uris);
    }
    break;

    default: break;
  }

  enqueue_early_exit:

  this->private->clear = FALSE;
  this->private->start_playback = FALSE;
  this->private->start_playback_track = 0;

  bmp_window_set_idle (GTK_WIDGET(window));
}

//FIXME: Make this a little cleaner, perhaps?
static void
on_button_play_clicked (GtkButton	    *button,
			BmpLibrary	    *this)
{
    GladeXML  *xml = this->private->xml;

    this->private->clear = TRUE;
    this->private->start_playback = TRUE;
    this->private->start_playback_track = 0;

    on_button_enqueue_clicked (GTK_BUTTON(glade_xml_get_widget_warn (xml, "button_enqueue")), this);
}

static gboolean
iter_matches (BmpLibrary	*this,
	      GtkTreeModel	*model,
	      GtkTreeIter	*iter,
	      const gchar	*text,
	      BmpMetadatumId	 metadatum_id)
{
    HTuple	      *tuple;
    HTupleElem	      *elem;
    gchar	      *uri, *check;
    gboolean	       rval;
#ifdef HAVE_HAL
    gboolean	       is_present;
#endif

    rval  = TRUE;
    check = NULL;
    tuple = NULL;

    gtk_tree_model_get (model, iter, BROWSE_COL_URI, &uri, -1);
    if (!uri)
      {
	goto iter_matches_early_exit;
      }

    tuple = bmp_metadata_cache_get_cached (bmp_metadata_cache, uri);
    if (!tuple) 
      {
	goto iter_matches_early_exit;
      }

#ifdef HAVE_HAL
    if (!bmp_metadata_cache_check_tuple_volume	(bmp_metadata_cache,
						 tuple,
						 &is_present))
      {
	//FIXME: Error
      }
    else if (!is_present)
      {
	rval = FALSE;
	goto iter_matches_early_exit;
      }
#endif

    elem = h_tuple_get (tuple, bmp_metadata_get_id_static (metadatum_id));
    if (!elem->value)
      {
	rval = TRUE;
	goto iter_matches_early_exit;
      }

    //FIXME: Possible check for UINT
    if (bmp_metadata_get_type (metadatum_id) == G_TYPE_INT)
	check = g_strdup_printf ("%d", GPOINTER_TO_INT(elem->value));
    else
	check = g_strdup (elem->value);

    if ((!check) || !strlen(check))
      {
	rval = TRUE;
	goto iter_matches_early_exit;
      }

    if ((!text) || match_keys (check, text))
      {
	  rval = TRUE;
	  goto iter_matches_early_exit; 
      }
    else
      {
	  //If it doesn't match the keys, then we return FALSE regardless of the tag
	  rval = FALSE;
	  goto iter_matches_early_exit;
      }

    iter_matches_early_exit:
 
    if (check) g_free (check); 
    if (tuple) h_tuple_free (tuple);

    return rval;
}

static gboolean
match_entry 	(gpointer   _k,
		 gpointer   _v,
		 gpointer   _d)
{
    return match_keys ((gchar*)_k, (gchar*)_d);
}

static gboolean
map_find_title 	(gpointer   _k,
		 gpointer   _v,
		 gpointer   _d)
{
    GtkTreeRowReference *reference = (GtkTreeRowReference *)_v;
    GtkTreeModel	*model;
    GtkTreePath		*path;
    GHashTable		*keys;
    GtkTreeIter		 iter;
    gpointer		 match;
    gboolean	         rval;

    model = gtk_tree_row_reference_get_model (reference);
    path = gtk_tree_row_reference_get_path (reference);   
    gtk_tree_model_get_iter (model, &iter, path); 
    gtk_tree_path_free (path);

    gtk_tree_model_get (model, &iter, BROWSE_COL_KEYS, &keys, -1);	

    match = g_hash_table_find (keys, (GHRFunc) match_entry, _d);
    rval = (match != NULL) ? TRUE:FALSE;
    return rval; 
}

static gboolean
tree_store_browse_visible_func (GtkTreeModel  *model,
                                GtkTreeIter   *iter,
                                BmpLibrary    *this) 
{
    GtkWidget	      *cbox_metadatum_id,
		      *entry_value;
    GtkTreeModel      *model_cbox;
    GtkTreeIter	       iter_cbox;
    GtkTreePath	      *path;
    const gchar	      *text;
    gboolean	       rval;
    gint	       depth;
    BmpMetadatumId     metadatum_id;

    path =
	gtk_tree_model_get_path (model, iter);
    depth =
	gtk_tree_path_get_depth (path)-1;

    gtk_tree_path_free (path);  

    cbox_metadatum_id =
	glade_xml_get_widget_warn (this->private->xml, "cbox_metadatum_id");
    entry_value =
	glade_xml_get_widget_warn (this->private->xml, "entry_value");

    model_cbox =
	gtk_combo_box_get_model (GTK_COMBO_BOX(cbox_metadatum_id));

    rval  = TRUE;

    if (!gtk_combo_box_get_active_iter (GTK_COMBO_BOX(cbox_metadatum_id), &iter_cbox))
      {
	goto visible_func_early_exit;
      }
    else
      {
	gtk_tree_model_get (model_cbox, &iter_cbox, 1, &metadatum_id, -1);
      }

    text = gtk_entry_get_text (GTK_ENTRY(entry_value));

    if ((metadatum_id == browse_levels[0]) && (depth == 0))
      { 
	  rval = iter_matches (this, model, iter, text, metadatum_id);
	  goto visible_func_early_exit;
      }

    if ((metadatum_id == browse_levels[1]) && (depth == 0))
      {
	  GHashTable  *keys;
	  gpointer     match;

	  gtk_tree_model_get (model, iter, BROWSE_COL_KEYS, &keys, -1);	
	  match = g_hash_table_find (keys, (GHRFunc) match_entry, (gchar *)text);
	  rval = (match != NULL) ? TRUE:FALSE;
	  goto visible_func_early_exit;
      }


    if ((metadatum_id == browse_levels[2]) && (depth == 1))
      {
	  GHashTable  *keys;
	  gpointer     match;

	  gtk_tree_model_get (model, iter, BROWSE_COL_KEYS, &keys, -1);	
	  match = g_hash_table_find (keys, (GHRFunc) match_entry, (gchar *)text);
	  rval = (match != NULL) ? TRUE:FALSE;
	  goto visible_func_early_exit;
      }

    if ((metadatum_id == browse_levels[2]) && (depth == 0))
      {
	  GHashTable	    *map;
	  gpointer	     match;

	  gtk_tree_model_get (model, iter, BROWSE_COL_MAP, &map, -1);	
	  match = g_hash_table_find (map, (GHRFunc) map_find_title, (gpointer)text); 
	  rval = (match != NULL) ? TRUE:FALSE;
	  goto visible_func_early_exit; 
      }

    rval = iter_matches (this, model, iter, text, metadatum_id);

    visible_func_early_exit:
    return rval;
}

static gboolean
on_dialog_library_key_press (GtkWidget	    *widget,
			     GdkEventKey    *event,
			     BmpLibrary	    *this)
{
    g_return_val_if_fail (GTK_IS_WIDGET (widget), FALSE);

    if (event->keyval == GDK_Escape)
      {
	gtk_widget_hide (widget);
	return TRUE;
      }

    if (event->keyval == GDK_F6)
      {
	gtk_widget_grab_focus (glade_xml_get_widget_warn (this->private->xml, "entry_value"));
	return TRUE;
      }

    if (event->keyval == GDK_F7)
      {
	GtkWidget *widget;
	gint page;

  	widget = glade_xml_get_widget_warn (this->private->xml, "notebook_library");
	page = gtk_notebook_get_current_page (GTK_NOTEBOOK(widget));

	gtk_tree_view_expand_all (GTK_TREE_VIEW(glade_xml_get_widget_warn (this->private->xml, page_names[page])));
	return TRUE;
      }

    if (event->keyval == GDK_F8)
      {
	GtkWidget *widget;
	gint page;

  	widget = glade_xml_get_widget_warn (this->private->xml, "notebook_library");
	page = gtk_notebook_get_current_page (GTK_NOTEBOOK(widget));

	gtk_tree_view_collapse_all (GTK_TREE_VIEW(glade_xml_get_widget_warn (this->private->xml, page_names[page])));
	return TRUE;
      }

    if (event->keyval == GDK_Down)
      {
	GtkWidget *entry, *widget;
	gint length, n;

  	entry = glade_xml_get_widget_warn (this->private->xml, "entry_value");
	if (!GTK_WIDGET_HAS_FOCUS(entry)) return FALSE; 

  	widget = glade_xml_get_widget_warn (this->private->xml, "cbox_metadatum_id");
	n = gtk_combo_box_get_active (GTK_COMBO_BOX(widget));
	length = gtk_tree_model_iter_n_children (gtk_combo_box_get_model (GTK_COMBO_BOX(widget)), NULL);
	if (!(n>=length))
	  {
	      n++;
	      gtk_combo_box_set_active (GTK_COMBO_BOX(widget), n);
	      return TRUE;
	  }
      }

    if (event->keyval == GDK_Up)
      {
	GtkWidget *entry, *widget;
	gint n;

  	entry = glade_xml_get_widget_warn (this->private->xml, "entry_value");
	if (!GTK_WIDGET_HAS_FOCUS(entry)) return FALSE; 

  	widget = glade_xml_get_widget_warn (this->private->xml, "cbox_metadatum_id");
	n = gtk_combo_box_get_active (GTK_COMBO_BOX(widget));
	if ((n-1)>=0)
	  {
	      n--;
	      gtk_combo_box_set_active (GTK_COMBO_BOX(widget), n);
	      return TRUE;
	  }
      }

    return FALSE;
}

void
length_renderer_func (GtkTreeViewColumn	  *tree_column,
                      GtkCellRenderer	  *cell,
                      GtkTreeModel	  *tree_model,
                      GtkTreeIter	  *iter,
                      gpointer		   data)
{
    HTuple *tuple;
    HTupleElem *elem;
    GtkTreePath *path;
    gchar *cell_data;

    path = gtk_tree_model_get_path (tree_model, iter);

    if (gtk_tree_path_get_depth (path) < 3)
      {
	g_object_set (G_OBJECT(cell), "visible", FALSE, NULL);
	g_object_set (G_OBJECT(cell), "text", "0", NULL);
      }
    else
      {
	gint	  ttime,
		  minutes,  
		  seconds;
	gchar	 *uri;

	gtk_tree_model_get (tree_model, iter, BROWSE_COL_URI, &uri, -1);
	tuple = bmp_metadata_cache_get_cached (bmp_metadata_cache, uri);
	if (tuple) 
	  {
	      g_object_set (G_OBJECT(cell), "visible", TRUE, NULL);
	      elem = h_tuple_get (tuple, bmp_metadata_get_id_static (BMP_DATUM_TIME));
	      ttime = GPOINTER_TO_INT(elem->value);
	      minutes = ttime / 60;
	      seconds = ttime % 60;
	      cell_data = g_strdup_printf ("%d:%-2.2d", minutes, seconds); 
	      g_object_set (G_OBJECT(cell), "text", cell_data, NULL);
	      g_free (cell_data);
	      h_tuple_free (tuple);
	  }
	else
	  {
	      g_object_set (G_OBJECT(cell), "text", "", NULL);
	  }
	g_free (uri);
      }
}

void
tracknumber_renderer_func (GtkTreeViewColumn  *tree_column,
                           GtkCellRenderer    *cell,
                           GtkTreeModel	      *tree_model,
                           GtkTreeIter	      *iter,
                           gpointer	       data)
{
    HTuple *tuple;
    HTupleElem *elem;
    GtkTreePath *path;
    gchar *cell_data;

    path = gtk_tree_model_get_path (tree_model, iter);

    if (gtk_tree_path_get_depth (path) < 3)
      {
	g_object_set (G_OBJECT(cell), "visible", FALSE, NULL);
	g_object_set (G_OBJECT(cell), "text", "0", NULL);
      }
    else
      {
	gchar *uri;

	g_object_set (G_OBJECT(cell), "visible", TRUE,  NULL);
	gtk_tree_model_get (tree_model, iter, BROWSE_COL_URI, &uri, -1);
	tuple = bmp_metadata_cache_get_cached (bmp_metadata_cache, uri);
	if (tuple)
	  {
	    elem = h_tuple_get (tuple, bmp_metadata_get_id_static (BMP_DATUM_TRACK));
	    cell_data = g_strdup_printf ("%d", GPOINTER_TO_INT(elem->value));
	    g_object_set (G_OBJECT(cell), "text", cell_data, NULL);
	    g_free (cell_data);
	    h_tuple_free (tuple);
	  }
	else
	  {
	    g_object_set (G_OBJECT(cell), "text", "", NULL);
	  }
	g_free (uri);
      }
}

void
bitrate_renderer_func (GtkTreeViewColumn  *tree_column,
                       GtkCellRenderer	  *cell,
                       GtkTreeModel	  *tree_model,
                       GtkTreeIter	  *iter,
                       gpointer		   data)
{
    HTuple *tuple;
    HTupleElem *elem;
    GtkTreePath *path;
    gchar *cell_data;

    path = gtk_tree_model_get_path (tree_model, iter);

    if (gtk_tree_path_get_depth (path) < 3)
      {
	g_object_set (G_OBJECT(cell), "visible", FALSE, NULL);
	g_object_set (G_OBJECT(cell), "text", "0", NULL);
      }
    else
      {
	gchar *uri;

	gtk_tree_model_get (tree_model, iter, BROWSE_COL_URI, &uri, -1);
	tuple = bmp_metadata_cache_get_cached (bmp_metadata_cache, uri);
	if (tuple)
	  {
	    g_object_set (G_OBJECT(cell), "visible", TRUE,  NULL);
	    elem = h_tuple_get (tuple, bmp_metadata_get_id_static (BMP_DATUM_BITRATE));
	    cell_data = g_strdup_printf ("%d", GPOINTER_TO_INT(elem->value));
	    g_object_set (G_OBJECT(cell), "text", cell_data, NULL);
	    g_free (cell_data);
	    h_tuple_free (tuple);
	  }
	else
	  {
	      g_object_set (G_OBJECT(cell), "text", "", NULL);
	  }
	g_free (uri);
      }
}

void
genre_renderer_func (GtkTreeViewColumn	*tree_column,
                     GtkCellRenderer	*cell,
                     GtkTreeModel	*tree_model,
                     GtkTreeIter	*iter,
                     gpointer	         data)
{
    HTuple *tuple;
    HTupleElem *elem;
    GtkTreePath *path;

    path = gtk_tree_model_get_path (tree_model, iter);

    if (gtk_tree_path_get_depth (path) < 3)
      {
	g_object_set (G_OBJECT(cell), "visible", FALSE, NULL);
	g_object_set (G_OBJECT(cell), "text", "0", NULL);
      }
    else
      {
	gchar *uri;

	gtk_tree_model_get (tree_model, iter, BROWSE_COL_URI, &uri, -1);
	tuple = bmp_metadata_cache_get_cached (bmp_metadata_cache, uri);
	if (tuple)
	  {
	    g_object_set (G_OBJECT(cell), "visible", TRUE,  NULL);
	    elem = h_tuple_get (tuple, bmp_metadata_get_id_static (BMP_DATUM_GENRE));
	    g_object_set (G_OBJECT(cell), "text", elem->value, NULL);
	    h_tuple_free (tuple);
	  }
	else
	  {
	      g_object_set (G_OBJECT(cell), "text", "", NULL);
	  }
	g_free (uri);
      }
}

static gboolean
on_dialog_library_delete (GtkWidget   *widget,
			  GdkEvent    *event,
                          gpointer     data)
{
  gtk_widget_hide (widget);
  return TRUE;
}

static void
on_library_notebook_switch_page (GtkNotebook	  *notebook,
				 GtkNotebookPage  *notebook_page,
				 guint		   page,
				 BmpLibrary	  *this)
{
    GtkTreeView	      *tree_view;
    GtkTreeSelection  *tree_selection;

    GladeXML	      *xml;

    GtkWidget	      *b_tag, 
		      *b_play, 
		      *b_enqueue,
		      *b_delete;

    GList	      *paths;

	
    g_return_if_fail (GTK_IS_NOTEBOOK(notebook));

    xml = this->private->xml; 

    b_tag = 
	glade_xml_get_widget_warn (xml, "button_tag");
    b_play = 
	glade_xml_get_widget_warn (xml, "button_play");
    b_enqueue = 
	glade_xml_get_widget_warn (xml, "button_enqueue");
    b_delete = 
	glade_xml_get_widget_warn (xml, "button_tag_delete");

    if (!strlen(page_names[page]))
      {
	  gtk_widget_set_sensitive (GTK_WIDGET(b_tag), FALSE); 
	  gtk_widget_set_sensitive (GTK_WIDGET(b_play), FALSE); 
	  gtk_widget_set_sensitive (GTK_WIDGET(b_enqueue), FALSE); 
	  return;
      }

    tree_view = GTK_TREE_VIEW(glade_xml_get_widget_warn (xml, page_names[page])); 
    tree_selection = gtk_tree_view_get_selection (tree_view);  

    paths = gtk_tree_selection_get_selected_rows (tree_selection, NULL);

    if (page == TAB_TAGS) 
      {
	  if (paths && (gtk_tree_path_get_depth(paths->data)>1))
	    gtk_widget_set_sensitive (b_delete, FALSE); 
	  else
	    gtk_widget_set_sensitive (b_delete, (gtk_tree_selection_count_selected_rows (tree_selection) != 0) ? TRUE:FALSE);

	  gtk_widget_set_sensitive (GTK_WIDGET(b_tag), FALSE); 
      }
    else
      {
	  if (g_list_length (paths)==1)
	      gtk_widget_set_sensitive (GTK_WIDGET(b_tag), TRUE); 
	  else
	      gtk_widget_set_sensitive (GTK_WIDGET(b_tag), FALSE); 
      }

#if 0
    if (paths && (page == TAB_TAGS) && (gtk_tree_path_get_depth(paths->data)<3))
      {
	gtk_widget_set_sensitive (GTK_WIDGET(b_play), FALSE);
	gtk_widget_set_sensitive (GTK_WIDGET(b_enqueue), FALSE); 
      }
    else
      {
#endif
	gtk_widget_set_sensitive (GTK_WIDGET(b_play), (!gtk_tree_selection_count_selected_rows (tree_selection)) ? FALSE : TRUE);
	gtk_widget_set_sensitive (GTK_WIDGET(b_enqueue), (!gtk_tree_selection_count_selected_rows (tree_selection)) ? FALSE : TRUE);
#if 0
      }      
#endif

    g_list_foreach (paths, (GFunc)gtk_tree_path_free, NULL);
}

static void
on_tree_selection_changed (GtkTreeSelection *tree_selection,
			   BmpLibrary	    *this)
{
    GladeXML	      *xml;

    enum {
	B_TAG,
	B_PLAY,
	B_ENQUEUE,
	N_BS
    };

    GtkWidget	      *b[N_BS];
    GtkWidget	      *notebook;
    GList	      *paths;

    xml = this->private->xml; 

    notebook =
	glade_xml_get_widget_warn (xml, "notebook_library");
    b[B_TAG] = 
	glade_xml_get_widget_warn (xml, "button_tag");
    b[B_PLAY] = 
	glade_xml_get_widget_warn (xml, "button_play");
    b[B_ENQUEUE] = 
	glade_xml_get_widget_warn (xml, "button_enqueue");

    paths = gtk_tree_selection_get_selected_rows (tree_selection, NULL);

    if (gtk_notebook_get_current_page (GTK_NOTEBOOK(notebook)) == TAB_BROWSE)
      {
	  GtkWidget *b_music_remove;
          b_music_remove = 
	      glade_xml_get_widget_warn (xml, "b_music_remove");
	  gtk_widget_set_sensitive (GTK_WIDGET(b_music_remove), (!gtk_tree_selection_count_selected_rows (tree_selection)) ? FALSE : TRUE);
      }

    if (g_list_length(paths) == 1)
	gtk_widget_set_sensitive (GTK_WIDGET(b[B_TAG]), TRUE); 
    else
	gtk_widget_set_sensitive (GTK_WIDGET(b[B_TAG]), FALSE); 

    gtk_widget_set_sensitive (GTK_WIDGET(b[B_PLAY]), (!gtk_tree_selection_count_selected_rows (tree_selection)) ? FALSE : TRUE);
    gtk_widget_set_sensitive (GTK_WIDGET(b[B_ENQUEUE]), (!gtk_tree_selection_count_selected_rows (tree_selection)) ? FALSE : TRUE);

    g_list_foreach (paths, (GFunc)gtk_tree_path_free, NULL);
}

#if 0
//FIXME:
static DNDQueryTuple *qtuple = NULL;

/* DnD API implemenations */
static GdkPixmap *
tv_browse_create_row_drag_icon (GtkWidget   *tree_view,
                                BmpLibrary  *this)
{
    GladeXML	      *xml;
    GtkTreeModel      *tree_model;
    GtkTreeSelection  *tree_selection;
    GtkTreePath	      *tree_path;

    PangoLayout	      *layout;
    GdkDrawable	      *drawable;
    cairo_t	      *cr;
    gint	       width;
    gint	       height;
    gint	       depth;
    gint	       tree_path_depth;
    gchar	      *name,
		      *uri;
    GList	      *paths;


    xml = this->private->xml;
    depth = gdk_visual_get_best_depth ();

    tree_selection = gtk_tree_view_get_selection (GTK_TREE_VIEW(tree_view));
    paths = gtk_tree_selection_get_selected_rows (tree_selection, &tree_model);
    
    if (!paths)
	return NULL;

    //FIXME: This is a little whackish because we assume there is only one selected path, _AND_ that it is the one being dragged
    tree_path = (GtkTreePath*)paths->data; 
    tree_path_depth = gtk_tree_path_get_depth (tree_path);
  
    query_create (tree_model, tree_path, &name, &uri);
 
    qtuple = g_new0 (DNDQueryTuple,1);
    qtuple->name = g_strdup(name);
    qtuple->uri  = g_strdup(uri);

    layout = gtk_widget_create_pango_layout (GTK_WIDGET(tree_view), name);
    pango_layout_get_pixel_size (layout, &width, &height);
    
    height = 28;
    width += 16 + 16 + 8;
    drawable = gdk_pixmap_new (NULL, width, height, depth);

    gtk_paint_box                   (GTK_WIDGET(tree_view)->style,
                                             drawable, 
                                             GTK_STATE_NORMAL, 
                                             GTK_SHADOW_OUT, 
                                             NULL, 
                                             GTK_WIDGET(tree_view), 
                                             "", 
                                             0, 
                                             0, 
                                             width,
                                             height);


    gtk_paint_layout                (GTK_WIDGET(tree_view)->style,
                                             drawable, 
                                             GTK_STATE_NORMAL, 
                                             TRUE, 
                                             NULL, 
                                             GTK_WIDGET(tree_view), 
                                             "", 
                                             32, 
                                             6, 
                                             layout); 

    cr = gdk_cairo_create (drawable);

    gdk_cairo_set_source_pixbuf     (cr,
                                     this->private->pb[tree_path_depth-1], 
                                     10, 
				     6); 

    cairo_paint (cr);
    cairo_destroy (cr);

    return drawable;
}

static void
tv_browse_drag_begin (GtkWidget	      *widget,
		      GdkDragContext  *ctx,
                      BmpLibrary      *this) 
{
    GdkPixmap *pixmap;
    gint       w, 
	       h;

    pixmap = tv_browse_create_row_drag_icon (widget, this); 
    
    if (!pixmap) gtk_drag_finish (ctx, FALSE, FALSE, 0L);
    gdk_drawable_get_size (pixmap, &w, &h);

    gtk_drag_set_icon_pixmap (ctx,
                              gdk_drawable_get_colormap (pixmap),
                              pixmap, 
			      NULL, 
			      0, 0); 
}

static void
drag_end (GtkWidget      *widget,
          GdkDragContext *drag_context,
          gpointer        data) 
{
    if (qtuple)
      {
	  g_free (qtuple->name);
	  g_free (qtuple->uri);
	  g_free (qtuple);
      }
}

static void
tv_browse_drag_data_get (GtkWidget        *widget,
                         GdkDragContext   *drag_context,
                         GtkSelectionData *data,
                         guint             info,
                         guint             time,
                         BmpLibrary       *this)
{
    gtk_selection_data_set          (data,
                                     gdk_atom_intern("application/x-bmpx-db-query", FALSE), 
                                     8, 
                                     (gpointer) &qtuple, 
                                     sizeof(gpointer)); 
}
#endif

static void
on_button_tag_new_clicked (GtkWidget	*widget,
			   BmpLibrary	*this)
{
    GError	  *error = NULL;
    GladeXML	  *xml = this->private->xml;
    GtkEntry	  *entry_tagname = GTK_ENTRY(glade_xml_get_widget_warn(xml, "entry_tagname"));
    const gchar	  *text = gtk_entry_get_text (entry_tagname);

    if ((!text) || (!strlen(text)))
      {
	GtkWidget   *message_dialog;

	message_dialog = gtk_message_dialog_new (GTK_WINDOW(glade_xml_get_widget_warn (xml, "dialog_library")),
						 GTK_DIALOG_MODAL,
						 GTK_MESSAGE_ERROR, 
						 GTK_BUTTONS_OK, 
						 _("Please supply a name for the new tag!")); 
	gtk_dialog_run (GTK_DIALOG(message_dialog));
	gtk_widget_destroy (message_dialog);
	return;
      }

    if (!bmp_metadata_cache_tag_new (bmp_metadata_cache,
				     text,
				     &error))
      {
	 if (g_error_matches (error, 
			      g_quark_from_static_string ("BmpMetadataTags"),
			      BMP_TAGS_EEXISTS)) 
	    {
	  	GtkWidget   *message_dialog;

		message_dialog = gtk_message_dialog_new (GTK_WINDOW(glade_xml_get_widget_warn (xml, "dialog_library")),
						 GTK_DIALOG_MODAL,
						 GTK_MESSAGE_ERROR, 
						 GTK_BUTTONS_OK, 
						 _("A tag with this name already exists. Please supply a different name.")); 
		gtk_dialog_run (GTK_DIALOG(message_dialog));
		gtk_widget_destroy (message_dialog);
		return;
	    }
  
	  // What could have happened here?!?! uh.
	  return;
      }
}

static void
on_button_tag_delete_clicked (GtkWidget	    *widget,
			      BmpLibrary    *this)
{
    GError	      *error = NULL;
    GladeXML	      *xml = this->private->xml; 
    GtkTreeView	      *tree_view_tags = GTK_TREE_VIEW(glade_xml_get_widget_warn(xml, "tree_view_tags"));
    GtkTreeSelection  *tree_selection = gtk_tree_view_get_selection (tree_view_tags);
    GtkTreeModel      *model_tags;
    GtkTreeIter	       iter;
    gchar	      *text;

    if (!gtk_tree_selection_get_selected (tree_selection, &model_tags, &iter))
      return;

    gtk_tree_model_get (model_tags, &iter, 1, &text, -1); 

    if (!bmp_metadata_cache_tag_delete (bmp_metadata_cache,
				     text,
				     &error))
      {
	 if (g_error_matches (error, 
			      g_quark_from_static_string ("BmpMetadataTags"),
			      BMP_TAGS_ENOEXIST)) 
	    {
	  	GtkWidget   *message_dialog;

		message_dialog = gtk_message_dialog_new (GTK_WINDOW(glade_xml_get_widget_warn (xml, "dialog_library")),
						 GTK_DIALOG_MODAL,
						 GTK_MESSAGE_ERROR, 
						 GTK_BUTTONS_OK, 
						 _("A tag with this name does not exist")); 
		gtk_dialog_run (GTK_DIALOG(message_dialog));
		gtk_widget_destroy (message_dialog);
		return;
	    }
  
	  // What could have happened here?!?! uh.
	  return;
      }
}

static void
on_tree_view_tags_selection_changed (GtkTreeSelection	*tree_selection,
				     BmpLibrary		*this)
{
    GladeXML	      *xml = this->private->xml; 

    GtkWidget	      *b_tag, 
		      *b_play, 
		      *b_enqueue,
		      *b_delete;

    GList	      *paths;
    
    paths = gtk_tree_selection_get_selected_rows (tree_selection, NULL);
    if (!paths)
      return;

    b_tag = 
	glade_xml_get_widget_warn (xml, "button_tag");
    b_play = 
	glade_xml_get_widget_warn (xml, "button_play");
    b_enqueue = 
	glade_xml_get_widget_warn (xml, "button_enqueue");
    b_delete = 
	glade_xml_get_widget_warn (xml, "button_tag_delete");

    if (gtk_tree_path_get_depth(paths->data)>1)
      {
	  gtk_widget_set_sensitive (b_delete, FALSE); 
      }
    else
      {
	  gtk_widget_set_sensitive (b_delete, (gtk_tree_selection_count_selected_rows (tree_selection) != 0) ? TRUE:FALSE);
      }

    gtk_widget_set_sensitive (GTK_WIDGET(b_tag), FALSE); 

#if 0
    if (gtk_tree_path_get_depth(paths->data)<3)
      {
	gtk_widget_set_sensitive (GTK_WIDGET(b_play), FALSE);
	gtk_widget_set_sensitive (GTK_WIDGET(b_enqueue), FALSE); 
      }
    else
      {
#endif
	gtk_widget_set_sensitive (GTK_WIDGET(b_play), (!gtk_tree_selection_count_selected_rows (tree_selection)) ? FALSE : TRUE);
	gtk_widget_set_sensitive (GTK_WIDGET(b_enqueue), (!gtk_tree_selection_count_selected_rows (tree_selection)) ? FALSE : TRUE);
#if 0
      }      
#endif
}

static void
on_renderer_modify_toggled (GtkCellRendererToggle *cell_renderer,
                            gchar                 *path,
                            BmpLibrary		  *this)
{
    GladeXML	      *xml = this->private->xml; 
    GtkTreeView	      *tree_view_modify_tags = GTK_TREE_VIEW(glade_xml_get_widget_warn (xml, "tree_view_modify_tags"));
    GtkTreeModel      *model = gtk_tree_view_get_model (tree_view_modify_tags);
    GtkTreePath	      *tree_path = gtk_tree_path_new_from_string (path);
    GtkTreeIter	       iter;
    gboolean	       value;

    if (!tree_path)
      return;

    if (!gtk_tree_model_get_iter (model, &iter, tree_path))
      {
	gtk_tree_path_free (tree_path);
	return;
      }

    gtk_tree_model_get (model, &iter, 0, &value, -1); 

    value = value ? FALSE : TRUE;

    gtk_list_store_set (GTK_LIST_STORE(model), &iter, 0, value, -1);
}

static void
on_entry_tagname_activate (GtkWidget	*widget,
			   BmpLibrary	*this)
{
    GladeXML	  *xml = this->private->xml; 
    GtkWidget	  *button_tag_new = glade_xml_get_widget_warn (xml, "button_tag_new");

    gtk_button_clicked (GTK_BUTTON(button_tag_new));
    gtk_entry_set_text (GTK_ENTRY(widget), "");
}

static void
on_cbox_metadatum_id_changed (GtkWidget	    *widget,
			      BmpLibrary    *this)
{
    gtk_tree_model_filter_refilter (GTK_TREE_MODEL_FILTER(this->private->tree_store_browse_filter));
}

#if 0
static gboolean
on_tb_playlist_new_toggled (GtkToggleButton *button,
			    BmpPlaylist	    *this)
{
  GtkWidget	  *menu;
  GdkWindow	  *parent;
  GtkRequisition   req;

  gint x_ro, y_ro;
  gint x_widget, y_widget;
  gint x_size, y_size;

  switch (gtk_toggle_button_get_active (button))
    {
	case TRUE:
	  {
	      parent = gtk_widget_get_parent_window(GTK_WIDGET(button));
	      gdk_drawable_get_size(parent, &x_size, &y_size);	 
	      gdk_drawable_get_size(GTK_WIDGET(button)->window, &x_size, &y_size);	 
	      gdk_window_get_root_origin(GTK_WIDGET(button)->window, &x_ro, &y_ro); 
	      gdk_window_get_position(GTK_WIDGET(button)->window, &x_widget, &y_widget);
 
	      menu = ui_manager_get_popup (bmp_ui_manager, "/popup-plmanager/menu-plmanager-new");
	      gtk_widget_size_request (menu, &req);
	      bmp_menu_popup (GTK_MENU(menu), 
			      x_ro + GTK_WIDGET(button)->allocation.x + 2, 
			      y_ro + y_size - GTK_WIDGET(button)->allocation.height - 6, 
			      1, 0L); 
	  }
	  break;

#if 0
	case FALSE:
	  {
	      menu = ui_manager_get_popup (bmp_ui_manager, "/popup-plmanager/menu-plmanager-new");
	      gtk_widget_hide (menu);
	  }
	  break;
#endif

	default: break;
    }

  return FALSE;
}

static void
disable_tb_playlist_new (GtkMenuShell	*shell,
		         BmpLibrary     *this)
{
  GladeXML    *xml = this->private->xml; 
  GtkWidget   *tb;

  tb = glade_xml_get_widget_warn (xml, "tb_playlist_new");

  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(tb), FALSE);
}

static void
relation_tuple_cell_data_func (GtkTreeViewColumn *tree_column,
                               GtkCellRenderer *cell,
                               GtkTreeModel *tree_model,
                               GtkTreeIter *iter,
                               gpointer data)
{
  HTuple         *tuple = NULL;
  HTupleElem	 *elem = NULL;
  gchar	         *v_string = NULL;
  BmpMetadatumId  metadatum_id = GPOINTER_TO_INT(data);

  gtk_tree_model_get (tree_model, iter, 0 /* this isn't safe */, &tuple, -1);
  g_assert (tuple);

  elem = h_tuple_get (tuple, bmp_metadata_get_id_static (metadatum_id));
  g_assert (elem);

  if (elem->type == G_TYPE_INT)
    {
      v_string = g_strdup_printf ("%d", GPOINTER_TO_INT(elem->value)); 
    }
  else
    {
      v_string = g_strdup ((gchar*)elem->value);
    }

  g_object_set (G_OBJECT(cell), "text", v_string, NULL);

  g_free (v_string);
}

static void
on_tree_view_playlist_cursor_changed (GtkTreeView     *tree_view_p,
                                      BmpLibrary      *this)
{
    GladeXML	      *xml = this->private->xml; 
    GtkWidget	      *tree_view_t;

    GtkTreeModel      *model,
		      *model_list;

    GtkTreeSelection  *selection;

    GtkTreeIter	       iter;

    gchar	      *name;

    tree_view_t = glade_xml_get_widget_warn (xml, "tree_view_tracks"); 
    selection = gtk_tree_view_get_selection (tree_view_p);

    if (!gtk_tree_selection_get_selected (selection, &model, &iter))
      return;

    gtk_tree_model_get (model, &iter, 0, &name, -1);
    model_list = GTK_TREE_MODEL(pl_manager_list_get_model (bmp_pl_manager,
					    name));

    gtk_tree_view_set_model (GTK_TREE_VIEW(tree_view_t), model_list);
}
#endif

static gboolean
tree_view_browse_drag_motion      (GtkWidget      *widget,
                                   GdkDragContext *context,
                                   gint            x,
                                   gint            y,
                                   guint           time,
                                   BmpLibrary     *this) 
{
    this->private->drag_inside = TRUE;
    gtk_drag_highlight (widget);
    return TRUE;
}

static void 
tree_view_browse_drag_leave      (GtkWidget      *widget,
                                  GdkDragContext *context,
                                  guint           time,
                                  BmpLibrary     *this)    
{
    gtk_drag_unhighlight (widget);
    this->private->drag_inside = FALSE;
}

static void 
tree_view_browse_drag_data_received        (GtkWidget        *widget,
                                            GdkDragContext   *context,
                                            gint              x,
                                            gint              y,
                                            GtkSelectionData *data,
                                            guint             info,
                                            guint             time_,
                                            BmpLibrary       *this)
{
    GtkWidget	  *notebook_library_add,
		  *window; 
    gchar	 **uri_strv;
    gchar	  *uri_string;
    gint	   n;

    g_return_if_fail (data != NULL);

    /* accept the drop */
    if (!data->data)
      {
  	  gtk_drag_finish (context, FALSE, FALSE, time_);
          return;
      }

    this->private->adding = TRUE;

    uri_string = g_strdup((gchar*)data->data);
    g_strstrip (uri_string);

    gtk_drag_finish (context, TRUE, FALSE, time_);

    uri_strv = g_uri_list_extract_uris (uri_string);
    g_free (uri_string);


    window = glade_xml_get_widget_warn (this->private->xml, "dialog_library");
    notebook_library_add = glade_xml_get_widget_warn (this->private->xml, "notebook_library_add");

    gtk_notebook_set_current_page (GTK_NOTEBOOK(notebook_library_add), 1);
    bmp_window_set_busy (GTK_WIDGET(window));

    n = 0;
    while (uri_strv[n]) 
      {
	BMPURI *bmpuri;
	    
	bmpuri = bmp_uri_new (uri_strv[n]);
	bmp_uri_unescape (bmpuri);

	if (g_file_test (bmpuri->path, G_FILE_TEST_IS_DIR))
	  {
	    GList  *items;
	    GError *error;
      
	    items = NULL;
	    error = NULL;
	    dir_foreach (bmpuri->path, (DirForeachFunc)collect_path, &items, &error);

	    for ( ; items ; items = items->next)
	      {
		  gchar *uri;
		  uri = g_filename_to_uri (items->data, NULL, NULL);
		  bmp_metadata_cache_metadata_cache (bmp_metadata_cache, uri);
		  g_free (uri);

		  while (gtk_events_pending()) gtk_main_iteration();
	      }
	    g_list_foreach (g_list_first(items), (GFunc)g_free, NULL);
	  }
	else
	  {
	    bmp_metadata_cache_metadata_cache (bmp_metadata_cache, uri_strv[n]);
	  }
  
	bmp_uri_delete (bmpuri);

	while (gtk_events_pending()) gtk_main_iteration();
	n++;
      }

    g_strfreev (uri_strv);

    gtk_notebook_set_current_page (GTK_NOTEBOOK(notebook_library_add), 0);
    bmp_window_set_idle (GTK_WIDGET(window));

    this->private->adding = FALSE;
}

static void
pb_cell_data_func (GtkTreeViewColumn  *tree_column,
		   GtkCellRenderer    *cell,
		   GtkTreeModel	      *tree_model,
                   GtkTreeIter	      *iter,
                   BmpLibrary	      *this)
{
    GtkTreePath *path;
    gint	 depth;

    path = gtk_tree_model_get_path (tree_model, iter);
    depth = gtk_tree_path_get_depth (path)-1;
    gtk_tree_path_free (path);

    g_object_set (G_OBJECT(cell), "pixbuf", this->private->pb[depth], NULL); 
}

static void
window_library_create (BmpLibrary *this)
{
    GtkWidget		*widget;
    static const gchar  *glade_file = DATA_DIR "/glade/dialog_library.glade";
    gint		 n;

    this->private->xml = glade_xml_new_or_die(N_("Library Dialog"), glade_file, NULL, NULL);
    if (!this->private->xml) return;
	
    glade_xml_signal_autoconnect (this->private->xml);

    /* Add actions */
#if 0
    gtk_action_group_add_actions (bmp_actions, bmp_actions_LIBRARY, G_N_ELEMENTS (bmp_actions_LIBRARY), this->private->xml);
    if (!gtk_ui_manager_add_ui_from_file (bmp_ui_manager,
      DATA_DIR G_DIR_SEPARATOR_S BMP_UI_BASENAME G_DIR_SEPARATOR_S "plmanager.ui", &error))
      {
		g_message ("building menus failed: %s", error->message);
		g_error_free (error);
		error = NULL;
      }
#endif

    this->window = glade_xml_get_widget_warn (this->private->xml, "dialog_library");

    RM_ITEM_NEW("windows", "window-dialog-library", this->window, NULL);

    gtk_widget_realize (this->window);
    g_object_connect (G_OBJECT(this->window), 
			  "signal::delete-event", 
			  G_CALLBACK(on_dialog_library_delete),
			  this,
			  NULL);

    bmp_window_set_icon_list (GTK_WIDGET(this->window), "library");

    gtk_window_set_position (GTK_WINDOW(this->window), GTK_WIN_POS_CENTER);
    g_object_connect (G_OBJECT(this->window), 
			  "signal::key-press-event", 
			  G_CALLBACK(on_dialog_library_key_press), 
			  this, 
			  NULL);
    
    /* Various */
    widget = glade_xml_get_widget_warn (this->private->xml, "notebook_library");
    g_object_connect (G_OBJECT(widget),
			  "signal::switch-page",
			  G_CALLBACK(on_library_notebook_switch_page),
			  this, 
			  NULL);

    gtk_notebook_set_current_page (GTK_NOTEBOOK(widget), 0);

    widget = glade_xml_get_widget_warn (this->private->xml, "image_library");
    gtk_image_set_from_file (GTK_IMAGE (widget), DATA_DIR "/images/header-library.png");

    widget = glade_xml_get_widget_warn (this->private->xml, "image_wait_adding");
    gtk_image_set_from_file (GTK_IMAGE(widget), DATA_DIR "/images/wait.gif");

    widget = glade_xml_get_widget_warn (this->private->xml, "image_wait_library_inner");
    gtk_image_set_from_file (GTK_IMAGE(widget), DATA_DIR "/images/wait.gif");

    widget = glade_xml_get_widget_warn (this->private->xml, "image_tag");
    gtk_image_set_from_file (GTK_IMAGE(widget), DATA_DIR "/images/icon-bookmark.png");


    widget = glade_xml_get_widget_warn (this->private->xml, "b_music_add");	
    g_object_connect (G_OBJECT(widget), 
			  "signal::clicked", 
			  G_CALLBACK(on_music_add), 
			  this, 
			  NULL);

    widget = glade_xml_get_widget_warn (this->private->xml, "b_music_remove");	
    g_object_connect (G_OBJECT(widget), 
			  "signal::clicked", 
			  G_CALLBACK(on_music_remove), 
			  this, 
			  NULL);

    /* Setup Tags tab */
    widget = glade_xml_get_widget_warn (this->private->xml, "image_tag1");
    gtk_image_set_from_file (GTK_IMAGE(widget), DATA_DIR "/images/icon-bookmark.png");

    widget = glade_xml_get_widget_warn (this->private->xml, "tree_view_tags");
    g_object_connect (G_OBJECT(gtk_tree_view_get_selection(GTK_TREE_VIEW(widget))),
			  "signal::changed",
			  G_CALLBACK(on_tree_view_tags_selection_changed),
			  this, 
			  NULL);

    column_tags[0] = gtk_tree_view_column_new ();

    gtk_tree_view_column_set_title (column_tags[0], _("Tag Name"));
    gtk_tree_view_column_set_sort_column_id (column_tags[0], 1);

    renderer_tags[0] =
		gtk_cell_renderer_pixbuf_new ();
    renderer_tags[1] =
		gtk_cell_renderer_text_new ();

    gtk_cell_renderer_set_fixed_size (renderer_tags[0], 16, 16);

    g_object_set (G_OBJECT(renderer_tags[0]), "xalign", 0.0, NULL);
    g_object_set (G_OBJECT(renderer_tags[1]), "xpad", 6, NULL);

    gtk_tree_view_column_pack_start
		(column_tags[0],
		 renderer_tags[0],
		 FALSE);

    gtk_tree_view_column_pack_start
		(column_tags[0],
		 renderer_tags[1],
		 TRUE);


    gtk_tree_view_column_set_attributes
		(column_tags[0],
		 renderer_tags[0],
		 "pixbuf",
		  0,
		  NULL);

    gtk_tree_view_column_set_attributes
		(column_tags[0],
		 renderer_tags[1],
		 "text",
		  1,
		  NULL);

    gtk_tree_view_column_set_expand
		(column_tags[0],
		 FALSE);

    gtk_tree_view_append_column
		(GTK_TREE_VIEW (widget),
		 GTK_TREE_VIEW_COLUMN (column_tags[0]));


	/* Setup BROWSE Tab */

	/* Filter combobox */
	{
	  GtkListStore *store;
	  GtkCellRenderer *cell;
	  GtkTreeIter iter;

	  widget = glade_xml_get_widget_warn (this->private->xml, "cbox_metadatum_id");
	  cell = gtk_cell_renderer_text_new ();
	  gtk_cell_layout_clear (GTK_CELL_LAYOUT(widget));
	  gtk_cell_layout_pack_start (GTK_CELL_LAYOUT(widget), cell, TRUE);
	  gtk_cell_layout_set_attributes (GTK_CELL_LAYOUT(widget), cell, "text", 0, NULL);

	  store = gtk_list_store_new (2, G_TYPE_STRING, G_TYPE_INT);

	  for (n = 0; n < BMP_DATUM_NONE; n++)
	    {
		if (bmp_metadata_get_filter (n))
		  {
		      gtk_list_store_append (store, &iter);
		      gtk_list_store_set (store, &iter, 0, bmp_metadata_get_title (n), 1, n, -1);
		  }
	    }

	  gtk_combo_box_set_model (GTK_COMBO_BOX(widget), GTK_TREE_MODEL(store));
	  gtk_combo_box_set_active (GTK_COMBO_BOX(widget), 1); //FIXME: hardcoded

	  g_object_connect(G_OBJECT(widget), 
			 "signal::changed", 
			 G_CALLBACK(on_cbox_metadatum_id_changed), 
			 this, 
			 NULL);
	}


	widget = glade_xml_get_widget_warn (this->private->xml, "tree_view_browse");
	g_object_connect(G_OBJECT(widget), 
			 "swapped-signal::row-activated", 
			 G_CALLBACK(gtk_window_activate_default), 
			 GTK_WINDOW(this->window), 
			 NULL);
	g_object_connect(G_OBJECT(gtk_tree_view_get_selection (GTK_TREE_VIEW(widget))),
			 "signal::changed",
			 G_CALLBACK(on_tree_selection_changed),
			 this, 
			 NULL);


	gtk_tree_selection_set_mode
	    (gtk_tree_view_get_selection (GTK_TREE_VIEW(widget)),
	     GTK_SELECTION_MULTIPLE);

	/* Main column */ 
	column_browse[0] =
		gtk_tree_view_column_new ();

	gtk_tree_view_column_set_title (column_browse[0], _("Artist / Album / Title"));
	gtk_tree_view_column_set_resizable (column_browse[0], TRUE);
	gtk_tree_view_column_set_min_width (column_browse[0],300);

	/* The cell renderer for the artist/album/title icon */
	renderer_browse[0] =
		gtk_cell_renderer_pixbuf_new ();

	g_object_set (renderer_browse[0], "width", 16, NULL);
	g_object_set (renderer_browse[0], "height", 16, NULL);

	gtk_tree_view_column_pack_start
		(column_browse[0],
		 renderer_browse[0],
		 FALSE);

	gtk_tree_view_column_set_cell_data_func 
		(column_browse[0],
		 renderer_browse[0],
		 (GtkTreeCellDataFunc) pb_cell_data_func,
		 this,
		 NULL);

#if 0
	gtk_tree_view_column_set_attributes
		(column_browse[0],
		 renderer_browse[0],
		 "pixbuf",
		  0,
		  NULL);
#endif

	/* The cell renderer for the actual text */
	renderer_browse[1] =
		gtk_cell_renderer_text_new ();

	g_object_set (G_OBJECT(renderer_browse[1]), 
		      "xpad", 4,
		      "ellipsize", PANGO_ELLIPSIZE_END,
		      NULL);

	gtk_tree_view_column_pack_start
		(column_browse[0],
		 renderer_browse[1],
		 TRUE);

	gtk_tree_view_column_set_attributes
		(column_browse[0],
		 renderer_browse[1],
		 "text",
		  1,
		  NULL);

	gtk_tree_view_column_set_expand
		(column_browse[0],
		 FALSE);

	gtk_tree_view_append_column
		(GTK_TREE_VIEW (widget),
		 GTK_TREE_VIEW_COLUMN (column_browse[0]));

	/* Track number column */ 
	column_browse[1] =
		gtk_tree_view_column_new ();

	gtk_tree_view_column_set_title (column_browse[1], _("Track"));

	renderer_browse[2] =
		gtk_cell_renderer_text_new ();

	g_object_set (G_OBJECT(renderer_browse[2]),
	              "xalign", 1.0,
		      NULL);

	gtk_tree_view_column_pack_start
		(column_browse[1],
		 renderer_browse[2],
		 FALSE);

	gtk_tree_view_column_set_expand
		(column_browse[1],
		 FALSE);

	gtk_tree_view_column_set_cell_data_func
		(column_browse[1],
		 renderer_browse[2],
		 (GtkTreeCellDataFunc) tracknumber_renderer_func,
                 NULL, 
                 NULL); 

	gtk_tree_view_append_column
		(GTK_TREE_VIEW (widget),
		 GTK_TREE_VIEW_COLUMN (column_browse[1]));

	/* Length */ 
	column_browse[2] =
		gtk_tree_view_column_new ();

	gtk_tree_view_column_set_title (column_browse[2], _("Length"));

	renderer_browse[3] =
		gtk_cell_renderer_text_new ();

	g_object_set (G_OBJECT(renderer_browse[2]),
	              "xalign", 0.5,
		      NULL);

	gtk_tree_view_column_pack_start
		(column_browse[2],
		 renderer_browse[3],
		 FALSE);

	gtk_tree_view_column_set_expand
		(column_browse[2],
		 FALSE);

	gtk_tree_view_column_set_cell_data_func
		(column_browse[2],
		 renderer_browse[3],
		 (GtkTreeCellDataFunc) length_renderer_func,
                 NULL, 
                 NULL); 

	gtk_tree_view_append_column
		(GTK_TREE_VIEW (widget),
		 GTK_TREE_VIEW_COLUMN (column_browse[2]));

	/* Genre Column */
	column_browse[3] =
		gtk_tree_view_column_new ();

	gtk_tree_view_column_set_title (column_browse[3], _("Genre"));

	renderer_browse[4] =
		gtk_cell_renderer_text_new ();

	gtk_tree_view_column_pack_start
		(column_browse[3],
		 renderer_browse[4],
		 FALSE);

	gtk_tree_view_column_set_expand
		(column_browse[3],
		 FALSE);

      	gtk_tree_view_column_set_cell_data_func
		(column_browse[3],
		 renderer_browse[4],
		 (GtkTreeCellDataFunc) genre_renderer_func,
                 NULL, 
                 NULL); 

	gtk_tree_view_append_column
		(GTK_TREE_VIEW (widget),
		 GTK_TREE_VIEW_COLUMN (column_browse[3]));

	/* Bitrate Column */
	column_browse[4] =
		gtk_tree_view_column_new ();

	gtk_tree_view_column_set_title (column_browse[4], _("Bitrate"));

	renderer_browse[5] =
		gtk_cell_renderer_text_new ();

	gtk_tree_view_column_pack_start
		(column_browse[4],
		 renderer_browse[5],
		 FALSE);

	gtk_tree_view_column_set_expand
		(column_browse[4],
		 FALSE);

      	gtk_tree_view_column_set_cell_data_func
		(column_browse[4],
		 renderer_browse[5],
		 (GtkTreeCellDataFunc) bitrate_renderer_func,
                 NULL, 
                 NULL); 

	gtk_tree_view_append_column
		(GTK_TREE_VIEW (widget),
		 GTK_TREE_VIEW_COLUMN (column_browse[4]));

#if 0
	/* Setup Playlists tab */
	widget = glade_xml_get_widget_warn (this->private->xml, "image_playlists");
	gtk_image_set_from_file (GTK_IMAGE(widget), DATA_DIR "/images/icon-playlist.png");

	widget = glade_xml_get_widget_warn (this->private->xml, "tb_playlist_new");
	g_object_connect (G_OBJECT(widget),
			  "signal::toggled",
			  G_CALLBACK(on_tb_playlist_new_toggled),
			  this, 
			  NULL);

	widget = ui_manager_get_popup (bmp_ui_manager, "/popup-plmanager/menu-plmanager-new");
	g_object_connect (G_OBJECT(widget),
			  "signal::cancel",
			  G_CALLBACK(disable_tb_playlist_new),
			  this, 
			  "signal::selection-done",
			  G_CALLBACK(disable_tb_playlist_new),
			  this, 
			  NULL);


	{	
	  GtkTreeViewColumn   *column;
	  GtkCellRenderer     *cell;
	  gint		       n;

	  widget = glade_xml_get_widget_warn (this->private->xml, "tree_view_playlists"); 
	  column = gtk_tree_view_column_new (); 
	  gtk_tree_view_column_set_title (column, _("Playlist Name")); 
	  g_object_connect (G_OBJECT(widget),
			    "signal::cursor-changed",
			    G_CALLBACK(on_tree_view_playlist_cursor_changed),
			    this,
			    NULL);

	  this->private->map_root_playlists = attr_row_map_new ();

	  cell = gtk_cell_renderer_text_new ();
	  gtk_tree_view_column_pack_start
		(column,
		 cell, 
		 FALSE);

	  gtk_tree_view_column_set_expand
		(column,
		 FALSE);

	  gtk_tree_view_column_set_attributes
		(column,
		 cell, 
		 "text",
		  0,
		  NULL);

	  gtk_tree_view_column_set_sort_column_id (column, 0);
	  gtk_tree_view_append_column (GTK_TREE_VIEW(widget), column);
      
	  { 
	      GtkListStore    *liststore;
	      GtkTreeIter      iter;
	      GList	      *playlists;

	      liststore = gtk_list_store_new (1, G_TYPE_STRING);
	      playlists = pl_manager_get_lists (bmp_pl_manager);

	      for ( ; playlists ; playlists = playlists->next)
		{
		    gtk_list_store_append (liststore, &iter);
		    gtk_list_store_set (liststore, &iter, 0, playlists->data, -1);
		    attr_row_map_mapping_insert (this->private->map_root_playlists, playlists->data, GTK_TREE_MODEL(liststore), &iter);
		}
	      
	      gtk_tree_view_set_model (GTK_TREE_VIEW(widget), GTK_TREE_MODEL(liststore)); 
	  }

	  widget = glade_xml_get_widget_warn (this->private->xml, "tree_view_tracks"); 
	  for (n = 0; n < BMP_DATUM_NONE; n++)
	    {

		column = gtk_tree_view_column_new (); 
		gtk_tree_view_column_set_title (column, gettext(bmp_metadata_get_title_static (n)));

		cell = gtk_cell_renderer_text_new ();
		gtk_tree_view_column_pack_start
		(column,
		 cell, 
		 FALSE);

		gtk_tree_view_column_set_expand
		(column,
		 FALSE);

		gtk_tree_view_column_set_cell_data_func
                                            (column,
                                             cell, 
                                             (GtkTreeCellDataFunc) relation_tuple_cell_data_func,
                                             GINT_TO_POINTER(n), 
                                             NULL);

		gtk_tree_view_append_column (GTK_TREE_VIEW(widget), column);
	    }

	    gtk_tree_view_set_model (GTK_TREE_VIEW(widget), NULL); 
	}
#endif

	/* Setup modify tags dialog */
	widget = glade_xml_get_widget_warn (this->private->xml, "tree_view_modify_tags");
	
	column_modify[0] =
	      gtk_tree_view_column_new (); 

	gtk_tree_view_column_set_title (column_modify[0], _("Tag Name"));

	renderer_modify[0] =
	      gtk_cell_renderer_toggle_new ();
	renderer_modify[1] =
	      gtk_cell_renderer_text_new ();

	gtk_tree_view_column_pack_start
		(column_modify[0],
		 renderer_modify[0],
		 FALSE);

	gtk_tree_view_column_pack_start
		(column_modify[0],
		 renderer_modify[1],
		 FALSE);

	gtk_tree_view_column_set_expand
		(column_modify[0],
		 FALSE);

	gtk_tree_view_column_set_attributes
		(column_modify[0],
		 renderer_modify[0],
		 "active",
		  0,
		  NULL);

	gtk_tree_view_column_set_attributes
		(column_modify[0],
		 renderer_modify[1],
		 "text",
		  1,
		  NULL);

	gtk_tree_view_append_column (GTK_TREE_VIEW(widget), column_modify[0]);

	g_object_connect (G_OBJECT(renderer_modify[0]),
			  "signal::toggled",
			  G_CALLBACK(on_renderer_modify_toggled),
			  this, 
			  NULL);

	/* Common widgets */
	widget = glade_xml_get_widget_warn (this->private->xml, "button_tag");
	g_object_connect(G_OBJECT(widget),
			 "signal::clicked", 
			 G_CALLBACK(on_button_tag_clicked),
			 this, 
			 NULL);

	widget = glade_xml_get_widget_warn (this->private->xml, "button_enqueue");
	g_object_connect(G_OBJECT(widget),
			 "signal::clicked",
		         G_CALLBACK(on_button_enqueue_clicked),
			 this, 
			 NULL);

	widget = glade_xml_get_widget_warn (this->private->xml, "button_play");
	g_object_connect(G_OBJECT(widget),
			 "signal::clicked",
			 G_CALLBACK(on_button_play_clicked),
			 this, 
			 NULL);

	widget = glade_xml_get_widget_warn (this->private->xml, "entry_tagname");
	g_object_connect (G_OBJECT(widget),
			  "signal::activate",
			  G_CALLBACK(on_entry_tagname_activate),
			  this, 
			  NULL);

	widget = glade_xml_get_widget_warn (this->private->xml, "sw_browse");

        gtk_drag_dest_set(widget,
			  GTK_DEST_DEFAULT_MOTION | GTK_DEST_DEFAULT_DROP, 
			  bmp_drop_types, 3, 
			  GDK_ACTION_COPY);

	g_object_connect (G_OBJECT(widget),

			  "signal::drag-motion",
			  G_CALLBACK(tree_view_browse_drag_motion),
			  this, 

			  "signal::drag-leave",
			  G_CALLBACK(tree_view_browse_drag_leave),
			  this, 

			  "signal::drag-data-received",
			  G_CALLBACK(tree_view_browse_drag_data_received),
			  this, 

			  NULL);
}

static gboolean
init_window (BmpLibrary *this)
{
    GtkWidget *widget;


    
    this->private->tree_store_browse = GTK_TREE_MODEL(gtk_tree_store_new (6, 
							        GDK_TYPE_PIXBUF, 
								G_TYPE_STRING,
							        G_TYPE_STRING, 
								G_TYPE_STRING, 
							        G_TYPE_HASH_TABLE,
								G_TYPE_HASH_TABLE));

    gtk_tree_sortable_set_sort_func         (GTK_TREE_SORTABLE (this->private->tree_store_browse),
					     1,
                                             (GtkTreeIterCompareFunc) tree_store_browse_sort_func,
                                             NULL, 
                                             NULL); 

   
    gtk_tree_sortable_set_sort_column_id    (GTK_TREE_SORTABLE (this->private->tree_store_browse),
					     1, 
					     GTK_SORT_ASCENDING);


    this->private->tree_store_browse_filter = gtk_tree_model_filter_new (GTK_TREE_MODEL(this->private->tree_store_browse), NULL);
    this->private->map_root_browse = attr_row_map_new ();

    bmp_metadata_cache_load_rtfs (bmp_metadata_cache);
    bmp_metadata_cache_for_each_tuple (bmp_metadata_cache, (HTupleFunc)browse_view_insert_tuple_proxy_func, this);

    signal_handler_pool_add_connection (this->private->pool,
					bmp_metadata_cache,
					"insert-tuple",
					G_CALLBACK(browse_view_insert_tuple),
				        this,
					0);

    signal_handler_pool_add_connection (this->private->pool,
					bmp_metadata_cache,
					"delete-tuple",
					G_CALLBACK(browse_view_delete_tuple),
				        this,
					0);

    this->private->tree_store_tags = GTK_TREE_MODEL(gtk_tree_store_new (5, 
					  GDK_TYPE_PIXBUF,
					  G_TYPE_STRING, 
					  G_TYPE_POINTER,
					  G_TYPE_POINTER,
					  G_TYPE_INT)); 

    this->private->map_root_tags = attr_row_map_new ();

    signal_handler_pool_add_connection (this->private->pool,
					bmp_metadata_cache,
					"tag-added",
					G_CALLBACK(on_tag_added),
				        this,
					0);

    signal_handler_pool_add_connection (this->private->pool,
					bmp_metadata_cache,
					"tag-deleted",
					G_CALLBACK(on_tag_deleted),
				        this,
					0);

    signal_handler_pool_add_connection (this->private->pool,
					bmp_metadata_cache,
					"tag-value-insert",
					G_CALLBACK(on_tag_value_insert),
				        this,
					0);

    signal_handler_pool_add_connection (this->private->pool,
					bmp_metadata_cache,
					"tag-value-remove",
					G_CALLBACK(on_tag_value_remove),
				        this,
					0);

    bmp_metadata_cache_load_tags (bmp_metadata_cache);

    /* Tags common setup */
    // New tag
    widget = glade_xml_get_widget_warn (this->private->xml, "button_tag_new");
    g_object_connect (G_OBJECT(widget),
			 "signal::clicked",
			 G_CALLBACK(on_button_tag_new_clicked),
			 this, 
			 NULL);

    // Delete a tag
    widget = glade_xml_get_widget_warn (this->private->xml, "button_tag_delete");
    g_object_connect (G_OBJECT(widget),
			 "signal::clicked",
			 G_CALLBACK(on_button_tag_delete_clicked),
			 this, 
			 NULL);

    widget = glade_xml_get_widget_warn (this->private->xml, "entry_value");
    g_object_connect(G_OBJECT(widget), 
		     "swapped-signal::changed", 
		     G_CALLBACK(gtk_tree_model_filter_refilter), 
		     this->private->tree_store_browse_filter, 
		     NULL);

    widget = glade_xml_get_widget_warn (this->private->xml, "notebook_library_inner");
    gtk_notebook_set_current_page (GTK_NOTEBOOK (widget), 1);

    widget = glade_xml_get_widget_warn (this->private->xml, "tree_view_browse");
    gtk_tree_model_filter_set_visible_func (GTK_TREE_MODEL_FILTER(this->private->tree_store_browse_filter),
						(GtkTreeModelFilterVisibleFunc)tree_store_browse_visible_func,
						this,	
						NULL);

#ifdef HAVE_HAL
    g_object_connect(G_OBJECT(bmp_hal), 
		     "swapped-signal::volume-added", 
		     G_CALLBACK(gtk_tree_model_filter_refilter), 
		     this->private->tree_store_browse_filter, 
		     NULL);

    g_object_connect(G_OBJECT(bmp_hal), 
		     "swapped-signal::volume-removed", 
		     G_CALLBACK(gtk_tree_model_filter_refilter), 
		     this->private->tree_store_browse_filter, 
		     NULL);
#endif

    gtk_tree_view_set_model (GTK_TREE_VIEW (widget), GTK_TREE_MODEL(this->private->tree_store_browse_filter));
    gtk_widget_show_all (GTK_WIDGET (widget));

    widget = glade_xml_get_widget_warn (this->private->xml, "tree_view_tags");
    gtk_tree_view_set_model (GTK_TREE_VIEW(widget), GTK_TREE_MODEL(this->private->tree_store_tags));
    gtk_widget_show_all (GTK_WIDGET (widget));

    widget = glade_xml_get_widget_warn (this->private->xml, "notebook_library");
    gtk_widget_set_sensitive (widget, TRUE);

    signal_handler_pool_add_connection (this->private->pool,
					bmp_system_control,
					"shutdown-request",
					G_CALLBACK(on_system_control_shutdown_request),
				        this,
					0);

    this->private->shown = TRUE;
    this->private->running = FALSE;

    return FALSE;
}

void
bmp_library_show (BmpLibrary *this)
{
    if ((!this->private->shown) && (!this->private->running))
      {
	GtkWidget *widget;

	widget = glade_xml_get_widget_warn (this->private->xml, "notebook_library");
	gtk_widget_set_sensitive (widget, FALSE);

	this->private->running = TRUE;

	g_idle_add ((GSourceFunc) init_window, this);
      }

    gtk_window_present (GTK_WINDOW(this->window));
}

#if 0
static void
ui_dialog_library_playlist_new_from_file (GtkAction	*action,
					  BmpLibrary    *this)
{
    GladeXML	  *xml = this->private->xml; 
    GtkWidget	  *chooser,
		  *parent;

    GtkFileFilter *file_filter;

    GList	  *containers = NULL;

    gint	   rv = 0;

    parent = glade_xml_get_widget_warn (xml, "dialog_library");
    chooser = gtk_file_chooser_dialog_new (_("Select playlist to load... - BMP"),
					   GTK_WINDOW(parent),
					   GTK_FILE_CHOOSER_ACTION_OPEN,
					   GTK_STOCK_OPEN,
					   GTK_RESPONSE_OK,
					   GTK_STOCK_CANCEL,
					   GTK_RESPONSE_CANCEL,
					   NULL);

    gtk_file_chooser_set_select_multiple (GTK_FILE_CHOOSER(chooser), FALSE);

    /* Set Path */
    gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER(chooser), xcs_key_fetch (xcs, "bmp", "file-chooser-path"));

    containers = plugins_find_containers (); 

    for ( ; containers; containers=containers->next)
      {
	  ContainerHTuple *tuple = (ContainerHTuple*)containers->data;
	  gchar *glob, *glob_suffix;

	  glob = create_glob (tuple->suffix);
	  glob_suffix = g_strconcat ("*.", glob, NULL);
	  g_free (glob);

	  file_filter = gtk_file_filter_new();
	  gtk_file_filter_set_name (file_filter, tuple->desc);
	  gtk_file_filter_add_pattern(file_filter, glob_suffix); 
	  gtk_file_chooser_add_filter (GTK_FILE_CHOOSER(chooser), file_filter);

	  g_free (glob_suffix);
      }


    rv = gtk_dialog_run (GTK_DIALOG(chooser));

    switch (rv)
      {
	  case GTK_RESPONSE_CANCEL:
	    {
		gtk_widget_destroy (chooser);
		return;
	    }

	  case GTK_RESPONSE_OK:
	    {
		GtkWidget    *widget = glade_xml_get_widget_warn (xml, "tree_view_tracks"); 
		GtkListStore *store;
		GtkTreeIter   iter;

		gchar  *uri,
		       *filename,
		       *basename;

		gchar **fragments;

		uri = gtk_file_chooser_get_uri (GTK_FILE_CHOOSER(chooser));
		filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER(chooser));
		basename = g_path_get_basename (filename);
		fragments = g_strsplit (basename, ".", -1);
		g_free (basename);

		//FIXME: PLManager should make use of GError
		if (!pl_manager_list_new_from_uri (bmp_pl_manager,
						   fragments[0], 
						   uri)) 
		  {
		    g_log (G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Unable to create playlist from URI '%s'", uri);
		  }

		store = pl_manager_list_get_model (bmp_pl_manager, 
						   fragments[0]);

		g_free (uri);
		g_free (filename);
		gtk_widget_destroy (GTK_WIDGET(chooser));

		widget = glade_xml_get_widget_warn (xml, "tree_view_playlists"); 
		gtk_list_store_append (GTK_LIST_STORE(gtk_tree_view_get_model(GTK_TREE_VIEW(widget))), &iter);
		gtk_list_store_set (GTK_LIST_STORE(gtk_tree_view_get_model(GTK_TREE_VIEW(widget))), &iter, 0, fragments[0], -1);
		attr_row_map_mapping_insert (this->private->map_root_playlists, fragments[0], gtk_tree_view_get_model(GTK_TREE_VIEW(widget)), &iter);
	
		g_strfreev (fragments);	

		return;
	    }

	  default: break;
      }

    gtk_widget_destroy (GTK_WIDGET(chooser));
}
#endif
