/*  BMPx - The Dumb Music Player
 *  Copyright (C) 2005 BMPx development team.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  $Id$
 *
 * The BMPx project hereby grant permission for non-gpl compatible GStreamer
 * plugins to be used and distributed together with GStreamer and BMPx. This
 * permission are above and beyond the permissions granted by the GPL license
 * BMPx is covered by.
 *
 */

#include <config.h>

#include <revision.h>
#include <build.h>

#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>

#include <limits.h>

#include <sys/time.h>
#include <sys/types.h>

#include <glib.h>
#include <glib/gi18n.h>
#include <gtk/gtk.h>
#include <glade/glade.h>

#include <gst/gst.h>
#include <gst/gstelement.h>

#include <bmp/util.h>

#include <libskinned/bmp_tooltips.h>
#include <libchroma/chroma-list.h>

#include <xcs/xcs.h>
#include <goa/libgoa.h>

#include "glade.h"
#include "loader.h"
#include "play.h"
#include "ui.h"
#include "ui_main.h"
#include "ui_playlist.h"
#include "ui_util.h"

#include "ui_dialog_preferences.h"
#include "signal_handler_pool.h"
#include "plugin_host.h"

#ifdef HAVE_ALSA
#  include "alsa.h"
#endif

#include "main.h"

#define THUMBNAIL_WIDTH  90
#define THUMBNAIL_HEIGHT 40

enum SkinViewCols {
    SKIN_VIEW_COL_PREVIEW,
    SKIN_VIEW_COL_NAME,
    SKIN_VIEW_COL_PATH,
    SKIN_VIEW_N_COLS
};

enum CategoryViewCols {
    CATEGORY_VIEW_COL_ICON,
    CATEGORY_VIEW_COL_NAME,
    CATEGORY_VIEW_COL_ID,

    CATEGORY_VIEW_N_COLS
};

#define SKIN_NODE(x)  ((SkinNode *)(x))
struct _SkinNode {
    gchar *name;
    gchar *path;
    GTime *time;
};
typedef struct _SkinNode SkinNode;

typedef struct {
    const gchar *icon_path;
    const gchar *name;
    gint id;
} Category;

typedef struct {
    const gchar *name;
    const gchar *tag;
}
TitleFieldTag;

static struct {
    gchar *description;
    gchar *sinkname;
    gint   tab_id;
} audio_systems[]= {
#ifdef HAVE_ALSA
    {"ALSA (Advanced Linux Sound Architecture)", "alsasink", 1},
#endif
    {"OSS (Open Sound System)","osssink", 2},
#ifdef HAVE_SUNAUDIO
    {"Sun/Solaris Audio", "sunaudiosink", 3},
#endif
};

static GHashTable *map_skins;

static Category categories[] = {
    {DATA_DIR "/images/preferences/appearance.png", N_("Appearance"),   0},
/*    {DATA_DIR "/images/preferences/eq.png",         N_("Equalizer"),	1}, */
    {DATA_DIR "/images/preferences/playlist.png",   N_("Playlist"),	2},
    {DATA_DIR "/images/preferences/playback.png",   N_("Audio"),	3}, 
    {DATA_DIR "/images/preferences/plugins.png",    N_("Plugins"),	4},
    {DATA_DIR "/images/preferences/lastfm.png",     N_("Last.FM"),	5},
    {DATA_DIR "/images/preferences/library.png",    N_("Library"),	6}
};
static gint n_categories = G_N_ELEMENTS(categories);

static BmpTooltipEntry window_preferences_tooltips[] = {
    { "titlestring_entry",
      N_("Allows you to customize how text is displayed in the playlist"),
      GTK_STOCK_INFO },

    { "skin_view",
      N_("Here you can choose a skin to determine how BMPx looks like"),
      GTK_STOCK_INFO },

    { "font-per-skin",
      N_("Enabling this will make BMPx remember the font for each skin individually, giving you more ease when theming BMPx"),
      GTK_STOCK_INFO },

    { "font_button",
      N_("Click on this button to select a font for BMPx to use"),
      GTK_STOCK_INFO }
};
static gint n_window_preferences_tooltips = G_N_ELEMENTS (window_preferences_tooltips);

#if 0
static gchar* eq_sliders[] = {
    "eq_30",
    "eq_60",
    "eq_125",
    "eq_250",
    "eq_500",
    "eq_1000",
    "eq_2000",
    "eq_4000",
    "eq_8000",
    "eq_16000"
};
static gint n_eq_sliders = G_N_ELEMENTS(eq_sliders);
#endif

static void
update_skinlist(BmpPreferences *this);

static GdkPixbuf *
skin_get_thumbnail(const gchar * path);

/* GObject stuff */
G_DEFINE_TYPE (BmpPreferences, bmp_preferences, G_TYPE_OBJECT)

struct _BmpPreferencesPrivate {

    gboolean		   dispose_has_run;

    GladeXML		  *xml;
    gchar		  *skins_path;
    GList		  *skinlist;

    GtkWidget		  *skin_view;
    GtkWidget		  *skin_view_notebook;
  
    GtkWidget		  *columns_view;
    GtkListStore	  *columns_store;

    BmpPluginHost	  *host;
    SignalHandlerPool	  *pool;
    gboolean		   updated;
};

void
window_preferences_create (BmpPreferences *this);

static void
set_font (BmpPreferences *this, const gchar *font_name)
{
    GtkWidget *font_button;

    font_button = glade_xml_get_widget (this->private->xml, "font_button");
    gtk_font_button_set_font_name(GTK_FONT_BUTTON(font_button), font_name);
}

BmpPreferences*
bmp_preferences_new (void)
{
    BmpPreferences *this;

    this = g_object_new (BMP_TYPE_PREFERENCES, NULL);
    return this;
}

static GObject *
bmp_preferences_constructor (GType type,
	                     guint n_construct_properties,
         	             GObjectConstructParam * construct_properties)
{
    GObject *obj;

    {
        /*
         * Invoke parent constructor.
         */
        BmpPreferencesClass *klass;
        GObjectClass *parent_class;
        klass = BMP_PREFERENCES_CLASS (g_type_class_peek (BMP_TYPE_PREFERENCES));
        parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
        obj =
            parent_class->constructor (type,
                                       n_construct_properties,
                                       construct_properties);
    }

    /*
     * do stuff.
     */
    return obj;
}

static void
bmp_preferences_dispose (GObject *object)
{
    BmpPreferences *this = (BmpPreferences *)object;

    signal_handler_pool_free (this->private->pool);
    if (this->private->xml) g_object_unref (this->private->xml);
    if (this->window) gtk_widget_destroy (this->window);
}

static void
bmp_preferences_class_init (BmpPreferencesClass * klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

#if 0
    gobject_class->set_property = bmp_play_set_property;
    gobject_class->get_property = bmp_play_get_property;
    gobject_class->finalize = bmp_play_finalize;
#endif
    gobject_class->dispose     = bmp_preferences_dispose;
    gobject_class->constructor = bmp_preferences_constructor;

}

static void
bmp_preferences_init (BmpPreferences *this)
{
    this->private = g_new (BmpPreferencesPrivate, 1);

    this->private->dispose_has_run = FALSE;
    this->private->skins_path = NULL;
    this->private->skinlist = NULL;
    this->private->host = bmp_plugin_host_new ();
    this->private->updated = FALSE;
    this->private->pool = signal_handler_pool_new ("UiPreferences");

    window_preferences_create (this);
}

/* Our stuff */
static void
change_category(GtkNotebook * notebook,
                GtkTreeSelection * selection)
{
    GtkTreeModel *model;
    GtkTreeIter iter;
    gint index;

    if (!gtk_tree_selection_get_selected(selection, &model, &iter))
        return;

    gtk_tree_model_get(model, &iter, CATEGORY_VIEW_COL_ID, &index, -1);
    gtk_notebook_set_current_page(notebook, index);
}

//FIMXE: Incorporate into xcs too?
static void
titlestring_binding (GObject *object, GValue *value)
{
  g_value_init (value, G_TYPE_STRING);
  g_value_set_string (value, gtk_combo_box_get_active_text (GTK_COMBO_BOX(object)));
}

static void
fontbutton_binding (GObject *object, GValue *value)
{
  g_value_init (value, G_TYPE_STRING);
  g_value_set_string (value, gtk_font_button_get_font_name (GTK_FONT_BUTTON(object)));
}

static void
on_font_button_font_set(GtkFontButton * button,
                                 gpointer data)
{
    const gchar *fontname;

    fontname = gtk_font_button_get_font_name (button);

    /* Now the font-per-skin handling */
    if (xcs_key_fetch_boolean (xcs, "bmp", "font-per-skin"))
    {
	    gchar *basename,
		  *basename_archive;

#ifndef USE_GCONF
	    xcs_domain_new (xcs, "skin-fonts");
#endif

	    basename = 
	      g_path_get_basename (xcs_key_fetch (xcs, "bmp", "skin"));
	    basename_archive = 
	      archive_basename (basename);

	    xcs_key_set_string (xcs, "skin-fonts", basename_archive, fontname);
    }
    else
    {
	    xcs_key_set_string (xcs, "bmp", "font-save", fontname);
    }
}

void
on_font_button_realize(GtkFontButton * button,
                                gpointer data)
{
    gtk_font_button_set_font_name(button,  xcs_key_fetch (xcs, "bmp", "font"));
}

static void
on_custom_cursors_toggled (GtkToggleButton *button,
                           gpointer data)
{
    BmpWindowMain *window_main;
    BmpWindowPlaylist *window_playlist;

    window_main = RM_GET_ITEM("windows","window-main", NOLOCK);
    window_playlist = RM_GET_ITEM("windows","window-playlist", NOLOCK);

    if (!window_main) return;
    if (!window_playlist) return;

    bmp_window_main_configure (window_main);
    bmp_window_playlist_configure (window_playlist);
}

static void
font_changed (GtkFontButton	   *button,
              gpointer		    data)
{
    const gchar* value;

    value = xcs_key_fetch (XCS_CONFIGURATION(data), "bmp", "font");

    gtk_font_button_set_font_name(GTK_FONT_BUTTON(button), value);
}

static void
use_custom_cursors_changed (GtkToggleButton      *button,
                            gpointer	          data)
{
    BmpWindowPlaylist *w_pl;
    BmpWindowMain     *w_main;
    gboolean	       value;

    value = xcs_key_fetch_boolean (xcs, "bmp", "use-custom-cursors");
    gtk_toggle_button_set_active (button, value);

    w_pl = RM_GET_ITEM("windows","window-playlist", NOLOCK);
    w_main = RM_GET_ITEM("windows","window-main", NOLOCK);

    bmp_window_main_configure (w_main);
    bmp_window_playlist_configure (w_pl);
}

static void
font_per_skin_changed (GtkToggleButton      *button,
                       gpointer	             data)
{
    gboolean value;

    value = xcs_key_fetch_boolean (xcs, "bmp", "font-per-skin");
    gtk_toggle_button_set_active (button, value);
}

static void
skin_changed (BmpPreferences  *this,
              gpointer	             data)
{
    GtkTreeModel      *model;
    GtkTreeSelection  *selection;
    GtkTreeIter	       iter;
    GtkTreePath	      *path;
    const gchar	      *value;

    model = gtk_tree_view_get_model (GTK_TREE_VIEW(this->private->skin_view));
    selection = gtk_tree_view_get_selection (GTK_TREE_VIEW(this->private->skin_view));
    value = xcs_key_fetch (xcs, "bmp", "skin");
    attr_row_map_mapping_get (map_skins, value, model, &iter);

    gtk_tree_selection_select_iter (selection, &iter); 

    path = gtk_tree_model_get_path (model, &iter);

    gtk_tree_view_scroll_to_cell    (GTK_TREE_VIEW(this->private->skin_view),
                                     path, 
                                     NULL,
				     TRUE,
				     0.5, 0.0); 
    gtk_tree_path_free (path);
}


static void
display_tooltips_changed (GtkToggleButton      *button,
                          gpointer	        data)
{
    BmpTooltips *tooltips;
    gboolean value;

    value = xcs_key_fetch_boolean (xcs, "bmp", "display-tooltips");

    tooltips = RM_GET_ITEM ("tooltips","tooltips-preferences", LOCK);
    if (! value)
	bmp_tooltips_disable (tooltips);
    else
	bmp_tooltips_enable  (tooltips);
    RM_ITEM_UNLOCK("tooltips","tooltips-preferences");

    tooltips = RM_GET_ITEM ("tooltips","tooltips-main", LOCK);
    if (! value)
	bmp_tooltips_disable (tooltips);
    else
	bmp_tooltips_enable  (tooltips);
    RM_ITEM_UNLOCK("tooltips","tooltips-main");

    gtk_toggle_button_set_active (button, value);
}

static void
display_tracklist_numbers_changed (BmpPreferences *button,
				   gpointer	   data)
{
    chroma_list_set_use_numbers (CHROMA_LIST(c_tracklist), xcs_key_fetch_boolean (xcs, "bmp", "display-tracklist-numbers"));
}

static void
on_font_per_skin_toggled (GtkToggleButton      *button,
                          BmpPreferences *this)
{
    const gchar *skin_font;
    gchar *basename, *basename_archive;
    gboolean value;

    value =
      gtk_toggle_button_get_active (button);

    if (value == TRUE)
    {
	const gchar *aux;

	aux = xcs_key_fetch (xcs, "bmp", "font-save");
	if (!aux)
	{
	    aux = xcs_key_fetch (xcs, "bmp", "font");
	    xcs_key_set_string (xcs, "bmp", "font-save", aux);
	}

#ifndef USE_GCONF
	xcs_domain_new (xcs, "skin-fonts");
#endif

	basename = 
	  g_path_get_basename (xcs_key_fetch (xcs, "bmp", "skin"));
	basename_archive = 
	  archive_basename (basename);

	skin_font = xcs_key_fetch (xcs, "skin-fonts", basename_archive);

	if (skin_font)
	{
	    set_font (this, skin_font);
	    xcs_key_set_string (xcs, "bmp", "font", skin_font);
	}
	else
	{
	    skin_font = xcs_key_fetch (xcs, "bmp", "font-save");

	    if (skin_font)
	    {
		set_font (this, skin_font);
		xcs_key_set_string (xcs, "bmp", "font", skin_font);
	    }
	}
    }
    else
    {
	skin_font = xcs_key_fetch (xcs, "bmp", "font-save");
	if (skin_font)
	{
	    set_font (this, skin_font);
	    xcs_key_set_string (xcs, "bmp", "font", skin_font);
	}
    }

}

static void
skin_view_update (BmpPreferences *this)
{
    GtkWidget *dialog_preferences;
    GtkTreeView *treeview = GTK_TREE_VIEW(this->private->skin_view);
    GtkTreeSelection *selection = NULL;
    GtkListStore *store;
    GtkTreeIter iter, iter_current_skin;
    GtkTreePath *path;
    GdkPixbuf *thumbnail;
    const gchar *current_skin;
    gchar *name, *basename = NULL, *current_skin_basename = NULL;
    GList *entry;

    dialog_preferences = glade_xml_get_widget_warn (this->private->xml, "dialog_preferences");

    current_skin = xcs_key_fetch (xcs, "bmp", "skin");
    if (current_skin)
      {
	basename = g_path_get_basename (current_skin);
	current_skin_basename = archive_basename (basename);
	if (!current_skin_basename &&
	    g_file_test (current_skin, G_FILE_TEST_IS_DIR))
	  current_skin_basename = basename;
      }

    bmp_window_set_busy (dialog_preferences);
    gtk_notebook_set_current_page (GTK_NOTEBOOK(this->private->skin_view_notebook), 1);

    store = GTK_LIST_STORE(gtk_tree_view_get_model(treeview));
    gtk_list_store_clear(store);

    update_skinlist(this);

    for (entry = this->private->skinlist; entry; entry = g_list_next(entry)) {
        thumbnail = skin_get_thumbnail(SKIN_NODE(entry->data)->path);

        if (!thumbnail)
            continue;

        name = SKIN_NODE(entry->data)->name;

        gtk_list_store_append(store, &iter);
        gtk_list_store_set(store, &iter,
                           SKIN_VIEW_COL_PREVIEW, thumbnail,
                           SKIN_VIEW_COL_NAME, name,
			   SKIN_VIEW_COL_PATH, SKIN_NODE(entry->data)->path, -1);
        g_object_unref(thumbnail);
	attr_row_map_mapping_insert (map_skins, SKIN_NODE(entry->data)->path, GTK_TREE_MODEL(store), &iter);  

        if ( current_skin_basename && (!strcmp(current_skin_basename,
                    name))) {
            iter_current_skin = iter;
        }

        while (gtk_events_pending())
            gtk_main_iteration();
    }

    selection = gtk_tree_view_get_selection(treeview);
    gtk_tree_selection_select_iter(selection, &iter_current_skin);

    path = gtk_tree_model_get_path(GTK_TREE_MODEL(store), &iter_current_skin);
    gtk_tree_view_scroll_to_cell(treeview, path, NULL, TRUE, 0.5, 0.5);
    gtk_tree_path_free(path);

    gtk_notebook_set_current_page (GTK_NOTEBOOK(this->private->skin_view_notebook), 0);
    bmp_window_set_idle (dialog_preferences);
}

/* Skins */
static gchar *
get_thumbnail_filename (const gchar * path)
{
    gchar *basename, *pngname, *thumbname;

    g_return_val_if_fail(path != NULL, NULL);

    basename = g_path_get_basename(path);
    pngname = g_strconcat(basename, ".png", NULL);

    thumbname = g_build_filename(BMP_PATH_SKIN_THUMB_DIR,
                                 pngname, NULL);

    g_free(basename);
    g_free(pngname);

    return thumbname;
}


static GdkPixbuf *
skin_get_preview (const gchar * path)
{
    GdkPixbuf *preview = NULL;
    gchar *dec_path, *preview_path;
    gboolean is_archive = FALSE;

    if (file_is_archive(path))
      {
        if (!(dec_path = archive_decompress(path)))
            return NULL;

        is_archive = TRUE;
      }
    else
      {
        dec_path = g_strdup(path);
      }

    preview_path = find_file_recursively (dec_path, "main.bmp", 1);

    if (preview_path)
      {
        preview = gdk_pixbuf_new_from_file (preview_path, NULL);
        g_free(preview_path);
      }
    else
      {
	// g_log (G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "%s: Unable to find main.bmp in path '%s'", G_STRLOC, dec_path);
      }

    if (is_archive)
        del_directory(dec_path);

    g_free(dec_path);

    return preview;
}


static GdkPixbuf *
skin_get_thumbnail (const gchar * path)
{
    GdkPixbuf *scaled = NULL;
    GdkPixbuf *preview;
    gchar *thumbname;

    g_return_val_if_fail (path != NULL, NULL);

    if (g_str_has_suffix (path, BMP_SKIN_THUMB_DIR_BASENAME))
        return NULL;

    thumbname = get_thumbnail_filename (path);

    if (g_file_test(thumbname, G_FILE_TEST_EXISTS))
      {
        scaled = gdk_pixbuf_new_from_file (thumbname, NULL);
        g_free(thumbname);
        return scaled;
    }

    if (!(preview = skin_get_preview(path)))
      {
        g_free(thumbname);
        return NULL;
      }

    scaled = gdk_pixbuf_scale_simple(preview,
                                     THUMBNAIL_WIDTH, THUMBNAIL_HEIGHT,
                                     GDK_INTERP_BILINEAR);
    g_object_unref(preview);

    gdk_pixbuf_save(scaled, thumbname, "png", NULL, NULL);
    g_free(thumbname);

    return scaled;
}

static void
skinlist_add (BmpPreferences *this, const gchar * filename)
{
    SkinNode *node;
    gchar *basename;

    g_return_if_fail(filename != NULL);

    node = g_new0(SkinNode, 1);
    node->path = g_strdup(filename);

    basename = g_path_get_basename(filename);

    if (file_is_archive(filename))
      {
        node->name = archive_basename(basename);
        g_free(basename);
      }
    else
      {
        node->name = basename;
      }


    this->private->skinlist = g_list_prepend(this->private->skinlist, node);
}

static gboolean
scan_skindir_func(const gchar * path, const gchar * basename, BmpPreferences *this)
{

    if (g_file_test(path, G_FILE_TEST_IS_REGULAR)) {
        if (file_is_archive(path)) {
            skinlist_add(this, path);
        }
    }
    else if (g_file_test(path, G_FILE_TEST_IS_DIR)) {
        skinlist_add(this, path);
    }

    return FALSE;
}

static void
scan_skindir(BmpPreferences *this, const gchar * path)
{
    GError *error = NULL;

    g_return_if_fail(path != NULL);

    if (path[0] == '.')
        return;

    if (!dir_foreach(path, (DirForeachFunc)scan_skindir_func, this, &error)) {
        g_warning("Failed to open directory (%s): %s", path, error->message);
        g_error_free(error);
        return;
    }
}

#if 0
static gint
skinlist_compare_func(gconstpointer a, gconstpointer b)
{
    g_return_val_if_fail(a != NULL && SKIN_NODE(a)->name != NULL, 1);
    g_return_val_if_fail(b != NULL && SKIN_NODE(b)->name != NULL, 1);
    return g_ascii_strcasecmp(SKIN_NODE(a)->name, SKIN_NODE(b)->name);
}
#endif

static void
skin_free_func(gpointer data)
{
    g_return_if_fail(data != NULL);
    g_free(SKIN_NODE(data)->name);
    g_free(SKIN_NODE(data)->path);
    g_free(data);
}

static void
skinlist_clear(BmpPreferences *this)
{
    if (!this->private->skinlist)
        return;

    g_list_foreach(this->private->skinlist, (GFunc) skin_free_func, NULL);
    g_list_free(this->private->skinlist);
    this->private->skinlist = NULL;
}

void
update_skinlist(BmpPreferences *this)
{
    skinlist_clear(this);

    scan_skindir(this, DATA_DIR G_DIR_SEPARATOR_S BMP_SKIN_DIR_BASENAME);
    scan_skindir(this, BMP_PATH_USER_SKIN_DIR);

    g_assert(this->private->skinlist != NULL);
}

static void
skin_view_on_cursor_changed(GtkTreeView * treeview,
                            BmpPreferences *this)
{
    GtkTreeModel *model;
    GtkTreeSelection *selection;
    GtkTreeIter iter;
    const gchar *aux;
    gchar *path;

    selection = gtk_tree_view_get_selection(treeview);
    if (!gtk_tree_selection_get_selected(selection, &model, &iter))
        return;

    gtk_tree_model_get(model, &iter, SKIN_VIEW_COL_PATH, &path, -1);

    aux = xcs_key_fetch (xcs, "bmp", "skin");
    if(!strcmp(aux, path))
	return;

    /* This will invoke the signal handler in bmp_ui.c for the configuration's skin path value */
    xcs_key_set_string (xcs, "bmp", "skin", path);

    /* Now the font-per-skin handling */
    if (xcs_key_fetch_boolean (xcs, "bmp", "font-per-skin"))
    {
	    const gchar *skin_font;
	    gchar *basename, *basename_archive;

#ifndef USE_GCONF
	    xcs_domain_new (xcs, "skin-fonts");
#endif

	    basename = g_path_get_basename (path);
	    basename_archive = archive_basename (basename);

	    skin_font = xcs_key_fetch(xcs, "skin-fonts", basename_archive);

	    if (skin_font)
	    {
		set_font (this, skin_font);
		xcs_key_set_string (xcs, "bmp", "font", skin_font);
	    }
	    else
	    {
		skin_font = xcs_key_fetch (xcs, "bmp", "font-save");
		if (skin_font)
		{
		    set_font (this, skin_font);
		    xcs_key_set_string (xcs, "bmp", "font", skin_font);
		}
	    }
	    g_free (basename);
	    g_free (basename_archive);
    }

}

static gint
skin_list_sort_func (GtkTreeModel *model,
                     GtkTreeIter *iter_a,
                     GtkTreeIter *iter_b,
                     gpointer user_data)
{
    gchar *string_a, *string_b;
    gint cmp_val;

    gtk_tree_model_get(model, iter_a, SKIN_VIEW_COL_NAME, &string_a, -1);
    gtk_tree_model_get(model, iter_b, SKIN_VIEW_COL_NAME, &string_b, -1);

    cmp_val = g_ascii_strcasecmp(string_a, string_b);

    return cmp_val;
}

static void
skin_added_func (const gchar *path, BmpPreferences *this)
{
    GtkListStore *store = GTK_LIST_STORE(gtk_tree_view_get_model(GTK_TREE_VIEW(this->private->skin_view)));
    GdkPixbuf *thumbnail;
    GtkTreeIter iter;
    gchar *name, *basename;

    thumbnail = skin_get_thumbnail(path);
    if (!thumbnail)
	    return;
    
    basename = g_path_get_basename(path);

    if (file_is_archive(path))
      {
        name = archive_basename(basename);
        g_free(basename);
      }
    else
      {
        name = basename;
      }

    gtk_list_store_append(store, &iter);
    gtk_list_store_set(store, &iter,
	    SKIN_VIEW_COL_PREVIEW, thumbnail,
	    SKIN_VIEW_COL_NAME, name,
	    SKIN_VIEW_COL_PATH, path, -1);
    g_object_unref(thumbnail);
    attr_row_map_mapping_insert (map_skins, path, GTK_TREE_MODEL(store), &iter);  

#if 0
    g_free (name);
    g_free (path);
#endif
}

static void
skin_removed_func (const gchar *path, BmpPreferences *this)
{
	GtkTreeModel *store = gtk_tree_view_get_model(GTK_TREE_VIEW(this->private->skin_view));
	GtkTreeIter iter;
	gchar *name;
	gboolean loop = TRUE;

	if (!gtk_tree_model_get_iter_first (store, &iter)) return;

	while (loop)
	  {
		gtk_tree_model_get (store, &iter, SKIN_VIEW_COL_PATH, &name, -1);
		if (!strcmp(path,name))
		  {
			gchar *path;

		        gtk_tree_model_get (GTK_TREE_MODEL(store), &iter, SKIN_VIEW_COL_PATH, &path, -1);
			attr_row_map_mapping_remove (map_skins, path);
			gtk_list_store_remove (GTK_LIST_STORE(store), &iter);
#if 0
			g_free (path);
			g_free (name);
#endif
			loop = FALSE;
		  }
		else
		  {
			if (!gtk_tree_model_iter_next (store, &iter)) loop = FALSE;
		  }
	  }
#if 0
       g_free(path);
#endif
}

static void
column_toggled  (GtkCellRendererToggle *cell,
                 gchar                 *path_str,
                 BmpPreferences        *this)
{
    BmpWindowPlaylist *window_playlist;
    GtkTreePath	      *path;
    GtkTreeIter	       iter;
    gboolean	       new_state;
    gint	       column;

    window_playlist = RM_GET_ITEM("windows","window-playlist", NOLOCK);

    path = gtk_tree_path_new_from_string (path_str); 
    gtk_tree_model_get_iter (GTK_TREE_MODEL(this->private->columns_store), &iter, path);
    column = gtk_tree_path_get_indices (path)[0];
    gtk_tree_path_free (path);

    new_state = gtk_cell_renderer_toggle_get_active (cell) ? FALSE : TRUE;
    gtk_list_store_set (this->private->columns_store, &iter, 0, new_state, -1);
    bmp_window_playlist_column_set_active (window_playlist, column, new_state); 
}

static void
columns_view_setup (BmpPreferences *this)
{
    GtkTreeView *treeview = GTK_TREE_VIEW(this->private->columns_view);
    GtkTreeViewColumn *column;
    GtkCellRenderer *renderer;
    GtkTreeIter iter;
    gint n;

    this->private->columns_store = gtk_list_store_new(3, G_TYPE_BOOLEAN, G_TYPE_STRING, G_TYPE_INT);

    for (n = 0; n < vcolumns_get_n_items (); n++)
      {
	  gchar	    *key_name;
	  gboolean   visible;

	  key_name = g_strdup_printf ("column_%d_visible", n);
	  visible = xcs_key_fetch_boolean (xcs, "tracklist", key_name);
	  g_free (key_name);
 
	  gtk_list_store_append (this->private->columns_store, &iter);
	  gtk_list_store_set (this->private->columns_store, &iter, 0, visible, 1, bmp_metadata_get_title_static (vcolumns_get_item_nth(n)), -1);
      }

    column = gtk_tree_view_column_new ();
    renderer = gtk_cell_renderer_toggle_new ();
    gtk_tree_view_column_pack_start (column, renderer, FALSE);
    gtk_tree_view_column_set_attributes (column, renderer, "active", 0, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(treeview), GTK_TREE_VIEW_COLUMN(column));

    signal_handler_pool_add_connection (this->private->pool,
					renderer,	
					"toggled",
					G_CALLBACK(column_toggled),
					this,	
					0);	


    column = gtk_tree_view_column_new ();
    renderer = gtk_cell_renderer_text_new();
    gtk_tree_view_column_pack_start(column, renderer, FALSE);
    gtk_tree_view_column_set_attributes(column, renderer, "text", 1, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(treeview), GTK_TREE_VIEW_COLUMN(column));

    gtk_tree_view_set_model(treeview, GTK_TREE_MODEL(this->private->columns_store));
}


static void
skin_view_setup (BmpPreferences *this)
{
    GtkTreeView *treeview = GTK_TREE_VIEW(this->private->skin_view);
    GtkListStore *store;
    GtkTreeViewColumn *column;
    GtkCellRenderer *renderer;
    GtkTreeSelection *selection;

    gtk_widget_show_all(GTK_WIDGET(treeview));

    gtk_tree_view_set_rules_hint(GTK_TREE_VIEW(treeview), TRUE);
    gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(treeview), FALSE);

    store = gtk_list_store_new(SKIN_VIEW_N_COLS, GDK_TYPE_PIXBUF,
                               G_TYPE_STRING, G_TYPE_STRING);
    gtk_tree_view_set_model(treeview, GTK_TREE_MODEL(store));

    map_skins = attr_row_map_new ();

    column = gtk_tree_view_column_new();
    gtk_tree_view_column_set_sizing(column, GTK_TREE_VIEW_COLUMN_AUTOSIZE);
    gtk_tree_view_column_set_spacing(column, 16);
    gtk_tree_view_append_column(GTK_TREE_VIEW(treeview),
                                GTK_TREE_VIEW_COLUMN(column));

    renderer = gtk_cell_renderer_pixbuf_new();
    gtk_tree_view_column_pack_start(column, renderer, FALSE);
    gtk_tree_view_column_set_attributes(column, renderer, "pixbuf",
                                        SKIN_VIEW_COL_PREVIEW, NULL);
    g_object_set (G_OBJECT(renderer),"xpad",6,NULL);
    g_object_set (G_OBJECT(renderer),"ypad",2,NULL);

    renderer = gtk_cell_renderer_text_new();
    gtk_tree_view_column_pack_start(column, renderer, TRUE);
    gtk_tree_view_column_set_attributes(column, renderer, "text",
                                        SKIN_VIEW_COL_NAME, NULL);

    selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeview));
    gtk_tree_selection_set_mode(selection, GTK_SELECTION_BROWSE);

    gtk_tree_sortable_set_sort_func (GTK_TREE_SORTABLE(store),
                                     SKIN_VIEW_COL_NAME,
                                     skin_list_sort_func,
                                     NULL,
                                     NULL);

    gtk_tree_sortable_set_sort_column_id (GTK_TREE_SORTABLE(store), SKIN_VIEW_COL_NAME, GTK_SORT_ASCENDING);

    signal_handler_pool_add_connection (this->private->pool,
					treeview,	
					"cursor-changed",
					G_CALLBACK(skin_view_on_cursor_changed),
					this,	
					0);	
}

/* Plugins */
static void
on_button_plugins_show_all  (GtkWidget   *button,
			     gpointer	  data)
{
    BmpPreferences  *this = BMP_PREFERENCES (data);
    GtkWidget	    *notebook;

    notebook = glade_xml_get_widget_warn (this->private->xml, "plugins_notebook_main");

    gtk_notebook_set_current_page (GTK_NOTEBOOK(notebook), 0);
    gtk_widget_set_sensitive (button, FALSE);
}

static void 
on_plugins_icon_view_item_activated (GtkIconView *iconview,
                                     GtkTreePath *path,
                                     gpointer	  data)
{
    BmpPreferences  *this = BMP_PREFERENCES (data);
    GtkWidget	    *button, *notebook;

    notebook = glade_xml_get_widget_warn (this->private->xml, "plugins_notebook_main");
    button = glade_xml_get_widget_warn (this->private->xml, "b_plugins_show_all");

    gtk_notebook_set_current_page (GTK_NOTEBOOK(notebook),
				   gtk_tree_path_get_indices (path)[0]+1); 
    gtk_widget_set_sensitive (button, TRUE);
}

typedef struct _PluginActivationTuple PluginActivationTuple;
struct _PluginActivationTuple {
    BmpPreferences    *this;
    PluginType	       type;
};

static void 
on_plugin_sub_icon_view_item_activated (GtkIconView	      *iconview,
					GtkTreePath	      *path,
					PluginActivationTuple *tuple) 
{
  GtkTreeModel *model = gtk_icon_view_get_model (iconview);
  GtkTreeIter   iter;
  gchar	       *name;
  
  gtk_tree_model_get_iter (model, &iter, path);
  gtk_tree_model_get (model, &iter, 2, &name, -1);

  bmp_plugin_host_gui_show (tuple->this->private->host,
				     tuple->type, 
				     name); 
}

static void
append_to_list (gpointer _k, gpointer _v, GList **list)
{
    if (BMP_IS_PLUG_PROPERTY(_v))
	  *list = g_list_append (*list, _k); 
}

static void
icon_view_set_bmpx_defaults (GtkIconView *icon_view)
{
    gtk_icon_view_set_text_column   (icon_view, 0);
    gtk_icon_view_set_pixbuf_column (icon_view, 1);
    gtk_icon_view_set_item_width    (icon_view, 128);
    gtk_icon_view_set_orientation   (icon_view, GTK_ORIENTATION_VERTICAL);
    gtk_icon_view_set_columns	    (icon_view, -1);
}

static void
on_plugins_iconview_realize (GtkIconView *icon_view,
			     gpointer	  data)
{
    BmpPreferences	    *this = BMP_PREFERENCES (data);
    GtkListStore	    *icon_store, *icon_store_sub; 
    GtkWidget		    *icon_view_sub;
    GdkPixbuf		    *pb;
    GtkTreeIter		     iter;
    PluginActivationTuple   *tuple;
    GList		    *plugins = NULL;
    gint		     n;

    g_return_if_fail (GTK_IS_ICON_VIEW (icon_view));  

    icon_store = gtk_list_store_new (3, G_TYPE_STRING, GDK_TYPE_PIXBUF, G_TYPE_INT);
    for (n = 0; n < BMP_N_PLUGIN_TYPES; n++)
      {
	pb = gdk_pixbuf_new_from_file (DATA_DIR "/images/preferences/plugins.png", NULL); 
	gtk_list_store_append (icon_store, &iter);	
	gtk_list_store_set (icon_store, &iter, 0, plugin_type_get_desc (n), 1, pb, 2, n, -1);

	g_object_unref (pb);
      }

    gtk_icon_view_set_model (icon_view, GTK_TREE_MODEL(icon_store));
    icon_view_set_bmpx_defaults (icon_view);

    signal_handler_pool_add_connection (this->private->pool,
					icon_view,	
					"item-activated",
					G_CALLBACK(on_plugins_icon_view_item_activated),
					this,	
					0);	

    signal_handler_pool_add_connection (this->private->pool,
					glade_xml_get_widget_warn (this->private->xml, "b_plugins_show_all"),
					"clicked",
					G_CALLBACK(on_button_plugins_show_all),
					this,	
					0);	

    /* Now set up the sub-icon-views */
    //Transport
    icon_view_sub = glade_xml_get_widget_warn (this->private->xml, "iconview_transport");    
    icon_store_sub = gtk_list_store_new (3, G_TYPE_STRING, GDK_TYPE_PIXBUF, G_TYPE_STRING);
    g_hash_table_foreach (plugins_table[BMP_PLUGIN_TYPE_TRANSPORT], (GHFunc)append_to_list, &plugins);

    for ( ; plugins ; plugins = plugins->next)
      {
	GObject	     *object;
	const gchar **metadata;
    
	object = g_hash_table_lookup (plugins_table[BMP_PLUGIN_TYPE_TRANSPORT], plugins->data);
	metadata = bmp_plug_metadata_get (BMP_PLUG_METADATA (object));
	pb = gdk_pixbuf_new_from_file (DATA_DIR "/images/preferences/plugins.png", NULL); 

	gtk_list_store_append (icon_store_sub, &iter);	
	gtk_list_store_set (icon_store_sub, &iter, 0, metadata[METADATA_FIELD_NAME], 1, pb, 2, (gchar*)plugins->data, -1);

	bmp_plugin_host_create_gui_for_plugin (this->private->host,
							BMP_PLUGIN_TYPE_TRANSPORT,	
							(const gchar *)plugins->data);	

	g_object_unref (pb);
      }

    gtk_icon_view_set_model (GTK_ICON_VIEW(icon_view_sub), GTK_TREE_MODEL(icon_store_sub));
    icon_view_set_bmpx_defaults (GTK_ICON_VIEW(icon_view_sub));

    tuple = g_new0 (PluginActivationTuple,1);
    tuple->this = this;
    tuple->type = BMP_PLUGIN_TYPE_TRANSPORT;

    signal_handler_pool_add_connection (this->private->pool,
					icon_view_sub,	
					"item-activated",
					G_CALLBACK(on_plugin_sub_icon_view_item_activated),
					tuple,	
					0);	

    //Flow
    plugins = NULL;
    icon_view_sub = glade_xml_get_widget_warn (this->private->xml, "iconview_flow");    
    icon_store_sub = gtk_list_store_new (3, G_TYPE_STRING, GDK_TYPE_PIXBUF, G_TYPE_STRING);
    g_hash_table_foreach (plugins_table[BMP_PLUGIN_TYPE_FLOW], (GHFunc)append_to_list, &plugins);

    for ( ; plugins ; plugins = plugins->next)
      {
	GObject	     *object;
	const gchar **metadata;
    
	object = g_hash_table_lookup (plugins_table[BMP_PLUGIN_TYPE_FLOW], plugins->data);
	metadata = bmp_plug_metadata_get (BMP_PLUG_METADATA (object));
	pb = gdk_pixbuf_new_from_file (DATA_DIR "/images/preferences/plugins.png", NULL); 

	gtk_list_store_append (icon_store_sub, &iter);	
	gtk_list_store_set (icon_store_sub, &iter, 0, metadata[METADATA_FIELD_NAME], 1, pb, 2, (gchar*)plugins->data, -1);

	bmp_plugin_host_create_gui_for_plugin (this->private->host,
							BMP_PLUGIN_TYPE_FLOW,	
							(const gchar *)plugins->data);	

	g_object_unref (pb);
      }

    gtk_icon_view_set_model (GTK_ICON_VIEW(icon_view_sub), GTK_TREE_MODEL(icon_store_sub));
    icon_view_set_bmpx_defaults (GTK_ICON_VIEW(icon_view_sub));

    tuple = g_new0 (PluginActivationTuple,1);
    tuple->this = this;
    tuple->type = BMP_PLUGIN_TYPE_FLOW;

    signal_handler_pool_add_connection (this->private->pool,
					icon_view_sub,	
					"item-activated",
					G_CALLBACK(on_plugin_sub_icon_view_item_activated),
					tuple,	
					0);	
}

static void
on_category_view_realize (GtkTreeView	  *treeview,
                          BmpPreferences  *this)
{
    GtkListStore      *store;
    GtkCellRenderer   *renderer;
    GtkTreeViewColumn *column;
    GtkTreeSelection  *selection;
    GtkWidget	      *notebook;
    GdkPixbuf	      *img;
    GtkTreeIter	       iter;
    gint	       i;

    notebook = glade_xml_get_widget_warn (this->private->xml, "category_notebook");
    column = gtk_tree_view_column_new();
    gtk_tree_view_append_column(treeview, column);
    gtk_tree_view_column_set_spacing(column, 2);

    renderer = gtk_cell_renderer_pixbuf_new();
    gtk_tree_view_column_pack_start(column, renderer, FALSE);
    g_object_set (G_OBJECT(renderer),"xpad",2,NULL);
    g_object_set (G_OBJECT(renderer),"ypad",2,NULL);
    gtk_tree_view_column_set_attributes(column, renderer, "pixbuf", CATEGORY_VIEW_COL_ICON, NULL);

    renderer = gtk_cell_renderer_text_new();
    gtk_tree_view_column_pack_start(column, renderer, FALSE);
    g_object_set (G_OBJECT(renderer),"xpad",2,NULL);
    g_object_set (G_OBJECT(renderer),"ypad",2,NULL);
    gtk_tree_view_column_set_attributes(column, renderer, "text", CATEGORY_VIEW_COL_NAME, NULL);


    store = gtk_list_store_new (CATEGORY_VIEW_N_COLS,
                                GDK_TYPE_PIXBUF, 
				G_TYPE_STRING, 
				G_TYPE_INT);

    gtk_tree_view_set_model(treeview, GTK_TREE_MODEL(store));

    for (i = 0; i < n_categories; i++) {
        img = gdk_pixbuf_new_from_file(categories[i].icon_path, NULL);
        gtk_list_store_append(store, &iter);
        gtk_list_store_set(store, &iter,
                           CATEGORY_VIEW_COL_ICON,
			   img,
                           CATEGORY_VIEW_COL_NAME,
                           gettext(categories[i].name),
			   CATEGORY_VIEW_COL_ID,
                           categories[i].id,
			   -1);
        g_object_unref(img);
    }

    gtk_tree_view_set_headers_visible (treeview, FALSE);

    selection = gtk_tree_view_get_selection(treeview);

    signal_handler_pool_add_connection (this->private->pool,
					selection,	
					"changed",
					G_CALLBACK(change_category),
					notebook,	
					G_CONNECT_SWAPPED);	
}

/* FIXME: complete the map */
FUNC_MAP_BEGIN(prefswin_func_map)
    FUNC_MAP_ENTRY(on_custom_cursors_toggled)
    FUNC_MAP_ENTRY(on_font_per_skin_toggled)
    FUNC_MAP_ENTRY(on_font_button_font_set)
    FUNC_MAP_ENTRY(on_font_button_realize)
    FUNC_MAP_ENTRY(on_plugins_iconview_realize)
FUNC_MAP_END

#if 0
static void
on_eq_value_changed (GtkWidget *widget, gpointer user_data)
{
    gdouble value;
    gint value_int;

    g_return_if_fail (GTK_IS_WIDGET(widget));

    value = gtk_range_get_value (GTK_RANGE(widget));
    value_int = (gint)value;

    bmp_system_control_eq_set (bmp_system_control, GPOINTER_TO_INT(user_data), value_int, NULL);
}
#endif

static void
lastfm_enable_changed	    (BmpPreferences *this,
			     gpointer	     data) 
{
    GtkWidget *checkbutton;
    gboolean   active;

    active = xcs_key_fetch_boolean (xcs, "lastfm", "enable");
    checkbutton = glade_xml_get_widget_warn (this->private->xml, "last_fm_enable");

    if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (checkbutton)) != active)
        gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (checkbutton), active);
}

static void
lastfm_general_enable_changed	    (BmpPreferences *this,
				     gpointer	     data) 
{
    GtkWidget *widget;
    gboolean   active;
    gint       n;

    static gchar* widget_names[]={
	"table_lastfm_credentials",
	"hbox_lastfm_explanation",
	"last_fm_enable",
	"label_lastfm_login"
    };

    active = xcs_key_fetch_boolean (xcs, "lastfm", "general-enable");

    for (n = 0; n < G_N_ELEMENTS(widget_names); n++)
      {
	widget = glade_xml_get_widget_warn (this->private->xml, widget_names[n]); 
	gtk_widget_set_sensitive (widget, active);
      }
}


static void
on_cbox_audio_system_changed	(GtkWidget	*cbox,
				 BmpPreferences *this)
{
    GtkWidget *widget;
    GtkTreeModel *model;
    GtkTreePath *path;
    GtkTreeIter iter;
    gint active, tab_id;

    active = gtk_combo_box_get_active (GTK_COMBO_BOX(cbox));
    model = gtk_combo_box_get_model (GTK_COMBO_BOX(cbox));

    path = gtk_tree_path_new_from_indices (active, -1);
    gtk_tree_model_get_iter (model, &iter, path);
    gtk_tree_path_free (path);

    gtk_tree_model_get (model, &iter, 2, &tab_id, -1); 

    widget = glade_xml_get_widget_warn (this->private->xml, "notebook_audio_system");
    gtk_notebook_set_current_page (GTK_NOTEBOOK(widget), tab_id); 
}

#ifdef HAVE_ALSA
static void
on_cbox_alsa_card_changed	(GtkWidget	*cbox,
				 BmpPreferences *this)
{
    GtkWidget *cbox_alsa_device;
    GtkTreeModel *model;
    GtkListStore *store;
    GtkTreePath *path;
    GtkTreeIter iter;
    BmpAlsaCard *bmp_alsa_card;
    GList *devices;
    gint active;

    active = gtk_combo_box_get_active (GTK_COMBO_BOX(cbox));
    model = gtk_combo_box_get_model (GTK_COMBO_BOX(cbox));

    cbox_alsa_device = glade_xml_get_widget_warn (this->private->xml, "cbox_alsa_device"); 
    store = gtk_list_store_new (2, G_TYPE_STRING, BMP_TYPE_ALSA_DEVICE);
    gtk_combo_box_set_model (GTK_COMBO_BOX(cbox_alsa_device), GTK_TREE_MODEL(store));

    if (active == 0)
      {
	  gtk_widget_set_sensitive (GTK_WIDGET(cbox_alsa_device), FALSE);
      }
    else
      {
	path = gtk_tree_path_new_from_indices (active, -1);
	gtk_tree_model_get_iter (model, &iter, path);
	gtk_tree_path_free (path);

	gtk_tree_model_get (model, &iter, 1, &bmp_alsa_card, -1); 

	devices = bmp_alsa_card->devices; 

	for ( ; devices ; devices = devices->next )
	  {
	    BmpAlsaDevice *device;

	    device = BMP_ALSA_DEVICE(devices->data);

	    gtk_list_store_append (GTK_LIST_STORE(store), &iter);
	    gtk_list_store_set (GTK_LIST_STORE(store), &iter,
				0,
				device->description,
				1,
				device,
				-1); 
	  }

	gtk_widget_set_sensitive (GTK_WIDGET(cbox_alsa_device), TRUE);
	gtk_combo_box_set_active (GTK_COMBO_BOX(cbox_alsa_device), 0);
      }
}
#endif

static void
set_apply_sensitive	(GtkWidget	*cbox,
			 BmpPreferences *this)
{
    gtk_widget_set_sensitive (glade_xml_get_widget_warn (this->private->xml, "b_audio_system_apply"), TRUE);
    gtk_widget_set_sensitive (glade_xml_get_widget_warn (this->private->xml, "hbox_warning"), TRUE);
}

static void
on_b_audio_system_apply	 (GtkWidget	  *widget,
			  BmpPreferences  *this)
{
    GtkWidget	  *cbox_audio_system,
		  *cbox_device;

    gchar	  *sink,
		  *device = NULL;
    GtkTreeModel  *model;
    GtkTreeIter	   iter;

    gtk_widget_set_sensitive (glade_xml_get_widget_warn (this->private->xml, "b_audio_system_apply"), FALSE);
    gtk_widget_set_sensitive (glade_xml_get_widget_warn (this->private->xml, "hbox_warning"), FALSE);

    cbox_audio_system = glade_xml_get_widget_warn (this->private->xml, "cbox_audio_system");
    model = gtk_combo_box_get_model (GTK_COMBO_BOX(cbox_audio_system));
    gtk_combo_box_get_active_iter (GTK_COMBO_BOX(cbox_audio_system), &iter);

    gtk_tree_model_get (model, &iter, 1, &sink, -1);

    //FIXME: Why do we do string comparisons here?
    if (!strcmp(sink,"alsasink"))
      {
#ifdef HAVE_ALSA
	  BmpAlsaDevice *bmp_alsa_device;
	  gint active;

	  cbox_device = glade_xml_get_widget_warn (this->private->xml, "cbox_alsa_device");
	  model = gtk_combo_box_get_model (GTK_COMBO_BOX(cbox_device));
	  active = gtk_combo_box_get_active (GTK_COMBO_BOX(cbox_device));

	  if (active > 0)
	    {
	      gtk_combo_box_get_active_iter (GTK_COMBO_BOX(cbox_device), &iter);
	      gtk_tree_model_get (model, &iter, 1, &bmp_alsa_device, -1);
	      device = g_strdup (bmp_alsa_device->dev);
	    }
	  else
	    {
	      device = g_strdup ("default");
	    }
#endif
      }
    else if (!strcmp(sink,"osssink"))
      {
	  cbox_device = glade_xml_get_widget_warn (this->private->xml, "cbe_oss_device");
	  device = gtk_combo_box_get_active_text (GTK_COMBO_BOX(cbox_device));
      }
#ifdef HAVE_SUNAUDIO
    else if (!strcmp(sink,"sunaudiosink"))
      {
	  cbox_device = glade_xml_get_widget_warn (this->private->xml, "cbe_sun_device");
	  device = gtk_combo_box_get_active_text (GTK_COMBO_BOX(cbox_device));
      }
#endif

    g_object_set (G_OBJECT(bmp_play_engine), "playstatus", PLAYSTATUS_STOPPED, NULL);

    xcs_key_set_string (xcs, "audio", "sink", sink);
    xcs_key_set_string (xcs, "audio", "device", device);

    g_free (sink);
    g_free (device);

    bmp_play_reset_pipes (bmp_play_engine);
}
 
void
window_preferences_create (BmpPreferences *this)
{
    BmpTooltips		*tooltips;
    GtkWidget		*widget;
    const gchar		*glade_file = DATA_DIR "/glade/dialog_preferences.glade";
    gint		 n;

    /* o Initialize tooltips */
    tooltips = bmp_tooltips_new ();

    RM_ITEM_NEW("tooltips", "tooltips-preferences", tooltips, NULL);

    /* o Load the interface */
    this->private->xml = glade_xml_new_or_die("Preferences Window", glade_file, NULL, NULL);

    /* o Connect the signals in the interface */
    glade_xml_signal_autoconnect_map (this->private->xml, prefswin_func_map, this);

    this->window = glade_xml_get_widget_warn (this->private->xml, "dialog_preferences");

    /* o Set icon to window */
    bmp_window_set_icon_list (GTK_WIDGET(this->window), "player");

    /* o Create Category View */
    widget = glade_xml_get_widget_warn (this->private->xml, "category_view");

    signal_handler_pool_add_connection (this->private->pool,
					widget,	
					"realize",
					G_CALLBACK(on_category_view_realize),
					this,	
					G_CONNECT_AFTER);	

    /*  o Skin View */
    this->private->skin_view = glade_xml_get_widget_warn (this->private->xml, "skin_view");
    this->private->skin_view_notebook = glade_xml_get_widget_warn (this->private->xml, "skins_notebook");
    skin_view_setup (this);

    /*  o Columns Treeview */
    this->private->columns_view = glade_xml_get_widget_warn (this->private->xml, "columns_view");
    columns_view_setup (this);
 
    /* o Titlestring */
    widget = glade_xml_get_widget_warn (this->private->xml, "titlestring_entry");
    gtk_entry_set_text (GTK_ENTRY(GTK_BIN(widget)->child), xcs_key_fetch(xcs,"bmp","titlestring-format"));

    xcs_bind_object (xcs,
		     G_OBJECT(widget),
		     titlestring_binding,
		     "changed",
		     "bmp",
		     "titlestring-format"); 

    /* o Images */
    widget = glade_xml_get_widget_warn (this->private->xml, "image_building");
    gtk_image_set_from_file (GTK_IMAGE(widget), DATA_DIR "/images/wait.gif");

    widget = glade_xml_get_widget_warn (this->private->xml, "last_fm_header_logo");
    gtk_image_set_from_file (GTK_IMAGE(widget), DATA_DIR "/images/lastfm-logo.png");

    widget = glade_xml_get_widget_warn (this->private->xml, "last_fm_poweredby");
    gtk_image_set_from_file (GTK_IMAGE(widget), DATA_DIR "/images/audioscrobbler.png");


#if 0
    /* Connect eq sliders */
    for (i = 0; i < n_eq_sliders; i++)
      {
	widget = glade_xml_get_widget_warn (this->private->xml, eq_sliders[i]);
	g_signal_connect (G_OBJECT(widget), "value-changed", G_CALLBACK(on_eq_value_changed), GINT_TO_POINTER(i));
      }
#endif

    for (n = 0; n < n_window_preferences_tooltips; n++)
      {
	widget = glade_xml_get_widget_warn (this->private->xml, window_preferences_tooltips[n].widget_name);
	bmp_tooltips_set_tip (tooltips,
				widget,
				window_preferences_tooltips[n].stock_id,
				gettext (window_preferences_tooltips[n].tooltip_text),
			        NULL,
				TRUE);
      }

    widget = glade_xml_get_widget_warn (this->private->xml, "label_version");
    gtk_label_set_markup (GTK_LABEL(widget), 
		  "<span size=\"small\" color=\"#606060\">BMPx " RV_LAST_CHANGED_DATE "-R" RV_REVISION " [" BUILD_DATE "] (" BUILD_BUILDUSER ")</span>"); 

    /* Bind widgets/other objects */ 
    static gchar* toggle_buttons[] = {
      "ui-esc-trayconify",
      "library-add-automatically",
      "library-no-incomplete-add",
      "resume-on-startup",
      "use-custom-cursors",
      "display-tooltips",
      "display-tracklist-numbers",
      "font-per-skin"
    };
  
    for (n = 0; n < G_N_ELEMENTS(toggle_buttons); n++)
      {
	  xcs_bind_toggle_button (xcs,
				  GTK_TOGGLE_BUTTON(glade_xml_get_widget_warn (this->private->xml, toggle_buttons[n])),
				  "bmp",
				  toggle_buttons[n]); 
      }


    xcs_bind_object (xcs,
		     G_OBJECT(glade_xml_get_widget_warn (this->private->xml,"font_button")),
		     fontbutton_binding,
		     "font_set",
		     "bmp",
		     "font"); 

    /* Last.FM stuff */
    widget = glade_xml_get_widget_warn (this->private->xml, "last_fm_username");
    xcs_bind_entry  (xcs,
		     GTK_ENTRY(widget),
		     "lastfm",
		     "username"); 

    widget = glade_xml_get_widget_warn (this->private->xml, "last_fm_password");
    xcs_bind_entry (xcs,
		    GTK_ENTRY(widget),
		     "lastfm",
		     "password"); 

    signal_handler_pool_add_connection (this->private->pool,
					xcs,	
					"changed::lastfm-enable",
					G_CALLBACK(lastfm_enable_changed),
				        this,
					G_CONNECT_SWAPPED);	

    widget = glade_xml_get_widget_warn (this->private->xml, "last_fm_enable");
    xcs_bind_toggle_button (xcs,
			    GTK_TOGGLE_BUTTON(widget),
			    "lastfm",
			    "enable");

    signal_handler_pool_add_connection (this->private->pool,
					xcs,	
					"changed::lastfm-general-enable",
					G_CALLBACK(lastfm_general_enable_changed),
				        this,
					G_CONNECT_SWAPPED);	

    widget = glade_xml_get_widget_warn (this->private->xml, "last_fm_general_enable");
    xcs_bind_toggle_button (xcs,
			    GTK_TOGGLE_BUTTON(widget),
			    "lastfm",
			    "general-enable");

    /* Various watches */
    signal_handler_pool_add_connection (this->private->pool,
					xcs,	
					"changed::bmp-display-tracklist-numbers",
					G_CALLBACK(display_tracklist_numbers_changed),
				        this,
					G_CONNECT_SWAPPED);	

    signal_handler_pool_add_connection (this->private->pool,
					xcs,	
					"changed::bmp-display-tooltips",
					G_CALLBACK(display_tooltips_changed),
					glade_xml_get_widget_warn (this->private->xml, "display-tooltips"),
					G_CONNECT_SWAPPED);	

    signal_handler_pool_add_connection (this->private->pool,
					xcs,	
					"changed::bmp-font",
					G_CALLBACK(font_changed),
					glade_xml_get_widget_warn (this->private->xml, "font_button"),
					G_CONNECT_SWAPPED);	

    signal_handler_pool_add_connection (this->private->pool,
					xcs,	
					"changed::bmp-use-custom-cursors",
					G_CALLBACK(use_custom_cursors_changed),
					glade_xml_get_widget_warn (this->private->xml, "use-custom-cursors"),
					G_CONNECT_SWAPPED);	

    signal_handler_pool_add_connection (this->private->pool,
					xcs,	
					"changed::bmp-font-per-skin",
					G_CALLBACK(font_per_skin_changed),
					glade_xml_get_widget_warn (this->private->xml, "font-per-skin"),
					G_CONNECT_SWAPPED);	

    signal_handler_pool_add_connection (this->private->pool,
					xcs,	
					"changed::bmp-skin",
					G_CALLBACK(skin_changed),
					this,	
					G_CONNECT_SWAPPED);	

    /* Create Audio System prefs */
    {
      GtkListStore *list_store_audio_systems;
      GtkCellRenderer *cell;
      gchar *sink;
      gint current;


      widget = glade_xml_get_widget_warn (this->private->xml, "cbox_audio_system");
      list_store_audio_systems = gtk_list_store_new (3, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT);
      sink = xcs_key_fetch_string (xcs, "audio", "sink");
      current = -1;
    
      for (n = 0; n < G_N_ELEMENTS(audio_systems); n++) 
	{
	    GstElement  *element;
	    GtkTreeIter  iter;
  
	    element = gst_element_factory_make (audio_systems[n].sinkname, "sinktest0"); 
	    if (!element) continue;
	    gst_object_unref (element);

	    gtk_list_store_append (GTK_LIST_STORE(list_store_audio_systems), &iter);
	    gtk_list_store_set (GTK_LIST_STORE(list_store_audio_systems), &iter,
				0, audio_systems[n].description,
				1, audio_systems[n].sinkname,
				2, audio_systems[n].tab_id,
			      -1);

	    if (!strcmp(audio_systems[n].sinkname, sink))
	      {
		current = n;
	      }

	}

      cell = gtk_cell_renderer_text_new ();
      gtk_cell_layout_pack_start (GTK_CELL_LAYOUT(widget), cell, TRUE);
      gtk_cell_layout_set_attributes (GTK_CELL_LAYOUT(widget), cell, "text", 0, NULL);

      gtk_combo_box_set_model (GTK_COMBO_BOX(widget), GTK_TREE_MODEL(list_store_audio_systems));
      gtk_combo_box_set_active (GTK_COMBO_BOX(widget), current);

      signal_handler_pool_add_connection (this->private->pool,
					  widget,	
					  "changed",
					  G_CALLBACK(on_cbox_audio_system_changed),
					  this,	
				          0);	


      widget = glade_xml_get_widget_warn (this->private->xml, "notebook_audio_system");
      if (current == -1)
	{
	  gtk_notebook_set_current_page (GTK_NOTEBOOK(widget), 0);
	}
      else
	{
	  gtk_notebook_set_current_page (GTK_NOTEBOOK(widget), audio_systems[current].tab_id);
	}

#ifdef HAVE_ALSA

	GtkListStore *list_store_cards;
	GtkTreeIter iter;
	GList *cards;

	cards = bmp_alsa_get_cards ();
	list_store_cards = gtk_list_store_new (2, G_TYPE_STRING, BMP_TYPE_ALSA_CARD);

	gtk_list_store_append (list_store_cards, &iter);
	gtk_list_store_set (list_store_cards, &iter, 0, "System Default", 1, NULL, -1);

	for ( ; cards ; cards = cards->next )
	  {
	      BmpAlsaCard *bmp_alsa_card;

	      bmp_alsa_card = BMP_ALSA_CARD(cards->data);
	      gtk_list_store_append (list_store_cards, &iter);
	      gtk_list_store_set (list_store_cards, &iter, 0, bmp_alsa_card->description, 1, bmp_alsa_card, -1);
	  }

	widget = glade_xml_get_widget_warn (this->private->xml, "cbox_alsa_card");
	cell = gtk_cell_renderer_text_new ();
	gtk_cell_layout_pack_start (GTK_CELL_LAYOUT(widget), cell, TRUE);
	gtk_cell_layout_set_attributes (GTK_CELL_LAYOUT(widget), cell, "text", 0, NULL);
	gtk_combo_box_set_model (GTK_COMBO_BOX(widget), GTK_TREE_MODEL(list_store_cards));

        signal_handler_pool_add_connection (this->private->pool,
					  widget,	
					  "changed",
					  G_CALLBACK(on_cbox_alsa_card_changed),
					  this,	
				          0);	

        widget = glade_xml_get_widget_warn (this->private->xml, "cbox_alsa_device"); 
        cell = gtk_cell_renderer_text_new ();
        gtk_cell_layout_pack_start (GTK_CELL_LAYOUT(widget), cell, TRUE);
	gtk_cell_layout_set_attributes (GTK_CELL_LAYOUT(widget), cell, "text", 0, NULL);

	//We have to parse the current device here in case in case sink is alsasink
	if (!strcmp(sink,"alsasink"))
	  {
	      gchar **parts;
	      gchar  *device;

	      widget = glade_xml_get_widget_warn (this->private->xml, "cbox_alsa_card");
	      device = xcs_key_fetch_string (xcs, "audio", "device");

	      if (!strcmp(device,"default"))
		{
		  gtk_combo_box_set_active (GTK_COMBO_BOX(widget), 0);
		}
	      else
		{
		  parts = g_strsplit_set (device, ":,", -1);
		  if (parts[0] && strcmp(parts[0],"hw"))
		    {
		      g_message ("Invalid identifier '%s", parts[0]);
		    }
		  else if (parts[0] && !strcmp(parts[0],"hw"))
		    {
		      gint card, device;
  
		      card = atoi (parts[1]) + 1;
		      device = atoi (parts[2]);

		      gtk_combo_box_set_active (GTK_COMBO_BOX(widget), card);
		      widget = glade_xml_get_widget_warn (this->private->xml, "cbox_alsa_device");
		      gtk_combo_box_set_active (GTK_COMBO_BOX(widget), device);
		    }
		  g_strfreev (parts);
		}
	  }
#endif

	//Set device in case sink is ossink
	if (!strcmp(sink,"osssink"))
	  {
	      gchar  *device;

	      widget = glade_xml_get_widget_warn (this->private->xml, "cbe_oss_device");
	      device = xcs_key_fetch_string (xcs, "audio", "device");
    
	      gtk_entry_set_text (GTK_ENTRY(GTK_BIN(widget)->child), device);
	  }

      xcs_bind_entry (xcs,
		     GTK_ENTRY(GTK_BIN(glade_xml_get_widget_warn (this->private->xml,"cbe_oss_device"))->child),
		     "audio",
		     "device"); 

      xcs_bind_entry (xcs,
		     GTK_ENTRY(GTK_BIN(glade_xml_get_widget_warn (this->private->xml,"cbe_sun_device"))->child),
		     "audio",
		     "device"); 

#ifdef HAVE_ALSA
      xcs_bind_spin_button (xcs,
		     GTK_SPIN_BUTTON(glade_xml_get_widget_warn (this->private->xml,"alsa_buffer_time")),
		     "audio",
		     "alsa-buffer-time"); 
#endif

      xcs_bind_spin_button (xcs,
		     GTK_SPIN_BUTTON(glade_xml_get_widget_warn (this->private->xml,"oss_buffer_time")),
		     "audio",
		     "oss-buffer-time"); 

#ifdef HAVE_SUNAUDIO
      xcs_bind_spin_button (xcs,
		     GTK_SPIN_BUTTON(glade_xml_get_widget_warn (this->private->xml,"sun_buffer_time")),
		     "audio",
		     "sun-buffer-time"); 
#endif

      widget = glade_xml_get_widget_warn (this->private->xml, "cbe_oss_device");
      signal_handler_pool_add_connection (this->private->pool,
					  widget,	
					  "changed",
					  G_CALLBACK(set_apply_sensitive),
					  this,	
				          0);	

#ifdef HAVE_SUNAUDIO
      widget = glade_xml_get_widget_warn (this->private->xml, "cbe_sun_device");
      signal_handler_pool_add_connection (this->private->pool,
					  widget,	
					  "changed",
					  G_CALLBACK(set_apply_sensitive),
					  this,	
				          0);	
#endif

#ifdef HAVE_ALSA
      widget = glade_xml_get_widget_warn (this->private->xml, "alsa_buffer_time");
      signal_handler_pool_add_connection (this->private->pool,
					  widget,	
					  "value-changed",
					  G_CALLBACK(set_apply_sensitive),
					  this,	
				          0);	
#endif

      widget = glade_xml_get_widget_warn (this->private->xml, "oss_buffer_time");
      signal_handler_pool_add_connection (this->private->pool,
					  widget,	
					  "value-changed",
					  G_CALLBACK(set_apply_sensitive),
					  this,	
				          0);	

#ifdef HAVE_SUNAUDIO
      widget = glade_xml_get_widget_warn (this->private->xml, "sun_buffer_time");
      signal_handler_pool_add_connection (this->private->pool,
					  widget,	
					  "value-changed",
					  G_CALLBACK(set_apply_sensitive),
					  this,	
				          0);	
#endif

      widget = glade_xml_get_widget_warn (this->private->xml, "cbox_audio_system");
      signal_handler_pool_add_connection (this->private->pool,
					  widget,	
					  "changed",
					  G_CALLBACK(set_apply_sensitive),
					  this,	
				          0);	

#ifdef HAVE_ALSA
      widget = glade_xml_get_widget_warn (this->private->xml, "cbox_alsa_card");
      signal_handler_pool_add_connection (this->private->pool,
					  widget,	
					  "changed",
					  G_CALLBACK(set_apply_sensitive),
					  this,	
				          0);	

      widget = glade_xml_get_widget_warn (this->private->xml, "cbox_alsa_device");
      signal_handler_pool_add_connection (this->private->pool,
					  widget,	
					  "changed",
					  G_CALLBACK(set_apply_sensitive),
					  this,	
				          0);	
#endif

      widget = glade_xml_get_widget_warn (this->private->xml, "b_audio_system_apply");
      signal_handler_pool_add_connection (this->private->pool,
					  widget,	
					  "clicked",
					  G_CALLBACK(on_b_audio_system_apply),
					  this,	
				          0);	

      //Setup support status indicators
      {
        const gchar *stock_id;

	if (g_object_get_boolean (bmp_play_engine, "has_cdda"))
	  stock_id = GTK_STOCK_YES;
	else
	  stock_id = GTK_STOCK_NO;

	gtk_image_set_from_stock (GTK_IMAGE(glade_xml_get_widget_warn (this->private->xml, "img_status_cdda")),
				  stock_id,
				  GTK_ICON_SIZE_MENU);

	if (g_object_get_boolean (bmp_play_engine, "has_http"))
	  stock_id = GTK_STOCK_YES;
	else
	  stock_id = GTK_STOCK_NO;

	gtk_image_set_from_stock (GTK_IMAGE(glade_xml_get_widget_warn (this->private->xml, "img_status_http")),
				  stock_id,
				  GTK_ICON_SIZE_MENU);
      }

    }

    gtk_widget_realize (this->window);

    if (xcs_key_fetch_boolean (xcs, "bmp", "display-tooltips"))
	bmp_tooltips_enable (tooltips);
    else
	bmp_tooltips_disable (tooltips);

    gtk_window_set_position (GTK_WINDOW (this->window), GTK_WIN_POS_CENTER);
}

void
bmp_preferences_show (BmpPreferences *this)
{
    gtk_window_present (GTK_WINDOW(this->window));

    if (!this->private->updated)
    {
        this->private->updated = TRUE;
        skin_view_update(this);
	gtk_widget_show_all(this->private->skin_view);

	this->private->skins_path = g_build_filename(DATA_DIR,G_DIR_SEPARATOR_S,BMP_SKIN_DIR_BASENAME,NULL);

	fam_monitor_watch_new (bmp_fam_monitor,
				"skins",
				this->private->skins_path,
				(FamAddedFunc)skin_added_func,
				(FamRemovedFunc)skin_removed_func,
				this,
				NULL);

	fam_monitor_watch_new (bmp_fam_monitor,
				"user skins",
				BMP_PATH_USER_SKIN_DIR,
				(FamAddedFunc)skin_added_func,
				(FamRemovedFunc)skin_removed_func,
				this,
				NULL);

    }
}
