/*  BMPx - The Dumb Music Player
 *  Copyright (C) 2005 BMPx development team.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * The BMPx project hereby grant permission for non-gpl compatible GStreamer
 * plugins to be used and distributed together with GStreamer and BMPx. This
 * permission are above and beyond the permissions granted by the GPL license
 * BMPx is covered by.
 */

#include <config.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <signal.h>
#include <getopt.h>
#include <pthread.h>
#include <sys/time.h>
#include <sys/select.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <glib.h>
#include <glib-object.h>
#include <glib/gstdio.h>

#ifdef HAVE_GUI
#  include <gtk/gtk.h>
#endif

#include <libxml/tree.h>
#include <libxml/parser.h>
#include <libxml/xmlreader.h>
#include <libxml/xpath.h>
#include <libxml/xpathInternals.h>

#include "xcs.h"
#include "xml.h"

#define XCS_CONFIG_BASENAME "config.xml"

G_DEFINE_TYPE (Xcs, xcs, G_TYPE_OBJECT)

typedef enum {
         XCS_CONFIGURATION_SIGNAL_CHANGED,
         XCS_CONFIGURATION_SIGNAL_REMOVED,
         XCS_CONFIGURATION_SIGNAL_ADDED,

         N_SIGNALS
} XcsSignals;

static guint signals[N_SIGNALS] = { 0, 0, 0 };

typedef enum {
    N_PROPS
} XcsProperties;

struct _XcsPrivate {
    gboolean dispose_has_run;

    xmlDocPtr xmlconfig;
};

static gboolean new_configuration = TRUE;

typedef struct {
    Xcs	    *xcs;
    GObject *object;
    BindingDataFunc func;    
    gchar *domain;
    gchar *key;
    gulong handler;
} GObjectBinding;

#ifdef HAVE_GUI
static void
toggle_action_binding (GObject *object, GValue *value)
{
  g_value_init (value, G_TYPE_BOOLEAN);
  g_value_set_boolean (value, gtk_toggle_action_get_active (GTK_TOGGLE_ACTION(object)));
}

static void
toggle_button_binding (GObject *object, GValue *value)
{
  g_value_init (value, G_TYPE_BOOLEAN);
  g_value_set_boolean (value, gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(object)));
}

static void
gtk_entry_binding     (GObject *object, GValue *value)
{
  const gchar *text;
  text = gtk_entry_get_text (GTK_ENTRY(object));
  g_value_init (value, G_TYPE_STRING);
  g_value_set_string (value, text);

}

static void
gtk_spin_button_binding (GObject *object, GValue *value)
{
  gdouble value_d; 
  gint value_i;

  value_d = gtk_spin_button_get_value (GTK_SPIN_BUTTON(object));
  value_i = (gint)value_d;

  g_value_init (value, G_TYPE_INT);
  g_value_set_int (value, value_i);

}
#endif

static void
binding_object_signal_handler_invoked (GObject *object, gpointer user_data)
{
    GObjectBinding *binding;
    GValue value = {0,};

    g_return_if_fail (G_IS_OBJECT(object));

    binding = (GObjectBinding*)user_data; 

    binding->func (object, &value);

    xcs_key_set (binding->xcs,
			       binding->domain,
			       binding->key,
			       &value);
			       
}

static void
object_binding_marshal (GClosure     *closure,
			GValue       *return_value,
			guint         n_param_values,
			const GValue *param_values,
			gpointer      invocation_hint,
			gpointer      marshal_data)
{
  typedef gboolean (*GMarshalBindingFunc) (gpointer     data1,
					   gpointer     data2);

  register GMarshalBindingFunc callback;
  register GCClosure *cc = (GCClosure*) closure;
  register gpointer data1, data2;

  data1 = g_value_peek_pointer (param_values + 0);
  data2 = closure->data;

  callback = (GMarshalBindingFunc) (cc->callback);

  callback (data1,
	    data2);
}

static void
object_destroyed (GObjectBinding *binding, GObject *no_object)
{
    g_free (binding->domain);
    g_free (binding->key);
}

void
xcs_bind_object (Xcs		      *configuration,
		 GObject	      *object, 
		 BindingDataFunc       func, 
		 const gchar	      *signal_spec, 
		 const gchar	      *domain, 
		 const gchar	      *key)
{
    GClosure *closure;
    GObjectBinding *binding;

    g_return_if_fail (XCS_IS_CONFIGURATION(configuration));

    binding = g_new0 (GObjectBinding, 1);

    binding->object = object;
    binding->func   = func;
    binding->domain = g_strdup (domain);
    binding->key    = g_strdup (key);
    binding->xcs    = configuration;

    /* FIXME: Create binding_object_destroy */ 
    closure = g_cclosure_new (G_CALLBACK(binding_object_signal_handler_invoked), binding, NULL);
    g_closure_set_marshal (closure, (GClosureMarshal)object_binding_marshal);
    binding->handler = g_signal_connect_closure ((gpointer)object, signal_spec, closure, TRUE);    

    g_object_weak_ref (object, (GWeakNotify)object_destroyed, binding); 
}

#ifdef HAVE_GUI
void
xcs_bind_toggle_button (Xcs		*configuration,
		        GtkToggleButton	*button,
			const gchar	*domain,
			const gchar	*key)
{
    gchar *sigdetail;

    g_return_if_fail (XCS_IS_CONFIGURATION(configuration));

    gtk_toggle_button_set_active (button, xcs_key_fetch_boolean (configuration, domain, key));

    sigdetail = g_strdup_printf ("%s-%s", domain, key);

    g_signal_emit (G_OBJECT (configuration),
                   signals[XCS_CONFIGURATION_SIGNAL_CHANGED],
                   g_quark_from_string (sigdetail), NULL);
    g_free (sigdetail);

    xcs_bind_object (configuration,
		     G_OBJECT(button),
		     toggle_button_binding,
		     "toggled",
		     domain, 
		     key); 
}

void
xcs_bind_toggle_action (Xcs		*configuration,
		        GtkToggleAction	*action,
			const gchar	*domain,
			const gchar	*key)
{
    gchar *sigdetail;

    g_return_if_fail (XCS_IS_CONFIGURATION(configuration));

    gtk_toggle_action_set_active (action, xcs_key_fetch_boolean (configuration, domain, key));

    sigdetail = g_strdup_printf ("%s-%s", domain, key);

    g_signal_emit (G_OBJECT (configuration),
                   signals[XCS_CONFIGURATION_SIGNAL_CHANGED],
                   g_quark_from_string (sigdetail), NULL);
    g_free (sigdetail);


    xcs_bind_object (configuration,
		     G_OBJECT(action),
		     toggle_action_binding,
		     "toggled",
		     domain, 
		     key); 
}

void
xcs_bind_entry	      (Xcs		*configuration,
		       GtkEntry		*entry,
		       const gchar	*domain,
		       const gchar	*key)
{
    gchar *text;

    g_return_if_fail (XCS_IS_CONFIGURATION(configuration));

    text = xcs_key_fetch_string (configuration, domain, key);
    if (text) gtk_entry_set_text (entry, text);
    g_free (text);

    xcs_bind_object (configuration,
		     G_OBJECT(entry),
		     gtk_entry_binding,
		     "changed",
		     domain, 
		     key); 
}

void
xcs_bind_spin_button  (Xcs		*configuration,
		       GtkSpinButton	*sbutton,
		       const gchar	*domain,
		       const gchar	*key)
{
    g_return_if_fail (XCS_IS_CONFIGURATION(configuration));

    gtk_spin_button_set_value (sbutton, (gdouble)xcs_key_fetch_int (configuration, domain, key));

    xcs_bind_object (configuration,
		     G_OBJECT(sbutton),
		     gtk_spin_button_binding,
		     "value-changed",
		      domain, 
		      key); 
}
#endif

gboolean
xcs_save (Xcs *configuration)
{
    gchar *filename;
    FILE *file_xml;

    filename =
        g_build_filename (g_get_home_dir(), XCS_RCPATH, XCS_CONFIG_BASENAME, NULL);
    file_xml = g_fopen (filename, "w+");
    g_free (filename);

    xmlDocFormatDump (file_xml,
                      configuration->priv->xmlconfig, 1);

    fclose (file_xml);

    return TRUE;
}

static GObject *xcs_constructor (GType type,
                                             guint n_construct_properties,
                                             GObjectConstructParam *
                                             construct_properties);

static void
xcs_init (Xcs *configuration)
{
    xmlNodePtr node;
    gchar *filename;

    configuration->priv = g_new (XcsPrivate, 1);

    configuration->priv->dispose_has_run = FALSE;

    filename = g_build_filename (g_get_home_dir(), XCS_RCPATH, NULL);

    /* FIXME: As not only the configuration is stored there, this shouldn't be the
       function to create the XCS_RCPATH directory */

    if (!g_file_test (filename, G_FILE_TEST_EXISTS))
        mkdir (filename, S_IRWXU);

    g_free (filename);

    filename = g_build_filename (g_get_home_dir(), XCS_RCPATH, "config.xml", NULL);

    if (g_file_test (filename, G_FILE_TEST_EXISTS))
      {
	xmlXPathObjectPtr xpathobj;
	xmlDocPtr xmldoc;

	xmldoc = xmlParseFile (filename);
	new_configuration = FALSE;

	xpathobj = xml_execute_xpath_expression (xmldoc, BAD_CAST ("/" XCS_ROOT_NODE_NAME), NULL); 
	
	if ((!xpathobj) || (!xpathobj->nodesetval) || (!xpathobj->nodesetval->nodeNr))
	  {
	    new_configuration = TRUE;
	    xmlFreeDoc (xmldoc);
	  }
	else
	  {
	      configuration->priv->xmlconfig = xmldoc;
	  }

	xmlXPathFreeObject (xpathobj);
      }

    if (new_configuration)
      {
         configuration->priv->xmlconfig = xmlNewDoc ((unsigned char*)"1.0");

         node = xmlNewNode (NULL, (unsigned char*)XCS_ROOT_NODE_NAME);
         xmlSetProp (node, (unsigned char*)"version", (unsigned char*)VERSION);

         xmlAddChild (node, xmlNewText ((unsigned char*)"\n"));

         xmlDocSetRootElement (configuration->priv->
                               xmlconfig, node);
      }

    g_free (filename);
}

static xmlXPathObjectPtr
xcs_get_from_xpath (Xcs *configuration,
		    const gchar * xpath_expr)
{
    xmlXPathObjectPtr xpathObj = NULL;

    g_return_val_if_fail (XCS_IS_CONFIGURATION (configuration), NULL);

    xpathObj =
        xml_execute_xpath_expression ((xmlDocPtr)configuration->priv->
                                         xmlconfig,
                                         (unsigned char*)xpath_expr, NULL);
    return xpathObj;
}

gboolean
xcs_domain_new (Xcs *configuration,
                            const gchar * domain_id)
{
    xmlNodePtr node;
    xmlNodeSetPtr nv;
    xmlXPathObjectPtr xpathObj;
    gchar *xpath_expr;

    g_return_val_if_fail (XCS_IS_CONFIGURATION (configuration), FALSE);
    g_return_val_if_fail ((domain_id != NULL), FALSE);

    if (xcs_domain_exists (configuration, domain_id))
      {
        return TRUE;
      }

    xpath_expr = 
      g_strdup_printf ("/%s",XCS_ROOT_NODE_NAME);
    xpathObj = 
      xcs_get_from_xpath (configuration, xpath_expr);
    if (!xpathObj) return FALSE;
    g_free (xpath_expr);
    nv = 
      xpathObj->nodesetval;

    if (!nv)
      {
	xmlXPathFreeObject (xpathObj);
        return FALSE;
      }

    if (!nv->nodeNr)
      {
	xmlXPathFreeObject (xpathObj);
        return FALSE;
      }

    node = xmlNewNode (NULL, (unsigned char*)"domain");

    xmlSetProp (node, (unsigned char*)"id", (unsigned char*)domain_id);
    xmlAddChild (nv->nodeTab[0], node);
    xmlAddChild (nv->nodeTab[0], xmlNewText ((unsigned char*)"\n\n"));
    xmlAddChild (node, xmlNewText ((unsigned char*)"\n"));

    xmlXPathFreeObject (xpathObj);
    xcs_save (configuration);


    return TRUE;
}

gboolean
xcs_key_new (Xcs *configuration,
                           const gchar	    *domain_id,
                           const gchar	    *key_id, 
			   GValue	    *key_value)
{
    xmlNodePtr node;
    xmlNodeSetPtr nv;
    xmlXPathObjectPtr xpathObj;
    gchar *xpath_expr;
    gchar *sigdetail;
    gchar *key_value_str = NULL;
    const gchar *type_id = NULL;

    g_return_val_if_fail (XCS_IS_CONFIGURATION (configuration), FALSE);
    g_return_val_if_fail (domain_id != NULL, FALSE);

    xpath_expr = g_strdup_printf ("/%s/domain[@id = '%s']", XCS_ROOT_NODE_NAME, domain_id); 
    xpathObj = xcs_get_from_xpath (configuration, xpath_expr);
    if (!xpathObj) return FALSE;
    g_free (xpath_expr);
    nv = xpathObj->nodesetval;

    if (!nv)
      {
	xmlXPathFreeObject (xpathObj);
        return FALSE;
      }

    if (!nv->nodeNr)
      {
	xmlXPathFreeObject (xpathObj);
        return FALSE;
      }

    if (G_VALUE_HOLDS_STRING (key_value))
    {
        key_value_str = g_value_dup_string (key_value);
	type_id = "string";
    }
    else
    if (G_VALUE_HOLDS_INT (key_value))
    {
        key_value_str = g_strdup_printf ("%d", g_value_get_int (key_value));
	type_id = "integer";
    }
    else
    if (G_VALUE_HOLDS_BOOLEAN (key_value))
    {
	key_value_str  = g_value_get_boolean (key_value) ? g_strdup("TRUE") : g_strdup("FALSE");
	type_id = "boolean";
    }
    else
    if (G_VALUE_HOLDS_DOUBLE (key_value))
    {
	key_value_str = g_strdup_printf ("%lf", g_value_get_double (key_value));
	type_id = "float";
    }

    node = xmlNewTextChild (nv->nodeTab[0], NULL, (unsigned char*)"key", (unsigned char*)key_value_str);
    g_free (key_value_str);

    xmlAddSibling (node, xmlNewText ((unsigned char*)"\n"));

    if (key_id)
     {
         xmlSetProp (node, (unsigned char*)"id", (unsigned char*)key_id);
         xmlSetProp (node, (unsigned char*)"type", (unsigned char*)type_id);
         sigdetail = g_strdup_printf ("%s-%s", domain_id, key_id);
     }
    else
     {
         sigdetail = g_strdup (domain_id);
     }

    g_signal_emit (G_OBJECT (configuration),
                   signals[XCS_CONFIGURATION_SIGNAL_ADDED],
                   g_quark_from_string (sigdetail), NULL);
    g_free (sigdetail);

    xmlXPathFreeObject (xpathObj);
    xcs_save (configuration);
    return TRUE;
}

/* This fetches all keys of a given domain */
GHashTable *
xcs_domain_fetch (Xcs *configuration,
                                         const gchar * domain_id)
{
    GHashTable *ht;
    xmlNodeSetPtr nv;
    xmlNodePtr node;
    xmlXPathObjectPtr xpathObj;
    gchar *xpath_expr, *key_value, *key_id;
    gint n;

    g_return_val_if_fail (XCS_IS_CONFIGURATION (configuration), NULL);

    xpath_expr =
        g_strdup_printf ("/%s/domain[@id = '%s']/key", XCS_ROOT_NODE_NAME, domain_id);
    xpathObj = 
	xcs_get_from_xpath (configuration, xpath_expr);
    if (!xpathObj) return FALSE;
    g_free (xpath_expr);
    nv = 
	xpathObj->nodesetval;	

    if (!nv)
      {
	xmlXPathFreeObject (xpathObj);
        return NULL;
      }

    if (!nv->nodeNr)
      {
	xmlXPathFreeObject (xpathObj);
        return NULL;
      }

    ht = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);

    for (n = 0; n < nv->nodeNr; ++n)
     {
         node = nv->nodeTab[n];

         if (!node->children)
             continue;

         key_value = (char*)XML_GET_CONTENT (node->children);

         /* _replace() will overwrite keys with the same id; but since
          * it's an id, it should be duplicate anyhow. Still i think
          */

         if (!xmlGetProp (node, (unsigned char*)"id"))
          {
              key_id = g_strdup_printf ("%d", n);
          }
         else
          {
              key_id = g_strdup ((char*)xmlGetProp (node, (unsigned char*)"id"));
          }

         g_hash_table_replace (ht, key_id, g_strdup (key_value));
     }

    xmlXPathFreeObject (xpathObj);
    return ht;
}

void
xcs_ht_destroy (Xcs *configuration,
                              GHashTable *ht)
{
    g_return_if_fail (XCS_IS_CONFIGURATION (configuration));
    g_hash_table_destroy (ht);
}

gboolean
xcs_key_unset (Xcs *configuration,
                         const gchar * domain_id,
                         const gchar * key_id,
                         gboolean create)
{
    xmlNodeSetPtr nv;
    xmlNodePtr node_cur;
    xmlXPathObjectPtr xpathObj;
    gchar *xpath_expr;

    g_return_val_if_fail (XCS_IS_CONFIGURATION (configuration), FALSE);

    xpath_expr =
        g_strdup_printf ("/%s/domain[@id = '%s']/key[@id = '%s']", XCS_ROOT_NODE_NAME, domain_id, key_id);
    xpathObj =
        xcs_get_from_xpath (configuration, xpath_expr);
    if (!xpathObj) return FALSE;
    g_free (xpath_expr);
    nv =
	xpathObj->nodesetval;  

    if (!nv)
      {
	xmlXPathFreeObject (xpathObj);
        return FALSE;
      }


    if (!nv->nodeNr)
    {
	xmlXPathFreeObject (xpathObj);
        return xcs_key_new (configuration,
                                          domain_id,
                                          key_id,
                                          NULL);
    }

    node_cur =
        nv->nodeTab[0];

    if (node_cur->children)
    {
         xmlUnlinkNode (node_cur->children);
         xmlFreeNode (node_cur->children);
    }


    xmlXPathFreeObject (xpathObj);
    xcs_save (configuration);
    return TRUE;
}

gboolean
xcs_key_set (Xcs	    *configuration,
             const gchar    *domain_id,
             const gchar    *key_id,
             GValue	    *key_value) 

{
    xmlNodeSetPtr nv;
    xmlNodePtr node_cur;
    xmlNodePtr node_new;
    xmlXPathObjectPtr xpathObj;
    gchar *xpath_expr, *key_value_str = NULL, *sigdetail;
    const gchar *type_id = NULL;
    gboolean create = FALSE;

    g_return_val_if_fail (XCS_IS_CONFIGURATION (configuration), FALSE);

    if (!xcs_domain_exists (configuration, domain_id))
      {
  	 xcs_domain_new (configuration, domain_id);
      }

    xpath_expr = g_strdup_printf ("/%s/domain[@id = '%s']/key[@id = '%s']", XCS_ROOT_NODE_NAME, domain_id, key_id);
    xpathObj = xcs_get_from_xpath (configuration, xpath_expr);	

    if (!xpathObj)
      {
	create = TRUE;
      }
    else
      {
	nv = xpathObj->nodesetval;
	if (!nv) 
	  {
	      create = TRUE;
	  }
	else
	  {
	    if (!nv->nodeNr) create = TRUE;
	  }
      }

    g_free (xpath_expr);

    if (create)
      {
	      xmlXPathFreeObject (xpathObj);
              return xcs_key_new (configuration, domain_id, key_id, key_value);
      }

    if (G_VALUE_HOLDS_STRING (key_value))
     {
        key_value_str =
             g_value_dup_string (key_value);
	type_id = "string";
     }
    else
    if (G_VALUE_HOLDS_INT (key_value))
     {
        key_value_str =
             g_strdup_printf ("%d", g_value_get_int (key_value));
	type_id = "integer";
     }
     else
     if (G_VALUE_HOLDS_BOOLEAN (key_value))
     {
        key_value_str =
             g_strdup_printf ("%s", g_value_get_boolean (key_value) ? "TRUE" : "FALSE");
	type_id = "boolean";
     }
    else
    if (G_VALUE_HOLDS_DOUBLE (key_value))
     {
        key_value_str =
             g_strdup_printf ("%lf", g_value_get_double (key_value));
	type_id = "float";
     }

    node_cur =
        nv->nodeTab[0];
    node_new =
        xmlNewText ((unsigned char*)key_value_str);

    if (node_cur->children)
         xmlReplaceNode (node_cur->children, node_new);
    else
         xmlAddChild (node_cur, node_new);

    xmlSetProp(node_cur,(unsigned char*)"type",(unsigned char*)type_id);

    sigdetail = g_strdup_printf ("%s-%s", domain_id,key_id);

    g_signal_emit (G_OBJECT (configuration),
                   signals[XCS_CONFIGURATION_SIGNAL_CHANGED],
                   g_quark_from_string (sigdetail), NULL);

    g_free (sigdetail);
    g_free (key_value_str);
    xmlXPathFreeObject (xpathObj);
    xcs_save (configuration);

    return TRUE;
}

gdouble
xcs_key_fetch_float  (Xcs *configuration,
                      const gchar * domain_id,
                      const gchar * key_id)
{
    xmlNodeSetPtr nv;
    xmlXPathObjectPtr xpathObj;
    gchar *xpath_expr, *aux;
    gdouble val = 0; 

    g_return_val_if_fail (XCS_IS_CONFIGURATION (configuration), -1);

    xpath_expr = g_strdup_printf ("/%s/domain[@id = '%s']/key[@id = '%s']", XCS_ROOT_NODE_NAME, domain_id, key_id);
    xpathObj = xcs_get_from_xpath (configuration, xpath_expr);
	
    g_free (xpath_expr);
    if (!xpathObj) return -1;
    nv = xpathObj->nodesetval;	

    if ((!nv) || (!nv->nodeNr))
      {
	xmlXPathFreeObject (xpathObj);
        return -1;
      }

    if (nv->nodeTab[0]->children)
    {
       aux =
            (char*)XML_GET_CONTENT((xmlNodePtr) (nv->nodeTab[0])->children);
       val =
	    strtod (aux, NULL);
    }

    xmlXPathFreeObject (xpathObj);

    return val;
}

gint
xcs_key_fetch_int (Xcs *configuration,
                                 const gchar * domain_id,
                                 const gchar * key_id)
{
    xmlNodeSetPtr nv;
    xmlXPathObjectPtr xpathObj;
    gchar *xpath_expr, *aux;
    gint val = -1;

    g_return_val_if_fail (XCS_IS_CONFIGURATION (configuration), -1);

    xpath_expr =
        g_strdup_printf ("/%s/domain[@id = '%s']/key[@id = '%s']", XCS_ROOT_NODE_NAME, domain_id, key_id);
    xpathObj = 
	xcs_get_from_xpath (configuration, xpath_expr);
    g_free (xpath_expr);
    if (!xpathObj) return -1;
    nv = 
	xpathObj->nodesetval;	

    if ((!nv) || (!nv->nodeNr))
      {
	xmlXPathFreeObject (xpathObj);
        return -1;
      }

    if (nv->nodeTab[0]->children)
    {
       aux =
            (char*)XML_GET_CONTENT((xmlNodePtr) (nv->nodeTab[0])->children);
       val =
	    strtol(aux,NULL,10);
    }

    xmlXPathFreeObject (xpathObj);
    return val;
}

gboolean
xcs_key_fetch_boolean (Xcs *configuration,
                                 const gchar * domain_id,
                                 const gchar * key_id)
{
    xmlNodeSetPtr nv;
    xmlXPathObjectPtr xpathObj;
    gchar *xpath_expr, *aux;
    gboolean val = FALSE;

    g_return_val_if_fail (XCS_IS_CONFIGURATION (configuration), FALSE);

    xpath_expr =
        g_strdup_printf ("/%s/domain[@id = '%s']/key[@id = '%s']", XCS_ROOT_NODE_NAME, domain_id, key_id);
    xpathObj = 
	xcs_get_from_xpath (configuration, xpath_expr);
    g_free (xpath_expr);
    if (!xpathObj) return FALSE;
    nv = 
	xpathObj->nodesetval;	

    if ((!nv) || (!nv->nodeNr))
      {
	xmlXPathFreeObject (xpathObj);
        return FALSE;
      }

    if (nv->nodeTab[0]->children)
    {
	aux = g_strdup((char*)XML_GET_CONTENT((xmlNodePtr)(nv->nodeTab[0]->children)));
	g_strstrip(aux);
	if (g_ascii_strcasecmp(aux, "TRUE") == 0)
		val = TRUE;
	if (g_ascii_strcasecmp(aux, "1") == 0)
		val = TRUE;
	g_free (aux);
    }

    xmlXPathFreeObject (xpathObj);
    return val;
}

gchar*
xcs_key_fetch_string            (Xcs *configuration,
                                 const gchar * domain_id,
                                 const gchar * key_id)
{
    xmlNodeSetPtr nv;
    xmlXPathObjectPtr xpathObj;
    gchar *xpath_expr;

    g_return_val_if_fail (XCS_IS_CONFIGURATION (configuration), NULL);

    xpath_expr =
        g_strdup_printf ("/%s/domain[@id = '%s']/key[@id = '%s']", XCS_ROOT_NODE_NAME, domain_id, key_id);
    xpathObj = 
	xcs_get_from_xpath (configuration, xpath_expr);
    if (!xpathObj) return NULL;
    g_free (xpath_expr);
    nv = 
	xpathObj->nodesetval;	

    if ((!nv) || (!nv->nodeNr))
      {
	xmlXPathFreeObject (xpathObj);
        return NULL;
      }

    if (nv->nodeTab[0]->children)
      {
	gchar *rv;
        rv =
            g_strdup ((char*)XML_GET_CONTENT
                      ((xmlNodePtr) (nv->nodeTab[0])->children));
	xmlXPathFreeObject (xpathObj);
	return rv;
      } 
    else
      {
	xmlXPathFreeObject (xpathObj);
        return NULL;
      }

    xmlXPathFreeObject (xpathObj);
    return NULL;
}

gboolean
xcs_domain_exists (Xcs *configuration,
                               const gchar * domain_id)
{
    xmlNodeSetPtr nv;
    xmlXPathObjectPtr xpathObj;
    gchar *xpath_expr;

    xpath_expr = 
	g_strdup_printf ("/%s/domain[@id = '%s']", XCS_ROOT_NODE_NAME, domain_id);
    xpathObj = 
	xcs_get_from_xpath (configuration, xpath_expr);
    g_free (xpath_expr);

    if (!xpathObj)
        return FALSE;

    nv = 
	xpathObj->nodesetval;

    if (!nv)
      {
	xmlXPathFreeObject (xpathObj);
        return FALSE;
      }

    if (!nv->nodeNr)
      {
	xmlXPathFreeObject (xpathObj);
        return FALSE;
      }

    xmlXPathFreeObject (xpathObj);
    return TRUE;
}

gboolean
xcs_domain_has_keys (Xcs *configuration,
                                 const gchar * domain_id)
{
    xmlNodeSetPtr nv;
    xmlXPathObjectPtr xpathObj;
    gchar *xpath_expr;
    gboolean key_found = FALSE;

    xpath_expr =
        g_strdup_printf ("/%s/domain[@id = '%s']/key", XCS_ROOT_NODE_NAME, domain_id);
    xpathObj = 
	xcs_get_from_xpath (configuration, xpath_expr);
    g_free (xpath_expr);
    if (!xpathObj)
      return FALSE;
    nv = 
	xpathObj->nodesetval;	

    if (!nv)
        return FALSE;

    if (!nv->nodeNr)
        key_found = FALSE;
    else
        key_found = TRUE;

    xmlXPathFreeObject (xpathObj);
    return key_found;
}

gboolean
xcs_key_find (Xcs *configuration,
                          const gchar * domain_id,
                          const gchar * key_id, const gchar * key_value)
{
    xmlNodePtr node;
    xmlNodeSetPtr nv;
    xmlXPathObjectPtr xpathObj;
    gchar *xpath_expr;
    gchar *key_value_cmp = NULL;
    gint n;

    gboolean key_found = FALSE;

    g_return_val_if_fail (XCS_IS_CONFIGURATION (configuration), FALSE);

    if (key_id)
     {
         xpath_expr =
             g_strdup_printf ("/%s/domain[@id = '%s']/key[@id = '%s']", XCS_ROOT_NODE_NAME, domain_id, key_id);
         xpathObj = 
	      xcs_get_from_xpath (configuration, xpath_expr);
         if (!xpathObj)
	    return FALSE;
         g_free (xpath_expr);
         nv = 
	     xpathObj->nodesetval; 

	 key_found = !!nv;
    
	 xmlXPathFreeObject (xpathObj);
         return key_found;
     }
    else if (key_value)
     {
         xpath_expr =
             g_strdup_printf ("/%s/domain[@id = '%s']/key", XCS_ROOT_NODE_NAME, domain_id);
         xpathObj = 
	     xcs_get_from_xpath (configuration, xpath_expr);
         if (!xpathObj)
	    return FALSE;
         g_free (xpath_expr);
         nv = 
	     xpathObj->nodesetval; 

         if ((!nv) || (!nv->nodeNr))
	   {
	     xmlXPathFreeObject (xpathObj);
             return FALSE;
	   }

         for (n = 0; n < nv->nodeNr; ++n)
          {
              node = nv->nodeTab[n];

	      if (!key_value_cmp) continue;

              key_value_cmp = (char*)XML_GET_CONTENT (node->children);

              if (strcmp (key_value_cmp, key_value) == 0)
               {
                   key_found = TRUE;
                   break;
               }
          }
	 xmlXPathFreeObject (xpathObj);
         return key_found;
     }

    g_assert_not_reached ();
}

gboolean
xcs_remove_domain (Xcs *configuration,
                               const gchar * domain_id)
{
    xmlNodeSetPtr nv;
    xmlXPathObjectPtr xpathObj;
    gchar *xpath_expr, *sigdetail;

    g_return_val_if_fail (XCS_IS_CONFIGURATION (configuration), FALSE);
    g_return_val_if_fail ((domain_id != NULL), FALSE);

    xpath_expr = 
	g_strdup_printf ("/%s/domain[@id = '%s']", XCS_ROOT_NODE_NAME, domain_id);
    xpathObj = 
	xcs_get_from_xpath (configuration, xpath_expr);
    if (!xpathObj)
      return FALSE;
    g_free (xpath_expr);
    nv = 
	xpathObj->nodesetval;	

    if ((!nv) || (!nv->nodeNr))
      {
	xmlXPathFreeObject (xpathObj);
        return FALSE;
      } 

    xmlUnlinkNode (nv->nodeTab[0]);
    xmlFreeNode (nv->nodeTab[0]);

    sigdetail = g_strdup_printf ("%s", domain_id);
    g_signal_emit (G_OBJECT (configuration),
                   signals[XCS_CONFIGURATION_SIGNAL_REMOVED],
                   g_quark_from_string (sigdetail), NULL);
    g_free (sigdetail);

    xmlXPathFreeObject (xpathObj);
    xcs_save (configuration);
    return TRUE;
}

gboolean
xcs_remove_key (Xcs *configuration,
                            const gchar * domain_id,
                            const gchar * key_id, const gchar * key_value)
{
    xmlNodePtr node;
    xmlNodeSetPtr nv;
    xmlXPathObjectPtr xpathObj = NULL;

    gchar *xpath_expr, *key_value_cmp, *sigdetail;
    gint n;

    g_return_val_if_fail (XCS_IS_CONFIGURATION (configuration), FALSE);
    g_return_val_if_fail ((domain_id != NULL), FALSE);
    g_return_val_if_fail ((key_value != NULL), FALSE);

    if (key_id)
     {
         xpath_expr =
             g_strdup_printf ("/%s/domain[@id = '%s']/key[@id = '%s']", XCS_ROOT_NODE_NAME, domain_id, key_id);
         xpathObj = 
	     xcs_get_from_xpath (configuration, xpath_expr);
         if (!xpathObj)
            return FALSE;
         g_free (xpath_expr);
         nv = 
	     xpathObj->nodesetval; 

         if ((!nv) || (!nv->nodeNr))
	   {
	     xmlXPathFreeObject (xpathObj);
             return FALSE;
	   }

         xmlUnlinkNode (nv->nodeTab[0]);
         xmlFreeNode (nv->nodeTab[0]);

         sigdetail = g_strdup_printf ("%s-%s", domain_id, key_id);
         g_signal_emit (G_OBJECT (configuration),
                        signals[XCS_CONFIGURATION_SIGNAL_REMOVED],
                        g_quark_from_string (sigdetail), NULL);
         g_free (sigdetail);
	 xmlXPathFreeObject (xpathObj);
     }
    else if (key_value)
     {
         xpath_expr =
             g_strdup_printf ("/%s/domain[@id = '%s']/key", XCS_ROOT_NODE_NAME, domain_id);
         xpathObj = 
	     xcs_get_from_xpath (configuration, xpath_expr);
         if (!xpathObj)
            return FALSE;
         g_free (xpath_expr);
         nv = 
	     xpathObj->nodesetval; 

         if ((!nv) || (!nv->nodeNr))
	   {
	     xmlXPathFreeObject (xpathObj);
             return FALSE;
	   }

         /*
          * Unlink node here
          */
         for (n = 0; n < nv->nodeNr; ++n)
          {
              node = nv->nodeTab[n];

              key_value_cmp = (char*)XML_GET_CONTENT (node->children);

	      if (!key_value_cmp) continue;

              if (strcmp (key_value_cmp, key_value) == 0)
               {
                   xmlUnlinkNode (node);
                   xmlFreeNode (node);
               }
          }
         sigdetail = g_strdup_printf ("%s__key", domain_id);
         g_signal_emit (G_OBJECT (configuration),
                        signals[XCS_CONFIGURATION_SIGNAL_REMOVED],
                        g_quark_from_string (sigdetail), NULL);
         g_free (sigdetail);
     }

    if (xpathObj)
      xmlXPathFreeObject (xpathObj);
    xcs_save (configuration);
    return TRUE;
}


static void
xcs_set_property (GObject * object,
                              guint property_id, const GValue * value,
                              GParamSpec * pspec)
{
    switch (property_id)
     {
     }
}

static void
xcs_get_property (GObject * object,
                              guint property_id, GValue * value,
                              GParamSpec * pspec)
{
    switch (property_id)
     {
     }
}

static void
xcs_dispose (GObject * obj)
{
    Xcs *configuration = (Xcs *) obj;

    g_return_if_fail (!configuration->priv->dispose_has_run);

    configuration->priv->dispose_has_run = TRUE;
}

static void
xcs_finalize (GObject * obj)
{
    Xcs *configuration = (Xcs *) obj;

    g_free (configuration->priv);
}


Xcs*
xcs_new (gboolean *_new_configuration)
{
    Xcs *configuration;

    configuration = 
      g_object_new (XCS_TYPE_CONFIGURATION, NULL);

    *_new_configuration = new_configuration;

    return configuration;
}

static GObject *
xcs_constructor (GType type,
                             guint n_construct_properties,
                             GObjectConstructParam * construct_properties)
{
    GObject *obj;

    {
        XcsClass *klass;
        GObjectClass *parent_class;
        klass =
            XCS_CONFIGURATION_CLASS (g_type_class_peek
                                        (XCS_TYPE_CONFIGURATION));
        parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
        obj =
            parent_class->constructor (type,
                                       n_construct_properties,
                                       construct_properties);
    }
    return obj;
}

static void
xcs_class_init (XcsClass * klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

    gobject_class->set_property = xcs_set_property;
    gobject_class->get_property = xcs_get_property;
    gobject_class->dispose = xcs_dispose;
    gobject_class->finalize = xcs_finalize;
    gobject_class->constructor = xcs_constructor;

    signals[XCS_CONFIGURATION_SIGNAL_REMOVED] =
        g_signal_new ("removed",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
                      G_STRUCT_OFFSET (XcsClass, removed),
                      NULL, NULL, g_cclosure_marshal_VOID__VOID, G_TYPE_NONE,
                      0);

    signals[XCS_CONFIGURATION_SIGNAL_CHANGED] =
        g_signal_new ("changed",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
                      G_STRUCT_OFFSET (XcsClass, changed),
                      NULL, NULL, g_cclosure_marshal_VOID__VOID, G_TYPE_NONE,
                      0);


    signals[XCS_CONFIGURATION_SIGNAL_ADDED] =
        g_signal_new ("added",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
                      G_STRUCT_OFFSET (XcsClass, added),
                      NULL, NULL, g_cclosure_marshal_VOID__VOID, G_TYPE_NONE,
                      0);

}

void
xcs_key_set_boolean (Xcs *configuration,
                     const gchar * domain_id,
                     const gchar * key_id,
                     gboolean value_arg)
{
    GValue value = {0,};

    g_value_init (&value, G_TYPE_BOOLEAN);
    g_value_set_boolean (&value, value_arg);

    xcs_key_set (configuration, domain_id, key_id, &value);
}

void
xcs_key_set_int (Xcs *configuration,
                 const gchar * domain_id,
                 const gchar * key_id,
                 gint value_arg)
{
    GValue value = {0,};

    g_value_init (&value, G_TYPE_INT);
    g_value_set_int (&value, value_arg);

    xcs_key_set (configuration, domain_id, key_id, &value);
}

void
xcs_key_set_float  (Xcs		*configuration,
                    const gchar *domain_id,
		    const gchar *key_id,
		    gdouble	 value_arg)
{
    GValue value = {0,};

    g_value_init (&value, G_TYPE_DOUBLE);
    g_value_set_double (&value, value_arg);

    xcs_key_set (configuration, domain_id, key_id, &value);
}

void
xcs_key_set_string (Xcs *configuration,
                    const gchar * domain_id,
                    const gchar * key_id,
                    const gchar *value_arg)
{
    GValue value = {0,};

    g_value_init (&value, G_TYPE_STRING);
    g_value_set_string (&value, value_arg);

    xcs_key_set (configuration, domain_id, key_id, &value);
}
